///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      ABCACHE.CPP      
//
//  Description
//      Implementation of object property cache
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#include "abpab.h"
#include "abcache.h"
#include "list.inl"
#include "abdb.h"


///////////////////////////////////////////////////////////////////////////////
//
//    operator ==
//
//    Parameters                        
//          lhs         reference to an SPropValue, left hand side     
//          rhs         reference to an SPropValue, right hand side
//    Purpose
//          Tests for equality of two SPropValues based on the property
//          tag. This is useful for the list's Insert/Delete which need
//          to have == overloaded for all nodes' data types
//      
//    Return Value
//          TRUE if the proptags are the same, FALSE otherwise

inline BOOL operator == (const SPropValue & lhs, const SPropValue &rhs)
{
    return PROP_ID(lhs.ulPropTag) == PROP_ID(rhs.ulPropTag);

}


///////////////////////////////////////////////////////////////////////////////
//
//    CPropCache()
//
//    Parameters                        
//          pDB         pointer to a database object
//
//    Purpose
//          Constructor for CPropCache
//      
//      
//    Return Value
//          none.

CPropCache :: CPropCache(PDATABASE  pDB)
{
    m_pDB = pDB;
    m_pDB -> AddRef();
    m_bDirty = TRUE;
}
///////////////////////////////////////////////////////////////////////////////
//
//    Init()
//
//    Parameters                        
//          none.
//
//    Purpose
//        Creates the cache object
//      
//      
//    Return Value
//          none.

STDMETHODIMP CPropCache :: Init()
{
    m_pProps = new CFlatPropList;

    if (!m_pProps)
        return MAPI_E_NOT_ENOUGH_MEMORY;

    return S_OK;
} 
///////////////////////////////////////////////////////////////////////////////
//
//    ~CPropCache()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Destructor for CPropCache object
//      
//      
//    Return Value
//          none.

CPropCache :: ~CPropCache()
{
    if (m_pProps)
        delete m_pProps;

    if (m_pDB)
        m_pDB -> Release();
}


///////////////////////////////////////////////////////////////////////////////
//
//    Flush()
//
//    Parameters                        
//          RecNum          Reference to a record number for record to flush   
//  
//    Purpose
//          Writes RecNum's property stream to the database file,
//          overwriting any previous data. Sets the dirty flag to FALSE.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise

STDMETHODIMP CPropCache :: Flush(RECORD_NUMBER & RecNum)
{   
    HRESULT hRes = S_OK;
                                                                               
    TraceMessage("CPropCache::Flush");

    if (SUCCEEDED(hRes = Update(RecNum, fldPROPS, m_pProps)))
        m_bDirty = FALSE;
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Load()
//
//    Parameters                        
//          RecNum          Reference to a record number of the record whose
//                          properties we want to cache.   
//  
//    Purpose
//          Reads the propety stream for this record into the cache.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CPropCache :: Load(RECORD_NUMBER & RecNum)
{
    HRESULT         hRes     = S_OK;

    TraceMessage("CPropCache::Load");

    if (m_pProps)
        m_pProps -> Destroy();
 
    if (SUCCEEDED(hRes = Fetch(RecNum, fldPROPS, m_pProps)))        
        m_bDirty = FALSE;  
    
    if (!m_pProps -> Size())
        hRes = MAPI_E_NOT_FOUND;                        
 
    return hRes;

}

///////////////////////////////////////////////////////////////////////////////
//
//    UnFlatten()
//
//    Parameters                        
//          pfpv     - pointer to a FlatPropNode to unflatten
//          pParent  - pointer to a block of memory to link the
//                     newly allocated SPropValue.Value.XX to
//          spv      - reference to an SPropValue that contains
//                     the unflattened result          
//    Purpose
//          Takes a flattened SProp, unflattens it and returns an SPropValue          
//          that MAPI can handle.  We use MAPIAllocate/More because we only 
//          unflatten to pass back to MAPI (e.g.GetProps). The Prop must be 
//          allocated by the caller, we copy the FlatSProp's SPropValue into
//          it, then allocate additional memory if required (e.g. for Value.lpb)
//          and copy the flattened data into it. We then adjust Prop's Value.XX
//          pointer to point at the new allocation. The new allocation is made
//          with MAPIAllocateMore if pAllocParent and is non-null. 
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP UnFlatten(PFlatPropValue pfpv, LPVOID pParent,SPropValue & spv)
{
    HRESULT     hRes = S_OK;
    ULONG       ulDataLen = pfpv -> cbData;
    LPBYTE      pbTemp = NULL;
    SPropValue  spvTemp = pfpv -> spv;

    // ulDataLen is the size of the data only, e.g. the     
    // length of the *Value.LPSZ or *Value.bin.lpb.
    if (ulDataLen)
    {
        if (pParent)            
            hRes = gAllocMore(ulDataLen, pParent,(LPVOID *) & pbTemp);
        else
            hRes = gAllocBuff(ulDataLen,(LPVOID *) & pbTemp);
        
        if (!pbTemp)
        {
            hRes = MAPI_E_NOT_ENOUGH_MEMORY;
            goto Quit;
        }
        
        memcpy(pbTemp, pfpv -> Data, ulDataLen);

        // if AllocXX fails, temp will be NULL instead of garbage,
        // later we can FreeBuff the whole pParent
        switch (PROP_TYPE(spvTemp.ulPropTag))
        {
            case PT_STRING8:
                spvTemp.Value.lpszA = (LPSTR) pbTemp;
                break;
        
            case PT_UNICODE:
                spvTemp.Value.lpszW = (LPWSTR) pbTemp;
                break;

            case PT_CLSID:
                spvTemp.Value.lpguid = (LPGUID) pbTemp;
                break;

            case PT_BINARY:
                spvTemp.Value.bin.cb = ulDataLen;
                spvTemp.Value.bin.lpb = pbTemp;
                break;

            default:    // e.g. multi-valued properties not handled
                ASSERT(0);
        }
    }   
    
    spv = spvTemp;    

Quit:
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    UnFlatten()
//
//    Parameters                        
//          pFlatPropNode     - pointer to a FlatPropNode to unflatten
//          pAllocParent      - pointer to a block of memory to link the
//                              newly allocated SPropValue.Value.XX to
//          Prop              - reference to an SPropValue that contains
//                              the unflattened result          
//
//    Purpose
//          Takes a flattened prop node, unflattens it and returns an SPropValue          
//          that MAPI can handle.  Calls UnFlatten (the flat prop Value version)
//          to do the work. If pAllocParent is non-NULL, we use AllocateMore to
//          allocate memory for properties.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP UnFlatten(const PFlatPropNode   pFlatPropNode, 
                       LPVOID                pAllocParent,
                       SPropValue          & Prop)
{

    return UnFlatten(&pFlatPropNode -> fpv,pAllocParent, Prop);
}


///////////////////////////////////////////////////////////////////////////////
//
//    Flatten()
//
//    Parameters                        
//          pspv            - pointer to SPropValue to flatten
//          ppfsp           - pointer to newly allocated FlatSPropNode
//                            containing a flattened version of *pspv
//
//    Purpose
//          Converts an SPropValue into a FlatSPropNode we can insert
//          into our list of flattened SPropValues. Flattening means
//          allocating a contiguous buffer that holds the node (its
//          header data, next pointer, etc.), the proptag and Value,
//          and any additional data (e.g. for a string or PT_BINARY 
//          property), chasing down the Value.XX pointer and copying 
//          the data into the new buffer.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Flatten(const LPSPropValue pspv, PFlatPropNode * ppfsp)
{
    LPBYTE           ptr = NULL;
    ULONG            cnt = 0,
                     cbTotal = 0,       // including hdr data and padding
                     ulDLong = 0;

    PFlatPropNode    pFlatPropNode = NULL;

    *ppfsp = NULL;
               
    switch (PROP_TYPE(pspv -> ulPropTag))
    {
        case PT_STRING8:
            cnt = CbtszsizeA(pspv -> Value.lpszA);
            ptr = (LPBYTE) pspv -> Value.lpszA;
            break;
        case PT_UNICODE:
            cnt = CbtszsizeW(pspv -> Value.lpszW);
            ptr = (LPBYTE) pspv -> Value.lpszW;
            break;
        
        case PT_CLSID:
            cnt = sizeof GUID;
            ptr = (LPBYTE) pspv -> Value.lpguid;
            break;

        case PT_BINARY:
            cnt = pspv -> Value.bin.cb;
            ptr = pspv -> Value.bin.lpb;
            break;     

        default:
            if (ISMV_TAG(pspv -> ulPropTag))
                return MAPI_E_NO_SUPPORT;

            cnt = 0;        
            ptr = NULL;
    }
    
    // calculate size data + padding. should be multiple of ALIGNMENT
    cbTotal = CEIL(cnt,ALIGNMENT);

    // sizeof a FlatPropValue if data <= FPN+BYTES    
    if (cbTotal <= FPN_BYTES)
        cbTotal = sizeof FlatPropValue;
    else
        cbTotal += sizeof FlatPropValue - FPN_BYTES;
    
    // # of DWLONGs including pNext, ulPad
    ulDLong = (cbTotal + sizeof PFlatPropNode + sizeof ULONG)/sizeof DWORDLONG;

    // allocate a node with trailing space for data and padding
    // force the buffer onto an 8-byte boundary
    pFlatPropNode = (PFlatPropNode) new DWORDLONG[ulDLong];

    if (!pFlatPropNode)
        return MAPI_E_NOT_ENOUGH_MEMORY;

    pFlatPropNode -> fpv.spv     = *pspv;
    pFlatPropNode -> pNext       = NULL; 
    pFlatPropNode -> fpv.cbData  = cnt;     // size of actual data without padding
    pFlatPropNode -> fpv.cbTotal = cbTotal;

    if (ptr)
        memcpy(pFlatPropNode -> fpv.Data, ptr, cnt);

    *ppfsp = pFlatPropNode;

    return S_OK;
   
}

///////////////////////////////////////////////////////////////////////////////
//
//      CFlatPropList object.
//      
//      A linked list of an object's properties. Each SPropValue is 
//      serialized so that it can easily be read or written to a stream.
//      NOTE: a CFlatPropList object isn't inherently reentrant. It's intended
//      to be a member of an AB object and relies on its parent's 
//      synchronization mechanisms to be thread-safe, i.e. a list object is 
//      only manipulated inside its parent's critical section. 
//


///////////////////////////////////////////////////////////////////////////////
//
//    CFlatPropList()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class constructor.
//      
//    Return Value
//          none.
//

CFlatPropList :: CFlatPropList()
{
    m_pHead = NULL;
    m_ulCnt = 0;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Destroy()
//
//    Parameters                        
//          none. 
//
//    Purpose
//          Walks the list destroying each node and resetting the list to
//          'empty'
//      
//    Return Value
//          none.
//
   
VOID  WINAPI CFlatPropList :: Destroy()
{
    PFlatPropNode  pTemp = m_pHead;

    while (NULL != pTemp)
    {   
        m_pHead = m_pHead -> pNext;
        m_ulCnt--;
        delete pTemp;
        pTemp = m_pHead;    
    }    

    m_pHead = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CFlatPropList()
//
//    Parameters                        
//          none.
//
//    Purpose
//          class destructor
//      
//    Return Value
//          none.
//
CFlatPropList :: ~CFlatPropList()
{
    Destroy();
}


///////////////////////////////////////////////////////////////////////////////
//
//    InsertNode()
//
//    Parameters                        
//          pfpnInsertee    - pointer to a FlatPropNode to insert     
//
//    Purpose
//          Inserts an FlatPropNodw into the list WITHOUT copying it. The
//          caller is responsible for allocating memory for the node. After
//          insertion, the node's memory belongs to the list, the caller 
//          shouldn't modify or free it.The insert replaces any existing node 
//          having the same prop ID, or appends to the list end if not already
//          in the list.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

void WINAPI CFlatPropList :: InsertNode(PFlatPropNode pfpnInsertee)
{
    PFlatPropNode   pfpnTemp1 = m_pHead,
                    pfpnTemp2;
    
    pfpnInsertee -> pNext = NULL;
    
    // empty list case
    if (!m_pHead)
    {   
        m_pHead = pfpnInsertee;
        m_ulCnt++;
        return;
    }

    // hit on head of list
    if (PROP_ID(pfpnTemp1 -> fpv.spv.ulPropTag) == 
        PROP_ID(pfpnInsertee -> fpv.spv.ulPropTag))
    {
        pfpnInsertee -> pNext = m_pHead -> pNext;
        m_pHead = pfpnInsertee;
        delete pfpnTemp1;
        return;
    }     

    pfpnTemp2 = m_pHead -> pNext;
    
    // scan the list        
    while (pfpnTemp2)
    {
        // hit in the middle
        if (PROP_ID(pfpnTemp2 -> fpv.spv.ulPropTag) == 
            PROP_ID(pfpnInsertee -> fpv.spv.ulPropTag))
        {
            pfpnInsertee -> pNext = pfpnTemp2 -> pNext;
            pfpnTemp1 -> pNext = pfpnInsertee;
            delete pfpnTemp2;
            return;
        }

        pfpnTemp1 = pfpnTemp2;
        pfpnTemp2 = pfpnTemp2 -> pNext;
    }

    // not in list, put at end    
    pfpnTemp1 -> pNext = pfpnInsertee;
    m_ulCnt++;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Insert()
//
//    Parameters                        
//          pSProp          - pointer to an SPropValue to insert     
//
//    Purpose
//          Inserts an SPropValue into the list by first flattening it into
//          a FlatSPropNode and inserting. The insert replaces any existing
//          node with the same prop tag, or appends to the list end if not
//          already in the list.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CFlatPropList :: Insert(const LPSPropValue pSProp)
{
    HRESULT         hRes;
    PFlatPropNode   pfpnTarget;
                  
    // create a flat prop node
    if (FAILED(hRes = Flatten(pSProp, & pfpnTarget)))
        return hRes;

    // initialize to new node    
    pfpnTarget -> pNext = NULL;
    InsertNode(pfpnTarget);

    return S_OK;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Insert()
//
//    Parameters                        
//          pSProp          - pointer to an FlatPropValue to insert     
//
//    Purpose
//          Inserts a flat prop node into the list by COPYING it into memory
//          owned by the list.The insert replaces any existing
//          node with the same prop ID, or appends to the list end if not
//          already in the list.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CFlatPropList :: Insert(const PFlatPropValue pfpvProp)
{
    HRESULT         hRes  = S_OK;
    PFlatPropNode   pfpnTarget;
    ULONG           ulDwords;
    
    ulDwords = (pfpvProp -> cbTotal + sizeof PFlatPropNode + sizeof ULONG) / sizeof DWORDLONG;
                  
    // allocate our memory for new flat prop node (on 8-byte boundary)
    pfpnTarget =  (PFlatPropNode) new DWORDLONG[ulDwords];

    if (!pfpnTarget)
        return MAPI_E_NOT_ENOUGH_MEMORY;
            
    // initialize to new node    
    memcpy((LPVOID) &pfpnTarget -> fpv,(LPVOID) pfpvProp, pfpvProp -> cbTotal);
    pfpnTarget -> pNext = NULL;
    InsertNode(pfpnTarget);
    return S_OK;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Delete()
//
//    Parameters                        
//          ulPropTag       - prop tag of node to delete     
//          ppRet           - pointer to pointer to flat prop node that
//                            will the deleted node
//    Purpose
//          Searches the list for a node with ulPropTag's property tag. If
//          found the node is deleted from the list and passed back in *ppRet.
//          If not found, NULL is passed back in *ppRet. 
//      
//    Return Value
//          FALSE on allocation error, TRUE otherwise.

STDMETHODIMP CFlatPropList :: Delete(const ULONG ulPropTag, PFlatPropNode * ppRet)
{
    PFlatPropNode  pfpnTemp1 = m_pHead,
                   pfpnTemp2;

    *ppRet = NULL;
        
    if (!pfpnTemp1)
        return S_OK;

    if (PROP_ID(pfpnTemp1 -> fpv.spv.ulPropTag) == PROP_ID(ulPropTag))
    {
        m_pHead = m_pHead -> pNext;
        *ppRet = pfpnTemp1;
        m_ulCnt--;
        return S_OK;
     }

    pfpnTemp2 = m_pHead -> pNext;

    while (pfpnTemp2)
    {
        if (PROP_ID(pfpnTemp2 -> fpv.spv.ulPropTag) == PROP_ID(ulPropTag))
        {
            pfpnTemp1 -> pNext = pfpnTemp2 -> pNext;
            *ppRet = pfpnTemp2;
            m_ulCnt--;
            return S_OK;
        }

        pfpnTemp1 = pfpnTemp2;
        pfpnTemp2 = pfpnTemp2 -> pNext;
    }

    return S_OK;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Find()
//
//    Parameters                        
//          ulPropTag       - prop tag of node to delete     
//
//    Purpose
//          returns a pointer to the node w/o disturbing the list                   
//          caller should regard the node as read only, it's still in the list 
//      
//    Return Value
//          pointer to the node if found, NULL otherwise
//
PFlatPropNode WINAPI CFlatPropList :: Find(const ULONG ulPropTag)
{
    PFlatPropNode pfpnTemp = m_pHead;

    if (!pfpnTemp)
        return NULL;

    while (pfpnTemp)
    {
        if (PROP_ID(pfpnTemp -> fpv.spv.ulPropTag) == PROP_ID(ulPropTag))
            return pfpnTemp;
        else
            pfpnTemp = pfpnTemp -> pNext;
    }
    return NULL;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Scan()
//
//    Parameters                        
//          bWalking        - defaults to TRUE, a flag that controls whether
//                            we just started the walk or are in process of
//                            walking     
//
//    Purpose
//          Each call causes you to get a pointer to the next node in the list.
//          Call in a loop until NULL is returned, Pass 0 on the first iteration, 
//          then while NULL != Scan() returns a ptr to successive nodes of the list. 
//          This function DOESN'T return a copy of the node, the caller mustn't free 
//          or otherwise write to the returned pointer.
//      
//    Return Value
//          A pointer to the next list node or NULL if at end of list.
//          

PFlatPropNode WINAPI CFlatPropList :: Scan(BOOL bWalking)
{

    PFlatPropNode  pfpnRet = NULL;
    
    if (!bWalking)
        m_pTemp = m_pHead;

    if (m_pTemp)
    {
        pfpnRet = m_pTemp;
        m_pTemp = m_pTemp -> pNext;        
    }                    
    
    return pfpnRet;       
}

// End of file for ABCACHE.CPP
