///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      ABDB.CPP      
//
//  Description
//      Wraps CDataBase with handy I/F for ABs
//      
//      The following per record streams are implemented:
//
//      1) fldPROPS     --the record's properties
//      2) fldPARENTS   --the record's list of parent containers
//      3) fldCHILDREN  --the record's list of children (N/A for MailUsers)
//      4) fldTABLEROWS --foreign entry's display table rows (N/A for native entries)
//      5) fldNAMEMAP   --name to ID mapping for named properties
//      
//      The organization of each stream is given below. Padding is added to
//      variable length fields when written to allow 8-byte alignment when the stream is
//      read.
//
//      >>fldPROPS or fldNAMEMAP:
//
//      0------>4------>8 ------------>32---------->8N--------->8N+24 . . .  cBytes-1
//      -----------------------------------------------------------------------------
//      ULONG   ULONG   FlatPropHdr    BYTE[varies] FlatPropHdr BYTE[varies] . . . . 
//      cBytes  cProps  Prop1          Data+padding Prop2       Data+padding . . . .
//
//
//      >>fldPARENTS or fldCHILDREN:
//
//      0------>4------>8 -------> . .  cBytes-1
//      -----------------------------------------------------------------------------
//      ULONG   ABPAB_ID ABPAB_ID    
//      cBytes  ID1      ID2        .
//
//
//      >>fldTABLEROWS:
//      
//      0------>4------>8------>12------>16--------->40---------->8M--------->8M+24.... 
//      ----------------------------------------------------------------------------------
//      ULONG   ULONG   ULONG   ULONG    FlatPropHdr BYTE[varies] FlatPropHdr BYTE[varies]
//      ulFlags cRows   cBytes  cProps   r0,Prop1    Data+padding r0,Prop2    Data+padding
//                      in r0   in r0
//
//      8N------>8N+12-->8N+16------>8N+40------->8R--------->8R+24....    
//      ------------------------------------------------------------------ 
//      ULONG   ULONG    FlatPropHdr BYTE[varies] FlatPropHdr BYTE[varies] 
//      cBytes  cProps   r1,Prop1    Data+padding r1,Prop2    Data+padding 
//      in r1   in r1                                                      
//
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#include <windows.h>

#include "abpab.h"
#include "abcache.h"
#include "abdb.h"
#include "list.inl"

#define RECFLD_TABLEROWS RECFLD_NAMES

const ABPAB_ID   DIR_ID = {DIR_ID_NUM,MAPI_ABCONT};
const ABPAB_ID   ROOT_ID = {ROOT_ID_NUM,MAPI_ABCONT};


///////////////////////////////////////////////////////////////////////////////
//
//    Close()
//
//    Parameters                        
//          RecNum      - Reference to a record number for record to close
//
//    Purpose
//          Releases all open streams on the record and invalidates the
//          record number's stream pointers. The record given by RecNum
//          must have been opened by a call to Open(), or its stream pointers
//          initialized to NULL. Using the DECLARE_REC_NUM macro insures that
//          you can successfully Close an unopened RecNum (for error case, e.g.)
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Close(RECORD_NUMBER & RecNum)
{
    HRESULT     hRes = S_OK;

    TraceString1("               Close %x", RecNum.ID.Num);

    for (int i = fldPROPS; i < NUM_REC_FLDS; i++)
    {
        if (RecNum.pCols[i])
            RecNum.pCols[i] -> Release();

        RecNum.pCols[i] = NULL;
    }
    
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    DeleteRec()
//
//    Parameters                        
//          pDB     - pointer to the database object
//          RecNum  - reference to the record number of the record to delete
//
//    Purpose
//          Deletes a record's stream from the database file.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP DeleteRec(PDATABASE pDB,RECORD_NUMBER & RecNum)
{

    Close(RecNum);

    return pDB -> DeleteRecord(RecNum.ID.Num);
}

///////////////////////////////////////////////////////////////////////////////
//
//    InsertRec()
//
//    Parameters                        
//          pDB     - pointer to the database object
//          rnDest  - reference to the destination record number
//          rnSrc   - reference to the source record number
//    Purpose
//          This is how we maintain child/parent relations between
//          a container and subcontainers or MailUsers. This function
//          adds rnSrc's ID to rnDest's list of children and rnDest's 
//          ID to rnSrc's list of parents. Both records must have 
//          been created and opened by a call to Open before this call
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP InsertRec(PDATABASE        pDB, 
                       RECORD_NUMBER  & rnDest,
                       RECORD_NUMBER  & rnSrc)
{
    HRESULT         hRes = S_OK;
    CList<ABPAB_ID> Relations;

    TraceMessage("CDatabase::InsertRec");

    if (MY_ABCONT != rnDest.ID.Type && MY_DISTLIST != rnDest.ID.Type)
        return MAPI_E_CALL_FAILED;

    // if my parent has other children, add me to his family
    if (SUCCEEDED(hRes = Fetch(rnDest,fldCHILDREN, &Relations)))
    {        
        if (FAILED(hRes = Relations.Insert(rnSrc.ID)))
            goto Quit;

        if (FAILED(hRes = Update(rnDest, fldCHILDREN, Relations)))
            goto Quit;

        Relations.Destroy();
    }
    
    // now add my new parent to my list of parents
    if (SUCCEEDED(hRes = Fetch(rnSrc,fldPARENTS, &Relations)))
    {        
        if (FAILED(hRes = Relations.Insert(rnDest.ID)))
            goto Quit;

        if (FAILED(hRes = Update(rnSrc, fldPARENTS, Relations)))
            goto Quit;

        Relations.Destroy();
    }

Quit:
    if (FAILED(hRes))
        TraceResult("Insert", hRes);

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Open()
//
//    Parameters                        
//          pDB      - pointer to the database object
//          ulFlags  - controls how opened: CREATE_IF_NOT_EXIST,OVERWRITE,OPEN_ONLY:         
//          ID       - reference to the ID of the record to open/create
//          RecNum   - reference to the newly opened/created record number
//    Purpose
//          Opens an existing record or creates a new record depending on 
//          the value of ulFlags. CREATE_IF_NOT_EXIST creates a new record if
//          ID isn't already in the database, and opens it if it is. OVERWRITE
//          creates a new record if it doesn't exist and truncates an existing
//          record. OPEN_ONLY opens an existing record or fails the call if it
//          doesn't exist. ID must be a valid ID of a record in the database
//          for OPEN_ONLY, or one obtained via NextID for a new record. A valid
//          record number is returned in RecNum on success.
//      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Open(CDataBase *pDB,ULONG ulFlags, const ABPAB_ID &ID, RECORD_NUMBER & RecNum)
{
    HRESULT     hRes = S_OK;
    LPSTREAM    pStr[NUM_REC_FLDS] = {NULL,NULL,NULL,NULL};
    BOOL        fCreate = (OVERWRITE == ulFlags);
    ULONG       Zero = 0,
                retries = 0;
    int         i;

    TraceString1("               Open %x", ID.Num);
    ZeroMemory((LPVOID) &RecNum, sizeof RecNum);
    ASSERT(ID.Num);

    if (ulFlags != CREATE_IF_NOT_EXIST &&
        ulFlags != OVERWRITE           &&
        ulFlags != OPEN_ONLY)
    {
        return MAPI_E_CALL_FAILED;
    }        

     // Open EXISTING first, if that fails Open Create and increment retries.
     // later, if retries == 1,this is a new record:truncate stream.
     while (retries < 2)
     {
         if (NAME_MAP_ID == ID.Num)
             hRes = pDB -> OpenRecordField(ID.Num,
                                           fCreate,
                                           RECFLD_TABLEROWS,
                                           &pStr[fldTABLEROWS]);
         else
         {
             if (SUCCEEDED(hRes = pDB -> OpenRecordField(ID.Num,
                                                         fCreate,
                                                         RECFLD_PROPERTIES,
                                                         &pStr[fldPROPS])))
                 if (SUCCEEDED(hRes = pDB -> OpenRecordField(ID.Num,
                                                             fCreate,
                                                             RECFLD_PARENTS,
                                                             &pStr[fldPARENTS])))
                 {
                     if (MY_ABCONT    == ID.Type    || 
                         MY_DISTLIST  == ID.Type    ||
                         FOREIGN_ABCONT == ID.Type  ||
                         FOREIGN_DL     == ID.Type)
                         hRes = pDB -> OpenRecordField(ID.Num,
                                                       fCreate,
                                                       RECFLD_CHILDREN,
                                                       &pStr[fldCHILDREN]);
                     if (SUCCEEDED(hRes))    
                         if (ID.Type & FOREIGN_ENTRY)
                             hRes = pDB -> OpenRecordField(ID.Num,
                                                           fCreate,
                                                           RECFLD_TABLEROWS,
                                                           &pStr[fldTABLEROWS]);
                 }
         }

         if (FAILED(hRes) && CREATE_IF_NOT_EXIST == ulFlags)
         {
             retries++;
             fCreate = TRUE;
         }
         else
             break;
     }


    if (SUCCEEDED(hRes))
        // overwrite or newly created, truncate streams
        if (OVERWRITE == ulFlags || 1 == retries)
        {
            const LARGE_INTEGER   liZero = {0,0};
            const ULARGE_INTEGER  uliSize = {sizeof ULONG,0};

            for (int i = fldPROPS; i < NUM_REC_FLDS; i++)
                if (pStr[i])
                {
                    if (FAILED(hRes = pStr[i] -> SetSize(uliSize)))
                        goto Quit;
            
                    if (FAILED(hRes = pStr[i] -> Seek(liZero, STREAM_SEEK_SET,NULL)))
                        goto Quit;

                    if (FAILED(hRes = pStr[i] -> Write((LPVOID) &Zero, sizeof Zero,NULL)))
                        goto Quit;

                }
        }

Quit:
    TraceResult("CDatabase::Open", hRes);

    if (FAILED(hRes))
        for (i = fldPROPS; i < NUM_REC_FLDS; i++)
            if (pStr[i])
            {
                pStr[i] -> Release();
                pStr[i] = NULL;
            }
    
    for (i = fldPROPS; i < NUM_REC_FLDS; i++)
        RecNum.pCols[i] = pStr[i];                    

    RecNum.ID = ID;
    return hRes;                           
}


///////////////////////////////////////////////////////////////////////////////
//
//    Fetch()
//
//    Parameters                        
//          RecNum      - reference to record number whose properties we want
//          Field       - fldPROPS is the only valid value
//          pFlatProps  - pointer to a flat prop list to hold returned
//                        properties
//    Purpose
//          Retrieves RecNum's property stream and converts it to a list of
//          FlatSPropNodes. fldProps is the only valid value for Field, and
//          is provided only for syntactic consistency with other Fetch 
//          overloads. RecNum must have been opened by a call to Open.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Fetch(RECORD_NUMBER  & RecNum,
                   FIELD            Field,
                   CFlatPropList  * pFlatProps)
{
    HRESULT         hRes         = S_OK;
    LARGE_INTEGER   liZero       = {0,0};
    LPSPropValue    pSProp       = NULL;
    LPBYTE          pBuff        = NULL,
                    pData        = NULL,
                    pTemp        = NULL;
    PFlatPropValue  pFlatPropVal = NULL;
    DWORD           i;
    ULONG           cBytes,
					cProps;

    // props stream:
    // 0 - - - 4 - - - 8 - - - ...
    // # bytes  # props 1st prop    2nd prop... etc.

    if (fldPROPS != Field || !pFlatProps)
        return MAPI_E_CALL_FAILED;

    // validate RecNum.pStrmProps

  	if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liZero, STREAM_SEEK_SET,NULL)))
    	goto Quit;

	// read # bytes
    if (FAILED(hRes = RecNum.pCols[Field] -> Read((LPVOID) & cBytes, sizeof cBytes, NULL)))
        goto Quit;

	if (!cBytes)		// 0 bytes of data
		return hRes;

    // new buffer to hold them all
    // force alignment on 8-byte boundary,make sure cBytes is multiple of 8
    pBuff = (LPBYTE) new DWORDLONG[CEIL(cBytes,ALIGNMENT)/ALIGNMENT];

    if (!pBuff)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

	// get # props
   	if (FAILED(hRes = RecNum.pCols[Field] -> Read((LPVOID) & cProps, sizeof cProps, NULL)))
        goto Quit;

	// read prop 1,.., prop N
    if (FAILED(hRes = RecNum.pCols[Field] -> Read(pBuff, cBytes,NULL)))
        goto Quit;

    pTemp = pBuff;

    for (i = 0; i < cProps; i++)
    {
        pFlatPropVal = (PFlatPropValue) pTemp;
        pFlatProps -> Insert(pFlatPropVal);
        pTemp += pFlatPropVal -> cbTotal;
    }

Quit:

    if (pBuff)
        delete [] pBuff;
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Fetch()
//
//    Parameters                        
//          RecNum     - Record number of record whose details we want
//          Field      - field number, must be fldTABLEROWS
//          ppRows     - pointer to an LPRowSet where the returned
//                       table rows are returned. 
//    Purpose
//          Reads RecNum's table rows stream and converts it into
//          a row set. This is for foreign entries we copy into our
//          database. We also copy their details table and store it
//          in the fldTABLEROWS stream.
//
//          The table rows start at the 4th byte. Bytes 0-3 are flags,
//          e.g. the object type OR'ed with FOREIGN_ENTRY | HAS_TEMPLATEID. On 
//          OpenEntry we get the flags first to know what type of object to create
//      
//          SRowSet stream:
//          --------------
//                   V---start here
//          0  .  .  4  .  .  8  .  .  12  .        m  . . .  n       ..etc
//          capablty cRows    row 0    row 0    row 0, ...,  row 0     row 1
//                            cBytes   cProps   prop 0 ...,  prop i    cBytes

//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
//
STDMETHODIMP Fetch(RECORD_NUMBER  & RecNum,
                   FIELD            Field,
                   LPSRowSet      * ppRows)
{
    HRESULT     hRes = S_OK;
    ULONG       cRows,
                i,
                ulRead;
    LPSRowSet   pRows = NULL;
    LARGE_INTEGER   liStart = {sizeof ULONG,0};
    *ppRows = NULL;

    if (fldTABLEROWS != Field)
        return MAPI_E_CALL_FAILED;

    // skip the first 4 bytes (flags)
  	if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liStart, STREAM_SEEK_SET,NULL)))
    	goto Quit;

	// read # rows
    if (FAILED(hRes = RecNum.pCols[Field] -> Read((LPVOID) & cRows, sizeof cRows, &ulRead)))
        goto Quit;

	if (!cRows)		// 0 bytes of data
		return MAPI_E_NOT_FOUND;

    if (FAILED(hRes = gAllocBuff(sizeof SRowSet + (cRows - 1) * sizeof SRow,
                                 (LPVOID *) &pRows)))
        goto Quit;

    pRows -> cRows = 0;     // in case we fail

    for (i = 0; i < cRows; i++, pRows -> cRows++)
        if (FAILED(hRes = ReadRow(RecNum.pCols[Field],&pRows -> aRow[i])))
            goto Quit;

Quit:
    if (FAILED(hRes))
        FreeProws(pRows);
    else
        *ppRows = pRows;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Fetch()
//
//    Parameters                        
//          RecNum          - record number of record whose children or parent's
//                            field we want to read    
//          Field           - ID of field, must be fldCHILDREN or fldPARENTS
//          pRelatedRecs    - returned list of parents or children
//    Purpose
//          Reads the records Parents or Children field, depending on the value
//          of Field, into a list of record IDs
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//        
STDMETHODIMP Fetch(RECORD_NUMBER   & RecNum,
                   FIELD             Field,
                   CList<ABPAB_ID> * pRelatedRecs)
{
    HRESULT         hRes = S_OK;
    ULONG           cBytes,
                    ulNumEntries,
                    ulRead;
    LARGE_INTEGER   li = {0,0};
    LPSTREAM        pStr = NULL;
    PABPAB_ID       pID  = NULL,
                    pBuf = NULL;

    if (fldCHILDREN    == Field && 
        !(MY_ABCONT   & RecNum.ID.Type) && 
        !(MY_DISTLIST & RecNum.ID.Type))
        return MAPI_E_CALL_FAILED;

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(li,STREAM_SEEK_SET,NULL)))
        goto Quit;
     
    if (FAILED(hRes = RecNum.pCols[Field] -> Read((LPVOID) &cBytes, sizeof ULONG, &ulRead)))
        goto Quit;

//!!////////////////////////////////////////////////////////////////////
//  FOUND bug !!! causing hardcode breakpoint !!!!!
//
//
//   RTL: Free Heap block 4b9c00 modified at 4b9c2c after it was freed
//      this is CStream.m_pdwSize = 0x004b9c2c
//
/////////////////////////////////////////////////////////////////////////

    if (cBytes)
    {
        ulNumEntries = cBytes / sizeof ABPAB_ID;
        pBuf = new ABPAB_ID[ulNumEntries];

        if (!pBuf)
        {
            hRes = MAPI_E_NOT_ENOUGH_MEMORY;
            goto Quit;
        }

        if (FAILED(hRes = RecNum.pCols[Field] -> Read((LPVOID) pBuf, cBytes, NULL)))
            goto Quit;

        pID = pBuf;

        for (ULONG i = 0; i < ulNumEntries; i++)
        {
            pRelatedRecs -> Insert(*pID);
            pID++;    
        }
    }
    
Quit:

    if (pBuf)
        delete [] pBuf;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    NextID()
//
//    Parameters                        
//          pDB     - pointer to the database object
//          ulID    - reference to the next available ID number
//
//    Purpose
//          Obtains the next available ID number in the database's 
//          internal sequence. This is the ID of the next record
//          that can be created.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP NextID(PDATABASE pDB, ULONG & ulID)
{

    return pDB -> CreateEmptyRecord(&ulID);

}

///////////////////////////////////////////////////////////////////////////////
//
//    ReadRow()
//
//    Parameters                        
//          pRow     - pointer to destination SRow where the read row is stored
//          pStream  - pointer to the source stream whence the row
//
//    Purpose
//          Reads a single SRow from pStream into pRow -> lpProps. pRow is 
//          allocated by the caller, we allocate the memory for the property
//          array. Caller must insure that the seek ptr is at the beginning of
//          the row to be read, for the 0th row, this means the caller must
//          skip 8 bytes past the flags and cRows.
//      
//          SRowSet stream:
//          --------------
//                            V---rows start here
//          0  .  .  4  .  .  8  .  .  12  .        m  . . .  n       ..etc
//          flags    cRows    row 0    row 0    row 0, ...,  row 0     row 1
//                            cBytes   cProps   prop 0 ...,  prop i    cBytes
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP ReadRow(LPSTREAM pStream,LPSRow pRow)
{
    HRESULT             hRes      = S_OK;
    PFlatPropValue      pFlatProp = NULL;
    LPBYTE              pTemp     = NULL,
                        pBuff     = NULL;
    SRow                srTemp    = {0,0,NULL};  
    ULONG               cBytes    = 0,
                        i,
                        ulRead;
    
    if (FAILED(hRes = pStream -> Read((LPVOID) & cBytes, 
                                       sizeof cBytes,
        	                           &ulRead)))
        goto Quit;            
 
    if (FAILED(hRes = pStream -> Read((LPVOID) & srTemp.cValues, 
                                      sizeof srTemp.cValues,
        	                          &ulRead)))
        goto Quit;            

	if (!cBytes)
        return MAPI_E_NOT_FOUND;
	
    if (FAILED(hRes = gAllocBuff((sizeof SPropValue) * (srTemp.cValues),
                                 (LPVOID *) &srTemp.lpProps)))
        goto Quit; 
                                                
    // new buffer to hold them all. force alignment on
    // 8-byte boundary,make sure cBytes is multiple of 8
    pBuff = (LPBYTE) new DWORDLONG[CEIL(cBytes,ALIGNMENT)/ALIGNMENT];

    if (!pBuff)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

	// read prop 1,.., prop N
    if (FAILED(hRes = pStream -> Read(pBuff, cBytes,NULL)))
        goto Quit;

    pTemp = pBuff;

    for (i = 0; i < srTemp.cValues; i++)
    {
        pFlatProp = (PFlatPropValue) pTemp;

        if (FAILED(hRes = UnFlatten(pFlatProp,(LPVOID)srTemp.lpProps, srTemp.lpProps[i])))
            goto Quit;

        pTemp += pFlatProp -> cbTotal;
    }

Quit:
    if (pBuff)
        delete [] pBuff;

    if (FAILED(hRes))
    {
        gFreeBuff((LPVOID) srTemp.lpProps);        // no leaks cause UnFlatten uses gAllocMore   
        return hRes;
    }

    *pRow = srTemp; 
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    WriteRow()
//
//    Parameters                        
//          pRow     - pointer to source SRow to write
//          pStream  - pointer to the destination whither the row
//
//    Purpose
//          Write a single SRow from pRow -> lpProps into pStream. This function
//          must flatten each SPropValue in pRow -> lpProps to write it in the
//          format expected.Caller must insure that the seek ptr is at the 
//          beginning of the row to be read, for the 0th row, this means the 
//          caller must skip 8 bytes past the flags and cRows.
//      
//          SRowSet stream:
//          --------------
//                   V---rows start here
//          0  .  .  4  .  .  8  .  .  12  .    m  . . .  n       ..etc
//          flags    cRows    row 0    row 0    row 0, ...row 0     row 1
//                            cBytes   cProps   prop 0 ...prop i    cBytes
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP    WriteRow(LPSRow pRow, LPSTREAM pStream)
{
    HRESULT             hRes       = S_OK;
    LARGE_INTEGER       liBookmark = {0,0};
    LARGE_INTEGER       liZero     = {0,0};
    ULONG               cBytes     = 0,
    					cProps     = pRow -> cValues,
                        ulBlobLen,
                        i,
                        ulWritten;
    LPBYTE              pByte;
    PFlatPropNode       pFlatPropNode;


    // set our bookmark at the beginning of this row
    if (FAILED(hRes = pStream -> Seek(liZero,STREAM_SEEK_CUR,(ULARGE_INTEGER *) &liBookmark)))
        goto Quit;

    if (FAILED(hRes = pStream -> Write((LPVOID) & cBytes, 
                                       sizeof cBytes,
        	                           &ulWritten)))
        goto Quit;            

    if (FAILED(hRes = pStream -> Write((LPVOID) & cProps, 
                                       sizeof cProps,
        	                           &ulWritten)))
        goto Quit;            
	
    // write row's  existing columns to the DB 
    for (i = 0; i < cProps; i++)
    {
        pFlatPropNode = NULL;

        if (FAILED(hRes = Flatten(&pRow -> lpProps[i], & pFlatPropNode)))
            goto Quit;

        pByte = (LPBYTE) &(pFlatPropNode -> fpv);
        ulBlobLen = pFlatPropNode -> fpv.cbTotal;                
        hRes = pStream -> Write((LPVOID) pByte, 
                                ulBlobLen,
                                NULL);
        delete pFlatPropNode;

        if (FAILED(hRes))
        {
            cBytes = 0;
            break;
        } 
        cBytes += ulBlobLen;
    }

    
    // go back to our bookmark
     if (FAILED(hRes = pStream -> Seek(liBookmark,STREAM_SEEK_SET,NULL)))
        goto Quit;

    if (FAILED(hRes = pStream -> Write((LPVOID) &cBytes, 
                                       sizeof cBytes,
                                       &ulWritten)))
        goto Quit;

    // calculate where the end of the stream is
    liBookmark.LowPart += cBytes + sizeof cBytes + sizeof cProps;
    
    // go there
    hRes = pStream -> Seek(liBookmark, STREAM_SEEK_SET,NULL);

Quit:

    return hRes;

}

///////////////////////////////////////////////////////////////////////////////
//
//    Update()
//
//    Parameters                        
//          RecNum      - reference to record number whose TABLEROWS 
//                        we want to update
//          Field       - field ID, must be fldTABLEROWS
//          pRows       - pointer to the RowSet for the details
//                        table we want to save
//    Purpose
//          Given an SRowSet, flattens it and writes it to fldTABLEROWS stream.
//          This is used for foreign entries whose details table we save in
//          the database. The fldTABLEROWS stream is a flattened version
//          of the table's rows.RecNum must have been created/opened via Open
//
//          SRowSet stream:
//          --------------
//                   V---rows start here
//          0  .  .  4  .  .  8  .  .  12  .    m  . . .  n       ..etc
//          flags    cRows    row 0    row 0    row 0, .. row 1
//                   cBytes   cProps   prop 0 ..prop i    cBytes
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP Update(RECORD_NUMBER & RecNum,
                    FIELD           Field,
                    LPSRowSet       pRows)
{
    HRESULT         hRes     = S_OK;
    LARGE_INTEGER   liOffset = {sizeof ULONG,0};
    ULONG           cRows    = pRows -> cRows,
                    i,
                    ulWritten;
                                                                               
    if (fldTABLEROWS != Field)
        return MAPI_E_INVALID_PARAMETER;

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liOffset,STREAM_SEEK_SET,NULL)))
        goto Quit;
	    
    // write # rows
    if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) & cRows, 
    	                                           sizeof cRows,
        	                                       &ulWritten)))
        goto Quit;            

    for (i = 0; i < cRows; i++)
        if (FAILED(hRes = WriteRow(&pRows -> aRow[i], RecNum.pCols[Field])))
        {
            cRows = i + 1;
            break;
        }

    if (cRows < pRows -> cRows)
    {
        if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liOffset,STREAM_SEEK_SET,NULL)))
            goto Quit;
	    
        // write # rows
        if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) & cRows, 
    	                                               sizeof cRows,
        	                                           &ulWritten)))
            goto Quit;            
    }
Quit:
    
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    Update()
//
//    Parameters                        
//          RecNum      - reference to record number whose properties 
//                        we want to update
//          Field       - field ID, must be fldPROPS
//          pFlatProps  - list of flattened properties we want to write
//    Purpose
//          Writes this record's property field to the fldPROPS stream.
//          The properties are in a FlatPropList.RecNum must have been created 
//          and opened via Open
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Update(RECORD_NUMBER & RecNum,
                    FIELD           Field,
                    CFlatPropList * pFlatProps)
{

    HRESULT             hRes   = S_OK;
    LARGE_INTEGER       liZero = {0,0};
    ULONG               cBytes = 0,
                        ulBlobLen,
    					cProps;
    LPBYTE              pByte;
    PFlatPropNode       pFlatPropNode;
                                                                               
    // props stream:
    // 0 - - - 4 - - - 8 - - - ...
    // # bytes # props 1st prop    2nd prop... etc.

    if (fldPROPS != Field)
        return MAPI_E_INVALID_PARAMETER;

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liZero,STREAM_SEEK_SET,NULL)))
        goto Quit;
	    
    // write # bytes -this is 0,but it adjusts the seek ptr to the right spot, we overwrite later
	// after we tally up the number of bytes we can't just Seek because this might be first write 
	// and stream might be only 4-bytes
    if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) & cBytes, 
    	                                           sizeof cBytes,
        	                                       NULL)))
        goto Quit;            

    cProps = pFlatProps -> Size();

	if (!cProps)
        return hRes;

    if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) & cProps, 
                                                   sizeof cProps,
        	                                       NULL)))
        goto Quit;            
                                            
    for (pFlatPropNode = pFlatProps -> Scan(0); 
         pFlatPropNode; 
         pFlatPropNode = pFlatProps -> Scan())
    {
        pByte = (LPBYTE) &(pFlatPropNode -> fpv);

        ulBlobLen = pFlatPropNode -> fpv.cbTotal;                
        cBytes += ulBlobLen;       // total length of data + padding
        
        if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) pByte, 
                                                       ulBlobLen,
                                                       NULL)))
            goto Quit;            
    }

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(liZero,STREAM_SEEK_SET,NULL)))
        goto Quit;

    if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) &cBytes, 
                                                   sizeof cBytes,
                                                   NULL)))
        goto Quit;            
        
Quit:
    
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    Update()
//
//    Parameters                        
//          RecNum      - reference to record number whose child or parent 
//                        stream we want to update
//          Field       - field ID, must be fldPARENTS or fldCHILDREN
//          RelatedRecs - list of parent or child IDs we want to write
//    Purpose
//          Writes this record's list of parents or children property field to 
//          the fldPARENTS or fldCHILDREN stream..RecNum must have been created 
//          and opened via Open
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP Update(RECORD_NUMBER   & RecNum,
                    FIELD             Field,
                    CList<ABPAB_ID> & RelatedRecs)
{
    HRESULT         hRes   = S_OK;
    ULONG           cBytes;
    LARGE_INTEGER   li     = {sizeof cBytes,0};
    ABPAB_ID      * pID;

    if (fldCHILDREN == Field && 
        MY_ABCONT   != RecNum.ID.Type && 
        MY_DISTLIST != RecNum.ID.Type)
        return MAPI_E_INVALID_PARAMETER;

    cBytes = (sizeof ABPAB_ID) * (RelatedRecs.Size());

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(li,STREAM_SEEK_SET,NULL)))
        goto Quit;

    for (pID = RelatedRecs.Scan(0);
         pID;
         pID = RelatedRecs.Scan())
    {
        if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) pID, sizeof ABPAB_ID, NULL)))
            goto Quit;
    }    

    li.LowPart = 0;

    if (FAILED(hRes = RecNum.pCols[Field] -> Seek(li,STREAM_SEEK_SET,NULL)))
        goto Quit;


    if (FAILED(hRes = RecNum.pCols[Field] -> Write((LPVOID) &cBytes, sizeof ULONG, NULL)))
        goto Quit;
Quit:
    return hRes;
}

// End of file for ABDB.CPP
