///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      BASEPROP.CPP      
//
//  Description
//      Implementation of base IMAPIProp object
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//

#include "abpab.h"
#include "abcache.h"
#include "cdbfile.h"
#include "abdb.h"
#include "itable.h"
#include "list.inl"




///////////////////////////////////////////////////////////////////////////////
//
//    CBaseProp()
//
//    Parameters                        
//          pLogon              [IN]    the logon object
//          pCreateContainerEID [IN]    pointer to entry ID of container object 
//                                      was created in (used by CreateEntry only)
//          pEID                [IN]    this object's entry ID
//          ulCreateFlags       [IN]    flags for duplicate removal      
//          pParentCont         [IN]    pointer to parent container object        
//          bnewEntry           [IN]    TRUE if this is a new entry
//
//    Purpose
//          Class constructor for base IMAPIProp implementation.
//    Return Value
//          none.
//

 CBaseProp :: CBaseProp(LPCABLogon  pLogon, 
                        PABPAB_EID  pCreateContainerEID,
                        PABPAB_EID  pEID,
                        ULONG       ulCreateFlags,
                        LPCABCont   pParentCont,
                        BOOL        bNewEntry)
 {
    TraceString1("CBaseProp::CBaseProp %x", pEID -> ID.Num);

    m_pLogon        = pLogon;
    m_fDirtyFlags   = 0;     
    m_ulCreateFlags = ulCreateFlags; 
    m_bNewEntry     = bNewEntry;     
    m_cRef          = 0;             
    m_szError[0]    = '\0';
    m_pLogon -> AddRef();            

    if (pParentCont)                 
        pParentCont -> AddRef();     
    
    m_pParentCont = pParentCont;     
    ZeroMemory(&m_MyEID, sizeof ABPAB_EID);
    ZeroMemory(&m_CreateContainerEID, sizeof ABPAB_EID);

    if (pCreateContainerEID)
        memcpy((LPBYTE) &m_CreateContainerEID, 
               (LPBYTE) pCreateContainerEID,   
               sizeof ABPAB_EID);              

    if (pEID)                    
        memcpy((LPBYTE) &m_MyEID,
               (LPBYTE) pEID,    
               sizeof ABPAB_EID);
 

    InitializeCriticalSection(&m_cs);
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CBaseProp()
//
//    Parameters                        
//          none
//
//    Purpose
//          Class destructor for base IMAPIProp implementation
//    Return Value
//          none.
//

CBaseProp :: ~CBaseProp()
{
    TraceString1("CBaseProp::~CBaseProp %x", m_MyEID.ID.Num);

    if (m_pPropCache)
        delete m_pPropCache;

    if (m_pParentCont)
        m_pParentCont -> Release();

    m_pLogon -> Release();   
    DeleteCriticalSection(&m_cs);
}

///////////////////////////////////////////////////////////////////////////////
//
//    Init()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Creates the prop cache object.
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//

STDMETHODIMP CBaseProp :: Init()
{
    if (IsBadReadPtr(m_pLogon -> m_pDB, sizeof PDATABASE))
        return MAPI_E_CALL_FAILED;

    m_pPropCache = new CPropCache(m_pLogon -> m_pDB);

    if (! m_pPropCache)
        return MAPI_E_NOT_ENOUGH_MEMORY;

    return m_pPropCache -> Init();
}



///////////////////////////////////////////////////////////////////////////////
//
//    DeleteDups()
//
//    Parameters                        
//          pParentCont         [in] pointer to object's parent container
//
//    Purpose
//          Processes the CreateEntry 'ulCreateFlags' parameter. Based on
//          the value of m_ulCreateFlags, either returns (no duplicate
//          replacement) or deletes an existing entry with either the same
//          display name or the same display name and search key. CREATE_
//          CHECK_DUP_STRICT means that we delete an existing entry if both 
//          display name and search key match. XX_LOOSE means we delete an
//          existing entry if only the display name matches. Note that
//          even if two entries are duplicates (i.e. same name and/or search
//          key) they are unique entries in the database, (different entry IDs).
//          So a container can have duplicates (two children with same name/
//          search key) but only a single child with a given entry ID.
//          We make a restriction on the contents table to find the duplicates.
//          A hit is:
//          LOOSE :     PR_DISPLAY_NAME == ours && PR_ENTRYID != ours
//          STRICT:     PR_SEARCH_KEY   == ours && PR_ENTRYID != ours
//          We qualify the the search with our entry ID so we can delete the
//          dupes after we add ourself. This way, if the insert fails we haven't
//          deleted anything.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//

STDMETHODIMP CBaseProp :: DeleteDups(LPCABCont pParentCont)
{

    HRESULT         hRes = S_OK;
    LPMAPITABLE     pParentCTbl = NULL;
    enum {NAME,EID,KEY,NUM_COLS};

    static SizedSPropTagArray(NUM_COLS,sptCols) = {NUM_COLS, PR_DISPLAY_NAME, 
                                                             PR_ENTRYID,
                                                             PR_SEARCH_KEY};
    SRestriction sr,
                 srAnd[3];            
    LPTSTR       szTarget = NULL;
    LPSPropValue pProps   = NULL;
    LPSRowSet    pRows    = NULL;
    ULONG        ulCnt;
    LPSBinary    pEIDs    = NULL;
    SBinaryArray sbEIDList;
    if (!(m_ulCreateFlags & (CREATE_REPLACE | CREATE_CHECK_DUP_LOOSE | CREATE_CHECK_DUP_STRICT)))
        return S_OK;
    
    // get my name, entry id, search key    
    if (FAILED(hRes = GetProps((LPSPropTagArray) &sptCols, 0, &ulCnt, &pProps)))
        goto Quit;

    if (PR_DISPLAY_NAME  != pProps[NAME].ulPropTag ||
        PR_SEARCH_KEY    != pProps[KEY].ulPropTag ||
        PR_ENTRYID       != pProps[EID].ulPropTag)
    {
        hRes = MAPI_E_NOT_FOUND;
        goto Quit;
    }

    sr.rt = RES_AND;
    sr.res.resAnd.lpRes = srAnd;

    srAnd[NAME].rt                        = RES_PROPERTY;
    srAnd[NAME].res.resProperty.relop     = RELOP_EQ;
    srAnd[NAME].res.resProperty.ulPropTag = PR_DISPLAY_NAME;
    srAnd[NAME].res.resProperty.lpProp    = &pProps[NAME];

    // NE so we don't delete ourself
    srAnd[EID].rt                         = RES_PROPERTY;
    srAnd[EID].res.resProperty.relop      = RELOP_NE;
    srAnd[EID].res.resProperty.ulPropTag  = PR_ENTRYID;
    srAnd[EID].res.resProperty.lpProp     = &pProps[EID];

    srAnd[KEY].rt                         = RES_PROPERTY;
    srAnd[KEY].res.resProperty.relop      = RELOP_EQ;
    srAnd[KEY].res.resProperty.ulPropTag  = PR_SEARCH_KEY;
    srAnd[KEY].res.resProperty.lpProp     = &pProps[KEY];

    // set restriction based on display name and ~my entry id
    if (m_ulCreateFlags & CREATE_CHECK_DUP_LOOSE)
        sr.res.resAnd.cRes = 2;
    else
        // restriction based on display name, ~my entry id,
        // and my search key
        sr.res.resAnd.cRes = 3;

    // get create container's contents table 
    if (FAILED(hRes = pParentCont -> GetContentsTable(0, &pParentCTbl)))
        goto Quit;

    // get the hits
    if (FAILED(hRes = HrQueryAllRows(pParentCTbl, 
                                     (LPSPropTagArray) &sptCols, 
                                     &sr,
                                     NULL,
                                     0, 
                                     &pRows)))
        goto Quit;

    // turn the rows into an LPENTRYLIST
    if (pRows -> cRows)
    {
        if (FAILED(hRes = gAllocBuff((sizeof SBinary) * (pRows -> cRows), 
                                     (LPVOID *) & pEIDs)))
            goto Quit;

        sbEIDList.cValues = 0;
        sbEIDList.lpbin = pEIDs;
       
        for (ulCnt = 0; ulCnt < pRows -> cRows; ulCnt++)
        {   
            if (PR_ENTRYID == pRows -> aRow[ulCnt].lpProps[EID].ulPropTag)
                pEIDs[sbEIDList.cValues++] = 
                    pRows -> aRow[ulCnt].lpProps[EID].Value.bin;
        }    

        // delete the hits
        if (FAILED(hRes = pParentCont -> DeleteEntries((LPENTRYLIST) &sbEIDList,0)))
            goto Quit;
   } 
Quit:
    if (pParentCTbl)
        pParentCTbl -> Release();

    FreeProws(pRows);
    gFreeBuff(pProps);
    gFreeBuff(pEIDs);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Sets detailed error information based on hRes.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: SetLastError(HRESULT       hRes,    
                                       ULONG         ulLine,    
                                       LPSTR         szFile)
{
    if (FAILED(hRes))
    {
        EnterCriticalSection(&m_cs);
        m_ulContext = ulLine + IE_IABLOGON;
        lstrcpyn(m_szError,szFile,MAX_PATH);
        m_szError[MAX_PATH - 1] = '\0';
        LeaveCriticalSection(&m_cs);
    }
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    GetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Gets detailed error information based on hResult.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: GetLastError(HRESULT       hResult,    
                                       ULONG         ulFlags,    
                                       LPMAPIERROR * lppMAPIError)
{
    HRESULT hRes;
    TraceMessage("CBaseProp :: GetLastError");
    Validate_IMAPIProp_GetLastError(this,
                                    hResult,     
                                    ulFlags,     
                                    lppMAPIError);
    EnterCriticalSection(&m_cs);
    hRes =::GetLastError(m_szError,SZ_PSS_SAMPLE_AB,m_ulContext,lppMAPIError);
    LeaveCriticalSection(&m_cs);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SaveChanges()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Writes changes to the object to the database, i.e.
//          updates the object's record in the database. In some
//          cases, we have to create the object's parent also (e.g.
//          an entry that is a member of a new DL.) We keep a list
//          of orphan objects and add the parent to it; if it
//          is later released without saving changes we delete all
//          the orphan objects.
//
//          After the database is updated, we notify our parents that
//          we've changed
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: SaveChanges(ULONG ulFlags)
{
    HRESULT         hRes = S_OK;
    PABPAB_ID       pID = NULL;

    DECLARE_REC_NUM(ParentRecNum);
    DECLARE_REC_NUM(RecNum);

    TraceMessage("CBaseProp :: SaveChanges");
    Validate_IMAPIProp_SaveChanges(this,ulFlags);

    EnterCriticalSection(&m_cs);

    // We're a oneoff, we can't be saved. This can only happen if a client
    // gets the oneoff EID from oneoff table and does an OpenEntry then SaveChanges.
    // CreateEntry changes the EID to a new one. We treat this case as a noop.
    if (m_MyEID.ID.Type & ONEOFF_ENTRY)
        return S_OK;

    if (FAILED(hRes = Open(m_pLogon -> m_pDB,
                           OPEN_ONLY,
                           m_MyEID.ID, 
                           RecNum)))
    {      
        if (MAPI_E_NOT_FOUND == hRes) // e.g., a one off hasn't been yet
        {
            if (M_FAIL(hRes = Open(m_pLogon -> m_pDB,
                                   CREATE_IF_NOT_EXIST,
                                   m_MyEID.ID,
                                   RecNum)))  
                goto Quit;
    
        }
        else
            goto Quit;
    }
    // It's possible that we get saved before the container we're 
    // created in does, e.g. if we're 'Add Members'd' to a new DL
    // If so, we create a record for the unsaved container and
    // add it to the orphan list in case it isn't saved later
    // Note that this only applies for a newly created entry, not
    // an existing entry that's just been OpenEntry'd and modified. 
    // In this case we know we have at least one parent and we don't know
    // which container we were created in anyway, so we skip this.
    if (m_CreateContainerEID.ID.Num)        // 0 if OpenEntry existing record
    {
        if (M_FAIL(hRes = Open(m_pLogon -> m_pDB,
                               OPEN_ONLY,
                               m_CreateContainerEID.ID,
                               ParentRecNum)))  
        {   // my parent might not have been saved yet, e.g. create a new DL
            // add me as a member before SaveChanges on him
            if (MAPI_E_NOT_FOUND == hRes)                
                if (SUCCEEDED(hRes = Open(m_pLogon -> m_pDB,
                                          CREATE_IF_NOT_EXIST,
                                          m_CreateContainerEID.ID,
                                          ParentRecNum)))  
                    m_pLogon -> m_OrphanRecs.Insert(m_CreateContainerEID.ID);
                else
                    goto Quit;

            else
                goto Quit;
        }

        // insert me in my parent
        if (M_FAIL(hRes = InsertRec(m_pLogon -> m_pDB,
                                    ParentRecNum,
                                    RecNum)))
            goto Quit;

        Close(ParentRecNum);
        DeleteDups(m_pParentCont);  // do last so we don't delete entries if insert fails
    }

    m_pLogon -> m_OrphanRecs.Delete(m_MyEID.ID, &pID);

    if (pID)
        delete pID;

    // write all the new properties to the DB              
    if (m_fDirtyFlags & PROPS)
    {
        LPSPropTagArray psptTags = NULL;

        if (M_FAIL(hRes = m_pPropCache -> Flush(RecNum)))
            goto Quit; 

        if (SUCCEEDED(GetPropList(0, &psptTags)))
        {
            m_pLogon  -> Notify(fnevObjectModified,
                                NULL, 
                                &m_MyEID,
                                psptTags);
            gFreeBuff((LPVOID) psptTags);
        }

    }

    hRes = NotifyParents(RecNum,
                         m_bNewEntry ? ABPAB_CHILD_ADDED : ABPAB_CHILD_MODIFIED,
                         m_pPropCache -> m_pProps);
    m_bNewEntry = FALSE;                           
Quit:

    Close(RecNum);   
    m_fDirtyFlags = 0; // clean
    LeaveCriticalSection(&m_cs);

    return hRes;
}     

///////////////////////////////////////////////////////////////////////////////
//
//    GetAllProps()
//
//    Parameters                        
//          pcValues    - pointer to a ULONG where we write the count of 
//                        returned properties
//          ppProps     - pointer to an SPropValue where we write the returned
//                        properties
//    Purpose
//          Retrieve all the object's properties from the cache, unflatten
//          them and pass them back in an SPropValue array.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: GetAllProps(ULONG        * pcValues,
                                      LPSPropValue * ppProps)
{
    HRESULT         hRes = S_OK;
    PFlatPropNode   pFlatPropNode = NULL;
    LPSPropValue    pProps = NULL;
    ULONG           cValues;
        
    TraceMessage("CBaseProp :: GetAllProps");
    
    cValues = m_pPropCache -> m_pProps -> Size();

    if (!cValues )
        return hRes;        
      
    if (FAILED(hRes = gAllocBuff(cValues * (sizeof SPropValue),
                                (LPVOID *) &pProps)))
        return MAPI_E_NOT_ENOUGH_MEMORY;

    cValues = 0;

    for (pFlatPropNode = m_pPropCache -> m_pProps -> Scan(0);
         pFlatPropNode;
         pFlatPropNode = m_pPropCache -> m_pProps -> Scan())
    {
        if (PROP_ID(PR_INSTANCE_KEY) != PROP_ID(pFlatPropNode -> fpv.spv.ulPropTag))
        {
            if (FAILED(UnFlatten(pFlatPropNode,
                                 pProps,
                                 pProps[cValues])))
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                gFreeBuff((LPVOID)pProps);
                goto Quit;
            }
            cValues++;
        }
    }

Quit:
    if (SUCCEEDED(hRes))
    {
        *ppProps = pProps;
        *pcValues = cValues;
    }

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Retrieve one or more properties from the object's property
//          cache. The cache is a linked list of FlatPropNodes,
//          we unflatten each node into the output LPSPropValue array
//          if/when the target is found. We use the list's methods
//          to search for the property.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: GetProps(LPSPropTagArray lpPropTagArray,     
                                   ULONG           ulFlags,    
                                   ULONG         * lpcValues,  
                                   LPSPropValue  * lppPropArray)
{
    HRESULT         hRes = S_OK;
    PFlatPropNode   pFlatPropNode = NULL;
    LPSPropValue    pProps = NULL;
    ULONG           cValues = 0;
    BOOL            bFound;
    int             i;
        
    TraceString1("CBaseProp :: GetProps %x\n",m_MyEID.ID.Num);
    Validate_IMAPIProp_GetProps(this,
                                lpPropTagArray,
                                ulFlags,      
                                lpcValues,    
                                lppPropArray); 
    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    EnterCriticalSection(&m_cs);

    if (!lpPropTagArray)
    {   // doesn't depend on order like GetPropList so scans once
        hRes = GetAllProps(&cValues, &pProps);
        goto Quit;
    }

    if (M_FAIL(hRes = gAllocBuff((lpPropTagArray -> cValues) * (sizeof SPropValue),
                                 (LPVOID *) &pProps)))
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

    cValues = lpPropTagArray -> cValues;

    for (i = 0; i < (int)cValues; i++)
    {
        if (PR_NULL == lpPropTagArray -> aulPropTag[i])
            pProps[i].ulPropTag = PR_NULL;
        else
        {
            TraceString1("    %08x\n",lpPropTagArray -> aulPropTag[i]);

            bFound = FALSE;
            
            // exclude PR_INSTANCE_KEY from returnable props
            if (PROP_ID(PR_INSTANCE_KEY) != PROP_ID(lpPropTagArray -> aulPropTag[i]))
            {
                pFlatPropNode = m_pPropCache -> 
                                m_pProps     -> Find(lpPropTagArray -> aulPropTag[i]);
        
                if (pFlatPropNode)
                {
                    if (M_FAIL(UnFlatten(pFlatPropNode,
                                         pProps,
                                         pProps[i])))
                    {
                        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                        goto Quit;
                    }
                    bFound = TRUE;
                }
            }

            if (!bFound)
                hRes = HrPropNotFound(pProps[i],lpPropTagArray -> aulPropTag[i]);
        }
    }

Quit:
    LeaveCriticalSection(&m_cs);
    
    if (SUCCEEDED(hRes))
    {    
        *lppPropArray = pProps;
        *lpcValues = cValues;
    }
    else
        gFreeBuff((LPVOID)pProps);

    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    GetPropList()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Scan the property cache and return a list of all the
//          property tags we find therein,
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: GetPropList(ULONG             ulFlags,    
                                      LPSPropTagArray * lppPropTagArray)
{
    HRESULT         hRes          = S_OK;
    ULONG           ulCnt;
    LPSPropTagArray pTags         = NULL;
    PFlatPropNode   pFlatPropNode = NULL;

    TraceMessage("CBaseProp :: GetPropList");
    Validate_IMAPIProp_GetPropList(this,ulFlags,lppPropTagArray);
    
    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    EnterCriticalSection(&m_cs);
    ulCnt = m_pPropCache -> m_pProps -> Size();

    if (M_FAIL(hRes = gAllocBuff(CbNewSPropTagArray(ulCnt),
                                 (LPVOID *) & pTags)))
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

    pTags -> cValues = ulCnt;

    ulCnt = 0;

    for (pFlatPropNode = m_pPropCache -> m_pProps -> Scan(0);
         pFlatPropNode;
         pFlatPropNode = m_pPropCache -> m_pProps -> Scan())
    {
        // don't list PR_INSTANCE_KEY
        if (PROP_ID(PR_INSTANCE_KEY) != PROP_ID(pFlatPropNode -> fpv.spv.ulPropTag))
        {
            pTags -> aulPropTag[ulCnt] = pFlatPropNode -> fpv.spv.ulPropTag;
            ulCnt++;    
        }
    }
    
Quit:
    LeaveCriticalSection(&m_cs);
    
    if (FAILED(hRes))
        gFreeBuff((LPVOID) pTags);
    else
        *lppPropTagArray = pTags;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    HrGetOneProp()
//
//    Parameters                        
//          ulTag   - property sought
//          ppProp  - pointer where we write the property if found
//
//    Purpose
//          Retrieve a single property given by ulTag from the object's
//          property cache into *ppProp. Caller must free *ppProp
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: HrGetOneProp(ULONG ulTag, LPSPropValue * ppProp)
{
    HRESULT         hRes = S_OK;
    PFlatPropNode   pNode;
    LPSPropValue    pProp = NULL;

    *ppProp = NULL;

    if (PR_NULL == ulTag || PT_ERROR == PROP_TYPE(ulTag)) 
        return hRes;

    if (PT_OBJECT == ulTag)
        return MAPI_E_NO_SUPPORT;

    pNode = m_pPropCache -> m_pProps -> Find(ulTag);

    if (!pNode)
        return MAPI_E_NOT_FOUND;

    if (FAILED(hRes = gAllocBuff(sizeof SPropValue, (LPVOID *) & pProp)))
        return MAPI_E_NOT_ENOUGH_MEMORY;
    
    if (SUCCEEDED(hRes = UnFlatten(pNode, pProp, *pProp)))
        *ppProp = pProp;
    else
        gFreeBuff(pProp);

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CopyProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details. We use the support
//          object to do the work.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise

STDMETHODIMP CBaseProp :: CopyProps(LPSPropTagArray       pIncludeProps,             
                                    ULONG                 ulUIParam,                  
                                    LPMAPIPROGRESS        pProgress,                 
                                    LPCIID                pInterface,                
                                    LPVOID                pDestObj,                  
                                    ULONG                 ulFlags,                    
                                    LPSPropProblemArray * ppProblems)
{   
    TraceString1("CBaseProp <%x> :: CopyProps", m_MyEID.ID.Num);
    Validate_IMAPIProp_CopyProps(this,
                                 pIncludeProps,
                                 ulUIParam,    
                                 pProgress,    
                                 pInterface,   
                                 pDestObj,     
                                 ulFlags,      
                                 ppProblems);
    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;
                                
    return m_pLogon -> m_pSupObj -> DoCopyProps(&IID_IABContainer,
                                                this,
                                                pIncludeProps,
                                                ulUIParam,
                                                pProgress,
                                                pInterface,
                                                pDestObj,
                                                ulFlags,
                                                ppProblems);
}         


///////////////////////////////////////////////////////////////////////////////
//
//    PropSettable()
//
//    Parameters                        
//          ulPropTag   - property tag of property to test
//
//    Purpose
//          Tests if a property is writeable by a client.
//      
//    Return Value
//          S_OK if the property is writable a MAPI error 
//          code otherwise
//

STDMETHODIMP CBaseProp :: PropSettable(ULONG ulPropTag)
{
    // no multi-valued properties
    if (ISMV_TAG(ulPropTag))
        return MAPI_E_NO_SUPPORT;

    if (FBadPropTag(ulPropTag))
        return MAPI_E_UNEXPECTED_ID;

    if (PT_OBJECT == PROP_TYPE(ulPropTag))
        return MAPI_E_NO_ACCESS;

    switch(PROP_ID(ulPropTag))
    {
        case PROP_ID(PR_SEARCH_KEY):
        case PROP_ID(PR_ENTRYID):
        case PROP_ID(PR_INSTANCE_KEY):
        case PROP_ID(PR_OBJECT_TYPE):
        case PROP_ID(PR_DISPLAY_TYPE):
        case PROP_ID(PR_CONTAINER_FLAGS):
        case PROP_ID(PR_AB_PROVIDER_ID):
        case PROP_ID(PR_ACCESS):
        case PROP_ID(PR_ACCESS_LEVEL):
            return MAPI_E_COMPUTED;
        default:
            return S_OK;
    }    
}      


///////////////////////////////////////////////////////////////////////////////
//
//    HrSetOneProp()
//
//    Parameters                        
//          pProp   - property to set   
//  
//    Purpose
//          Sets one property into the prop cache and updates any
//          properties derived from it.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CBaseProp :: HrSetOneProp(LPSPropValue pProp)
{
    HRESULT     hRes = S_OK;

    if (PR_NULL == pProp -> ulPropTag || PT_ERROR==PROP_TYPE(pProp -> ulPropTag))
        return hRes;

    if (FAILED(hRes = PropSettable(pProp -> ulPropTag)))
        return hRes;

    if (FAILED(hRes = UpdateComputedProps(pProp)))
        return hRes;

    m_fDirtyFlags |= PROPS;

    return m_pPropCache -> m_pProps -> Insert(pProp);
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Add one or more properties to the property cache. Only settable
//          properties can be set here. Computed properties are updated
//          if any property they're derived from changes. Lastly, the
//          object's dirty flag is set for it's properties.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
//
STDMETHODIMP CBaseProp :: SetProps(ULONG                 cValues,    
                                   LPSPropValue          lpPropArray,
                                   LPSPropProblemArray * lppProblems)
{   
    HRESULT             hRes = S_OK;
    LPSPropProblemArray pProblems = NULL;
    ULONG               ulBufLen;
    BOOL                bProblems = FALSE;

    TraceString1("CBaseProp :: SetProps %x\n", m_MyEID.ID.Num);

    if (lppProblems)            // caller can pass NULL if doesn't care about problems
    {
        *lppProblems = NULL;
        ulBufLen = (sizeof SPropProblemArray) + ((cValues - 1) * sizeof SPropProblem);
    
        if (M_FAIL(gAllocBuff(ulBufLen,(LPVOID *) &pProblems)))
            return MAPI_E_NOT_ENOUGH_MEMORY;   

        pProblems -> cProblem = 0;
    }

    EnterCriticalSection(&m_cs);
                                        
    for (ULONG i = 0; i < cValues; i++)
    {
        TraceString1("    %08x\n",lpPropArray[i].ulPropTag);

        if (M_FAIL(hRes = HrSetOneProp(&lpPropArray[i])))
        {
            HrSetPropProblem(pProblems, i, lpPropArray[i].ulPropTag, hRes);
            bProblems = TRUE;
        }
    }

    if (lppProblems)
    {
        if (bProblems)    
            *lppProblems = pProblems;
        else
            gFreeBuff((LPVOID)pProblems);
    }

    m_fDirtyFlags |= PROPS;
    LeaveCriticalSection(&m_cs);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    UpdateComputedProps()
//
//    Parameters                        
//          pProp   - pointer to the property from which the computed
//                    property is derived.
//
//    Purpose
//          Some properties are computed from other settable ones.
//          This fcn updates the computed properties derived from the
//          settable ones every time the settable property changes.
//          This function is virtual in the base class. Overide with
//          one appropriate to the derived object.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
// 
STDMETHODIMP CBaseProp :: UpdateComputedProps(LPSPropValue pProp)
{
    HRESULT         hRes = S_OK;
    TCHAR           szTemp[MAX_SEARCH_KEY];
    LPTSTR          szPtr = NULL;
    ULONG           ulLen;
    SPropValue      spvComputed;

    PFlatPropNode   pFlatPNode = NULL;

    switch(pProp -> ulPropTag)
    {
        case PR_EMAIL_ADDRESS:

            pFlatPNode = m_pPropCache -> m_pProps -> Find(PR_ADDRTYPE);

            if (!pFlatPNode)
                break;

            szPtr = (LPTSTR) pFlatPNode -> fpv.Data;
            lstrcpy(szTemp, szPtr);
            ulLen = lstrlen(szPtr);
            szTemp[ulLen] = ':';
            ulLen++;
            lstrcpy(&szTemp[ulLen],pProp -> Value.LPSZ);
            Normalize(szTemp,&ulLen);
            spvComputed.ulPropTag     = PR_SEARCH_KEY;
            spvComputed.Value.bin.cb  = ulLen + 1;
            spvComputed.Value.bin.lpb = (LPBYTE) szTemp;

            hRes = m_pPropCache -> m_pProps -> Insert(&spvComputed);
            break;

        case PR_ADDRTYPE:

            pFlatPNode = m_pPropCache -> m_pProps -> Find(PR_EMAIL_ADDRESS);

            if (!pFlatPNode)
                break;
            
            // personal DL has no email address,search key defaults to EID
            if (0 == lstrcmpi(pProp -> Value.lpszA,"MAPIDL"))
                break;

            lstrcpy(szTemp, pProp -> Value.LPSZ);
            ulLen = lstrlen(szTemp);
            szTemp[ulLen] = ':';
            ulLen++;
            szPtr = (LPTSTR) pFlatPNode -> fpv.Data;
            lstrcpy(&szTemp[ulLen], szPtr);
            Normalize(szTemp,&ulLen);
            spvComputed.ulPropTag     = PR_SEARCH_KEY;
            spvComputed.Value.bin.cb  = ulLen + 1;
            spvComputed.Value.bin.lpb = (LPBYTE) szTemp;

            hRes = m_pPropCache -> m_pProps -> Insert(&spvComputed);
            break;
            
        case PR_DISPLAY_NAME:
            spvComputed.ulPropTag     = PR_7BIT_DISPLAY_NAME;
            spvComputed.Value.lpszA   = pProp -> Value.lpszA;
            hRes = m_pPropCache -> m_pProps -> Insert(&spvComputed);
            break;

        default:
            return S_OK;
    }

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    HrSetPropProblem()
//
//    Parameters                        
//          pProbs      - pointer to the problem array to set
//          ulIdx       - index into the problem array for given property
//          ulTag       - tag of errant property
//          hRes        - error code to set in problem array
//    Purpose
//          Helper function for setting errors in a prop problem array.
//          Sets the next space in the problem array with the error
//          code for the erroneous property given by ulTag.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
void HrSetPropProblem(LPSPropProblemArray pProbs,
                      ULONG               ulIdx, 
                      ULONG               ulTag,
                      HRESULT             hRes)
{
    ULONG j;

    if (!pProbs)
        return;

    j = pProbs -> cProblem++;
    // the index of the errant propArray element
    pProbs -> aProblem[j].ulIndex = ulIdx; 
    pProbs -> aProblem[j].ulPropTag = ulTag;
    pProbs -> aProblem[j].scode = hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Returns an object when requested a supported property
//          of type PT_OBJECT. This is virtual in the base class.
//          Derived classes should declare an override appropriate
//          to the object type.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
//

STDMETHODIMP CBaseProp :: OpenProperty(ULONG       ulPropTag,  
                                       LPCIID      lpiid,      
                                       ULONG       ulInterfaceOptions, 
                                       ULONG       ulFlags,    
                                       LPUNKNOWN * lppUnk)
{                                         
    TraceMessage("CBaseProp :: OpenProperty");
    Validate_IMAPIProp_OpenProperty(this,
                                    ulPropTag,         
                                    lpiid,             
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    lppUnk);            
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//
//    DeleteProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Deletes the properties given in lpPropTagArray from the
//          property cache. We do a Delete on the property cache's list for 
//          each tag and destroy the node if found. Lastly, set the dirty
//          flag for this object's properties.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: DeleteProps(LPSPropTagArray       lpPropTagArray,     
                                      LPSPropProblemArray * lppProblems)
{
    HRESULT             hRes = S_OK;
    PFlatPropNode       pFlatPropNode = NULL;
    LPSPropProblemArray pProblems = NULL;
    ULONG               ulBufLen,
                        i;
    BOOL                bProblems = FALSE;

    TraceMessage("CBaseProp :: DeleteProps");
    Validate_IMAPIProp_DeleteProps(this,
                                   lpPropTagArray,
                                   lppProblems);   

    if (lppProblems)            // caller passes NULL if doesn't care about problems
    {
        *lppProblems = NULL;
        ulBufLen = CbNewSPropProblemArray(lpPropTagArray -> cValues); 
    
        if (M_FAIL(gAllocBuff(ulBufLen,(LPVOID *) &pProblems)))
            return MAPI_E_NOT_ENOUGH_MEMORY;   

        pProblems -> cProblem = 0;
    }

    EnterCriticalSection(&m_cs);

    for (i = 0; i < lpPropTagArray -> cValues; i++)
    {
        m_pPropCache -> 
        m_pProps     -> Delete(lpPropTagArray -> aulPropTag[i],
                               &pFlatPropNode);
        if (!pFlatPropNode)        
            if (pProblems)
            {
                HrSetPropProblem(pProblems, 
                                 i, 
                                 lpPropTagArray -> aulPropTag[i],
                                 MAPI_E_NOT_FOUND);
                bProblems = TRUE;
            }
        delete pFlatPropNode;
    }

    if (lppProblems)
    {
        if (bProblems)    
            *lppProblems = pProblems;
        else
            gFreeBuff((LPVOID)pProblems);
    }

    m_fDirtyFlags |= PROPS;
    LeaveCriticalSection(&m_cs);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CopyTo()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CBaseProp :: CopyTo(ULONG                 ciidExclude,
                                 LPCIID                rgiidExclude,       
                                 LPSPropTagArray       lpExcludeProps,     
                                 ULONG                 ulUIParam,  
                                 LPMAPIPROGRESS        lpProgress, 
                                 LPCIID                lpInterface,
                                 LPVOID                lpDestObj,  
                                 ULONG                 ulFlags,    
                                 LPSPropProblemArray * lppProblems)
{
    TraceMessage("CBaseProp :: CopyTo");
    Validate_IMAPIProp_CopyTo(this,
                              ciidExclude,   
                              rgiidExclude,  
                              lpExcludeProps,
                              ulUIParam,     
                              lpProgress,    
                              lpInterface,   
                              lpDestObj,     
                              ulFlags,       
                              lppProblems);

    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CopyNamedID
//
//    Parameters                        
//          
//
//    Purpose
//          Copies a row from our mapping table into a newly allocated
//          MAPINAMEID struct and fills in the fields, and returns the 
//          MAPINAMEDID, which the caller must free.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//  
STDMETHODIMP CopyNamedID(LPVOID pParent, LPSRow pSRow, LPMAPINAMEID * ppNameID)
{
    HRESULT      hRes;
    LPMAPINAMEID pName = NULL;

    if (FAILED(hRes = gAllocMore(sizeof MAPINAMEID, pParent,(LPVOID *) & pName)))
        goto Quit;

    if (FAILED(hRes = gAllocMore(sizeof MAPIUID, pParent,(LPVOID *) & pName -> lpguid)))
        goto Quit;

    memcpy(pName -> lpguid, 
           pSRow -> lpProps[MAP_UID].Value.bin.lpb, 
           sizeof MAPIUID);

    pName -> ulKind =  pSRow -> lpProps[MAP_KIND].Value.l;

    if (MNID_ID == pName -> ulKind) 
        pName -> Kind.lID = pSRow -> lpProps[MAP_ULNAME].Value.l;
    else
    {
        ULONG len = CbtszsizeW(pSRow -> lpProps[MAP_SZNAME].Value.lpszW);

        if (FAILED(hRes = gAllocMore(len, pParent,(LPVOID *) & pName -> Kind.lpwstrName)))
            goto Quit;

        memcpy(pName -> Kind.lpwstrName,
               pSRow -> lpProps[MAP_SZNAME].Value.lpszW,
               len);
    }

    *ppNameID = pName;
Quit:
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetAllNames()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//  

STDMETHODIMP CBaseProp :: GetAllNames(LPSRestriction pRes,LPSPropTagArray *ppPropTags, LPMAPINAMEID ** pppNIDs)
{
    HRESULT         hRes,
                    hr      = S_OK;
    LPSRowSet       psr     = NULL;
    LPSPropTagArray pspt    = NULL;
    LPMAPINAMEID  * ppNIDs   = NULL;
    int             i;
    ULONG           ulID;

    *ppPropTags = NULL;
    *pppNIDs    = NULL;

    EnterCriticalSection(&m_pLogon -> m_MapCS);
    hRes = HrQueryAllRows(m_pLogon -> m_pMapView,
                          (LPSPropTagArray)&g_sptMAPTags,
                          pRes,
                          NULL,
                          0,
                          &psr);

    // if we restricted for PS_MAPI, remove restriction
    if (pRes)   
        m_pLogon -> m_pMapView -> Restrict(NULL,0);
    LeaveCriticalSection(&m_pLogon -> m_MapCS);
    
    if (FAILED(hRes))
        goto Quit;

    if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(psr -> cRows),(LPVOID*) &pspt)))
        goto Quit;

    if (FAILED(hRes = gAllocBuff((sizeof LPMAPINAMEID) * (psr -> cRows),(LPVOID*) &ppNIDs)))
        goto Quit;

    pspt -> cValues = 0;

    for (i = 0; i < (int) psr -> cRows; i++)
    {

        if (PR_MAP_ID != psr -> aRow[0].lpProps[MAP_ID].ulPropTag)
        {
            hr = MAPI_W_ERRORS_RETURNED;
            pspt -> aulPropTag[i] = PROP_TAG(PT_ERROR,0);
            continue;
        }

        ulID = psr -> aRow[i].lpProps[MAP_ID].Value.l;

        // skip the header with ID == @ 0x8000
        if (ulID > BASE_NAMED_ID)
        {
            pspt -> aulPropTag[i] = PROP_TAG(PT_UNSPECIFIED, ulID); 
        
            if (FAILED(hRes = CopyNamedID(ppNIDs, &psr -> aRow[i], &ppNIDs[pspt -> cValues])))
                goto Quit;

            pspt -> cValues++;
        }
    }

    *ppPropTags = pspt;
    *pppNIDs    = ppNIDs;

Quit:
    
    FreeProws(psr);   
    
    if (FAILED(hRes))
        gFreeBuff(ppNIDs);
        
    return hRes == S_OK ? hr : hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetNamesFromIDs()
//
//    Parameters                        
//          See the MAPI documentation for details.
//    Purpose
//          See the MAPI documentation for details.      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise 
//
STDMETHODIMP CBaseProp :: GetNamesFromIDs(LPSPropTagArray * ppPropTags,
                                          LPGUID            lpPropSetGuid,      
                                          ULONG             ulFlags,    
                                          ULONG           * lpcPropNames,       
                                          LPMAPINAMEID  * * lpppPropNames)
{
    SPropValue      spv;
    HRESULT         hRes,
                    hr = S_OK;
    SRestriction    sr,
                  * pRes = NULL;  
    LPSRowSet       psr = NULL;
    LPMAPINAMEID  * ppNames = NULL;
    int             cnt = 0;

    TraceMessage("CBaseProp :: GetNamesFromIDs");    
    Validate_IMAPIProp_GetNamesFromIDs(this,
                                       ppPropTags,    
                                       lpPropSetGuid, 
                                       ulFlags,       
                                       lpcPropNames,  
                                       lpppPropNames); 

    if (! m_pLogon -> m_pMapView)
        if (M_FAIL(hRes = m_pLogon -> GetMapTable()))
            return hRes;

    sr.rt = RES_PROPERTY;
    sr.res.resProperty.relop  = RELOP_EQ;
    sr.res.resProperty.lpProp = &spv;

    if (!*ppPropTags)
    {
        if (lpPropSetGuid && PS_MAPI == *lpPropSetGuid)
        {

            spv.ulPropTag = PR_MAP_UID;
            spv.Value.lpguid = (LPGUID) &PS_MAPI;

            sr.res.resProperty.ulPropTag = PR_MAP_UID;
            pRes = & sr;
        }

        M_FAIL(hRes = GetAllNames(pRes,ppPropTags,lpppPropNames));
        *lpcPropNames = (*ppPropTags) -> cValues;
        return hRes;
    }

    int cNames = (int) (*ppPropTags) -> cValues;

    for (int i = 0; i < cNames; i++)
    {
        if (FAILED(hRes = gAllocBuff(sizeof LPMAPINAMEID * cNames, (LPVOID *) &ppNames)))
            goto Quit;

        sr.res.resProperty.ulPropTag = PR_MAP_ID;

        spv.ulPropTag = PR_MAP_ID;
        spv.Value.l   = PROP_ID((*ppPropTags) -> aulPropTag[i]);
        TraceString1("    %08x\n",spv.Value.l);

        EnterCriticalSection(& m_pLogon -> m_MapCS);

        if (SUCCEEDED(hRes = m_pLogon -> m_pMapView -> FindRow(&sr,BOOKMARK_BEGINNING,0)))
            hRes = m_pLogon -> m_pMapView -> QueryRows(1,0,&psr);
        
        LeaveCriticalSection(& m_pLogon -> m_MapCS);
        
        if (MAPI_E_NOT_FOUND == hRes)
        {
            ppNames[i] = NULL;            
            hr = MAPI_W_ERRORS_RETURNED;
        }
        else
        {
            if (M_FAIL(hRes))
                goto Quit;

            if (M_FAIL(hRes = CopyNamedID((LPVOID) ppNames,&psr -> aRow[0], &ppNames[i])))
                goto Quit;
            
            cnt++;
            FreeProws(psr);
            psr = NULL;
        }
    }

    *lpppPropNames = ppNames;
    *lpcPropNames  = cnt;
Quit:
    FreeProws(psr);

    if (FAILED(hRes))
        gFreeBuff(ppNames);

    return hRes == S_OK ? hr : hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    NewMapping()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//  
STDMETHODIMP CBaseProp :: NewMapping(LPMAPINAMEID pName, ULONG * pulID)
{
    HRESULT         hRes;
    SRestriction    sr;
    SPropValue      spv;
    SizedSRowSet(2,srNewRows);
    LPSRowSet       psr = NULL;
    ULONG           ulID;

    *pulID = 0;
    spv.ulPropTag = PR_MAP_ID;
    spv.Value.l   = BASE_NAMED_ID;

    sr.rt = RES_PROPERTY;
    sr.res.resProperty.relop     = RELOP_EQ;
    sr.res.resProperty.ulPropTag = PR_MAP_ID;
    sr.res.resProperty.lpProp    = &spv;

    srNewRows.aRow[0].lpProps = srNewRows.aRow[1].lpProps = NULL;

    if (FAILED(hRes = m_pLogon  -> 
                      m_pMapView -> FindRow(&sr,BOOKMARK_BEGINNING,0)))
    {
        // no header row, bail
        if (MAPI_E_NOT_FOUND == hRes)
            goto Quit;
    }
    else
    {
        if (FAILED(hRes = m_pLogon   ->
                          m_pMapView -> QueryRows(1,0,&psr)))
            goto Quit;

        ulID = psr -> aRow[0].lpProps[MAP_NEXTID].Value.l;
        FreeProws(psr);

        if (ulID > 0xFFFFF)
        {
            hRes = MAPI_E_NOT_ENOUGH_RESOURCES;
            goto Quit;
        }
    }

    srNewRows.cRows = 2;

    // initialize the table w/ 2 rows: the header row for this prop set
    // and the row containing the next ID.

    // first row 
    if (FAILED(hRes = gAllocBuff(sizeof SPropValue * NUM_MAP_PROPS, 
                                 (LPVOID *) &srNewRows.aRow[0].lpProps)))
        goto Quit;

    // UID of our property set
    srNewRows.aRow[0].lpProps[MAP_UID].ulPropTag     = PR_MAP_UID;
    srNewRows.aRow[0].lpProps[MAP_UID].Value.bin.cb  = sizeof PS_ABPAB;
    srNewRows.aRow[0].lpProps[MAP_UID].Value.bin.lpb = (LPBYTE) &PS_ABPAB;

    // ID for this row
    srNewRows.aRow[0].lpProps[MAP_ID].ulPropTag      = PR_MAP_ID;
    srNewRows.aRow[0].lpProps[MAP_ID].Value.l        = BASE_NAMED_ID;

    // we don't have a Unicode string name
    srNewRows.aRow[0].lpProps[MAP_SZNAME].ulPropTag   = PR_MAP_SZNAME;
    srNewRows.aRow[0].lpProps[MAP_SZNAME].Value.lpszW = L"";

    // numeric name for this mapping
    srNewRows.aRow[0].lpProps[MAP_ULNAME].ulPropTag   = PR_MAP_ULNAME;
    srNewRows.aRow[0].lpProps[MAP_ULNAME].Value.l     = BASE_NAMED_ID;
    
    // Increment column for next ID for our prop set. Only found 
    // in the first row of our prop set. In other rows its the ulKind
    srNewRows.aRow[0].lpProps[MAP_NEXTID].ulPropTag   = PR_MAP_NEXTID;
    srNewRows.aRow[0].lpProps[MAP_NEXTID].Value.l     = ulID + 1;

    srNewRows.aRow[0].cValues = NUM_MAP_PROPS;

    // this row has data for this name mapping)
    if (FAILED(hRes = gAllocBuff(sizeof SPropValue * NUM_MAP_PROPS, 
                                 (LPVOID *) &srNewRows.aRow[1].lpProps)))
    {
        gFreeBuff(srNewRows.aRow[1].lpProps);
        goto Quit;
    }

    // UID of our property set
    srNewRows.aRow[1].lpProps[MAP_UID].ulPropTag     = PR_MAP_UID;
    srNewRows.aRow[1].lpProps[MAP_UID].Value.bin.cb  = sizeof MAPIUID;
    srNewRows.aRow[1].lpProps[MAP_UID].Value.bin.lpb = (LPBYTE) pName -> lpguid;

    // ID for this row
    srNewRows.aRow[1].lpProps[MAP_ID].ulPropTag      = PR_MAP_ID;
    srNewRows.aRow[1].lpProps[MAP_ID].Value.l        = ulID;

    // store Unicode string name
    srNewRows.aRow[1].lpProps[MAP_SZNAME].ulPropTag   = PR_MAP_SZNAME;
    srNewRows.aRow[1].lpProps[MAP_SZNAME].Value.lpszW = 
        MNID_ID == pName -> ulKind ? L"" : pName -> Kind.lpwstrName;

    // store numeric name
    srNewRows.aRow[1].lpProps[MAP_ULNAME].ulPropTag   = PR_MAP_ULNAME;
    srNewRows.aRow[1].lpProps[MAP_ULNAME].Value.l     =
        MNID_ID == pName -> ulKind ? pName -> Kind.lID : 0;
    
    // save the ulKind info: string or numeric
    srNewRows.aRow[1].lpProps[MAP_KIND].ulPropTag     = PR_MAP_KIND;
    srNewRows.aRow[1].lpProps[MAP_KIND].Value.l       = pName -> ulKind;
    
    srNewRows.aRow[1].cValues = NUM_MAP_PROPS;

    if (SUCCEEDED(hRes = m_pLogon   -> 
                         m_pMapTable -> HrModifyRows(0,(LPSRowSet)&srNewRows)))
        *pulID = ulID;

Quit:
    gFreeBuff(srNewRows.aRow[0].lpProps);
    gFreeBuff(srNewRows.aRow[1].lpProps);
    return hRes;

}

///////////////////////////////////////////////////////////////////////////////
//
//    GetAllIDs()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//  

STDMETHODIMP CBaseProp :: GetAllIDs(LPSPropTagArray *ppPropTags)
{
    HRESULT         hRes,
                    hr      = S_OK;
    LPSRowSet       psr     = NULL;
    LPSPropTagArray pspt    = NULL;
    int             i;

    if (FAILED(hRes = HrQueryAllRows(m_pLogon -> m_pMapView,
                                     (LPSPropTagArray)&g_sptMAPTags,
                                     NULL,
                                     NULL,
                                     0,
                                     &psr)))
        goto Quit;
        
    if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(psr -> cRows),(LPVOID*) &pspt)))
        goto Quit;

    pspt -> cValues = psr -> cRows;

    for (i = 0; i < (int) psr -> cRows; i++)
    {
        if (PR_MAP_ID != psr -> aRow[0].lpProps[MAP_ID].ulPropTag)
        {
            hr = MAPI_W_ERRORS_RETURNED;
            pspt -> aulPropTag[i] = PROP_TAG(PT_ERROR,0);
            continue;
        }
        
        pspt -> aulPropTag[i] = PROP_TAG(PT_UNSPECIFIED, 
                                         psr -> aRow[i].lpProps[MAP_ID].Value.l); 
    }

    *ppPropTags = pspt;
Quit:
    
    FreeProws(psr);        
    return hRes == S_OK ? hr : hRes;

}

///////////////////////////////////////////////////////////////////////////////
//
//    GetIDsFromNames()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//  

STDMETHODIMP CBaseProp :: GetIDsFromNames(ULONG             cPropNames, 
                                          LPMAPINAMEID    * lppPropNames,       
                                          ULONG             ulFlags,    
                                          LPSPropTagArray * lppPropTags)
{
    HRESULT         hRes,
                    hr = S_OK;
    LPSPropTagArray pspt = NULL;
    LPSRowSet       psr = NULL;
    BOOL            bNewMapping = FALSE;
    SRestriction    sr[3];
    SPropValue      spv[2];
    int             i;

    TraceMessage("CBaseProp::GetIDsFromNames");
    Validate_IMAPIProp_GetIDsFromNames(this,
                                       cPropNames,  
                                       lppPropNames,
                                       ulFlags,     
                                       lppPropTags); 


    // create a view shared by all instances
    if (! m_pLogon -> m_pMapView)
        if (M_FAIL(hRes = m_pLogon -> GetMapTable()))
            goto Quit;

    // return the whole mapping
    if (!lppPropNames)
        return GetAllIDs(lppPropTags);

    if (M_FAIL(hRes = gAllocBuff(CbNewSPropTagArray(cPropNames),(LPVOID*) &pspt)))
        goto Quit;

    pspt -> cValues = cPropNames;

    sr[0].rt = RES_AND;
    sr[0].res.resAnd.cRes = 2;
    sr[0].res.resAnd.lpRes = &sr[1];

    sr[1].rt = RES_PROPERTY;
    sr[1].res.resProperty.relop     = RELOP_EQ;
    sr[1].res.resProperty.ulPropTag = PR_MAP_UID;
    sr[1].res.resProperty.lpProp    = &spv[0];

    // look up the name in our mapping table
    for (i = 0; i < (int) cPropNames; i++)
    {
        ULONG   ulID = 0;
        LPMAPINAMEID pName = lppPropNames[i];

        //!! if guid == PS_MAPI just return ID
        // construct a restriction based on UID and name
        spv[0].ulPropTag     = PR_MAP_UID;
        spv[0].Value.bin.cb  = sizeof GUID;
        spv[0].Value.bin.lpb = (LPBYTE) pName -> lpguid;

        sr[2].rt = RES_PROPERTY;
        sr[2].res.resProperty.relop     = RELOP_EQ;
        sr[2].res.resProperty.lpProp    = &spv[1];

        if (MNID_ID == pName -> ulKind)                 // numeric name
        {
            sr[2].res.resProperty.ulPropTag = PR_MAP_ULNAME;
            spv[1].ulPropTag = PR_MAP_ULNAME;
            spv[1].Value.l = pName -> Kind.lID;
        }   
        else
        {                                               // Unicode name
            sr[2].res.resProperty.ulPropTag = PR_MAP_SZNAME;
            spv[1].ulPropTag = PR_MAP_SZNAME;
            spv[1].Value.lpszW = pName -> Kind.lpwstrName;
        }

        EnterCriticalSection(&m_pLogon -> m_MapCS);

        if (FAILED(hRes = m_pLogon   ->
                          m_pMapView -> FindRow(&sr[0],BOOKMARK_BEGINNING,0)))
        {
            if (MAPI_E_NOT_FOUND == hRes && (ulFlags & MAPI_CREATE))
            {
                // Add this mapping
                hRes = NewMapping(pName,&ulID);
                bNewMapping = TRUE;
            }
        }
        else
        {
            if (SUCCEEDED(hRes = m_pLogon   ->
                                 m_pMapView -> QueryRows(1, 0, &psr)))
            {
                ulID = psr -> aRow[0].lpProps[MAP_ID].Value.l;
                FreeProws(psr);
                psr = NULL;
            }    

        }
        LeaveCriticalSection(&m_pLogon -> m_MapCS);

        if (SUCCEEDED(hRes))
            pspt -> aulPropTag[i] = PROP_TAG(PT_UNSPECIFIED, ulID); 
        else
        {
            hr = MAPI_W_ERRORS_RETURNED;
            pspt -> aulPropTag[i] = PROP_TAG(PT_ERROR,0);
        }
    }
    
    if (bNewMapping)
    {
        ABPAB_ID   NameMapID = {NAME_MAP_ID,0};
        DECLARE_REC_NUM(NameIDRec);

        FreeProws(psr);
        psr = NULL;

        // save new map table to database
        if (FAILED(hRes = Open(m_pLogon-> m_pDB,OPEN_ONLY,NameMapID, NameIDRec)))
            goto Quit;

        EnterCriticalSection(&m_pLogon -> m_MapCS);

        if (SUCCEEDED(hRes = HrQueryAllRows(m_pLogon -> m_pMapView,
                                            NULL,
                                            NULL,
                                            NULL,
                                            0,
                                            &psr)))
        {
            hRes = Update(NameIDRec,fldTABLEROWS,(LPSRowSet)psr);
        }
        
        LeaveCriticalSection(&m_pLogon -> m_MapCS);
        Close(NameIDRec);
    }

Quit:
    FreeProws(psr);

    if (FAILED(hRes))
    {
        hr = hRes;
        gFreeBuff(pspt);
    }

    *lppPropTags = pspt;
    return hr;
}

///////////////////////////////////////////////////////////////////////////////
//
//    NotifyParents()
//
//    Parameters                        
//          MyRecNum        - reference to record number of 
//                            object whose parents we notify
//          ulEvent         - table event type
//          pFlatPropList   - FlatPropList containing the modified
//                            row in the parent's contents table for
//                            this child
//    Purpose
//          Notify this container's parents that the object has changed and
//          send them a list of properties they can put in their contents
//          tables for this child.
//
//          We use a table event because then we can pass an SRow for the
//          child's row in contents table to its parent. This is slightly
//          backwards: the child isn't a table, the table is in the parent.
//          If we use an extended notification, we have to flatten the row
//          into a contiguous blob. If we pass a table notification, we    
//          can pass the row and let MAPI worry about copying the row data
//          in Notify()
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CBaseProp :: NotifyParents(RECORD_NUMBER & MyRecNum,
                                        ULONG           ulEvent,
                                        PCFlatPropList  pFlatPropList)
{
    HRESULT             hRes = S_OK;
    CList<ABPAB_ID>     ParentIDList;
    ABPAB_ID          * pID;
    SizedNOTIFKEY(sizeof ABPAB_EID,Key);
    PABPAB_EID          pEID = (PABPAB_EID) Key.ab;
    NOTIFICATION        Notif;
    ULONG               ulFlags;
    LPSPropValue        pCTblRow = NULL;
    SRow                srChanged;

    // convert my cached flat props into a contents table row
    // that we can pass to parent in the notification
    if (FAILED(hRes = m_pLogon -> GetCTblRow(pFlatPropList,
                                             &pCTblRow)))
        goto Quit;                                    

    srChanged.cValues = NUM_CTBL_COLS;
    srChanged.lpProps = pCTblRow;

    ZeroMemory(&Notif,sizeof Notif);
    Notif.ulEventType           = fnevChildEvent;
    Notif.info.tab.ulTableEvent = ulEvent; 
    Notif.info.tab.propIndex    = pCTblRow[CT_INST_KEY];
    Notif.info.tab.row          = srChanged;
  
    if (FAILED(hRes = Fetch(MyRecNum, 
                            fldPARENTS,
                            &ParentIDList)))
        goto Quit;

    if (ParentIDList.Size())
    {     
        for (pID = ParentIDList.Scan(0);
             pID;
             pID = ParentIDList.Scan())
        {
            Key.cb = sizeof ABPAB_EID;
            // Create the parent object's EID on the fly.
            // The only unique fields per object are the
            // the object type (we don't care) and the ID.
            // The muid is the same for all objects in this instance,
            // so copy our muid, get the ID from the parent list,
            // and DONT_CARE the obj type
            memcpy((LPVOID) Key.ab, &m_MyEID, sizeof ABPAB_EID);
            pEID -> ID.Type = DONT_CARE;                
            pEID -> ID.Num  = pID -> Num;

            ulFlags = 0;
            hRes = m_pLogon -> m_pSupObj -> Notify((LPNOTIFKEY) &Key,
                                                   1,
                                                   &Notif,
                                                   &ulFlags);
            if (FAILED(hRes))
                goto Quit;
        }
    }

Quit:
    // MAPI copies the notification fields
    gFreeBuff(pCTblRow);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SaveForeignDetails()
//
//    Parameters                        
//          pDB             [in] pointer to database
//          ID              [in] ID of host object
//          pRemoteObj      [in] pointer to an object native to another AB
//          pHostObj        [in] pointer to host IMAPIProp
//          ppTbl           [out]pointer to pointer with cached details table
//
//    Purpose
//          Gets a foreign DL's details table, extracts the rows and
//          saves them to our database. Later we can retrieve them to
//          construct a details table for the foreign entry.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP SaveForeignDetails(PDATABASE        pDB,
                                const ABPAB_ID & ID,
                                LPMAPIPROP       pRemoteObj,
                                LPMAPIPROP       pHostObj,
                                LPMAPITABLE    * ppTbl)
{
    HRESULT         hRes   = S_OK;
    LPSRowSet       pRows  = NULL;
    int             i;
    CList<ULONG *>  PropTagList;
    LPSPropTagArray pTags = NULL;
    LPMAPINAMEID   * ppNames = NULL;
  
    DECLARE_REC_NUM(RecNum);
    
    enum {XPOS,YPOS,DELTAX,DELTAY,TYPE,FLAGS,STRUCT,CTL_ID,ROW_ID,NUM_COLS};
    SizedSPropTagArray(NUM_COLS,sptCols) = {NUM_COLS,{PR_XPOS,
                                                      PR_YPOS,
                                                      PR_DELTAX,
                                                      PR_DELTAY,
                                                      PR_CONTROL_TYPE,
                                                      PR_CONTROL_FLAGS,
                                                      PR_CONTROL_STRUCTURE,
                                                      PR_CONTROL_ID,
                                                      PR_ROWID}};

    // get the foreign entry's details table
    if (FAILED(hRes = pRemoteObj -> OpenProperty(PR_DETAILS_TABLE,
                                                 &IID_IMAPITable,
                                                 0,
                                                 0,
                                                (LPUNKNOWN *) ppTbl)))
        goto Quit;
    
    if (FAILED(hRes = HrQueryAllRows(*ppTbl,
                                     (LPSPropTagArray)&sptCols,
                                     NULL,
                                     NULL,
                                     0,
                                     &pRows)))
        goto Quit;

    for (i = 0; i < (int)pRows -> cRows; i++)
    {
        // check if PR_ROWID is there, if not (e.g. MSMAIL) add ours
        if (PR_ROWID != pRows -> aRow[i].lpProps[ROW_ID].ulPropTag)
        {
            pRows -> aRow[i].lpProps[ROW_ID].ulPropTag = PR_ROWID;
            pRows -> aRow[i].lpProps[ROW_ID].Value.l = i;
        }

        // MSMAIL doesn't set include this requried property
        if (PR_CONTROL_ID != pRows -> aRow[i].lpProps[CTL_ID].ulPropTag)
        {
            pRows -> aRow[i].lpProps[CTL_ID].ulPropTag = PR_NULL;
            pRows -> aRow[i].lpProps[CTL_ID].Value.l = 0;
        }
        // check each control for named properties. if a named property is
        // found, save a pointer to the tag in the DTCTL 
        switch(pRows -> aRow[i].lpProps[TYPE].Value.l)
        {
            case DTCT_CHECKBOX:
            {
                LPDTBLCHECKBOX  pCtl;
                pCtl = (LPDTBLCHECKBOX) pRows -> aRow[i].lpProps[STRUCT].Value.bin.lpb;

                if (PROP_ID(pCtl -> ulPRPropertyName) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPRPropertyName));

                break;
            }
            case DTCT_DDLBX:
            {
                LPDTBLDDLBX pCtl = (LPDTBLDDLBX) pRows -> aRow[i].lpProps[STRUCT].Value.bin.lpb;

                if (PROP_ID(pCtl -> ulPRSetProperty) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPRSetProperty));

                if (PROP_ID(pCtl -> ulPRTableName) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPRTableName));

                break;
            }
            case DTCT_EDIT:
            {
                LPDTBLEDIT pCtl = (LPDTBLEDIT) pRows -> aRow[i].lpProps[STRUCT].Value.bin.lpb;

                if (PROP_ID(pCtl -> ulPropTag) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPropTag));
                
                break;
            }
            case DTCT_LBX:
            {
                LPDTBLLBX pCtl = (LPDTBLLBX) pRows -> aRow[i].lpProps[STRUCT].Value.bin.lpb;

                if (PROP_ID(pCtl -> ulPRSetProperty) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPRSetProperty));
                
                break;
            }
           // DTCT_MVDDLBX  
            case DTCT_RADIOBUTTON:
            {
                LPDTBLRADIOBUTTON  pCtl = (LPDTBLRADIOBUTTON) pRows -> aRow[i].lpProps[STRUCT].Value.bin.lpb;

                if (PROP_ID(pCtl -> ulPropTag) > 0x7FFF)
                    PropTagList.Insert(&(pCtl -> ulPropTag));
                
                break;
            }

            case DTCT_LABEL :      
            case DTCT_COMBOBOX :
            case DTCT_GROUPBOX :   
            case DTCT_BUTTON :      
            case DTCT_PAGE :        
            case DTCT_MVLISTBOX :   
            case DTCT_MVDDLBX :
                break;//


            default:


                ;// the rest either have no proptag or are PT_OBJECT
        }
    }

    i = PropTagList.Size();

    if (i)  // any named props in the controls?
    {
        // Convert the list of pointers to ULONGs in the DTCTL blobs
        // to an SPropTagArray we can pass to ReMap.

        ULONG ** ppPropTag;

        if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(i),
                                     (LPVOID *) &pTags))) 
            goto Quit;

        // convert linked list to SPropTagArray
        for (ppPropTag = PropTagList.Scan(0),pTags -> cValues = i = 0;
             ppPropTag;
             ppPropTag = PropTagList.Scan())
        {
            pTags -> aulPropTag[i] = **ppPropTag;
            pTags -> cValues = ++i;
        }
        
        // Map named props
        if (FAILED(hRes = pRemoteObj -> GetNamesFromIDs(&pTags,
                                                        NULL,
                                                        MAPI_NO_STRINGS,
                                                        (ULONG *) &i,
                                                        (LPMAPINAMEID **)&ppNames)))
            goto Quit;

        gFreeBuff((LPVOID) pTags);    
        pTags = NULL;

        if (FAILED(hRes = pHostObj -> GetIDsFromNames(i,
                                                      ppNames,
                                                      MAPI_CREATE,
                                                      &pTags)))
            goto Quit;

        // replace foreign ID with ID from our mapping
        for (ppPropTag = PropTagList.Scan(0), i = 0;
             ppPropTag;
             ppPropTag = PropTagList.Scan())
        {
            SWAP_IDS(**ppPropTag, pTags -> aulPropTag[i]);
        }
    }

    // save it all back to database
    if (SUCCEEDED(hRes = Open(pDB, 
                              OVERWRITE,
                              ID,
                              RecNum)))
    {
        hRes = Update(RecNum, fldTABLEROWS,pRows);
        Close(RecNum);
    }
Quit:

    gFreeBuff((LPVOID) pTags);    
    gFreeBuff(ppNames);
    
    if (FAILED(hRes))
	{
        if (*ppTbl)
            (*ppTbl) -> Release();
	 
	 	*ppTbl = NULL;
	}

    FreeProws(pRows);
    return hRes;    
}
