///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      IABCONT.CPP
//
//  Description
//      IABContainer, IDistList implementation.
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
//
#include <mapiutil.h>
#include "abpab.h"
#include "abcache.h"
#include "cdbfile.h"
#include "abdb.h"

#include "itable.h"
#include "list.inl"

#define ONEOFF_ENTRY  0x4000
#define ONEOFF_USER_ID 0xfffffffd
#define ONEOFF_DL_ID   0xfffffffc

///////////////////////////////////////////////////////////////////////////////
//
//    HrGetView()
//
//    Parameters                        
//          pTblData         [in]  pointer to underlying table data
//          lpSSortOrderSet  [in]  view sort order
//          lpfCallerRelease [in]  view release callback
//          ulCallerData     [in]  data to call callback with
//          lppMAPITable     [out] pointer to pointer where view is
//                                 returned
//    Purpose
//          Returns a view of our wrapper table implementation.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: HrGetView(LPTABLEDATA         pTblData,
                                       LPSSortOrderSet     lpSSortOrderSet, 
                                       CALLERRELEASE FAR * lpfCallerRelease, 
                                       ULONG               ulCallerData, 
                                       LPMAPITABLE FAR   * lppMAPITable)
{
    HRESULT      hRes = S_OK;
    CMAPITable * pTbl = NULL;
    LPMAPITABLE  pTblTemp = NULL;

    if (FAILED(hRes = pTblData -> HrGetView(lpSSortOrderSet,
                                            lpfCallerRelease,
                                            ulCallerData,
                                            &pTblTemp)))
        return hRes;
    
    pTbl = new CMAPITable(pTblTemp,m_MyEID.ID.Num);

    if (!pTbl)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }
    
    pTbl -> AddRef();
    *lppMAPITable = (LPMAPITABLE) pTbl;
    return hRes;

Quit:

    if (pTblTemp)
        pTblTemp -> Release();

    if (pTbl)
        pTbl -> Release();
    
    return hRes;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Sets detailed error information based on hRes.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: SetLastError(HRESULT       hRes,    
                                          ULONG         ulLine,    
                                          LPSTR         szFile)
{
    if (FAILED(hRes))
    {
        EnterCriticalSection(&m_cs);
        m_ulContext = ulLine + IE_IABLOGON;
        lstrcpyn(m_szError,szFile,MAX_PATH);
        m_szError[MAX_PATH - 1] = '\0';
        LeaveCriticalSection(&m_cs);
    }
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    GetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Gets detailed error information based on hResult.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: GetLastError(HRESULT        hResult,
                                          ULONG          ulFlags,                    
                                          LPMAPIERROR  * ppMAPIError)
{
    HRESULT hRes;
    TraceString1("CABContainer <%x> :: GetLastError", m_MyEID.ID.Num);
    Validate_IMAPIProp_GetLastError(this,
                                    hResult,    
                                    ulFlags,    
                                    ppMAPIError);
    EnterCriticalSection(&m_cs);
    hRes =::GetLastError(m_szError,SZ_PSS_SAMPLE_AB,m_ulContext,ppMAPIError);
    LeaveCriticalSection(&m_cs);
    return hRes;
}        


///////////////////////////////////////////////////////////////////////////////
//  
//      The following functions all delegate down to the CBaseProp
//      object. This is necessary because the OLE COM interfaces can't
//      be derived virtual. (e.g. class IABContainer : virtual public 
//      IMAPIContainer) so if we also derive from a MAPI base class, there 
//      will be ambiguities between our base IMAPIProp implementation and 
//      the abstract IMAPIProp base in IMAPIContainer

///////////////////////////////////////////////////////////////////////////////
//
//    SaveChanges()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  SaveChanges(ULONG  ulFlags)                        
{                                                                                     
    return CBaseProp::SaveChanges(ulFlags);             
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    GetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  GetProps(LPSPropTagArray  lpPropTagArray,                     
                      ULONG            ulFlags,                            
                      ULONG  *         lpcValues,                          
                      LPSPropValue  *  lppPropArray)                       
{                                                                                     
    return CBaseProp::GetProps(lpPropTagArray,           
                           ulFlags,                  
                           lpcValues,                
                           lppPropArray);             
}                                                                                     

///////////////////////////////////////////////////////////////////////////////
//
//    SetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: SetProps(ULONG                 cValues,    
                                      LPSPropValue          lpPropArray,
                                      LPSPropProblemArray * lppProblems)
{                                                                                     
    return CBaseProp::SetProps(cValues,lpPropArray, lppProblems);
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetPropList()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
                                                                                      
STDMETHODIMP  CABContainer ::  GetPropList(ULONG ulFlags,LPSPropTagArray * lppPropTagArray)
{                                                                                     
    return CBaseProp::GetPropList(ulFlags,lppPropTagArray);       
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    DeleteProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  DeleteProps(LPSPropTagArray       lpPropTagArray,                  
                                           LPSPropProblemArray * lppProblems)                     
{                                                                                     
    return CBaseProp::DeleteProps(lpPropTagArray,        
                                   lppProblems);
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    CopyTo()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  CopyTo(ULONG                 ciidExclude,                          
                                      LPCIID                rgiidExclude,                         
                                      LPSPropTagArray       lpExcludeProps,                       
                                      ULONG                 ulUIParam,                            
                                      LPMAPIPROGRESS        lpProgress,                           
                                      LPCIID                lpInterface,                          
                                      LPVOID                lpDestObj,                            
                                      ULONG                 ulFlags,                              
                                      LPSPropProblemArray * lppProblems)                          
{                                                                                     
    return CBaseProp::CopyTo(ciidExclude,                
                                 rgiidExclude,               
                                 lpExcludeProps,             
                                 ulUIParam,                  
                                 lpProgress,                 
                                 lpInterface,                
                                 lpDestObj,                  
                                 ulFlags,                    
                                 lppProblems);                
}                                                                                     

///////////////////////////////////////////////////////////////////////////////
//
//    CopyProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
                                                                                      
STDMETHODIMP  CABContainer ::  CopyProps(LPSPropTagArray       lpIncludeProps,                    
                                         ULONG                 ulUIParam,                         
                                         LPMAPIPROGRESS        lpProgress,                        
                                         LPCIID                lpInterface,                       
                                         LPVOID                lpDestObj,                         
                                         ULONG                 ulFlags,                           
                                         LPSPropProblemArray * lppProblems)                       
{                                                                                     
    return CBaseProp::CopyProps(lpIncludeProps,          
                                    ulUIParam,               
                                    lpProgress,              
                                    lpInterface,             
                                    lpDestObj,               
                                    ulFlags,                 
                                    lppProblems);             
}                                                                                        
                                                                                         
///////////////////////////////////////////////////////////////////////////////
//
//    GetNamesFromIDs()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  GetNamesFromIDs(LPSPropTagArray * lppPropTags,                 
                                               LPGUID            lpPropSetGuid,               
                                               ULONG             ulFlags,                     
                                               ULONG  *          lpcPropNames,                
                                               LPMAPINAMEID **   lpppPropNames)               
{                                                                                     
    return CBaseProp::GetNamesFromIDs(lppPropTags,       
                                          lpPropSetGuid,     
                                          ulFlags,           
                                          lpcPropNames,      
                                          lpppPropNames);     
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    GetIDsFromNames()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CABContainer ::  GetIDsFromNames(ULONG             cPropNames,                  
                                               LPMAPINAMEID    * lppPropNames,                
                                               ULONG             ulFlags,                     
                                               LPSPropTagArray * lppPropTags)                 
{                                                                                     
    return CBaseProp::GetIDsFromNames(cPropNames,        
                                          lppPropNames,      
                                          ulFlags,           
                                          lppPropTags);       
}                                                                                        

///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Returns an object when requested a supported property
//          of type PT_OBJECT. This overrides the base IMAPIProp's stub.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: OpenProperty(ULONG       ulPropTag,                  
                                          LPCIID      piid,                      
                                          ULONG       ulInterfaceOptions,         
                                          ULONG       ulFlags,                    
                                          LPUNKNOWN * ppUnk)
{
    HRESULT     hRes = S_OK;
    LPMAPITABLE pTbl = NULL;

    TraceString1("CABContainer <%x> :: OpenProperty",m_MyEID.ID.Num);
    Validate_IMAPIProp_OpenProperty(this,
                                    ulPropTag,
                                    piid,              
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    ppUnk);

     
    *ppUnk = NULL;

    EnterCriticalSection(&m_cs);
    
    switch (ulPropTag)
    {
        case PR_CREATE_TEMPLATES:
            if (IID_IMAPITable != *piid)
            {
                hRes = MAPI_E_INTERFACE_NOT_SUPPORTED;
                break;
            }

            hRes = m_pLogon -> OpenTemplates(0,&pTbl);
            *ppUnk = (LPUNKNOWN) pTbl;
            break;

        case PR_DETAILS_TABLE:
            // return details table of directory container or DL
            if (IID_IMAPITable != *piid)
            {
                hRes = MAPI_E_INTERFACE_NOT_SUPPORTED;
                break;
            }

            if (DIR_ID_NUM == m_MyEID.ID.Num)
                hRes = m_pLogon -> m_pRootDetails -> BuildDisplayTable(&pTbl);
            else                
                hRes = m_pLogon -> m_pDLDetails -> BuildDisplayTable(&pTbl);
            *ppUnk = (LPUNKNOWN) pTbl;
            break;

        case PR_CONTAINER_CONTENTS:
            if (IID_IMAPITable != *piid)
            {
                hRes = MAPI_E_INTERFACE_NOT_SUPPORTED;
                break;
            }

            hRes = GetContentsTable(ulInterfaceOptions, &pTbl);
            *ppUnk = (LPUNKNOWN) pTbl;            
            break;

        case PR_EDIT_BUTTON: 
            {
                CMembersButton * pDLMembersButton = new CMembersButton(this);
 
                if (! pDLMembersButton)
                    hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                else
                {
                    *ppUnk = (LPUNKNOWN) pDLMembersButton;
                    (*ppUnk) -> AddRef();
                }
            }
            break;

        default:
            hRes = MAPI_E_NO_SUPPORT;
            break;
    }
    
    LeaveCriticalSection(&m_cs);
    return hRes;
}              


///////////////////////////////////////////////////////////////////////////////
//
//    CreateCTbl()
//
//    Parameters                        
//          ppTable     - out param where we return the contents table
//                        view.
//          RecNum      - reference to the record number for the container
//                        whose contents table we want
//    Purpose
//          Gets a container's list of children, then opens each child 
//          in turn, gets its properties and extracts the contents table's
//          column set from them, creates a table row and adds the row
//          to the contents table's underlying data. Returns a view of
//          the constructed contents table and caches a pointer to the
//          underlying data in the object.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: CreateCTbl(LPMAPITABLE   * ppTable,
                                        RECORD_NUMBER & RecNum)
{
    LPMAPITABLE         pCTbl      = NULL;   
    HRESULT             hRes       = S_OK;
    PFlatPropNode       pFlatNode  = NULL;
    CPropCache        * pPropCache = NULL;        
    LPTABLEDATA         pTblData = NULL;
    SRow                sr;
    ABPAB_ID          * pChildID;
    ABPAB_EID           EIDChild;
    CList<ABPAB_ID>     ChildIDList; 
    CFlatPropList       FlatPropList;

    DECLARE_REC_NUM(ChildRecNum); 
    
    if (FAILED(hRes = Fetch(RecNum, fldCHILDREN, &ChildIDList)))
        goto Quit;
        
    if (FAILED(hRes = CreateTable(&IID_IMAPITableData,
                                  gAllocBuff,
                                  gAllocMore,
                                  gFreeBuff,
                                  NULL,
                                  TBLTYPE_DYNAMIC,
                                  PR_INSTANCE_KEY,
                                  (LPSPropTagArray)&g_sptCntTbl,
                                  &pTblData)))
        return hRes;

    // get each child's row data
    if (ChildIDList.Size())
    {     

        for (pChildID = ChildIDList.Scan(0);
             pChildID;
             pChildID = ChildIDList.Scan())
        {
            
            EIDChild.ID = *pChildID;
            sr.cValues = NUM_CTBL_COLS;
            sr.lpProps = NULL;

            if (SUCCEEDED(hRes = Open(m_pLogon -> m_pDB, 
                                      OPEN_ONLY,
                                      EIDChild.ID,
                                      ChildRecNum)))
            {
                FlatPropList.Destroy();
                hRes = Fetch(ChildRecNum, fldPROPS, &FlatPropList);
                Close(ChildRecNum);

                if (SUCCEEDED(hRes))  // don't need this for dir, it hasn't any props on creation and root doesn't need a contents table
                    if (SUCCEEDED(hRes = m_pLogon -> GetCTblRow(&FlatPropList, &sr.lpProps)))
                        hRes = pTblData -> HrModifyRow(&sr); 
            }            

            gFreeBuff(sr.lpProps);
        }          
    }    
    
    // return a view
    if (ppTable) 
    {
        if (FAILED(hRes = HrGetView(pTblData,NULL,NULL,0,&pCTbl)))
            goto Quit;
        *ppTable = pCTbl;
    }

    // cache the underlying data        
    if (m_pCntsTblData)
        m_pCntsTblData -> Release();

    m_pCntsTblData = pTblData;
    return hRes;

Quit:
    if (pTblData)
        pTblData -> Release();

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetContentsTable()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details. If we already have
//          the underlying data in our cache, return a view, otherwise
//          create the underlying data, cache it and return a view. 
//          Contents tables are sorted ascending on PR_DISPLAY_NAME
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: GetContentsTable(ULONG         ulFlags,  
                                              LPMAPITABLE * ppTable)
{

    HRESULT         hRes   = S_OK;
    LPMAPITABLE     pTable = NULL;
    
    DECLARE_REC_NUM(RecNum);
    SizedSSortOrderSet(1,SortOrder) = {1,0,0,{PR_DISPLAY_NAME,TABLE_SORT_ASCEND}};

    TraceString1("CABContainer <%x> :: GetContentsTable", m_MyEID.ID.Num);
    Validate_IMAPIContainer_GetContentsTable(this,
                                             ulFlags, 
                                             ppTable); 

    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    *ppTable = NULL;

    EnterCriticalSection(&m_cs);

    if (m_pCntsTblData)
    { 
        if (M_FAIL(hRes = HrGetView(m_pCntsTblData,NULL,NULL,0,&pTable)))
            goto Quit;
    }
    else 
    {   
        if (FAILED(hRes = Open(m_pLogon -> m_pDB, 
                               OPEN_ONLY,   
                               m_MyEID.ID,
                               RecNum)))
        {   // create an empty m_pCntsTbl (TAD) for one-off
            if (MAPI_E_NOT_FOUND == hRes)
            {
                Close(RecNum);

                if (M_FAIL(hRes = CreateTable(&IID_IMAPITableData,
                                              gAllocBuff,
                                              gAllocMore,
                                              gFreeBuff,
                                              NULL,
                                              TBLTYPE_DYNAMIC,
                                              PR_INSTANCE_KEY,
                                              (LPSPropTagArray)&g_sptCntTbl,
                                              &m_pCntsTblData)))
                    goto Quit;

                if (M_FAIL(hRes = HrGetView(m_pCntsTblData,NULL,NULL,0,&pTable)))
                    goto Quit;
            }
        }
        else
        {   // create m_pCntsTbl and populate from DB
            hRes = CreateCTbl(&pTable, RecNum);   // side effect is to update m_pCntsTblData
            Close(RecNum);

            if (M_FAIL(hRes))
                goto Quit;
        }
    }

    if (M_FAIL(hRes = pTable -> SortTable((LPSSortOrderSet)&SortOrder,0)))
        goto Quit;

    *ppTable = pTable;
    
Quit:
    LeaveCriticalSection(&m_cs);
    if (FAILED(hRes))
        if (pTable)
            pTable -> Release();

    return hRes;
}   

///////////////////////////////////////////////////////////////////////////////
//
//    CreateRootHTbl()
//
//    Parameters                        
//          ppHTbl  - pointer to the returned view
//
//    Purpose
//          Creates the underlying data for the root container's 
//          hierarchy table, caches it,and returns a view.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: CreateRootHTbl(LPMAPITABLE *ppHTbl)
{
  
    LPMAPITABLE   pHTbl     = NULL;
    LPTABLEDATA   pTblData  = NULL;
    HRESULT       hRes      = S_OK;
    LPSPropValue  pProps    = NULL;
    ABPAB_ID      ID        = {DIR_ID_NUM,MAPI_ABCONT}; 
    ULONG         i;
    SRow          sr;
    CFlatPropList fplDirProps;
    DECLARE_REC_NUM(RecNum);

    if (FAILED(hRes = gAllocBuff(sptHierTbl.cValues * (sizeof SPropValue),
                                   (LPVOID *) &pProps)))
        goto Quit;

    // ROOT has a single container, DIR, Open DIRs record
    if (FAILED(hRes = Open(m_pLogon -> m_pDB,
                           OPEN_ONLY,
                           ID, 
                           RecNum)))
        goto Quit;

    // get DIRs properties
    if (FAILED(hRes = Fetch(RecNum, fldPROPS, &fplDirProps)))        
        goto Quit;

    for (i = 0; i < sptHierTbl.cValues; i++)
    {
        PFlatPropNode   pFlatPropNode = NULL;
        
        pFlatPropNode = fplDirProps.Find(sptHierTbl.aulPropTag[i]);        

        if (pFlatPropNode)
        {
            hRes = UnFlatten(pFlatPropNode,
                             pProps,
                             pProps[i]);
        }
        else
            hRes = MAPI_E_NOT_FOUND;
        
        if (hRes)
            hRes = HrPropNotFound(pProps[i],sptHierTbl.aulPropTag[i]);
    }

    // create a table
    if (FAILED(hRes = CreateTable (&IID_IMAPITableData,
                                   gAllocBuff,
                                   gAllocMore,
                                   gFreeBuff,
                                   NULL,
                                   TBLTYPE_DYNAMIC,
                                   PR_INSTANCE_KEY,
                                   (LPSPropTagArray)&sptHierTbl,
                                   &pTblData)))
        goto Quit;

    sr.cValues = NUM_HTBL_PROPS;
    sr.lpProps = pProps;

    // populate the table with DIR's properties, voila, an HTbl
    if (FAILED(hRes = pTblData -> HrModifyRow (&sr)))
        goto Quit;        

    // we can return a view if asked to
    if (ppHTbl)
        hRes = HrGetView(pTblData,NULL,NULL,0,&pHTbl);
Quit:
    Close(RecNum);
    gFreeBuff((LPVOID) pProps);

    if (FAILED(hRes))
    {
        if (pTblData)
            pTblData -> Release();

        if (pHTbl)
            pHTbl -> Release();
        
        return hRes;
    }
            
    if (ppHTbl)
        *ppHTbl = pHTbl;

    m_pHierTblData = pTblData;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetHierarchyTable()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details. Returns a view of
//          the cached hierarchy table data if it exists. If it doesn't
//          the table data is created and cached first. 
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: GetHierarchyTable(ULONG         ulFlags,  
                                               LPMAPITABLE * ppTable)
{

    HRESULT         hRes   = S_OK;
    LPMAPITABLE     pTable = NULL;

    TraceString1("CABContainer <%x> :: GetHierarchyTable", m_MyEID.ID.Num);
    Validate_IMAPIContainer_GetHierarchyTable(this,
                                              ulFlags,
                                              ppTable);
    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    *ppTable = NULL;

    if (ROOT_ID_NUM != m_MyEID.ID.Num)
        return MAPI_E_NO_SUPPORT;

    EnterCriticalSection(&m_cs);

    if (m_pHierTblData)
    { 
        if (FAILED(hRes = HrGetView(m_pHierTblData,NULL,NULL,0,&pTable)))
            goto Quit;
    }
    else 
        if (FAILED(hRes = CreateRootHTbl(&pTable)))
            goto Quit;

    *ppTable = pTable;
    
Quit:
    LeaveCriticalSection(&m_cs);

    if (FAILED(hRes))
        if (pTable)
            pTable -> Release();

    return hRes;

}   

///////////////////////////////////////////////////////////////////////////////
//
//    OpenEntry()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          MAPI_E_NO_SUPPORT
//
STDMETHODIMP CABContainer :: OpenEntry(ULONG       cbEntryID,
                                       LPENTRYID   pEntryID,
                                       LPCIID      pInterface,   
                                       ULONG       ulFlags,  
                                       ULONG     * pulObjType,   
                                       LPUNKNOWN * ppUnk)
{
    TraceString1("CABContainer <%x> :: OpenEntry", m_MyEID.ID.Num);
    Validate_IMAPIContainer_OpenEntry(this,
                                      cbEntryID, 
                                      pEntryID,  
                                      pInterface,
                                      ulFlags,   
                                      pulObjType,
                                      ppUnk);     
    return MAPI_E_NO_SUPPORT;
} 

///////////////////////////////////////////////////////////////////////////////
//
//    SetSearchCriteria()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: SetSearchCriteria(LPSRestriction pRestriction, 
                                               LPENTRYLIST    pContainerList,   
                                               ULONG          ulSearchFlags)
{
    TraceString1("CABContainer <%x> :: SetSearchCriteria", m_MyEID.ID.Num);
    Validate_IMAPIContainer_SetSearchCriteria(this,
                                              pRestriction,  
                                              pContainerList,
                                              ulSearchFlags); 

    return MAPI_E_NO_SUPPORT;
}  

///////////////////////////////////////////////////////////////////////////////
//
//    GetSearchCriteria()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: GetSearchCriteria(ULONG            ulFlags,  
                                               LPSRestriction * ppRestriction,
                                               LPENTRYLIST    * ppContainerList,  
                                               ULONG          * pulSearchState) 
{
    TraceString1("CABContainer <%x> :: GetSearchCriteria", m_MyEID.ID.Num);
    Validate_IMAPIContainer_GetSearchCriteria(this,
                                              ulFlags,        
                                              ppRestriction,  
                                              ppContainerList,
                                              pulSearchState); 
    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    return MAPI_E_NO_SUPPORT;
}  

///////////////////////////////////////////////////////////////////////////
//
//    AddRef()
//
//    Parameters
//          See the OLE documentation for details.
//
//    Purpose
//      To increase the reference count of this object, and of any contained 
//      (or aggregated) object within
//
//    Return Value
//      Usage count for this object
//
STDMETHODIMP_ (ULONG) CABContainer::AddRef (void)
{   
    ULONG   ulRef;

    // Increment object's internal counter
    EnterCriticalSection(&m_cs);
    ulRef = ++m_cRef;
    TraceString2("CABContainer::AddRef ID %x, cRef %d",m_MyEID.ID.Num,m_cRef);
    LeaveCriticalSection(&m_cs);
       
    return ulRef;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Release()
//
//    Parameters
//          See the OLE documentation for details.
//
//    Purpose
//      Decreases the usage count of this object. When the count is zero,
//      the object deletes itself.
//
//    Return Value
//      Usage of this object
//
STDMETHODIMP_ (ULONG) CABContainer::Release(void)
{
    ULONG  ulRefCount;

    EnterCriticalSection(&m_cs);
    ulRefCount = --m_cRef;
    TraceString2("CABContainer::Release ID %x, cRef %d",
                 m_MyEID.ID.Num,
                 ulRefCount);
    LeaveCriticalSection(&m_cs); 

    // If nobody is using this object, delete it
    if (0 == ulRefCount)
    {
        // Destroy the object
        delete this;        // don't need to leave the critical section
        return ulRefCount;  // its deleted when the object is destroyed
    }
        
    return ulRefCount;
}

///////////////////////////////////////////////////////////////////////////////
//
//    QueryInterface()
//
//    Parameters                        
//          See the OLE documentation for details.
//
//    Purpose
//          Returns an interface if supported.
//      
//      
//    Return Value
//          NOERROR on success, E_NOINTERFACE if the interface
//          can't be returned.
//
STDMETHODIMP CABContainer::QueryInterface(REFIID   riid, 
                                          LPVOID * ppvObj)
{
    TraceString1("CABContainer <%x> :: AddRef",m_MyEID.ID.Num);
    
    if (IsBadWritePtr(ppvObj,sizeof LPUNKNOWN))
        return MAPI_E_INVALID_PARAMETER;

    *ppvObj = NULL;

    // If this is one of our IID's return an interface pointer
    if (riid == IID_IUnknown        || 
        riid == IID_IABContainer    ||
        riid == IID_IMAPIContainer  ||
        riid == IID_IMAPIProp       ||
        riid == IID_IDistList)
    {
        EnterCriticalSection(&m_cs);
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        m_cRef++;
        LeaveCriticalSection(&m_cs);
        return NOERROR;
    }
    return E_NOINTERFACE;
}


///////////////////////////////////////////////////////////////////////////////
//
//    CABContainer()
//
//    Parameters                        
//          pLogon              [in] pointer to the logon object
//          pCreateContainerEID [in] pointer to the container we were created int
//          pMyEID              [in] pointer to our entry ID
//          ulCreateFlags       [in] file creation flags
//    Purpose
//          Class constructor.
//      
//    Return Value
//          none.

CABContainer :: CABContainer(LPCABLogon pLogon, 
                             PABPAB_EID pCreateContainerEID,
                             PABPAB_EID pMyEID,
                             ULONG      ulCreateFlags,
                             LPCABCont  pParentCont,
                             BOOL       bNewEntry) : CBaseProp(pLogon,             
                                                                pCreateContainerEID,
                                                                pMyEID,             
                                                                ulCreateFlags,      
                                                                pParentCont,        
                                                                bNewEntry)
{
    PABPAB_EID      pEID = NULL;

    TraceString1("CABContainer <%x> :: CABContainer", 
                 pMyEID ? pMyEID -> ID.Num : m_MyEID.ID.Num);

    ASSERT((pMyEID -> ID.Num));

    m_pCntsTblData  = NULL;
    m_bCTblDirty    = TRUE;
    m_pHierTblData  = NULL;
    m_ulCnx         = 0;       
    m_szError[0]    = '\0';
    m_ulVersion     = 0;
    m_ulContext     = 0;
    m_pAdviseSink   = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Init()
//
//    Parameters                        
//          none
//
//    Purpose
//          Creates a new advisesink to listen for changes to this 
//          container's children
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: Init()
{
    HRESULT     hRes = S_OK;

    if (FAILED(hRes = CBaseProp::Init()))
        return hRes;
   
    m_pAdviseSink = new CContAdviseSink(this);    

    if (! m_pAdviseSink)
        return MAPI_E_NOT_ENOUGH_MEMORY;

    m_pAdviseSink -> AddRef();

    if (SUCCEEDED(hRes))    
        AddRef();

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    InitContainer()
//
//    Parameters                        
//          none
//
//    Purpose
//          Initializes the new container and adds it to the database in
//          a 'limbo' state, i.e. it has a record and streams for parents
//          properties and children, but its also in the orphan record list.
//          The child stays in the 'limbo' state until SaveChanges when it is
//          deleted from the orphan list. If SaveChanges isn't called, it stays
//          becomes an orphan and is deleted at Logoff. Containers (DLs) are
//          added this way since they can be root nodes of subtrees (i.e. they
//          can have children, grandchildren, etc) when members are added. The
//          members can be added before SaveChanges on the parent.Rather
//          than try to keep the subtree with the object, we write it to the
//          database as if it had been committed, and delete the whole thing later
//          if the user cancels.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: InitContainer()
{
    HRESULT     hRes = S_OK;
    DECLARE_REC_NUM(RecNum);

    if (FAILED(hRes = Init()))
        goto Quit;

    if (FAILED(hRes = Open(m_pLogon -> m_pDB,
                           OPEN_ONLY,
                           m_MyEID.ID, 
                           RecNum)))
    {
        if (hRes != MAPI_E_NOT_FOUND)
            goto Quit;

        if (FAILED(hRes = Open(m_pLogon -> m_pDB,
                               CREATE_IF_NOT_EXIST,
                               m_MyEID.ID, 
                               RecNum)))
            goto Quit;

        m_pLogon -> m_OrphanRecs.Insert(m_MyEID.ID);
    }
    
    // Do I need this Load????
    if (FAILED(hRes = m_pPropCache -> Load(RecNum)))
         if (FAILED(hRes = SetDefaultProps(&m_MyEID)))
             goto Quit;

    if (FAILED(hRes = CreateCTbl(NULL,RecNum)))
         goto Quit;
     
    // create a cnx so this object can listen 
    // for changes to its children
    hRes = Subscribe();
    
Quit:
    Close(RecNum);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CABContainer()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class destructor. Releases all outstanding objects.
//      
//    Return Value
//          none.

CABContainer :: ~CABContainer()
{
    TraceString1("CABContainer <%x> :: ~CABContainer", m_MyEID.ID.Num);
  
    // close internal notification connections
    if (m_ulCnx)                             
        m_pLogon -> m_pSupObj -> Unsubscribe(m_ulCnx);

    if (m_pCntsTblData)
        m_pCntsTblData -> Release();

    if (m_pHierTblData)
        m_pHierTblData -> Release();


    if (m_pAdviseSink)
        m_pAdviseSink -> Release();

}

///////////////////////////////////////////////////////////////////////////////
//
//    Subscribe()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Asks MAPI to send us notifications for a fnevChildEvent.This
//          is to update our contents table when a child is added/deleted
//          or changed. We use our entry ID as the key, it's unique, and 
//          our child can generate it easily by getting his parent list 
//          when he wishes to notify us.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: Subscribe()
{
    SizedNOTIFKEY(sizeof ABPAB_EID,Key);    
    PABPAB_EID   pEID = (PABPAB_EID) Key.ab;

    Key.cb = sizeof ABPAB_EID;
    memcpy((LPVOID)pEID, (LPVOID) &m_MyEID, sizeof ABPAB_EID);

    // Later when our child (e.g.) does a notify, it'll construct 
    // the entry ID on the fly knowing only the ID number. We won't 
    // know or care what the object type is
    pEID -> ID.Type = DONT_CARE;
    
    return m_pLogon -> m_pSupObj -> Subscribe((LPNOTIFKEY)&Key,
                                              fnevChildEvent,
                                              0,
                                              m_pAdviseSink,  
                                              &m_ulCnx);
}

///////////////////////////////////////////////////////////////////////////////
//
//    CreateEntry()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Creates a new entry in this container. A new entry can be a
//          mailuser, a DL (a container), a foreign DL or foreign mailuser.
//          The entry ID tells us whether the entry is foreign or native.      
//          Creates a new record in the database, It's in a limbo state,
//          (so we add to orphan rec list, until SaveChanges, then     
//          insert it in it's parent.                                   
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: CreateEntry(ULONG        cbEntryID,              
                                         LPENTRYID    lpEntryID,              
                                         ULONG        ulCreateFlags,          
                                         LPMAPIPROP * lppMAPIPropEntry)
{   
    HRESULT         hRes = S_OK;
    PABPAB_EID      pEID = (PABPAB_EID)lpEntryID,
                    pOOEID = NULL;
                
    ULONG           ulEntryType;

    *lppMAPIPropEntry = NULL;

    TraceString2("CABContainer <%x> :: CreateEntry<%x>", m_MyEID.ID.Num,pEID -> ID.Num);
    Validate_IABContainer_CreateEntry(this,
                                      cbEntryID,       
                                      lpEntryID,       
                                      ulCreateFlags,   
                                      lppMAPIPropEntry);

    ulEntryType = m_pLogon -> CheckEID(cbEntryID,lpEntryID);
    
    // convert fixed oneoff entry ID to meaninful database ID
    // this only applies if its one of our oneoffs
    if (ulEntryType & ONEOFF_ENTRY)
    {
        if (M_FAIL(hRes = m_pLogon -> CreateEID(ulEntryType,&pOOEID)))
            goto Quit;

        pEID = pOOEID;      // so we can free new EID later
    }

    switch (ulEntryType)
    {
        case MY_ABCONT:
            // don't suport containers other than ROOT, DIR
            hRes = MAPI_E_NO_SUPPORT;
            break;
        case MY_DISTLIST:
        {
            LPCABCont pCont = new CABContainer(m_pLogon,
                                               &m_MyEID,
                                               pEID,
                                               ulCreateFlags,
                                               this,
                                               TRUE);
            if (!pCont)
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                goto Quit;
            }                

            if (M_FAIL(hRes = pCont -> InitContainer()))
            {
                delete pCont;
                goto Quit;
            }
            else
                *lppMAPIPropEntry = pCont;

            break;                 
        }              
        case MY_MAILUSER:
        {
            PMailUser pMailUser = new CMailUser(m_pLogon, 
                                                &m_MyEID, 
                                                pEID,
                                                ulCreateFlags,
                                                this,
                                                TRUE);

            if (!pMailUser)
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                goto Quit;
            }                
                           
            if (M_FAIL(hRes = pMailUser -> Init()))
            {
                delete pMailUser;
                goto Quit;
            }
            else
                *lppMAPIPropEntry =  (LPMAPIPROP) pMailUser;

            break;
        }
        case FOREIGN_ENTRY :
        {
            LPMAPIPROP pProp = NULL;

            if (M_FAIL(hRes = CreateTemplateEntry(cbEntryID,   
                                                  lpEntryID,      
                                                  ulCreateFlags,  
                                                  &pProp)))
                goto Quit;
            else
                *lppMAPIPropEntry = pProp;
            break;
        }
        case FOREIGN_ENTRY | HAS_TEMPLATEID:  // will never happen, we won't know if it
            ASSERT(0);                        // supports PR_TEMPLATEID until we open it 
            break;
        default:            
            hRes = MAPI_E_INVALID_ENTRYID;        
    }

Quit:
    gFreeBuff(pOOEID);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    InitTemplateDL()
//
//    Parameters                        
//          pForeignObj     [in] pointer to the foreign DL object
//          pHostDL         [in] pointer to new local DL object
//    Purpose
//          Initializes a local copy of a foreign address book entry
//          that we want to add to our database. We set the default
//          properties, (we'll overwrite some later with the foreign entry's
//          props), get the foreigner's details table and save it,
//          then request notifications. 
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: InitTemplateDL(LPMAPIPROP     pForeignObj, 
                                            CTemplateDL *  pHostDL)
{
    HRESULT         hRes = S_OK;

    TraceString2("CABContainer <%x> :: InitTemplateDL %x", 
                 m_MyEID.ID.Num, 
                 pHostDL -> m_MyEID.ID.Num);

    if (FAILED(hRes = pHostDL -> Init()))
        goto Quit;

    if (FAILED(hRes = pHostDL -> SetDefaultProps(&pHostDL -> m_MyEID)))
        goto Quit;

    if (FAILED(hRes = pHostDL -> SetForeignDetails(pForeignObj)))
        goto Quit;

    hRes = pHostDL -> Subscribe();
    
Quit:
    if (FAILED(hRes))
    {
        if (pHostDL)
            delete pHostDL;
    }

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    InitTemplateUser()
//
//    Parameters                        
//          pForeignObj     [in] pointer to the foreign mailuser object
//          pHostDL         [in] pointer to new local mailuser object
//    Purpose
//          Initializes a local copy of a foreign address book entry
//          that we want to add to our database. We set the default
//          properties, (we'll overwrite some later with the foreign entry's
//          props), get the foreigner's details table and save it,
//          then request notifications. 
//      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: InitTemplateUser(LPMAPIPROP      pForeignObj, 
                                              CTemplateUser * pHostUser)
{
    HRESULT         hRes;
 
    TraceString2("CABContainer <%x> :: InitTemplateUser %x", 
                 m_MyEID.ID.Num, 
                 pHostUser -> m_MyEID.ID.Num);
    
    if (FAILED(hRes = pHostUser -> Init()))
        goto Quit;

    if (FAILED(hRes = pHostUser -> SetForeignDetails(pForeignObj)))
        goto Quit;

Quit:
    if (FAILED(hRes))
    {
        if (pHostUser)
            delete pHostUser;
    }

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CreateTemplateEntry()
//
//    Parameters                        
//          cbEntryID           [in]  byte count of entry ID of foreign object
//          lpEntryID           [in]  pointer to foreigner's entry ID
//          ulCreateFlags       [in]  creation flags
//          lppMAPIPropEntry    [out] pointer to newly created foreign entry
//    Purpose
//          Creates an entry in our AB that is native to another AB, i.e.
//          creates a foreign entry in our AB. The foreign entry is a DL 
//          object derived from CMAPIProp interface. It contains a pointer to 
//          a remoted IMAPIProp interface for the foreign object in it's
//          native AB. If the remote interface supports OpenTemplateID, we
//          can have calls to our object routed first to the remote interface.
//          Now calls to our object pass first to the remote interface which
//          decides which ones it wants to handle and which it delegates down
//          to our local interface. The remote interface can supply us with 
//          the members of a DL,e.g., when we display details on a foreign DL
//          that is copied to our AB. 
//
//          First we open the foreign object and get its properties. We then
//          create a foreign entry. If the PR_TEMPLATE_ID is present on the
//          remote object, we call OpenTemplateID and pass back the pointer to
//          the remote interface. The remote AB will delegate to us. If there's
//          no template ID, we return our object. The CTemplateDL object is
//          derived from our vanilla CABContainer (DL) with a couple of methods
//          overridden.
//      
//    NOTE:!!! Some foreign ABs use named properties for their entries. This
//    version doesn't support named properties (yet). The foreign named 
//    properties must be remapped into our name space before copying. Some
//    foreign AB entry details will not be properly displayed.
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: CreateTemplateEntry(ULONG        cbEntryID,              
                                                 LPENTRYID    lpEntryID,              
                                                 ULONG        ulCreateFlags,          
                                                 LPMAPIPROP * lppMAPIPropEntry)
{
    HRESULT          hRes           = S_OK;
    LPMAPIPROP       pForeignObj    = NULL;
    CTemplateDL    * pHostDL        = NULL;
    CTemplateUser  * pHostUser      = NULL;
    LPSPropValue     pSrcProps      = NULL;
    PABPAB_EID       pEID           = NULL;
    ULONG            ulEntryType    = (ULONG) MAPI_E_NOT_FOUND,
                     ulCnt,
                     ulID;
    SPropValue       SIG,
                     spvTID;
    LPSPropTagArray  pNamedTags     = NULL;
    LPMAPINAMEID   * ppNames = NULL;
    LPSPropTagArray  pTags = NULL;
    int              i,
                     NamedCnt,
                   * Idx = NULL;





    spvTID.ulPropTag = 0;
    spvTID.Value.err = MAPI_E_NOT_FOUND;
    
    *lppMAPIPropEntry = NULL;
    SIG.ulPropTag = PR_NULL;

    // open the foreign entry
    if (FAILED(hRes = m_pLogon  -> 
                      m_pSupObj -> OpenEntry(cbEntryID,
                                             lpEntryID,
                                             NULL,
                                             0,
                                             &ulEntryType,
                                             (LPUNKNOWN *) &pForeignObj)))
        goto Quit;

    // stash the src props we need, copy the rest to our obj
    if (FAILED(hRes = pForeignObj -> GetProps(NULL,
                                              0,
                                              &ulCnt,
                                              &pSrcProps)))
        goto Quit;

    if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(ulCnt), (LPVOID *) &pNamedTags)))
        goto Quit;

    pNamedTags -> cValues = 0;

    if (NULL == (Idx = new int[ulCnt]))
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

    // set the ones we don't want to copy to PR_NULL    
    for (i = 0; i < (int) ulCnt; i++)
    {
        switch (pSrcProps[i].ulPropTag)
        {
            // save whether this is a foreign mailuser, DL
            case PR_OBJECT_TYPE:
                ulEntryType = ENTRY_TYPE(pSrcProps[i].Value.l) | FOREIGN_ENTRY;
                break;    
            case PR_TEMPLATEID:
                spvTID = pSrcProps[i]; 
                //bHasTemplateID = TRUE;
                break;
            case PR_MAPPING_SIGNATURE:
                SIG = pSrcProps[i];                     // save data for later
                pSrcProps[i].ulPropTag = PR_NULL;       // so SetProps won't copy this one
                break;
            case PR_ENTRYID:
                pSrcProps[i].ulPropTag = PR_NULL;
                break;
            default:
                {
                    if (PT_OBJECT == PROP_TYPE(pSrcProps[i].ulPropTag))
                        pSrcProps[i].ulPropTag = PR_NULL;
 
                    if (PROP_ID(pSrcProps[i].ulPropTag) > 0x7FFF)
                    {
                        // copy foreign proptag to array
                        pNamedTags -> aulPropTag[pNamedTags -> cValues] = 
                                                    pSrcProps[i].ulPropTag;

                        // save index in SPropValue array for later replacement
                        Idx[pNamedTags -> cValues++] = i;
                  }                                             
                } 
        }
    }

    // PR_OBJECT_TYPE is missing, bail
    if (MAPI_E_NOT_FOUND == ulEntryType)
    {
        hRes = MAPI_E_NOT_FOUND;
        goto Quit;
    }
    
    // save whether the object supports OpenTemplateID
    ulEntryType |= (MAPI_E_NOT_FOUND != spvTID.Value.err &&
                    PR_TEMPLATEID   == spvTID.ulPropTag ? HAS_TEMPLATEID : 0);

    // only open foreign entries here
    if (FOREIGN_DL    != ulEntryType &&
        FOREIGN_USER  != ulEntryType &&
        TEMPLATE_DL   != ulEntryType &&
        TEMPLATE_USER != ulEntryType)
    {
        hRes = MAPI_E_NO_SUPPORT;
        goto Quit;
    }

    // we need a new ID for our local object
    if (FAILED(hRes = NextID(m_pLogon -> m_pDB,ulID)))
       goto Quit;

    if (FAILED(hRes = m_pLogon -> NewEID(0, 
                                         ulID, 
                                         ulEntryType, 
                                         NULL,
                                         &pEID)))
        goto Quit;

    switch (ulEntryType)
    {  
        case FOREIGN_DL:
        case TEMPLATE_DL:
            pHostDL = new CTemplateDL(m_pLogon,                        
                                      &m_MyEID,
                                      pEID,
                                      ulCreateFlags,
                                      this,
                                      TRUE);
            
            if (!pHostDL)
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                goto Quit;
            }
                            
            if (FAILED(hRes = InitTemplateDL(pForeignObj,
                                             pHostDL)))
                goto Quit;  
    
            if (TEMPLATE_DL == ulEntryType)   // has PR_TEMPLATEID in spvTID
            {
                hRes = m_pLogon  -> 
                       m_pSupObj -> OpenTemplateID(spvTID.Value.bin.cb,
                                                   (LPENTRYID) spvTID.Value.bin.lpb,
                                                   FILL_ENTRY,
                                                   pHostDL,
                                                   NULL,
                                                   &(pHostDL -> m_pRemoteIProp),
                                                   NULL);
                if (FAILED(hRes))
                    goto Quit;

                // if OpenTemplateID returns a wrapped IMAPIProp with
                // our pHostDL embedded, it'll AddRefs us, so Release()
                // if it doesn't wrap us, it just returns our pHostDL
                // w/o AddRef'ing us
                if (pHostDL != pHostDL -> m_pRemoteIProp)
                    pHostDL -> Release();

               *lppMAPIPropEntry = pHostDL -> m_pRemoteIProp;
            }
            else
                // server won't delegate, all OpenProperty calls go
                // first to CTemplateUser::, then CMAPIProp::
                *lppMAPIPropEntry = pHostDL;
            break;

        case FOREIGN_USER:
        case TEMPLATE_USER:
            pHostUser = new CTemplateUser(m_pLogon,                        
                                          &m_MyEID,
                                          pEID,
                                          ulCreateFlags,
                                          this,
                                          TRUE);
            if (!pHostUser)
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                goto Quit;
            }
  
            if (FAILED(hRes = InitTemplateUser(pForeignObj,
                                               pHostUser)))
                goto Quit;  
           
            if (TEMPLATE_USER == ulEntryType)     // has PR_TEMPLATEID in spvTID
            {
                hRes = m_pLogon  -> 
                       m_pSupObj -> OpenTemplateID(spvTID.Value.bin.cb,
                                                   (LPENTRYID) spvTID.Value.bin.lpb,
                                                   FILL_ENTRY,
                                                   (LPMAPIPROP)pHostUser,
                                                   NULL,
                                                   (LPMAPIPROP*) &(pHostUser -> m_pRemoteIProp),
                                                   NULL);
                if (FAILED(hRes))
                    goto Quit;

                // if OpenTemplateID returns a wrapped IMAPIProp with
                // our pHostDL embedded, it'll AddRefs us, so Release()
                // if it doesn't wrap us, it just returns our pHostDL
                // w/o AddRef'ing us
                if (pHostUser -> m_pRemoteIProp != (LPMAPIPROP)pHostUser)
                    pHostUser -> Release();

                *lppMAPIPropEntry = pHostUser -> m_pRemoteIProp;
            }
            else
                // server won't delegate, all OpenProperty calls go
                // first to CTemplateUser::, then CMAPIProp::
                
               *lppMAPIPropEntry = (LPMAPIPROP)pHostUser;
            break;
        default:
            ASSERT(0); 
    }

    // this is where we remap named properties
    if (pNamedTags -> cValues)
    {
        if (FAILED(hRes = pForeignObj -> GetNamesFromIDs(&pNamedTags,
                                                         NULL,
                                                         MAPI_NO_STRINGS,
                                                         (ULONG *) &NamedCnt,
                                                         (LPMAPINAMEID **)&ppNames)))
            goto Quit;
        
        if (FAILED(hRes = GetIDsFromNames(NamedCnt,
                                          ppNames,
                                          MAPI_CREATE,
                                          &pTags)))
            goto Quit;

        // replace foreign ID with ID from our mapping
        for (i = 0; i < (int) pNamedTags -> cValues; i++)                                                           
            SWAP_IDS(pSrcProps[Idx[i]].ulPropTag, pTags -> aulPropTag[i]);
    }

    // copy some of the foreigner's properties to our object
    hRes = (*lppMAPIPropEntry) -> SetProps(ulCnt,
                                           pSrcProps,
                                           NULL);
Quit:

    if (FAILED(hRes))
    {
        if (*lppMAPIPropEntry)
            (*lppMAPIPropEntry) -> Release();

        *lppMAPIPropEntry = NULL;

        if (pHostUser)
            pHostUser -> Release();

        if (pHostDL)
            pHostDL -> Release();                        
    }

    if (pForeignObj)
        pForeignObj -> Release();

    if (Idx)
        delete Idx;

    gFreeBuff(pTags);
    gFreeBuff(ppNames);
    gFreeBuff(pNamedTags);
    gFreeBuff(pSrcProps);
    gFreeBuff(pEID);

    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    UpdateCTbl()
//
//    Parameters                        
//          NewEntryRecNum      [in] record number of new entry
//          pEntryTblProps      [in] list of contents table column props
//    Purpose
//          Adds a new entry's row to the contents table.
//          If pNewTblProps is NULL, gets the new entry's table 
//          columns from the DB, otherwise from pEntryTblProps, 
//          and creates a new row for this container's contents table. 
//          Modifies the table data. The record must be Open()'d prior
//          to this call if pEntryTblProps is NULL.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: UpdateCTbl(RECORD_NUMBER   NewEntryRecNum,
                                        PCFlatPropList  pEntryTblProps)
{
    HRESULT         hRes = S_OK;
    CFlatPropList   EntryTblProps;
    SRow            sr;

    sr.cValues = NUM_CTBL_COLS;
    sr.lpProps = NULL;

    if (!pEntryTblProps)
    {
        if (FAILED(hRes = Fetch(NewEntryRecNum, fldPROPS, &EntryTblProps)))
            goto Quit;
    
        if (FAILED(hRes = m_pLogon -> GetCTblRow(&EntryTblProps,
                                                 &sr.lpProps)))
            goto Quit;                                    
    }
    else
        if (FAILED(hRes = m_pLogon -> GetCTblRow(pEntryTblProps,
                                                 &sr.lpProps)))
            goto Quit;                                    

        
    hRes =  m_pCntsTblData -> HrModifyRow(&sr);    

Quit:
    gFreeBuff((LPVOID)sr.lpProps);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CopyEntries()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: CopyEntries(LPENTRYLIST        lpEntries,              
                                         ULONG              ulUIParam,              
                                         LPMAPIPROGRESS     lpProgress,             
                                         ULONG              ulFlags) 
{   
    LPMAPIPROP          pNewEntry  = NULL;
    ULONG               ulFailures = 0;
    int                 i;

    DECLARE_REC_NUM(ParentRecNum);

    TraceString1("CABContainer <%x> :: CopyEntries", m_MyEID.ID.Num);
    Validate_IABContainer_CopyEntries(this,
                                      lpEntries, 
                                      ulUIParam, 
                                      lpProgress,
                                      ulFlags);

    // If the entry ID is a foreigner, create a foreign entry in this
    // container. If it's one of ours, and it doesn't exist in the database
    // (i.e. a one-off) create it, SaveChanges and Release.
    // SaveChanges will add it to the database as a child of this container.
    // If the entry already exists in the database, make it a child of this
    // container.

    EnterCriticalSection(&m_cs);

    for (i = 0; i < (int)lpEntries -> cValues; i++)
    {
        PABPAB_EID      pEID = (PABPAB_EID) lpEntries -> lpbin[i].lpb;
        ULONG           ulEntryType;

        DECLARE_REC_NUM(ChildRecNum);
        
        ulEntryType = m_pLogon -> CheckEID(lpEntries -> lpbin[i].cb,
                                         (LPENTRYID)lpEntries -> lpbin[i].lpb);

        switch(ulEntryType)
        {
            case MAPI_E_INVALID_ENTRYID:
                ulFailures++;
                break;

            case MY_DISTLIST:
            case MY_MAILUSER:
            case MY_ABCONT:
            case TEMPLATE_USER:
            case TEMPLATE_DL:
            case FOREIGN_USER:
            case FOREIGN_DL:
                // If entry is already in the database, add it as a child of
                // this container, else create it. We don't have to worry about
                // deleting duplicates, duplicate checking (per create flags) 
                // happens when an entry is created in the database. When created,
                // each entry is unique (entryid) even if two have the same name
                // and/or search key.
                if (SUCCEEDED(Open(m_pLogon -> m_pDB, 
                                   OPEN_ONLY,
                                   m_MyEID.ID,
                                   ParentRecNum)))
                {
                    if (SUCCEEDED(Open(m_pLogon -> m_pDB, 
                                       OPEN_ONLY,
                                       pEID -> ID,
                                       ChildRecNum)))
                    {
                        if (SUCCEEDED(InsertRec(m_pLogon -> m_pDB,
                                                ParentRecNum,
                                                ChildRecNum)))
                        {
                            if (SUCCEEDED(UpdateCTbl(ChildRecNum,NULL)))
                                m_pLogon  -> Notify(fnevObjectCopied,
                                                    &m_MyEID,
                                                    pEID,
                                                    NULL);
                            else
                                ulFailures++;
                        }
                        else
                            ulFailures++;
                        Close(ChildRecNum);
                        Close(ParentRecNum);
                        break;
                    }
                    else    
                    {       
                        Close(ParentRecNum);
                        ulFailures++;   
                        break;          
                    }
                }                      
                else 
                {                
                    ulFailures++;
                    break;
                }   // fall through

            case MY_ONEOFF_USER:
            case MY_ONEOFF_DL:           
            case FOREIGN_ENTRY:     // another AB's entry ID, not in DB yet
                if (SUCCEEDED(CreateEntry(lpEntries -> lpbin[i].cb,                
                                          (LPENTRYID)lpEntries -> lpbin[i].lpb,
                                          ulFlags,
                                          & pNewEntry)))
                {
                    if (M_FAIL(pNewEntry -> SaveChanges(0)))
                        ulFailures++;

                    pNewEntry -> Release();                   
                }
                else
                    ulFailures++;

                break;
        }
    }
    LeaveCriticalSection(&m_cs);

    return (0 == ulFailures                    ? S_OK : 
            ulFailures == lpEntries -> cValues ? MAPI_E_INVALID_ENTRYID : 
            MAPI_W_PARTIAL_COMPLETION);
}

///////////////////////////////////////////////////////////////////////////////
//
//    DeleteChildren()
//
//    Parameters                        
//          MyID            [in] parent container's ID     
//          pDeletions      [in] list of children to delete
//    Purpose
//          Recursively traverses the tree of children and deletes them.
//          (Leaf nodes are MailUsers, DLs can have children, i.e. members)
//          Opens the parent's record, gets a list of its children, then
//          DeleteChildren's each child in the list.
//      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: DeleteChildren(ABPAB_ID        & MyID, 
                                            CList<ABPAB_ID> * pDeletions)
{
    HRESULT             hRes        = S_OK;
    PABPAB_ID           pID         = NULL;
    CList<ABPAB_ID>   * pDeleteList = NULL,
                        Parents,
                        Children;

    DECLARE_REC_NUM(RecNum);

    if (MY_DISTLIST == MyID.Type || MY_ABCONT == MyID.Type)
    {

        if (FAILED(hRes = Open(m_pLogon -> m_pDB, 
                               OPEN_ONLY,
                               MyID,
                               RecNum)))
            return hRes;
        
        if (SUCCEEDED(hRes = Fetch(RecNum,fldCHILDREN, &Children)))
        {
            if (pDeletions)         // only some of my children will be deleted
            {
                Children -= *pDeletions;
                pDeleteList = pDeletions;
            }
            else        
            {
                pDeleteList = & Children;       // i.e. delete all my children
                Children.Destroy();             // write back an empty list of children
            }

            hRes = Update(RecNum, fldCHILDREN, Children);
        }
        
        Close(RecNum);
        
        if (FAILED(hRes))      // E_NOT_FOUND isn't an error, we've reached an empty DL
            return (MAPI_E_NOT_FOUND == hRes ? S_OK : hRes);
        
        // pDeleteList depends on the container's critical section
        for (pID = pDeleteList -> Scan(0);
             pID;
             pID = pDeleteList -> Scan())            
        {
            PABPAB_ID   pDel = NULL;

            if (FAILED(hRes = DeleteChildren(*pID,NULL)))
                break;

            // remove me from my childrens' lists of parents, i.e
            // make this child "no child o' mine"
            if (FAILED(hRes = Open(m_pLogon -> m_pDB, 
                                   OPEN_ONLY,
                                   *pID,
                                   RecNum)))
                break;

            if (SUCCEEDED(hRes = Fetch(RecNum,fldPARENTS, &Parents)))
            {
                Parents.Delete(MyID,&pDel);

                if (pDel)
                    delete pDel;

                if (Parents.Size())            
                    hRes = Update(RecNum, fldPARENTS,Parents);
                else
                {
                    // if child is an orphan, delete it from DB
                    // and notify listeners that it's gone                    
                    if (SUCCEEDED(hRes = DeleteRec(m_pLogon -> m_pDB, RecNum)))
                    {
                        ABPAB_EID eidMe,
                                  eidChild;
                        
                        memcpy((LPBYTE) &eidMe, (LPBYTE) &m_MyEID, sizeof m_MyEID);
                        eidMe.ID = MyID;
                        memcpy((LPBYTE) &eidChild, (LPBYTE) &m_MyEID, sizeof m_MyEID);
                        eidChild.ID = *pID;
                                        
                        m_pLogon  -> Notify(fnevObjectDeleted,
                                            &eidMe,
                                            &eidChild,
                                            NULL);
                    }
                }
            }                   

            Close(RecNum);

            if (FAILED(hRes))
                break;
        }
    }
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    DeleteEntries()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details. Converts the
//          lpEntries into a list and deletes each child.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: DeleteEntries(LPENTRYLIST  lpEntries,              
                                           ULONG        ulFlags)
{
    HRESULT             hRes = S_OK,
                        hr   = S_OK;
    ULONG               i;
    CList<ABPAB_ID>     DeleteList;
    SPropValue          Prop;
    SizedSPropTagArray(1,sptTag) = {1, PR_CONTAINER_CONTENTS};
       
    TraceString1("CABContainer <%x> :: DeleteEntries", m_MyEID.ID.Num);
    Validate_IABContainer_DeleteEntries(this,
                                        lpEntries,
                                        ulFlags);  

    EnterCriticalSection(&m_cs);

    for (i = 0; i < lpEntries -> cValues; i++)
    {

        PABPAB_EID  pEID = (PABPAB_EID) lpEntries -> lpbin[i].lpb;        

        if (FOREIGN_ENTRY == m_pLogon -> CheckEID(lpEntries -> lpbin[i].cb, 
                                                  (LPENTRYID)lpEntries -> lpbin[i].lpb))
        {
            hRes = MAPI_E_INVALID_ENTRYID;            
            continue;
        }

        Prop.ulPropTag = PR_INSTANCE_KEY;
        Prop.Value.bin.cb = sizeof (pEID -> ID.Num);
        Prop.Value.bin.lpb = (LPBYTE) &(pEID -> ID.Num);

        // make a list of the deletions
        DeleteList.Insert(pEID -> ID);

        if (m_pCntsTblData)
            if (M_FAIL(m_pCntsTblData -> HrDeleteRow(&Prop)))
                hr = MAPI_W_PARTIAL_COMPLETION;
    }

    // this does the work recursively
    hRes = DeleteChildren(m_MyEID.ID, &DeleteList);
    m_pLogon -> Notify(fnevObjectModified, 
                       NULL,
                       &m_MyEID,
                       (LPSPropTagArray) &sptTag);

    LeaveCriticalSection(&m_cs);

    if (FAILED(hRes) || S_OK != hr)
        return MAPI_W_PARTIAL_COMPLETION;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetNextToken()
//
//    Parameters                        
//          pszInput    [in-out] input: buffer to tokenize, 
//                               output:buffer with tokenized prefix
//                               removed suitable to pass to GetNextToken
//
//    Purpose
//          Destructively obtains the next non-whitespace substring
//          in the input buffer. Equivalent to strtok().
//      
//    Return Value
//          The next substring if found, NULL if at end of string
//
LPTSTR GetNextToken(LPTSTR * pszInput)
{
    TCHAR * szPtr = *pszInput,
          * szTmp;
    
    if (! szPtr)
        return NULL;

    // skip whitespace
    while (isspace(*szPtr))
        szPtr++;

    szTmp = szPtr;

    // skip blackspace
    while ('\0'!= *szPtr && (!isspace(*szPtr)))
        szPtr++;

    // end of input ?
    if ('\0' == *szPtr)
        *pszInput = NULL;
    else
    {
        *szPtr = (TCHAR) 0;
        *pszInput = szPtr + 1;
    }

    return CharUpper(szTmp);
} 

///////////////////////////////////////////////////////////////////////////////
//
//    class DECK
//
//    A simple stack for holding tokens used in ambiguous name resolution.
//    The only reason its a deck and not a true stack is that we put the
//    tokens on in FIFO order instead of LIFO
//    

enum        {INIT,AMBIG,NOMATCH,MATCH,MATCHING};
#define     MAX_TOKS    64

class DECK
{
    public:
        DECK()
        {
            m_bot = m_idx = -1;
            m_top = 0;
        }

        ~DECK(){};

        LPSTR   Pop()
        {
            if (Empty())
                return NULL;

            return m_toks[m_top++];
        }
        
        BOOL    Queue(LPSTR szTok)
        {
            if (m_bot >= (MAX_TOKS - 1))
                return FALSE;

            m_toks[++m_bot] = szTok;
            return TRUE;
        }

        BOOL    Empty()
        {  
            return m_top > m_bot;
        }

        BOOL    Contains(LPSTR szTok)
        {
            if (Empty())
                return FALSE;

            for (m_idx = m_top; m_idx <= m_bot; m_idx++)
                if (!lstrcmp(szTok,m_toks[m_idx]))         
                    return TRUE;

            return FALSE;
        }

        void PopHandle()
        {
            m_top = m_idx + 1;
        }
        
        LPSTR Top()
        {
            return m_toks[m_top];
        }
    private:
        int     m_idx,
                m_top,
                m_bot;
        LPSTR   m_toks[MAX_TOKS];

};

///////////////////////////////////////////////////////////////////////////////
//
//    RunStringState()
//
//    Parameters                        
//          ulPropTag       [in] property in pRows with target string
//          szString        [in] name we're resolving
//          pRows           [in] row from table we're resolving against
//
//    Purpose
//          Tokenizes the sought for string and the target string
//          (in the pRows) and compares them token by token. If 
//          szStrings tokens are a proper substring of pRows[?],
//          the match is ambiguous. If they are the same, token
//          for token, it's an exact match, else it's no match.
//          This function is a simple state machine.
//      
//    Return Value
//          AMBIG, MATCH, or NO_MATCH as appropriate
//
ULONG WINAPI RunStringState(ULONG ulPropTag, LPTSTR szString,LPSRow pRow)
{

    int          State = INIT;
    DECK         Stack;    
    LPSTR        szTok = NULL;
    TCHAR        szBuf1[MAX_PATH],
                 szBuf2[MAX_PATH];
    LPSTR        szIn  = szBuf1,
                 szRow = szBuf2;

    LPSPropValue pProp = PpropFindProp(pRow -> lpProps,
                                       pRow -> cValues,
                                       ulPropTag);

    if (!pProp)
        return NOMATCH;

    lstrcpy(szRow,pProp -> Value.lpszA);
    lstrcpy(szIn, szString);

    // insert row's tokens in stack in R-to-L order
    while (NULL != (szTok = GetNextToken(&szRow)))
        Stack.Queue(szTok);

    szTok = GetNextToken(&szIn);
    
    if (!lstrcmp(szTok, Stack.Top()))
    {
        State = MATCHING;

        while (1)
        {
            Stack.Pop();
            szTok = GetNextToken(&szIn);
            
            // both input and row tokens gone, match
            if (!szTok && Stack.Empty())
                return MATCH;

            // input OR row tokens gone, partial match
            if (!szTok || Stack.Empty())
                return AMBIG;
            
            // both have tokens left but they stopped matching, continue
            if (lstrcmp(szTok, Stack.Top()))
                break;
        }
    }
    else
        State = NOMATCH;

    while (1)
    {
        // scan stack for current token
        if (Stack.Contains(szTok))
        {
            // pop a handle from stack
            Stack.PopHandle();
            szTok = GetNextToken(&szIn);

            // we were matching before, but now one stream is empty
            if (!szTok || (Stack.Empty() && MATCHING == State))
                return AMBIG;
        }
        else
        {
            szTok = GetNextToken(&szIn);

            if (!szTok)
                return NOMATCH;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//    DoANR()
//
//    Parameters                        
//          szTarget        [in]  recipient name we're resolving
//          pSearchTbl      [in]  table we're looking for recipient in
//          psrResolved     [out] resolved ADRENTRY
//          ulNameCol       [in]  index into table rows props where we find
//                                the string property for the comparison
//    Purpose
//          This fcn does ambiguous name resolution on szTarget in
//          pSearchTable. We return a single row in *psrResolved if we find
//          a match and pass back NULL and return MAPI_E_AMBIGUOUS or 
//          MAPI_E_NOT_FOUND otherwise.
//
//
//          If we don't find a match,  We restrict the container's
//          contents table, to rows that have the target string's first token as a 
//          substring in the display name. Then we run a simple parser on each row
//          to determine if the rest of the tokens match exactly, kind of match 
//          (ambiguous) or don't match at all
//
//          We define a match to be one and only one exact match (token for token)
//          or exactly 1 ambiguous match.
//
//          We do this because MAPI defines an ambiguous match to be > 1 match. 
//          The implementation determines what is meant by a match. If we have
//          a single non-exact match, i.e. return a single row in Restrict w/
//          PR_ANR, the CheckNames dialog treats it like a match anyway.
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP  DoANR(LPTSTR        szTarget,
                    LPMAPITABLE   pSearchTbl,
                    LPSRow        psrResolved,
                    ULONG         ulNameCol)
{
    HRESULT         hRes               = S_OK;
    LPSRowSet       pHitRow            = NULL;        
    BOOL            bSearch            = TRUE;
    TCHAR           szBuf[MAX_PATH];
    LPSTR           szTemp             = szBuf;
    ULONG           ulMatches          = 0,
                    ulState,
                    ulAmbig = 0;
    int             i,idxMatch = -1;
    SPropValue      spTemp;
    SRestriction    SRes;


    lstrcpyn(szTemp,szTarget, MAX_PATH);
    szTemp[MAX_PATH - 1] = '\0';
    spTemp.ulPropTag   = PR_DISPLAY_NAME;
	spTemp.Value.lpszA = GetNextToken(&szTemp);

    // find all rows with the target string's first token
	SRes.rt = RES_CONTENT;
  	SRes.res.resContent.ulPropTag = PR_DISPLAY_NAME;
	SRes.res.resContent.lpProp = &spTemp;
    SRes.res.resContent.ulFuzzyLevel = FL_SUBSTRING | FL_IGNORECASE;

    if (FAILED(hRes = HrQueryAllRows(pSearchTbl,
                                     NULL,
                                     &SRes,
                                     NULL,
                                     0,
                                     &pHitRow)))
        goto Quit;

    for (bSearch = TRUE, i = 0; i < (int)pHitRow -> cRows && bSearch; i++)
    {   
        // run our parser and see if it matches
        switch (ulState = RunStringState(PR_DISPLAY_NAME, 
                                         szTarget,
                                         &pHitRow -> aRow[i]))
        {   
            case MATCH:
                if (++ulMatches > 1)
                    bSearch = FALSE;

                idxMatch = i;
                break;
            case AMBIG:
                ulAmbig++;
                idxMatch = i;
                break;
            case NOMATCH:               
                break;
        }
    }
    
Quit:
    if (1 == ulMatches || !ulMatches && 1 == ulAmbig)
    {   // pull matching row out of rowset, save it
        *psrResolved = pHitRow -> aRow[idxMatch];
        pHitRow -> aRow[idxMatch].cValues = 0;
        pHitRow -> aRow[idxMatch].lpProps = NULL;
        hRes = S_OK;
    }
    else
    {   // if we have a single ambiguous hit
        if (ulAmbig > 1 || ulMatches > 1)
            hRes = MAPI_E_AMBIGUOUS_RECIP;
        else
            hRes = MAPI_E_NOT_FOUND;
    }
    FreeProws(pHitRow);
    return hRes;                                
}


///////////////////////////////////////////////////////////////////////////////
//
//    CopyAdrEntry()
//
//    Parameters                        
//     
//
//    Purpose
//          Copies an SRow's properties into an ADRENTRY without duplicates,
//          reallocating a larger ADRENTRY if necessary. The old ADRENTRY 
//          property array is freed and the new one is plugged into its
//          place. 
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: CopyAdrEntry(LPSPropTagArray pTags, 
                                          LPSRow          pRow,
                                          LPADRENTRY      pAdrEntry)
{
    HRESULT         hRes      = S_OK;
    LPSPropValue    pNewProps = NULL;

    CList<ULONG>    MergedProps;
    ULONG           cExistingProps,
                    cTotalProps,
                  * pulPropTag;

    // Make a merged list of existing props on the unresolved AdrEntry,
    // and requested props on the resolved one. Eliminate the duplicates.
    // (i.e. pTags and *ppAdrEntry can have tags in common). We really 
    // just want a count of the tags without counting common ones

    for (ULONG i = 0; i < pAdrEntry -> cValues; i++)
        MergedProps.Insert(pAdrEntry -> rgPropVals[i].ulPropTag);

    // cTotalProps is in case we fail, use it to free
    cExistingProps = MergedProps.Size();

    for (i = 0; i < pTags -> cValues; i++)
        MergedProps.Insert(pTags -> aulPropTag[i]);

    cTotalProps = MergedProps.Size();        

    if (FAILED(hRes = gAllocBuff(cTotalProps * sizeof SPropValue,
                                 (LPVOID *) & pNewProps)))
        goto Quit;

    // if we fail, FreeBuff won't complain
    ZeroMemory(pNewProps, cTotalProps * sizeof SPropValue);

    // copy existing props on the adr entry to new AdrEntry
    for (i = 0; i < cExistingProps; i++)
    {
        pulPropTag = MergedProps.Dequeue();  // remove from list
        delete pulPropTag;

        if (FAILED(hRes = CopyProp(&(pNewProps[i]), &(pAdrEntry -> rgPropVals[i]),pNewProps)))
            goto Quit;
    }

    // copy props from SRow to new adrentry
    // we'll never hit a prop in the row that's already in
    // the adrentry 'cause the merged list has no duplicates
    for (i = cExistingProps; i < cTotalProps; i++)
    {
        BOOL bFound = FALSE;
        pulPropTag = MergedProps.Dequeue();

        for (ULONG j = 0; j < pRow -> cValues; j++)
            if (PROP_ID(*pulPropTag) == PROP_ID(pRow -> lpProps[j].ulPropTag))
            { 
                bFound = TRUE;

                // uses MAPIAllocateMore, so we can free it later
                if (FAILED(hRes = CopyProp(&(pNewProps[i]), 
                                           &(pRow -> lpProps[j]),
                                           pNewProps)))
                    goto Quit;
                break;
            }    

        // this can never happen        
        if (!bFound)
            HrPropNotFound(pNewProps[i],*pulPropTag);
        
        delete pulPropTag;
    }    

Quit:

    if (FAILED(hRes))
    {
        gFreeBuff(pNewProps);
        return hRes;
    }

    gFreeBuff(pAdrEntry -> rgPropVals);
    pAdrEntry -> rgPropVals = pNewProps;
    pAdrEntry -> cValues    = cTotalProps;
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    ResolveNames()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          See the MAPI Documentation for details. Resolves names
//          against the container's contents table.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CABContainer :: ResolveNames(LPSPropTagArray lpPropTagArray,         
                                          ULONG           ulFlags,                
                                          LPADRLIST       lpAdrList,              
                                          LPFlagList      lpFlagList)
{      
    HRESULT           hRes     = S_OK; 
    LPMAPITABLE       pCntTbl  = NULL;
    ULONG             CurrIdx  = 0;
    SRow              ResolvedRow = {0,0,NULL};
    LPSPropTagArray   pCols;
    int               i;

    TraceString1("CABContainer <%x> :: ResolveNames", m_MyEID.ID.Num);
    Validate_IABContainer_ResolveNames(this,
                                       lpPropTagArray,
                                       ulFlags,       
                                       lpAdrList,     
                                       lpFlagList);    

    if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    pCols = (!lpPropTagArray ? (LPSPropTagArray) &g_sptCntTbl : lpPropTagArray);
    EnterCriticalSection(&m_cs);
                                    
    // contents table is also sorted
    if (M_FAIL(hRes = GetContentsTable(0,&pCntTbl)))
        goto Quit;

    if (M_FAIL(hRes = pCntTbl -> SetColumns(pCols, TBL_BATCH)))
        goto Quit;

    
    // if a name is unresolved, try to parse it
    for (i = 0; i < (int)lpFlagList -> cFlags; i++)       
    {   
        if (MAPI_UNRESOLVED == lpFlagList -> ulFlag[i])
        {
            LPSPropValue pProp = PpropFindProp(lpAdrList -> aEntries[i].rgPropVals,
                                               lpAdrList -> aEntries[i].cValues,
                                               PR_DISPLAY_NAME);
            if (!pProp)
                continue;

            hRes = DoANR(pProp -> Value.lpszA,
                         pCntTbl, 
                         &ResolvedRow,
                         0);
            switch (hRes)
            {
                case S_OK : // copy props to ADRENTRY
                    lpFlagList -> ulFlag[i] = MAPI_RESOLVED;
                    hRes = CopyAdrEntry(pCols, 
                                        &ResolvedRow, 
                                        &(lpAdrList -> aEntries[i]));
                    break;

                case MAPI_E_AMBIGUOUS_RECIP:
                    lpFlagList -> ulFlag[i] = MAPI_AMBIGUOUS;
                    hRes = S_OK;
                    break;
                case MAPI_E_NOT_FOUND:
                    hRes = S_OK;
                    break;
                default:
                    break;
            }
        } 
    }        

Quit:
    gFreeBuff((LPVOID)ResolvedRow.lpProps);

    if (pCntTbl)
        pCntTbl -> Release();   
         
    LeaveCriticalSection(&m_cs);

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetDefContProps()
//
//    Parameters                        
//          pEID            [in]     pointer to entry ID of container        
//          pFlatPropList   [in/out] pointer to list to hold default properties
//    Purpose
//          Initialize a list of properties with default values for a 
//          container. Any properties in pFlatPropList that are defaults are replaced
//          by the hard-coded defaults.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP SetDefContProps(PABPAB_EID      pEID, 
                             CFlatPropList * pFlatPropList)
{
    HRESULT     hRes = S_OK;
    enum {CONT_FLAGS,
          EID,
          INST_KEY,
          OBJ_TYPE,
          DISP_TYPE,                        
          NAME,
          DEPTH,  
          ACCESS,
          STATUS,  
          PROV_UID,
          ABP_ID,
          DEF_USER,
          DEF_DL,
          SIG,
          NUM_PROPS};

    ABPAB_EID  eidDefDL,
               eidDefUser;
    SPropValue Props[NUM_PROPS];
    
    ZeroMemory((LPVOID) &eidDefDL,sizeof eidDefDL);
    ZeroMemory((LPVOID) &eidDefUser,sizeof eidDefUser);
    eidDefDL.muid = eidDefUser.muid = pEID -> muid;

    eidDefDL.ID.Num    = ONEOFF_DL_ID;    
    eidDefDL.ID.Type   = MY_DISTLIST | ONEOFF_ENTRY;

    eidDefUser.ID.Num  = ONEOFF_USER_ID;
    eidDefUser.ID.Type = MY_MAILUSER | ONEOFF_ENTRY;

    Props[CONT_FLAGS].ulPropTag   = PR_CONTAINER_FLAGS;
    Props[CONT_FLAGS].Value.l     = AB_MODIFIABLE | AB_RECIPIENTS;
          
    Props[EID].ulPropTag          = PR_ENTRYID;
    Props[EID].Value.bin.cb       = sizeof(*pEID);
    Props[EID].Value.bin.lpb      = (LPBYTE) pEID;
    
    Props[INST_KEY].ulPropTag     = PR_INSTANCE_KEY;
    Props[INST_KEY].Value.bin.cb  = sizeof ULONG;
    Props[INST_KEY].Value.bin.lpb = (LPBYTE) &(pEID -> ID.Num);

    Props[OBJ_TYPE].ulPropTag     = PR_OBJECT_TYPE;
    Props[OBJ_TYPE].Value.l       = MAPI_ABCONT;

    Props[DISP_TYPE].ulPropTag    = PR_DISPLAY_TYPE;
    Props[DISP_TYPE].Value.l      = DT_MODIFIABLE;

    // IAddrBook::Details complains if you don't have this
    Props[NAME].ulPropTag         = PR_DISPLAY_NAME;
    Props[NAME].Value.LPSZ        = DIR_ID_NUM  == pEID -> ID.Num || 
                                    ROOT_ID_NUM == pEID -> ID.Num ? 
                                        SZ_PSS_SAMPLE_AB          :
                                        NULL_STRING;

    Props[DEPTH].ulPropTag        = PR_DEPTH;
    Props[DEPTH].Value.l          = 0;

    Props[ACCESS].ulPropTag       = PR_ACCESS;
    Props[ACCESS].Value.l         = MAPI_MODIFY;

    Props[STATUS].ulPropTag       = PR_STATUS;
    Props[STATUS].Value.l         = 0;

    Props[PROV_UID].ulPropTag     = PR_PROVIDER_UID;
    Props[PROV_UID].Value.bin.cb  = sizeof MAPIUID;
    Props[PROV_UID].Value.bin.lpb = (LPBYTE) & ABPAB_PROVIDER_UID;

    Props[ABP_ID].ulPropTag       = PR_AB_PROVIDER_ID;
    Props[ABP_ID].Value.bin.cb    = sizeof MAPIUID;
    Props[ABP_ID].Value.bin.lpb   = (LPBYTE)& (pEID -> muid);
                           
    Props[DEF_DL].ulPropTag       = PR_DEF_CREATE_DL;
    Props[DEF_DL].Value.bin.cb    = sizeof ABPAB_EID;
    Props[DEF_DL].Value.bin.lpb   = (LPBYTE) &eidDefDL;

    Props[DEF_USER].ulPropTag     = PR_DEF_CREATE_MAILUSER;
    Props[DEF_USER].Value.bin.cb  = sizeof ABPAB_EID;
    Props[DEF_USER].Value.bin.lpb = (LPBYTE) &eidDefUser;

    Props[SIG].ulPropTag          = PR_MAPPING_SIGNATURE;
    Props[SIG].Value.bin.cb       = sizeof MAPIUID;
    Props[SIG].Value.bin.lpb      = (LPBYTE) & (pEID -> muid);

    for (int i = CONT_FLAGS; i < NUM_PROPS; i++)                 
        if (FAILED(hRes = pFlatPropList -> Insert(&Props[i])))
            return hRes;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetDefaultContProps()
//
//    Parameters                        
//          pEID            [in]     pointer to entry ID of container        
//    Purpose
//          Initialize a container's property cache with default values.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: SetDefaultContProps(PABPAB_EID pEID)
{
    m_fDirtyFlags |= PROPS;
    return SetDefContProps(pEID, m_pPropCache -> m_pProps);
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetDefaultDLProps()
//
//    Parameters                        
//          pEID            [in]     pointer to entry ID of DL
//    Purpose
//          Initialize a DL's property cache with default values.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: SetDefaultDLProps(PABPAB_EID pEID)
{
    HRESULT     hRes = S_OK;

    enum {EID,
          INST_KEY,
          SEARCH_KEY,
          OBJ_TYPE,
          DISP_TYPE,
          DISP_NAME,                       
          ADDRTYPE,
          SIG,
          NUM_PROPS
         };

    SPropValue      Props[NUM_PROPS];

    Props[EID].ulPropTag            = PR_ENTRYID;
    Props[EID].Value.bin.cb         = sizeof(*pEID);
    Props[EID].Value.bin.lpb        = (LPBYTE) pEID;

    Props[INST_KEY].ulPropTag       = PR_INSTANCE_KEY;
    Props[INST_KEY].Value.bin.cb    = sizeof(ULONG);
    Props[INST_KEY].Value.bin.lpb   = (LPBYTE) & (pEID -> ID.Num);

    Props[SEARCH_KEY].ulPropTag     = PR_SEARCH_KEY;
    Props[SEARCH_KEY].Value.bin.cb  = sizeof (*pEID);
    Props[SEARCH_KEY].Value.bin.lpb = (LPBYTE)pEID;

    Props[OBJ_TYPE].ulPropTag       = PR_OBJECT_TYPE;
    Props[OBJ_TYPE].Value.l         = MAPI_DISTLIST;

    Props[DISP_TYPE].ulPropTag      = PR_DISPLAY_TYPE;
    Props[DISP_TYPE].Value.l        = DT_PRIVATE_DISTLIST;

    Props[DISP_NAME].ulPropTag      = PR_DISPLAY_NAME;
    Props[DISP_NAME].Value.LPSZ     = "";

    Props[ADDRTYPE].ulPropTag       = PR_ADDRTYPE;
    Props[ADDRTYPE].Value.LPSZ      = "MAPIPDL";

    Props[SIG].ulPropTag            = PR_MAPPING_SIGNATURE;
    Props[SIG].Value.bin.cb         = sizeof MAPIUID;
    Props[SIG].Value.bin.lpb        = (LPBYTE) & pEID -> muid;

    for (int i = EID; i < NUM_PROPS; i++)
        if (FAILED(hRes = m_pPropCache -> m_pProps -> Insert(&Props[i])))
            return hRes;

    m_fDirtyFlags |= PROPS;

    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    SetDefaultProps()
//
//    Parameters                        
//          pEID        [in]  pointer to object's entry ID     
//
//    Purpose
//          Initializes a container or DL's default property
//          property cache with default values.      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABContainer :: SetDefaultProps(PABPAB_EID pEID)
{
    HRESULT     hRes;

    m_fDirtyFlags |= PROPS;

    switch (pEID -> ID.Type)
    {
        case MY_ABCONT:
            hRes = SetDefaultContProps(pEID);
            break;
        case MY_DISTLIST:
        case FOREIGN_DL:
        case TEMPLATE_DL:
        case MY_ONEOFF_DL:
            hRes = SetDefaultDLProps(pEID);
            break;
        default:
            hRes = MAPI_E_INVALID_PARAMETER;
    }
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    CContAdviseSink()
//
//    Parameters                        
//          pCont   [in] pointer to container to receive notifications
//
//    Purpose
//          Constructor for an advise sink object for a container. 
//      
//    Return Value
//          none.
//
CContAdviseSink :: CContAdviseSink(LPCABCont pCont)
{
    m_pCont = pCont;
    m_cRef = 0;
    InitializeCriticalSection(&m_cs);
}

///////////////////////////////////////////////////////////////////////////////
//
//    OnNotify()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          See the MAPI Documentation for details.
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP_ (ULONG) CContAdviseSink :: OnNotify(ULONG cNotif, LPNOTIFICATION pNotifs)
{
    HRESULT     hRes = S_OK;

    TraceString1("CContAdviseSink::OnNotify. cont: %x",m_pCont -> m_MyEID.ID.Num);

    for (ULONG i = 0; i < cNotif; i++)
        switch (pNotifs[i].ulEventType)
        {
            case fnevChildEvent:
                switch (pNotifs[i].info.tab.ulTableEvent)
                {  

                    case ABPAB_CHILD_ADDED:
                        // if a child was created in me,I've changed.Notify
                        m_pCont -> m_pLogon -> Notify(fnevObjectModified,
                                                      NULL, 
                                                      &m_pCont -> m_MyEID,
                                                      NULL);
                        // fall through
                    case ABPAB_CHILD_CHANGED: 
                    case ABPAB_CHILD_MODIFIED:
                        // update contents table on any change in my children
                        SRow    sr = pNotifs[i].info.tab.row;

                        if (m_pCont -> m_pCntsTblData)
                            hRes = m_pCont -> m_pCntsTblData -> HrModifyRow(&sr);
                        break;

                    default:
                        hRes = MAPI_E_NO_SUPPORT;
                        break;
                }
                break;
            case fnevObjectModified:
                break;
            case fnevObjectCreated:
                break;
            case fnevObjectDeleted:
                break;
            default: 
                hRes = MAPI_E_NO_SUPPORT;
    }

    return hRes;

}

///////////////////////////////////////////////////////////////////////////////
//    CAdviseSink::Release()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          Decrease usage count of this object. If the count goes to zero,
//          the object deletes itself.
//
//    Return Value
//      Usage of this object
//
STDMETHODIMP_ (ULONG) CContAdviseSink::Release (void)
{    
    ULONG ulRefCount;

    EnterCriticalSection(&m_cs);
    ulRefCount = --m_cRef;
    TraceString2("CContAdviseSink::Release. cont: %x ref %x",m_pCont -> m_MyEID.ID.Num,m_cRef);
    LeaveCriticalSection(&m_cs); 

    // If nobody is using this object, delete it
    if (0 == ulRefCount)
    {
        // Destroy the object
        delete this;        // don't need to leave the critical section
        return ulRefCount;  // its deleted when the object is destroyed
    }
        
    return ulRefCount;
}


///////////////////////////////////////////////////////////////////////////////
//
//    QueryInterface()
//
//    Parameters                        
//          See the OLE documentation for details.
//
//    Purpose
//          Returns an interface if supported.
//      
//      
//    Return Value
//          NOERROR on success, E_NOINTERFACE if the interface can't be
//          returned.
//
STDMETHODIMP CContAdviseSink::QueryInterface(REFIID   riid, 
                                             LPVOID * ppvObj)
{
    TraceString1("CContAdviseSink::QueryInterface. cont: %x",m_pCont -> m_MyEID.ID.Num);
   
    if (IsBadWritePtr(ppvObj,sizeof LPUNKNOWN))
        return MAPI_E_INVALID_PARAMETER;
   
    *ppvObj = NULL;

    // If this is one of the two IID return an interface pointer to it
    if (riid == IID_IUnknown || riid == IID_IMAPIAdviseSink)
    {
        EnterCriticalSection(&m_cs);
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        m_cRef++;
        LeaveCriticalSection(&m_cs);
        return NOERROR;
    }

    return E_NOINTERFACE;
}


///////////////////////////////////////////////////////////////////////////////
//    CContAdviseSink::AddRef()
//
//    Parameters
//          See the OLE documentation for details.
//
//    Purpose
//      To increase the reference count of this object, and of any contained 
//      (or aggregated) object within
//
//    Return Value
//      Usage count for this object
//
STDMETHODIMP_ (ULONG) CContAdviseSink::AddRef (void)
{   
    ULONG   ulRef;
    // Increment object's internal counter
    EnterCriticalSection(&m_cs);
    ulRef = ++m_cRef;
    TraceString1("CContAdviseSink::AddRef ID cRef %d",m_cRef);
    LeaveCriticalSection(&m_cs);
       
    return ulRef;
}

///////////////////////////////////////////////////////////////////////////////
//
//    operator ==()
//
//    Parameters                        
//          lhs         [in] reference to the left-hand side entry ID
//          rhs         [in] reference to the right-hand side entry ID
//    Purpose
//          Equality operator for entry IDs.
//      
//    Return Value
//          TRUE if both are equal, FALSE otherwise.
//

BOOL operator ==(const ABPAB_EID & lhs, const ABPAB_EID & rhs)
{
    return (lhs.ID == rhs.ID  && (0 == memcmp((LPVOID) & lhs.muid, (LPVOID) & rhs.muid, sizeof MAPIUID)));
}

///////////////////////////////////////////////////////////////////////////////
//
//    operator ==()
//
//    Parameters                        
//          lhs         [in] reference to the left-hand side ID
//          rhs         [in] reference to the right-hand side ID
//    Purpose
//          Equality operator for IDs.
//      
//    Return Value
//          TRUE if both are equal, FALSE otherwise.
//

BOOL operator ==(const ABPAB_ID & lhs, const ABPAB_ID & rhs)
{

    return (lhs.Num == rhs.Num && lhs.Type == rhs.Type);
}


///////////////////////////////////////////////////////////////////////////////
//
//    CTemplateDL()
//
//    Parameters                        
//          pLogon              [in] pointer to a logon object                 
//          pCreateContainerEID [in] pointer to entry ID of 
//                                   container this DL was created in
//          pMyEID              [in] pointer to this DL's entry ID
//          ulCreateFlags       [in] creation flags
//    Purpose
//          Class constructor for a template DL, i.e. a DL we create from
//          a foreign AB's entry that supports OpenTemplateID. This class
//          is derived from CABContainer. The only difference is we have
//          a pointer to the remote IProp interface, our OpenProperty
//          retrieves the details page from disk instead of creating it on
//          the fly. 
//
//    Return Value
//          none.
CTemplateDL :: CTemplateDL(LPCABLogon pLogon, 
                           PABPAB_EID pCreateContainerEID,
                           PABPAB_EID pMyEID,
                           ULONG      ulCreateFlags,
                           LPCABCont  pParentCont,
                           BOOL       bNewEntry) : CABContainer(pLogon,
                                                                pCreateContainerEID,
                                                                pMyEID,
                                                                ulCreateFlags,
                                                                pParentCont,
                                                                bNewEntry)
{
    m_pRemoteIProp = NULL;
    m_pDetails = NULL;
    
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CTemplateDL()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class destructor.
//      
//      
//    Return Value
//          none.
//
CTemplateDL :: ~CTemplateDL()
{
    if (m_pDetails)
        m_pDetails -> Release();
}

///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Returns an object when requested a supported property
//          of type PT_OBJECT. This is an override of base class to support
//          special handling of PR_DETAILS_TABLE which we get from the
//          remote object and save to disk, instead of creating from hard-
//          wired values.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
//
STDMETHODIMP CTemplateDL :: OpenProperty(ULONG       ulPropTag,  
                                         LPCIID      lpiid,      
                                         ULONG       ulInterfaceOptions, 
                                         ULONG       ulFlags,    
                                         LPUNKNOWN * lppUnk)
{
    HRESULT     hRes = S_OK;
    Validate_IMAPIProp_OpenProperty(this,
                                    ulPropTag,         
                                    lpiid,             
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    lppUnk);            

    EnterCriticalSection(&m_cs);

    switch (ulPropTag)
    {
        case PR_DETAILS_TABLE:
            if (IID_IMAPITable != *lpiid)
            {
                hRes = MAPI_E_INTERFACE_NOT_SUPPORTED;
                break;
            }

            // if table is cached, return, else get it from the DB
            if (m_pDetails)
                *lppUnk = (LPUNKNOWN) m_pDetails;
            else
            {
                if (FAILED(hRes = GetForeignDetails(&m_pDetails)))
                    break;

                *lppUnk = (LPUNKNOWN) m_pDetails;
            }    

            m_pDetails -> AddRef();
            break;

        default:
            hRes = MAPI_E_NO_SUPPORT;
    }
    LeaveCriticalSection(&m_cs);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    GetForeignDetails()
//
//    Parameters                        
//          ppTbl   [out] pointer to a new details table
//
//    Purpose
//          Gets the entry's details table from disk and creates a display
//          table view from it.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//
STDMETHODIMP CTemplateDL :: GetForeignDetails(LPMAPITABLE * ppTbl)
{

    HRESULT         hRes     = S_OK;
    LPSRowSet       pRowSet  = NULL;
    LPSPropTagArray pTags    = NULL;
    LPTABLEDATA     pTblData = NULL;
    LPMAPITABLE     pTbl     = NULL;
    ULONG           i,
                    cTags;

    DECLARE_REC_NUM(RecNum);

    if (ppTbl)
    {
        if (IsBadWritePtr((LPVOID) ppTbl, sizeof LPMAPITABLE))
            return MAPI_E_INVALID_OBJECT;
    
        *ppTbl = NULL;
    }                 

    // open this DL's database record
    if (FAILED(hRes = Open(m_pLogon -> m_pDB, 
                           OPEN_ONLY,
                           m_MyEID.ID,
                           RecNum)))
        goto Quit;
    
    // get the raw details data
    hRes = Fetch(RecNum, fldTABLEROWS, &pRowSet);
    Close(RecNum);

    if (FAILED(hRes))
        goto Quit;
    
    // make a tag array
    cTags = pRowSet -> aRow[0].cValues;

    if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(cTags),(LPVOID *) &pTags)))
        goto Quit;

    pTags -> cValues = cTags;

    for (i = 0; i < cTags; i++)
        pTags -> aulPropTag[i] = pRowSet -> aRow[0].lpProps[i].ulPropTag;        

    // create a details table data
    if (FAILED(hRes = CreateTable(&IID_IMAPITableData,
                                  gAllocBuff,
                                  gAllocMore,
                                  gFreeBuff,
                                  NULL,
                                  TBLTYPE_SNAPSHOT,
                                  PR_ROWID,
                                  pTags,
                                  &pTblData)))
        goto Quit;

    // insert the rows retrieved from disk
    if (FAILED(hRes = pTblData -> HrModifyRows(0, pRowSet)))
        goto Quit;

    // make a view
    if (FAILED(hRes = pTblData -> HrGetView(NULL,
                                            NULL,
                                            NULL,
                                            &pTbl)))
        goto Quit;

    *ppTbl = pTbl;
Quit:
    gFreeBuff((LPVOID) pTags);    

    if (pTblData)
        pTblData -> Release();

    FreeProws(pRowSet);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    SetForeignDetails()
//
//    Parameters                        
//          pRemoteObj      [in] pointer to a DL native to another AB
//
//    Purpose
//          Gets a foreign DL's details table, extracts the rows and
//          saves them to our database. Later we can retrieve them to
//          construct a details table for the foreign entry.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CTemplateDL :: SetForeignDetails(LPMAPIPROP pRemoteObj)
{
    HRESULT hRes = SaveForeignDetails(m_pLogon -> m_pDB, 
                                      m_MyEID.ID,
                                      pRemoteObj,
                                      this,
                                      &m_pDetails);

    return hRes;
 }

// End of file for IABCONT.CPP
