///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      IABPROV.CPP
//
//  Description
//      IABProvider implementation
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
//

#define INITGUID
#define USES_IID_IABProvider

#include "abpab.h"
#include "cdbfile.h"
#include "abdb.h"
#include "abcache.h"
#include "traces.h"
#include "list.inl"

CTL3DFCN            pfnCtl3dAutoSubClass = NULL;
CTL3DFCN            pfnCtl3dRegister     = NULL;
CTL3DFCN            pfnCtl3dUnregister   = NULL;
LPALLOCATEBUFFER    gAllocBuff;  // MAPIAllocateBuffer function
LPALLOCATEMORE      gAllocMore;  // MAPIAllocateMore function    
LPFREEBUFFER        gFreeBuff;   // MAPIFreeBuffer function      


///////////////////////////////////////////////////////////////////////////////
//
//    Register3DCtls()
//
//    Parameters                        
//          hInstMod    [in]  module instance handle
//          pInstLib    [out] pointer to the library instance handle
//    Purpose
//          Dynamically loads CTL3D32 (for NT). If the library isn't
//          found, we assume we're running in Win95.
//      
//    Return Value
//          S_OK on success or if the library isn't found,
//          MAPI_E_CALL_FAILED if a function can't be bound.
//
STDMETHODIMP    Register3DCtls(HINSTANCE hInstMod,HINSTANCE * pInstLib)
{
    HRESULT     hRes = S_OK;

	// We need to load the Ctl3d DLL manually for NT
	*pInstLib = LoadLibrary("CTL3D32.DLL");

	if (! *pInstLib)        // most likely Win95
		return hRes;
	
    if (!pfnCtl3dAutoSubClass)
        pfnCtl3dAutoSubClass = (CTL3DFCN) GetProcAddress(*pInstLib,
                                                         "Ctl3dAutoSubclass");

    if (!pfnCtl3dAutoSubClass)
    {
        hRes = MAPI_E_CALL_FAILED;
        goto Quit;
    }                

    if (!pfnCtl3dRegister)
        pfnCtl3dRegister = (CTL3DFCN) GetProcAddress(*pInstLib,
                                                     "Ctl3dRegister");
    if (!pfnCtl3dRegister)
    {
        hRes = MAPI_E_CALL_FAILED;
        goto Quit;
    }                

    if (!pfnCtl3dUnregister)
        pfnCtl3dUnregister = (CTL3DFCN) GetProcAddress(*pInstLib,
                                                       "Ctl3dUnregister");

    if (!pfnCtl3dUnregister)
    {
        hRes = MAPI_E_CALL_FAILED;
        goto Quit;
    }                
	
	pfnCtl3dRegister((HINSTANCE) hInstMod); 
    pfnCtl3dAutoSubClass((HINSTANCE)hInstMod);

Quit:
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    UnRegister3DCtls()
//
//    Parameters                        
//          hInstMod    [in] module instance handle
//          hInst3DLib  [in] library instance handle
//    Purpose
//          Unloads CTL3D32.DLL, unbinds the 3D CTL functions.
//      
//    Return Value
//          S_OK on success or if the library was never loaded.
//          MAPI_E_CALL_FAILED if an error unloading
//
STDMETHODIMP    UnRegister3DCtls(HINSTANCE hInstMod,HINSTANCE hInst3DLib)
{
	if (!hInst3DLib)
		return S_OK;
		    
    if (pfnCtl3dUnregister)
        pfnCtl3dUnregister(hInstMod);
    
    pfnCtl3dAutoSubClass = NULL;
    pfnCtl3dRegister     = NULL;    
    pfnCtl3dUnregister   = NULL;  
    
    return FreeLibrary(hInst3DLib) ? S_OK : MAPI_E_CALL_FAILED;
}

///////////////////////////////////////////////////////////////////////////////
//
//    ABProviderInit()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          See the MAPI documentation for details.
//          This function provides an exception handler for
//          IABProvider object's constructor.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise. 
//
STDINITMETHODIMP ABProviderInit(HINSTANCE         hInst,
                                LPMALLOC          pMalloc,
                                LPALLOCATEBUFFER  pAllocBuff,
                                LPALLOCATEMORE    pAllocMore,
                                LPFREEBUFFER      pFreeBuff,
                                ULONG             ulFlags,
                                ULONG             ulMAPIVers,
                                ULONG           * pulABVers,
                                LPABPROVIDER    * ppABProvider)
{
    InitTraces(TRACES_CONSOLE | TRACES_NO_COM_OUTPUT);

    TraceMessage("ABProviderInit");
    HRESULT     hRes    = S_OK;
    LPCABProv   pABProv = NULL;

    *ppABProvider = NULL;

     
    
    // never use this        
    pMalloc -> Release();

    // this must come before first use of new, since the debug
    // version overloads new with MAPI allocators (for leak detection)
    gAllocBuff = pAllocBuff;
    gAllocMore = pAllocMore;
    gFreeBuff  = pFreeBuff;


    // Version check. If we're more recent than caller (we have a larger
    // version number) and we want to run with them, we should return 
   	// their version number back to them. If we're older (we have a smaller 
   	// version number), we return our version, and let them decide if they 
   	// can call us.

    if (ulMAPIVers < CURRENT_SPI_VERSION)
    {

        *pulABVers = CURRENT_SPI_VERSION;
        return MAPI_E_VERSION;
    }

    *pulABVers = CURRENT_SPI_VERSION;
    pABProv = new CABProvider(hInst);

    if (! pABProv)
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;

    if (SUCCEEDED(hRes))
    {
        pABProv -> AddRef();
        *ppABProvider = (LPABPROVIDER) pABProv;
    }
    
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CABProvider()
//
//    Parameters                        
//          hInst   [in] module instance handle
//
//    Purpose
//          Class constructor.
//      
//    Return Value
//          none.
CABProvider :: CABProvider(HINSTANCE  hInst)
{
    TraceMessage("CABProvider :: CABProvider");

    m_cRef       = 0;
    m_hInst      = hInst;
    
    InitializeCriticalSection(&m_cs);
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CABProvider()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class destructor.
//      
//      
//    Return Value

CABProvider :: ~CABProvider()
{
    TraceMessage("CABProvider :: ~CABProvider");

    DeleteCriticalSection(&m_cs);
}

///////////////////////////////////////////////////////////////////////////////
//
//    QueryInterface()
//
//    Parameters                        
//          See the OLE documentation for details.
//
//    Purpose
//          Returns an interface if supported.
//      
//      
//    Return Value
//          NOERROR on success, E_NOINTERFACE if the interface can't be
//          returned.
//
STDMETHODIMP CABProvider::QueryInterface(REFIID   riid, 
                                         LPVOID * ppvObj)
{
    TraceMessage("CABProvider::QueryInterface");
   
    if (IsBadWritePtr(*ppvObj,sizeof LPUNKNOWN))
        return MAPI_E_INVALID_PARAMETER;
   
    *ppvObj = NULL;
    // If this is one of our two IID return, interface pointer
    if (riid == IID_IUnknown || riid == IID_IABProvider)
    {
        EnterCriticalSection(&m_cs);
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        m_cRef++;
        LeaveCriticalSection(&m_cs);
        return NOERROR;
    }

    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CABProvider::AddRef()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          To increase the reference count of this object,
//          and of any contained (or aggregated) object within
//
//    Return Value
//          Usage count for this object
//
STDMETHODIMP_ (ULONG) CABProvider::AddRef (void)
{   
    ULONG   ulRef;

    // Increment object's internal counter
    EnterCriticalSection(&m_cs);
    ulRef = ++m_cRef;
    TraceString1("CABProvider :: AddRef cRef %d",m_cRef);
    LeaveCriticalSection(&m_cs);
       
    return ulRef;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CABProvider::Release()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          Decreases the usage count of this object. If the count is 
//          zero, object deletes itself.
//
//    Return Value
//          Usage of this object
//
STDMETHODIMP_ (ULONG) CABProvider::Release (void)
{
    ULONG ulRefCount;

    EnterCriticalSection(&m_cs);
    ulRefCount = --m_cRef;
    LeaveCriticalSection(&m_cs); 

    TraceString1("CABProvider :: Release", ulRefCount);

    // If nobody is using this object, delete it
    if (0 == ulRefCount)
    {
        // Destroy the object
        delete this;        // don't need to leave the critical section
        return ulRefCount;  // its deleted when the object is destroyed
    }
        
    return ulRefCount;
}



///////////////////////////////////////////////////////////////////////////////
//
//    CABProvider::Shutdown()
//
//    Parameters
//          Refer to MAPI Documentation on this method
//
//    Purpose
//          To terminate all sessions attached to this provider.
//
//    Return Value
//      S_OK                    If no error
//      MAPI_E_UNKNOWN_FLAGS    If flags are passed to this method
//
STDMETHODIMP CABProvider::Shutdown (ULONG * pulFlags)
{
    TraceMessage("CABProvider :: Shutdown");
    CheckParameters_IABProvider_Shutdown(this,pulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    OpenDatabase()
//
//    Parameters                        
//          ulFlags     [in]     UI flags, same as in ServiceEntry
//          hWnd        [in]     application's window handle
//          DBUID       [in/out] reference to the database UID
//          ppDB        [out]    ptr to ptr to new database object
//
//    Purpose
//          Gets the DB filename, name, comment from the profile section
//          and attempts to open the DB file. If the file can't be opened,
//          a dialog is displayed allowing user to select a new file. On
//          success, a new database object is returned in *ppDB. Since the
//          properties can change even if the file isn't newly created, 
//          we go ahead and create the ROOT and DIR records and save the 
//          profile properties in them, as if they are newly created. If the
//          file must be created, we have to create the ROOT and DIR records
//          first.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CABProvider :: OpenDatabase(ULONG          ulFlags, 
                                         HWND           hWnd,
                                         MAPIUID      & DBUID,
                                         LPMAPISUP      pSupObj,
                                         CDataBase   ** ppDB)
{
    HRESULT             hRes       = S_OK;
    LPSPropValue        pProps     = NULL;
    CDataBase         * pDB        = NULL;
    LPSPropProblemArray pPropErrs  = NULL;
    LPPROFSECT          pProfSect  = NULL;
    HINSTANCE           hInst3DLib = NULL;
    ULONG               ulPropCount;
    ABPAB_ID            NameMapID = {NAME_MAP_ID,0};

    DECLARE_REC_NUM(DirRecNum);
    DECLARE_REC_NUM(RootRecNum);
    DECLARE_REC_NUM(NameIDRec);


    TraceMessage("OpenDatabase");

    *ppDB = NULL;
    Register3DCtls(m_hInst, &hInst3DLib);
    
    // get the saved profile properties
    if (FAILED(hRes = pSupObj -> OpenProfileSection(NULL,
                                                      MAPI_MODIFY,
                                                      &pProfSect)))
       goto Quit;

    if (FAILED(hRes = pProfSect -> GetProps((LPSPropTagArray) &sptLogonTags, 
                                            0,
                                            &ulPropCount,
                                            &pProps)))
        goto Quit;

    // Yup, this is a kludge. pProps[LOGON_DB_PATH] is set to a NULL string
    // but later allocated with MapiAllocateMore after you get the real path
    // in GetValidFileName. This guarantees that the open file dialog is displayed
    // since GetValidFileName on a null string always brings the dialog up.    
    if (PR_SMP_AB_PATH != pProps[LOGON_DB_PATH].ulPropTag)
    {
        pProps[LOGON_DB_PATH].ulPropTag = PR_SMP_AB_PATH;  // could be error if not found
        pProps[LOGON_DB_PATH].Value.LPSZ = NULL_STRING;    
    }

    if (PR_DISPLAY_NAME != pProps[LOGON_DISP_NAME].ulPropTag)
        pProps[LOGON_DISP_NAME].Value.lpszA = SZ_PSS_SAMPLE_AB;

    if (PR_COMMENT != pProps[LOGON_COMMENT].ulPropTag)
        pProps[LOGON_COMMENT].Value.LPSZ = NULL_STRING;

    if (FAILED(hRes = GetValidFileName(&pProps[LOGON_DB_PATH].Value.LPSZ,
                                       hWnd,
                                       (LPVOID) pProps,
                                       ulFlags)))
        goto Quit;
     
    // try to open existing first, if we can't, create DB file
    if (FAILED(CommonOpenFile(pProps[LOGON_DB_PATH].Value.LPSZ,
                              NULL,
                              NULL,
                              DBUID,
                              &pDB,
                              FALSE)))
    {  
        if (FAILED(hRes = pSupObj -> NewUID(&DBUID)))
            goto Quit;
                                                       
        if (FAILED(hRes = CreateDataBaseFile(pProps[LOGON_DB_PATH].Value.LPSZ, 
                                             pProps[LOGON_DISP_NAME].Value.LPSZ,
                                             pProps[LOGON_COMMENT].Value.LPSZ, 
                                             DBUID,
                                             &pDB)))
            goto Quit;
    }
    
    // set up our name-to_ID mapping table
    if (FAILED(hRes = Open(pDB,OPEN_ONLY,NameMapID, NameIDRec)))
    {
        SizedSRowSet(2,sr);                

        sr.aRow[0].lpProps = sr.aRow[1].lpProps = NULL;

        if (FAILED(hRes = Open(pDB,CREATE_IF_NOT_EXIST,NameMapID, NameIDRec))) 
            goto Quit;

        // initialize the table w/ 2 rows: our starting ID and PR_HOME_PAGE
        // use HrQueryRow to allocate an empty row set for us
        if (FAILED(hRes = gAllocBuff(sizeof SPropValue * NUM_MAP_PROPS, 
                                     (LPVOID *) &sr.aRow[0].lpProps)))
            goto Quit;
        
        sr.aRow[0].lpProps[MAP_ID].ulPropTag      = PR_MAP_ID;
        sr.aRow[0].lpProps[MAP_ID].Value.bin.cb   = BASE_NAMED_ID;

        sr.aRow[0].lpProps[MAP_UID].ulPropTag     = PR_MAP_UID;
        sr.aRow[0].lpProps[MAP_UID].Value.bin.cb  = sizeof PS_ABPAB;
        sr.aRow[0].lpProps[MAP_UID].Value.bin.lpb = (LPBYTE) &PS_ABPAB;
        

        // we don't have a Unicode string name
        sr.aRow[0].lpProps[MAP_SZNAME].ulPropTag     = PR_MAP_SZNAME;
        sr.aRow[0].lpProps[MAP_SZNAME].Value.lpszW   = L"";
        
        // The first row in our prop set is at 0x8000, it tells our guid
        // and keeps the counter for the next ID 
        sr.aRow[0].lpProps[MAP_ULNAME].ulPropTag     = PR_MAP_ULNAME;
        sr.aRow[0].lpProps[MAP_ULNAME].Value.l       = BASE_NAMED_ID;
        
        // This column gives the next ID for our prop set. Only found 
        // in the first row of our prop set. In other rows its the ulKind
        sr.aRow[0].lpProps[MAP_NEXTID].ulPropTag     = PR_MAP_NEXTID;
        sr.aRow[0].lpProps[MAP_NEXTID].Value.l       = BASE_NAMED_ID + 2;
        
        sr.aRow[0].cValues = NUM_MAP_PROPS;
        
        if (FAILED(hRes = gAllocBuff(sizeof SPropValue * NUM_MAP_PROPS, 
                                     (LPVOID *) &sr.aRow[1].lpProps)))
        {
            gFreeBuff(sr.aRow[1].lpProps);
            goto Quit;
        }
        
        // ID for this name
        sr.aRow[1].lpProps[MAP_ID].ulPropTag      = PR_MAP_ID;
        sr.aRow[1].lpProps[MAP_ID].Value.bin.cb   = BASE_NAMED_ID + 1;

        sr.aRow[1].lpProps[MAP_UID].ulPropTag     = PR_MAP_UID;
        sr.aRow[1].lpProps[MAP_UID].Value.bin.cb  = sizeof PS_ABPAB;
        sr.aRow[1].lpProps[MAP_UID].Value.bin.lpb = (LPBYTE) &PS_ABPAB;
        
        // we don't have a Unicode string name
        sr.aRow[1].lpProps[MAP_SZNAME].ulPropTag   = PR_MAP_SZNAME;
        sr.aRow[1].lpProps[MAP_SZNAME].Value.lpszW = L"";
        
        // The first row in our prop set is at 0x8000, it tells our guid
        // and keeps the counter for the next ID 
        sr.aRow[1].lpProps[MAP_ULNAME].ulPropTag   = PR_MAP_ULNAME;
        sr.aRow[1].lpProps[MAP_ULNAME].Value.l     = URL_NAME;
        
        // This column gives the next ID for our prop set. Only found 
        // in the first row of our prop set. In other rows its the ulKind
        sr.aRow[1].lpProps[MAP_KIND].ulPropTag     = PR_MAP_KIND;
        sr.aRow[1].lpProps[MAP_KIND].Value.l       = MNID_ID;
        
        sr.aRow[1].cValues = NUM_MAP_PROPS;
        
        
        sr.cRows = 2;
        hRes = Update(NameIDRec,fldTABLEROWS,(LPSRowSet)&sr);
        gFreeBuff(sr.aRow[0].lpProps);      
        gFreeBuff(sr.aRow[1].lpProps);      
        
        if (FAILED(hRes))
            goto Quit;
    }

    if (FAILED(hRes = pProfSect -> SetProps(LOGON_NUM_PROPS,
                                            pProps,
                                            &pPropErrs)))
        goto Quit;                    

    // initialize the ROOT and DIR containers' records
    if (FAILED(hRes = Open(pDB,CREATE_IF_NOT_EXIST,ROOT_ID, RootRecNum)))
        goto Quit;

    if (SUCCEEDED(hRes = Open(pDB,CREATE_IF_NOT_EXIST,DIR_ID, DirRecNum)))
    {   
        // set default properties
        CFlatPropList   PropList;

        if (SUCCEEDED(hRes = InsertRec(pDB, RootRecNum, DirRecNum)))
        {   // create new EID for DIR
            ABPAB_EID   EID;
            SPropValue  spv;

            ZeroMemory((LPVOID) &EID, sizeof ABPAB_EID);
            memcpy((LPVOID) &(EID.muid), (LPVOID)& DBUID, sizeof MAPIUID);
            EID.ID.Num  = DIR_ID_NUM;
            EID.ID.Type = MAPI_ABCONT;

            // save config props to a list                     
            for (int i = LOGON_DB_PATH; i <= LOGON_COMMENT; i++)
                PropList.Insert(&pProps[i]);
            
            // set them and the defaults in the DIR's prop stream
            SetDefContProps(&EID,&PropList);

            if (FAILED(hRes = Update(DirRecNum,fldPROPS, &PropList)))
                goto Quit;
            
            // now set the ROOT's config/default props
            // only the entry ID is different,replace it
            EID.ID.Num        = ROOT_ID_NUM;
            spv.ulPropTag     = PR_ENTRYID;
            spv.Value.bin.cb  = sizeof EID;
            spv.Value.bin.lpb = (LPBYTE) & EID;
            PropList.Insert(&spv);

            if (FAILED(hRes = Update(RootRecNum,fldPROPS, &PropList)))
                goto Quit;
        }
    }
    
Quit:
    Close(NameIDRec);
    Close(DirRecNum);
    Close(RootRecNum);
    UnRegister3DCtls(m_hInst, hInst3DLib);    

    if (pProfSect)
        pProfSect -> Release();

    gFreeBuff(pProps);

    if (FAILED(hRes))
        if (pDB)
        {
            pDB -> Release();
            pDB = NULL;
        }

    gFreeBuff(pPropErrs);
    *ppDB = pDB;

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CABProvider::Logon()
//
//    Parameters
//          Refer to MAPI Documentation on this method
//
//    Purpose
//          Display the logon dialog to show the options saved in the
//          profile for this provider and allow changes to it. Save 
//          new configuration settings back in the profile. Try and 
//          open the DB file, or create it if not found. Create a 
//          new CABLogon object and return it to the caller. We try
//          to get the logon UID from the profile (the UID is unique
//          to the database file), create a new one if not found. Set
//          the provider UID.   We catch any exceptions thrown in the
//          ABLogon constructor.
//
//    Return Value
//      S_OK                        If no error
//      MAPI_E_NO_ACCESS            If transport is logged with the 
//                                  LOGON_NO_DIALOG flag
//      MAPI_E_NOT_ENOUGH_MEMORY    If could not allocate new CXPLogon object
//      hResult                     With any other error code
//
STDMETHODIMP CABProvider :: Logon (LPMAPISUP        lpMAPISup,
                                   ULONG            ulUIParam,
                                   LPTSTR           lpszProfileName,
                                   ULONG            ulFlags,
                                   ULONG *          lpulpcbSecurity,
                                   LPBYTE *         lppbSecurity,
                                   LPMAPIERROR *    lppMAPIError,
                                   LPABLOGON *      lppABLogon)
{
    HRESULT     hRes      = S_OK;    
    LPCABLogon  pLogonObj = NULL;
    MAPIUID     UID       = {0}; 
    PDATABASE   pDB       = NULL;

    TraceMessage("CABProvider :: Logon");    

    CheckParameters_IABProvider_Logon(this,
                                      lpMAPISup,       
                                      ulUIParam,      
                                      lpszProfileName,
                                      ulFlags,        
                                      lpulpcbSecurity,
                                      lppbSecurity,   
                                      lppMAPIError,   
                                      lppABLogon);

     if (ulFlags & MAPI_UNICODE)
        return MAPI_E_BAD_CHARWIDTH;

    *lppABLogon   = NULL;
    *lppMAPIError = NULL;   

    // IAddrBook expects us to AddRef, even tho' we don't keep it around
    lpMAPISup -> AddRef();

    // this opens/creates database and root and DIR records
    if (FAILED(hRes = OpenDatabase(ulFlags, 
                                   (HWND)ulUIParam,
                                   UID,
                                   lpMAPISup,
                                   &pDB)))
        goto Quit;
 
    pLogonObj = new CABLogon(ulFlags,
                             m_hInst,
                             lpMAPISup,
                             UID,
                             pDB);

        
    if (!pLogonObj)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }    

    if (FAILED(hRes = pLogonObj -> Init(&GDLDetailsPage,
                                        &GConfigDetailsPage,
                                        &GUserDetailsPage)))
        goto Quit;                                            


    if (FAILED(hRes = lpMAPISup -> SetProviderUID((LPMAPIUID) &UID, 0)))
        goto Quit;

    hRes = pLogonObj -> SetStatusRow(TRUE);

Quit:
    if (FAILED(hRes))
    {
        if (pLogonObj)
        {
            // releases stuff not destroyed in destructor
            pLogonObj -> Logoff(0);    
            delete pLogonObj;
        }
    }
    else
    {
        //for caller
        pLogonObj -> AddRef();
        *lppABLogon = (LPABLOGON) pLogonObj;
    }

    if (pDB)
        pDB -> Release();

    // we don't use this, but CABLogon copies and AddRefs
    if (lpMAPISup)
        lpMAPISup -> Release();
    
    return hRes;
}                                        

// End of file for IABPROV.CPP




