///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      IPROP.CPP
//
//  Description
//      IMailUser implementation
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
//
#include "abpab.h"
#include "abcache.h"
#include "abdb.h"
#include "itable.h"
#include "list.inl"
#include <mapiwin.h>


///////////////////////////////////////////////////////////////////////////////
//
//    CMailUser()
//
//    Parameters                        
//          pLogon              [in] pointer to logon object           
//          pCreateContainerEID [in] pointer to entry ID of container
//                                   object created in
//          pEID                [in] pointer to object's entry ID
//          ulCreateFlags       [in] creation flags
//          pParentCont         [in] pointer to parent container
//    Purpose
//          Class constructor.
//      
//    Return Value
//          none.
//
 CMailUser :: CMailUser(LPCABLogon  pLogon, 
                        PABPAB_EID  pCreateContainerEID,
                        PABPAB_EID  pEID,
                        ULONG       ulCreateFlags,
                        LPCABCont   pParentCont,
                        BOOL        bNewEntry) : CBaseProp(pLogon,             
                                                           pCreateContainerEID,
                                                           pEID,             
                                                           ulCreateFlags,      
                                                           pParentCont,        
                                                           bNewEntry)
{
    TraceString1("CMailUser::CMailUser %x", pEID -> ID.Num);
    m_szError[0] = '\0';
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CMailUser()
//
//    Parameters                        
//          none.
//
//    Purpose
//          class destructor.
//      
//    Return Value
//          none.
//
CMailUser :: ~CMailUser()
{
    TraceString1("CMailUser::~CMailUser %x", m_MyEID.ID.Num);
}

///////////////////////////////////////////////////////////////////////////
//    CMailUser::AddRef()
//
//    Parameters
//          See the OLE documentation for details.
//
//    Purpose
//      To increase the reference count of this object, and of any contained 
//      (or aggregated) object within
//
//    Return Value
//      Usage count for this object
//
STDMETHODIMP_ (ULONG) CMailUser::AddRef (void)
{   
    ULONG   ulRef;

    // Increment object's internal counter
    EnterCriticalSection(&m_cs);
    ulRef = ++m_cRef;
    TraceString2("CMailUser::AddRef ID %x, cRef %d",m_MyEID.ID.Num,m_cRef);
    LeaveCriticalSection(&m_cs);
       
    return ulRef;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CBaseProp::Release()
//
//    Parameters
//          See the OLE documentation for details.
//
//    Purpose
//      Decreases the usage count of this object. When the count is zero,
//      the object deletes itself.
//
//    Return Value
//      Usage of this object
//
STDMETHODIMP_ (ULONG) CMailUser::Release (void)
{
    // decrement the object's internal counter

    ULONG ulRefCount;
    EnterCriticalSection(&m_cs);
    ulRefCount = --m_cRef;
    TraceString2("CMailUser::Release ID %x, cRef %d",m_MyEID.ID.Num,m_cRef);
    LeaveCriticalSection(&m_cs); 

    // If nobody is using this object, delete it
    if (0 == ulRefCount)
    {
        // Destroy the object
        delete this;        // don't need to leave the critical section
        return ulRefCount;  // its deleted when the object is destroyed
    }
        
    return ulRefCount;
}

///////////////////////////////////////////////////////////////////////////////
//
//    QueryInterface()
//
//    Parameters                        
//          See the OLE documentation for details.
//
//    Purpose
//          Returns an interface if supported.
//      
//      
//    Return Value
//          NOERROR on success, E_NOINTERFACE if the interface can't be
//          returned.
//

STDMETHODIMP CMailUser ::QueryInterface(REFIID   riid, 
                                        LPVOID * ppvObj)
{
    TraceMessage("CMailUser :: QueryInterface");

    if (IsBadWritePtr(ppvObj,sizeof LPUNKNOWN))
        return MAPI_E_INVALID_PARAMETER;

    *ppvObj = NULL;

    // If this is one of the two IID return an interface pointer to it
    if (riid == IID_IUnknown  || 
        riid == IID_IMAPIProp ||
        riid == IID_IMailUser)
    {
        EnterCriticalSection(&m_cs);
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        m_cRef++;
        LeaveCriticalSection(&m_cs);
        return NOERROR;
    }

    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Init()
//
//    Parameters                        
//          none
//
//    Purpose
//          Initializes the new mailuser, the base IProp and sets default 
//          properties. We don't have to add it to the database in
//          a 'limbo' state because mailusers are always leaf nodes of a
//          container (DL) hierarchy.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CMailUser :: Init()
{
    HRESULT     hRes = S_OK;

    if (FAILED(hRes = CBaseProp::Init()))
        return hRes;

    if (FAILED(hRes = SetDefaultProps(&m_MyEID)))
        return hRes;
    
    AddRef();

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Sets detailed error information based on hRes.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CMailUser :: SetLastError(HRESULT       hRes,    
                                       ULONG         ulLine,    
                                       LPSTR         szFile)
{
    if (FAILED(hRes))
    {
        EnterCriticalSection(&m_cs);
        m_ulContext = ulLine + IE_IABLOGON;
        lstrcpyn(m_szError,szFile,MAX_PATH);
        m_szError[MAX_PATH - 1] = '\0';
        LeaveCriticalSection(&m_cs);
    }
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetLastError()
//
//    Parameters                        
//          See the MAPI Documentation for details.
//
//    Purpose
//          Gets detailed error information based on hResult.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CMailUser :: GetLastError(HRESULT       hResult,    
                                       ULONG         ulFlags,    
                                       LPMAPIERROR * lppMAPIError)
{
    HRESULT hRes;
    TraceMessage("CMailUser :: GetLastError");
    Validate_IABLogon_GetLastError(this,
                                   hResult,     
                                   ulFlags,     
                                   lppMAPIError);
    EnterCriticalSection(&m_cs);
    hRes =::GetLastError(m_szError,SZ_PSS_SAMPLE_AB,m_ulContext,lppMAPIError);
    LeaveCriticalSection(&m_cs);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Returns an object when requested a supported property
//          of type PT_OBJECT.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CMailUser :: OpenProperty(ULONG       ulPropTag,  
                                       LPCIID      lpiid,      
                                       ULONG       ulInterfaceOptions, 
                                       ULONG       ulFlags,    
                                       LPUNKNOWN * lppUnk)
{
    HRESULT     hRes = S_OK;

    TraceMessage("CMailUser :: OpenProperty");
    Validate_IMAPIProp_OpenProperty(this,                              
                                    ulPropTag,         
                                    lpiid,             
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    lppUnk);            
    EnterCriticalSection(&m_cs);
    switch (ulPropTag)
    {
        case PR_DETAILS_TABLE:
            if (IID_IMAPITable != *lpiid)
            {
                hRes = MAPI_E_INTERFACE_NOT_SUPPORTED;
                break;
            }                    
            hRes = m_pLogon       -> 
                   m_pUserDetails -> BuildDisplayTable((LPMAPITABLE *) lppUnk);

            break;
    //!!    case PR_HOMEPAGE: 
    //        {
    //            CHomePageButton * pButton = new CHomePageButton(this);
    //            
    //            if (!pButton)
    //                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
    //            else
    //            {
    //                *lppUnk = (LPUNKNOWN) pButton;
    //                (*lppUnk) -> AddRef();
    //            }
    //        }
    //        break;

        default:
            hRes = MAPI_E_NO_SUPPORT;
            break;
    }

    LeaveCriticalSection(&m_cs);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    Normalize()
//
//    Parameters                        
//          szIn    [in]  string to normalize
//          ulLen   [out] length of output string
//
//    Purpose
//          Normalizes an input string; i.e. removes leading, trailing
//          and excess embedded whitespace, and converts to upper case. 
//          The normalized string is placed back in szIn.
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
//
void Normalize(LPTSTR szIn, ULONG * pulLen)
{
    ULONG       ulTemp = 0;
    LPTSTR      szTemp = NULL,
                szSave = szIn;  //GetNextToken alters input, save start address
    
    *pulLen = 0;
    
    for (szTemp = GetNextToken(&szIn);
         szTemp;
         szTemp = GetNextToken(&szIn))
    {
        ulTemp = lstrlen(szTemp);
        memcpy((LPVOID) &szSave[*pulLen],(LPVOID) szTemp, ulTemp);
        *pulLen += ulTemp;
    }

    szSave[*pulLen] = '\0';
    return;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetDefaultProps()
//
//    Parameters                        
//          pEID        [in]  pointer to object's entry ID     
//
//    Purpose
//          Initializes a MailUser's property
//          cache with default values.      
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP CMailUser :: SetDefaultProps(PABPAB_EID pEID)
{
    HRESULT     hRes = S_OK;

    enum {EID,
          INST_KEY,
          REC_KEY,
          SEARCH_KEY,
          OBJ_TYPE,
          DISP_TYPE, 
          SIG,
          NUM_PROPS
         };   

    m_fDirtyFlags |= PROPS;

    memcpy((LPBYTE) &m_MyEID,(LPBYTE) pEID, sizeof ABPAB_EID);

    SPropValue      Props[NUM_PROPS];
    
    Props[EID].ulPropTag            = PR_ENTRYID;
    Props[EID].Value.bin.cb         = sizeof(*pEID);
    Props[EID].Value.bin.lpb        = (LPBYTE) pEID;

    Props[INST_KEY].ulPropTag       = PR_INSTANCE_KEY;
    Props[INST_KEY].Value.bin.cb    = sizeof(ULONG);
    Props[INST_KEY].Value.bin.lpb   = (LPBYTE) & (pEID -> ID.Num);

    
    Props[REC_KEY].ulPropTag        = PR_RECORD_KEY;
    Props[REC_KEY].Value.bin.cb     = sizeof(*pEID);
    Props[REC_KEY].Value.bin.lpb    = (LPBYTE) pEID;

    Props[SEARCH_KEY].ulPropTag     = PR_SEARCH_KEY;
    Props[SEARCH_KEY].Value.bin.cb  = 1;
    Props[SEARCH_KEY].Value.bin.lpb = (LPBYTE) "";

    Props[OBJ_TYPE].ulPropTag       = PR_OBJECT_TYPE;
    Props[OBJ_TYPE].Value.l         = MAPI_MAILUSER;

    Props[DISP_TYPE].ulPropTag      = PR_DISPLAY_TYPE;
    Props[DISP_TYPE].Value.l        = DT_MAILUSER;

    Props[SIG].ulPropTag            = PR_MAPPING_SIGNATURE;
    Props[SIG].Value.bin.cb         = sizeof MAPIUID;
    Props[SIG].Value.bin.lpb        = (LPBYTE) & (pEID -> muid);

    for (int i = EID; i < NUM_PROPS; i++)
        if (FAILED(hRes = m_pPropCache -> m_pProps -> Insert(&Props[i])))
            return hRes;

    return hRes;
} 


///////////////////////////////////////////////////////////////////////////////
//  
//      The following functions all delegate down to the CBaseProp
//      object. This is necessary because the OLE COM interfaces can't
//      be derived virtual. (e.g. class IABContainer : virtual public 
//      IMAPIContainer) so if we also derive from a MAPI base class, there 
//      will be ambiguities between our base IMAPIProp implementation and 
//      the abstract IMAPIProp base in IMAPIContainer

///////////////////////////////////////////////////////////////////////////////
//
//    SaveChanges()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  SaveChanges(ULONG  ulFlags)                        
{                                                                                     
    return CBaseProp::SaveChanges(ulFlags);             
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    GetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  GetProps(LPSPropTagArray  lpPropTagArray,                     
                                     ULONG            ulFlags,                            
                                     ULONG  *         lpcValues,                          
                                     LPSPropValue  *  lppPropArray)                       
{                                                                                     
    return CBaseProp::GetProps(lpPropTagArray,           
                           ulFlags,                  
                           lpcValues,                
                           lppPropArray);             
}                                                                                     

///////////////////////////////////////////////////////////////////////////////
//
//    SetProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CMailUser :: SetProps(ULONG                 cValues,    
                                   LPSPropValue          lpPropArray,
                                   LPSPropProblemArray * lppProblems)
{                                                                                     
    return CBaseProp::SetProps(cValues,lpPropArray, lppProblems);
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetPropList()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
                                                                                      
STDMETHODIMP  CMailUser ::  GetPropList(ULONG             ulFlags,
                                        LPSPropTagArray * lppPropTagArray)
{                                                                                     
    return CBaseProp::GetPropList(ulFlags,lppPropTagArray);       
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    DeleteProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  DeleteProps(LPSPropTagArray       lpPropTagArray,                  
                                        LPSPropProblemArray * lppProblems)                     
{                                                                                     
    return CBaseProp::DeleteProps(lpPropTagArray,        
                                   lppProblems);
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    CopyTo()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  CopyTo(ULONG                 ciidExclude,                          
                                   LPCIID                rgiidExclude,                         
                                   LPSPropTagArray       lpExcludeProps,                       
                                   ULONG                 ulUIParam,                            
                                   LPMAPIPROGRESS        lpProgress,                           
                                   LPCIID                lpInterface,                          
                                   LPVOID                lpDestObj,                            
                                   ULONG                 ulFlags,                              
                                   LPSPropProblemArray * lppProblems)                          
{                                                                                     
    return CBaseProp::CopyTo(ciidExclude,                
                             rgiidExclude,               
                             lpExcludeProps,             
                             ulUIParam,                  
                             lpProgress,                 
                             lpInterface,                
                             lpDestObj,                  
                             ulFlags,                    
                             lppProblems);                
}                                                                                     

///////////////////////////////////////////////////////////////////////////////
//
//    CopyProps()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
                                                                                      
STDMETHODIMP  CMailUser ::  CopyProps(LPSPropTagArray       lpIncludeProps,                    
                                      ULONG                 ulUIParam,                         
                                      LPMAPIPROGRESS        lpProgress,                        
                                      LPCIID                lpInterface,                       
                                      LPVOID                lpDestObj,                         
                                      ULONG                 ulFlags,                           
                                      LPSPropProblemArray * lppProblems)                       
{                                                                                     
    return CBaseProp::CopyProps(lpIncludeProps,          
                                ulUIParam,               
                                lpProgress,              
                                lpInterface,             
                                lpDestObj,               
                                ulFlags,                 
                                lppProblems);             
}                                                                                        
                                                                                         
///////////////////////////////////////////////////////////////////////////////
//
//    GetNamesFromIDs()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  GetNamesFromIDs(LPSPropTagArray * lppPropTags,                 
                                            LPGUID            lpPropSetGuid,               
                                            ULONG             ulFlags,                     
                                            ULONG  *          lpcPropNames,                
                                            LPMAPINAMEID **   lpppPropNames)               
{                                                                                     
    return CBaseProp::GetNamesFromIDs(lppPropTags,       
                                      lpPropSetGuid,     
                                      ulFlags,           
                                      lpcPropNames,      
                                      lpppPropNames);     
}                                                                                     
                                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//    GetIDsFromNames()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//         Delegates to base prop interface.
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//

STDMETHODIMP  CMailUser ::  GetIDsFromNames(ULONG             cPropNames,                  
                                            LPMAPINAMEID    * lppPropNames,                
                                            ULONG             ulFlags,                     
                                            LPSPropTagArray * lppPropTags)                 
{                                                                                     
    return CBaseProp::GetIDsFromNames(cPropNames,        
                                      lppPropNames,      
                                      ulFlags,           
                                      lppPropTags);       
}                                                                                        
///////////////////////////////////////////////////////////////////////////////
//
//    CTemplateUser()
//
//    Parameters                        
//          pLogon              [in] pointer to a logon object                 
//          pCreateContainerEID [in] pointer to entry ID of 
//                                   container this DL was created in
//          pMyEID              [in] pointer to this DL's entry ID
//          ulCreateFlags       [in] creation flags
//          pCreateCont         [in] pointer to container created in
//    Purpose
//          Class constructor for a template User i.e. a user we create from
//          a foreign AB's entry that supports OpenTemplateID. This class
//          is derived from CMailUser. The only difference is we have
//          a pointer to the remote IProp interface, our OpenProperty
//          retrieves the details page from disk instead of creating it on
//          the fly. 
//
//    Return Value
//          none.
//
//
CTemplateUser :: CTemplateUser(LPCABLogon pLogon, 
                               PABPAB_EID pCreateContainerEID,
                               PABPAB_EID pEID,
                               ULONG      ulCreateFlags,
                               LPCABCont  pCreateCont,
                               BOOL       bNewEntry) : CMailUser(pLogon,
                                                                 pCreateContainerEID,
                                                                 pEID,
                                                                 ulCreateFlags,
                                                                 pCreateCont,
                                                                 bNewEntry)
{
    m_pRemoteIProp = NULL;
    m_pDetails = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CTemplateUser()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class destructor.
//      
//    Return Value
//
CTemplateUser :: ~CTemplateUser()
{
    if (m_pDetails)
        m_pDetails -> Release();
}


///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          See the MAPI documentation for details.
//
//    Purpose
//          Returns an object when requested a supported property
//          of type PT_OBJECT. This is an override of base class to support
//          special handling of PR_DETAILS_TABLE which we get from the
//          remote object and save to disk, instead of creating from hard-
//          wired values.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CTemplateUser :: OpenProperty(ULONG       ulPropTag,  
                                           LPCIID      lpiid,      
                                           ULONG       ulInterfaceOptions, 
                                           ULONG       ulFlags,    
                                           LPUNKNOWN * lppUnk)
{                                         
    HRESULT     hRes = S_OK;              
    TraceMessage("CTemplateUser :: OpenProperty");
    Validate_IMAPIProp_OpenProperty(this,
                                    ulPropTag,         
                                    lpiid,             
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    lppUnk);            
    EnterCriticalSection(&m_cs);                       

    switch (ulPropTag)
    {
        case PR_DETAILS_TABLE:
            if (IID_IMAPITable != *lpiid)
            {
                M_FAIL(hRes = MAPI_E_INTERFACE_NOT_SUPPORTED);
                break;
            }
            // if table is cached, return, else get it from the DB
            if (m_pDetails)
                *lppUnk = (LPUNKNOWN) m_pDetails;
            else
            {
                if (M_FAIL(hRes = GetForeignDetails(&m_pDetails)))
                    return hRes;

                *lppUnk = (LPUNKNOWN) m_pDetails;
            }    

            m_pDetails -> AddRef();
            break;

        default: 
            M_FAIL(hRes = MAPI_E_NO_SUPPORT);
            return hRes;
       
    }
    
    LeaveCriticalSection(&m_cs);
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetForeignDetails()
//
//    Parameters                        
//          ppTbl   [out] pointer to a new details table
//
//    Purpose
//          Gets the entry's details table from disk and creates a display
//          table view from it.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//

STDMETHODIMP CTemplateUser :: GetForeignDetails(LPMAPITABLE * ppTbl)
{

    HRESULT     hRes = S_OK;
    LPSRowSet   pRowSet = NULL;
    ULONG       i,
                cTags;
    LPSPropTagArray pTags = NULL;
    LPTABLEDATA pTblData = NULL;
    LPMAPITABLE pTbl = NULL;
    
    DECLARE_REC_NUM(RecNum);


    if (IsBadWritePtr((LPVOID) ppTbl, sizeof LPMAPITABLE))
        return MAPI_E_INVALID_PARAMETER;
      
    if (FAILED(hRes = Open(m_pLogon -> m_pDB, 
                           OPEN_ONLY,
                           m_MyEID.ID,
                           RecNum)))
        goto Quit;
    
    hRes = Fetch(RecNum, fldTABLEROWS, &pRowSet);
    Close(RecNum);

    if (FAILED(hRes))
        goto Quit;
    
    cTags = pRowSet -> aRow[0].cValues;

    if (FAILED(hRes = gAllocBuff(CbNewSPropTagArray(cTags), //sizeof SPropTagArray + (cTags - 1) * sizeof ULONG,
                                 (LPVOID *) & pTags)))
        goto Quit;

    pTags -> cValues = cTags;

    for (i = 0; i < cTags; i++)
        pTags -> aulPropTag[i] = pRowSet -> aRow[0].lpProps[i].ulPropTag;        

    if (FAILED(hRes = CreateTable(&IID_IMAPITableData,
                                  gAllocBuff,
                                  gAllocMore,
                                  gFreeBuff,
                                  NULL,
                                  TBLTYPE_SNAPSHOT,
                                  PR_ROWID,
                                  pTags,
                                  &pTblData)))
        goto Quit;

    if (FAILED(hRes = pTblData -> HrModifyRows(0, pRowSet)))
        goto Quit;

    if (FAILED(hRes = pTblData -> HrGetView(NULL,
                                            NULL,
                                            NULL,
                                            &pTbl)))
        goto Quit;

    *ppTbl = pTbl;
Quit:
    gFreeBuff((LPVOID) pTags);    

    if (pTblData)
        pTblData -> Release();

    if (pRowSet)
        FreeProws(pRowSet);

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    SetForeignDetails()
//
//    Parameters                        
//          pRemoteObj      [in] pointer to a DL native to another AB
//
//    Purpose
//          Gets a foreign DL's details table, extracts the rows and
//          saves them to our database. Later we can retrieve them to
//          construct a details table for the foreign entry.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
//
STDMETHODIMP CTemplateUser :: SetForeignDetails(LPMAPIPROP pRemoteObj)
{
    HRESULT hRes = SaveForeignDetails(m_pLogon -> m_pDB, 
                                      m_MyEID.ID,
                                      pRemoteObj,
                                      this,
                                      &m_pDetails);

    return hRes;
    

/*
    HRESULT         hRes     = S_OK;
    LPSRowSet       pRows    = NULL;
    ULONG           i;
    LPSPropTagArray pTags = NULL;
    LONG            lFound = -1;

    DECLARE_REC_NUM(RecNum);

    if (FAILED(hRes = pRemoteObj -> OpenProperty(PR_DETAILS_TABLE,
                                                 &IID_IMAPITable,
                                                 0,
                                                 0,
                                                (LPUNKNOWN *) &m_pDetails)))
        goto Quit;


    if (FAILED(hRes = m_pDetails -> QueryColumns(TBL_ALL_COLUMNS, & pTags)))
        goto Quit;
    
    if (FAILED(hRes = m_pDetails -> SetColumns(pTags,0)))
        goto Quit;

    for (i = 0; i < pTags -> cValues; i++)
        if (PROP_ID(PR_ROWID) == PROP_ID(pTags -> aulPropTag[i]))
        {
            lFound = i;   
            break;
        }
    
    if (FAILED(hRes = HrQueryAllRows(m_pDetails,NULL,NULL,NULL,0,&pRows)))
        goto Quit;

    if (lFound < 0)
    {               
        // we don't know if this table has an index column
        // so add our own
        for (i = 0; i < pRows -> cRows; i++)
        {
            LPSPropValue    pNewRow = NULL;
            ULONG           cSrcCols  = pRows -> aRow[i].cValues,
                            ulSrcSize = sizeof SPropValue * cSrcCols;
        
            // allocate new, bigger row
            if (FAILED(hRes = gAllocBuff(ulSrcSize + sizeof SPropValue,
                                         (LPVOID *) &pNewRow)))
                goto Quit;

            if (FAILED(hRes = CopySProps(pNewRow, pRows -> aRow[i].lpProps,cSrcCols)))
                goto Quit;

            gFreeBuff(pRows -> aRow[i].lpProps);

            // use row ID for index prop
            pNewRow[cSrcCols].ulPropTag = PR_ROWID;
            pNewRow[cSrcCols].Value.l   = i;
            pRows -> aRow[i].lpProps = pNewRow;
            pRows -> aRow[i].cValues++;
        }
        
    }
    if (SUCCEEDED(hRes = Open(m_pLogon -> m_pDB, 
                              OVERWRITE,
                              m_MyEID.ID,
                              RecNum)))
    {
        hRes = Update(RecNum, fldTABLEROWS,pRows);
        Close(RecNum);
    }
Quit:

    gFreeBuff(pTags);

    if (FAILED(hRes))
	{
        if (m_pDetails)
            m_pDetails -> Release();
	 	m_pDetails = NULL;
	}

    if (pRows)
        FreeProws(pRows);

    return hRes;    
    */
}

// End of file for IPROP.CPP
