///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      LIST.INL
//
//  Description
//      Template list class.
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#ifndef _LIST_INL_
#define _LIST_INL_

#include "template.h"

///////////////////////////////////////////////////////////////////////////////
//
//    CNode()
//
//    Parameters                        
//          val
//
//    Purpose
//
//
//    Return Value

template <class NodeType>
CNode<NodeType> :: CNode(const NodeType & val) : m_Item(val)
{
    m_pNext = NULL;
}
///////////////////////////////////////////////////////////////////////////////
//
//    CList()
//
//    Parameters                        
//          none.
//
//    Purpose
//          List node constructor.
//
//    Return Value
//          none.
//
template <class NodeType> 
CList<NodeType> :: CList()
{
    m_pHead  = m_pTail = NULL;
    m_ulCnt  = 0;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Destroy()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Walks the list deleting each node, then deletes the list.
//
//    Return Value
//          none.
//

template <class NodeType>
VOID  WINAPI
CList<NodeType> :: Destroy()
{
    CNode<NodeType> * temp = m_pHead;

    while (NULL != temp)
    {   
        m_pHead = m_pHead -> m_pNext;
        m_ulCnt--;
        delete temp;
        temp = m_pHead;    
    }    

    m_pHead = m_pTail = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    ~CList()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Class destructor.
//
//    Return Value
//          none.
//

template <class NodeType> 
CList<NodeType> :: ~CList()
{
    Destroy();
}

///////////////////////////////////////////////////////////////////////////////
//
//    InsertAsc()
//
//    Parameters                        
//          val     [in] reference to a parameterized type value to insert
//                       in the list.
//
//    Purpose
//          Consructs a list node containing val and inserts it in ascending
//          order in the list. NodeType must have a >= overload (this is what
//          determines the interpretation of 'ascending'
//
//    Return Value
//          TRUE on success, FALSE otherwise.
//
template <class NodeType> 
BOOL WINAPI
CList<NodeType> :: InsertAsc(const NodeType & val)
{
    CNode<NodeType> * temp1   = m_pHead,
                    * temp2,
                    * target  = new CNode<NodeType>(val);

    if (!target)
        return FALSE;

    m_ulCnt++;

    // empty list
    if (!m_pHead)
    {   
        m_pHead = m_pTail = target;
        return TRUE;
    }

    // insert before head
    if (temp1 -> m_Item >= val)
    {
        target -> m_pNext = temp1;
        m_pHead = target;
        return TRUE;
    }     

    temp2 = m_pHead -> m_pNext;
        
    while (temp2)
    {
        // hit in the middle
        if (temp2 -> m_Item >= val)
        {
            target -> m_pNext = temp2;
            temp1 -> m_pNext = target;
            return TRUE;
        }

        temp1 = temp2;
        temp2 = temp2 -> m_pNext;
    }

    // not in list, put at end    
    temp1 -> m_pNext = target;
    target -> m_pNext = NULL;
    m_pTail = target;
    
    return TRUE;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Insert()
//
//    Parameters                        
//          val     [in] reference to a parameterized type value to insert
//                       in the list.
//
//    Purpose
//          Consructs a list node containing val and inserts it in the list. 
//          val replaces any node whose data is 'equal' to val. NodeType must 
//          have an == overload (this is what determines the interpretation 
//          of 'equal'
//
//    Return Value
//          TRUE on success, FALSE otherwise.
//
template <class NodeType> 
BOOL WINAPI
CList<NodeType> :: Insert(const NodeType & val)
{
    CNode<NodeType> * temp1   = m_pHead,
                    * temp2,
                    * target  = new CNode<NodeType>(val);

    if (!target)
        return FALSE;

    // don't incr count if you replace
    if (!m_pHead)
    {   
        m_pHead = m_pTail = target;
        m_ulCnt++;
        return TRUE;
    }

    // hit on head of list
    if (temp1 -> m_Item == val)
    {
        target -> m_pNext = m_pHead -> m_pNext;
        m_pHead = target;
        delete temp1;
        return TRUE;
    }     

    temp2 = m_pHead -> m_pNext;
        
    while (temp2)
    {
        // hit in the middle
        if (temp2 -> m_Item == val)
        {
            target -> m_pNext = temp2 -> m_pNext;
            temp1 -> m_pNext = target;
            
            // hit on last node
            if (!target -> m_pNext)
                m_pTail = target;

            delete temp2;
            return TRUE;
        }

        temp1 = temp2;
        temp2 = temp2 -> m_pNext;
    }

    // not in list, put at end    
    temp1 -> m_pNext = target;
    target -> m_pNext = NULL;
    m_pTail = target;
    m_ulCnt++;
    
    return TRUE;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Queue()
//
//    Parameters                        
//          val     [in] reference to a parameterized type value to insert
//                       in the list.
//    Purpose
//          Consructs a list node containing val and queues it tp the list. 
//
//    Return Value
//          TRUE on success, FALSE otherwise.
//

template <class NodeType> 
BOOL WINAPI
CList<NodeType> :: Queue(const NodeType & val)
{
    CNode<NodeType> * node = new CNode<NodeType>(val);

    if (!node)
        return FALSE;

    m_ulCnt++;

    if (!m_pHead) 
        m_pHead = m_pTail = node;
    else
    {
        m_pTail -> m_pNext = node;
        m_pTail = node;
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Size()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Returns the number of nodes in the list.
//
//    Return Value
//          Size of the list.
//
template <class NodeType> 
ULONG WINAPI
CList<NodeType> :: Size()
{
    return m_ulCnt;
}

// return FALSE on allocation error, TRUE otherwise.
// *ret == NULL if not found, pts to node data otherwise

///////////////////////////////////////////////////////////////////////////////
//
//    Delete()
//
//    Parameters                        
//          SearchTarget [in]  reference to a value of type NodeType we
//                             seek in the list.
//          ret          [out] pointer to a pointer where we write the found
//                             node's value
//    Purpose
//          Walks the list looking for SearchTarget. If found, copies the
//          found node's data into a newly allocated buffer and returns the
//          buffer in *ret. If not found, *ret <- NULL. Note that Node type
//          must have an == overload that determines what a match is. e.g.
//          a complex struct might base the comparison on a subset of the
//          struct members, so SearchTarget == *ret even though not all
//          members are identical.
//
//    Return Value
//          TRUE on if the node is not in list after the call, FALSE on error.
//          Note: you must check *ret as well as error code, since its not an
//          error to delet a nonexistent node
//

template <class NodeType> 
BOOL WINAPI
CList<NodeType> :: Delete(const NodeType & SearchTarget,
                          NodeType      ** ret)
{
    CNode<NodeType> * temp1 = m_pHead,
                    * temp2;

    *ret = NULL;
        
    if (!temp1)
        return TRUE;
    
    // hit on head
    if (temp1 -> m_Item == SearchTarget)
    {
        *ret = new NodeType(temp1 -> m_Item);
        if (*ret)
        {
            m_pHead = m_pHead -> m_pNext;
            
            // single node list
            if (!m_pHead)
                m_pTail = NULL;

            delete temp1;            
            m_ulCnt--;
            return TRUE;
        }
        return FALSE;      
    }     

    temp2 = m_pHead -> m_pNext;

    while (temp2)
    {
        // hit in middle
        if (temp2 -> m_Item == SearchTarget)
        {
            *ret = new NodeType(temp2 -> m_Item);
            if (*ret)
            {
                temp1 -> m_pNext = temp2 -> m_pNext;
                m_ulCnt--;

                // just deleted last node
                if (!temp1 -> m_pNext)
                    m_pTail = temp1;

                delete temp2;
                return TRUE;
            }
            return FALSE;
        }

        temp1 = temp2;
        temp2 = temp2 -> m_pNext;
    }
    
    // not in list
    return TRUE;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Find()
//
//    Parameters                        
//          SearchTarget [in]  reference to a value of type NodeType we
//                             seek in the list.
//    Purpose
//          Walks the list looking for SearchTarget. If found, returns
//          a pointer to the node's data, otherwise returns NULL.Node type
//          must have an == overload that determines what a match is. e.g.
//          a complex struct might base the comparison on a subset of the
//          struct members, so SearchTarget == *ret even though not all
//          members are identical. This method 'peeks' at the node, caller
//          should not modify what the return value points at or delete the
//          returned pointer.
//   
//    Return Value
//          Pointer to the node's value field, if found, NULL if not.
//

template <class NodeType> 
NodeType * WINAPI
CList<NodeType> :: Find(const NodeType & SearchTarget)
{
    CNode<NodeType> * temp = m_pHead;
    NodeType        * ret = NULL;

    if (!temp)
        return NULL;

    while (temp)
    {
        if (temp -> m_Item == SearchTarget)    
        {
            ret = new NodeType(temp -> m_Item);
            return ret;
        }
        else
            temp = temp -> m_pNext;

    }
    return ret;    
}

///////////////////////////////////////////////////////////////////////////////
//
//    Dequeue()
//
//    Parameters                        
//          none.
//
//    Purpose
//         Returns a copy of the node's data at the head of the list
//         Destructively dequeues the head.
//
//    Return Value
//         Pointer to value at head of list, or NULL on error or empty list.
//

template <class NodeType>
NodeType * WINAPI
CList<NodeType> :: Dequeue()
{
    NodeType        * ret  = NULL;
    CNode<NodeType> * temp = m_pHead;

    if (!m_pHead)
        return NULL;

    ret = new NodeType(temp -> m_Item);

    if (ret)
    {
        m_ulCnt--;
        m_pHead = m_pHead -> m_pNext;

        if (!m_pHead)
            m_pTail = NULL;

        delete temp;
    }

    return ret;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Walk()
//
//    Parameters                        
//          bWalking  [in] set to FALSE on first iteration, then defaults
//                         to TRUE to continue the walk
//
//    Purpose
//          Called in a loop until returns NULL, First time pass 0, then 
//          while NULL != Walk() returns a ptr to a COPY of the current
//          node's data, (which caller must free)
//
//    Return Value
//          Pointer to the current node's data or NULL if at end of list.
//
template <class NodeType> 
NodeType * WINAPI
CList<NodeType> :: Walk(BOOL bWalking)
{

    NodeType        * ret = NULL;
    
    if (!bWalking)
        m_pTemp = m_pHead;

    if (m_pTemp)
    {
        ret = new NodeType(m_pTemp -> m_Item);
        m_pTemp = m_pTemp -> m_pNext;        
    }                    
    
    return ret;       
}

// called in a loop until returns NULL, First time pass 0, then while NULL != Walk()
// returns a ptr to existing current node's data, caller mustn't free
///////////////////////////////////////////////////////////////////////////////
//
//    Scan()
//
//    Parameters                        
//          bWalking  [in] set to FALSE on first iteration, then defaults
//                         to TRUE to continue the walk
//
//    Purpose
//          Called in a loop until returns NULL, First time pass 0, then 
//          while NULL != Walk() returns a ptr to the current
//          node's data. Caller must not free or modify the memory at the
//          returned pointer, it's still in the list.
//
//    Return Value
//          Pointer to the current node's data or NULL if at end of list.
//
template <class NodeType> 
NodeType * WINAPI
CList<NodeType> :: Scan(BOOL bWalking)
{

    NodeType        * ret = NULL;
    
    if (!bWalking)
        m_pTemp = m_pHead;

    if (m_pTemp)
    {
        ret = &(m_pTemp -> m_Item);
        m_pTemp = m_pTemp -> m_pNext;        
    }                    
    
    return ret;       
}

///////////////////////////////////////////////////////////////////////////////
//
//    operator +()
//
//    Parameters                        
//          Suffix  [in] reference to a list of NodeType to append to
//                       this list
//
//    Purpose
//          Destructively appends Suffix to 'this'. If an error occurs,
//          'this' may be messed up, return NULL
//
//    Return Value
//          The modified list with Suffix appended
//      
template <class NodeType> 
CList<NodeType> &
CList<NodeType> :: operator +(CList<NodeType> & Suffix)
{
    
    for (NodeType * pNode = Suffix.Scan(0);
         pNode;
         pNode = Suffix.Scan())   

        if (!(this -> Insert(*pNode)))
            return NULL;

    return *this;
}

///////////////////////////////////////////////////////////////////////////////
//
//    operator -=()
//
//    Parameters                        
//          Deletions   [in] reference to a list of values to delete
//                           from 'this'
//
//    Purpose
//          Destructively deletes any members of Deletions from 'this'
//          if they are found in 'this', The deleted values are freed.
//
//    Return Value
//          The modified list.
//

template <class NodeType> 
CList<NodeType> &
CList<NodeType> :: operator -=(CList<NodeType> & Deletions)
{

    NodeType * pTemp;
                                                         
    for (NodeType * pNode = Deletions.Scan(0);
         pNode;
         pNode = Deletions.Scan())   
    {
        Delete(*pNode, &pTemp);

        if (pTemp)
            delete pTemp;
    }

    return *this;
}

#endif

// End of file for LIST.INL
