///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      SERVICE.CPP
//
//  Description
//      Configuration/profile related code.
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#include <WINDOWS.H>

#include <mapispi.h>
#include <mapiutil.h>
#include <mapiwin.h>
#include <mapiwz.h>

#include "abpab.h"
#include "traces.h"
#include "resource.h"
#include "cdbfile.h"

#define USES_IID_IMAPIProp
#define USES_IID_IMailUser
#define USES_IID_IMAPIPropData
#define USES_IID_IMAPIContainer
#define USES_IID_IMAPITable
#define USES_IID_IMAPITableData
#define USES_IID_IMAPIAdviseSink
#define USES_IID_IABContainer
#define USES_IID_IABLogon
#define USES_IID_IDistList
#define USES_IID_IMAPIControl
#define USES_PS_MAPI
#define USES_PS_PUBLIC_STRINGS

#define INITGUID
#include <initguid.h>
#include <mapiguid.h>

// Prototypes of the function in this module
extern "C"
{
    WIZARDENTRY          WizardEntry;
    SERVICEWIZARDDLGPROC WizardDlgProc;
};

void WINAPI TogglePage (HWND, BOOL);

// Global data used in the WizardDlgProc function
LPMAPIPROP  pLocalPropObj = NULL;
HINSTANCE   hInstance     = NULL;
SPropValue  spProp;



///////////////////////////////////////////////////////////////////////////////
//    
//    ~ConfigObj
//
//    Parameters
//          none.
//
//    Purpose
//          class destructor.
//    Return Value
//          none.
//
ConfigObj :: ~ConfigObj()
{
    if (m_pProps)
        gFreeBuff(m_pProps);
      
    // Release objects
    if (m_pProfSecn)
        m_pProfSecn -> Release();

    if (m_pDispTbl)
        m_pDispTbl -> Release();

    if (m_pDispTblData)
        m_pDispTblData -> Release();

    if (m_pPropData)
        m_pPropData -> Release();

}

///////////////////////////////////////////////////////////////////////////////
//    
//    Init
//
//    Parameters
//          none.
//
//    Purpose
//          Intializes the ConfigObj
//    Return Value
//          S_OK ofn success, a MAPI error code otherwise.
//
STDMETHODIMP ConfigObj :: Init(LPSPropTagArray pLogonTags)
{
    LPPROFSECT  pProfSecn = NULL;
    HRESULT     hRes = S_OK;
    ULONG       ulPropCnt;
    LPMAPITABLE pProvTbl = NULL;
    LPSPropValue pProps = NULL;
    LPPROPDATA 	    pPropData = NULL;
    SPropTagArray   sptProvider = { 1, { PR_PROVIDER_UID } };
    LPSRowSet       pRows = NULL;

    // Get the provider table where the settings are stored
    if (FAILED(hRes = m_pAdminProv -> GetProviderTable(0, &pProvTbl)))
        return hRes;

    if (FAILED(hRes = HrQueryAllRows(pProvTbl,
                                     &sptProvider,
                                     NULL,
                                     NULL,
                                     1,
                                     &pRows)))
        goto Quit;

    
    if (!pRows -> cRows || 
        PR_PROVIDER_UID != pRows -> aRow -> lpProps[0].ulPropTag)
    {
        hRes = MAPI_E_CALL_FAILED;
        goto Quit;
    }

    // Can't open profile section, bail
    if (FAILED(hRes = m_pAdminProv -> OpenProfileSection((LPMAPIUID)pRows->aRow->lpProps->Value.bin.lpb,
                                                         NULL,
                                                         MAPI_MODIFY,
                                                         &pProfSecn)))
        goto Quit;

    // get the current config settings
    if (FAILED(hRes = pProfSecn -> GetProps(pLogonTags,
                                            0,
                                            &ulPropCnt,
                                            &pProps)))
        goto Quit;

    hRes = CreateIProp((LPIID)&IID_IMAPIPropData,
                         gAllocBuff,
                         gAllocMore,
                         gFreeBuff,
                         NULL,
                         &pPropData);
Quit:
    FreeProws(pRows);

    if (pProvTbl)
        pProvTbl -> Release();

    if (FAILED(hRes))
    {
        if (pProps)
            gFreeBuff((LPVOID) m_pProps);

        if (pPropData)
            pPropData -> Release();

        if (pProfSecn)
            pProfSecn -> Release();
        
        return hRes;
    }

    m_pProfSecn = pProfSecn;
    m_pPropData = pPropData;
    m_pProps = pProps;
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    ConfigObj()
//
//    Parameters                        
//          hInst         [in]  module instance handle
//          hWnd,         [in]  main application window handle
//          pAdminProv    [in]  pointer to provider admin object
//          pMapiSup      [in]  pointer to support object
//    Purpose
//          Class constructor.  This class packages a bunch of data and a 
//          a couple methods used in configuration of the provider.
//      
//    Return Value
//          none.
//


ConfigObj :: ConfigObj(HINSTANCE       hInst,
                       HWND            hWnd,
                       LPPROVIDERADMIN pAdminProv,
                       LPMAPISUP       pMapiSup,
                       LPDTPAGE        pDetailsPage)
{
    m_pAdminProv   = pAdminProv;       // passed to us by MAPI, don't AddRef or Release
    m_pSupObj      = pMapiSup;
    m_hInst        = hInst;
    m_hWnd         = hWnd;
    m_pDTPage      = pDetailsPage;
    m_pProps       = NULL;
    m_pPropData    = NULL;
    m_pDispTbl     = NULL;
    m_pDispTblData = NULL;
    m_pProfSecn    = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    DoLogonDlg()
//
//    Parameters                        
//          none.
//
//    Purpose
//          Brings up a dialog for provider configuration. Creates a 
//          a property interface and a display table for our page,
//          then brings up the dialog. The properties are set in
//          the dialog, then saved to an SPropValue array which is
//          later written to the profile section.
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//

STDMETHODIMP ConfigObj :: DoLogonDlg()
{ 
    LPSPropValue    pTemp = NULL;
    HRESULT         hRes  = S_OK;
    ULONG           ulPropCnt;
    int             i;
    // validate an existing DB filename, or get new one from user
    if (!GetValidFileName(&m_pProps[LOGON_DB_PATH].Value.LPSZ)) 
        return MAPI_E_USER_CANCEL;

    // set any existing values    
    if (FAILED(hRes = SetProps(LOGON_NUM_PROPS, m_pProps, NULL)))
        goto Quit;

    // enable the controls (make them writable). 
    // DT_CTL_XX are enums, order is significant
    for (i = DT_CTL_NAME; i < DT_CTL_BROWSE; i++)
        m_pDTPage -> lpctl[i].ulCtlFlags |= DT_EDITABLE;
    
    // Create the display table for the logon dialog. It's based on dialog
    // resources plus display table struct
    if (FAILED(hRes = BuildDisplayTable(gAllocBuff,
                                        gAllocMore,
                                        gFreeBuff,
                                        NULL,
                               		    m_hInst,
                                		1,
                                		m_pDTPage,
                                		0,
                                 		&m_pDispTbl,       
                                   		&m_pDispTblData)))
		goto Quit;

    // Display the dialog/property sheets
    if (FAILED(hRes = m_pSupObj -> DoConfigPropsheet((ULONG) m_hWnd,
                                                     0,
                                                     SZ_PSS_SAMPLE_AB,
                                                     m_pDispTbl,
                                                     (LPMAPIPROP) this,
                                                     0)))
        goto Quit;

    // get any changes
    hRes = GetProps((LPSPropTagArray)&sptLogonTags,
                    0,
                    &ulPropCnt,
                    &pTemp);

Quit:
    gFreeBuff((LPVOID)m_pProps);
    m_pProps = pTemp;
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    OpenProperty()
//
//    Parameters                        
//          see MAPI documentation.
//    Purpose
//          The whole reason ConfigObj is derived from IMAPIProp is
//          to provide this method, so we can get one stinking Browse
//          button to work. This method handles the OpenProperty with
//          PR_BROWSE_BUTTON for the display table
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise.
//
STDMETHODIMP    ConfigObj :: OpenProperty(ULONG       ulPropTag,  
                                          LPCIID      lpiid,      
                                          ULONG       ulInterfaceOptions, 
                                          ULONG       ulFlags,    
                                          LPUNKNOWN * lppUnk)
{                                         
    HRESULT     hRes = S_OK;              
    Validate_IMAPIProp_OpenProperty(this,
                                    ulPropTag,         
                                    lpiid,             
                                    ulInterfaceOptions,
                                    ulFlags,           
                                    lppUnk);            
    *lppUnk = NULL;
    switch (ulPropTag)
    {
    
        case PR_BROWSE_BUTTON:
        {
            CBrowseButton * pButton = new CBrowseButton(this);

            if (!pButton)
                return MAPI_E_NOT_ENOUGH_MEMORY;
            
            *lppUnk = (LPUNKNOWN) pButton;
            pButton -> AddRef(); 
            break;
        }
        default: 
            return MAPI_E_NO_SUPPORT;
    }

    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    GetValidFileName()
//
//    Parameters                        
//          pszFile     [in/out]  filename to validate (in), valid filename(out)
//          hWnd,       [in]      application main window
//          pParent     [in]      block of memory to link any additional property
//                                allocations to (a la MAPIAllocateMore)  
//    Purpose
//          Checks if the file given in *pszFile is valid. If not, it's released
//          and a dialog is brought up where the user selects a new filename.
//          The new filename buffer is is allocated and the old filename copied
//          into the buffer. The Browse dialog overwrites whatever was in the
//          buffer with the user's selection.
//      
//    Return Value
//          S_OK on success, a MAPI error code or an exception (fatal error).

STDMETHODIMP GetValidFileName(LPTSTR   * pszFile,
                              HWND       hWnd,
                              LPVOID     pParent,
                              ULONG      ulFlags)      
{
    HANDLE      hFile;
    HRESULT     hRes;
       
    hFile = CreateFile(*pszFile,
                       GENERIC_READ,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_RANDOM_ACCESS,
                       NULL);

    if (INVALID_HANDLE_VALUE == hFile)
    {
        if (ulFlags & AB_NO_DIALOG)
            return MAPI_E_DISK_ERROR;

        if (FAILED(hRes = BrowseFile(hWnd,pszFile,pParent)))
            return hRes;         
    }

    CloseHandle(hFile);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    GetValidFileName()
//
//    Parameters                        
//          pszFile     [in/out]  pointer to filename string to validate (in),
//                                valid filename(out)     
//
//    Purpose
//          Returns right away if the filename passed in is valid, if not
//          brings up an OpenFilename dialog for the user to select a
//          valid filename and returns the name in *pszFile.
//      
//    Return Value
//          TRUE on success, FALSE if the user cancels the OpenFile dialog
//
BOOL WINAPI ConfigObj :: GetValidFileName(LPTSTR * pszFile)
{            
    TraceMessage("ConfigObj::GetValidFileName");

    return S_OK == ::GetValidFileName(pszFile, m_hWnd, (LPVOID)m_pProps,0);

}

///////////////////////////////////////////////////////////////////////////////
//
//    UseCannedInProp()
//
//    Parameters                        
//          pProp       [in] pointer to property to set
//          LogonTag    [in] tag of property to set
//    Purpose
//          Utility function for initializing the configuration properties
//          to default values.
//      
//    Return Value
//          none.
//

STDMETHODIMP ConfigObj :: UseCannedInProp(LPSPropValue pProp,ULONG LogonTag)
{
    pProp -> ulPropTag = LogonTag;

    switch (LogonTag)
    { 
        // We use gAllocMore because m_props comes from GetProps, all the
        // Value.lpszA members are linked to m_props. When we replace
        // them, we don't have to copy and free m_props, just overwrite
        // the members then free the whole hairball later. This way we
        // can leave dangling pointers and know that they get freed.
        case PR_SMP_AB_PATH:
        case PR_COMMENT:
            // the proptag is PT_ERROR, lpszA is garbage
            // so allocate a null string we can free later
            if (FAILED(gAllocMore(Cbtszsize(NULL_STRING),
                                  (LPVOID)m_pProps,  
                                  (LPVOID *) &pProp -> Value.lpszA)))
                return MAPI_E_NOT_ENOUGH_MEMORY;

            lstrcpy(pProp -> Value.lpszA, NULL_STRING);                    
            break;

         case PR_DISPLAY_NAME:
             if (FAILED(gAllocMore(Cbtszsize(SZ_PSS_SAMPLE_AB),
                                   (LPVOID) m_pProps,
                                  (LPVOID *) &pProp -> Value.lpszA)))
                return MAPI_E_NOT_ENOUGH_MEMORY;

            lstrcpy(pProp -> Value.lpszA,SZ_PSS_SAMPLE_AB);
            break;

        default :
            ASSERT(0);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    DefaultLogonProps()
//
//    Parameters                        
//          pLogonTags          [in] array of tags of requried configuration
//                                   properties
//          pWizardProps        [in] array of properties passed to us in Wizard
//          ulWizardPropCount   [in] count of Wizard properties
//    Purpose
//          Merges the properties passed to us by the Wizard with our canned-in
//          defaults.
//      
//    Return Value
//          none.
//
STDMETHODIMP ConfigObj :: DefaultLogonProps(LPSPropTagArray pLogonTags,
                                            LPSPropValue    pWizardProps,
                                            ULONG           ulWizardPropCount)
{
    LPTSTR      szTemp = NULL;
    BOOL        bFound;
    HRESULT     hRes = S_OK;

    for (int i = LOGON_DB_PATH; i < LOGON_NUM_PROPS; i++)
    {
        bFound = FALSE;

        // if property was passed by Wizard, save it
        for (ULONG j = 0; j < ulWizardPropCount; j++)
            if (PROP_ID(pWizardProps[j].ulPropTag) == 
                PROP_ID(pLogonTags -> aulPropTag[i]))
            {
                bFound = TRUE;
                m_pProps[i] = pWizardProps[j];
                break;       
            }    
        
        // if not passed by Wizard and not in profile, use canned value 
        if (!bFound)
            if (m_pProps[i].ulPropTag != pLogonTags -> aulPropTag[i])        
               if (FAILED(hRes = UseCannedInProp(&m_pProps[i],
                                                 pLogonTags -> aulPropTag[i])))
                    goto Quit;
    }
Quit:
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//    DefaultLogonProps()
//
//    Parameters                        
//          pLogonTags    [in] array of tags of required logon properties     
//
//    Purpose
//          Sets default logon properties.
//      
//    Return Value
//          none.
//

STDMETHODIMP ConfigObj:: DefaultLogonProps(LPSPropTagArray pLogonTags)
{
    LPTSTR  szTemp = NULL;
    HRESULT hRes;

    // if property not in profile section, use canned-in default
    for (int i = LOGON_DB_PATH; i < LOGON_NUM_PROPS; i++)
        if (m_pProps[i].ulPropTag != pLogonTags -> aulPropTag[i])
            if (FAILED(hRes = UseCannedInProp(&m_pProps[i], pLogonTags -> aulPropTag[i])))
                return hRes;

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    UpdateProfile()
//
//    Parameters
//      none.
//
//    Purpose
//          Writes changes in our configuration properties to profile section.
//
//    Return Value
//          S_OK  on success, MAPI error code otherwise
//

STDMETHODIMP  ConfigObj :: UpdateProfile()
{
    HRESULT hRes;

    if (FAILED(hRes = m_pProfSecn -> SetProps(LOGON_NUM_PROPS,
                                              m_pProps,
                                              NULL)))

        return hRes;

    return m_pProfSecn -> SaveChanges(0);
}                

///////////////////////////////////////////////////////////////////////////////
//
//    ServiceEntry()
//
//    Parameters
//          Refer to MAPI Documentation on this method.
//
//    Purpose
//          Called by the profile setup adminitration tool to display the provider
//          configuration properties for this transport provider. Provides 
//          exception handler for configuration exceptions.
//
//    Return Value
//          MAPI_E_NO_ACCESS    If the function is called with unknown flags, or
//                              if the funciton is called in a context other than
//                              MSG_SERVICE_CONFIGURE
//          hResult             Handle with the error code of the method or 
//                              function that failed
//          S_OK                Otherwise
//
HRESULT STDAPICALLTYPE ServiceEntry(HINSTANCE          hInst,
                                    LPMALLOC           pMallocObj,
                                    LPMAPISUP          pMAPISup,
                                    ULONG              ulUIParam,
                                    ULONG              ulFlags,
                                    ULONG              ulContext,
                                    ULONG              ulCfgPropCount,
                                    LPSPropValue       pCfgProps,
                                    LPPROVIDERADMIN    pAdminProv,
                                    LPMAPIERROR *      ppMAPIError)
{
    HRESULT         hRes     = S_OK;
    LPConfigObj     pCfgObj  = NULL;
    MAPIUID         FileUID  = {0};
    CDataBase     * pDB      = NULL;
    HINSTANCE       hInstLib = NULL;

    BOOL            bNoUI;  
    BOOL            bNoCfgProps;

    // for our debug traces
    InitTraces(TRACES_CONSOLE | TRACES_NO_COM_OUTPUT);

    bNoUI       = !(ulFlags & (SERVICE_UI_ALWAYS | SERVICE_UI_ALLOWED));
    bNoCfgProps = IsBadReadPtr(pCfgProps, 
                               ulCfgPropCount * sizeof SPropValue) || NULL == pCfgProps; 

    // if we're in NT, this does a dynamic load
    Register3DCtls(hInst, &hInstLib);

    // save the allocation routines
    if (FAILED(hRes = pMAPISup -> GetMemAllocRoutines(&gAllocBuff,
                                                      &gAllocMore,
                                                      &gFreeBuff)))
        goto Quit;

    switch (ulContext)
    {
        case MSG_SERVICE_DELETE :
        case MSG_SERVICE_INSTALL :
        case MSG_SERVICE_UNINSTALL :
            break;
        
        case MSG_SERVICE_CREATE :
            // Wizard calling to create, no work to do
            if (bNoUI)
                break;  
            // fall through

        case MSG_SERVICE_CONFIGURE :
            pCfgObj = new ConfigObj(hInst,
                                   (HWND) ulUIParam,
                                    pAdminProv,
                                    pMAPISup,
                                    &GConfigDetailsPage);

            if (!pCfgObj)
            {
                hRes = MAPI_E_NOT_ENOUGH_MEMORY;
                goto Quit;
            }

            if (FAILED(hRes = pCfgObj -> Init((LPSPropTagArray) &sptLogonTags)))
                goto Quit;

            if (bNoCfgProps)
                hRes = pCfgObj -> DefaultLogonProps((LPSPropTagArray) &sptLogonTags);
            else
                hRes = pCfgObj -> DefaultLogonProps((LPSPropTagArray) &sptLogonTags, 
                                                    pCfgProps,
                                                    ulCfgPropCount);                
            if (FAILED(hRes))
                goto Quit;

            if (!bNoUI)
                hRes = pCfgObj -> DoLogonDlg();
           
            if (MAPI_E_USER_CANCEL == hRes)
                break;
   
            // see if the database exists, create it if not 
            if (FAILED(CommonOpenFile(pCfgObj -> m_pProps[LOGON_DB_PATH].Value.LPSZ, 
                                      NULL,
                                      NULL,
                                      FileUID,
                                      &pDB,
                                      FALSE)))
            {  
                
                if (FAILED(hRes = pMAPISup -> NewUID(&FileUID)))
                    goto Quit;

                if (FAILED(hRes = CreateDataBaseFile(pCfgObj -> m_pProps[LOGON_DB_PATH].Value.LPSZ, 
                                                     pCfgObj -> m_pProps[LOGON_DISP_NAME].Value.LPSZ,
                                                     pCfgObj -> m_pProps[LOGON_COMMENT].Value.LPSZ, 
                                                     FileUID,
                                                     &pDB)))
                    goto Quit;
            }

            hRes = pCfgObj -> UpdateProfile();
            break;

        default:
            hRes = MAPI_E_NO_SUPPORT;
            break;
    }
   
Quit:
    if (pCfgObj)    
        delete pCfgObj;

    if (pDB)
        pDB -> Release();
     
    UnRegister3DCtls(hInst,hInstLib);
    UnInitTraces();
        
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//                                             
//    WizardEntry()
//
//    Parameters
//          Refer to MAPI Documentation on this method.
//
//    Purpose
//          Returns the name of the resource where the controls for the wizard
//          pages are defined and outlined. The Wizard gives us a pointer to an
//          IMAPIProp object. We will release this object after we are finished
//          in the dialog procedure of our configuration pages.
//
//    Return Value
//      S_OK    Always
//
ULONG STDAPICALLTYPE WizardEntry(HINSTANCE       hProviderDLLInstance,
                                 LPSTR         * ppszResourceName,
                                 DLGPROC       * ppDlgProc,
                                 LPMAPIPROP      pPropObj,
                                 LPVOID          pMAPISupObj)
{   
    TraceMessage("WizardEntry function called");
        
    // Quick sanity check on the input parameters
    ASSERT (ppszResourceName);
    ASSERT (ppDlgProc);
    ASSERT (pPropObj);
    // The resource name needs to be an actual text string not
    // something generated with the MAKEINTRESOURCE() macro
    *ppszResourceName = "WIZARD_PAGES";
    *ppDlgProc = (FARPROC)WizardDlgProc;
    pLocalPropObj = pPropObj;
    hInstance = hProviderDLLInstance;
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//    WizardDlgProc()
//
//    Parameters
//          Refer to MAPI and Win32 SDK Documentation on this method.
//
//    Purpose
//          This is the function that the profile wizard will call while
//          interacting with our configuration pages.
//
//    Return Value
//          TRUE if we handle a message, FALSE if we don't or there was an error
//          in the data
//
BOOL STDAPICALLTYPE WizardDlgProc(HWND           hDlg,
                                  UINT           message,
                                  WPARAM         wParam,
                                  LPARAM         lParam)
{
    static BOOL       fInitalizeControl = TRUE;
    static int        pageNum           = 1; 
    static HINSTANCE  hInstLib;
    static TCHAR      szABPath[MAX_PATH];

    switch (message)
    {
        case WM_INITDIALOG:
            TogglePage (hDlg, TRUE);

            // if the control have not been initialized do so
    		if (fInitalizeControl)
    		{
                // Set the flag to avoid re-initialization
    			fInitalizeControl = FALSE;
                // Limit the maximum number of character users can put in the edit
    			SendMessage(GetDlgItem (hDlg, IDC_ABPATH), EM_LIMITTEXT, MAX_PATH,0);
                Register3DCtls(hInstance, &hInstLib);
    		}
    		return TRUE;

        case WIZ_QUERYNUMPAGES :
            // Return the number of pages the our provider needs to display
            return 1;

        case WM_CLOSE :
            // We are closing down
            UnRegister3DCtls(hInstance, hInstLib);
		    return TRUE;

        case WM_COMMAND :
		    switch (LOWORD(wParam))
		    {
                case IDC_BROWSE :
                {
                    LPSTR szTemp = szABPath;

                    if (S_OK == BrowseFile(hDlg, &szTemp, NULL))                   
                        SetWindowText(GetDlgItem(hDlg, IDC_ABPATH), szABPath);
                   
                    return TRUE;
                }
		        case WIZ_NEXT:
                
		            if (1 == pageNum)
                    {
                        pageNum++;
                        SetFocus(GetDlgItem(hDlg, IDC_ABPATH));		                
                        return TRUE;
                    }

                    if (!GetWindowText(GetDlgItem(hDlg,IDC_ABPATH), 
                                       szABPath, 
                                       MAX_PATH))
            		    return FALSE;

            		spProp.ulPropTag = PR_SMP_AB_PATH;
                    spProp.Value.LPSZ = szABPath;

                    // Since this is the last page in our configuration,
                    // we must save the obtained data into the
                    // IMAPIProp object that the wizard gave us. In
                    // turn this data will be saved in the profile
                    // section of this provider
                    ASSERT (pLocalPropObj);

                    if (pLocalPropObj)
                    {
				        HRESULT hResult = pLocalPropObj -> SetProps (1, 
				                                                     &spProp,
				                                                     NULL);
                        if (FAILED(hResult))
                        {
                            // Something went wrong
                            TraceResult ("WizardDlgPrc: Failed to set the prop", hResult);
                            return FALSE;
                        }
                    }
                
                    return TRUE;

        		case WIZ_PREV:
                    // Disable the page we were in and enable the next one
                    // before it is display by the wizard

                    if (2 == pageNum)
                    {
        			    pageNum--;
                        SetWindowText(GetDlgItem (hDlg, IDC_ABPATH), szABPath);
                        SetFocus(GetDlgItem(hDlg, IDC_ABPATH));
                    }
        			return TRUE;
                
                default:
			        return FALSE;
		    }
		    break;

        default:
            return FALSE;
    }
   // return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//
//    TogglePage()
//
//    Parameters
//          hDlg     [in] Handle to the dialogs where the controls are located
//          fState   [in] State of the controls: TRUE shows them, FALSE hides them
//
//    Purpose
//          Enables/Disables controls.
//    Return Value
//          None
//
void WINAPI TogglePage (HWND hDlg, BOOL fState)
{
    // Toggle the controls ON or OFF
    EnableWindow (GetDlgItem (hDlg, IDC_ABPATH), FALSE);
    ShowWindow   (GetDlgItem (hDlg, IDC_ABPATH), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_BROWSE), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_BROWSE), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_STATIC1), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_STATIC1), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_INFO), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_INFO), fState ? SW_SHOW : SW_HIDE);
}



///////////////////////////////////////////////////////////////////////////////
//
//    BrowseFile()
//
//    Parameters
//          hWnd    [in]     handle of main application window
//          pszFile [in/out] in - filename displayed in GetOpenFileName
//                                dialog
//                           out- newly allocated buffer containing 
//                                user selected new filename
//          pParen  [in]     memory the allocation is linked to
//    Purpose
//          Wrapper for GetOpenFileName(). Pass in an existing filename
//          string in *pszFile and non-null in pParent, and on output
//          *pszFile has the user's selection in a newly allocated buffer
//          linked to pParent. If pParent is NULL, the selection is copied
//          into *pszFile, which must be allocated by the caller and of
//          sufficient size.
//
//    Return Value
//          TRUE on success, FALSE on CANCEL or error
//

STDMETHODIMP BrowseFile(HWND hWnd,LPTSTR *pszFile, LPVOID pParent)
{
    HRESULT       hRes   = S_OK;
    LPTSTR        szTemp = NULL;
    OPENFILENAME  ofn    = {0,0,0,0,0,0,0,0,0,0,
                            0,0,0,0,0,0,0,0,0,0};
    
    if (pParent)
    {
        if (FAILED(hRes = gAllocMore(MAX_PATH * sizeof(TCHAR),
                                     pParent,
                                     (LPVOID *) &szTemp)))
            return hRes;

        lstrcpy(szTemp,*pszFile);
    } 
    else
        szTemp = *pszFile;

    ofn.lStructSize = sizeof(ofn);
    ofn.hwndOwner   = hWnd;
    ofn.lpstrFilter = SZ_OFN_FILTER;
    ofn.lpstrFile   = szTemp;
    ofn.nMaxFile    = MAX_PATH;
    ofn.lpstrTitle  = SZ_PSS_SAMPLE_AB;
    ofn.Flags       = OFN_HIDEREADONLY | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST;
    ofn.lpstrDefExt = SZ_AB_DEF_EXT;
        
    if (GetOpenFileName(&ofn))
        // don't need to free old *pszFile, its linked to
        // pParent when passed to us, so is szTemp
        *pszFile = szTemp;

    else
        // if we're here, user must have cancelled browse,
        // don't free szTemp, it's still linked to parent
        hRes = MAPI_E_USER_CANCEL;

    return hRes;
}

// End of file for SERVICE.CPP
