///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      UTIL.CPP      
//
//  Description
//      Some handy utility functions.
//
//  Author
//      Les Thaler
//
//  Revision: 1.00
//
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#include "abpab.h"
#include "list.inl"

///////////////////////////////////////////////////////////////////////////////
//
//    CopySRow()
//
//    Parameters                        
//         psrDest  [in/out]   pointer to an SRow to copy to
//         psrSrc   [in]       pointer to an SRow to copy from
//    Purpose
//          Copy an SRow. The caller should allocate and free the
//          destination SRowfree. The lpProps member of the 
//          destination row should be freed with a single call to 
//          MAPIFreeBuffer when done.         
//      
//    Return Value
//          S_OK on success, a MAPI error code otherwise
// 
STDMETHODIMP    CopySRow(LPSRow psrDest, LPSRow psrSrc)
{
    HRESULT     hRes = S_OK;
    LPSRow      psrTemp = NULL;

    if (FAILED(hRes = gAllocBuff((sizeof SPropValue) * (psrSrc -> cValues), 
                                 (LPVOID *) &psrDest ->lpProps)))
        goto Quit;

    psrDest -> cValues = psrSrc -> cValues;
    
    hRes = CopySProps(psrDest -> lpProps, 
                      psrSrc  -> lpProps, 
                      psrSrc   -> cValues);
Quit:
    if (FAILED(hRes))
    {
        gFreeBuff((LPVOID)psrDest -> lpProps);
        psrDest -> lpProps = NULL;
    }
                
    return hRes;

}

///////////////////////////////////////////////////////////////////////////////
//
//    CopySProps()
//
//    Parameters                        
//          pspvDest  [in/out]  destination SPropValue array 
//          pspvSrc   [in]      source SPropValue array
//          cSrcVals  [in]      count of source properties
//    Purpose
//          Copies an SRow from psrSrc to psrDest          
//          The caller should allocate the dest SRow       
//          with MAPIAllocateBuffer. The dest can be       
//          larger than the src. This fcn chases down      
//          the nested pointers, allocates memory for them 
//          with MAPIAllocateMore, linked to psrDest       
//
//    Return Value
//          S_OK on success, a MAPI error code otherwise
// 
STDMETHODIMP CopySProps(LPSPropValue pspvDest, 
                        LPSPropValue pspvSrc,
                        ULONG        cSrcVals)
{
    HRESULT     hRes = S_OK;
    ULONG       i,
                ulTag;

    for (i = 0; i < cSrcVals; i++)
    {
        ulTag = pspvDest[i].ulPropTag;

        if (FAILED(hRes = CopyProp(&pspvDest[i], &pspvSrc[i], pspvDest)))        
            pspvDest[i].ulPropTag = CHANGE_PROP_TYPE(ulTag,PT_ERROR);
    }
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    CopyProp()
//
//    Parameters                        
//          pspDest   [in/out]  pointer to SPropValue to copy to
//          pspSrc    [in]      pointer to SPropValue to copy from
//          pParent   [in]      parent block to link allocations to
//    Purpose
//          copy an SPropValue from pspSrc to spsDest, including any 
//          sub structures. The resulting SPropValue sub structures 
//          are linked to pParent so the whole mess can be freed by 
//          freeing pParent                              
//    Return Value
//          S_OK on success, a MAPI error code otherwise
// 
STDMETHODIMP CopyProp(LPSPropValue pspDest, LPSPropValue pspSrc, LPSPropValue pParent) 
{

    HRESULT         hRes = S_OK;
    ULONG           ulTag = pspSrc -> ulPropTag,
                    ulLen;
    LPBYTE          pbSrc = NULL,
                  * ppbDest = NULL;

    // in case we fail, FreeBuff won't cack on garbage        
    ZeroMemory(pspDest, sizeof SPropValue);
    pspDest -> ulPropTag = pspSrc -> ulPropTag;


    switch(PROP_TYPE(ulTag))
    {
        case PT_BINARY:
            pbSrc  = pspSrc -> Value.bin.lpb;
            ulLen  = pspSrc -> Value.bin.cb;
            ppbDest = &(pspDest -> Value.bin.lpb);
            pspDest -> Value.bin.cb = ulLen;
            break;

        case PT_TSTRING:
            ulLen = lstrlen(pspSrc -> Value.LPSZ) + 1;
            pbSrc  = (LPBYTE) pspSrc -> Value.LPSZ;
            ppbDest = (LPBYTE *) &(pspDest -> Value.LPSZ);
            break;

        case PT_CLSID:
            ulLen = sizeof GUID;
            pbSrc  = (LPBYTE) pspSrc -> Value.lpguid;
            ppbDest = (LPBYTE *) &(pspDest -> Value.lpguid);
            break;

        default:
            if (ISMV_TAG(ulTag))
                return MAPI_E_NO_SUPPORT;

            ulLen = 0;
            pbSrc = NULL;
            ppbDest = NULL;
            pspDest -> Value = pspSrc -> Value;
            break;
    }

    if (ulLen && pbSrc && ppbDest)
        if (SUCCEEDED(hRes = gAllocMore(ulLen,(LPVOID) pParent,(LPVOID *) ppbDest)))
            memcpy((LPVOID)*ppbDest, (LPVOID)pbSrc, ulLen);

    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//    HrPropNotFound()
//
//    Parameters                        
//          pspv      [in/out]  pointer to SPropValue to mark 'not found'
//          ulTag     [in]      property tag of missing property
//    Purpose
//          Marks a property value as NOT FOUND by changing the tag to
//          (ulTag | PT_ERROR) and the Value.err to MAPI_E_NOT_FOUND
//
//    Return Value
//          MAPI_W_ERRORS_RETURNED
// 

STDMETHODIMP HrPropNotFound(SPropValue & spv, ULONG ulTag)
{

    spv.ulPropTag = CHANGE_PROP_TYPE(ulTag,PT_ERROR);
    spv.Value.err = MAPI_E_NOT_FOUND;
    return MAPI_W_ERRORS_RETURNED;
}

//#define _DEBUG

void * operator new(size_t size)
{
#ifdef _DEBUG
    LPVOID pMem;
    gAllocBuff((ULONG)size,&pMem);
    //TraceString2("\tnew block @ %08x, size %08x",(ULONG) pMem,size);
    return pMem;
#else
    return (LPVOID)HeapAlloc(GetProcessHeap(),0,size);
#endif
}

void operator delete(void * pObj)
{
#ifdef _DEBUG
   // TraceString1("\tdelete block @ %08x",(ULONG)pObj);
    gFreeBuff(pObj);
#else    
    HeapFree(GetProcessHeap(),0,pObj);
#endif

}
// End of file for UTIL.CPP
