///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      ABLOGON.CPP
//
//  Description
//      This file implements the IABLogon interface with the methods specified
//      in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "ABWDS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceInfoMessage(a)
#define InfoTraceRaw(a)     TraceRaw(a)
#define InfoTraceProp(a)    TraceProp(a)
#else
#define InfoTrace(a)        0
#define InfoTraceRaw(a)     0
#define InfoTraceProp(a)    0
#endif // DO_INFO_TRACES

enum
{
    OO_INST,
    OO_ENTRYID,
    OO_DEPTH,
    OO_OBJTYPE,
    OO_DISPTYPE,
    OO_DISPNAME,
    OO_ADDRTYPE,
    OO_SELECTABLE,
    NUM_ONE_OFF_TABLE_PROPS
};

SizedSPropTagArray(NUM_ONE_OFF_TABLE_PROPS, sptOneOffProps) =
{
    NUM_ONE_OFF_TABLE_PROPS,
    {
        PR_INSTANCE_KEY,
        PR_ENTRYID,
        PR_DEPTH,
        PR_OBJECT_TYPE,
        PR_DISPLAY_TYPE,
        PR_DISPLAY_NAME,
        PR_ADDRTYPE,
        PR_SELECTABLE
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::CABLogon()
//
//    Parameters
//      hInstance               Handle to the instance of the SP DLL
//      pBEobj                  Pointer to the object where the data comes from
//      pNPList                 Pointer to the named property map list
//      pSupObj                 MAPI's support object
//      dwConnectionID          ID of the connection with the server
//      pUserMailboxInfo        Pointer to a strucuture with information for
//                              the user that is logging on.
//      dwMailboxID             Server ID of the mailbox we are login onto
//      pszRemoteServer         Name of the remote host
//      ConnectionType          Type of connection to the host
//      pszLocalABD             Full path to the local file with the AB cache
//                              for the server receipients
//      pszLocalABDLastUpdate   String with the date and time the cache was
//                              last updated by the user.
//      hUIMutex                Handle to a cross-process mutex to syncronize
//                              access to the configuration property sheets of
//                              this service provider.
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CABLogon::CABLogon (HINSTANCE           hInstance,
                    CBackEnd *          pBEobj,
                    CNamePropMgr *      pNPList,
                    LPMAPISUP           pSupObj,
                    DWORD               dwConnectionID,
                    MAILBOX_INFO *      pUserMailboxInfo,
                    DWORD               dwMailboxID,
                    LPTSTR              pszRemoteServer,
                    CONNECTION_TYPE     ConnectionType,
                    LPTSTR              pszLocalABD,
                    LPTSTR              pszLocalABDLastUpdate,
                    HANDLE              hUIMutex) : CGeneric()
{
    InfoTrace ("CABLogon: Constructor called");
    DWORD dwThreadID;
    HRESULT hResult;
    char szMailslotName[64];
    // If the connection ID is 0, then the server could not set a notification link
    // to us, so don't botther creating a mailslot and listening thread.
    if (dwConnectionID)
    {
        wsprintf (szMailslotName,
                  CLIENT_MAILSLOT_SINK_NAME_FORMAT,
                  AB_WINDS_NOTIFICATION_MAILSLOT,
                  dwConnectionID);
        m_hMailslot = CreateMailslot (szMailslotName, 0, MAILSLOT_WAIT_FOREVER, NULL);
        if (INVALID_HANDLE_VALUE == m_hMailslot)
        {
            hResult = HRESULT_FROM_WIN32(::GetLastError()); // The Win32 API
            TraceResult ("CABLogon: Failed to create notification mailslot", hResult);
            throw CException (hResult);
        }
        else
        {
            HANDLE hThread = CreateThread (NULL,
                                           0,
                                           (LPTHREAD_START_ROUTINE)ABMailslotListenThreadProc,
                                           (LPVOID)this,
                                           CREATE_SUSPENDED,
                                           &dwThreadID);
            if (hThread)
            {
                SetThreadPriority (hThread, THREAD_PRIORITY_LOWEST);
                ResumeThread (hThread);
                CloseHandle (hThread);
            }
            else
            {
                hResult = HRESULT_FROM_WIN32(::GetLastError()); // The Win32 API BEFORE the CloseHandle() call
                CloseHandle (m_hMailslot);
                TraceResult  ("CABLogon: Failed to create listen thread", hResult);
                throw CException (hResult);
            }
        }
    }
    // Initialize the entry ID of the identity of the user in this service
    ZeroMemory (&m_UserEID, CB_PRIVATE_EID);
    m_UserEID.uidGlobal = pBEobj->GetBackEndUID();
    m_UserEID.bVersion = ENTRYID_VERSION;
    m_UserEID.bObject = MAPI_MAILUSER;
    m_UserEID.dwObjID = dwMailboxID;
    m_UserInfo = *pUserMailboxInfo;
    lstrcpy (m_szServer, pszRemoteServer);
    lstrcpy (m_szLocalABD, pszLocalABD);
    lstrcpy (m_szLocalABDLastUpdate, pszLocalABDLastUpdate);

    m_hInstance = hInstance;
    m_hUIMutex = hUIMutex;
    m_pSupObj = pSupObj;
    m_pSupObj->AddRef();
    m_pBEobj = pBEobj;
    m_pBEobj->AddRef();
    m_pOneOffTable = NULL;
    m_pGALTable = NULL;
    m_pNPList = pNPList;

    m_pStatusObj = NULL;

    m_dwServerConnectionID = dwConnectionID;
    m_ConnectionType = ConnectionType;
    InitializeCriticalSection (&m_csGALCache);
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::~CABLogon()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CABLogon.
//
//    Return Value
//      None
//
CABLogon::~CABLogon ()
{
    InfoTrace ("CABLogon: Destructor called");
    if (m_hMailslot)
    {
        CloseHandle (m_hMailslot);
    }
    m_pSupObj->Release();
    m_pSupObj = NULL;
    m_pBEobj->Release();
    m_pBEobj = NULL;
    if (m_pOneOffTable)
    {
        m_pOneOffTable->Release();
    }
    if (m_pGALTable)
    {
        m_pGALTable->Release();
    }
    // Delete the contained CMAPIStatus object, if any.
    if (m_pStatusObj)
    {
        m_pStatusObj->Release();
        m_pStatusObj = NULL;
    }
    if (m_hUIMutex)
    {
        CloseHandle (m_hUIMutex);
    }
    if (m_pNPList)
    {
        delete m_pNPList;
    }
    DeleteCriticalSection (&m_csGALCache);
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    CheckParameters_IUnknown_QueryInterface (this, riid, ppvObj);
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IABLogon || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IABLogon virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return a MAPIERROR structure filled up with meaningful descriptions
//      about an error that occurred in our provider. The caller is responsible
//      for freeing the memory returned.
//      In our implementation we delegate the call up to the base class which
//      implements this call.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::GetLastError (HRESULT        hError,
                                     ULONG          ulFlags,
                                     LPMAPIERROR *  ppMAPIError)
{
    InfoTrace ("CABLogon::GetLastError method called");
    CheckParameters_IABLogon_GetLastError (this, hError, ulFlags, ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::Logoff()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Closes down any open resources. In the case of this provider we
//      terminate any remote connections with the remote host.
//
//    Return Value
//      S_OK always.
//
STDMETHODIMP CABLogon::Logoff (ULONG ulFlags)
{
    InfoTrace ("CABLogon::Logoff method called");
    CheckParameters_IABLogon_Logoff (this, ulFlags);
    // Close down the mailslot handle. This will cause the waiting thread to terminate it self;
    CloseHandle (m_hMailslot);
    m_hMailslot = NULL;
    // Tell the server to remove us for it's list of notifications
    TerminateServerNotifications (m_szServer, m_dwServerConnectionID);
    // Release the RPC resources to connect to the remote host
    TerminateRemoteConnections();
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::OpenEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Open an object in this address book regardless of the container the
//      object is in. Here we also return the root container of this address
//      book when the caller specifies 0 and NULL for the entry ID parameters.
//      The returned objects are ready to use and the caller is responsible
//      for releasing them.
//      This implementation allows opening READ-ONLY instances of objects
//      meaning that we don't allow the MAPI_MODIFY flag to be passed in.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::OpenEntry (ULONG             cbEntryID,
                                  LPENTRYID         pEntryID,
                                  LPCIID            pInterface,
                                  ULONG             ulFlags,
                                  ULONG *           pulObjType,
                                  LPUNKNOWN *       ppUnk)
{
    InfoTrace ("CABLogon::OpenEntry method called");
    CheckParameters_IABLogon_OpenEntry (this,
                                        cbEntryID,
                                        pEntryID,
                                        pInterface,
                                        ulFlags,
                                        pulObjType,
                                        ppUnk);
    //In out implementation we only support read-only access to objects
    if (ulFlags & MAPI_MODIFY)
    {
        TraceMessage ("CABLogon::OpenEntry: Entries are READ-ONLY, access denied");
        return E_ACCESSDENIED;
    }
    HRESULT hResult = S_OK;
    PRIVATE_ENTRYID eidXPEntry = { 0 }, * pEID = cbEntryID ? (PRIVATE_ENTRYID *)pEntryID : NULL;
    if (CB_PRIVATE_XP_EID == cbEntryID)
    {
        PRIVATE_XP_ENTRYID * pXPEID = (PRIVATE_XP_ENTRYID *)pEntryID;
        if (!IsValidXPEntryID (pXPEID))
        {
            TraceMessage ("CABLogon::OpenEntry: Invalid Transport requested, entry ID passed in");
            return MAPI_E_INVALID_ENTRYID;
        }
        hResult = GetABEntryIDFromXPEntryID (*pXPEID, &eidXPEntry);
        if (hResult)
        {
            return hResult;
        }
        pEID = &eidXPEntry;
        cbEntryID = CB_PRIVATE_EID;
    }
    if (pEID)
    {
        if (CB_PRIVATE_EID != cbEntryID || !m_pBEobj->IsValidEntryID (pEID))
        {
            TraceMessage ("CABLogon::OpenEntry: Invalid entry ID passed in");
            return MAPI_E_INVALID_ENTRYID;
        }
        if (pInterface)
        {
            BOOL fBadInterface = FALSE;
            switch (pEID->bObject)
            {
                case MAPI_DISTLIST :
                case MAPI_ABCONT :
                    if (IID_IABContainer != *pInterface &&
                        IID_IDistList != *pInterface &&
                        IID_IMAPIContainer != *pInterface &&
                        IID_IMAPIProp != *pInterface &&
                        IID_IUnknown != *pInterface)
                    {
                        fBadInterface = TRUE;
                    }
                    break;
                case MAPI_MAILUSER :
                    if (IID_IMailUser != *pInterface &&
                        IID_IMAPIProp != *pInterface &&
                        IID_IUnknown != *pInterface )
                    {
                        fBadInterface = TRUE;
                    }
                    break;
            }
            if (fBadInterface)
            {
                TraceMessage ("CABLogon::OpenEntry: Unsupported interface requested");
                return E_NOINTERFACE;
            }
        }
    }

    // Create an entry ID of the root and have it readly available to
    // supply it as the parent of the entries.
    PRIVATE_ENTRYID eidRoot = { 0 };
    eidRoot.uidGlobal = m_pBEobj->GetBackEndUID();
    eidRoot.bVersion = ENTRYID_VERSION;
    eidRoot.bObject = MAPI_ABCONT;
    eidRoot.dwObjID = ROOT_CONTAINER_ID;
    // This is the entry ID of the ROOT container  { 0, NULL }
    if (!pEID || MAPI_ABCONT == pEID->bObject)
    {
        if (NULL == pInterface ||
            IID_IABContainer == *pInterface ||
            IID_IDistList == *pInterface ||
            IID_IMAPIContainer == *pInterface ||
            IID_IMAPIProp == *pInterface ||
            IID_IUnknown == *pInterface)
        {
            if (!pEID || ROOT_CONTAINER_ID == pEID->dwObjID)
            {
                pEID = &eidRoot;
            }
        }
        else
        {
            TraceMessage ("CABLogon::OpenEntry: Unsupported interface requested for root container");
            return E_NOINTERFACE;
        }
    }

    CABContainer * pContObj = NULL;
    CMailUser * pUserObj = NULL;
    CPropList * pList = NULL;
    BOOL fIsOneOff = FALSE;
    
    EnterCriticalSection (&m_csObj);
    
    // If the entry ID is for a MAILUSER object, check to see if the ID is for a ONE-OFF
    if ((MAPI_MAILUSER == pEID->bObject) && (0xFFFFFFFF == pEID->dwObjID))
    {
        // Bingo, this is a special case for us so go get the properties for the one-off.
        fIsOneOff = TRUE;
        hResult = GetOneOffProperties (&pList);
    }
    if (!hResult && !fIsOneOff)
    {
        // Looks like the object is a regular one. Ask the Back end to supply the properties of it.
        hResult = m_pBEobj->OpenObject (*pEID, &pList);
    }

    PRIVATE_ENTRYID * pParentEID = &eidRoot;
    if (!hResult && (MAPI_ABCONT == pEID->bObject || MAPI_DISTLIST == pEID->bObject))
    {
        // Allocate a new container object. Since distribution list and address book container
        // have the same interfaces they were implemented in the sample class. When behaviour
        // needs to be different between the two, we check the flags of the object and
        // the entry ID's object type field
        pContObj = new CABContainer (this,
                                     pParentEID,
                                     pEID,
                                     (MAPI_ABCONT == pEID->bObject ? MAPI_ABCONT : MAPI_DISTLIST),
                                     pList,
                                     m_pNPList,
                                     MAPI_ACCESS_READ);
        if (pContObj)
        {
            // These types of containers only have containers
            if (ROOT_CONTAINER_ID == pEID->dwObjID || GATEWAY_CONTAINERS_ID == pEID->dwObjID)
            {
                pContObj->m_ulObjFlags |= OBJECT_MAY_HAVE_CONTAINERS;
            }
            else
            {
                // These types of containers only have recipients
                pContObj->m_ulObjFlags |= OBJECT_MAY_HAVE_RECIPIENTS;
            }
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    if (!hResult && (MAPI_MAILUSER == pEID->bObject))
    {
        pUserObj = new CMailUser (this, pEID, pList, m_pNPList, MAPI_ACCESS_READ);
        if (pUserObj)
        {
            if (fIsOneOff)
            {
                pUserObj->m_ulObjFlags |= OBJECT_IS_ONE_OFF;
            }
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    LeaveCriticalSection (&m_csObj);
    if (hResult)
    {
        TraceResult ("CABLogon::OpenEntry", hResult);
        if (pList)
        {
            delete pList;
        }
    }
    else
    {
        // Return the object to the caller with the appropiate object type set.
        // Note that the constructor of the object placed the reference count
        // of the objects to 1 meaning that the object has been AddRef()'ed
        // and the caller is responsible for releasing it.
        if (!pEID || MAPI_ABCONT == pEID->bObject)
        {
            *pulObjType = MAPI_ABCONT;
            *ppUnk = (LPUNKNOWN)pContObj;
        }
        else
        {
            if (MAPI_DISTLIST == pEID->bObject)
            {
                *pulObjType = MAPI_DISTLIST;
                *ppUnk = (LPUNKNOWN)pContObj;
            }
            else
            {
                ASSERT (MAPI_MAILUSER == pEID->bObject);
                *pulObjType = MAPI_MAILUSER;
                *ppUnk = (LPUNKNOWN)pUserObj;
            }            
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::CompareEntryIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Given two entry ID's of some of our objects, return TRUE or FALSE
//      indicating weather or not the two entry ID's refer to the same
//      object in our provider.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::CompareEntryIDs (ULONG           cbEntryID1,
                                        LPENTRYID       pEntryID1,
                                        ULONG           cbEntryID2,
                                        LPENTRYID       pEntryID2,
                                        ULONG           ulFlags,
                                        ULONG *         pulResult)
{
    InfoTrace ("CABLogon::CompareEntryIDs method called");
    CheckParameters_IABLogon_CompareEntryIDs (this,
                                              cbEntryID1,
                                              pEntryID1,
                                              cbEntryID2,
                                              pEntryID2,
                                              ulFlags,
                                              pulResult);
    if ((CB_PRIVATE_EID != cbEntryID1) || (CB_PRIVATE_EID != cbEntryID2))
    {
        TraceMessage ("CABLogon::CompareEntryIDs: Getting called to compare two foreign entryIDs? Why?");
        return MAPI_E_UNKNOWN_ENTRYID;
    }
    PRIVATE_ENTRYID * pEID1 = (PRIVATE_ENTRYID *)pEntryID1;
    PRIVATE_ENTRYID * pEID2 = (PRIVATE_ENTRYID *)pEntryID2;
    HRESULT hResult = S_OK;
    __try
    {
        // All the fields in the private entry ID structure must be exaclty the same or there is no hit.
        if (pEID1 &&
            pEID2 &&
            cbEntryID1 == cbEntryID2 &&
            memcmp (&pEID1->uidGlobal, &pEID2->uidGlobal, sizeof(MAPIUID)) == 0 &&
            pEID1->bVersion == pEID2->bVersion &&
            pEID1->bObject == pEID2->bObject &&
            pEID1->dwObjID == pEID2->dwObjID)
        {
            *pulResult = TRUE;
        }
        else
        {
            *pulResult = FALSE;
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CABLogon::CompareEntryIDs: An exception occured. Exception code: %d", GetExceptionCode());
        hResult = MAPI_E_INVALID_ENTRYID;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::Advise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't support this in our provider implementation.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABLogon::Advise (ULONG                cbEntryID,
                               LPENTRYID            pEntryID,
                               ULONG                ulEventMask,
                               LPMAPIADVISESINK     pAdviseSink,
                               ULONG *              pulConnection)
{
    InfoTrace ("CABLogon::Advise method called");
    CheckParameters_IABLogon_Advise (this,
                                     cbEntryID,
                                     pEntryID,
                                     ulEventMask,
                                     pAdviseSink,
                                     pulConnection);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::Unadvise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't support this in our provider implementation.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABLogon::Unadvise (ULONG ulConnection)
{
    InfoTrace ("CABLogon::Unadvise method called");
    CheckParameters_IABLogon_Unadvise (this, ulConnection);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::OpenStatusEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return our implementation of IMAPIStatus to the client so that they
//      can get status and general information about the state and external
//      connections of this provider.
//      Only one status object need to be create for each Logon object so that
//      changes on the session can be updated to the one instance of
//      IMAPIStatus.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::OpenStatusEntry (LPCIID          pInterface,
                                        ULONG           ulFlags,
                                        ULONG *         pulObjType,
                                        LPMAPISTATUS *  ppEntry)
{
    InfoTrace ("CABLogon::OpenStatusEntry method called");
    CheckParameters_IABLogon_OpenStatusEntry (this,
                                              pInterface,
                                              ulFlags,
                                              pulObjType,
                                              ppEntry);
    if (MAPI_MODIFY & ulFlags)
    {
        TraceMessage ("CABLogon::OpenStatusEntry: We don't support WRITE access to the status object");
        return E_ACCESSDENIED;
    }
    if (pInterface)
    {
        if (IID_IMAPIStatus != *pInterface && IID_IMAPIProp != *pInterface)
        {
            TraceMessage ("CABLogon::OpenStatusEntry: We don't support the interface requested");
            return E_NOINTERFACE;
        }
    }
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
	// Now, if we already have created a status object on this logon context, 
	// we'll just AddRef() it and return it to the caller.
    if (!m_pStatusObj)
	{
        // Get the profile section of the PROVIDER so that we may get some properties,
        // assigned by MAPI to this provider, directly on the status object.
        LPPROFSECT pProfileObj = NULL;
        m_pSupObj->OpenProfileSection (NULL, 0, &pProfileObj);
        // If we don't have an object, create it, and save a copy in the logon object
        m_pStatusObj = new CABMAPIStatus (this, pProfileObj);
        if (!m_pStatusObj)
        {
            TraceMessage ("CABLogon::OpenStatusEntry: Could not allocate new CMAPIStatus object");
            hResult = E_OUTOFMEMORY;
        }
        // The constructor of CMAPIStatus AddRef()'ed this object
        if (pProfileObj)
        {
            pProfileObj->Release();
        }
	}
    if (!hResult)
    {
        m_pStatusObj->AddRef();
        *pulObjType = MAPI_STATUS;
        *ppEntry = m_pStatusObj;
    }
	//	Release the critical section
	LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::OpenTemplateID()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Open a templace for objects in our address book, so that the right
//      property pages can be displayed in foreign address book.
//      In our implementation, only distribution list have a template ID so
//      that if they are copied and details table open in a foreign address
//      book, the foreign address book, routes the call back to us.
//      (actually we wrap the foreign address book)
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::OpenTemplateID (ULONG        cbTemplateID,
                                       LPENTRYID    pTemplateID,
                                       ULONG        ulTemplateFlags,
                                       LPMAPIPROP   pMAPIPropData,
                                       LPCIID       pInterface,
                                       LPMAPIPROP * ppMAPIPropNew,
                                       LPMAPIPROP   pMAPIPropSibling)
{
    InfoTrace ("CABLogon::OpenTemplateID method called");
    CheckParameters_IABLogon_OpenTemplateID (this,
                                             cbTemplateID,
                                             pTemplateID,
                                             ulTemplateFlags,
                                             pMAPIPropData,
                                             pInterface,
                                             ppMAPIPropNew,
                                             pMAPIPropSibling);
    // Make sure the entry ID is for a DL object in our address book
    PRIVATE_ENTRYID * pTempID = (PRIVATE_ENTRYID *)pTemplateID;
    if (CB_PRIVATE_EID != cbTemplateID || NULL == pTempID || 
        EID_OBJTYPE_DL_TEMPLATE != pTempID->bObject || ENTRYID_VERSION != pTempID->bVersion)
    {
        TraceMessage ("CABLogon::OpenTemplateID: The caller gave us a bogus template ID was passed");
        return MAPI_E_INVALID_ENTRYID;
    }

    // If we are getting called to FILL the entry ID, the we just return the passed IMAPIProp
    // because we don't implement this (we don't need to)
    if (FILL_ENTRY & ulTemplateFlags)
    {
        *ppMAPIPropNew = pMAPIPropData;
        return S_OK;
    }

    // Allocate the wrapper IDistList interface which will defer all the methods called to
    // the contained IMAPIProp, except OpenProperty where we intercept it to handle PR_CONTENTS_TABLE.
    PRIVATE_ENTRYID ObjectEID = *pTempID;
    ObjectEID.bObject = MAPI_DISTLIST;
    HRESULT hResult = S_OK;
    CDLTemplate * pDLTempID = new CDLTemplate (pMAPIPropData, this, m_pBEobj, ObjectEID);
    if (pDLTempID)
    {
        // This is the object the foreign address book returns to the client doing
        // an open entry on it.
        *ppMAPIPropNew = (LPMAPIPROP)pDLTempID;
    }
    else
    {
        hResult = E_OUTOFMEMORY;
        TraceMessage ("CABLogon::OpenTemplateID: Failed to allocate a new DL template ID wrapper");
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::GetOneOffTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Returns an IMAPITable with all the type of address templates that we
//      support. In our implementation, we only return one template in the
//      ONE-OFF table. The entry ID is fixed so that we may recognize it
//      in the call to IABLogon::OpenEntry().
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::GetOneOffTable (ULONG            ulFlags,
                                       LPMAPITABLE *    ppTable)
{
    InfoTrace ("CABLogon::GetOneOffTable method called");
    CheckParameters_IABLogon_GetOneOffTable (this, ulFlags, ppTable);
    
    HRESULT hResult = S_OK;
    if (!m_pOneOffTable)
    {
        EnterCriticalSection (&m_csObj);
        hResult = CreateTable (&IID_IMAPITableData,
                               GetMAPIAllocBuffer(),
                               GetMAPIAllocMore(),
                               GetMAPIFreeBuffer(),
                               NULL,
                               TBLTYPE_DYNAMIC,
                               PR_INSTANCE_KEY,
                               (LPSPropTagArray)&sptOneOffProps,
                               &m_pOneOffTable);
        if (!hResult)
        {
            DWORD dwInstKey = 1;
            SPropValue spvProps[NUM_ONE_OFF_TABLE_PROPS] = { 0 };
            for (ULONG i=0; i<NUM_ONE_OFF_TABLE_PROPS; i++)
            {
                spvProps[i].ulPropTag = sptOneOffProps.aulPropTag[i];
            }
            PRIVATE_ENTRYID eidOneOff = { 0 };
            eidOneOff.uidGlobal = m_pBEobj->GetBackEndUID();
            eidOneOff.bVersion = ENTRYID_VERSION;
            eidOneOff.bObject = MAPI_MAILUSER;
            eidOneOff.dwObjID = 0xFFFFFFFF;
            spvProps[OO_INST].Value.bin.cb = sizeof(DWORD);
            spvProps[OO_INST].Value.bin.lpb = (LPBYTE)&dwInstKey;
            spvProps[OO_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
            spvProps[OO_ENTRYID].Value.bin.lpb = (LPBYTE)&eidOneOff;
            spvProps[OO_DEPTH].Value.l = 0;
            spvProps[OO_OBJTYPE].Value.l = MAPI_MAILUSER;
            spvProps[OO_DISPTYPE].Value.l = DT_MAILUSER;
            spvProps[OO_DISPNAME].Value.lpszA = "WINDS Address";
            spvProps[OO_ADDRTYPE].Value.lpszA = WINDS_ADDRESS_TYPE;
            spvProps[OO_SELECTABLE].Value.l = MAPI_MAILUSER;
            SRow row = { 0 };
            row.cValues = NUM_ONE_OFF_TABLE_PROPS;
            row.lpProps = spvProps;
            HRESULT hTableError = m_pOneOffTable->HrModifyRow (&row);
            TraceResult ("CABLogon::GetOneOffTable: Failed to add our row", hTableError);
        }
        LeaveCriticalSection (&m_csObj);
    }
    if (!hResult)
    {
        hResult = m_pOneOffTable->HrGetView (NULL, NULL, 0, ppTable);
    }
    TraceResult ("CABLogon::GetOneOffTable", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::PrepareRecips()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Replace the short term entry ID of the recipient for the long term one.
//      Request properties in the recipients, needed for the submission of
//      a message (i.e. PR_SEARCH_KEY for each recipient)
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::PrepareRecips (ULONG             ulFlags,
                                      LPSPropTagArray   pPropTagArray,
                                      LPADRLIST         pRecipList)
{
    InfoTrace ("CABLogon::PrepareRecips method called");
    CheckParameters_IABLogon_PrepareRecips (this,
                                            ulFlags,
                                            pPropTagArray,
                                            pRecipList);
    if (!pPropTagArray)
    {
        // If the array of property tags is NULL, MAPI's IAddrBook, is only
        // trying to conver the short term entry ID's of our objects, to long
        // term. Since our entries are ALWAYS long term, there is nothing for us to do.
        return S_OK;
    }
    BOOL fUseObjProp;
    ULONG ulCountRecipProps, ulObjType, ulObjProps, ulProp;
    PRIVATE_ENTRYID * pPrivEID;
    LPMAPIPROP pPropObj;
    LPSPropValue pRecipProps, pOneProp, pObjProps = NULL, pNewEntryProps = NULL;
    HRESULT hResult = S_OK;
    for (ULONG ulRecip=0; ulRecip<pRecipList->cEntries; ulRecip++)
    {
        pRecipProps = pRecipList->aEntries[ulRecip].rgPropVals;
        ulCountRecipProps = pRecipList->aEntries[ulRecip].cValues;
        // Get the entry ID of each ADRENTRY we are processing
        pOneProp = PpropFindProp (pRecipProps, ulCountRecipProps, PR_ENTRYID);
        if (NULL == pOneProp)
        {
            // We can't process a recipient without it's entry ID.
            continue;
        }
        pPrivEID = (PRIVATE_ENTRYID *)pOneProp->Value.bin.lpb;
        if (CB_PRIVATE_EID != pOneProp->Value.bin.cb || !m_pBEobj->IsValidEntryID (pPrivEID))
        {
            // We can't and don't need to process a recipient that is not an entry in our address book
            continue;
        }

        // This is a recipient in our address book. Open it.
        hResult = OpenEntry (CB_PRIVATE_EID,
                             (LPENTRYID)pPrivEID,
                             NULL,
                             0,
                             &ulObjType,
                             (LPUNKNOWN *)&pPropObj);
        if (hResult)
        {
             // Couldn't open it. Process the next entry.
            hResult = S_OK;
            continue;
        }
        ASSERTMSG (ulObjType == pPrivEID->bObject, "But this is not the object I asked for?!?!");
        hResult = pPropObj->GetProps (pPropTagArray, 0, &ulObjProps, &pObjProps);
        pPropObj->Release();
        if (FAILED(hResult)) // In case we get warnings
        {
            // Could not get the properties of this object. Process the next entry.
            hResult = S_OK;
            continue;
        }
        ASSERTMSG (ulObjProps == pPropTagArray->cValues, "But this is not how many properties I asked for?!?!");

        // Merge the two property sets: pObjProps and pRecipProps.  Note that
        // both of these sets may have the same property - chances are they do.
        // For these conflicts, pObjProps should be the one we get the property
        // from. Calculate how big the resultant combined SPropValue array is,
        // and allocate one of that size.
        hResult = PrivMAPIAllocate ((ulObjProps + ulCountRecipProps)*sizeof(SPropValue), (LPVOID *)&pNewEntryProps);
        if (hResult)
        {
            // It's Ok to fail the call here. If we're running into out of memory
            // conditions we're all in trouble.
            goto ErrorExit;
        }

        // Copy pObjProps properties over to pNewEntryProps. Check each property
        // in pObjProps to ensure that it isn't PT_ERROR, if so find the propval
        // in pRecipProps (the [in] recip prop val array), if it exists and
        // use that property.
        for (ulProp=0; ulProp<ulObjProps; ulProp++)
        {
            fUseObjProp = TRUE;
            if (PT_ERROR == PROP_TYPE(pObjProps[ulProp].ulPropTag))
            {
                pOneProp = PpropFindProp (pRecipProps, ulCountRecipProps, pPropTagArray->aulPropTag[ulProp]);
                if (pOneProp)
                {
                    hResult = PropCopyMore (pNewEntryProps + ulProp, pOneProp, GetMAPIAllocMore(), pNewEntryProps);
                    fUseObjProp = FALSE;

                }
            }
            if (fUseObjProp)
            {
                hResult = PropCopyMore (&pNewEntryProps[ulProp], &pObjProps[ulProp], GetMAPIAllocMore(), pNewEntryProps);
            }
            if (E_OUTOFMEMORY == hResult)
            {
                goto ErrorExit;
            }
        }

        // We are done with pObjProps
        PrivMAPIFree ((LPVOID *)&pObjProps);

        // Copy those properties that aren't already in pNewEntryProps
        // from pRecipProps. Don't copy over the PT_ERROR prop vals
        for (ulProp=0; ulProp<ulCountRecipProps; ulProp++)
        {
            if (PpropFindProp (pNewEntryProps, ulObjProps, pRecipProps[ulProp].ulPropTag) ||
                PT_ERROR == PROP_TYPE (pRecipProps[ulProp].ulPropTag))
            {
                continue;
            }
            hResult = PropCopyMore (&pNewEntryProps[ulObjProps], &pRecipProps[ulProp], GetMAPIAllocMore(), pNewEntryProps);
            if (E_OUTOFMEMORY == hResult)
            {
                goto ErrorExit;
            }
            ulObjProps++;
        }
        // Replace the ADRENTRY in the ADRLIST with pNewEntryProps. And don't forget the cValues!
        pRecipList->aEntries[ulRecip].rgPropVals = pNewEntryProps;
        pRecipList->aEntries[ulRecip].cValues    = ulObjProps;
        pNewEntryProps = NULL;
        ulObjProps = 0;
        // Finally, free up the processed entry
        PrivMAPIFree ((LPVOID *)&pRecipProps);
    }
ErrorExit:
    PrivMAPIFree ((LPVOID *)&pObjProps);
    PrivMAPIFree ((LPVOID *)&pNewEntryProps);
    TraceResult ("CABLogon::PrepareRecips", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::GetOneOffProperties()
//
//    Parameters
//      ppList      Where we return a newly allocated property cache for the
//                  One-Off entry in the foreign address book.
//
//    Purpose
//      Creates a new property cache for One-off recipients opened from the
//      entry ID in the One-Off table of this provider. The properties are
//      empty place holder. The actual values are entered by the user in
//      the empty-template create in the address book where this one-off
//      entry will be saved.
//
//    Return Value
//      E_OUTOFMEMORY if we failed to allocate the property cache.
//      S_OK otherwise.
//
STDMETHODIMP CABLogon::GetOneOffProperties (CPropList ** ppList)
{
    *ppList = new CPropList;
    if (!(*ppList))
    {
        TraceMessage ("CABLogon::GetOneOffProperties: Failed to allocate property cache");
        return E_OUTOFMEMORY;
    }
    // This is a special search key in our address book
    char szSearchKey[64];
    wsprintf (szSearchKey, "%s:\\\\_server\\one-off", WINDS_ADDRESS_TYPE);
    CharUpper (szSearchKey);
    // Set empty strings and default values for the properties of a one-off mailuser
    SPropValue spvProps[13] = { 0 };
    DWORD i = 0;
    spvProps[i].ulPropTag       = PR_DISPLAY_TYPE;
    spvProps[i++].Value.l       = DT_MAILUSER;
    spvProps[i].ulPropTag       = PR_ACCOUNT;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_DISPLAY_NAME;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_TRANSMITABLE_DISPLAY_NAME;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_TITLE;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_OFFICE_LOCATION;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_OFFICE_TELEPHONE_NUMBER;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_OTHER_TELEPHONE_NUMBER;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_PRIMARY_FAX_NUMBER;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_COMMENT;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_EMAIL_ADDRESS;
    spvProps[i++].Value.lpszA   = szBlank;
    spvProps[i].ulPropTag       = PR_SEND_RICH_INFO;
    spvProps[i++].Value.b       = TRUE;
    spvProps[i].ulPropTag       = PR_SEARCH_KEY;
    spvProps[i].Value.bin.cb    = Cbtszsize (szSearchKey);
    spvProps[i++].Value.bin.lpb = (LPBYTE)szSearchKey;
    // Save the properties in the cache
    for (UINT j=0; j<i; j++)
    {
        (*ppList)->Insert (&spvProps[j]);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::InitializeStatusRow()
//
//    Parameters
//      ulFlag      0 if we are adding the row for the first time in this
//                  session. STATUSROW_UPDATE we are changing values while
//                  the session is running
//
//    Purpose
//      Adds or modifies the status row of this provider on the MAPI
//      subsystem. The status row is where the provider register the identity
//      of the user in this service and also where we inform MAPI and clients
//      of the state and status of this provider.
//      Among the things that a provider could have in the table are:
//      State of the connection with the server, Name of the server
//      connecting to, name of the provider and/or service, etc.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::InitializeStatusRow (ULONG ulFlag)
{
    #define NUM_STATUS_ROW_PROPS    8
    SPropValue spvStatusRow[NUM_STATUS_ROW_PROPS] = { 0 };
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_PROVIDER_DISPLAY property: The transport readable name
    ULONG i = 0;
    spvStatusRow[i].ulPropTag = PR_PROVIDER_DISPLAY;
    spvStatusRow[i++].Value.lpszA = PROVIDER_DISPLAY_NAME_STRING;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_RESOURCE_METHODS property. These are the methods implemented
    // in the our IMAPIStatus implementation (CMAPIStatus class.)
    spvStatusRow[i].ulPropTag = PR_RESOURCE_METHODS;
    spvStatusRow[i++].Value.l = STATUS_SETTINGS_DIALOG | STATUS_CHANGE_PASSWORD;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_STATUS_CODE property.
    spvStatusRow[i].ulPropTag = PR_STATUS_CODE;
    // The status of the the provider is dictated by the state of the connection
    // and the availability of a local address book directory. Without both of
    // these data sources, the address book won't be able to resolve names,
    // rendering it useless.
    spvStatusRow[i++].Value.l = (NOT_AVAIL_NO_ABD == m_ConnectionType ? STATUS_OFFLINE : STATUS_AVAILABLE);
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_STATUS_STRING property
    char szBuffer[64];
    DWORD dwStringID = IDS_AB_STATUS_STRING_OFFSET + (DWORD)m_ConnectionType;
    LoadString (m_hInstance, dwStringID, szBuffer, 64);
    spvStatusRow[i].ulPropTag = PR_STATUS_STRING;
    spvStatusRow[i++].Value.lpszA = szBuffer;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_DISPLAY_NAME property
    char szDisplayName[64];
    wsprintf (szDisplayName,
              "%s (%s)",
              PROVIDER_DISPLAY_NAME_STRING,
              m_szServer);
    spvStatusRow[i].ulPropTag = PR_DISPLAY_NAME;
    spvStatusRow[i++].Value.lpszA = szDisplayName;
    
    // The identity of the user is actually an entry in this address book
    // so get the ID of the user's mailbox and put it in the status row.

    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_IDENTITY_DISPLAY property
    spvStatusRow[i].ulPropTag = PR_IDENTITY_DISPLAY;
    spvStatusRow[i++].Value.lpszA = GetUserName();

    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_IDENTITY_ENTRYID property
    spvStatusRow[i].ulPropTag = PR_IDENTITY_ENTRYID;
    spvStatusRow[i].Value.bin.cb = CB_PRIVATE_EID;
    spvStatusRow[i++].Value.bin.lpb = (LPBYTE)GetUserEID();
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_IDENTITY_SEARCH_KEY property
    // Search keys are in the format: "<ADDRESS TYPE>:<EMAIL ADDRESS>", in
    // upper case, so our search keys will look like:
    //      "WINDS:\\SERVERNAME\MAILBOX"
    char szSearchKey[128];
    wsprintf (szSearchKey,
              "%s:%s\\%s",
              WINDS_ADDRESS_TYPE,
              m_szServer, // this is already in the format "\\<servername>"
              m_UserInfo.szMailboxName);
    CharUpper (szSearchKey);
    spvStatusRow[i].ulPropTag = PR_IDENTITY_SEARCH_KEY;
    spvStatusRow[i].Value.bin.cb = Cbtszsize (szSearchKey);
    spvStatusRow[i++].Value.bin.lpb = (LPBYTE)szSearchKey;
    
    ASSERT (NUM_STATUS_ROW_PROPS == i);

    // Write the entries on the provider's session status row
    HRESULT hResult = m_pSupObj->ModifyStatusRow (i, spvStatusRow, 0);
    TraceResult ("CABLogon::InitializeStatusRow: Failed to register the row", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::IsValidXPEntryID()
//
//    Parameters
//      pXPEID      Pointer to the entry ID structure we will verify.
//
//    Purpose
//      Given a pointer to a private transport entry ID structure, determines
//      if its contents would be something in this address book.
//
//    Return Value
//      TRUE if the entry is not null and the UID matches the UID of this
//      provider. FALSE otherwise.
//
BOOL WINAPI CABLogon::IsValidXPEntryID (PRIVATE_XP_ENTRYID * pXPEID)
{
    // Check the members of the structure against our known values.
    BOOL bValid = TRUE;
    __try
    {
        if (!pXPEID ||
            memcmp (&pXPEID->uidGlobal, &guidXPEntries, sizeof(MAPIUID)) ||
            memcmp (&pXPEID->uidWINDSEntries, &guidABEntries, sizeof(MAPIUID)))
        {
            bValid = FALSE;
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CABLogon::IsValidXPEntryID: An exception occured. Exception code: %d", GetExceptionCode());
        bValid = FALSE;
    }
    return bValid;
}

///////////////////////////////////////////////////////////////////////////////
//    CABLogon::GetABEntryIDFromXPEntryID()
//
//    Parameters
//      XPEID   Entry ID passed by a WINDS transport
//      pEID    Pointer to a private entry ID where we return the information
//              mapped from XPEID
//
//    Purpose
//      We return an address book entry ID from the entry ID given by a WINDS
//      transport. To map the entry ID, we use the contents table of the GAL
//      to find the entry, if found, we return it's entry ID.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABLogon::GetABEntryIDFromXPEntryID (PRIVATE_XP_ENTRYID XPEID, 
                                                  PRIVATE_ENTRYID * pEID)
{
    HRESULT hResult;
    LPTABLEDATA pTableData;
    LPMAPITABLE pTableView;
    GetCachedGAL (&pTableData);
    if (NULL == pTableData)
    {
        PRIVATE_ENTRYID eidGAL = { 0 };
        eidGAL.uidGlobal = m_pBEobj->GetBackEndUID();
        eidGAL.bVersion = ENTRYID_VERSION;
        eidGAL.bObject = MAPI_ABCONT;
        eidGAL.dwObjID = GAL_CONTAINER_ID;
    
        hResult = m_pBEobj->GetTable (TABLE_CONTENTS,
                                      eidGAL,
                                      0,
                                      NULL,
                                      (LPVOID)FALSE,
                                      &pTableData);
        if (hResult)
        {
            TraceResult ("CABLogon::GetABEntryIDFromXPEntryID: Failed to get the GAL table", hResult);
            return hResult;
        }
        SetCachedGAL (pTableData);
    }

    // Get a view of the table
    hResult = pTableData->HrGetView (NULL, NULL, 0, &pTableView);
    if (!hResult)
    {
        LPSRowSet pRows;
        // Create the restriction. The transport entry ID has the object alias
        // which maps to the PR_ACCOUNT property in WINDS AB entries. So we
        // find a row that matches exactly the object alias passed in.
        // Note that there could only be ONE row that matches this PR_ACCOUNT.
        // No two object in this address book can have the same value
        // in PR_ACCOUNT.
        SPropValue spvString = { 0 };
        SRestriction SRes = { 0 };
        SRes.rt = RES_CONTENT;
        SRes.res.resContent.lpProp = &spvString;
        SRes.res.resContent.ulFuzzyLevel = FL_FULLSTRING | FL_IGNORECASE;
        SRes.res.resContent.ulPropTag = spvString.ulPropTag = PR_ACCOUNT;
        spvString.Value.lpszA = XPEID.szObjectAlias;
        
        // Limit the columns in the table to what we are interested in.
        const static SizedSPropTagArray (2, sptMappedEntry) =
        {
            2,
            {
                PR_ENTRYID,
                PR_ACCOUNT
            }
        };
        hResult = pTableView->SetColumns ((LPSPropTagArray)&sptMappedEntry, TBL_BATCH);
        if (!hResult)
        {
            // Find the row now with the restriction constructed above
            hResult = pTableView->FindRow (&SRes, BOOKMARK_BEGINNING, 0);
            if (!hResult)
            {
                // Get the found row and copy the entry ID value
                hResult = pTableView->QueryRows (1, 0, &pRows);
                if (!hResult)
                {
                    ASSERTMSG (pRows->cRows == 1, "But I asked for 1 row in the View?!");
                    ASSERTMSG (pRows->aRow[0].cValues == 2, "But I asked for 2 properties from the View?!");
                    ASSERTMSG (pRows->aRow[0].lpProps->ulPropTag == PR_ENTRYID, "Where is PR_ENTRYID in the view?!");
                    CopyMemory (pEID, pRows->aRow[0].lpProps->Value.bin.lpb, CB_PRIVATE_EID);
                    FreeProws (pRows);
                }
            }
        }
        pTableView->Release();
    }
    pTableData->Release();
    TraceResult ("CABLogon::GetABEntryIDFromXPEntryID", hResult);
    return hResult;
}

// End of file for ABLOGON.CPP
