///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      ABPROV.CPP
//
//  Description
//      This file implements the IABProvider interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//

#define INITGUID
#define USES_IID_IABProvider
#define USES_IID_IMAPIControl
#define USES_IID_IMAPIStatus
#define USES_IID_IMAPIProp
#define USES_IID_IMAPIPropData
#define USES_IID_IMAPIContainer
#define USES_IID_IMAPITableData
#define USES_IID_IMAPITable
#define USES_IID_IABContainer
#define USES_IID_IABLogon
#define USES_IID_IMessage
#define USES_IID_IMailUser
#define USES_IID_IDistList
#define USES_IID_IXPLogon
#define USES_IID_IXPProvider
#define USES_IID_IMAPIFolder
#define USES_PS_MAPI
#define USES_PS_PUBLIC_STRINGS
#define USES_IID_IStreamDocfile

#include "ABWDS.H"
#include "ABSRVBE.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceInfoMessage(a)
#else
#define InfoTrace(a)        0
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::CABProvider()
//
//    Parameters
//      hInst       Handle to instance of this provider DLL
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CABProvider::CABProvider (HINSTANCE hInst)
{
    InfoTrace ("CABProvider: Constructor called");
    m_cRef = 1;
    m_hInstance = hInst;
    m_pABLogonList = NULL;
    InitializeCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::~CABProvider()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of the object. Release the contained objects and internal
//      resources used in this instance.
//
//    Return Value
//      None
//
CABProvider::~CABProvider()
{
    InfoTrace ("CABProvider: Destructor called");
    
    m_hInstance = NULL;
    ASSERT (NULL == m_pABLogonList);

    // Deallocate the resources for the critical section
    DeleteCriticalSection (&m_csObj);

    CloseMemoryHeap();

    // Close down the log file/com port/console window/etc.
    UnInitTraces();
}

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it 
//      returns NULL.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABProvider::QueryInterface (REFIID riid,  LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IABProvider || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMSProvider virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::Shutdown()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      To terminate all session objects attached to this provider object
//
//    Return Value
//      S_OK                    If no error
//      MAPI_E_UNKNOWN_FLAGS    If flags are passed to this method
//
STDMETHODIMP CABProvider::Shutdown (ULONG * pulFlags)
{
    InfoTrace ("CABProvider::Shutdown method called");
    CheckParameters_IABProvider_Shutdown (this, pulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::Logon()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      Return and IABLogon object to MAPI in the client context and on the
//      spooler context. MAPI uses this object to get the containers in this
//      address book, and to manipulate the address book in general.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABProvider::Logon (LPMAPISUP          pSupObj,
                                 ULONG              ulUIParam,
                                 LPTSTR             pszProfileName,
                                 ULONG              ulFlags,
                                 ULONG *            pcbSecurity,
                                 LPBYTE *           ppbSecurity,
                                 LPMAPIERROR *      ppMAPIError,
                                 LPABLOGON *        ppABLogon)
{
    InfoTrace ("CABProvider::Logon method called");
    CheckParameters_IABProvider_Logon (this,
                                       pSupObj,
                                       ulUIParam,
                                       pszProfileName,
                                       ulFlags,
                                       pcbSecurity,
                                       ppbSecurity,
                                       ppMAPIError,
                                       ppABLogon);
    CGeneric obj;
    *ppMAPIError = NULL;
    *ppbSecurity = NULL;
    *pcbSecurity = 0;

    LPSPropValue pProps = NULL;
    CBackEnd * pBEobj = NULL;
    CABLogon * pABLogonObj = NULL;
    
    CFGDLG CfgDialog = { 0 };
    MAILBOX_INFO UserMBInfo = { 0 };
    ULONG ulPropCount;
    DWORD dwConnectionID;
    LPPROFSECT pProfileObj;

    CNamePropMgr * pNPList = NULL;

    HRESULT hResult = OpenServiceProfileSection (pSupObj, &pProfileObj, GetMAPIFreeBuffer());
    if (hResult)
    {
        TraceResult ("CABLogon::Logon: Failed to open the service profile section", hResult);
        return hResult;
    }
    
    HANDLE hUIMutex = CreateMutex (NULL, FALSE, CONFIG_UI_MUTEX);
    if (NULL == hUIMutex)
    {
        TraceResult ("CABLogon::Logon: Failed to create UI mutext", HRESULT_FROM_WIN32(GetLastError()));
    }

    hResult = pProfileObj->GetProps ((LPSPropTagArray)&sptLogonProps,
                                     fMapiUnicode,
                                     &ulPropCount,
                                     &pProps);
    if (FAILED(hResult))
    {
        TraceResult ("CABProvider::Logon: Failed to get the logon props", hResult);
        goto ErrorExit;
    }

    // Fill in the logon UI structure
    CfgDialog.hInst    = m_hInstance;
    CfgDialog.hWnd     = (HWND)ulUIParam;
    CfgDialog.ppProps  = &pProps;
    CfgDialog.pSupObj  = pSupObj;
    CfgDialog.hUIMutex = hUIMutex;
    
    // In case we get MAPI_W_ERRORS_RETURNED, ignore it and reset 
    // to S_OK. Now display the logon configuration dialog
    if (MAPI_W_ERRORS_RETURNED == hResult)
    {
        if (PR_SMP_MAILBOX_ID      != pProps[MAILBOX_ID].ulPropTag     ||
            PR_SMP_CONNECTION_TYPE != pProps[NET_CON].ulPropTag        ||
            PR_SMP_LUL_LAST_UPDATE != pProps[LUL_LASTUPDATE].ulPropTag ||
            PR_SMP_LUL_FILENAME    != pProps[LUL_FILENAME].ulPropTag)
        {
            TraceMessage ("CABLogon::Logon: Where are the provider properties?");
            hResult = MAPI_E_UNCONFIGURED;
            if (!(LOGON_NO_DIALOG & ulFlags))
            {
                PrivInitialize3DCtl (m_hInstance);
                PrivateMessageBox (IDS_MSG_SERVICE_NOT_PROPERLY_CFG, (HWND)ulUIParam);
            }
            goto ErrorExit;
        }

        hResult = S_OK;
        if (PR_SMP_REMOTE_SERVER != pProps[SERVER_NAME].ulPropTag ||
            PR_SMP_MAILBOX_NAME != pProps[MAILBOX_NAME].ulPropTag ||
            PR_SMP_USER_NAME != pProps[USER_NAME].ulPropTag ||
            PR_SMP_MAILBOX_PASSWORD != pProps[PASSWORD].ulPropTag)
        {
            if (AB_NO_DIALOG & ulFlags)
            {
                TraceMessage ("CABProvider::Logon: UI not allowed but it is needed");
                hResult = MAPI_E_UNCONFIGURED;
                goto ErrorExit;
            }
            PrivInitialize3DCtl (m_hInstance);
ReStartLogonDlg :
            hResult = DoLogonDlg (&CfgDialog);
            if (hResult)
            {
                goto ErrorExit;
            }
            hResult = pProfileObj->SetProps (ulPropCount, pProps, NULL);
            if (hResult)
            {
                TraceResult ("CABProvider::Logon: Failed to set the properties", hResult);
                goto ErrorExit;
            }
        }
    }
    ASSERTMSG (PR_SMP_CONNECTION_TYPE == pProps[NET_CON].ulPropTag, "Where is the connection information for the logon?");
    if (LAN_CONNECT == pProps[NET_CON].Value.l)
    {
        hResult = DoServerLogon (&UserMBInfo,
                                 pProps,
                                 pProfileObj,
                                 !(AB_NO_DIALOG & ulFlags),
                                 (HWND)ulUIParam,
                                 TRUE,
                                 WINDS_NOTIF_ON_AB | WINDS_NOTIF_ON_USER,
                                 &dwConnectionID,
                                 FALSE);
        if (S_FALSE == hResult)
        {
            goto ReStartLogonDlg;
        }
    }
    if (OFFLINE_CONNECT == pProps[NET_CON].Value.l ||
        MAPI_W_NO_SERVICE == hResult ||     // This is the error DoServerLogon() returns in the CLIENT context
        MAPI_E_NETWORK_ERROR == hResult)    // This is the error DoServerLogon() returns in the SPOOLER context
    {
        hResult = S_OK;
        pProps[NET_CON].Value.l = OFFLINE_CONNECT;
        // To avoid create a mailslot notification sink in the CABLogon constructor
        dwConnectionID = 0;
    }
    if (!hResult)
    {
        pNPList = new CNamePropMgr (GetMAPIAllocBuffer(),
                                    GetMAPIAllocMore(),
                                    GetMAPIFreeBuffer());
        if (!pNPList)
        {
            hResult = E_OUTOFMEMORY;
        }
        else
        {
            pNPList->InitializeForInMemoryUse();
        }
    }
    if (hResult)
    {
        goto ErrorExit;
    }

    // Since we are going to allocate a new object and manipulate the list of
    // sessions, lock the provider until we are finished
    EnterCriticalSection (&m_csObj);
    hResult = OpenRemoteDataBase (pProps[SERVER_NAME].Value.lpszA,
                                  &guidABEntries,
                                  (OFFLINE_CONNECT == pProps[NET_CON].Value.l),
                                  pProps[LUL_FILENAME].Value.lpszA,
                                  &pBEobj);
    if (!hResult)
    {
        try
        {
            pABLogonObj = new CABLogon (m_hInstance,
                                        pBEobj,
                                        pNPList,
                                        pSupObj,
                                        dwConnectionID,
                                        &UserMBInfo,
                                        pProps[MAILBOX_ID].Value.l,
                                        pProps[SERVER_NAME].Value.lpszA,
                                        (CONNECTION_TYPE)pProps[NET_CON].Value.l,
                                        pProps[LUL_FILENAME].Value.lpszA,
                                        pProps[LUL_LASTUPDATE].Value.lpszA,
                                        hUIMutex);
            if (!pABLogonObj)
            {
                hResult = E_OUTOFMEMORY;
            }
        }
        catch (CException & Exception)
        {
            hResult = Exception.GetError();
        }
        if (!hResult)
        {
            AddWindsABNamedPropsToList (pNPList);
            // To avoid closing it below. If we fail, the destructor
            // in CABLogon will take care of this.
            pNPList = NULL;
            hUIMutex = NULL;

            // The guidABEntries is used in the entry ID of this provider. The UID is
            // fixed (hardcoded) because only ONE instance of the provider is allowed
            // to be in the profile at any time. If multiple instances of the same
            // address book are allowed, the a new UID provider must be generated for
            // each session (i.e. IMAPISupport::NewUID()) and save the UID in the
            // profile section of your provider.
            hResult = pSupObj->SetProviderUID (&guidABEntries, 0);
            if (!hResult)
            {
                // Also register support for this UID in entry ID's which WINDS transport
                // providers set to open entries in this address book.
                hResult = pSupObj->SetProviderUID (&guidXPEntries, 0);
                if (!hResult)
                {
                    pABLogonObj->InitializeStatusRow();
                }
            }
        }
    }
    // Unlock the critical section
    LeaveCriticalSection (&m_csObj);

ErrorExit:
    pProfileObj->Release();
    if (pBEobj)
    {
        pBEobj->Release();
    }
    PrivMAPIFree ((LPVOID *)&pProps);

    PrivUninitialize3DCtl (m_hInstance);

    if (hUIMutex)
    {
        CloseHandle (hUIMutex);
    }

    // If we failed cleanup before leaving
    if (hResult)
    {
        TraceResult ("CABProvider::Logon", hResult);
        if (pABLogonObj)
        {
            pABLogonObj->Release();
        }
        obj.GetLastError (hResult, fMapiUnicode, ppMAPIError, m_hInstance, PROVIDER_DISPLAY_NAME_STRING);

        if (pNPList)
        {
            delete pNPList;
        }
    }
    else
    {
        // We were successful, so copy the objects into the returning parameters
        ASSERT (pABLogonObj);
        *ppABLogon = (LPABLOGON)pABLogonObj;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABProvider::AddWindsABNamedPropsToList()
//
//    Parameters
//      [IN]    pNPList     Pointer to a CNamePropMgr object where we will add
//                          the named properties to
//
//    Purpose
//      This function add the named properties to the generic Named Property
//      manager used in the implementation of our IMailUser and IDistList
//      interfaces (common IMAPIProp for both). This will return the ID for
//      the named properties in our recipient objects for when entries are
//      copied to other address books.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABProvider::AddWindsABNamedPropsToList (CNamePropMgr * pNPList)
{
#pragma warning (disable : 4127)
    ULONG ulNewID;
    pNPList->GetIDForName ((LPGUID)&guidABEntries,
                           WINDS_AB_MANAGER_NAME,
                           0,
                           MNID_STRING,
                           &ulNewID,
                           TRUE);
    ASSERT (0x8000 == PROP_ID(PR_WINDS_AB_MANAGER_NAME));

    pNPList->GetIDForName ((LPGUID)&guidABEntries,
                           WINDS_AB_MANAGER_EID,
                           0,
                           MNID_STRING,
                           &ulNewID,
                           TRUE);
    ASSERT (0x8001 == PROP_ID(PR_WINDS_AB_MANAGER_ENTRYID));

    pNPList->GetIDForName ((LPGUID)&guidABEntries,
                           WINDS_AB_DL_OWNER_ACCT,
                           0,
                           MNID_STRING,
                           &ulNewID,
                           TRUE);
    ASSERT (0x8002 == PROP_ID(PR_WINDS_AB_DL_OWNER_ACCT));

    pNPList->GetIDForName ((LPGUID)&guidABEntries,
                           WINDS_AB_DL_OWNER_NAME,
                           0,
                           MNID_STRING,
                           &ulNewID,
                           TRUE);
    ASSERT (0x8003 == PROP_ID(PR_WINDS_AB_DL_OWNER_NAME));
#pragma warning (default : 4127)
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    ABMailslotListenThreadProc()
//
//    Parameters
//      pLogon      Pointer to the AB logon session the thread is
//                  listening for
//
//    Purpose
//      This is function runs in a separate low priority thread listening for
//      notification that a WINDS server send to client provider. When a
//      notification of interest is received, an appropiate action is taken.
//      The notification arrive on a mailslot created by the logon object.
//
//    Return Value
//      S_OK always.
//
DWORD WINAPI ABMailslotListenThreadProc (CABLogon * pLogon)
{
    HANDLE hMailslot = pLogon->GetListenMailslot();
    DWORD dwRead;
    FILETIME ftLastNotifTime = { 0 };
    WINDS_NOTIF_EVENT LastEvent = GENERAL_NOTIF_MIN;
    WINDS_NOTIFICATION Notif;
    HRESULT hReadError;
    LPTABLEDATA pTable;
    BOOL fContinue = TRUE;
    while (fContinue)
    {
        pTable = NULL;
        if (!ReadFile (hMailslot, &Notif, sizeof(WINDS_NOTIFICATION), &dwRead, NULL)) 
        { 
            hReadError = HRESULT_FROM_WIN32 (GetLastError());
            if (HRESULT_FROM_WIN32(ERROR_INVALID_HANDLE) == hReadError ||
                HRESULT_FROM_WIN32(ERROR_HANDLE_EOF) == hReadError)
            {
                break; // out of the WHILE() loop and terminate the thread
            }
            TraceResult ("ABMailslotListenThreadProc: ReadFile", HRESULT_FROM_WIN32(GetLastError()));
            continue;
        }
        // Mailslot are received in each network protocol stack installed on the system.
        // Check the time stamp and event of the last notification received. If it
        // is the same, drop it and continue listen for new and different incoming
        // data on the mailslot.
        if (ftLastNotifTime.dwLowDateTime  == Notif.ftEventTime.dwLowDateTime &&
	        ftLastNotifTime.dwHighDateTime == Notif.ftEventTime.dwHighDateTime &&
            LastEvent                      == Notif.Event)
        {
            continue;
        }
        // Save the last event information
        ftLastNotifTime = Notif.ftEventTime;
        LastEvent = Notif.Event;

        #ifdef DO_INFO_TRACES
        switch (Notif.Event)
        {
            case AB_USER_ADDED :
                InfoTrace ("AB_USER_ADDED notification received from the WINDS server");
                break;
            case AB_USER_MODIFIED :
                InfoTrace ("AB_USER_MODIFIED notification received from the WINDS server");
                break;
            case AB_USER_DELETED :
                InfoTrace ("AB_USER_DELETED notification received from the WINDS server");
                break;
            case AB_DL_ADDED :
                InfoTrace ("AB_DL_ADDED notification received from the WINDS server");
                break;
            case AB_DL_MODIFIED :
                InfoTrace ("AB_DL_MODIFIED notification received from the WINDS server");
                break;
            case AB_DL_DELETED :
                InfoTrace ("AB_DL_DELETED notification received from the WINDS server");
                break;
            case SERVER_IS_SHUTTING_DOWN :
                InfoTrace ("SERVER_IS_SHUTTING_DOWN notification received from the WINDS server");
                break;
            case RESET_LINKS_WITH_SERVER :
                InfoTrace ("RESET_LINKS_WITH_SERVER notification received from the WINDS server");
                break;
            case SERVER_HAS_RESTARTED :
                InfoTrace ("SERVER_HAS_RESTARTED notification received from the WINDS server");
                break;
            case AB_GET_LOCAL_ABDATA_NOW :
                InfoTrace ("AB_GET_LOCAL_ABDATA_NOW notification received from the WINDS server");
                break;
            default :
                InfoTrace ("UNKNOWN notification received from the WINDS server");
                break;
        }
        #endif // DO_INFO_TRACES

        __try
        {
            pLogon->GetCachedGAL (&pTable);
            if ((NULL == pTable) && (Notif.Event >= AB_NOTIF_MIN  && Notif.Event <= AB_NOTIF_MAX))
            {
                continue;
            }
            switch (Notif.Event)
            {
                case AB_USER_DELETED :
                case AB_DL_DELETED :
                    {
                        DWORD dwInstKey;
                        if (AB_USER_DELETED == Notif.Event)
                        {
                            dwInstKey = Notif.Info.MB.dwObjID;
                        }
                        else
                        {
                            if (AB_DL_DELETED == Notif.Event)
                            {
                                dwInstKey = Notif.Info.DL.dwObjID;
                            }
                        }

                        SPropValue spvInstKey = { 0 };
                        spvInstKey.ulPropTag = PR_INSTANCE_KEY;
                        spvInstKey.Value.bin.cb = sizeof(DWORD);
                        spvInstKey.Value.bin.lpb = (LPBYTE)&dwInstKey;
                        pTable->HrDeleteRow (&spvInstKey);
                    }
                    break;
                case AB_USER_ADDED :
                case AB_USER_MODIFIED :
                case AB_DL_ADDED :
                case AB_DL_MODIFIED :
                    {
                        SPropValue spvProps[CONTENTS_TABLE_PROPS] = { 0 };
                        SRow row = { 0 };
                        row.cValues = CONTENTS_TABLE_PROPS;
                        row.lpProps = spvProps;
                        PRIVATE_ENTRYID eidObject = { 0 };
                        eidObject.uidGlobal = pLogon->GetBackEndUID();
                        eidObject.bVersion = ENTRYID_VERSION;

                        DWORD dwInstKey;
                        for (ULONG i=0; i<CONTENTS_TABLE_PROPS; i++)
                        {
                            spvProps[i].ulPropTag = sptContainerContProps.aulPropTag[i];
                        }

                        TCHAR szEmailAddress[64], szSearchKey[128];
                        LPTSTR pszServer = pLogon->GetServerName();

                        spvProps[CONT_INST_KEY].Value.bin.cb = sizeof(DWORD);
                        spvProps[CONT_INST_KEY].Value.bin.lpb = (LPBYTE)&dwInstKey;
                        spvProps[CONT_ADDRTYPE].Value.lpszA = WINDS_ADDRESS_TYPE;
                        spvProps[CONT_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
                        spvProps[CONT_ENTRYID].Value.bin.lpb = (LPBYTE)&eidObject;
                        spvProps[CONT_RTF].Value.b = TRUE;
                        spvProps[CONT_EMAILADDR].Value.lpszA = szEmailAddress;

                        switch (Notif.Event)
                        {
                            case AB_USER_ADDED :
                            case AB_USER_MODIFIED :
                                spvProps[CONT_OBJTYPE].Value.l = MAPI_MAILUSER;
                                spvProps[CONT_DISPTYPE].Value.l = DT_MAILUSER;
                                spvProps[CONT_ACCOUNT].Value.lpszA = Notif.Info.MB.szMailboxName;
                                wsprintf (szEmailAddress, "%s\\%s", pszServer, Notif.Info.MB.szMailboxName);
                                spvProps[CONT_DISPNAME].Value.lpszA = Notif.Info.MB.szFullName;
                                spvProps[CONT_TITLE].Value.lpszA = Notif.Info.MB.szJobTitle;
                                spvProps[CONT_OFFICE_LOC].Value.lpszA = Notif.Info.MB.szOffice;
                                spvProps[CONT_OFFICE_PHONE].Value.lpszA = Notif.Info.MB.szPhone;
                                spvProps[CONT_OTHER_PHONE].Value.lpszA = Notif.Info.MB.szAltPhone;
                                spvProps[CONT_FAX].Value.lpszA = Notif.Info.MB.szFax;
                                spvProps[CONT_COMMENT].Value.lpszA = Notif.Info.MB.szComments;
                                eidObject.dwObjID = Notif.Info.MB.dwObjID;
                                spvProps[CONT_COMPANY].Value.lpszA = Notif.Info.MB.szCompany;
                                spvProps[CONT_DEPT].Value.lpszA = Notif.Info.MB.szDepartment;
                                spvProps[CONT_MANAGER].Value.lpszA = Notif.Info.MB.szManagerName;
                                break;
                            case AB_DL_ADDED :
                            case AB_DL_MODIFIED :
                                spvProps[CONT_DISPTYPE].Value.l = DT_DISTLIST;
                                spvProps[CONT_OBJTYPE].Value.l = MAPI_DISTLIST;
                                spvProps[CONT_ACCOUNT].Value.lpszA = Notif.Info.DL.szDLAlias;
                                wsprintf (szEmailAddress, "%s\\%s", pszServer, Notif.Info.DL.szDLAlias);
                                spvProps[CONT_DISPNAME].Value.lpszA = Notif.Info.DL.szDLFullName;
                                spvProps[CONT_TITLE].Value.lpszA = szBlank;
                                spvProps[CONT_OFFICE_LOC].Value.lpszA = szBlank;
                                spvProps[CONT_OFFICE_PHONE].Value.lpszA = szBlank;
                                spvProps[CONT_OTHER_PHONE].Value.lpszA = szBlank;
                                spvProps[CONT_FAX].Value.lpszA = szBlank;
                                spvProps[CONT_COMMENT].Value.lpszA = szBlank;
                                eidObject.dwObjID = Notif.Info.DL.dwObjID;
                                spvProps[CONT_COMPANY].Value.lpszA = szBlank;
                                spvProps[CONT_DEPT].Value.lpszA = szBlank;
                                spvProps[CONT_MANAGER].Value.lpszA = szBlank;
                                break;
                        }
                        wsprintf (szSearchKey, "%s:%s", WINDS_ADDRESS_TYPE, szEmailAddress);
                        CharUpper (szSearchKey);
                        spvProps[CONT_SKEY].Value.bin.cb = Cbtszsize(szSearchKey);
                        spvProps[CONT_SKEY].Value.bin.lpb = (LPBYTE)szSearchKey;
                        dwInstKey = eidObject.dwObjID;
                        eidObject.bObject = (BYTE)spvProps[CONT_OBJTYPE].Value.l;
                        pTable->HrModifyRow (&row);
                    }
                    break;
                case SERVER_IS_SHUTTING_DOWN :
                    {
                        DWORD dwFileSize = 0;
                        // Get the name of the local address book directory file.
                        LPTSTR pszLocalABD = pLogon->GetLocalABD();
                        HANDLE hABDFile = CreateFile (pszLocalABD,
                                                      GENERIC_READ,
                                                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                      NULL,
                                                      OPEN_ALWAYS,
                                                      FILE_ATTRIBUTE_NORMAL,
                                                      NULL);
                        if (INVALID_HANDLE_VALUE != hABDFile)
                        {
                            // If the file size is 0, or we failed to get the file size, we
                            // assume that the file is inaccessible or empty and can't be used
                            // to load a local address book directory.
                            dwFileSize = GetFileSize (hABDFile, NULL);
                            if (0xFFFFFFFF == dwFileSize)
                            {
                                dwFileSize = 0;
                            }
                            // However, if the contents table has already been cached, then the address book is still available
                            if (pTable)
                            {
                                dwFileSize = 1;
                            }
                            CloseHandle (hABDFile);
                        }
                        CONNECTION_TYPE Connection = (dwFileSize ? CONNECTION_BROKEN : NOT_AVAIL_NO_ABD);
                        pLogon->SetConnectionType (Connection);
                        ((CABServerBackEnd *)(pLogon->m_pBEobj))->GoOffLine();
                    }
                    break;
                case SERVER_HAS_RESTARTED :
                    pLogon->SetConnectionType (LAN_CONNECT);
                    ((CABServerBackEnd *)(pLogon->m_pBEobj))->GoOnLine();
                    break;
                case RESET_LINKS_WITH_SERVER :
                    {
                        pLogon->ResetServerNotifications();
                        DWORD dwConnectionID = pLogon->GetConnectionID();
                        TCHAR szMailslotName[64];
                        wsprintf (szMailslotName,
                                  CLIENT_MAILSLOT_SINK_NAME_FORMAT,
                                  AB_WINDS_NOTIFICATION_MAILSLOT,
                                  dwConnectionID);
                        hMailslot = CreateMailslot (szMailslotName, 0, MAILSLOT_WAIT_FOREVER, NULL);
                        if (INVALID_HANDLE_VALUE == hMailslot)
                        {
                            TraceResult ("ABMailslotListenThreadProc: Failed to create notification mailslot", HRESULT_FROM_WIN32(GetLastError()));
                        }
                        else
                        {
                            pLogon->SetListenMailslot (hMailslot);
                        }
                    }
                    break;
                case AB_GET_LOCAL_ABDATA_NOW :
                    // The server is instructing us to get the address book data
                    GetServerABD (pLogon->GetServerName(),
                                  pLogon->GetLocalABD(),
                                  NULL);
                    break;
            }
            if (pTable)
            {
                pTable->Release();
            }
        }
        __except (EXCEPTION_EXECUTE_HANDLER)
        {
            continue;
        }
    } 
    TraceMessage ("ABMailslotListenThreadProc: Thread is terminating");
    return S_OK;
}

// End of File for ABPROV.CPP
