///////////////////////////////////////////////////////////////////////////////
//
//  File Name
//      ABSTATUS.CPP 
//
//  Description
//      This file implements the methods for the IMAPIStatus interface.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "ABWDS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceInfoMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

#define NUM_AB_STATUS_PROPS        16
// Properties available in our implementation of IMAPIStatus
SizedSPropTagArray(NUM_AB_STATUS_PROPS, sptABStatusObj) =
{
    NUM_AB_STATUS_PROPS,
    {
        PR_ACCESS,
        PR_ACCESS_LEVEL,

        PR_DISPLAY_NAME,
        PR_PROVIDER_DISPLAY,
        PR_PROVIDER_DLL_NAME,

        PR_RESOURCE_METHODS,
        PR_RESOURCE_FLAGS,
        PR_RESOURCE_TYPE,

        PR_STATUS_CODE,
        PR_STATUS_STRING,

        PR_AB_PROVIDER_ID,

        PR_IDENTITY_ENTRYID,
        PR_IDENTITY_DISPLAY,
        PR_IDENTITY_SEARCH_KEY,

        PR_OBJECT_TYPE,

        PR_CURRENT_VERSION
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::CABMAPIStatus()
//
//    Parameters
//      pLogon          Parent logon session for this IMAPIStatus object
//      pProfSectObj    Profile section object for this provider
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CABMAPIStatus::CABMAPIStatus (CABLogon * pLogon,
                              LPPROFSECT pProfSectObj) : CGenericProp (pLogon->m_pBEobj,
                                                                       NULL,
                                                                       NULL,
                                                                       NULL,
                                                                       FALSE,
                                                                       MAPI_ACCESS_READ)
{
    InfoTrace ("CABMAPIStatus: Constructor called");
    m_cRef   = 1;
    m_pLogon = pLogon;
    m_pLogon->AddRef();
    m_pProfSectObj  = pProfSectObj;
    if (m_pProfSectObj)
    {
        m_pProfSectObj->AddRef();
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::~CABMAPIStatus()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of the object. Release the contained objects and internal
//      resources used in this instance.
//
//    Return Value
//      None
//
CABMAPIStatus::~CABMAPIStatus()
{
    InfoTrace ("CABMAPIStatus: Destructor called");
    if (m_pProfSectObj)
    {
        m_pProfSectObj->Release();
    }
    m_pLogon->Release();
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::QueryInterface()
//
//    Parameters
//      { Refer to OLE Documentation on this method }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it 
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    Validate_IUnknown_QueryInterface (this, riid, ppvObj);
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    // If this is one of the supported IIDs return an interface pointer to it
    if (riid == IID_IMAPIStatus || riid == IID_IMAPIProp || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        AddRef();
    }
    else
    {
        // This object does not support the interface requested
        hResult = E_NOINTERFACE;
    }
    LeaveCriticalSection (&m_csObj);
    return hResult;
}    

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions
//

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::GetLastError()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Return a MAPIERROR structure filled up with meaningful descriptions
//      about an error that occurred in our provider. The caller is responsible
//      for freeing the memory returned.
//      In our implementation we delegate the call up to the base class which
//      implements this call.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::GetLastError (HRESULT         hError,
                                          ULONG           ulFlags,
                                          LPMAPIERROR *   ppMAPIError)
{
    InfoTrace ("CABMAPIStatus::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this, hError, ulFlags, ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_pLogon->m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::SaveChanges()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our implementation is read-only so there are no changes
//      to commit. We return success all the time to be polite to the caller.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CABMAPIStatus::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CABMAPIStatus::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::GetProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Return the value of the properties requested in the list of tags
//      passed in. The returned property array is in the same order as
//      the tag array passed.
//      The memory for returned properties is allocated using the MAPI
//      allocation functions. The caller is responsible for free it with
//      MAPIFreeBuffer() or the equivalent.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::GetProps (LPSPropTagArray pPropTagArray,
                                      ULONG           ulFlags,
                                      ULONG *         pcValues,
                                      LPSPropValue *  ppPropArray)
{
    InfoTrace ("CABMAPIStatus::GetProps method called");
    Validate_IMAPIProp_GetProps (this, pPropTagArray, ulFlags, pcValues, ppPropArray);
    #ifdef DO_INFO_TRACES
    if (!pPropTagArray)
    {
        TraceInfoMessage ("CABMAPIStatus::GetProps: All properties requested");
    }
    else
    {
        TraceMessage ("CABMAPIStatus::GetProps Properties requested...");
        for (ULONG j=0; j<pPropTagArray->cValues; j++) { TraceProp (pPropTagArray->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES
    
    EnterCriticalSection (&m_csObj);
    BOOL fGotErrors = FALSE;
    // Ask the base class to do partial work for this call:
    //      1. Collect all the property tags of the object including the ones in the property cache
    //      2. Allocate a new SPropValue array for the returned properties
    //      3. Replace the tags of PT_UNSPECIFIED type in the tag passed in
    //         with the actual type of the store property (if it exists in this object)
    HRESULT hResult = CGenericProp::GetProps (pPropTagArray,
                                              ulFlags,
                                              pcValues,
                                              ppPropArray,
                                              (LPSPropTagArray)&sptABStatusObj);
    if (hResult)
    {
        goto ErrorExit;
    }
    TCHAR szBuffer[128];
    LPVOID pSrc;
    ULONG cbSize, i, ulStringID;
    // Get the values for all the properties requested
    for (cbSize=0, i=0; i<*pcValues; i++)
    {
        pSrc = NULL;
        hResult = S_OK;
        switch ((*ppPropArray)[i].ulPropTag)
        {
            // Get the PT_OBJECT/PR_NULL properties
            case PR_NULL :
                (*ppPropArray)[i].Value.err = S_OK;
                break;
            // Get the PT_LONG properties
            case PR_ACCESS :
                (*ppPropArray)[i].Value.l = m_lAccess;
                break;
            case PR_ACCESS_LEVEL :
                (*ppPropArray)[i].Value.l = 0;
                break;
            case PR_OBJECT_TYPE :
                (*ppPropArray)[i].Value.l = MAPI_STATUS;
                break;
            case PR_RESOURCE_METHODS :
		        (*ppPropArray)[i].Value.l = STATUS_SETTINGS_DIALOG | STATUS_CHANGE_PASSWORD;
			    break;
            case PR_RESOURCE_FLAGS :
                hResult = E_FAIL;
                if (m_pProfSectObj)
                {
                    LPSPropValue pResFlags;
                    SPropTagArray sptResFlags = { 1, { PR_RESOURCE_FLAGS } };
                    ULONG cValues;
                    hResult = m_pProfSectObj->GetProps (&sptResFlags, fMapiUnicode, &cValues, &pResFlags);
                    if (SUCCEEDED(hResult))
                    {
                        if (PR_RESOURCE_FLAGS == pResFlags->ulPropTag)
                        {
                            (*ppPropArray)[i].Value.l = pResFlags->Value.l;
                        }
                        PrivMAPIFree ((LPVOID *)&pResFlags);
                    }
                }
                if (hResult)
                {
				    (*ppPropArray)[i].Value.l = STATUS_PRIMARY_IDENTITY;
                    hResult = S_OK;
                }
			    break;
            case PR_RESOURCE_TYPE :
				(*ppPropArray)[i].Value.l = MAPI_AB_PROVIDER;
			    break;
            case PR_STATUS_CODE :
                (*ppPropArray)[i].Value.l = (NOT_AVAIL_NO_ABD == m_pLogon->GetConnectionType() ? STATUS_OFFLINE : STATUS_AVAILABLE);
			    break;
            // Get the PT_I8 properties
            case PR_CURRENT_VERSION :
                (*ppPropArray)[i].Value.li.LowPart = ADR_BOOK_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = ADR_BOOK_VERSION_MINOR;
                break;
            // Get the PT_TSTRING properties
            case PR_STATUS_STRING :
                ulStringID = IDS_AB_STATUS_STRING_OFFSET + (ULONG)m_pLogon->GetConnectionType();
                if (0 == LoadString (m_pLogon->m_hInstance, ulStringID, szBuffer, 128))
                {
                    wsprintf (szBuffer, "Status String Not Found - Connection State %d", m_pLogon->GetConnectionType());
                }
                pSrc = szBuffer;
                break;
            case PR_PROVIDER_DISPLAY :
                pSrc = (LPVOID)PROVIDER_DISPLAY_NAME_STRING;
			    break;
            case PR_DISPLAY_NAME :
                wsprintf (szBuffer,
                          "%s (%s)",
                          PROVIDER_DISPLAY_NAME_STRING,
                          m_pLogon->GetServerName());
                pSrc = (LPVOID)szBuffer;
                break;
            case PR_PROVIDER_DLL_NAME :
                pSrc = (LPVOID)AB_DLL_NAME_STRING;
                break;
            case PR_IDENTITY_DISPLAY :
                pSrc = (LPVOID)m_pLogon->GetUserName();
                break;
            // Get the PT_BINARY properties
            case PR_IDENTITY_SEARCH_KEY :
            case PR_IDENTITY_ENTRYID :
                cbSize = CB_PRIVATE_EID;
                pSrc = (LPVOID)m_pLogon->GetUserEID();
                break;
            case PR_AB_PROVIDER_ID :
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&guidABEntries;
                break;
            default :
                hResult = MAPI_E_NOT_FOUND;
                #ifdef DO_INFO_TRACES
                TraceRaw ("CABMAPIStatus::GetProps: Not Found - "); TraceProp ((*ppPropArray)[i].ulPropTag);
                #endif // #DO_INFO_TRACES
                break;
        }
        // Now that I have the value of the property in pSrc or the error in hResult,
        // copy the property or set the property tag and error if not found.
        if (S_OK != CGenericProp::GetProps (hResult,
                                            cbSize,
                                            pSrc,
                                            &(*ppPropArray)[i],
                                            *ppPropArray))
        {
            fGotErrors = TRUE;
        }
    }
ErrorExit :
    LeaveCriticalSection (&m_csObj);
    if (fGotErrors)
	{
        // I got at least one error. Return partial success.
		hResult = MAPI_W_ERRORS_RETURNED;
	}
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::GetPropList()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Duplicate the tag array with ALL the available properties
//      on this object. In our implementation we delegate the call down to
//      the base class.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::GetPropList (ULONG             ulFlags,
                                         LPSPropTagArray * ppAllTags)
{
    InfoTrace ("CABMAPIStatus::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppAllTags);
    return CGenericProp::GetPropList (ulFlags, (LPSPropTagArray)&sptABStatusObj, ppAllTags);
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::OpenProperty()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Place holder. Replace this function with actual implementation 
//      if needed.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::OpenProperty (ULONG       ulPropTag,
                                          LPCIID      piid,
                                          ULONG       ulInterfaceOptions,
                                          ULONG       ulFlags,
                                          LPUNKNOWN * ppUnk)
{
    InfoTrace ("CABMAPIStatus::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this,
                                     ulPropTag, 
                                     piid,
                                     ulInterfaceOptions,
                                     ulFlags,
                                     ppUnk);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::SetProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our IMAPIProp implementation is read-only.
//
//    Return Value
//      E_ACCESSDENIED always.
//
STDMETHODIMP CABMAPIStatus::SetProps (ULONG                 cValues,
                                      LPSPropValue          pPropArray,
                                      LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CABMAPIStatus::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    return E_ACCESSDENIED;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::DeleteProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our IMAPIProp implementation is read-only.
//
//    Return Value
//      E_ACCESSDENIED always.
//
STDMETHODIMP CABMAPIStatus::DeleteProps (LPSPropTagArray       pPropTagArray,
                                         LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CABMAPIStatus::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    return E_ACCESSDENIED;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::CopyTo()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Place holder. Replace this function with actual implementation 
//      if needed.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::CopyTo (ULONG                 ciidExclude,
                                    LPCIID                rgiidExclude,
                                    LPSPropTagArray       pExcludeProps,
                                    ULONG                 ulUIParam,
                                    LPMAPIPROGRESS        pProgress,
                                    LPCIID                pInterface,
                                    LPVOID                pDestObj,
                                    ULONG                 ulFlags,
                                    LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CABMAPIStatus::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,
                               rgiidExclude,
                               pExcludeProps,
                               ulUIParam,
                               pProgress,
                               pInterface,
                               pDestObj,
                               ulFlags,
                               ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::CopyProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Place holder. Replace this function with actual implementation 
//      if needed.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::CopyProps (LPSPropTagArray       pIncludeProps,
                                       ULONG                 ulUIParam,
                                       LPMAPIPROGRESS        pProgress,
                                       LPCIID                pInterface,
                                       LPVOID                pDestObj,
                                       ULONG                 ulFlags,
                                       LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CABMAPIStatus::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::GetNamesFromIDs()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our IMAPIProp implementation does not support named
//      properties.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                             LPGUID            pPropSetGuid,
                                             ULONG             ulFlags,
                                             ULONG *           pcPropNames,
                                             LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CABMAPIStatus::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::GetIDsFromNames()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our IMAPIProp implementation does not support named
//      properties.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::GetIDsFromNames (ULONG             cPropNames,
                                             LPMAPINAMEID *    ppPropNames,
                                             ULONG             ulFlags,
                                             LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CABMAPIStatus::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIStatus virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::ValidateState()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our implementation does not support this call.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::ValidateState (ULONG ulUIParam, ULONG ulFlags)
{   
    InfoTrace ("CABMAPIStatus::ValidateState method called");
    Validate_IMAPIStatus_ValidateState (this, ulUIParam, ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::SettingsDialog()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Called by a client to display the configuration settings of addres
//      book provider. This method could be used to display the current
//      setting in a READ-ONLY mode
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::SettingsDialog (ULONG ulUIParam, ULONG ulFlags)
{
    InfoTrace ("CABMAPIStatus::SettingsDialog method called");
    Validate_IMAPIStatus_SettingsDialog (this, ulUIParam, ulFlags);
    // Open the profile for our provider and get what ever is set there
    LPPROFSECT pProfileObj = NULL;
    ULONG ulPropCount;
    LPSPropValue pProps = NULL;
    CFGDLG CfgDialog = { 0 };
    HRESULT hResult = S_OK;

    PrivInitialize3DCtl (m_pLogon->m_hInstance);
    DWORD dwWait = WaitForSingleObject (m_pLogon->m_hUIMutex, 250); // Wait for the MUTEX for 1/4 second
    if (WAIT_TIMEOUT == dwWait)
    {
        PrivateMessageBox (IDS_MSG_CONFIG_UI_IN_USE, (HWND)ulUIParam);
        PrivUninitialize3DCtl (m_pLogon->m_hInstance);
        return MAPI_E_BUSY;
    }
    hResult = OpenServiceProfileSection (m_pLogon->m_pSupObj, &pProfileObj, GetMAPIFreeBuffer());
    if (hResult)
    {
        TraceResult ("CABMAPIStatus::SettingsDialog: Failed to open the profile section", hResult);
        goto ErrorExit;
    }
    
    hResult = pProfileObj->GetProps ((LPSPropTagArray)&sptLogonProps,
                                     fMapiUnicode,
                                     &ulPropCount,
                                     &pProps);
    if (FAILED(hResult))
    {
        TraceResult ("CABMAPIStatus::SettingsDialog: Failed to get profile section properties", hResult);
        goto ErrorExit;
    }

    // Get the Critical Section
	EnterCriticalSection (&m_pLogon->m_csObj);

	// Fill in the logon UI structure
    CfgDialog.hInst    = m_pLogon->m_hInstance;
	CfgDialog.hWnd     = (HWND)ulUIParam;
	CfgDialog.ppProps  = &pProps;
	CfgDialog.pSupObj  = m_pLogon->m_pSupObj;
	CfgDialog.ulFlags  = ulFlags;
    CfgDialog.hUIMutex = NULL; // So that the DoLogonDlg() won't have to wait.

ShowPropsAgain:
    hResult = DoLogonDlg (&CfgDialog);
    // If the dialog was displayed READ-only, we don't need to verify
    // anything because the user didn't change anything and we are
    // still runnig with the logon setting of the session.
    if (hResult || (ulFlags & UI_READONLY))
    {
        goto ErrorExit;
    }
    if (!PingRemoteServer ((HWND)ulUIParam, pProps))
    {
        goto ShowPropsAgain;
    }

    lstrcpy (m_pLogon->m_szServer, pProps[SERVER_NAME].Value.lpszA);
    lstrcpy (m_pLogon->m_UserInfo.szMailboxName, pProps[MAILBOX_NAME].Value.lpszA);
    lstrcpy (m_pLogon->m_UserInfo.szFullName, pProps[USER_NAME].Value.lpszA);
    lstrcpy (m_pLogon->m_UserInfo.szPassword, pProps[PASSWORD].Value.lpszA);
    m_pLogon->m_UserEID.dwObjID = pProps[MAILBOX_ID].Value.l;
    hResult = pProfileObj->SetProps (NUM_LOGON_PROPERTIES, pProps, NULL);
    TraceResult ("CABMAPIStatus::SetttingsDialog: Failed to save the properties back into the profile", hResult);

	// Set the new/changed identity in the status row of the provider
    hResult = m_pLogon->InitializeStatusRow (STATUSROW_UPDATE);   

ErrorExit:
	// Release the Critical Section.
	LeaveCriticalSection (&m_pLogon->m_csObj);
	if (pProfileObj)
    {
        pProfileObj->Release();
	}
	PrivMAPIFree ((LPVOID *)&pProps);
    PrivUninitialize3DCtl (m_pLogon->m_hInstance);
    ReleaseMutex (m_pLogon->m_hUIMutex);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::ChangePassword()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Changes the password for the mailbox the service is logged in.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABMAPIStatus::ChangePassword (LPTSTR pszOldPass,
                                            LPTSTR pszNewPass,
                                            ULONG  ulFlags)
{
    InfoTrace ("CABMAPIStatus::ChangePassword method called");
    Validate_IMAPIStatus_ChangePassword (this, pszOldPass, pszNewPass, ulFlags);
    HRESULT hResult = ChangeMBServerPassword (m_pLogon->m_szServer,
                                              m_pLogon->m_UserInfo.szMailboxName,
                                              pszOldPass,
                                              pszNewPass);
    if (hResult)
    {
        TraceResult ("CABMAPIStatus::ChangePassword", hResult);
        if (HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORD) == hResult)
        {
            hResult = E_ACCESSDENIED;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABMAPIStatus::FlushQueues()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method. Our implementation does not support this call.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CABMAPIStatus::FlushQueues (ULONG     ulUIParam,
                                         ULONG     cbTargetTransport,
                                         LPENTRYID pTargetTransport,
                                         ULONG     ulFlags)
{
    InfoTrace ("CABMAPIStatus::FlushQueues method called");
    Validate_IMAPIStatus_FlushQueues (this,
                                      ulUIParam,
                                      cbTargetTransport,
                                      pTargetTransport,
                                      ulFlags);
    return MAPI_E_NO_SUPPORT;
}

// End of File for ABSTATUS.CPP