///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      CFILESTG.CPP
//
//  Description
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.0
//
// Written for Microsoft Corporation
// Copyright (c) 1995 Microsoft Corporation. All rights reserved.
//
#include "MSLMS.H"
#include "CFILESTG.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceMessage(a)
#define InfoTraceRaw(a)     TraceRaw(a)
#define InfoTraceProp(a)    TraceProp(a)
#else
#define InfoTrace(a)        0
#define InfoTraceRaw(a)     0
#define InfoTraceProp(a)    0
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    OpenMsgStoreDataBase()
//
//      [IN]    pszFilePath         Pointer to a string with a fully qualified
//                                  file name and directory where the database
//                                  file will be created or opened.
//      [IN]    pszStoreName        Pointer to a string the caller gave us
//                                  with the name of the message store. This
//                                  can't be a NULL pointer, but it can be
//                                  and empty string.   
//      [IN]    pUID                Pointer to a MAPIUID allocated by the
//                                  caller where, if the file was created, the
//                                  caller gave us the UID to stamp the
//                                  database with. If the database was
//                                  just opened, we will return the save UID.
//      [IN]    pSupObj             Pointer to a support object given by
//                                  MAPI to the caller.
//
//    Purpose
//      Creates the database file to be used in a Message Store. The caller
//      must check if the file indicated by pszFilePath, exists or not. If
//      the file does not exist, a new UID must be generated (i.e.
//      IMAPISupport::NewUID) and give to us in pUID. If the file exists,
//      the caller just provides a MAPIUID structure (uninitialized) where
//      we return the UID stored in the database file.
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI OpenMsgStoreDataBase (LPMAPISUP      pSupObj,
                                     LPTSTR         pszFilePath,
                                     LPTSTR         pszStoreName,
                                     LPTSTR         pszStoreComment,
                                     MAPIUID *      pUID,
                                     BOOL           fCreateIfNotFound,
                                     CBackEnd **    ppBackEnd)
{
    CDataBase * pDBobj;
    BOOL bWasFileCreated = FALSE;
    HRESULT hResult = OpenDataBaseFile (pszFilePath, pszStoreName, pszStoreComment, *pUID, &pDBobj);
    if (hResult)
    {
        if (fCreateIfNotFound && HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND) == hResult)
        {
            bWasFileCreated = TRUE;
            hResult = CreateDataBaseFile (pszFilePath, pszStoreName, pszStoreComment, *pUID, &pDBobj);
        }
    }
    if (hResult)
    {
        return hResult;
    }
    // Allocate the new database object. The constructor initializes internal data in the object.
    CFileBackEnd * pFBEobj = new CFileBackEnd (pDBobj, pUID, pSupObj);
    pDBobj->Release(); // The constructor in CFileBackEnd did an AddRef() on this
    if (pFBEobj)
    {
        if (bWasFileCreated)
        {
            PRIVATE_ENTRYID InboxEID = { 0 };
            // Set the default receive folder for the IPM message class. The entry ID
            // used here is the entry ID of the INBOX which at this point it has not been created
            // (this is done is CFileBackEnd::CreateSpecialFolders)
            // This entry ID is fixed.
            InboxEID.uidGlobal = *pUID;
            InboxEID.bVersion  = ENTRYID_VERSION;
            InboxEID.bObject   = MAPI_FOLDER;
            InboxEID.dwObjID   = IPM_INBOX_ID;
            hResult = pFBEobj->SetRFSEntryID (TEXT("IPM"), InboxEID, NULL);
            if (!hResult)
            {
                hResult = pFBEobj->CreateSpecialFolders();
            }
        }
    }
    else
    {
        hResult = E_OUTOFMEMORY;
    }
    if (hResult)
    {
        if (bWasFileCreated)
        {
            // On failure, if the file was created, we don't want to leave traces of it, remove it.
            DeleteFile (pszFilePath);
        }
        TraceResult ("OpenMsgStoreDataBase", hResult);
    }
    else
    {
        *ppBackEnd = pFBEobj;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::Insert()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CIDList::Insert (DWORD dwObjID)
{
    POBJECT_ID pNewNode;
    HRESULT hResult = PrivAllocate (sizeof(OBJECT_ID), (LPVOID *)&pNewNode, "CIDList Object List Node");
    if (hResult)
    {
        TraceResult ("CIDList::Insert", hResult);
        return hResult;
    }
    pNewNode->dwObjID = dwObjID;
    InsertInList (&m_pList, pNewNode);
    m_dwNodeCount++;
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::Delete()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CIDList::Delete (DWORD dwObjID)
{
    HRESULT hResult = MAPI_E_NOT_FOUND;
    POBJECT_ID pPrev, pNode = m_pList;
    while (pNode)
    {
        if (dwObjID == pNode->dwObjID)
        {
            if (pNode == m_pList)
            {
                // If the node we are deleting is the head of the list, reset the list head pointer.
                m_pList = m_pList->pNext;
            }
            else
            {
                // Take this node out from the list, and make the previous node
                // point to the one in front of the current node.
                pPrev->pNext = pNode->pNext;
            }
            PrivFree ((LPVOID *)&pNode);
            hResult = S_OK;
            m_dwNodeCount--;
            break; // Out of the WHILE() loop
        }
        else
        {
            // Don't delete this node, save is as the "previous" node in the linear list
            pPrev = pNode;
            pNode = pNode->pNext;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::SaveToStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CIDList::SaveToStream (LPSTREAM pStream)
{
    if (!pStream) { return S_OK; }
    LARGE_INTEGER liZero = { 0 };
    pStream->Seek (liZero, STREAM_SEEK_SET, NULL);

    DWORD dwIDCount = m_dwNodeCount, dwWritten = 0;
    POBJECT_ID pNode = m_pList;
    ULARGE_INTEGER uliSize = { 0 };
    HRESULT hResult = pStream->Write (&dwIDCount, sizeof(DWORD), NULL);
    while (pNode)
    {
        hResult = pStream->Write (&(pNode->dwObjID), sizeof(DWORD), NULL);
        TraceResult ("CIDList::SaveToStream: Failed to write ID", hResult);
        if (!hResult)
        {
            dwWritten++;
        }
        pNode = pNode->pNext;
    }
    // A desperate attempt to write the actual written number of IDs
    if (dwWritten != dwIDCount)
    {
        pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
        pStream->Write (&dwWritten, sizeof(DWORD), NULL);
    }
    uliSize.LowPart = sizeof(DWORD) * (dwWritten + 1); // +1 for the size of the ID counter (dwIDCount)
    pStream->SetSize (uliSize);
    TraceResult ("CIDList::SaveToStream", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::LoadFromStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CIDList::LoadFromStream (LPSTREAM pStream)
{
    if (!pStream) { return S_OK; }
    DWORD dwIDCount = 0, dwObjID;
    HRESULT hResult = pStream->Read (&dwIDCount, sizeof(DWORD), NULL);
    if (!hResult && dwIDCount)
    {
        while (dwIDCount && !hResult)
        {
            hResult = pStream->Read (&dwObjID, sizeof(DWORD), NULL);
            if (!hResult)
            {
                hResult = Insert (dwObjID);
                dwIDCount--;
            }
        }
    }
    TraceResult ("CIDList::LoadFromStream", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::IsInList()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI CIDList::IsInList (DWORD dwObjID)
{
    BOOL bFound = FALSE;
    POBJECT_ID pNode = m_pList;
    while (pNode)
    {
        if (dwObjID == pNode->dwObjID)
        {
            bFound = TRUE;
            break; // out of the WHILE() loop
        }
        pNode = pNode->pNext;
    }
    return bFound;
}

///////////////////////////////////////////////////////////////////////////////
//    CIDList::Walk()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
DWORD WINAPI CIDList::Walk (WALKING_ACTION waAction)
{
    if (m_pList)
    {
        if (WALK_START == waAction)
        {
            m_pWalkNext = m_pList->pNext;
            return m_pList->dwObjID;
        }
        if (WALK_CONTINUE == waAction && m_pWalkNext)
        {
            DWORD dwObjID = m_pWalkNext->dwObjID;
            m_pWalkNext = m_pWalkNext->pNext;
            return dwObjID;
        }
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CFileBackEnd()
//
//    Parameters
//      
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CFileBackEnd::CFileBackEnd (CDataBase * pDBobj,
                            MAPIUID *   pUIDinFile,
                            LPMAPISUP   pSupObj) : CBackEnd (*pUIDinFile)
{
    InfoTrace ("CFileBackEnd: Constructor called");
    m_cRef = 1;
    m_pDBobj = pDBobj;
    m_pDBobj->AddRef();
    // Don't AddRef() this guy here, the container of this object has done this.
    m_pSupObj = pSupObj;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::~CFileBackEnd()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CFileBackEnd.
//
//    Return Value
//      None
//
CFileBackEnd::~CFileBackEnd ()
{
    InfoTrace ("CFileBackEnd: Destructor called");
    m_pDBobj->Release();
    m_pDBobj = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::GetStream (DWORD             dwRecordID,
                                      RECORD_FIELDS     rfStream,
                                      LPSTREAM *        ppStream,
                                      BOOL              fCreateIfNotFound,
                                      BOOL              fWriteZeroCount)
{
    ASSERT (dwRecordID);
    ASSERT (ppStream);
    *ppStream = NULL;
    DWORD dwZero = 0;
    LARGE_INTEGER liOffSet = { 0 };
    HRESULT hResult = m_pDBobj->OpenRecordField (dwRecordID, FALSE, rfStream, ppStream);
    if (MAPI_E_NOT_FOUND == hResult && fCreateIfNotFound)
    {
        hResult = m_pDBobj->OpenRecordField (dwRecordID, TRUE, rfStream, ppStream);
        if (!hResult && fWriteZeroCount)
        {
            hResult = (*ppStream)->Write (&dwZero, sizeof(DWORD), NULL);
            if (hResult)
            {
                m_pDBobj->DeleteRecord (dwRecordID);
            }
            else
            {
                (*ppStream)->Seek (liOffSet, STREAM_SEEK_SET, NULL);
            }
        }
    }
#ifdef _DEBUG
    if (MAPI_E_NOT_FOUND != hResult)
    {
        TraceResult ("CFileBackEnd::GetStream", hResult);
    }
#endif // _DEBUG
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::SaveObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::SaveObject (PRIVATE_ENTRYID ParentEID,
                                       PRIVATE_ENTRYID ObjectEID,
                                       CPropList *     pList,
                                       COpenPropList * pOpenList,
                                       LPTABLEDATA     pTable)
{
    LockDataBase();
    SPropValue spvSize = { 0 }, spvTime = { 0 };
    LPSTREAM pStream;
    STATSTG stat;
    HRESULT hResult = GetStream (ObjectEID.dwObjID, RECFLD_PROPERTIES, &pStream);
    if (!hResult)
    {
        if (pTable)
        {
            ASSERT (MAPI_MESSAGE == ObjectEID.bObject);
            hResult = SaveRecipTable (ObjectEID.dwObjID, pTable);
        }
        if (!hResult && pOpenList)
        {
            hResult = SaveLargeProps (pOpenList, pList);
        }
        if (!hResult)
        {
            hResult = pList->SaveToStream (pStream);
            // If the object we are saving is not a folder, we most calculate PR_xxx_SIZE
            if (!hResult && (MAPI_FOLDER != ObjectEID.bObject))
            {
                if (MAPI_MESSAGE == ObjectEID.bObject)
                {
                    spvSize.ulPropTag = PR_MESSAGE_SIZE;
                }
                else
                {
                    ASSERT (MAPI_ATTACH == ObjectEID.bObject);
                    spvSize.ulPropTag = PR_ATTACH_SIZE;
                }
                CalculateSize (ObjectEID, pList, spvSize.Value.l);
                pStream->Stat (&stat, STATFLAG_NONAME);
                spvSize.Value.l += stat.cbSize.LowPart;
                pList->Insert (&spvSize);
                pList->SaveToStream (pStream);
            }
            if (!hResult && (MAPI_ATTACH == ObjectEID.bObject))
            {
                UpdateAttachmentsCount (ParentEID, ObjectEID, FALSE);
            }
        }
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::SaveObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CalculateSize()
//
//    Parameters
//
//    Purpose
//      This function is RECURSIVE
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::CalculateSize (PRIVATE_ENTRYID   ObjectEID,
                                          CPropList *       pList,
                                          long &            lSize)
{
    STATSTG stat;
    PRIVATE_ENTRYID ItemEID = { 0 };
    LPSTREAM pIDsStream, pStream;
    PFlatSPropValue pProp = pList->Walk (WALK_START);
    while (pProp)
    {
        if (PT_STREAM_STORAGE == PROP_TYPE (pProp->ulPropTag))
        {
            if (!GetStream (pProp->Value.li.HighPart, RECFLD_RAWDATA, &pStream, FALSE))
            {
                pStream->Stat (&stat, STATFLAG_NONAME);
                pStream->Release();
                lSize += stat.cbSize.LowPart;
            }
        }
        pProp = pList->Walk (WALK_CONTINUE);
    }
    if (MAPI_MESSAGE == ObjectEID.bObject)
    {
        if (!GetStream (ObjectEID.dwObjID, RECFLD_RECIPIENTS, &pIDsStream, FALSE))
        {
            CIDList IDList (pIDsStream);
            pIDsStream->Release();
            ItemEID.bObject = MAPI_MAILUSER;
            ItemEID.dwObjID = IDList.Walk (WALK_START);
            while (ItemEID.dwObjID)
            {
                if (!GetStream (ItemEID.dwObjID, RECFLD_RECIPIENTS, &pStream, FALSE))
                {
                    CPropList ItemProps (pStream);
                    pStream->Release();
                    CalculateSize (ItemEID, &ItemProps, lSize);
                }
                ItemEID.dwObjID = IDList.Walk (WALK_CONTINUE);
            }
        }
        if (!GetStream (ObjectEID.dwObjID, RECFLD_ATTACHMENTS, &pIDsStream, FALSE))
        {
            CIDList IDList (pIDsStream);
            pIDsStream->Release();
            ItemEID.bObject = MAPI_ATTACH;
            ItemEID.dwObjID = IDList.Walk (WALK_START);
            while (ItemEID.dwObjID)
            {
                if (!GetStream (ItemEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE))
                {
                    CPropList ItemProps (pStream);
                    pStream->Release();
                    CalculateSize (ItemEID, &ItemProps, lSize);
                }
                ItemEID.dwObjID = IDList.Walk (WALK_CONTINUE);
            }
        }
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::SaveLargeProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::SaveLargeProps (COpenPropList * pOpenList,
                                           CPropList *     pList)
{
    SPropValue spvProp = { 0 };
    LPSTREAM pStream;
    BOOL fRecordCreated;
    HRESULT hResult = S_OK;
    POPEN_PROPS pNode = pOpenList->Walk (WALK_START);
    while (pNode && !hResult)
    {
        // If the property interface was opened read-only, don't
        // bother to save it
        if (PROP_READ_ONLY & pNode->ulFlags)
        {
            goto NextNode;
        }
        ASSERT (PROP_READ_WRITE & pNode->ulFlags);
        if (PROP_DELETE_THIS & pNode->ulFlags)
        {
            pList->Delete (pNode->ulPropTag);
            if (!(PROP_NEVER_SAVED & pNode->ulFlags))
            {
                m_pDBobj->DeleteRecord (pNode->dwID);
            }
            goto NextNode;
        }
        if (PROP_NEVER_SAVED & pNode->ulFlags)
        {
            fRecordCreated = TRUE;
            ASSERT (0 == pNode->dwID);
            hResult = m_pDBobj->CreateRecord (&pNode->dwID, &pStream);
        }
        else
        {
            fRecordCreated = FALSE;
            ASSERT (pNode->dwID);
            hResult = m_pDBobj->OpenRecordField (pNode->dwID, FALSE, RECFLD_PROPERTIES, &pStream);
        }
        if (!hResult)
        {
            ASSERT (PROP_STREAM & pNode->ulFlags || PROP_STORAGE & pNode->ulFlags);                    
            if (PROP_STREAM & pNode->ulFlags)
            {
                if (pNode->iface.pStream->IsDirty())
                {
                    hResult = pNode->iface.pStream->SaveToStream (pStream);
                    if (!hResult)
                    {
                        pNode->iface.pStream->SetDirty (FALSE);
                    }
                }
            }
            if (PROP_STORAGE & pNode->ulFlags)
            {
                if (pNode->iface.pStorage->IsDirty())
                {
                    ULARGE_INTEGER uliSize = { 0 };
                    LARGE_INTEGER liZero = { 0 }, liOldPos = { 0 };
                    uliSize.LowPart = 0xFFFFFFFF;
                    LPSTREAM pFileStream = pNode->iface.pStorage->GetRidingStream();
                    ASSERTMSG (pFileStream, "Where is the stream for this storage?");

                    // Save the current position in the stream
                    pFileStream->Seek (liZero, STREAM_SEEK_CUR, (ULARGE_INTEGER *)&liOldPos);
                    // Move the stream index pointer to the begining of the stream
                    pFileStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                    // Dump the stream's content onto the database stream
                    hResult = pFileStream->CopyTo (pStream, uliSize, NULL, NULL);
                    if (!hResult)
                    {
                        pNode->iface.pStorage->SetDirty (FALSE);
                    }
                    pFileStream->Seek (liOldPos, STREAM_SEEK_SET, NULL);
                }
            }
            pStream->Release();
            if (hResult)
            {
                if (fRecordCreated)
                {
                    m_pDBobj->DeleteRecord (pNode->dwID);
                }
            }
            else
            {
                // At this point the property has been save to the file, so we can remove
                // the bit in the mask. From here on we ignore any errors
                pNode->ulFlags &= ~PROP_NEVER_SAVED;
                pList->InsertOpenProp (pNode->ulPropTag, pNode->dwID, (WORD)(0xF & pNode->ulFlags));
            }
        }
    NextNode:
        pNode = pOpenList->Walk (WALK_CONTINUE);
    }
    TraceResult ("CFileBackEnd::SaveLargeProps", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteObjects()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteObjects (PRIVATE_ENTRYID   ContainerEID,
                                          ULONG             ulDeleteFlags,
                                          LPMAPIPROGRESS    pProgress,
                                          POPENOBJECTS_DATA pData)
{
    HRESULT hResult = S_OK;
    LockDataBase();
    if (CONTAINER_OBJECTS & ulDeleteFlags || EMPTY_CONTAINER & ulDeleteFlags)
    {
        __try
        {
            hResult = DeleteFolderContents (ContainerEID, pProgress, pData);
        }
        __except (EXCEPTION_EXECUTE_HANDLER)
        {
            TraceString1 ("CFileBackEnd::DeleteObjects: An exception occured. Exception code: %#08x", GetExceptionCode());
            hResult = HRESULT_FROM_WIN32(ERROR_EXCEPTION_IN_SERVICE);
        }
    }
    if (ATTACHMENT_OBJECT & ulDeleteFlags)
    {
        hResult = DeleteMessageAttachment (ContainerEID.dwObjID, (DWORD)pProgress);
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::DeleteObjects", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteMessageAttachment()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteMessageAttachment (DWORD dwMsgObjID, DWORD dwAttachmentID)
{
    CIDList AttachIDList;
    LPSTREAM pStream;
    HRESULT hResult = GetStream (dwMsgObjID, RECFLD_ATTACHMENTS, &pStream, FALSE);
    if (!hResult)
    {
        AttachIDList.LoadFromStream (pStream);
        hResult = m_pDBobj->DeleteRecord (dwAttachmentID);
        if (!hResult)
        {
            hResult = AttachIDList.Delete (dwAttachmentID);
            if (!hResult)
            {
                hResult = AttachIDList.SaveToStream (pStream);
            }
        }
        pStream->Release();
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::LoadProperties()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::LoadProperties (LPSTREAM pStream, CPropList ** ppList)
{
    HRESULT hResult;
    *ppList = new CPropList;
    if (*ppList)
    {
        hResult = (*ppList)->LoadFromStream (pStream);
        if (hResult)
        {
            delete *ppList;
            *ppList = NULL;
        }
    }
    else
    {
        hResult = E_OUTOFMEMORY;
    }
    TraceResult ("CFileBackEnd::LoadProperties", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::OpenPropertyInMemory()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::OpenPropertyInMemory (PRIVATE_ENTRYID  ObjectEID,
                                                 ULONG            ulMaxBytes,
                                                 LPVOID           pMemBlock,
                                                 LPSPropValue     pProp)
{
    LockDataBase();
    LPSTREAM pStream;
    STATSTG stat;
    LPVOID pSavedData;
    DWORD dwAllocSize;
    HRESULT hResult = GetStream (pProp->Value.li.HighPart, RECFLD_RAWDATA, &pStream, FALSE);
    if (!hResult)
    {
        pStream->Stat (&stat, STATFLAG_NONAME);
        if (stat.cbSize.HighPart || (stat.cbSize.LowPart > ulMaxBytes))
        {
            hResult = E_OUTOFMEMORY;
        }
        else
        {
            dwAllocSize = stat.cbSize.LowPart;
            if (PROP_TYPE(pProp->ulPropTag) == PT_TSTRING)
            {
                dwAllocSize += sizeof(TCHAR); // Add space for NULL terminator
            }
            hResult = PrivMAPIAllocate (dwAllocSize, &pSavedData, pMemBlock); // Alloc more
            if (!hResult)
            {
                hResult = pStream->Read (pSavedData, stat.cbSize.LowPart, NULL);
                if (!hResult)
                {
                    if (PROP_TYPE(pProp->ulPropTag) == PT_TSTRING)
                    {
                        pProp->Value.LPSZ = (LPTSTR)pSavedData;
                        pProp->Value.LPSZ[stat.cbSize.LowPart] = 0;
                    }
                    else
                    {
                        pProp->Value.bin.cb = stat.cbSize.LowPart;
                        pProp->Value.bin.lpb = (LPBYTE)pSavedData;
                    }
                }
            }
        }
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::OpenPropertyInMemory", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::OpenProperty()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::OpenProperty (PRIVATE_ENTRYID   ObjectEID,
                                         LPSPropValue      pProp,
                                         LPUNKNOWN *       ppUnk)
{
    LockDataBase();
    TCHAR szFileName[MAX_PATH];
    LARGE_INTEGER liZero = { 0 };
    LPSTREAM pStream, pNewStream;
    STATSTG stat;
    BOOL fDeleteUserUnk;
    HRESULT hResult = GetStream (pProp->Value.li.HighPart, RECFLD_RAWDATA, &pStream, FALSE);
    if (!hResult)
    {
        if (PROP_STREAM != HIWORD(pProp->Value.li.LowPart) &&
            PROP_STORAGE != HIWORD(pProp->Value.li.LowPart))
        {
            ASSERTNOW();
            hResult = E_NOINTERFACE;
        }
        if (PROP_STREAM == HIWORD(pProp->Value.li.LowPart))
        {
            pStream->Stat (&stat, STATFLAG_NONAME);
            ASSERT (0 == stat.cbSize.HighPart);
            // If the stream is bigger that 64K, then we will copy the stream to a temp file.
            if (stat.cbSize.LowPart > 65536)
            {
                fDeleteUserUnk = TRUE;
                GetTempPath (MAX_PATH, szFileName);
                hResult = OpenStreamOnFile (GetMAPIAllocBuffer(),
                                            GetMAPIFreeBuffer(),
                                            SOF_UNIQUEFILENAME | STGM_READWRITE | STGM_CREATE | STGM_DELETEONRELEASE,
                    		                szFileName,
                    		                NULL,
                    		                &pNewStream);
            }
            else
            {
                // The caller already allocated an in-memory stream where we can copy our stuff
                fDeleteUserUnk = FALSE;
                pNewStream = (LPSTREAM)(*ppUnk);
                pNewStream->AddRef();
            }
            if (!hResult)
            {
                hResult = pStream->CopyTo (pNewStream, stat.cbSize, NULL, NULL);
                if (!hResult)
                {
                    hResult = pNewStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                }
                if (hResult)
                {
                    pNewStream->Release();
                }
                else
                {
                    // If we allocated a new stream, we must release the one passed in.
                    if (fDeleteUserUnk)
                    {
                        (*ppUnk)->Release();
                    }
                    else
                    {
                        pNewStream->Release(); // To compesate the AddRef() abouve
                    }
                    *ppUnk= (LPUNKNOWN)pNewStream;
                }
            }
        }
        if (PROP_STORAGE == HIWORD(pProp->Value.li.LowPart))
        {
            ULARGE_INTEGER uliSize = { 0 };
            LARGE_INTEGER liZero = { 0 };
            uliSize.LowPart = 0xFFFFFFFF;
            char szTmpFile[512];
            LPSTREAM pFileStream;
            CStorage * pStg = (CStorage *)(*ppUnk);
            BOOL fWriteOk = pStg->IsWriteable();
            pStg->Release();
            *ppUnk = NULL;
            if (GetTempFullPathFileName (szTmpFile))
            {
                hResult = OpenStreamOnFile (GetMAPIAllocBuffer(),
                                            GetMAPIFreeBuffer(),
        		                            STGM_CREATE | STGM_READWRITE | STGM_DELETEONRELEASE,
        		                            szTmpFile,
        		                            NULL,
        		                            &pFileStream);
                if (!hResult)
                {
                    pStream->CopyTo (pFileStream, uliSize, NULL, NULL);
                    pFileStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                    hResult = CreatePrivateIStorage (&pStg, fWriteOk, pFileStream);
                    *ppUnk = pStg;
                    pFileStream->Release();
                }
            }
            else
            {
                hResult = E_FAIL;
            }
        }
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::OpenProperty", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::GetTable (const TABLE_TYPE  tblType,
                                     PRIVATE_ENTRYID   ContainerEID,
                                     long              lAccess,
                                     LPVOID            pData1,
                                     LPVOID            pData2,
                                     LPTABLEDATA *     ppTable)
{
    LPSPropTagArray pDefaultColumnSet;
    ULONG ulIndexProp = PR_INSTANCE_KEY;
    HRESULT hResult;
    if (!IsValidEntryID (&ContainerEID))
    {
        hResult = MAPI_E_INVALID_ENTRYID;
        goto ErrorExit;
    }
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            pDefaultColumnSet = (LPSPropTagArray)&sptFolderHierProps;
            break;
        case TABLE_CONTENTS :
            pDefaultColumnSet = (LPSPropTagArray)&sptFolderContProps;
            break;
        case TABLE_ATTACHMENTS :
            pDefaultColumnSet = (LPSPropTagArray)&sptAttTableProps;
            break;
        case TABLE_RECIPIENTS :
            pDefaultColumnSet = (LPSPropTagArray)&sptRecipTableProps;
            ulIndexProp = PR_ROWID;
            break;
        case TABLE_RECEIVE_FOLDERS :
            pDefaultColumnSet = (LPSPropTagArray)&sptRFSprops;
            break;
        default :
            hResult = MAPI_E_NO_SUPPORT;
            goto ErrorExit;
    }
    
    // Before locking the database, create a new In-Memory table
    hResult = CreateTable (&IID_IMAPITableData,
                           GetMAPIAllocBuffer(),
                           GetMAPIAllocMore(),
                           GetMAPIFreeBuffer(),
                           NULL,
                           TBLTYPE_DYNAMIC,
                           ulIndexProp,
                           pDefaultColumnSet,
                           ppTable);
    if (hResult)
    {
        goto ErrorExit;
    }

    LockDataBase();
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            hResult = FillHierarchyTable (ContainerEID,
                                          *ppTable,
                                          (LPSBinary)pData1,
                                          lAccess);
            break;
        case TABLE_CONTENTS :
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = FillContentsTable (ContainerEID,
                                         *ppTable,
                                         (LPSBinary)pData1,
                                         lAccess,
                                         (BOOL)pData2);
            break;
        case TABLE_ATTACHMENTS :
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = FillAttachmentsTable (ContainerEID, *ppTable, (LPSBinary)pData1, lAccess);
            break;
        case TABLE_RECIPIENTS :
            if (FALSE == (BOOL)pData2)
            {
                hResult = FillRecipientTable (ContainerEID, *ppTable, (DWORD *)pData1);
            }
            break;
        case TABLE_RECEIVE_FOLDERS :
            hResult = GetRFSTableData (*ppTable);
            break;
    }
    if (hResult)
    {
        (*ppTable)->Release();
        *ppTable = NULL;
    }
    UnlockDataBase();
ErrorExit:
    TraceResult ("CFileBackEnd::GetTable", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetTableProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::GetTableProps (const TABLE_TYPE     tblType,
                                          long                 lAccess,
                                          PRIVATE_ENTRYID *    pParentEID,
                                          PRIVATE_ENTRYID *    pObjectEID,
                                          CPropList *          pList,
                                          LPSPropValue *       ppProps,
                                          DWORD &              dwLastAllocated,
                                          LPTABLEDATA          pTable,
                                          LPVOID               pData1,
                                          LPVOID               pData2)
{
    HRESULT hResult;
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            hResult = GetHierarchyTableProps (pList,
                                              ppProps,
                                              *pObjectEID,
                                              *pParentEID,
                                              lAccess,
                                              (LPSBinary)pData1,
                                              (long)pData2,
                                              dwLastAllocated,
                                              pTable,
                                              m_uidBackEnd);
            break;
        case TABLE_CONTENTS :
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = GetContentsTableProps (pList,
                                             ppProps,
                                             *pObjectEID,
                                             lAccess,
                                             (LPSBinary)pData1,
                                             dwLastAllocated,
                                             pTable);
            break;
        case TABLE_ATTACHMENTS :
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = GetAttachmentTableProps (pList,
                                               ppProps,
                                               pObjectEID->dwObjID,
                                               lAccess,
                                               (LPSBinary)pData1,
                                               dwLastAllocated,
                                               pTable);
            break;
        default :
            hResult = MAPI_E_NO_SUPPORT;
    }
    TraceResult ("CFileBackEnd::GetTableProps", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::SaveRecipTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::SaveRecipTable (DWORD dwRecordID, LPTABLEDATA pRecipTable)
{
    ULONG i, j;
    LPSTREAM pStream, pRecipsStream = NULL;
    LPSRowSet pRows = NULL;
    LPMAPITABLE pView = NULL;
    LPSPropTagArray pTags = NULL;
    LPSPropValue pProps;
    CIDList RecipIDList;
    DWORD dwRecipRecID = 0;
    BOOL fReplaced = FALSE, fContinue = TRUE;
    HRESULT hResult = GetStream (dwRecordID, RECFLD_RECIPIENTS, &pRecipsStream, TRUE, TRUE);
    if (!hResult)
    {
        RecipIDList.LoadFromStream (pRecipsStream);
        hResult = pRecipTable->HrGetView (NULL, NULL, 0, &pView);
        if (!hResult)
        {
            hResult = pView->QueryColumns (TBL_ALL_COLUMNS, &pTags);
            if (!hResult)
            {
                #ifdef DO_INFO_TRACES
                TraceMessage ("SaveRecipTable:: Properties available in recip table...");
                for (ULONG j=0; j<pTags->cValues; j++)
                {
                    TraceProp (pTags->aulPropTag[j]);
                }
                #endif // DO_INFO_TRACES
                hResult = pView->SetColumns (pTags, 0);
            }
        }
    }
    if (hResult)
    {   
        goto ErrorExit;
    }
    dwRecipRecID = RecipIDList.Walk (WALK_START);
    while (fContinue && !hResult)
    {
        hResult = pView->QueryRows (15, 0, &pRows);
        if (!hResult)
        {
            if (pRows->cRows)
            {
                for (i=0; i<pRows->cRows; i++)
                {
                    if (dwRecipRecID)
                    {
                        fReplaced = TRUE;
                        hResult = m_pDBobj->OpenRecordField (dwRecipRecID, FALSE, RECFLD_PROPERTIES, &pStream);
                    }
                    else
                    {
                        fReplaced = FALSE;
                        hResult = m_pDBobj->CreateRecord (&dwRecipRecID, &pStream);
                    }
                    if (!hResult)
                    {
                        // This object needs to be defined here so that for every iteration
                        // of the outer FOR() loop, the destructor gets called and the
                        // internal data gets released.
                        CPropList List;
                        pProps = pRows->aRow[i].lpProps;
                        for (j=0; j<pRows->aRow[i].cValues; j++)
                        {
                            if (pProps[j].ulPropTag == PR_ROWID || pProps[j].ulPropTag == PR_INSTANCE_KEY)
                            {
                                continue;
                            }
                            List.Insert (&pProps[j]);
                        }
                        List.SaveToStream (pStream);
                        pStream->Release();
                        if (fReplaced)
                        {
                            dwRecipRecID = RecipIDList.Walk (WALK_CONTINUE);
                        }
                        else
                        {
                            RecipIDList.Insert (dwRecipRecID);
                        }
                    }
                }
            }
            else
            {
                fContinue = FALSE;
            }
            FreeProws (pRows);
        }
    }
    hResult = S_OK;
    while (fReplaced && dwRecipRecID)
    {
        RecipIDList.Delete (dwRecipRecID);
        dwRecipRecID = RecipIDList.Walk (WALK_CONTINUE);
    }
    RecipIDList.SaveToStream (pRecipsStream);
ErrorExit:
    if (pTags)
    {
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    if (pView)
    {
        pView->Release();
    }
    if (pRecipsStream)
    {
        pRecipsStream->Release();
    }
    TraceResult ("CFileBackEnd::SaveRecipTable", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::FillRecipientTable()
//
//    Parameters
//
//    Purpose
//      This function loads the table passed, with the properties found in
//      the attachments of the message indicated by the entry ID passed in.
//      If the message doesn't have any attachments we return 0 for the
//      attachment count.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::FillRecipientTable (PRIVATE_ENTRYID MessageEID,
                                               LPTABLEDATA     pTable,
                                               DWORD *         pdwRecipCount)
{
    *pdwRecipCount = 0;
    SRow srRow = { 0 };
    SPropValue spvIndexes[2] = { 0 };
    spvIndexes[0].ulPropTag = PR_ROWID;
    spvIndexes[1].ulPropTag = PR_INSTANCE_KEY;
    spvIndexes[1].Value.bin.cb = sizeof(ULONG);
    spvIndexes[1].Value.bin.lpb = (LPBYTE)&spvIndexes[0].Value.l;

    LPSTREAM pStream, pRecipsStream = NULL;
    CIDList RecipIDList;
    DWORD dwRecipRecID;
    HRESULT hResult = GetStream (MessageEID.dwObjID, RECFLD_RECIPIENTS, &pRecipsStream, FALSE);
    if (!hResult)
    {
        hResult = RecipIDList.LoadFromStream (pRecipsStream);
        pRecipsStream->Release();
    }
    if (hResult)
    {   
        if (MAPI_E_NOT_FOUND == hResult)    // No recipients in the message
        {
            hResult = S_OK;
        }
        TraceResult ("CFileBackEnd::FillRecipientTable", hResult);
        return hResult;
    }
    dwRecipRecID = RecipIDList.Walk (WALK_START);
    while (dwRecipRecID)
    {
        hResult = GetStream (dwRecipRecID, RECFLD_PROPERTIES, &pStream, FALSE);
        if (!hResult)
        {
            // This object needs to be defined here so that for every iteration
            // of the WHILE() loop, the descructor gets called and the
            // internal data gets released.
            CPropList List;
            hResult = List.LoadFromStream (pStream);
            if (!hResult)
            {
                spvIndexes[0].Value.l++;
                hResult = List.Insert (&spvIndexes[0]);
                if (!hResult)
                {
                    hResult = List.Insert (&spvIndexes[1]);
                    if (!hResult)
                    {
                        hResult = List.GetAllProps (&srRow.lpProps, &srRow.cValues);
                        if (!hResult)
                        {
                            hResult = pTable->HrModifyRow (&srRow);
                            if (!hResult)
                            {
                                (*pdwRecipCount)++;
                            }
                            PrivFree ((LPVOID *)&srRow.lpProps);
                        }
                    }
                }
            }
            pStream->Release();
        }
        TraceResult ("CFileBackEnd::FillRecipTable: An error occured while adding an entry", hResult);
        dwRecipRecID = RecipIDList.Walk (WALK_CONTINUE);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::FillAttachmentsTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::FillAttachmentsTable (PRIVATE_ENTRYID  MessageEID,
                                                 LPTABLEDATA      pTable,
                                                 LPSBinary        pStoreEID,
                                                 long             lAttachmentAccess)
{
    CIDList AttachIDList;
    LPSTREAM pStream;
    HRESULT hResult = GetStream (MessageEID.dwObjID, RECFLD_ATTACHMENTS, &pStream, FALSE);
    if (!hResult)
    {
        AttachIDList.LoadFromStream (pStream);
        pStream->Release();
    }
    if (hResult || AttachIDList.IsEmpty())
    {   
        if (MAPI_E_NOT_FOUND == hResult)    // No Attachments in the message
        {
            hResult = S_OK;
        }
        TraceResult ("CFileBackEnd::FillAttachmentsTable", hResult);
        return hResult;
    }
    DWORD dwObjID, dwLastAllocated = 0;
    LPSPropValue pProps = NULL;
    CPropList * pList;
    dwObjID = AttachIDList.Walk (WALK_START);
    while (dwObjID)
    {
        hResult = GetStream (dwObjID, RECFLD_PROPERTIES, &pStream);
        if (!hResult)
        {
            hResult = LoadProperties (pStream, &pList);
            if (!hResult)
            {
                hResult = GetAttachmentTableProps (pList,
                                                   &pProps,
                                                   dwObjID,
                                                   lAttachmentAccess,
                                                   pStoreEID,
                                                   dwLastAllocated,
                                                   pTable);
                delete pList;
            }
            pStream->Release();
        }
        TraceResult ("CFileBackEnd::FillAttachmentsTable: An error occured while adding an entry", hResult);
        dwObjID = AttachIDList.Walk (WALK_CONTINUE);
    }
    PrivFree ((LPVOID *)&pProps);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetAttachmentTableProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::GetAttachmentTableProps (CPropList *     pList,
                                                    LPSPropValue *  ppProps,
                                                    ULONG           ulAttachNum,
                                                    long            lAttachmentAccess,
                                                    LPSBinary       pStoreEID,
                                                    DWORD &         dwLastAllocated,
                                                    LPTABLEDATA     pTable)
{
    // PR_INSTANCE_KEY
    // PR_ACCESS_LEVEL
    // PR_STORE_ENTRYID
    // PR_STORE_RECORD_KEY
    // PR_RECORD_KEY
    // PR_MAPPING_SIGNATURE
    // PR_OBJECT_TYPE
    // PR_ATTACH_NUM
    #define ON_THE_FLY_ATTACH_TABLE_PROPS    8
    LPSPropTagArray pTags;
    HRESULT hResult = pList->GetTags (&pTags);
    if (hResult)
    {
        return hResult;
    }
    
    ULONG ulFoundProps = 0, i = pTags->cValues;
    while (i)
    {
        i--;
        if (IsAttachmentTableProp (pTags->aulPropTag[i]))
        {
            ulFoundProps++;
        }
    }
    PrivFree ((LPVOID *)&pTags);
    ulFoundProps += ON_THE_FLY_ATTACH_TABLE_PROPS;
    if (dwLastAllocated < ulFoundProps)
    {
        PrivFree ((LPVOID *)ppProps);
        dwLastAllocated = 0;
        HRESULT hResult = PrivAllocate (sizeof(SPropValue) * ulFoundProps, (LPVOID *)ppProps, "Properties in Attachments Table");
        if (hResult)
        {
            return hResult;
        }
        dwLastAllocated = ulFoundProps;
    }
    i = 0;
    PFlatSPropValue pProp;
    pProp = pList->Walk (WALK_START);
    while (pProp)
    {
        if (IsAttachmentTableProp (pProp->ulPropTag))
        {
            (*ppProps)[i].ulPropTag = pProp->ulPropTag;
            (*ppProps)[i++].Value   = pProp->Value;
        }
        pProp = pList->Walk (WALK_CONTINUE);
    }

    (*ppProps)[i].ulPropTag          = PR_INSTANCE_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(ULONG);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&ulAttachNum;
    
    (*ppProps)[i].ulPropTag          = PR_ACCESS_LEVEL;
    (*ppProps)[i++].Value.l          = lAttachmentAccess;

    (*ppProps)[i].ulPropTag          = PR_STORE_ENTRYID;
    (*ppProps)[i].Value.bin.cb       = pStoreEID->cb;
    (*ppProps)[i++].Value.bin.lpb    = pStoreEID->lpb;
    
    (*ppProps)[i].ulPropTag          = PR_STORE_RECORD_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;
    
    (*ppProps)[i].ulPropTag          = PR_RECORD_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;

    (*ppProps)[i].ulPropTag          = PR_MAPPING_SIGNATURE;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;

    (*ppProps)[i].ulPropTag          = PR_OBJECT_TYPE;
    (*ppProps)[i++].Value.l          = MAPI_ATTACH;

    (*ppProps)[i].ulPropTag          = PR_ATTACH_NUM;
    (*ppProps)[i++].Value.ul         = ulAttachNum;
    ASSERT (i == ulFoundProps);
    SRow srRow = { 0 };
    srRow.cValues = i;
    srRow.lpProps = *ppProps;
    return pTable->HrModifyRow (&srRow);
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CreateObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::CreateObject (PRIVATE_ENTRYID   ParentEID,
                                         PRIVATE_ENTRYID * pNewObjEID,
                                         CPropList **      ppList,
                                         ULONG             ulObjType,
                                         LPVOID            pData1,
                                         LPVOID            pData2)
{
    if (ulObjType != MAPI_MESSAGE && ulObjType != MAPI_FOLDER && ulObjType != MAPI_ATTACH)
    {
        TraceMessage ("CFileBackEnd::CreateObject: Invalid object to create");
        return MAPI_E_NO_SUPPORT;
    }
    if (!IsValidEntryID (&ParentEID))
    {
        TraceMessage ("CFileBackEnd::CreateObject: Invalid parent entry ID");
        return MAPI_E_INVALID_ENTRYID;
    }
    LockDataBase();
    
    // Initialize the entry ID of the new object.
    ZeroMemory (pNewObjEID, CB_PRIVATE_EID);
    pNewObjEID->uidGlobal = m_uidBackEnd;
    pNewObjEID->bVersion  = ENTRYID_VERSION;
    pNewObjEID->bObject   = (BYTE)ulObjType;

    HRESULT hResult = m_pDBobj->CreateEmptyRecord (&(pNewObjEID->dwObjID));
    if (!hResult)
    {
        if (MAPI_MESSAGE == ulObjType)
        {
            hResult = InitMessageProps ((BOOL)pData2, ParentEID, (LPBYTE)pData1, ppList);
        }
        if (MAPI_ATTACH == ulObjType)
        {
            hResult = InitAttachProps (ppList);
        }
        if (MAPI_FOLDER == ulObjType)
        {
            hResult = CreateFolderObject (ParentEID,
                                          *pNewObjEID,
                                          (LPTSTR)pData1,
                                          pData2 ? (LPTSTR)pData2 : TEXT(""),
                                          ppList);
            if (!hResult)
            {
                UpdateFolderCount (ParentEID, *pNewObjEID);
            }
        }
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::CreateObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::OpenObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::OpenObject (PRIVATE_ENTRYID  ObjectEID,
                                       CPropList **     ppList)
{
    LockDataBase();
    LPSTREAM pStream;
    HRESULT hResult = GetStream (ObjectEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE);
    if (!hResult)
    {
        // Get the properties from the object's storage
        hResult = LoadProperties (pStream, ppList);
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::OpenObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CopyObjects()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::CopyObjects (PRIVATE_ENTRYID     SrcParentContainerEID,
                                        PRIVATE_ENTRYID     SrcContainerEID,
                                        PRIVATE_ENTRYID     DstParentContainerEID,
                                        PRIVATE_ENTRYID     DstContainerEID,
                                        DWORD               dwCopyFlags,
                                        LPMAPIPROGRESS      pProgress,
                                        POPENOBJECTS_DATA   pData)
{
    HRESULT hResult;
    LockDataBase();
    if ((MAPI_FOLDER == SrcContainerEID.bObject) && (CONTAINER_OBJECTS & dwCopyFlags))
    {
        hResult = CopyMessagesInFolder (SrcContainerEID,
                                        DstContainerEID,
                                        dwCopyFlags,
                                        pProgress,
                                        pData);
    }
    else
    {
        if (MAPI_MESSAGE == SrcContainerEID.bObject)
        {
            ASSERT (MAPI_MESSAGE == DstContainerEID.bObject);
            hResult = CopyMessageToMessage (SrcContainerEID.dwObjID, DstContainerEID.dwObjID);
        }
        else
        {
            hResult = MAPI_E_NO_SUPPORT;
        }
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::CopyObjects", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::UpdateAttachmentsCount()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::UpdateAttachmentsCount (PRIVATE_ENTRYID MessageEID,
                                                   PRIVATE_ENTRYID AttachEID,
                                                   BOOL            fRemove)
{
    CIDList AttachIDList;
    LPSTREAM pStream;
    HRESULT hResult = GetStream (MessageEID.dwObjID, RECFLD_ATTACHMENTS, &pStream);
    if (!hResult)
    {
        hResult = AttachIDList.LoadFromStream (pStream);
        if (!hResult)
        {
            if (fRemove)
            {
                AttachIDList.Delete (AttachEID.dwObjID);
            }
            else
            {
                AttachIDList.Insert (AttachEID.dwObjID);
            }
            AttachIDList.SaveToStream (pStream);
        }
        pStream->Release();
    }
    TraceResult ("CFileBackEnd::UpdateAttachmentsCount", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CreateFolderObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::CreateFolderObject (PRIVATE_ENTRYID    ParentEID,
                                               PRIVATE_ENTRYID    NewFolderEID,
                                               LPTSTR             pszNewFolderName,
                                               LPTSTR             pszNewFolderComment,
                                               CPropList **       ppList)
{
    LPSTREAM pStream;
    // When creating a folder container in a message store, one needs to check for name collisions.
    // Pass the information for the new folder. If the names do not collide, then the
    // function will write the object ID information and the name, in the parent container.
    // If the names collide, then it will return the object ID of the colliding folder.
    HRESULT hResult = CheckFldNameCollision (ParentEID.dwObjID,
                                             pszNewFolderName,
                                             &(NewFolderEID.dwObjID),
                                             FLDNAME_ADD);
    if (!hResult)
    {
        hResult = m_pDBobj->CreateRecord (&NewFolderEID.dwObjID, &pStream);
        if (!hResult)
        {
            // Set default property values of this folder
            hResult = InitFolderProps (pszNewFolderName, pszNewFolderComment, ParentEID, ppList);
            if (!hResult)
            {
                // Save the properties to the database.
                hResult = (*ppList)->SaveToStream (pStream);
            }
            pStream->Release();
            if (hResult)
            {
                // If we failed to create the container, we must remove it from the database
                // TODO: We must also remove it from the folder names list.
                //                  IrvingD 4/1/95
                m_pDBobj->DeleteRecord (NewFolderEID.dwObjID);
            }
        }
    }
    TraceResult ("CFileBackEnd::CreateFolderObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CreateSpecialFolders()
//
//    Parameters
//      None
//
//    Purpose
//      This function creates the storage containers of the special folder
//      in this store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::CreateSpecialFolders()
{
    // NOTE: The order of how these folders are created must be maintained to
    // preserve their fixed entry ID's.
    #define SPECIAL_FOLDER_NAMES 4
    LPTSTR szFolders[SPECIAL_FOLDER_NAMES] =
    {
        TEXT("IPM Subtree"),
        TEXT("Common Views"),
        TEXT("Folder Views"),
        TEXT("Finder")
    };
    LPTSTR szFoldersComments[SPECIAL_FOLDER_NAMES] =
    {
        TEXT("Mail Client IPM Subtree"),
        TEXT("Global View Restrictions"),
        TEXT("Folder View Restrictions"),
        TEXT("Search Results Folders")
    };
    #define IPM_SUBTREE_NAMES 4
    LPTSTR szIpmSubTree[IPM_SUBTREE_NAMES] =
    {
        TEXT("Deleted Items"),
        TEXT("Inbox"),      // As the default receive folder for the class ""
        TEXT("Outbox"),
        TEXT("Sent Items")
    };
    LPTSTR szIpmSubTreeComments[IPM_SUBTREE_NAMES] =
    {
        TEXT("Discarded Messages Waste Basket"),
        TEXT("Default Receive Folder"),
        TEXT("Outgoing Client Queue"),
        TEXT("Already Sent Messages")
    };

    // Create the root folder
    CPropList * pList;
    PRIVATE_ENTRYID eidIPMFolder, eidNewFolder, eidRoot = { 0 };
    eidRoot.uidGlobal = m_uidBackEnd;
    eidRoot.bVersion  = ENTRYID_VERSION;
    eidRoot.bObject   = MAPI_FOLDER;
    eidRoot.dwObjID   = ROOT_FOLDER_ID;
    LPSTREAM pStream;
    HRESULT hResult = m_pDBobj->CreateRecord (&eidRoot.dwObjID, &pStream);
    if (!hResult)
    {
        // Set default property values of this folder
        hResult = InitFolderProps (TEXT("Root"), TEXT("Store Root Folder"), eidRoot, &pList);
        if (!hResult)
        {
            // Save the properties to the database.
            hResult = pList->SaveToStream (pStream);
            delete pList;
        }
        pStream->Release();
    }
    if (!hResult)
    {
        // Iterate through the names and create each.
        for (ULONG i=0; i<SPECIAL_FOLDER_NAMES; i++)
        {
            hResult = CreateObject (eidRoot, &eidNewFolder, &pList, MAPI_FOLDER, szFolders[i], szFoldersComments[i]);
            if (hResult)
            {
                goto ErrorExit;
            }
            delete pList;
            // If we are to create the IPM Subtree, use this as the parent folder and create the default folders.
            if (0 == i)
            {
                for (ULONG j=0; j<IPM_SUBTREE_NAMES; j++)
                {
                    hResult = CreateObject (eidNewFolder, &eidIPMFolder, &pList, MAPI_FOLDER, szIpmSubTree[j], szIpmSubTreeComments[j]);
                    if (hResult)
                    {
                        goto ErrorExit;
                    }
                    delete pList;
                }
            }
        }
    }
ErrorExit:
    TraceResult ("CFileBackEnd::CreateSpecialFolders", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CheckFldNameCollision()
//
//    Parameters
//
//    Purpose
//      This function checks for name collision for folder containers. If a
//      folder already exists with the same name as pszNewFolderName, a store
//      should not allow the new folder to be created.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::CheckFldNameCollision (DWORD     dwFolderID,
                                                  LPTSTR    pszNewFolderName,
                                                  DWORD *   pdwExistingFldID,
                                                  DWORD     dwFlags)
{
    ASSERT (pdwExistingFldID);
    DWORD dwOldNameCount, dwNamesCount, cbSize;
    TCHAR szName[256];
    LPSTREAM pStream, pNewStream = NULL;
    LARGE_INTEGER liZero = { 0 };
    DWORD dwFldID;
    // Open the stream in the folder where the folder names are.
    HRESULT hResult = GetStream (dwFolderID, RECFLD_NAMES, &pStream);
    if (!hResult)
    {
        if (FLDNAME_REPLACE & dwFlags)
        {
            hResult = CreateStreamOnHGlobal (NULL, TRUE, &pNewStream);
        }
        if (!hResult)
        {
            // How many folder names we have?
            hResult = pStream->Read (&dwNamesCount, sizeof(DWORD), NULL);
            if (!hResult)
            {
                // If replacing...
                if (pNewStream)
                {
                    pNewStream->Write (&dwNamesCount, sizeof(DWORD), NULL);
                }
                
                // Save this count, and start reading the names until we compared all
                // the names of child folders or until an error occurred
                dwOldNameCount = dwNamesCount;
                while (dwNamesCount && !hResult)
                {
                    // Read the ID of the folder we are about to compare
                    hResult = pStream->Read (&dwFldID, sizeof(DWORD), NULL);
                    if (!hResult)
                    {
                        // How many characters its name has?
                        hResult = pStream->Read (&cbSize, sizeof(DWORD), NULL);
                        if (!hResult)
                        {
                            // Read the name string
                            ASSERT (cbSize < 256);
                            hResult = pStream->Read (szName, cbSize, NULL);
                            if (!hResult)
                            {
                                // Add the NULL terminator. This is not stored in the file.
                                szName[cbSize] = 0;

                                // If we are REPLACING an entry, then we only need to compare
                                // the ID of the read folder. If they match, we need to
                                // replace this entry
                                if (FLDNAME_REPLACE & dwFlags)
                                {
                                    if (dwFldID == *pdwExistingFldID)
                                    {
                                        break; // Out of the WHILE() loop
                                    }
                                    // If this is not the entry we are replacing, save it to
                                    // the new stream we are creating.
                                    if (pNewStream)
                                    {
                                        pNewStream->Write (&dwFldID, sizeof(DWORD), NULL);
                                        pNewStream->Write (&cbSize, sizeof(DWORD), NULL);
                                        pNewStream->Write (szName, cbSize, NULL);
                                    }
                                }
                                else
                                {                                
                                    // Do a case-insensitive operation of the folder names
                                    if (!lstrcmpi (szName, pszNewFolderName))
                                    {
                                        // Bingo!
                                        hResult = MAPI_E_COLLISION;
                                        break; // Out of the WHILE() loop
                                    }
                                }
                            }
                        }
                    }
                    dwNamesCount--;
                }
                // Check for errors
                if (hResult)
                {
                    if (MAPI_E_COLLISION == hResult)
                    {
                        // Return the ID of the matched folder
                        *pdwExistingFldID = dwFldID;
                    }
                }
                if (!hResult && (dwFlags & FLDNAME_ADD))
                {
                    // Ok, no collision occurred. We are at the end of the stream so we will write the data
                    // of the new folder we are about to create
                    hResult = pStream->Write (pdwExistingFldID, sizeof(DWORD), NULL);
                    if (!hResult)
                    {
                        // Save the lenght (in bytes) of the string, but don't store the NULL terminator
                        cbSize = Cbtszsize (pszNewFolderName) - sizeof(TCHAR);
                        hResult = pStream->Write (&cbSize, sizeof(DWORD), NULL);
                        if (!hResult)
                        {
                            // Write the string
                            hResult = pStream->Write (pszNewFolderName, cbSize, NULL);
                            if (!hResult)
                            {
                                // Move to the beginning of the stream
                                pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                                // Save the new count of names in the stream
                                dwOldNameCount++;
                                hResult = pStream->Write (&dwOldNameCount, sizeof(DWORD), NULL);
                            }
                        }
                    }
                }
                if (!hResult && (dwFlags & FLDNAME_REPLACE))
                {
                    ULARGE_INTEGER uliBigNumber = { 0xFFFFFFFF, 0 }, uliWritten;
                    pStream->CopyTo (pNewStream, uliBigNumber, NULL, NULL);
                    pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                    pNewStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                    pNewStream->CopyTo (pStream, uliBigNumber, NULL, &uliWritten);
                    pStream->SetSize (uliWritten);
                }
            }
        }
        if (pNewStream)
        {
            pNewStream->Release();
        }
        pStream->Release();
    }
    if (hResult && hResult != MAPI_E_COLLISION)
    {
        TraceResult ("CFileBackEnd::CheckFldNameCollision", hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteFolderNameInParent()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteFolderNameInParent (DWORD  dwParentFolderID,
                                                     DWORD  dwFolderID)
{
    DWORD dwNamesCount, cbSize, dwStoredFldID, dwNewCount;
    TCHAR szName[256];
    LPSTREAM pStream, pNewStream;
    LARGE_INTEGER   liZero = { 0 };
    BOOL fFound = FALSE;
    HRESULT hResult = CreateStreamOnHGlobal (NULL, TRUE, &pNewStream);
    if (!hResult)
    {
        // Open the stream in the folder where the folder names are.
        hResult = GetStream (dwParentFolderID, RECFLD_NAMES, &pStream);
        if (!hResult)
        {
            // How many folder names we have?
            hResult = pStream->Read (&dwNamesCount, sizeof(DWORD), NULL);
            if (!hResult)
            {
                ASSERT (dwNamesCount > 0);
                dwNewCount = dwNamesCount;
                dwNewCount--;
                hResult = pNewStream->Write (&dwNewCount, sizeof(DWORD), NULL);
                do
                {
                    // Read the ID of the folder we are about to compare
                    hResult = pStream->Read (&dwStoredFldID, sizeof(DWORD), NULL);
                    if (!hResult)
                    {
                        // How many characters its name has?
                        hResult = pStream->Read (&cbSize, sizeof(DWORD), NULL);
                        if (!hResult)
                        {
                            // Read the name string
                            ASSERT (cbSize < 256);
                            hResult = pStream->Read (szName, cbSize, NULL);
                            if (!hResult)
                            {
                                szName[cbSize] = 0;
                                if (dwFolderID == dwStoredFldID)
                                {
                                    fFound = TRUE;
                                }
                            }
                        }
                    }
                    if (!hResult)
                    {
                        if (fFound)
                        {
                            // When the folder we are deleting has been found, we copy the rest of the
                            // original stream into the new stream. Setting up a big number for the
                            // number of bytes cause the IStream interface to copy "the rest" of the stream.
                            ULARGE_INTEGER uliBigNumber = { 0xFFFFFFFF, 0 }, uliWritten;
                            pStream->CopyTo (pNewStream, uliBigNumber, NULL, NULL);
                            pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                            pNewStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                            pNewStream->CopyTo (pStream, uliBigNumber, NULL, &uliWritten);
                            pStream->SetSize (uliWritten);
                        }
                        else
                        {
                            // New folder information we just read needs to be saved to the new stream.
                            pNewStream->Write (&dwStoredFldID, sizeof(DWORD), NULL);
                            pNewStream->Write (&cbSize, sizeof(DWORD), NULL);
                            pNewStream->Write (szName, cbSize, NULL);
                        }
                    }
                    dwNamesCount--;
                }
                while (!fFound && dwNamesCount && !hResult);
            }
        }
        pStream->Release();
        pNewStream->Release();
    }
    TraceResult ("CFileBackEnd::DeleteFolderNameInParent", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::FillHierarchyTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::FillHierarchyTable (PRIVATE_ENTRYID  ParentFolderEID,
                                               LPTABLEDATA      pTable,
                                               LPSBinary        pStoreEID,
                                               long             lFolderAccess)
{
    LPSTREAM pStream;
    CIDList FolderIDList;
    HRESULT hResult = GetStream (ParentFolderEID.dwObjID, RECFLD_HIERARCHY, &pStream, FALSE);
    if (!hResult)
    {
        hResult = FolderIDList.LoadFromStream (pStream);
        pStream->Release();
    }
    if (hResult || FolderIDList.IsEmpty())
    {
        if (MAPI_E_NOT_FOUND == hResult)
        {
            hResult = S_OK;
        }
        TraceResult ("CFileBackEnd::FillHierarchyTable", hResult);
        return hResult;
    }
    
    PRIVATE_ENTRYID FolderEID = { 0 };
    FolderEID.uidGlobal = m_uidBackEnd;
	FolderEID.bVersion = ENTRYID_VERSION;
    FolderEID.bObject = MAPI_FOLDER;
    FolderEID.dwObjID = FolderIDList.Walk (WALK_START);
    
    CPropList * pList;
    LPSPropValue pProps = NULL;
    DWORD dwLastAllocated = 0;
    while (FolderEID.dwObjID)
    {
        hResult = GetStream (FolderEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE);
        if (!hResult)
        {
            hResult = LoadProperties (pStream, &pList);
            if (!hResult)
            {
                hResult = GetHierarchyTableProps (pList,
                                                  &pProps,
                                                  FolderEID,
                                                  ParentFolderEID,
                                                  lFolderAccess, 
                                                  pStoreEID,
                                                  FOLDER_GENERIC,
                                                  dwLastAllocated,
                                                  pTable,
                                                  m_uidBackEnd);
                delete pList;
            }
            pStream->Release();
        }
        TraceResult ("CFileBackEnd::FillHierarchyTable: An error occured while adding an entry", hResult);
        FolderEID.dwObjID = FolderIDList.Walk (WALK_CONTINUE);
    }
    PrivFree ((LPVOID *)&pProps);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::FillContentsTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::FillContentsTable (PRIVATE_ENTRYID  FolderEID,
                                              LPTABLEDATA      pTable,
                                              LPSBinary        pStoreEID,
                                              long             lMessageAccess,
                                              BOOL             fGetAssocContTable)
{
    CIDList MsgsIDList;
    LPSTREAM pStream;
    HRESULT hResult = GetStream (FolderEID.dwObjID, fGetAssocContTable ? RECFLD_ASSOCIATED : RECFLD_CONTENTS, &pStream, FALSE);
    if (!hResult)
    {
        hResult = MsgsIDList.LoadFromStream (pStream);
        pStream->Release();
    }
    if (hResult || MsgsIDList.IsEmpty())
    {
        if (MAPI_E_NOT_FOUND == hResult)
        {
            hResult = S_OK;
        }
        TraceResult ("CFileBackEnd::FillContentsTable", hResult);
        return hResult;
    }
    PRIVATE_ENTRYID MessageEID = { 0 };
    MessageEID.uidGlobal = m_uidBackEnd;
	MessageEID.bVersion = ENTRYID_VERSION;
    MessageEID.bObject = MAPI_MESSAGE;
    MessageEID.dwObjID = MsgsIDList.Walk (WALK_START);

    CPropList * pList;
    LPSPropValue pProps = NULL;
    DWORD dwLastAllocated = 0;
    while (MessageEID.dwObjID)
    {
        hResult = GetStream (MessageEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE);
        if (!hResult)
        {
            hResult = LoadProperties (pStream, &pList);
            if (!hResult)
            {
                hResult = GetContentsTableProps (pList,
                                                 &pProps,
                                                 MessageEID,
                                                 lMessageAccess,
                                                 pStoreEID,
                                                 dwLastAllocated,
                                                 pTable);
                delete pList;
            }
            pStream->Release();
        }
        TraceResult ("CFileBackEnd::FillContentsTable: An error occurred filling a row", hResult);
        MessageEID.dwObjID = MsgsIDList.Walk (WALK_CONTINUE);
    }
    PrivFree ((LPVOID *)&pProps);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetRFSEntryID()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::GetRFSEntryID (LPTSTR            pszMsgClass,
                                          PRIVATE_ENTRYID * pFolderEID,
                                          LPTSTR            pszExplicitClass)
{
    LPSTREAM pStream;
    DWORD cbStringSize, dwClassesCount, dwFolderID;
    TCHAR szClass[256];
    BOOL bFound = FALSE;

    LockDataBase();
    HRESULT hResult = GetStream (RECEIVE_FOLDER_SETTINGS, RECFLD_PROPERTIES, &pStream, FALSE);
    if (!hResult)
    {
        hResult = pStream->Read (&dwClassesCount, sizeof(DWORD), NULL);
        if (!hResult)
        {
            while (dwClassesCount && !hResult)
            {
                hResult = pStream->Read (&cbStringSize, sizeof(DWORD), NULL);
                if (!hResult)
                {
                    ASSERT (cbStringSize < 256);
                    hResult = pStream->Read (szClass, cbStringSize, NULL);
                    if (!hResult)
                    {
                        hResult = pStream->Read (&dwFolderID, sizeof(DWORD), NULL);
                        if (!hResult)
                        {
                            szClass[cbStringSize] = 0; // Add NULL terminator not stored in file.
                            if (0 == lstrcmpi (szClass, pszMsgClass))
                            {
                                bFound = TRUE;
                                pFolderEID->uidGlobal = m_uidBackEnd;
                                pFolderEID->bVersion  = ENTRYID_VERSION;
                                pFolderEID->bObject   = MAPI_FOLDER;
                                pFolderEID->dwObjID   = dwFolderID;
                            }
                        }
                    }
                }
                dwClassesCount--;
            }
        }
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::GetRFSEntryID", hResult);
    if (!bFound && !hResult)
    {
        hResult = MAPI_E_NOT_FOUND;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::SetRFSEntryID()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::SetRFSEntryID (LPTSTR          pszMsgClass,
                                          PRIVATE_ENTRYID FolderEID,
                                          LPTABLEDATA     pTable)
{
    if (!IsValidEntryID (&FolderEID) || (FolderEID.bObject != MAPI_FOLDER))
    {
        TraceMessage ("CFileBackEnd::SetRFSEntryID: Invalid entry ID for receive folder");
        return MAPI_E_INVALID_ENTRYID;
    }
    LPSTREAM pStream;
    DWORD dwClassesCount, cbStringSize, dwRow = 1;
    TCHAR szClass[256];
    LARGE_INTEGER liZero = { 0 };

    LockDataBase();
    HRESULT hResult = GetStream (RECEIVE_FOLDER_SETTINGS, RECFLD_PROPERTIES, &pStream);
    if (!hResult)
    {
        hResult = pStream->Read (&dwClassesCount, sizeof(DWORD), NULL);
        if (!hResult)
        {
            while (dwRow <= dwClassesCount && !hResult)
            {
                hResult = pStream->Read (&cbStringSize, sizeof(DWORD), NULL);
                if (!hResult)
                {
                    ASSERT (cbStringSize < 256);
                    hResult = pStream->Read (szClass, cbStringSize, NULL);
                    if (!hResult)
                    {
                        // Add NULL terminator not stored in file.
                        szClass[cbStringSize] = 0;
                        // TODO: This class name comparison is not compliant with the MAPI specs.
                        //              IrvingD 3/30/95
                        if (0 == lstrcmpi (szClass, pszMsgClass))
                        {
                            hResult = pStream->Write (&FolderEID.dwObjID, sizeof(DWORD), NULL);
                            if (!hResult)
                            {
                                // Just so that we know we don't need to write this class at the end of the stream
                                hResult = MAPI_E_COLLISION;
                                // dwRow is used below to fill PR_INSTANCE_KEY. The instance key for this row in
                                // the table (if any was passed in) is offset'ed by 1, to account for the message class "".
                                dwRow++;
                            }
                        }
                    }
                }
                dwRow++;
            }
            if (!hResult)
            {
                cbStringSize = (lstrlen (pszMsgClass))*sizeof(TCHAR);
                hResult = pStream->Write (&cbStringSize, sizeof(DWORD), NULL);
                if (!hResult)
                {
                    hResult = pStream->Write (pszMsgClass, cbStringSize, NULL);
                    if (!hResult)
                    {
                        hResult = pStream->Write (&FolderEID.dwObjID, sizeof(DWORD), NULL);
                        if (!hResult)
                        {
                            pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
                            dwClassesCount++;
                            hResult = pStream->Write (&dwClassesCount, sizeof(DWORD), NULL);
                        }
                    }
                }
                if (!hResult && pTable)
                {
                    SRow srRow = { 0 };
                    SPropValue spvProps[NUM_RFS_TABLE_PROPS] = { 0 };
                    // Initialize the property tags in the row set
                    srRow.cValues = NUM_RFS_TABLE_PROPS;
                    srRow.lpProps = spvProps;
                    for (UINT i=0; i<NUM_RFS_TABLE_PROPS; i++)
                    {
                        spvProps[i].ulPropTag = sptRFSprops.aulPropTag[i];
                    }
                    spvProps[RFS_INST_KEY].Value.bin.cb = sizeof(DWORD);
                    // The instance key is generated using the counter for the number of rows
                    // added in to the table + 1. We need to add an extra row to the number of
                    // classes stored in the stream to account for the row in the table for
                    // the message class ""
                    dwClassesCount++;
                    spvProps[RFS_INST_KEY].Value.bin.lpb = (LPBYTE)&dwClassesCount;
                    spvProps[RFS_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
                    spvProps[RFS_ENTRYID].Value.bin.lpb = (LPBYTE)&FolderEID;
                    spvProps[RFS_RECORD_KEY].Value.bin.cb = CB_PRIVATE_EID;
                    spvProps[RFS_RECORD_KEY].Value.bin.lpb = (LPBYTE)&FolderEID;
                    spvProps[RFS_MSG_CLASS].Value.LPSZ = pszMsgClass;
                    hResult = pTable->HrModifyRow (&srRow);
                }
            }
            else
            {
                if (hResult == MAPI_E_COLLISION)
                {
                    hResult = S_OK;
                    if (pTable)
                    {
                        LPSRow pRow;
                        SPropValue Prop = { 0 };
                        Prop.ulPropTag = PR_INSTANCE_KEY;
                        Prop.Value.bin.cb = sizeof(DWORD);
                        Prop.Value.bin.lpb = (LPBYTE)&dwRow;
                        hResult = pTable->HrQueryRow (&Prop, &pRow, NULL);
                        if (!hResult)
                        {
                            LPSPropValue pProps = pRow->lpProps;
                            ASSERT (pProps[RFS_INST_KEY].ulPropTag == PR_INSTANCE_KEY);
                            ASSERT (pProps[RFS_ENTRYID].ulPropTag == PR_ENTRYID);
                            ASSERT (pProps[RFS_RECORD_KEY].ulPropTag == PR_RECORD_KEY);
                            ASSERT (pProps[RFS_MSG_CLASS].ulPropTag == PR_MESSAGE_CLASS);
                            pProps[RFS_ENTRYID].Value.bin.lpb = (LPBYTE)&FolderEID;
                            hResult = pTable->HrModifyRow (pRow);
                            PrivMAPIFree ((LPVOID *)pRow);
                        }
                    }
                }
            }
        }
        pStream->Release();
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::SetRFSEntryID", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetRFSTableData()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::GetRFSTableData (LPTABLEDATA pTable)
{
    SRow srRow = { 0 };
    SPropValue spvProps[NUM_RFS_TABLE_PROPS] = { 0 };
    // Initialize the property tags in the row set
    srRow.cValues = NUM_RFS_TABLE_PROPS;
    srRow.lpProps = spvProps;
    for (UINT i=0; i<NUM_RFS_TABLE_PROPS; i++)
    {
        spvProps[i].ulPropTag = sptRFSprops.aulPropTag[i];
    }

    DWORD dwClassesCount = 0, cbStringSize, dwFolderID;
    TCHAR szClass[256] = { 0 };
    // First, add a row to the table with the default inbox folder for the "" message class
    // Initialize the values for the CB members only once.
    PRIVATE_ENTRYID eid = { 0 };
    // Fill the entry ID structure values.
    eid.uidGlobal = m_uidBackEnd;
    eid.bVersion = ENTRYID_VERSION;
    eid.bObject = MAPI_FOLDER;
    eid.dwObjID = 3;    // This is the entry ID of the INBOX

    spvProps[RFS_INST_KEY].Value.bin.cb = sizeof(DWORD);
    spvProps[RFS_INST_KEY].Value.bin.lpb = (LPBYTE)&dwClassesCount;
    spvProps[RFS_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
    spvProps[RFS_ENTRYID].Value.bin.lpb = (LPBYTE)&eid;
    spvProps[RFS_RECORD_KEY].Value.bin.cb = CB_PRIVATE_EID;
    spvProps[RFS_RECORD_KEY].Value.bin.lpb = (LPBYTE)&eid;
    spvProps[RFS_MSG_CLASS].Value.LPSZ = szClass;
    LPSTREAM pStream;
    HRESULT hResult = pTable->HrModifyRow (&srRow);
    if (!hResult)
    {
        hResult = GetStream (RECEIVE_FOLDER_SETTINGS, RECFLD_PROPERTIES, &pStream, FALSE);
        if (!hResult)
        {
            hResult = pStream->Read (&dwClassesCount, sizeof(DWORD), NULL);
            if (!hResult)
            {
                while (dwClassesCount && !hResult)
                {
                    hResult = pStream->Read (&cbStringSize, sizeof(DWORD), NULL);
                    if (!hResult)
                    {
                        ASSERT (cbStringSize < 256);
                        hResult = pStream->Read (szClass, cbStringSize, NULL);
                        if (!hResult)
                        {
                            hResult = pStream->Read (&dwFolderID, sizeof(DWORD), NULL);
                            if (!hResult)
                            {
                                // The property array values will change because they
                                // have pointer to these variables
                                szClass[cbStringSize] = 0; // Add NULL terminator not stored in file.
                                eid.dwObjID = dwFolderID;
                                hResult = pTable->HrModifyRow (&srRow);
                            }
                        }   
                    }
                    dwClassesCount--;
                }
            }
            pStream->Release();
        }
    }
    TraceResult ("CFileBackEnd::GetRFSTableData", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetContentsTableProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::GetContentsTableProps (CPropList *     pList,
                                                  LPSPropValue *  ppProps,
                                                  PRIVATE_ENTRYID MessageEID,
                                                  long            lAccessLevel,
                                                  LPSBinary       pStoreEID,
                                                  DWORD &         dwLastAllocated,
                                                  LPTABLEDATA     pTable)
{
    // PR_INSTANCE_KEY
    // PR_ACCESS_LEVEL
    // PR_ENTRYID
    // PR_STORE_ENTRYID
    // PR_STORE_RECORD_KEY
    // PR_RECORD_KEY
    // PR_MAPPING_SIGNATURE
    // PR_OBJECT_TYPE
    // PR_HASATTACH
    #define ON_THE_FLY_CONTENT_TABLE_PROPS    9

    LPSPropTagArray pTags;
    HRESULT hResult = pList->GetTags (&pTags);
    if (hResult)
    {
        return hResult;
    }
    ULONG ulFoundProps = 0, i = pTags->cValues;
    while (i)
    {
        i--;
        if (IsContentTableProp (pTags->aulPropTag[i]))
        {
            ulFoundProps++;
        }
    }
    PrivFree ((LPVOID *)&pTags);
    
    ulFoundProps += ON_THE_FLY_CONTENT_TABLE_PROPS;

    if (dwLastAllocated < ulFoundProps)
    {
        PrivFree ((LPVOID *)ppProps);
        dwLastAllocated = 0;
        HRESULT hResult = PrivAllocate (sizeof(SPropValue) * ulFoundProps, (LPVOID *)ppProps, "Properties in Contents Table");
        if (hResult)
        {
            return hResult;
        }
        dwLastAllocated = ulFoundProps;
    }
    
    i = 0;
    PFlatSPropValue pProp;
    pProp = pList->Walk (WALK_START);
    while (pProp)
    {
        if (IsContentTableProp (pProp->ulPropTag))
        {
            (*ppProps)[i].ulPropTag = pProp->ulPropTag;
            (*ppProps)[i++].Value   = pProp->Value;
        }
        pProp = pList->Walk (WALK_CONTINUE);
    }
    DWORD key = MessageEID.dwObjID;
    (*ppProps)[i].ulPropTag          = PR_INSTANCE_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(DWORD);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&key;

    (*ppProps)[i].ulPropTag          = PR_ACCESS_LEVEL;
    (*ppProps)[i++].Value.l          = lAccessLevel;

    (*ppProps)[i].ulPropTag          = PR_ENTRYID;
    (*ppProps)[i].Value.bin.cb       = CB_PRIVATE_EID;
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&MessageEID;

    (*ppProps)[i].ulPropTag          = PR_STORE_ENTRYID;
    (*ppProps)[i].Value.bin.cb       = pStoreEID->cb;
    (*ppProps)[i++].Value.bin.lpb    = pStoreEID->lpb;

    (*ppProps)[i].ulPropTag          = PR_STORE_RECORD_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;

    (*ppProps)[i].ulPropTag          = PR_RECORD_KEY;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;

    (*ppProps)[i].ulPropTag          = PR_MAPPING_SIGNATURE;
    (*ppProps)[i].Value.bin.cb       = sizeof(MAPIUID);
    (*ppProps)[i++].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;

    (*ppProps)[i].ulPropTag          = PR_OBJECT_TYPE;
    (*ppProps)[i++].Value.l          = MAPI_MESSAGE;

    SPropValue spvOne;
    pList->Find (PR_MESSAGE_FLAGS, &spvOne);
    (*ppProps)[i].ulPropTag          = PR_HASATTACH;
    (*ppProps)[i++].Value.b          = (USHORT)!!(spvOne.Value.l & MSGFLAG_HASATTACH);

    ASSERT (i == ulFoundProps);
    SRow srRow = { 0 };
    srRow.cValues = i;
    srRow.lpProps = *ppProps;
    return pTable->HrModifyRow (&srRow);
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::IsContentTableProp()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI CFileBackEnd::IsContentTableProp (ULONG ulPropTag)
{
    switch (ulPropTag)
    {
        case PR_MESSAGE_FLAGS :
        case PR_CREATION_TIME :
        case PR_LAST_MODIFICATION_TIME :
        case PR_MESSAGE_SIZE :
        case PR_PRIORITY :
        case PR_SEARCH_KEY :
        case PR_MESSAGE_CLASS :
        case PR_PARENT_ENTRYID :
        case PR_SENDER_NAME :
        case PR_SENDER_ENTRYID :
        case PR_SENDER_SEARCH_KEY :
        case PR_SENDER_ADDRTYPE :
        case PR_SENDER_EMAIL_ADDRESS :
        case PR_RECEIPT_TIME :
        case PR_RECIPIENT_REASSIGNMENT_PROHIBITED :
        case PR_REDIRECTION_HISTORY :
        case PR_RELATED_IPMS :
        case PR_LANGUAGES :
        case PR_REPLY_TIME :
        case PR_REPORT_TAG :
        case PR_REPORT_TIME :
        case PR_RETURNED_IPM :
        case PR_SECURITY :
        case PR_INCOMPLETE_COPY :
        case PR_SUBJECT_IPM :
        case PR_REPORT_NAME :
        case PR_SENT_REPRESENTING_SEARCH_KEY :
        case PR_X400_CONTENT_TYPE :
        case PR_NON_RECEIPT_REASON :
        case PR_RECEIVED_BY_ENTRYID :
        case PR_RECEIVED_BY_NAME :
        case PR_SENT_REPRESENTING_ENTRYID :
        case PR_SENT_REPRESENTING_NAME :
        case PR_RCVD_REPRESENTING_ENTRYID :
        case PR_RCVD_REPRESENTING_NAME :
        case PR_REPORT_ENTRYID :
        case PR_READ_RECEIPT_ENTRYID :
        case PR_MESSAGE_SUBMISSION_ID :
        case PR_PROVIDER_SUBMIT_TIME :
        case PR_ORIGINAL_SUBJECT :
        case PR_DISC_VAL :
        case PR_ORIG_MESSAGE_CLASS :
        case PR_ORIGINAL_AUTHOR_ENTRYID :
        case PR_ORIGINAL_AUTHOR_NAME :
        case PR_ORIGINAL_SUBMIT_TIME :
        case PR_REPLY_RECIPIENT_ENTRIES :
        case PR_REPLY_RECIPIENT_NAMES :
        case PR_RECEIVED_BY_SEARCH_KEY :
        case PR_RCVD_REPRESENTING_SEARCH_KEY :
        case PR_READ_RECEIPT_SEARCH_KEY :
        case PR_REPORT_SEARCH_KEY :
        case PR_ORIGINAL_DELIVERY_TIME :
        case PR_ORIGINAL_AUTHOR_SEARCH_KEY :
        case PR_MESSAGE_TO_ME :
        case PR_MESSAGE_CC_ME :
        case PR_MESSAGE_RECIP_ME :
        case PR_ORIGINAL_SENDER_NAME :
        case PR_ORIGINAL_SENDER_ENTRYID :
        case PR_ORIGINAL_SENDER_SEARCH_KEY :
        case PR_ORIGINAL_SENT_REPRESENTING_NAME :
        case PR_ORIGINAL_SENT_REPRESENTING_ENTRYID :
        case PR_ORIGINAL_SENT_REPRESENTING_SEARCH_KEY :
        case PR_START_DATE :
        case PR_END_DATE :
        case PR_OWNER_APPT_ID :
        case PR_RESPONSE_REQUESTED :
        case PR_SENT_REPRESENTING_ADDRTYPE :
        case PR_SENT_REPRESENTING_EMAIL_ADDRESS :
        case PR_ORIGINAL_SENDER_ADDRTYPE :
        case PR_ORIGINAL_SENDER_EMAIL_ADDRESS :
        case PR_ORIGINAL_SENT_REPRESENTING_ADDRTYPE :
        case PR_ORIGINAL_SENT_REPRESENTING_EMAIL_ADDRESS :
        case PR_CONVERSATION_TOPIC :
        case PR_CONVERSATION_INDEX :
        case PR_IMPORTANCE :
        case PR_SENSITIVITY :
        case PR_SUBJECT :
        case PR_NORMALIZED_SUBJECT :
        case PR_CLIENT_SUBMIT_TIME :
        case PR_DISPLAY_TO :
        case PR_DISPLAY_CC :
        case PR_DISPLAY_BCC :
        case PR_BODY :
        case PR_MESSAGE_DELIVERY_TIME :
        case PR_MSG_STATUS :
            return TRUE;
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::IsAttachmentTableProp()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI CFileBackEnd::IsAttachmentTableProp (ULONG ulPropTag)
{
    switch (ulPropTag)
    {
        case PR_CREATION_TIME :
        case PR_LAST_MODIFICATION_TIME :
        case PR_DISPLAY_NAME :
        case PR_ATTACH_ENCODING :
        case PR_ATTACH_EXTENSION :
        case PR_ATTACH_FILENAME :
        case PR_ATTACH_LONG_FILENAME :
        case PR_ATTACH_PATHNAME :
        case PR_ATTACH_RENDERING :
        case PR_ATTACH_TAG :
        case PR_ATTACH_METHOD :
        case PR_RENDERING_POSITION :
        case PR_ATTACH_TRANSPORT_NAME :
            return TRUE;
    }
    return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::UpdateContentCount()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::UpdateContentCount (PRIVATE_ENTRYID ContainerEID,
                                               PRIVATE_ENTRYID ObjectEID,
                                               BOOL            fAssociated,
                                               BOOL            fUnread)
{
    LPSTREAM pStream;
    LockDataBase();
    HRESULT hResult = GetStream (ContainerEID.dwObjID, RECFLD_PROPERTIES, &pStream);
    if (!hResult)
    {
        SPropValue spvOne = { 0 };
        CPropList * pList;
        hResult = LoadProperties (pStream, &pList);
        if (!hResult)
        {
            pList->Find (fAssociated ? PR_ASSOC_CONTENT_COUNT : PR_CONTENT_COUNT, &spvOne);
            spvOne.Value.l++;
            pList->Insert (&spvOne);
            if (fUnread)
            {
                pList->Find (PR_CONTENT_UNREAD, &spvOne);
                spvOne.Value.l++;
                pList->Insert (&spvOne);
            }
            pList->SaveToStream (pStream);
            pStream->Release();
            delete pList;
        }
    }
    if (!hResult)
    {
        hResult = GetStream (ContainerEID.dwObjID, fAssociated ? RECFLD_ASSOCIATED : RECFLD_CONTENTS, &pStream);
        if (!hResult)
        {
            CIDList MsgsIDList;
            hResult = MsgsIDList.LoadFromStream (pStream);
            if (!hResult)
            {
                MsgsIDList.Insert (ObjectEID.dwObjID);
                MsgsIDList.SaveToStream (pStream);
            }
            pStream->Release();
        }
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::UpdateContentCount", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::UpdateFolderCount()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::UpdateFolderCount (PRIVATE_ENTRYID FolderEID,
                                              PRIVATE_ENTRYID NewFolderEID)
{
    LPSTREAM pStream;
    CIDList FolderIDList;
    HRESULT hResult = GetStream (FolderEID.dwObjID, RECFLD_HIERARCHY, &pStream);
    if (!hResult)
    {
        hResult = FolderIDList.LoadFromStream (pStream);
        if (!hResult)
        {
            FolderIDList.Insert (NewFolderEID.dwObjID);
            FolderIDList.SaveToStream (pStream);
        }
        pStream->Release();
    }
    TraceResult ("CFileBackEnd::UpdateFolderCount", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CopyObjectProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::CopyObjectProps (DWORD dwSrcMsgID, DWORD dwDstMsgID)
{
    LPSTREAM pSrc, pDst, pPropStream, pNewPropStream;
    CPropList SrcList, DstList;
    ULARGE_INTEGER uliSize = { 0 };
    uliSize.LowPart = 0xFFFFFFFF;
    SPropValue spvOld = { 0 }, spvNew = { 0 };
    HRESULT hResult = GetStream (dwSrcMsgID, RECFLD_PROPERTIES, &pSrc, FALSE);
    if (!hResult)
    {
        SrcList.LoadFromStream (pSrc);
        hResult = GetStream (dwDstMsgID, RECFLD_PROPERTIES, &pDst);
        if (!hResult)
        {
            hResult = DstList.LoadFromStream (pDst);
            PFlatSPropValue pProp = SrcList.Walk (WALK_START);
            while (!hResult && pProp)
            {
                spvNew.ulPropTag = pProp->ulPropTag;
                spvNew.Value = pProp->Value;
                if (PT_STREAM_STORAGE == PROP_TYPE (pProp->ulPropTag))
                {
                    if (!GetStream (pProp->Value.li.HighPart, RECFLD_RAWDATA, &pPropStream, FALSE))
                    {
                        HRESULT hOpError = S_OK;
                        DstList.Delete (pProp->ulPropTag, &spvOld);
                        hOpError = m_pDBobj->CreateEmptyRecord ((DWORD *)&(spvNew.Value.li.HighPart));
                        if (!hOpError)
                        {
                            hOpError = GetStream (spvNew.Value.li.HighPart, RECFLD_RAWDATA, &pNewPropStream);
                            if (!hOpError)
                            {
                                hOpError = pPropStream->CopyTo (pNewPropStream, uliSize, NULL, NULL);
                                if (!hOpError)
                                {
                                    DstList.Insert (&spvNew);
                                }
                            }
                        }
                        if (hOpError)
                        {
                            DstList.Insert (&spvOld);
                        }
                        pPropStream->Release();
                    }
                }
                else
                {
                    DstList.Insert (&spvNew);
                }
                pProp = SrcList.Walk (WALK_CONTINUE);
            }
            if (!hResult)
            {
                DstList.SaveToStream (pDst);
            }
            pDst->Release();
        }
        pSrc->Release();
    }
    if (MAPI_E_NOT_FOUND == hResult)
    {
        hResult = S_OK;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CopyMessageToMessage()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::CopyMessageToMessage (DWORD dwSrcMsgID, DWORD dwDstMsgID)
{
    LPSTREAM pSrc, pDst;
    HRESULT hResult = CopyObjectProps (dwSrcMsgID, dwDstMsgID);
    if (!hResult)
    {
        hResult = GetStream (dwSrcMsgID, RECFLD_RECIPIENTS, &pSrc, FALSE);
        if (!hResult)
        {
            hResult = GetStream (dwDstMsgID, RECFLD_RECIPIENTS, &pDst);
            if (!hResult)
            {
                CIDList SrcIDList(pSrc), DstIDList(pDst);
                DWORD dwNewRecipID, dwRecipID = SrcIDList.Walk (WALK_START);
                while (dwRecipID)
                {
                    if (!m_pDBobj->CreateEmptyRecord (&dwNewRecipID))
                    {
                        if (!CopyObjectProps (dwRecipID, dwNewRecipID))
                        {
                            DstIDList.Insert (dwNewRecipID);
                        }
                    }
                    dwRecipID = SrcIDList.Walk (WALK_CONTINUE);
                }
                DstIDList.SaveToStream (pDst);
                pDst->Release();
            }
            pSrc->Release();
        }
    }
    if (!hResult || (MAPI_E_NOT_FOUND == hResult))
    {
        hResult = GetStream (dwSrcMsgID, RECFLD_ATTACHMENTS, &pSrc, FALSE);
        if (!hResult)
        {
            hResult = GetStream (dwDstMsgID, RECFLD_ATTACHMENTS, &pDst);
            if (!hResult)
            {
                CIDList SrcIDList(pSrc), DstIDList(pDst);
                DWORD dwNewAttachID, dwAttachID = SrcIDList.Walk (WALK_START);
                ASSERT (dwAttachID);
                while (dwAttachID)
                {
                    if (!m_pDBobj->CreateEmptyRecord (&dwNewAttachID))
                    {
                        if (!CopyObjectProps (dwAttachID, dwNewAttachID))
                        {
                            DstIDList.Insert (dwNewAttachID);
                        }
                    }
                    dwAttachID = SrcIDList.Walk (WALK_CONTINUE);
                }
                DstIDList.SaveToStream (pDst);
                pDst->Release();
            }
            pSrc->Release();
        }
    }
    if (MAPI_E_NOT_FOUND == hResult)
    {
        hResult = S_OK;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::CopyMessagesInFolder()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CFileBackEnd::CopyMessagesInFolder (PRIVATE_ENTRYID   SrcFolderEID,
                                                 PRIVATE_ENTRYID   DstFolderEID,
                                                 DWORD             dwCopyFlags,
                                                 LPMAPIPROGRESS    pProgress,
                                                 POPENOBJECTS_DATA pData)
{
    // TODO: This function is not using the IMAPIPropgress interface
    //       to update the UI on the client. This needs to be done.
    //          IrvingD 2/23/96
    CIDList SrcAsscMsgsList, SrcMsgsList, DstAsscMsgsList, DstMsgsList;
    CPropList *pMsgProps;
    LPSTREAM pStream, pSrcContStream = NULL, pSrcAsscStream = NULL, pDstContStream = NULL, pDstAsscStream = NULL;
    PRIVATE_ENTRYID * pEID;
    BOOL fGotErrors = FALSE, fDeleteCopy, fPropsSaved, fParentFolderHierTableUpdated;
    LPTABLEDATA pTable;
    DWORD i, dwNewMsgID, dwMsgType, dwLastAllocated = 0;
    LPSPropValue pProps = NULL;
    SPropValue spvProps[3] = { 0 }, spvSrcCounts[3] = { 0 };
    SPropValue spvDstCounts[3] = { 0 }, spvMsgFlags = { 0 }, spvIndex = { 0 };
    spvIndex.ulPropTag = PR_INSTANCE_KEY;
    spvIndex.Value.bin.cb = sizeof(DWORD);

    HRESULT hOpError, hResult = GetStream (SrcFolderEID.dwObjID, RECFLD_ASSOCIATED, &pSrcAsscStream, FALSE);
    if (!hResult || (MAPI_E_NOT_FOUND == hResult))
    {
        SrcAsscMsgsList.LoadFromStream (pSrcAsscStream);
        hResult = GetStream (SrcFolderEID.dwObjID, RECFLD_CONTENTS, &pSrcContStream, FALSE);
        if (!hResult || (MAPI_E_NOT_FOUND == hResult))
        {
            SrcMsgsList.LoadFromStream (pSrcContStream);
            hResult = GetStream (DstFolderEID.dwObjID, RECFLD_ASSOCIATED, &pDstAsscStream);
            if (!hResult)
            {
                DstAsscMsgsList.LoadFromStream (pDstAsscStream);
                GetStream (DstFolderEID.dwObjID, RECFLD_CONTENTS, &pDstContStream);
                if (!hResult)
                {
                    DstMsgsList.LoadFromStream (pDstContStream);
                }
            }
        }
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    
    CMAPIFolder *pSrcFld, *pDstFld;
    pSrcFld = (CMAPIFolder *)pData->pOpenObjList->Walk (SrcFolderEID, WALK_START, NULL);
    pDstFld = (CMAPIFolder *)pData->pOpenObjList->Walk (DstFolderEID, WALK_START, NULL);
    ASSERT (pSrcFld);
    ASSERT (pDstFld);
    pSrcFld->m_pList->Find (PR_CONTENT_COUNT, &spvSrcCounts[0]);
    pSrcFld->m_pList->Find (PR_CONTENT_UNREAD, &spvSrcCounts[1]);
    pSrcFld->m_pList->Find (PR_ASSOC_CONTENT_COUNT, &spvSrcCounts[2]);

    pDstFld->m_pList->Find (PR_CONTENT_COUNT, &spvDstCounts[0]);
    pDstFld->m_pList->Find (PR_CONTENT_UNREAD, &spvDstCounts[1]);
    pDstFld->m_pList->Find (PR_ASSOC_CONTENT_COUNT, &spvDstCounts[2]);

    pSrcFld->Release();
    pDstFld->Release();   

    SYSTEMTIME st;
    GetSystemTime (&st);
    spvProps[0].ulPropTag = PR_CREATION_TIME;
    SystemTimeToFileTime (&st, &spvProps[0].Value.ft);
    spvProps[1].ulPropTag = PR_LAST_MODIFICATION_TIME;
    spvProps[1].Value.ft = spvProps[0].Value.ft;
    spvProps[2].ulPropTag = PR_PARENT_ENTRYID;
    spvProps[2].Value.bin.cb = CB_PRIVATE_EID;
    PRIVATE_ENTRYID MsgCopyEID;

    for (i=0; i<pData->pSrcEIDList->cValues; i++)
    {
        // Reset these variables in each iteration
        dwNewMsgID = 0;
        hOpError = S_OK;
        fDeleteCopy = FALSE;
        pMsgProps = NULL;
        pEID = (PRIVATE_ENTRYID *)pData->pSrcEIDList->lpbin[i].lpb;
        if (pEID && CB_PRIVATE_EID == pData->pSrcEIDList->lpbin[i].cb && IsValidEntryID (pEID))
        {
            MsgCopyEID = *pEID;
            // Search for the ID in the list of ID's of associated messages
            if (SrcAsscMsgsList.IsInList (pEID->dwObjID))
            {
                dwMsgType = OBJTYPE_ASSOCIATED;
            }
            else
            {
                ASSERT (SrcMsgsList.IsInList (pEID->dwObjID));
                dwMsgType = OBJTYPE_NORMAL;
            }
            if (COPYFLAG_COPY & dwCopyFlags)
            {
                // If we are copying a message, we must create a brand new message which requires a new unique ID
                hOpError = m_pDBobj->CreateEmptyRecord (&dwNewMsgID);
                if (!hOpError)
                {
                    hOpError = CopyMessageToMessage (pEID->dwObjID, dwNewMsgID);
                    if (!hOpError)
                    {
                        // Set this flag in case we fail down the line
                        fDeleteCopy = TRUE;
                    }
                }
            }
            if (!hOpError)
            {
                DWORD dwMsgID;
                if (COPYFLAG_MOVE & dwCopyFlags)
                {
                    dwMsgID = pEID->dwObjID;
                }
                if (COPYFLAG_COPY & dwCopyFlags)
                {
                    dwMsgID = dwNewMsgID;
                }
                hOpError = GetStream (dwMsgID, RECFLD_PROPERTIES, &pStream, FALSE);
            }
            if (!hOpError)
            {
                // So far so good. Now we must set some unique properties in the copied or moved message
                hOpError = LoadProperties (pStream, &pMsgProps);
                if (!hOpError)
                {
                    // For copied messages, since the targets are brand new, we must change some
                    // properties in them. In specific PR_CREATION_TIME, PR_LAST_MODIFICATION_TIME,
                    // and PR_PARENT_ENTRYID.
                    // On moved messages, only PR_PARENT_ENTRYID changes.
                    if (COPYFLAG_COPY & dwCopyFlags)
                    {
                        pMsgProps->Insert (&spvProps[0]);
                        pMsgProps->Insert (&spvProps[1]);
                    }
                    spvProps[2].Value.bin.lpb = (LPBYTE)&DstFolderEID;
                    pMsgProps->Insert (&spvProps[2]);
                    // Find out if the message we are copying or moving has been read or not
                    if (S_OK == pMsgProps->Find (PR_MESSAGE_FLAGS, &spvMsgFlags))
                    {
                        if (!(spvMsgFlags.Value.l & MSGFLAG_READ))
                        {
                            dwMsgType |= OBJTYPE_UNREAD;
                        }
                    }
                    hOpError = pMsgProps->SaveToStream (pStream);
                }
                pStream->Release();
            }
            if (hOpError && fDeleteCopy)
            {
                ASSERT (dwNewMsgID);
                m_pDBobj->DeleteRecord (dwNewMsgID);
            }
        }
        else
        {
            hOpError = MAPI_E_INVALID_ENTRYID;
        }
        if (hOpError)
        {
            fGotErrors = TRUE;
            TraceResult ("CFileBackEnd::CopyMessagesInFolder: An error occured while copying a message", hOpError);
            pData->pSrcEIDList->lpbin[i].cb = FALSE; // To indicate the this entry was NOT copied/moved
        }
        else
        {
            spvIndex.Value.bin.lpb = (LPBYTE)&(pEID->dwObjID);
            if (COPYFLAG_MOVE & dwCopyFlags)
            {
                if (OBJTYPE_ASSOCIATED & dwMsgType)
                {
                    if (spvSrcCounts[2].Value.l) { spvSrcCounts[2].Value.l--; }
                    SrcAsscMsgsList.Delete (pEID->dwObjID);
                }
                else
                {   
                    if (spvSrcCounts[0].Value.l) { spvSrcCounts[0].Value.l--; }
                    if (OBJTYPE_UNREAD & dwMsgType)
                    {
                        if (spvSrcCounts[1].Value.l) { spvSrcCounts[1].Value.l--; }
                    }
                    SrcMsgsList.Delete (pEID->dwObjID);
                }
            }
            if (COPYFLAG_COPY & dwCopyFlags)
            {
                MsgCopyEID.dwObjID = dwNewMsgID;
                pEID = &MsgCopyEID;
            }
            if (OBJTYPE_ASSOCIATED & dwMsgType)
            {
                spvDstCounts[2].Value.l++;
                DstAsscMsgsList.Insert (pEID->dwObjID);
            }
            else
            {
                spvDstCounts[0].Value.l++;
                if (OBJTYPE_UNREAD & dwMsgType)
                {
                    spvDstCounts[1].Value.l++;
                }
                DstMsgsList.Insert (pEID->dwObjID);
            }
            pSrcFld = (CMAPIFolder *)pData->pOpenObjList->Walk (SrcFolderEID, WALK_START, NULL);
            while (pSrcFld && (COPYFLAG_MOVE & dwCopyFlags))
            {
                if (OBJTYPE_ASSOCIATED & dwMsgType)
                {
                    pTable = pSrcFld->m_pAsscContTable;
                }
                else
                {
                    pTable = pSrcFld->m_pContTable;
                }
                if (pTable)
                {
                    pTable->HrDeleteRow (&spvIndex);
                }
                pSrcFld->Release();
                pSrcFld = (CMAPIFolder *)pData->pOpenObjList->Walk (SrcFolderEID, WALK_CONTINUE, pSrcFld);
            }

            pDstFld = (CMAPIFolder *)pData->pOpenObjList->Walk (DstFolderEID, WALK_START, NULL);
            while (pDstFld)
            {
                if (OBJTYPE_ASSOCIATED & dwMsgType)
                {
                    pTable = pSrcFld->m_pAsscContTable;
                }
                else
                {
                    pTable = pDstFld->m_pContTable;
                }
                if (pTable)
                {
                    GetContentsTableProps (pMsgProps,
                                           &pProps,
                                           *pEID,
                                           pData->lDstAccessLevel,
                                           pData->pStoreEID,
                                           dwLastAllocated,
                                           pTable);
                }
                pDstFld->Release();
                pDstFld = (CMAPIFolder *)pData->pOpenObjList->Walk (DstFolderEID, WALK_CONTINUE, pDstFld);
            }
        }
        delete pMsgProps;
    }
    PrivFree ((LPVOID *)&pProps);
    if (fGotErrors)
    {
        hResult = MAPI_W_PARTIAL_COMPLETION;
    }
    fPropsSaved = fParentFolderHierTableUpdated = FALSE;
    pSrcFld = (CMAPIFolder *)pData->pOpenObjList->Walk (SrcFolderEID, WALK_START, NULL);
    while (pSrcFld)
    {
        pSrcFld->m_pList->Insert (&spvSrcCounts[0]);
        pSrcFld->m_pList->Insert (&spvSrcCounts[1]);
        pSrcFld->m_pList->Insert (&spvSrcCounts[2]);
        if (!fPropsSaved)
        {
            if (S_OK == GetStream (SrcFolderEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE))
            {
                if (S_OK == pSrcFld->m_pList->SaveToStream (pStream))
                {
                    fPropsSaved = TRUE;
                }
                pStream->Release();
            }
        }
        if (!fParentFolderHierTableUpdated)
        {    
            pSrcFld->UpdateParentTable();
            fParentFolderHierTableUpdated = TRUE;
        }
        pSrcFld->Release();
        pSrcFld = (CMAPIFolder *)pData->pOpenObjList->Walk (SrcFolderEID, WALK_CONTINUE, pSrcFld);
    }
    fPropsSaved = fParentFolderHierTableUpdated = FALSE;
    pDstFld = (CMAPIFolder *)pData->pOpenObjList->Walk (DstFolderEID, WALK_START, NULL);
    while (pDstFld)
    {
        pDstFld->m_pList->Insert (&spvDstCounts[0]);
        pDstFld->m_pList->Insert (&spvDstCounts[1]);
        pDstFld->m_pList->Insert (&spvDstCounts[2]);
        if (!fPropsSaved)
        {
            if (S_OK == GetStream (DstFolderEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE))
            {
                if (S_OK == pDstFld->m_pList->SaveToStream (pStream))
                {
                    fPropsSaved = TRUE;
                }
                pStream->Release();
            }
        }
        if (!fParentFolderHierTableUpdated)
        {    
            pDstFld->UpdateParentTable();
            fParentFolderHierTableUpdated = TRUE;
        }
        pDstFld->Release();
        pDstFld = (CMAPIFolder *)pData->pOpenObjList->Walk (DstFolderEID, WALK_CONTINUE, pSrcFld);
    }

    DstAsscMsgsList.SaveToStream (pDstAsscStream);
    DstMsgsList.SaveToStream (pDstContStream);
    SrcAsscMsgsList.SaveToStream (pSrcAsscStream);
    SrcMsgsList.SaveToStream (pSrcContStream);

ErrorExit:
    if (pDstContStream) { pDstContStream->Release(); }
    if (pDstAsscStream) { pDstAsscStream->Release(); }
    if (pSrcContStream) { pSrcContStream->Release(); }
    if (pSrcAsscStream) { pSrcAsscStream->Release(); }
    TraceResult ("CFileBackEnd::CopyMessagesInFolder", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteFolderContents()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteFolderContents (PRIVATE_ENTRYID     FolderEID,
                                                 LPMAPIPROGRESS      pProgress,
                                                 POPENOBJECTS_DATA   pData)
{
    // TODO: This function is not using the IMAPIPropgress interface
    //       to update the UI on the client. This needs to be done.
    //          IrvingD 2/23/96
    CIDList AsscMsgsList, MsgsList, FolderList;
    LPSTREAM pAsscStream = NULL, pContStream = NULL, pFldStream = NULL, pPropStream;
    HRESULT hOpError, hResult = GetStream (FolderEID.dwObjID, RECFLD_CONTENTS, &pContStream , FALSE);
    if (!hResult || (MAPI_E_NOT_FOUND == hResult))
    {
        MsgsList.LoadFromStream (pContStream);
        if (!(OPENOBJDATA_SKIP_ASSOCIATED & pData->dwFlags))
        {
            hResult = GetStream (FolderEID.dwObjID, RECFLD_ASSOCIATED, &pAsscStream, FALSE);
            if (!hResult || (MAPI_E_NOT_FOUND == hResult))
            {
                AsscMsgsList.LoadFromStream (pAsscStream);
            }
        }
        if (!hResult || (MAPI_E_NOT_FOUND == hResult))
        {
            hResult = GetStream (FolderEID.dwObjID, RECFLD_HIERARCHY, &pFldStream, FALSE);
            if (!hResult || (MAPI_E_NOT_FOUND == hResult))
            {
                FolderList.LoadFromStream (pFldStream);
                hResult = S_OK;
            }
        }
    }
    ULONG i = 0;
    DWORD dwObjID;
    BOOL fGotErrors = FALSE, fObjectOpen;
    if (OPENOBJDATA_USE_MSGEIDLIST & pData->dwFlags)
    {
        pData->dwCount = pData->pSrcEIDList->cValues;
    }
    else
    {
        pData->dwCount = MsgsList.GetNodeCount() +
                         FolderList.GetNodeCount();
        if (!(OPENOBJDATA_SKIP_ASSOCIATED & pData->dwFlags))
        {
            pData->dwCount += AsscMsgsList.GetNodeCount();
        }
    }

    // If there are any objects to delete in this folder, allocate a structure
    // that returns the ID of the objects deleted. This is used to construct
    // the entry ID that goes for the notifications of the deleted objects.
    if (!hResult && 0 != pData->dwCount)
    {
        hResult = PrivAllocate ((pData->dwCount) * sizeof(SUB_ENTRYID),
                                (LPVOID *)&(pData->pSubEIDList),
                                "DeleteFolderContents EID List");
    }

    // If there was an error or there is nothing to delete, get out of here
    if (hResult || 0 == pData->dwCount)
    {
        if (pAsscStream) { pAsscStream->Release(); }
        if (pContStream) { pContStream->Release(); }
        if (pFldStream)  { pFldStream->Release();  }
        TraceResult ("CFileBackEnd::DeleteFolderContents", hResult);
        return hResult;
    }

    if (OPENOBJDATA_USE_MSGEIDLIST & pData->dwFlags)
    {
        PRIVATE_ENTRYID * pEID;
        for (ULONG j=0; j<pData->pSrcEIDList->cValues; j++)
        {
            pEID = (PRIVATE_ENTRYID *)pData->pSrcEIDList->lpbin[i].lpb;
            if (pEID &&
                CB_PRIVATE_EID == pData->pSrcEIDList->lpbin[i].cb &&
                IsValidEntryID (pEID) &&
                (MAPI_FOLDER == pEID->bObject || MAPI_MESSAGE == pEID->bObject))
            {
                pData->pSubEIDList[i].bObject = pEID->bObject;
                pData->pSubEIDList[i].dwObjID = pEID->dwObjID;
                pData->pSubEIDList[i].bAssociated = FALSE;
                // Search for the ID in the list of ID's of associated messages
                if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
                {
                    if (AsscMsgsList.IsInList (pData->pSubEIDList[i].dwObjID))
                    {
                        pData->pSubEIDList[i].bAssociated = TRUE;
                    }
                }
                i++;
            }
            else
            {
                fGotErrors = TRUE;
            }
        }
    }
    else
    {
        if (!(OPENOBJDATA_SKIP_ASSOCIATED & pData->dwFlags))
        {
            dwObjID = AsscMsgsList.Walk (WALK_START);
            while (dwObjID)
            {
                pData->pSubEIDList[i].bObject = MAPI_MESSAGE;
                pData->pSubEIDList[i].bAssociated = TRUE;
                pData->pSubEIDList[i].dwObjID = dwObjID;
                i++;
                dwObjID = AsscMsgsList.Walk (WALK_CONTINUE);
            }
        }
        dwObjID = MsgsList.Walk (WALK_START);
        while (dwObjID)
        {
            pData->pSubEIDList[i].bObject = MAPI_MESSAGE;
            pData->pSubEIDList[i].bAssociated = FALSE;
            pData->pSubEIDList[i].dwObjID = dwObjID;
            i++;
            dwObjID = MsgsList.Walk (WALK_CONTINUE);
        }
        dwObjID = FolderList.Walk (WALK_START);
        while (dwObjID)
        {
            pData->pSubEIDList[i].bObject = MAPI_FOLDER;
            pData->pSubEIDList[i].dwObjID = dwObjID;
            i++;
            dwObjID = FolderList.Walk (WALK_CONTINUE);
        }
    }

    LPTABLEDATA pTable;
    SPropValue spvCounts[3] = { 0 }, spvMsgFlags = { 0 }, spvIndex = { 0 };
    spvIndex.ulPropTag = PR_INSTANCE_KEY;
    spvIndex.Value.bin.cb = sizeof(DWORD);
    
    CMAPIFolder *pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (FolderEID, WALK_START, NULL);
    if (pFldObj)
    {
        pFldObj->m_pList->Find (PR_CONTENT_COUNT, &spvCounts[0]);
        pFldObj->m_pList->Find (PR_CONTENT_UNREAD, &spvCounts[1]);
        pFldObj->m_pList->Find (PR_ASSOC_CONTENT_COUNT, &spvCounts[2]);
        pFldObj->Release();
    }
    DWORD dwObjType;
    CPropList *pProps;
    for (i=0; i<pData->dwCount; i++)
    {
        hOpError = S_OK;
        dwObjID = pData->pSubEIDList[i].dwObjID;
        if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject &&
            pData->pSubEIDList[i].bAssociated)
        {
            dwObjType = OBJTYPE_ASSOCIATED;
        }
        else
        {
            dwObjType = OBJTYPE_NORMAL;
        }

        // If the message is a regular message, check its properties
        // and find out if it has been read or not
        if (OBJTYPE_NORMAL & dwObjType && MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
        {
            hOpError = GetStream (dwObjID, RECFLD_PROPERTIES, &pPropStream);
            if (!hOpError)
            {
                hOpError = LoadProperties (pPropStream, &pProps);
                pPropStream->Release();
                if (!hOpError)
                {
                    {
                        hOpError = pProps->Find (PR_MESSAGE_FLAGS, &spvMsgFlags);
                        if (!hOpError)
                        {
                            if (!(spvMsgFlags.Value.l & MSGFLAG_READ))
                            {
                                dwObjType |= OBJTYPE_UNREAD;
                            }
                        }
                    }
                    delete pProps;
                }
            }
        }
        if (MAPI_FOLDER == pData->pSubEIDList[i].bObject)
        {
            PRIVATE_ENTRYID SubFolderEID = FolderEID;
            SubFolderEID.dwObjID = pData->pSubEIDList[i].dwObjID;
            OPENOBJECTS_DATA SubData = { 0 };
            SubData.pOpenObjList = pData->pOpenObjList;
            SubData.dwFlags = OPENOBJDATA_SUB_CONTAINER;
            hOpError = DeleteFolderContents (SubFolderEID, NULL, &SubData);
            if (!hOpError && SubData.pSubEIDList)
            {
                ASSERT (SubData.dwCount);
                ASSERT (SubFolderEID.dwObjID);

                // When ever a subfolder is deleted, we need to save the information of the
                // objects that were deleted so that we may pass this to the caller which
                // in turn generates the notification fnevObjectDeleted for every
                // subobject deleted.
                if (pData->pSubFldInfo)
                {
                    DWORD dwNewSize = sizeof(SUBFOLDERS_DELETED_INFO) * (pData->dwSubFldCount + 1);
                    LPVOID pVoid = (LPVOID)pData->pSubFldInfo;
                    pData->pSubFldInfo = (SUBFOLDERS_DELETED_INFO *)HeapReAlloc (GetProcessHeap(),
                                                                                 HEAP_ZERO_MEMORY,
                                                                                 pVoid,
                                                                                 dwNewSize);
                    if (pData->pSubFldInfo)
                    {
                        pData->pSubFldInfo[pData->dwSubFldCount].dwFldObjID = SubFolderEID.dwObjID;
                        pData->pSubFldInfo[pData->dwSubFldCount].dwCount = SubData.dwCount;
                        pData->pSubFldInfo[pData->dwSubFldCount].pSubEIDList = SubData.pSubEIDList;
                        pData->dwSubFldCount++;
                    }
                    else
                    {
                        pData->pSubFldInfo = (SUBFOLDERS_DELETED_INFO *)pVoid;
                        PrivFree ((LPVOID *)&(SubData.pSubEIDList));
                    }
                }
                else
                {
                    pData->pSubFldInfo = (SUBFOLDERS_DELETED_INFO *)HeapAlloc (GetProcessHeap(),
                                                                               HEAP_ZERO_MEMORY,
                                                                               sizeof(SUBFOLDERS_DELETED_INFO));
                    if (pData->pSubFldInfo)
                    {
                        pData->pSubFldInfo[0].dwFldObjID = SubFolderEID.dwObjID;
                        pData->pSubFldInfo[0].dwCount = SubData.dwCount;
                        pData->pSubFldInfo[0].pSubEIDList = SubData.pSubEIDList;
                        pData->dwSubFldCount = 1;
                    }
                    else
                    {
                        PrivFree ((LPVOID *)&(SubData.pSubEIDList));
                    }
                }
            }
            if (hOpError)
            {
                fGotErrors = TRUE;
                hOpError = S_OK;
            }
        }
        if (!hOpError)
        {
            // If the object that we are deleting is currently open we can't delete
            // it's storage until the message is closed because the user can still do
            // a CopyTo, OpenProperty, OpenAttach, or other calls that requires access
            // to the file. If open, we simply mark it so that the object knows that
            // it must delete itself when its destructor get called.
            fObjectOpen = FALSE;
            if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
            {
                CMessage * pMsgObj = (CMessage *)pData->pOpenObjList->Walk (pData->pSubEIDList[i], WALK_START, NULL);
                while (pMsgObj)
                {
                    fObjectOpen = TRUE;
                    pMsgObj->m_ulObjFlags |= OBJECT_DELETED_IN_CONTAINER;
                    pMsgObj->Release();
                    pMsgObj = (CMessage *)pData->pOpenObjList->Walk (pData->pSubEIDList[i], WALK_CONTINUE, pMsgObj);
                }
            }
            else
            {
                pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (pData->pSubEIDList[i], WALK_START, NULL);
                while (pFldObj)
                {
                    fObjectOpen = TRUE;
                    pFldObj->m_ulObjFlags |= OBJECT_DELETED_IN_CONTAINER;
                    pFldObj->Release();
                    pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (pData->pSubEIDList[i], WALK_CONTINUE, pFldObj);
                }
            }
            if (!fObjectOpen)
            {
                // If the object we are deleting is an IMessage, we must deltet
                if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
                {
                    DeleteMessageAttachmentsStreams (dwObjID);
                    DeleteMessageRecipientsStreams (dwObjID);
                    DeleteLargePropsStreams (dwObjID);
                }
                hOpError = m_pDBobj->DeleteRecord (dwObjID);
            }
        }
        if (hOpError)
        {
            fGotErrors = TRUE;
            TraceResult ("CFileBackEnd::DeleteFolderContents: An error occured while deleting a message", hOpError);
            pData->pSubEIDList[i].dwObjID = 0; // To indicate the this entry was NOT deleted
        }
        else
        {
            spvIndex.Value.bin.lpb = (LPBYTE)&(dwObjID);
            if (OBJTYPE_ASSOCIATED & dwObjType)
            {
                ASSERT (MAPI_MESSAGE == pData->pSubEIDList[i].bObject);
                AsscMsgsList.Delete (dwObjID);
                if (spvCounts[2].Value.l) { spvCounts[2].Value.l--; }
            }
            else
            {
                if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
                {
                    if (spvCounts[0].Value.l) { spvCounts[0].Value.l--; }
                    if (OBJTYPE_UNREAD & dwObjType)
                    {
                        if (spvCounts[1].Value.l) { spvCounts[1].Value.l--; }
                    }
                    MsgsList.Delete (dwObjID);
                }
            }
            pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (FolderEID, WALK_START, NULL);
            while (pFldObj)
            {
                if (MAPI_MESSAGE == pData->pSubEIDList[i].bObject)
                {
                    if (OBJTYPE_ASSOCIATED & dwObjType)
                    {
                        pTable = pFldObj->m_pAsscContTable;
                    }
                    else
                    {
                        pTable = pFldObj->m_pContTable;
                    }
                }
                else
                {
                    pTable = pFldObj->m_pHierTable;
                }
                if (pTable)
                {
                    pTable->HrDeleteRow (&spvIndex);
                }
                pFldObj->Release();
                pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (FolderEID, WALK_CONTINUE, pFldObj);
            }
        }
    }
    if (fGotErrors)
    {
        hResult = MAPI_W_PARTIAL_COMPLETION;
    }
    BOOL fPropsSaved = FALSE, fParentFolderHierTableUpdated = FALSE;
    pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (FolderEID, WALK_START, NULL);
    while (pFldObj)
    {
        pFldObj->m_pList->Insert (&spvCounts[0]);
        pFldObj->m_pList->Insert (&spvCounts[1]);
        pFldObj->m_pList->Insert (&spvCounts[2]);
        if (!fPropsSaved)
        {
            if (S_OK == GetStream (FolderEID.dwObjID, RECFLD_PROPERTIES, &pPropStream, FALSE))
            {
                if (S_OK == pFldObj->m_pList->SaveToStream (pPropStream))
                {
                    fPropsSaved = TRUE;
                }
                pPropStream->Release();
            }
        }
        if (!fParentFolderHierTableUpdated)
        {    
            pFldObj->UpdateParentTable();
            fParentFolderHierTableUpdated = TRUE;
        }
        pFldObj->Release();
        pFldObj = (CMAPIFolder *)pData->pOpenObjList->Walk (FolderEID, WALK_CONTINUE, pFldObj);
    }
    AsscMsgsList.SaveToStream (pAsscStream);
    MsgsList.SaveToStream (pContStream);
    FolderList.SaveToStream (pFldStream);

    if (pAsscStream) { pAsscStream->Release(); }
    if (pContStream) { pContStream->Release(); }
    if (pFldStream)  { pFldStream->Release();  }

    // If this is the TOP level folder for which the function was called (as opposed to one
    // of the recurses reentrancies) update the folder props
    if (!(pData->dwFlags & OPENOBJDATA_SUB_CONTAINER))
    {
        LPSTREAM pStream;
        if (!GetStream (FolderEID.dwObjID, RECFLD_HIERARCHY, &pStream))
        {
            CIDList FolderIDList (pStream);
            pStream->Release();
            // If this is the only folder in the parent we must alter a property in the
            // folder which tells wheather or not the folder has subfolder. If we modify
            // the folder, then we must also update the "last modified time" property.
            if (FolderIDList.IsEmpty())
            {
                if (!GetStream (FolderEID.dwObjID, RECFLD_PROPERTIES, &pStream, FALSE))
                {
                    CPropList List(pStream);
                    SPropValue spvProps[2] = { 0 };
                    spvProps[0].ulPropTag = PR_SUBFOLDERS;
                    spvProps[0].Value.b = FALSE;
                    List.Insert (&spvProps[0]);
                    spvProps[1].ulPropTag = PR_LAST_MODIFICATION_TIME;
                    SYSTEMTIME st;
                    GetSystemTime (&st);
                    SystemTimeToFileTime (&st, &spvProps[1].Value.ft);
                    List.Insert (&spvProps[1]);
                    List.SaveToStream (pStream);
                    pStream->Release();
                }
            }
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteOneObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteOneObject (PRIVATE_ENTRYID ObjectEID,
                                            PRIVATE_ENTRYID ParentEID)
{
    HRESULT hResult = S_OK;
    LockDataBase();
    switch (ObjectEID.bObject)
    {
        case 0 : // Undefined type used to delete property streams and storages
            hResult = m_pDBobj->DeleteRecord (ObjectEID.dwObjID);
            break;
        case MAPI_FOLDER :
            hResult = DeleteFolderStorage (ParentEID.dwObjID, ObjectEID.dwObjID);
            break;
        case MAPI_MESSAGE :
            DeleteMessageAttachmentsStreams (ObjectEID.dwObjID);
            DeleteMessageRecipientsStreams (ObjectEID.dwObjID);
            DeleteLargePropsStreams (ObjectEID.dwObjID);
            m_pDBobj->DeleteRecord (ObjectEID.dwObjID);
            break;
        default :
            ASSERTNOW();
            hResult = E_FAIL;
            break;
    }
    UnlockDataBase();
    TraceResult ("CFileBackEnd::DeleteOneObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::GetNamedPropertiesStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::GetNamedPropertiesStream (LPSTREAM * ppStream)
{
    return m_pDBobj->OpenRecordField (NAMED_PROPERTIES_MAPPING,
                                      TRUE,
                                      RECFLD_NAMED_PROPS,
                                      ppStream);
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteMessageRecipientsStreams()
//
//    Parameters
//      dwMsgID     ID in the database of the message object whose recipient
//                  table we are going to wipe out.
//
//    Purpose
//      Delete the data storage associated with the recipient table
//      of a message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteMessageRecipientsStreams (DWORD dwMsgID)
{
    LPSTREAM pStream;
    // In the database, the stream of recipients in a message record is a
    // list of IDs pointing to other records in the database where the
    // properties for each recipients are saved. The properties in the
    // recipient table cannot be large so we don't
    // have to call DeleteLargePropsStreams()
    HRESULT hResult = GetStream (dwMsgID, RECFLD_RECIPIENTS, &pStream, FALSE);
    if (!hResult)
    {
        CIDList RecipientsIDList (pStream);
        pStream->Release();
        DWORD dwRecipientID = RecipientsIDList.Walk (WALK_START);
        while (dwRecipientID)
        {
            m_pDBobj->DeleteRecord (dwRecipientID);
            dwRecipientID = RecipientsIDList.Walk (WALK_CONTINUE);
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteMessageAttachmentsStreams()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteMessageAttachmentsStreams (DWORD dwMsgID)
{
    LPSTREAM pStream;
    HRESULT hResult = GetStream (dwMsgID, RECFLD_ATTACHMENTS, &pStream, FALSE);
    if (!hResult)
    {
        CIDList AttachIDList (pStream);
        pStream->Release();
        DWORD dwAttachID = AttachIDList.Walk (WALK_START);
        while (dwAttachID)
        {
            DeleteLargePropsStreams (dwAttachID);
            m_pDBobj->DeleteRecord (dwAttachID);
            dwAttachID = AttachIDList.Walk (WALK_CONTINUE);
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteLargePropsStreams()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteLargePropsStreams (DWORD dwObjID)
{
    LPSTREAM pStream;
    HRESULT hResult = GetStream (dwObjID, RECFLD_PROPERTIES, &pStream, FALSE);
    if (!hResult)
    {
        CPropList List(pStream);
        pStream->Release();
        PFlatSPropValue pProp = List.Walk (WALK_START);
        while (pProp)
        {
            // If the type of the property is our custom type, the data is
            // in the LARGE_INTEGER member of the VALUE structure. The
            // highpart of the value is a DWORD which is the ID of a record
            // in the database storing the RAW data of the IStream or IStorage
            // property. We must delete this record.
            if (PT_STREAM_STORAGE == PROP_TYPE (pProp->ulPropTag))
            {
                m_pDBobj->DeleteRecord (pProp->Value.li.HighPart);
            }
            pProp = List.Walk (WALK_CONTINUE);
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::DeleteFolderStorage()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::DeleteFolderStorage (DWORD dwParentFldObjID, DWORD dwFldObjID)
{
    LPSTREAM pStream;
    // When a folder is deleted we must delete its name from the list of names in the
    // parent folder, so that a new folder with the name of the deleted folder, may be created.
    HRESULT hResult = DeleteFolderNameInParent (dwParentFldObjID, dwFldObjID);
    // We must delete the ID of the deleted folder from the list of ID of subfolders in the
    // parent folder. This list of IDs is used in the parent folder to construct
    // the hierarchy table.
    hResult = GetStream (dwParentFldObjID, RECFLD_HIERARCHY, &pStream);
    if (!hResult)
    {
        CIDList FolderIDList (pStream);
        FolderIDList.Delete (dwFldObjID);
        FolderIDList.SaveToStream (pStream);
        pStream->Release();
        // If this is the only folder in the parent we must alter a property in the parent
        // folder which tells weather or not the parent folder has subfolder. If we modify
        // the parent folder, then we must also update the "last modified time" property.
        if (FolderIDList.IsEmpty())
        {
            hResult = GetStream (dwParentFldObjID, RECFLD_PROPERTIES, &pStream, FALSE);
            if (!hResult)
            {
                CPropList List(pStream);
                SPropValue spvProps[2] = { 0 };
                spvProps[0].ulPropTag = PR_SUBFOLDERS;
                spvProps[0].Value.b = FALSE;
                List.Insert (&spvProps[0]);
                spvProps[1].ulPropTag = PR_LAST_MODIFICATION_TIME;
                SYSTEMTIME st;
                GetSystemTime (&st);
                SystemTimeToFileTime (&st, &spvProps[1].Value.ft);
                List.Insert (&spvProps[1]);
                List.SaveToStream (pStream);
                pStream->Release();
            }
        }
    }
    m_pDBobj->DeleteRecord (dwFldObjID);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CFileBackEnd::SetObjNameInContainer()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CFileBackEnd::SetObjNameInContainer (LPTSTR             pszNewFolderName,
                                                  PRIVATE_ENTRYID    ParentFolderEID,
                                                  PRIVATE_ENTRYID    FolderEID)
{
    DWORD dwFolderID = FolderEID.dwObjID;
    LockDataBase();
    HRESULT hResult = CheckFldNameCollision (ParentFolderEID.dwObjID, 
                                             pszNewFolderName,
                                             &dwFolderID,
                                             FLDNAME_CHECK);
    if (MAPI_E_COLLISION == hResult)
    {
        // If we had a collision because the folders are 
        if (dwFolderID == FolderEID.dwObjID)
        {
            hResult = S_OK;
        }
    }
    else
    {
        hResult = CheckFldNameCollision (ParentFolderEID.dwObjID, 
                                         pszNewFolderName,
                                         &dwFolderID,
                                         FLDNAME_REPLACE);
    }
    UnlockDataBase();
    return hResult;
}

// End of file for CFILESTG.CPP