///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      GENPROP.CPP
//
//  Description
//      This file implements the CGeneric and CGenericProp classes used as
//      base classes for interfaces that support the IMAPIProp interfaces.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "GENPROP.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceMessage(a)
#define InfoTraceRaw(a)     TraceRaw(a)
#define InfoTraceProp(a)    TraceProp(a)
#else
#define InfoTrace(a)
#define InfoTraceRaw(a)
#define InfoTraceProp(a)
#endif // DO_INFO_TRACES

typedef struct _OBJDATA
{
    ULONG               ulObjFlags;
    long                lAccess;
    CPropList *         pList;
    COpenPropList **    ppOpenPropList;
    CBackEnd *          pBEobj;
    PRIVATE_ENTRYID *   pObjectEID;
    PRIVATE_ENTRYID *   pParentEID;
} OBJDATA, *POBJDATA;

///////////////////////////////////////////////////////////////////////////////
//    CGeneric::CGeneric()
//
//    Parameters
//      None
//
//    Purpose
//      None
//
//    Return Value
//      None
//
CGeneric::CGeneric()
{
    InfoTrace ("CGeneric: Constructor called");
    m_cRef            = 1;
    m_hLastError      = S_OK;
    m_ctxLastError    = NONE_ERROR;
    InitializeCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CGeneric::~CGeneric()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CGeneric.
//
//    Return Value
//      None
//
CGeneric::~CGeneric()
{
    InfoTrace ("CGeneric: Destructor called");
    DeleteCriticalSection (&m_csObj);
    m_cRef         = 0;
    m_hLastError   = S_OK;
    m_ctxLastError = NONE_ERROR;
}

///////////////////////////////////////////////////////////////////////////////
//    CGeneric::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Based on the error passed in, this function loads the full description
//      of the error. The function allocates a MAPIERROR structure and the
//      memory for the members in it. A client is responsible for releasing
//      this memory with the MAPI allocation functions.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGeneric::GetLastError (HRESULT        hError,
                                     ULONG          ulFlags,
                                     LPMAPIERROR *  ppMAPIError,
                                     HINSTANCE      hInstance,
                                     LPTSTR         pszComponent)
{
    // The following array maps a string identifier (IDS) to an hResult
    // The order of HRESULTs in the array has an external dependency:
    // the order of elements in the array is dictated by the IDS definitions
    // in RESOURCE.H. This implicit association must be maintained for the 
    // strings associated with string identifiers to make sense. Thus, if 
    // either this structure or the RESOURCE.H defines changes, the other 
    // must change to match it
    static const HRESULT MapOfhResults[] =
    {
    	S_OK,

    	E_FAIL,
    	E_OUTOFMEMORY,
    	E_INVALIDARG,
    	E_NOINTERFACE,
    	E_ACCESSDENIED,

        MAPI_E_BUSY,
    	MAPI_E_NO_SUPPORT,
    	MAPI_E_NOT_FOUND,
    	MAPI_E_UNKNOWN_FLAGS,
    	MAPI_E_VERSION,
    	MAPI_E_INVALID_ENTRYID,
        MAPI_E_USER_CANCEL,
        MAPI_E_DISK_ERROR,
        MAPI_E_CORRUPT_DATA,
        MAPI_E_BAD_CHARWIDTH,
        MAPI_E_UNCONFIGURED,
        MAPI_E_COMPUTED,
        MAPI_E_DECLINE_COPY,
        MAPI_E_INVALID_OBJECT,
        MAPI_E_OBJECT_DELETED,
        MAPI_E_CORRUPT_STORE,
        MAPI_E_NOT_IN_QUEUE,
        MAPI_E_NO_SUPPRESS,
        MAPI_E_COLLISION,
        MAPI_E_NOT_INITIALIZED,
        MAPI_E_NON_STANDARD,
        MAPI_E_NO_RECIPIENTS,
        MAPI_E_SUBMITTED,
        MAPI_E_HAS_FOLDERS,
        MAPI_E_HAS_MESSAGES,
        MAPI_E_FOLDER_CYCLE,
        MAPI_E_EXTENDED_ERROR,
        MAPI_E_NETWORK_ERROR,
        MAPI_W_PARTIAL_COMPLETION,
        MAPI_W_ERRORS_RETURNED,
        HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE),
        HRESULT_FROM_WIN32(ERROR_NO_SUCH_USER),
        HRESULT_FROM_WIN32(ERROR_WRONG_PASSWORD),
        HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORD),
        HRESULT_FROM_WIN32(ERROR_BAD_USERNAME),
        HRESULT_FROM_WIN32(ERROR_INVALID_ACCOUNT_NAME),
        HRESULT_FROM_WIN32(ERROR_BAD_NET_NAME),
        HRESULT_FROM_WIN32(ERROR_SERVICE_NO_THREAD),
        HRESULT_FROM_WIN32(ERROR_SHUTDOWN_IN_PROGRESS),
        HRESULT_FROM_WIN32(ERROR_CONNECTION_INVALID),
        HRESULT_FROM_WIN32(ERROR_ALREADY_EXISTS)
    };

    // See if we have the string for the error. If so, make a copy 
    // and pass it back. Search for the hResult in the HRESULT map.
    UINT uStringID = 0, uMaxNum = sizeof (MapOfhResults) / sizeof (HRESULT);
    for (UINT i=0; i<uMaxNum; i++ )
	{
		if (MapOfhResults[i] == hError)
		{
            break; // Out of the FOR() loop
		}
	}
    if (i == uMaxNum)
    {
        uStringID = IDS_E_UNKNOWN_ERROR;
    }
	else
    {
        uStringID += IDS_SUCCESS + i;
    }

    TCHAR szBuffer[256] = { 0 };
	if (0 == LoadString (hInstance, uStringID, szBuffer, 255))
    {
        lstrcpy (szBuffer, TEXT("No description is available for this error."));
    }
    // The caller is responsible for freeing this.
    HRESULT hResult = PrivMAPIAllocate (sizeof(MAPIERROR), (LPVOID *)ppMAPIError);
    if (!hResult)
    {
        (*ppMAPIError)->ulVersion = MAPI_ERROR_VERSION;
        // Allocate memory for the returning string. The memory for this is linked to the
        // memory of the structure for easier cleanup.
    	hResult = PrivMAPIAllocate (Cbtszsize(szBuffer), (LPVOID *)&(*ppMAPIError)->lpszError, *ppMAPIError); // Alloc more
    	if (!hResult)
        {
            lstrcpy ((*ppMAPIError)->lpszError, szBuffer);
            hResult = PrivMAPIAllocate (Cbtszsize (pszComponent), (LPVOID *)&(*ppMAPIError)->lpszComponent, *ppMAPIError); // Alloc more
            if (!hResult)
            {
                lstrcpy ((*ppMAPIError)->lpszComponent, pszComponent);
            }
        }
    }
    if (hResult)
    {
        PrivMAPIFree ((LPVOID *)ppMAPIError);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::CGenericProp()
//
//    Parameters
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CGenericProp::CGenericProp (CBackEnd *          pBEobj,
                            PRIVATE_ENTRYID *   pParentEID,
                            PRIVATE_ENTRYID *   pObjectEID,
                            CPropList *         pList,
                            BOOL                fDeferErrors,
                            long                lAccess) : CGeneric()
{
    InfoTrace ("CGenericProp: Constructor called");
    m_pBEobj        = pBEobj;
    m_pList         = pList;
    m_lAccess       = lAccess;
    m_ulObjFlags    = fDeferErrors ? OBJECT_DEFER_ERRORS : 0;
    m_pOpenPropList = NULL;
    m_pNPList       = NULL;

    ASSERT (m_pBEobj);
    m_pBEobj->AddRef();

    ZeroMemory (&m_ParentEID, CB_PRIVATE_EID);
    ZeroMemory (&m_ObjectEID, CB_PRIVATE_EID);
    if (pParentEID)
    {
        CopyMemory (&m_ParentEID, pParentEID, CB_PRIVATE_EID);
    }
    if (pObjectEID)
    {
        CopyMemory (&m_ObjectEID, pObjectEID, CB_PRIVATE_EID);
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::~CGenericProp()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CGenericProp.
//
//    Return Value
//      None
//
CGenericProp::~CGenericProp()
{
    InfoTrace ("CGenericProp: Destructor called");
    if (m_pBEobj)
    {
        m_pBEobj->Release();
        m_pBEobj = NULL;
    }
    if (m_pList)
    {
        delete m_pList;
        m_pList = NULL;
    }
    if (m_pOpenPropList)
    {
        delete m_pOpenPropList;
        m_pOpenPropList = NULL;
    }
    m_lAccess = 0;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::SaveChanges()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::SaveChanges (ULONG ulFlags, LPTABLEDATA pTable)
{
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        return E_ACCESSDENIED;
    }
    // Lock the object for exclusive access.
    EnterCriticalSection (&m_csObj);

    m_pList->CopySmallStreams (m_pOpenPropList);

    SPropValue spvLastModTime = { 0 };
    spvLastModTime.ulPropTag = PR_LAST_MODIFICATION_TIME;
    SYSTEMTIME st;
    GetSystemTime (&st);
    SystemTimeToFileTime (&st, &spvLastModTime.Value.ft);
    m_pList->Insert (&spvLastModTime);
    
    HRESULT hResult = m_pBEobj->SaveObject (m_ParentEID, m_ObjectEID, m_pList, m_pOpenPropList, pTable);
    if (!hResult)
    {
        // Close all the open properties that are NOT in use
        if (m_pOpenPropList)
        {
            m_pOpenPropList->FreeList (FALSE);
        }
        // Note that for the implementation of IMAPIProp for transacted objects,
        // we ignore the FORCE_SAVE flag and always close the object if the KEEP_xxx
        // flags where not passed in.
        if (ulFlags & KEEP_OPEN_READONLY)
        {
            // The user need only read access from now on.
            m_lAccess = MAPI_ACCESS_READ;
        }
        else
        {
            // If this flag was not passed in, close the object and change the access flags.
            if (!(ulFlags & KEEP_OPEN_READWRITE))
            {
                m_lAccess = 0;
            }
        }
        // The implemented object is responsible for sending a
        // notification that PR_ACCESS/PR_ACCESS_LEVEL
    }
    // Unlock the object.
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetPropList()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::GetPropList (const ULONG             ulFlags,
                                        const LPSPropTagArray   pObjectTags,
                                        LPSPropTagArray *       ppAllTags)
{
    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);

    LPSPropTagArray pListTags = NULL;
    if (m_pList)
    {
        m_pList->GetTags (&pListTags);
    }
    
    HRESULT hResult = CombineTags (pObjectTags, pListTags, ppAllTags, TRUE);
    PrivFree ((LPVOID *)&pListTags);

    SPropValue spvProp;
    for (ULONG i=0; i<(*ppAllTags)->cValues; i++)
    {
        if (PT_STREAM_STORAGE == PROP_TYPE ((*ppAllTags)->aulPropTag[i]))
        {
            m_pList->Find ((*ppAllTags)->aulPropTag[i], &spvProp);
            (*ppAllTags)->aulPropTag[i] = PROP_TAG(LOWORD(spvProp.Value.li.LowPart),
                                                   PROP_ID((*ppAllTags)->aulPropTag[i]));
        }
    }

    // Unlock the object
    LeaveCriticalSection (&m_csObj);

    // If an error occurred, save it in the object cache
    TraceResult ("GetPropList", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::GetProps (LPSPropTagArray     pPropTagArray,
                                     ULONG               ulFlags,
                                     ULONG *             pcValues,
                                     LPSPropValue *      ppPropArray,
                                     LPSPropTagArray     pObjectTags)
{
    BOOL fGetAllProperties = FALSE;
    if (!pPropTagArray)
    {
        fGetAllProperties = TRUE;
        pPropTagArray = pObjectTags;
    }
    LPSPropTagArray pAllTags, pListTags = NULL;
    if (m_pList)
    {
        m_pList->GetTags (&pListTags);
    }
    HRESULT hResult = CombineTags (pObjectTags, pListTags, &pAllTags);
    PrivFree ((LPVOID *)&pListTags);
    if (hResult)
    {
        return hResult;
    }
    if (fGetAllProperties)
    {
        *pcValues = pAllTags->cValues;
    }
    else
    {
        *pcValues = pPropTagArray->cValues;
    }
    
    // Allocate space for the requested properties
    hResult = PrivMAPIAllocate ((*pcValues)*sizeof(SPropValue), (LPVOID *)ppPropArray);
    if (hResult)
    {
        *pcValues = 0;
    }
    
    if (fGetAllProperties)
    {
        for (ULONG i=0; i<*pcValues; i++)
        {
            (*ppPropArray)[i].ulPropTag = pAllTags->aulPropTag[i];
        }
    }
    else
    {
        // Implementations of IMAPIProp should return the actual type of the property
        // request to the client. Clients that don't know the property type, set PT_UNSPECIFIED
        // in the tag, and the implementation is responsible to replace it with the actual
        // type of the property stored. If the type is not PT_UNSPECIFIED, we leave it intact.
        ULONG k, i;
        BOOL fAdded;
        for (k=0; k<pPropTagArray->cValues; k++)
        {
            fAdded = FALSE;
            for (i=0; i<pAllTags->cValues; i++)
            {
                if (PROP_ID(pPropTagArray->aulPropTag[k]) == PROP_ID(pAllTags->aulPropTag[i]))
                {
                    if (PT_UNSPECIFIED == PROP_TYPE(pPropTagArray->aulPropTag[k]))
                    {
                        (*ppPropArray)[k].ulPropTag = pAllTags->aulPropTag[i];
                        fAdded = TRUE;
                    }
                    break;
                }
            }
            if (!fAdded)
            {
                (*ppPropArray)[k].ulPropTag = pPropTagArray->aulPropTag[k];
            }
        }
    }
    PrivFree ((LPVOID *)&pAllTags);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::GetProps (LPSPropValue pProp, LPVOID pMemBlock)
{
    #define MAX_GETPROP_SIZE    16384
    HRESULT hResult = m_pList->Find (pProp->ulPropTag, pProp, FALSE, pMemBlock);
    if (!hResult && (PROP_TYPE(pProp->ulPropTag) == PT_STREAM_STORAGE))
    {
        ULONG ulOriginalTag = PROP_TAG (LOWORD(pProp->Value.li.LowPart), PROP_ID(pProp->ulPropTag));
        BOOL fLoadFromFile = TRUE;
        if (PROP_STREAM == HIWORD(pProp->Value.li.LowPart))
        {
            if (m_pOpenPropList)
            {
                hResult = m_pOpenPropList->Find (pProp->ulPropTag, pProp, pMemBlock, MAX_GETPROP_SIZE);
                if (MAPI_E_NOT_FOUND != hResult)
                {
                    fLoadFromFile = FALSE;
                }
            }
            pProp->ulPropTag = ulOriginalTag;
            if (fLoadFromFile)
            {
                hResult = m_pBEobj->OpenPropertyInMemory (m_ObjectEID, MAX_GETPROP_SIZE, pMemBlock, pProp);
            }
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::GetProps (HRESULT        hError,
                                     ULONG          cbSize,
                                     LPVOID         pSrc,
                                     LPSPropValue   pProp,
                                     LPVOID         pMemBlock)
{
    LPVOID pDst;
    HRESULT hResult = S_OK;
    if (pSrc && !hError)
    {
        if (PROP_TYPE(pProp->ulPropTag) == PT_TSTRING)
        {
            cbSize = Cbtszsize ((LPTSTR)pSrc);
        }
        hResult = PrivMAPIAllocate (cbSize, &pDst, pMemBlock); // Alloc More
        if (!hResult)
        {
            CopyMemory (pDst, pSrc, cbSize);
            if (PROP_TYPE(pProp->ulPropTag) == PT_BINARY)
            {
                pProp->Value.bin.cb = cbSize;
                pProp->Value.bin.lpb = (LPBYTE)pDst;
            }
            else
            {
                pProp->Value.lpszA = (LPSTR)pDst;
            }
        }
    }
    if (hResult)
    {
        hError = hResult;
    }
    if (hError)
    {
        pProp->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(pProp->ulPropTag));
        pProp->Value.err = hError;
    }
    return hError;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::CombineTags()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::CombineTags (LPSPropTagArray     pObjectTags,
                                        LPSPropTagArray     pListTags,
                                        LPSPropTagArray *   ppAllTags,
                                        BOOL                fUseMAPIAlloc)
{
    LPSPropTagArray pTags;
	HRESULT hResult;
    ULONG cTagsInList = pListTags ? pListTags->cValues : 0;
    ULONG cbSize = CbNewSPropTagArray(pObjectTags->cValues + cTagsInList);
    if (fUseMAPIAlloc)
    {
        hResult = PrivMAPIAllocate (cbSize, (LPVOID *)&pTags);
    }
    else
    {
        hResult = PrivAllocate (cbSize, (LPVOID *)&pTags, "Combine Tags");
    }
    if (hResult)
    {
        TraceResult ("CombineTags: Failed to allocate memory for the tags", hResult);
    	return hResult;
    }
    CopyMemory (pTags, pObjectTags, CbNewSPropTagArray(pObjectTags->cValues));
    ULONG j, k = pTags->cValues;
    BOOL fAddTag;
    for (ULONG i=0; i<cTagsInList; i++)
    {
        fAddTag = TRUE;
        j = pTags->cValues;
        while (j)
        {
            j--;
            if (pTags->aulPropTag[j] == pListTags->aulPropTag[i])
            {
                // If the property is already in the tag array, then don't add it.
                fAddTag = FALSE;
                break; // Out of the WHILE() loop
            }
        }
        if (fAddTag)
        {
            pTags->aulPropTag[k] = pListTags->aulPropTag[i];
            pTags->cValues++;
            k++;
        }
    }
    *ppAllTags = pTags;
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::DeleteProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::DeleteProps (LPSPropTagArray       pTags,
                                        LPSPropProblemArray * ppProblems)
{
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CGenericProp::DeleteProps: Cannot delete properties in a READ only object");
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    
    HRESULT hTagError, hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    __try
    {
        #ifdef DO_INFO_TRACES
        TraceMessage ("Properties to delete...");
        for (ULONG j=0; j<pTags->cValues; j++) TraceProp (pTags->aulPropTag[j]);
        #endif DO_INFO_TRACES
        for (ULONG i=0; i<pTags->cValues; i++)
        {
            if (PR_NULL == pTags->aulPropTag[i])
            {
                continue;
            }
            hTagError = GetPropAccess (pTags->aulPropTag[i]);
            if (!hTagError)
            {
                hTagError = m_pList->Delete (pTags->aulPropTag[i]);
            }
            if (hTagError && ppProblems)
            {
                AddTagToProbArray (pTags->cValues, ppProblems, i, pTags->aulPropTag[i], hTagError);
            }
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CGenericProp::DeleteProps: An exception occured. Exception code: %#08x", GetExceptionCode());
        hResult = E_FAIL;
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CGenericProp::DeleteProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::OpenProperty()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::OpenProperty (ULONG              ulPropTag,
                                         LPCIID             piid,
                                         ULONG              ulInterfaceOptions,
                                         ULONG              ulFlags,
                                         LPUNKNOWN *        ppUnk)
{
    // If the implementation of IMAPIProp does not allow writing, fail if that's the access requested.
    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & (MAPI_MODIFY | MAPI_CREATE)))
    {
        TraceMessage ("CGenericProp::OpenProperty: Cannot open properties with WRITE access in a READ only object");
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    #ifdef DO_INFO_TRACES
    TraceMessage ("CGenericProp::OpenProperty: Property to Open:");
    TraceProp (ulPropTag);
    TraceRaw ("CGenericProp::OpenProperty: Interface Requested: ");
    if (piid && *piid == IID_IStream)
    {
        TraceMessage ("IID_IStream");
    }
    else
    {
        if (piid && *piid == IID_IStorage)
        {
            TraceMessage ("IID_IStorage");
        }
        else
        {
            if (piid && *piid == IID_IMessage)
            {
                TraceMessage ("IID_IMessage");
            }
            else
            {
                if (piid && *piid == IID_IStreamDocfile)
                {
                    TraceMessage ("IID_IStreamDocfile");
                }
                else
                {
                    TraceMessage ("Unknown.");
                }
            }
        }
    }
    #endif // DO_INFO_TRACES

    // If the user requested MAPI_BEST_ACCESS, a provider should give the client the
    // maximum access level it can. In our object, if we are write-able, give the
    // user WRITE access. Is the easiest to handle for us.
    if (ulFlags & MAPI_BEST_ACCESS)
    {
        if (m_lAccess & MAPI_ACCESS_MODIFY)
        {
            ulFlags |= MAPI_MODIFY;
        }
    }
    if (ulFlags & MAPI_CREATE && !(ulFlags & MAPI_MODIFY))
    {
        TraceMessage ("CGenericProp::OpenProperty: MAPI_CREATE without MAPI_MODIFY is not allowed");
        SetLastError (E_INVALIDARG);
        return E_INVALIDARG;
    }
    if (!piid)
    {
        TraceMessage ("CGenericProp::OpenProperty: Invalid interface");
        SetLastError (E_INVALIDARG);
        return E_INVALIDARG;
    }

    LPUNKNOWN pUnk = NULL;
    BOOL fPropertyAlreadyOpen = FALSE, fDeleteInCache = FALSE;
    ULONG ulPropFlags;
    if (*piid == IID_IStream)
    {
        ulPropFlags = PROP_STREAM;
    }
    else
    {
        if (*piid == IID_IStorage)
        {
            ulPropFlags = PROP_STORAGE;
        }
        else
        {
            if (*piid == IID_IMessage)
            {
                // The class derived from CGenericProp must implement this, if it needs to.
                SetLastError (MAPI_E_NO_SUPPORT);
                return MAPI_E_NO_SUPPORT;
            }
            else
            {
                TraceMessage ("CGenericProp::OpenProperty: Unsupported interface requested");
                SetLastError (E_NOINTERFACE);
                return E_NOINTERFACE;
            }
        }
    }
    if (ulFlags & MAPI_MODIFY)
    {
        ulPropFlags |= PROP_READ_WRITE;
    }
    else
    {
        ulPropFlags |= PROP_READ_ONLY;
    }
    
    EnterCriticalSection (&m_csObj);
    DWORD dwNameID = 0;
    HRESULT hFound, hResult = S_OK;
    if (!m_pOpenPropList)
    {
        m_pOpenPropList = new COpenPropList;
        if (!m_pOpenPropList)
        {
            hResult = E_OUTOFMEMORY;
            goto ErrorExit;
        }
    }

    // Find the property in the set to see if we have this property stored in the database
    SPropValue spvProp;
    hFound = m_pList->Find (ulPropTag, &spvProp, FALSE);
    if (PROP_STREAM & ulPropFlags)
    {
        hResult = CreateStreamOnHGlobal (NULL, TRUE, (LPSTREAM *)&pUnk);
        TraceResult ("CGenericProp::OpenProperty: Failed to create memory stream", hResult);
    }
    if (PROP_STORAGE & ulPropFlags)
    {
        hResult = CreatePrivateIStorage ((CStorage **)&pUnk, ulFlags & MAPI_MODIFY);
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    if (MAPI_CREATE & ulFlags)
    {
        if (S_OK == hFound)
        {
            if (PT_STREAM_STORAGE == PROP_TYPE(spvProp.ulPropTag))
            {
                dwNameID = spvProp.Value.li.HighPart;
            }
            else
            {
                ulPropFlags |= PROP_NEVER_SAVED;
                fDeleteInCache = TRUE;
            }
        }
        else
        {
            ulPropFlags |= PROP_NEVER_SAVED;
        }
    }
    else
    {
        if (S_OK == hFound)
        {
            // What we found is a large property in the list
            if (PT_STREAM_STORAGE == PROP_TYPE(spvProp.ulPropTag))
            {
                // If the store property is of the same type, then open it from the database.
                // If the types are not the same the the property was not found.
                if (PROP_TYPE(ulPropTag) == LOWORD(spvProp.Value.li.LowPart))
                {
                    if (PROP_STREAM == HIWORD(spvProp.Value.li.LowPart) ||
                        PROP_STORAGE == HIWORD(spvProp.Value.li.LowPart))
                    {
                        ASSERT (m_pOpenPropList);
                        POPEN_PROPS pNode = m_pOpenPropList->Find (ulPropTag);
                        if (pNode)
                        {
                            ASSERT (PROP_NEVER_SAVED & pNode->ulFlags);
                            if (PROP_STREAM & pNode->ulFlags)
                            {
                                if (pNode->iface.pStream->IsStreamInUse())
                                {
                                    // Our implementation of OpenProperty does not support opening
                                    // the same property stream twice.
                                    hResult = E_ACCESSDENIED;
                                }
                                else
                                {
                                    LARGE_INTEGER liZero = { 0 };
                                    VERIFY (!pNode->iface.pStream->Seek (liZero, STREAM_SEEK_SET, NULL));
                                    fPropertyAlreadyOpen = TRUE;
                                    pUnk->Release();
                                    pUnk = (LPUNKNOWN)pNode->iface.pStream;
                                }
                            }
                            else
                            {
                                ASSERT (PROP_STORAGE & pNode->ulFlags);
                                if (pNode->iface.pStorage->IsStorageInUse())
                                {
                                    // Our implementation of OpenProperty does not support opening
                                    // the same property storage twice.
                                    hResult = E_ACCESSDENIED;
                                }
                                else
                                {
                                    fPropertyAlreadyOpen = TRUE;
                                    pUnk->Release();
                                    pUnk = (LPUNKNOWN)pNode->iface.pStorage;
                                }
                            }
                        }
                        else
                        {
                            hResult = m_pBEobj->OpenProperty (m_ObjectEID, &spvProp, &pUnk);
                        }
                    }
                    else
                    {
                        hResult = E_NOINTERFACE;
                    }
                }
                else
                {
                    hResult = MAPI_E_NOT_FOUND;
                }
            }
            else
            {
                if (PROP_STREAM & ulPropFlags)
                {
                    if (PROP_READ_WRITE & ulPropFlags)
                    {
                        fDeleteInCache = TRUE;
                        ulPropFlags |= PROP_NEVER_SAVED;
                        dwNameID = 0;
                    }
                    hResult = m_pList->CopyToStream (ulPropTag, (LPSTREAM)pUnk);
                }
                else
                {
                    // We don't allow opening an IStorage interface on a property that is not an object.
                    hResult = E_INVALIDARG;
                }
            }
        }
        else
        {
            hResult = MAPI_E_NOT_FOUND;
        }
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    // The stream object we return is wrapped in our CPropStream
    // interface so that we can find out the state of the stream
    if (PROP_STREAM & ulPropFlags && !fPropertyAlreadyOpen && (PROP_READ_WRITE & ulPropFlags))
    {
        CPropStream * pPrivStream = new CPropStream ((LPSTREAM)pUnk, ulFlags & MAPI_MODIFY);
        pUnk->Release(); // The CPropStream constructor AddRef the stream
        pUnk = NULL;     // To avoid releasing this below
        if (pPrivStream)
        {
            hResult = m_pOpenPropList->Insert (pPrivStream, ulPropFlags, dwNameID, ulPropTag);
            if (hResult)
            {
                // the object goes away...
                pPrivStream->Release();
            }
            else
            {
                if (PROP_NEVER_SAVED & ulPropFlags)
                {
                    pPrivStream->SetDirty (TRUE);
                }
                pUnk = (LPUNKNOWN)pPrivStream;
            }
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    if (PROP_STORAGE & ulPropFlags && !fPropertyAlreadyOpen && (PROP_READ_WRITE & ulPropFlags))
    {
        hResult = m_pOpenPropList->Insert (pUnk, ulPropFlags, dwNameID, ulPropTag);
        if (!hResult)
        {
            if (PROP_NEVER_SAVED & ulPropFlags)
            {
                ((CStorage *)pUnk)->SetDirty (TRUE);
            }
        }
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    // Copy the new object into the returning parameter
    (*ppUnk) = pUnk;
    (*ppUnk)->AddRef();
    pUnk = NULL;
    if (fDeleteInCache)
    {
        hFound = MAPI_E_NOT_FOUND;
        ASSERT (FALSE == fPropertyAlreadyOpen);
        m_pList->Delete (ulPropTag);
    }
    if (!hResult && !fPropertyAlreadyOpen && (MAPI_E_NOT_FOUND == hFound))
    {
        m_pList->InsertOpenProp (ulPropTag, dwNameID, (WORD)(0xF & ulPropFlags));
    }
ErrorExit:
    if (pUnk)
    {
        pUnk->Release();
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CGenericProp::OpenProperty", hResult);
	SetLastError (hResult);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::AddTagToProbArray()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::AddTagToProbArray (ULONG                 cNumberOfTags,
                                              LPSPropProblemArray * ppProbArray,
                                              UINT                  ulPropTagIndex,
                                              ULONG                 ulPropTag,
                                              HRESULT               hError)
{
    HRESULT hResult;
	if (!(*ppProbArray))
	{
		hResult = PrivMAPIAllocate (CbNewSPropProblemArray(cNumberOfTags), (LPVOID *)ppProbArray);
		if (hResult)
        {
            return hResult;
        }
        (*ppProbArray)->cProblem = 0;
	}
    ULONG ulProbIndex = (*ppProbArray)->cProblem;
    (*ppProbArray)->aProblem[ulProbIndex].ulIndex = ulPropTagIndex;
    (*ppProbArray)->aProblem[ulProbIndex].ulPropTag = ulPropTag;
    (*ppProbArray)->aProblem[ulProbIndex].scode = hError;
    (*ppProbArray)->cProblem++;
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::AddPropToList()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CGenericProp::AddPropToList (const LPSPropValue pProp, BOOL fForceIt)
{
    HRESULT hTagError = E_OUTOFMEMORY;
    if (!m_pList)
    {
        m_pList = new CPropList;
        if (!m_pList)
        {
            TraceMessage ("AddPropToList: Failed to create CPropList object");
            return hTagError;
        }
    }
    hTagError = GetPropAccess (pProp->ulPropTag);
    if (!hTagError || fForceIt)
    {
        hTagError = m_pList->Insert (pProp);
    }
    return hTagError;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetPropAccess()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::GetPropAccess  (ULONG ulPropTag)
{
    switch (ulPropTag)
    {
        /////////////////////////////////////////////////////////////////////////
        // Object properties for which specific methods exists to create or
        // retrieve from the provider.
        //
        case PR_MESSAGE_ATTACHMENTS :
        case PR_MESSAGE_RECIPIENTS :
        case PR_CONTAINER_HIERARCHY :
        case PR_CONTAINER_CONTENTS :
        case PR_FOLDER_ASSOCIATED_CONTENTS :
        case PR_RECEIVE_FOLDER_SETTINGS :
            InfoTraceRaw ("WARNING: Trying to set PT_OBJECT property. ");
            InfoTraceProp (ulPropTag);
            return E_INVALIDARG;
        
        /////////////////////////////////////////////////////////////////////////
        // Computed properties of an object generated by the provider
        //
        case PR_ROW_TYPE :
        case PR_ROWID :
        case PR_VALID_FOLDER_MASK :
        case PR_STORE_ENTRYID :
        case PR_STORE_RECORD_KEY :
        case PR_RECORD_KEY :
        case PR_INSTANCE_KEY :
        case PR_MDB_PROVIDER :
        case PR_MAPPING_SIGNATURE :
        case PR_ENTRYID :
        case PR_PARENT_ENTRYID :
        case PR_OBJECT_TYPE :
        case PR_FOLDER_TYPE :
        case PR_DISPLAY_TYPE :
        case PR_CREATION_VERSION :
        case PR_CURRENT_VERSION :
        case PR_SEARCH_KEY :
        case PR_STORE_SUPPORT_MASK :
        case PR_ATTACH_NUM :
        case PR_ACCESS :
        case PR_ACCESS_LEVEL :
        case PR_MESSAGE_SIZE :
        case PR_HASATTACH :
        case PR_ATTACH_SIZE :
        case PR_BODY_CRC :
        case PR_SUBFOLDERS :
        case PR_CONTENT_COUNT :
        case PR_CONTENT_UNREAD :
        case PR_ASSOC_CONTENT_COUNT :
        // Some implementation allow client and transport to set PR_CREATION_TIME
        // and PR_LAST_MODIFICATION_TIME. In our implementation, we don't allow
        // this. We (the database) computes the time an object was created and
        // when it was last modified).
        case PR_CREATION_TIME :
        case PR_LAST_MODIFICATION_TIME :
        // In our implementation, we don't allow setting the entry ID of the
        // special folders becuase they have fixed location in our database file.
        // On a call to IMsgStore::SetProps(), we will success the overall
        // function call, but return E_ACCESSDENIED for the specific property error.
        // Other implementation of message store could allow setting the entry ID of
        // these folders successfully.
        case PR_COMMON_VIEWS_ENTRYID :
        case PR_FINDER_ENTRYID :
        case PR_IPM_OUTBOX_ENTRYID :
        case PR_IPM_SUBTREE_ENTRYID :
        case PR_IPM_SENTMAIL_ENTRYID :
        case PR_IPM_WASTEBASKET_ENTRYID :
            InfoTraceRaw ("WARNING: Trying to set a computed property: ");
            InfoTraceProp (ulPropTag);
            return MAPI_E_COMPUTED;
    }
    return S_OK;
}


///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetNamesFromIDs()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                            LPGUID            pPropSetGuid,
                                            ULONG             ulFlags,
                                            ULONG *           pcPropNames,
                                            LPMAPINAMEID **   pppPropNames,
                                            LPSPropTagArray   pTags)
{
    // Eventually we will implement a garbage-collection algorithm to recover the
    // IDs of properties no longer in use by objects in our store.
    //                  IrvingD 12/3/95

    HRESULT hResult = S_OK;
    BOOL fReturnWarnings = FALSE;
    ULONG i;
    EnterCriticalSection (&m_csObj);

    if (!(*ppPropTags))
    {
        if (pPropSetGuid && (PS_MAPI == *pPropSetGuid))
		{
		    // If the property set GUID is PS_MAPI, the caller is requesting
            // to return an array of the named properties in this object
            // which are LESS than 0x8000. The caller gave us a list of all
            // the object properties in the pTags argument to this method.
			hResult = m_pNPList->GetMAPINamedPropsAndTags (pTags, ppPropTags, pppPropNames, pcPropNames);
            // We are done.
            goto ErrorExit;
		}

        hResult = m_pNPList->GetAllNamesAndTags (ppPropTags, pppPropNames, ulFlags, pcPropNames);
        // We are done.
        goto ErrorExit;
    }

    hResult = PrivMAPIAllocate ((*ppPropTags)->cValues * sizeof(LPMAPINAMEID), (LPVOID *)pppPropNames);
    if (hResult)
    {
        goto ErrorExit;
    }
    LPVOID pVoid;
    pVoid = (LPVOID)(*pppPropNames);
    LPMAPINAMEID *ppOneName;
    ppOneName = *pppPropNames;
    
    *pcPropNames = 0;

    for (i=0; i<(*ppPropTags)->cValues; i++)
    {
        if (m_pNPList->GetNameOfID ((*ppPropTags)->aulPropTag[i],
                                    pVoid,
                                    ppOneName))
        {
            fReturnWarnings = TRUE;
            *ppOneName = NULL;
        }
        else
        {
            (*pcPropNames)++;
        }
        ppOneName++;
    }
    if (fReturnWarnings)
    {
        hResult = MAPI_W_ERRORS_RETURNED;
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);
    if (FAILED(hResult))
    {
        if (*pppPropNames)
        {
            PrivMAPIFree ((LPVOID *)pppPropNames);
        }
        *pcPropNames = 0;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetIDsFromNames()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::GetIDsFromNames (ULONG               cPropNames,
                                            LPMAPINAMEID *      ppPropNames,
                                            ULONG               ulFlags,
                                            LPSPropTagArray *   ppPropTags,
                                            LPSPropTagArray     pTags,
                                            LPSTREAM            pStream)
{
    HRESULT hResult = S_OK;
    LPSPropTagArray pNewTags = NULL;
    BOOL fReturnWarnings = FALSE;
    LPMAPINAMEID pName;
    ULONG ulNewID, i;
    EnterCriticalSection (&m_csObj);

    if (MAPI_CREATE & ulFlags)
	{
        if (!(MAPI_ACCESS_MODIFY & m_lAccess))
        {
            hResult = E_ACCESSDENIED;
            goto ErrorExit;
        }
	}

    hResult = PrivMAPIAllocate (CbNewSPropTagArray (cPropNames), (LPVOID *)&pNewTags);
    if (hResult)
	{
		goto ErrorExit;
	}

    // If no properties where specified, we return all the
    // tags for all the named properties we have
    if (!cPropNames)
    {
        for (i=0; i<cPropNames; i++)
        {
            if (0x8000 >= PROP_ID(pTags->aulPropTag[i]))
            {
                pNewTags->aulPropTag[pNewTags->cValues] = pTags->aulPropTag[i];
                pNewTags->cValues++;
            }
        }
        goto ErrorExit;
    }

	pNewTags->cValues = cPropNames;
    for (i=0; i<cPropNames; i++)
    {
        // For clarity of code
        pName = (ppPropNames)[i];

        // If the property set is PS_MAPI, we need to make extra validation...
		if (PS_MAPI == *(pName->lpguid))
		{
            // For properties in this set, the kind must be numeric ID
            if (MNID_ID != pName->ulKind)
            {
                pNewTags->aulPropTag[i] = PROP_TAG (PT_ERROR, 0);
                fReturnWarnings = TRUE;
                continue;
			}
		}

		// The same goes for PS_PUBLIC_STRINGS. We need extra validation.
        if (PS_PUBLIC_STRINGS == *(pName->lpguid))
		{
			// For properties in this set, the kind must be string
			if (MNID_STRING != pName->ulKind)
			{
				pNewTags->aulPropTag[i] = PROP_TAG (PT_ERROR, 0);
                fReturnWarnings = TRUE;
				continue;
			}
		}

        hResult = m_pNPList->GetIDForName (pName->lpguid,
                                           (MNID_STRING == pName->ulKind ? pName->Kind.lpwstrName : NULL),
                                           (MNID_ID == pName->ulKind ? pName->Kind.lID : 0),
                                           pName->ulKind,
                                           &ulNewID,
                                           (ulFlags & MAPI_CREATE));
        if (hResult)
        {
            pNewTags->aulPropTag[i] = PROP_TAG (PT_ERROR, 0);
            fReturnWarnings = TRUE;
        }
        else
        {
            pNewTags->aulPropTag[i] = PROP_TAG (PT_UNSPECIFIED, ulNewID);
        }
    }
    if (m_pNPList->IsDirty() && pStream)
    {
        m_pNPList->SaveMappingToStream (pStream);
    }
    if (fReturnWarnings)
    {
        hResult = MAPI_W_ERRORS_RETURNED;
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);

    if (FAILED(hResult))
    {
        PrivMAPIFree ((LPVOID *)&pNewTags);
    }
    *ppPropTags = pNewTags;

    return hResult;
}


///////////////////////////////////////////////////////////////////////////////
//    CGenericProp::GetObjectData()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CGenericProp::GetObjectData (LPVOID * ppVoid)
{
    OBJDATA * pData;
    HRESULT hResult = PrivMAPIAllocate (sizeof(OBJDATA), (LPVOID *)&pData);
    if (!hResult)
    {
        pData->pList = m_pList;
        pData->ppOpenPropList = &m_pOpenPropList;
        pData->pBEobj = m_pBEobj;
        pData->pParentEID = &m_ParentEID;
        pData->pObjectEID = &m_ObjectEID;
        pData->ulObjFlags = m_ulObjFlags;
        pData->lAccess = m_lAccess;
        *ppVoid = (LPVOID)pData;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenObjectList::Insert()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP COpenObjectList::Insert (LPUNKNOWN          pUnkObj,
                                      PRIVATE_ENTRYID *  pObjectEID)
{
    POPEN_OBJECTS pNewNode;
    HRESULT hResult = PrivAllocate (sizeof(OPEN_OBJECTS), (LPVOID *)&pNewNode, "OPEN Object List Node");
    if (hResult)
    {
        return hResult;
    }
    pNewNode->pEID    = pObjectEID;
    pNewNode->pUnkObj = pUnkObj;
    pNewNode->pNext   = NULL;
    EnterCriticalSection (&m_csObj);
    InsertInList (&m_pHead, pNewNode);
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenObjectList::Delete()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
void WINAPI COpenObjectList::Delete (LPUNKNOWN         pUnkObj,
                                     PRIVATE_ENTRYID * pObjectEID)
{
    EnterCriticalSection (&m_csObj);
    POPEN_OBJECTS pPrev = m_pHead, pNode = m_pHead;
    while (pNode)
    {
        if (pNode->pEID == pObjectEID && pNode->pUnkObj == pUnkObj)
        {
            ASSERT (pNode->pEID->bObject == pObjectEID->bObject);
            ASSERT (pNode->pEID->dwObjID == pObjectEID->dwObjID);
            if (pNode == m_pHead)
            {
                // If the node we are deleting is the head of the list, reset the list head pointer.
                m_pHead = m_pHead->pNext;
            }
            else
            {
                // Take this node out from the list, and make the previous node
                // point to the one in front of the current node.
                pPrev->pNext = pNode->pNext;
            }
            PrivFree ((LPVOID *)&pNode);
            break; // Out of the WHILE loop
        }
        pPrev = pNode;
        pNode = pNode->pNext;
    }
    LeaveCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    COpenObjectList::Walk()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
LPUNKNOWN WINAPI COpenObjectList::Walk (SUB_ENTRYID     ObjectEID,
                                        WALKING_ACTION  waAction,
                                        LPUNKNOWN       pPreviousUnk)
{
    PRIVATE_ENTRYID eid = { 0 };
    eid.bObject = ObjectEID.bObject;
    eid.dwObjID = ObjectEID.dwObjID;
    return Walk (eid, waAction, pPreviousUnk);
}

///////////////////////////////////////////////////////////////////////////////
//    COpenObjectList::Walk()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
LPUNKNOWN WINAPI COpenObjectList::Walk (PRIVATE_ENTRYID ObjectEID,
                                        WALKING_ACTION  waAction,
                                        LPUNKNOWN       pPreviousUnk)
{
    LPUNKNOWN pUnk = NULL;
    EnterCriticalSection (&m_csObj);
    POPEN_OBJECTS pNode = m_pHead;
    if (WALK_CONTINUE == waAction)
    {
        while (pNode)
        {
            if (pNode->pUnkObj == pPreviousUnk)
            {
                ASSERT (pNode->pEID->bObject == ObjectEID.bObject);
                ASSERT (pNode->pEID->dwObjID == ObjectEID.dwObjID);
                pNode = pNode->pNext;
                break; // Out of the WHILE() loop
            }
            pNode = pNode->pNext;
        }
    }
    else
    {
        ASSERT (WALK_START == waAction);
        ASSERT (NULL == pPreviousUnk);
    }
    while (pNode)
    {
        if ((pNode->pEID->bObject == ObjectEID.bObject) && (pNode->pEID->dwObjID == ObjectEID.dwObjID))
        {
            pUnk = pNode->pUnkObj;
            break; // Out of the WHILE() loop
        }
        pNode = pNode->pNext;
    }
    if (pUnk)
    {
        pUnk->AddRef();
    }
    LeaveCriticalSection (&m_csObj);
    return pUnk;
}

// End of file for GENPROP.CPP
