///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSLOGON.CPP
//
//  Description
//      This file implements the IMSLogon interface with the methods specified
//      in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.0
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSLMS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::CMSLogon()
//
//    Parameters
//      [IN]    pStore  IMsgStore linked to the object.
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values. Further initialization is
//      defer to the contructor of the base class.
//
//    Return Value
//      None
//
CMSLogon::CMSLogon (CMsgStore * pStore) : CGeneric()
{
    InfoTrace ("CMSLogon: Constructor called");
    m_pNext  = NULL;
    m_pStore = pStore;
    m_pStore->AddRef();
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::~CMSLogon()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CMSLogon.
//
//    Return Value
//      None
//
CMSLogon::~CMSLogon ()
{
    InfoTrace ("CMSLogon: Destructor called");
    m_pStore->Release();
    m_pStore = NULL;
    m_pNext = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL.
//
//    Return Value
//      S_OK            If successful. With the interface pointer in *ppvObj
//      E_NOINTERFACE   If interface requested is not supported by this object
//
STDMETHODIMP CMSLogon::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IMSLogon || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMSLogon virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Delegates the call to the contained to the base class member which has
//      an implementation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSLogon::GetLastError (HRESULT       hError,
                                     ULONG         ulFlags,
                                     LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CMSLogon::GetLastError method called");
    Validate_IMSLogon_GetLastError (this,
                                    hError,
                                    ulFlags,
                                    ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_pStore->m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::Logoff()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't do anything in this method.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMSLogon::Logoff (ULONG * pulFlags)
{
    InfoTrace ("CMSLogon::Logoff method called");
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::OpenEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Delegates the call to the contained IMsgStore store object which has
//      an implementation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSLogon::OpenEntry (ULONG       cbEntryID,
                                  LPENTRYID   pEntryID,
                                  LPCIID      pInterface,
                                  ULONG       ulFlags,
                                  ULONG *     pulObjType,
                                  LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMSLogon::OpenEntry method called");
    HRESULT hResult = m_pStore->OpenEntry (cbEntryID,
                                           pEntryID,
                                           pInterface,
                                           ulFlags,
                                           pulObjType,
                                           ppUnk);
    TraceResult ("CMSLogon::OpenEntry", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::CompareEntryIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Delegates the call to the contained IMsgStore store object which has
//      an implementation.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMSLogon::CompareEntryIDs (ULONG     cbEntryID1,
                                        LPENTRYID pEntryID1,
                                        ULONG     cbEntryID2,
                                        LPENTRYID pEntryID2,
                                        ULONG     ulFlags,
                                        ULONG *   pulResult)
{
    InfoTrace ("CMSLogon::CompareEntryIDs method called");
    HRESULT hResult = m_pStore->CompareEntryIDs (cbEntryID1,
                                                 pEntryID1,
                                                 cbEntryID2,
                                                 pEntryID2,
                                                 ulFlags,
                                                 pulResult);
    TraceResult ("CMSLogon::Advise", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::Advise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Delegates the call to the contained IMsgStore store object which has
//      an implementation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSLogon::Advise (ULONG            cbEntryID,
                               LPENTRYID        pEntryID,
                               ULONG            ulEventMask,
                               LPMAPIADVISESINK pAdviseSink,
                               ULONG *          pulConnection)
{
    InfoTrace ("CMSLogon::Advise method called");
    HRESULT hResult = m_pStore->Advise (cbEntryID,
                                        pEntryID,
                                        ulEventMask,
                                        pAdviseSink,
                                        pulConnection);
    TraceResult ("CMSLogon::Advise", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::Unadvise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Delegates the call to the contained IMsgStore store object which has
//      an implementation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSLogon::Unadvise (ULONG ulConnection)
{
    InfoTrace ("CMSLogon::Unadvise method called");
    HRESULT hResult = m_pStore->Unadvise (ulConnection);
    TraceResult ("CMSLogon::Unadvise", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::OpenStatusEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub function. We don't implement the IMAPIStatus interface.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMSLogon::OpenStatusEntry (LPCIID   pInterface,
                                        ULONG    ulFlags,
                                        ULONG *  pulObjType,
                                        LPVOID * ppEntry)
{
    InfoTrace ("CMSLogon::OpenStatusEntry method called");
    TraceResult ("CMSLogon::OpenStatusEntry", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSLogon::InitializeStatusRow()
//
//    Parameters
//      None
//
//    Purpose
//      Adds the status row of this provider to the MAPI subsystem status
//      table. 
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSLogon::InitializeStatusRow()
{
    #define NUM_STATUS_ROW_PROPS    5
    SPropValue spvStatusRow[NUM_STATUS_ROW_PROPS] = { 0 };
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_PROVIDER_DISPLAY property: The message store readable name
    ULONG i = 0;
    spvStatusRow[i].ulPropTag = PR_PROVIDER_DISPLAY;
    spvStatusRow[i++].Value.lpszA = PROVIDER_DISPLAY_NAME_STRING;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_RESOURCE_METHODS property. These are the methods implemented
    // in the our IMAPIStatus implementation (CMAPIStatus class.)
    spvStatusRow[i].ulPropTag = PR_RESOURCE_METHODS;
    spvStatusRow[i++].Value.l = 0;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_STATUS_CODE property.
    spvStatusRow[i].ulPropTag = PR_STATUS_CODE;
    spvStatusRow[i++].Value.l = STATUS_AVAILABLE ;
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_STATUS_STRING property
    spvStatusRow[i].ulPropTag = PR_STATUS_STRING;
    spvStatusRow[i++].Value.lpszA = "Available";
    
    ///////////////////////////////////////////////////////////////////////////
    // Set the PR_DISPLAY_NAME property
    spvStatusRow[i].ulPropTag = PR_DISPLAY_NAME;
    spvStatusRow[i++].Value.lpszA = "Sample Mailbox";
    
    ASSERT (NUM_STATUS_ROW_PROPS == i);

    // Write the entries on the provider's session status row
    HRESULT hResult = m_pStore->m_pSupObj->ModifyStatusRow (i, spvStatusRow, 0);
    TraceResult ("CMSLogon::InitializeStatusRow", hResult);
    return hResult;
}

// End of file for MSLOGON.CPP