///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSMSG.CPP
//
//  Description
//      This file implements the IMessage interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.0
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSLMS.H"
#include "CFILESTG.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

#define NUM_MESSAGE_PROPS        21
// Default internal properties available in our implementation of IMessage
// Other properties could be available if a client or MAPI added them using
// IMessage::SetProps()
SizedSPropTagArray(NUM_MESSAGE_PROPS, sptMsgObj) =
{
    NUM_MESSAGE_PROPS,
    {
        PR_ACCESS,                      // READ only
        PR_ACCESS_LEVEL,                // READ only
        PR_ENTRYID,                     // READ only
        PR_PARENT_ENTRYID,              // READ only
        PR_STORE_ENTRYID,               // READ only
        PR_STORE_RECORD_KEY,            // READ only
        PR_RECORD_KEY,                  // READ only
        PR_SEARCH_KEY,                  // READ/WRITE (before the first SaveChanges(), READ only afterwards)
        PR_MAPPING_SIGNATURE,           // READ only
        PR_HASATTACH,                   // READ only
        PR_OBJECT_TYPE,                 // READ only
        PR_STORE_SUPPORT_MASK,          // READ only
        PR_CREATION_TIME,               // READ only
        PR_CREATION_VERSION,            // READ only
        PR_LAST_MODIFICATION_TIME,      // READ only
        PR_MESSAGE_ATTACHMENTS,         // READ only
        PR_MESSAGE_RECIPIENTS,          // READ only
        PR_MESSAGE_SIZE,                // READ only
        PR_MESSAGE_CLASS,               // READ/WRITE
        PR_MESSAGE_FLAGS,               // READ/WRITE (before the first SaveChanges(), READ only afterwards)
        PR_PRIORITY                     // READ/WRITE
    }
};

SizedSPropTagArray(2, sptSubmitRecipTable) =
{
    2,
    {
        PR_ROWID,
        PR_RESPONSIBILITY
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CMessage::CMessage()
//
//    Parameters
//      [IN]    pStore          Pointer to the container IMsgStore object
//      [IN]    pParentEID      Entry ID of the parent folder
//      [IN]    pFolderEID      Entry ID of this message
//      [IN]    pList           In-memory property cache for this object
//      [IN]    fDeferErrors    Weather or not we should defer errors
//                              in this instance
//      [IN]    lAccess         Access level of this object
//      [IN]    pParentAttObj   Parent attachment object for embeeded messages
//      [IN]    pNPList         Pointer to the named properties manager
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CMessage::CMessage (CMsgStore *       pStore,
                    PRIVATE_ENTRYID * pParentEID,
                    PRIVATE_ENTRYID * pMsgEID,
                    CPropList *       pList,
                    BOOL              fDeferErrors,
                    long              lAccess,
                    CAttach *         pParentAttObj,
                    CNamePropMgr *    pNPList) : CGenericProp (pStore->m_pBEobj,
                                                               pParentEID,
                                                               pMsgEID,
                                                               pList,
                                                               fDeferErrors,
                                                               lAccess)
{
    InfoTrace ("CMessage: Constructor called");
    m_pNPList           = pNPList;
    m_pAttTable         = NULL;
    m_pRecipTable       = NULL;
    m_fRecipTableDirty  = FALSE;
    m_dwRecipTableViews = 0;
    m_dwAttTableViews   = 0;
    m_pActionList       = NULL;
    // Don't AddRef the parent attachment in embedded messages.
    m_pParentAttObj     = pParentAttObj;
    
    SPropValue spvFlags = { 0 };
    pList->Find (PR_MESSAGE_FLAGS, &spvFlags);
    m_ulObjFlags |= ((MSGFLAG_ASSOCIATED & spvFlags.Value.l) ? OBJECT_ASSOCIATED : 0) |
                    ((MSGFLAG_READ & spvFlags.Value.l) ? 0 : OBJECT_UNREAD_MESSAGE) |
                    ((MSGFLAG_HASATTACH & spvFlags.Value.l) ? OBJECT_HAS_ATTACHMENT : 0);

    m_pStore = pStore;
    m_pStore->AddRef();
    if (NULL == m_pParentAttObj)
    {
        m_pStore->m_OpenObjList.Insert (this, &m_ObjectEID);
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::~CMessage()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CMessage.
//
//    Return Value
//      None
//
CMessage::~CMessage()
{
    InfoTrace ("CMessage: Destructor called");
    // Release the object's tables
    if (m_pAttTable)
    {
        m_pAttTable->Release();
        m_pAttTable = NULL;
    }
    if (m_pRecipTable)
    {
        m_pRecipTable->Release();
        m_pRecipTable = NULL;
    }
    if (m_dwRecipTableViews)
    {
        TraceString1 ("CMessage: Object going away with %d views left on the recipient table", m_dwRecipTableViews);
    }
    if (m_dwAttTableViews)
    {
        TraceString1 ("CMessage: Object going away with %d views left on the attachments table", m_dwAttTableViews);
    }
    // If we have a list of transacted actions, delete them. Some of the nodes
    // way have IAttach objects in them, so make sure to release them
    if (m_pActionList)
    {
        TRANSACTED_OBJECTS * pNode = m_pActionList;
        while (pNode)
        {
            if (TRANS_OBJ_ACTION_SAVE == pNode->Action)
            {
                ASSERT (MAPI_ATTACH == pNode->ObjEID.bObject);
                pNode->pUnkObj->Release();
            }
            pNode = pNode->pNext;
        }
        DeleteList (&m_pActionList);
    }
    // If the message is not an IAttach embeded message, that it out from the list of open messages
    if (NULL == m_pParentAttObj)
    {
        m_pStore->m_OpenObjList.Delete (this, &m_ObjectEID);
    }
    // If the object has been deleted in the parent folder, I must delete the file storage for this message....
    if (OBJECT_DELETED_IN_CONTAINER & m_ulObjFlags)
    {
        // ... however, the message storage should only go away if this is the LAST object open. I search the
        // list again trying to find another message with the same entry ID. If found, I don't delete it now,
        // Let that task for the last object instance opened by a client.
        BOOL fDeleteObject = TRUE;
        CMessage * pMsgObj = (CMessage *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
        while (pMsgObj)
        {
            fDeleteObject = FALSE;
            pMsgObj->Release();
            pMsgObj = NULL;
        }
        if (fDeleteObject)
        {
            m_pBEobj->DeleteOneObject (m_ObjectEID, m_ParentEID);
        }
    }
    m_pStore->Release();
    m_pStore = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IMessage || riid == IID_IMAPIProp || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method finds a string description for the last error cached in.
//      If the error cached does not match the error passed by the client, I
//      fail with MAPI_E_NOT_FOUND. The MAPIERROR structure is allocated
//      dynamically, so the caller is responsible for releasing it with
//      the MAPI allocation functions.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetLastError (HRESULT       hError,
                                     ULONG         ulFlags,
                                     LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CMessage::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this,
                                     hError,
                                     ulFlags,
                                     ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_pStore->m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SaveChanges()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function will save the properties on this object from the cache
//      into the message database file.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CMessage::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);
    HRESULT hResult = S_OK;
    if (OBJECT_DELETED_IN_CONTAINER & m_ulObjFlags)
    {
        hResult = MAPI_E_OBJECT_DELETED;
        SetLastError (hResult);
        return hResult;
    }
    if (OBJECT_MESSAGE_IN_MESSAGE & m_ulObjFlags)
    {
        ASSERT (m_pParentAttObj);
        __try
        {
            if (m_pParentAttObj->m_cRef < 1)
            {
                hResult = E_FAIL;
            }
        }
        __except (EXCEPTION_EXECUTE_HANDLER)
        {
            hResult = E_FAIL;
        }
        if (hResult)
        {
            TraceMessage ("CMessage::SaveChanges: The container attachment has been released");
            return S_OK;
        }
        EnterCriticalSection (&m_csObj);
        EnterCriticalSection (&m_pParentAttObj->m_csObj);
        TRANSACTED_OBJECTS * pNewNode;
        hResult = PrivAllocate (sizeof(TRANSACTED_OBJECTS), (LPVOID *)&pNewNode, "Emb Msg Action Node (Save)");
        if (!hResult)
        {
            pNewNode->Action = TRANS_OBJ_ACTION_SAVE;
            pNewNode->ObjEID = m_ObjectEID;
            pNewNode->pUnkObj = (LPUNKNOWN)this;
            pNewNode->pUnkObj->AddRef();
            InsertInList (&(m_pParentAttObj->m_pActionList), pNewNode);
            if (ulFlags & KEEP_OPEN_READONLY)
            {
                // The user need only read access from now on.
                m_lAccess = MAPI_ACCESS_READ;
            }
            else
            {
                // If this flag was not passed in, close the object and change the access flags.
                if (!(ulFlags & KEEP_OPEN_READWRITE))
                {
                    m_lAccess = 0;
                }
            }
        }
        LeaveCriticalSection (&m_pParentAttObj->m_csObj);
        LeaveCriticalSection (&m_csObj);
    }
    else
    {
        hResult = SaveChangesNow (ulFlags);
    }
    TraceResult ("CMessage::SaveChanges", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SaveChangesNow()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CMessage::SaveChangesNow (ULONG ulFlags)
{
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    long lOldObjectAccess = m_lAccess;
    m_lAccess = MAPI_ACCESS_MODIFY | MAPI_ACCESS_DELETE | MAPI_ACCESS_READ;

    // Generate PR_NORMALIZED_SUBJECT on large properties
    BOOL fNormalizeSubjectSet = FALSE;
    if (m_pOpenPropList)
    {
        LPTSTR pszNormSubject, pszSubject;
        SPropValue spvOne = { 0 };
        HRESULT hError = m_pList->Find (PR_SUBJECT, &spvOne, FALSE);
        if (S_OK == hError && (PROP_TYPE(spvOne.ulPropTag) == PT_STREAM_STORAGE))
        {
            hError = PrivAllocate (4096, (LPVOID *)&pszSubject, "SaveChanges Subject");
            if (!hError)
            {
                hError = PrivAllocate (4096, (LPVOID *)&pszNormSubject, "SaveChanges Normalized Subject");
                if (!hError)
                {
                    hError = m_pOpenPropList->Find (PR_SUBJECT, &spvOne, pszSubject, 4096);
                    if (!hError)
                    {
                        if (!SetNormalizedSubject (pszSubject, &pszNormSubject))
                        {
                            fNormalizeSubjectSet = TRUE;
                            spvOne.ulPropTag = PR_SUBJECT;
                            spvOne.Value.LPSZ = pszSubject;
                            m_pList->Insert (&spvOne);
                            spvOne.ulPropTag = PR_NORMALIZED_SUBJECT;
                            spvOne.Value.LPSZ = pszNormSubject;
                            m_pList->Insert (&spvOne);
                            m_pOpenPropList->Delete (PR_SUBJECT);
                        }
                    }
                    PrivFree ((LPVOID *)&pszNormSubject);
                }
                PrivFree ((LPVOID *)&pszSubject);
            }
        }
    }
    if (!fNormalizeSubjectSet)
    {
        // TODO: Set PR_NORMALIZED_SUBJECT and PR_SUBJECT_PREFIX
    }
    long lOldAccess = m_lAccess;
    // If this object has attachments, the actions perfomed in those objects are
    // not commited until the container object (this message) is committed. I
    // maintain a list of actions done on those objects and now that the
    // container is getting saved, we do the actual save of the attachment.
    if (m_pActionList)
    {
        TRANSACTED_OBJECTS * pNode = m_pActionList;
        while (pNode)
        {
            if (TRANS_OBJ_ACTION_DELETE == pNode->Action)
            {
                m_pBEobj->DeleteObjects (m_ObjectEID,
                                         ATTACHMENT_OBJECT,
                                         (LPMAPIPROGRESS)pNode->ObjEID.dwObjID,
                                         NULL);
                pNode->Action = TRANS_OBJ_ACTION_UNDEFINED;
            }
            if (TRANS_OBJ_ACTION_SAVE == pNode->Action)
            {
                ASSERT (MAPI_ATTACH == pNode->ObjEID.bObject);
                hResult = ((CAttach *)pNode->pUnkObj)->SaveChangesNow (FORCE_SAVE);
                pNode->pUnkObj->Release();
                pNode->pUnkObj = NULL;
                pNode->Action = TRANS_OBJ_ACTION_UNDEFINED;
            }
            pNode = pNode->pNext;
        }
        hResult = S_OK;
    }
    if (!hResult)
    {
        DeleteList (&m_pActionList);
        hResult = CGenericProp::SaveChanges (ulFlags, m_fRecipTableDirty ? m_pRecipTable : NULL);
    }
    if (!hResult)
    {
        NOTIFICATION notif = { 0 };
        // Update the parent folder but is the message is an embedded message (i.e. as an attachment), skip this
        if (!(OBJECT_MESSAGE_IN_MESSAGE & m_ulObjFlags))
        {
            // I don't need to save the last error code. The SaveChanges function cached the error, if any.
            m_fRecipTableDirty = FALSE;
            ZeroMemory (&notif, sizeof(NOTIFICATION));
            notif.info.obj.ulObjType  = MAPI_MESSAGE;
            notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
            notif.info.obj.lpEntryID  = (LPENTRYID)&m_ObjectEID;
            notif.info.obj.cbParentID = CB_PRIVATE_EID;
            notif.info.obj.lpParentID = (LPENTRYID)&m_ParentEID;
            if (OBJECT_NEVER_SAVED & m_ulObjFlags)
            {
                m_pBEobj->UpdateContentCount (m_ParentEID,
                                              m_ObjectEID,
                                              OBJECT_ASSOCIATED & m_ulObjFlags,
                                              OBJECT_UNREAD_MESSAGE & m_ulObjFlags);
                // Send a notification that a new object was created
                notif.ulEventType = fnevObjectCreated;
            }
            else
            {
                // Send a notification that this object has changed
                notif.ulEventType = fnevObjectModified;
            }
            m_pStore->NotifyClients (&m_ObjectEID, 1, &notif);
            
            // I need to get the opened parent object and update the appropiate
            // table so that clients with notifications on those object's tables get
            // updated and notified of the changes/additions/substractions done on it
            CMAPIFolder * pParent = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ParentEID, WALK_START, NULL); // Start the walk in the list looking for this entry ID
            if (pParent)
            {
                // For newly create messages ...
                if (OBJECT_NEVER_SAVED & m_ulObjFlags)
                {
                    // ... Add the message to the contents tables and update the folder count in the hierarchy
                    pParent->UpdateContentTables (UPD_TABLE_ADD_ROW,                    // Add a new message
                                                  OBJECT_ASSOCIATED & m_ulObjFlags,     // Weather or not a message is associated
                                                  OBJECT_UNREAD_MESSAGE & m_ulObjFlags, // Weather or not a message has been read
                                                  &m_ObjectEID,
                                                  m_pList);
                    // ... We also need to update the hierarchy of the parent of
                    // the parent folder of this message so that the new values for
                    // PR_CONTENT_COUNT, PR_CONTENT_UNREAD, and PR_ASSOC_CONTENT_COUNT show up
                    pParent->UpdateParentTable();
                    // ... Finally, send a notification that some object properties have been altered (on the parent)
                    ZeroMemory (&notif, sizeof(NOTIFICATION));
                    notif.ulEventType         = fnevObjectModified;
                    notif.info.obj.ulObjType  = MAPI_FOLDER;
                    notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
                    notif.info.obj.lpEntryID  = (LPENTRYID)&pParent->m_ObjectEID;
                    notif.info.obj.cbParentID = CB_PRIVATE_EID;
                    notif.info.obj.lpParentID = (LPENTRYID)&pParent->m_ParentEID;
                    m_pStore->NotifyClients (&pParent->m_ObjectEID, 1, &notif);
                }
                else
                {
                    // For messages that already exist we don't need to update the hierarchy
                    // of the parent of the parent folder of this message
                    pParent->UpdateContentTables (UPD_TABLE_REFRESH,
                                                  OBJECT_ASSOCIATED & m_ulObjFlags,     // Weather or not a message is associated
                                                  OBJECT_UNREAD_MESSAGE & m_ulObjFlags, // Weather or not a message has been read
                                                  &m_ObjectEID,
                                                  m_pList);
                }
                pParent->Release();
            }
            hResult = S_OK;
        }
        m_ulObjFlags &= ~OBJECT_NEVER_SAVED;
    }
    m_lAccess = lOldObjectAccess;
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the value of the properties the client especified in the
//      property tag array. If the tag array is NULL, the user wants all
//      the properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetProps (LPSPropTagArray pPropTagArray,
                                 ULONG           ulFlags,
                                 ULONG *         pcValues,
                                 LPSPropValue *  ppPropArray)
{
    InfoTrace ("CMessage::GetProps method called");
    Validate_IMAPIProp_GetProps (this,
                                 pPropTagArray,
                                 ulFlags,
                                 pcValues,
                                 ppPropArray);
    #ifdef DO_INFO_TRACES
    if (!pPropTagArray)
    {
        TraceMessage ("CMessage::GetProps: All properties requested");
    }
    else
    {
        TraceMessage ("CMessage::GetProps:: Properties requested...");
        for (ULONG j=0; j<pPropTagArray->cValues; j++) { TraceProp (pPropTagArray->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES

    EnterCriticalSection (&m_csObj);
    BOOL fGotErrors = FALSE;
    HRESULT hResult = CGenericProp::GetProps  (pPropTagArray,
                                               ulFlags,
                                               pcValues,
                                               ppPropArray,
                                               (LPSPropTagArray)&sptMsgObj);
    if (hResult)
    {
        goto ErrorExit;
    }

    LPVOID pSrc;
    ULONG cbSize, i;

    for (i=0; i<*pcValues; i++)
    {
        ASSERT (PT_STREAM_STORAGE != PROP_TYPE ((*ppPropArray)[i].ulPropTag));
    }

    // Get the values for all the properties requested
    for (i=0; i<*pcValues; i++)
    {
        pSrc = NULL;
        hResult = S_OK;
        switch ((*ppPropArray)[i].ulPropTag)
        {
            // Get the PT_OBJECT/PR_NULL properties
            case PR_NULL :
            case PR_MESSAGE_ATTACHMENTS :
            case PR_MESSAGE_RECIPIENTS :
                (*ppPropArray)[i].Value.err = S_OK;
                break;
            // Get the PT_LONG properties
            case PR_ACCESS :
                (*ppPropArray)[i].Value.l = m_lAccess;
                break;
            case PR_ACCESS_LEVEL :
                (*ppPropArray)[i].Value.l = (m_lAccess & MAPI_ACCESS_MODIFY ? MAPI_MODIFY : 0);
                break;
            case PR_OBJECT_TYPE :
                (*ppPropArray)[i].Value.l = MAPI_MESSAGE;
                break;
            case PR_STORE_SUPPORT_MASK:
                (*ppPropArray)[i].Value.l = SUPPORT_MASK;
                break;
            // Get the PT_I8 properties
            case PR_CREATION_VERSION :
                (*ppPropArray)[i].Value.li.LowPart = MSG_STORE_VERSION;
                (*ppPropArray)[i].Value.li.HighPart = 0;
                break;
            // Get the PT_BOOLEAN properties
            case PR_HASATTACH :
                (*ppPropArray)[i].Value.b = (USHORT)!!(OBJECT_HAS_ATTACHMENT & m_ulObjFlags);
                break;
            // Get the PT_BINARY properties
            case PR_STORE_RECORD_KEY :
            case PR_MAPPING_SIGNATURE :
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&m_pStore->m_StoreUID;
                break;
            case PR_PARENT_ENTRYID :
                cbSize = CB_PRIVATE_EID;
                pSrc = (LPVOID)&m_ParentEID;
                break;
            case PR_STORE_ENTRYID :
                cbSize = m_pStore->m_StoreEID.cb;
                pSrc = (LPVOID)m_pStore->m_StoreEID.lpb;
                break;
            // In our implementation, PR_RECORD_KEY and PR_ENTRYID
            // are binary equivalent
            case PR_RECORD_KEY :
            case PR_ENTRYID :
                cbSize = CB_PRIVATE_EID;
                pSrc = (LPVOID)&m_ObjectEID;
                break;
            // Oh no, the property was not found. Try the list of properties
            default :
                hResult = CGenericProp::GetProps (&(*ppPropArray)[i], *ppPropArray);
                // Message Stores implementing IMessage should always return at least
                // an empty string for PR_DISPLAY_TO, PR_DISPLAY_CC, and PR_DISPLAY_BCC
                // for messages without this type of recipients.
                if (MAPI_E_NOT_FOUND == hResult &&
                    (PR_DISPLAY_TO  == (*ppPropArray)[i].ulPropTag ||
                     PR_DISPLAY_CC  == (*ppPropArray)[i].ulPropTag ||
                     PR_DISPLAY_BCC == (*ppPropArray)[i].ulPropTag))
                {
                    pSrc = (LPVOID)szBlank;
                    hResult = S_OK;
                }
                #ifdef DO_INFO_TRACES
                if (MAPI_E_NOT_FOUND == hResult)
                {
                    TraceRaw ("CMessage::GetProps: Not Found - ");
                    TraceProp ((*ppPropArray)[i].ulPropTag);
                }
                #endif // #DO_INFO_TRACES
                break;
        }
        if (S_OK != CGenericProp::GetProps (hResult,
                                            cbSize,
                                            pSrc,
                                            &(*ppPropArray)[i],
                                            *ppPropArray))
        {
            fGotErrors = TRUE;
        }
    }
ErrorExit :
    LeaveCriticalSection (&m_csObj);
    if (fGotErrors)
	{
        hResult = MAPI_W_ERRORS_RETURNED;
    }
    TraceResult ("CMessage::GetProps", hResult);
    SetLastError (hResult);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetPropList()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return an array of property tags with all the properties
//      in this object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetPropList (ULONG             ulFlags,
                                    LPSPropTagArray * ppAllTags)
{
    InfoTrace ("CMessage::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppAllTags);
    return CGenericProp::GetPropList (ulFlags, (LPSPropTagArray)&sptMsgObj, ppAllTags);
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::OpenProperty()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::OpenProperty (ULONG       ulPropTag,
                                     LPCIID      piid,
                                     ULONG       ulInterfaceOptions,
                                     ULONG       ulFlags,
                                     LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMessage::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this,
                                     ulPropTag,
                                     piid,
                                     ulInterfaceOptions,
                                     ulFlags,
                                     ppUnk);
    HRESULT hResult = S_OK;
    if (ulPropTag == PR_MESSAGE_ATTACHMENTS || ulPropTag == PR_MESSAGE_RECIPIENTS) 
    {
        if (piid)
        {
            if (*piid != IID_IMAPITable || *piid != IID_IUnknown)
            {
                TraceResult ("CMessage::OpenProperty: Invalid interface requested for the recipient table", hResult);
                hResult = E_NOINTERFACE;
            }
        }
        if (!hResult && ulPropTag == PR_MESSAGE_RECIPIENTS)
        {
            hResult = GetRecipientTable (ulFlags, (LPMAPITABLE *)ppUnk);
        }
        if (!hResult && ulPropTag == PR_MESSAGE_ATTACHMENTS)
        {
            hResult = GetAttachmentTable (ulFlags, (LPMAPITABLE *)ppUnk);
        }
    }
    else
    {
        hResult = CGenericProp::OpenProperty (ulPropTag, piid, ulInterfaceOptions, ulFlags, ppUnk);
    }
    TraceResult ("CMessage::OpenProperty", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Modify the value of the properties in our object. On transacted
//      objects (such as this), the changes to the property cache are not
//      permanent until the SaveChanges() method is called.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::SetProps (ULONG                 cValues,
                                 LPSPropValue          pPropArray,
                                 LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMessage::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMessage::SetProps: Cannot set properties in a READ only object");
        TraceResult ("CMessage::SetProps", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    
    #ifdef DO_INFO_TRACES
    TraceMessage ("CMessage::SetProps: Properties to set...");
    for (ULONG j=0; j<cValues; j++) { TraceProp (pPropArray[j].ulPropTag); }
    #endif DO_INFO_TRACES

    HRESULT hTagError, hResult = S_OK;
    EnterCriticalSection (&m_csObj);

    for (ULONG i=0; i<cValues && !hResult; i++)
    {
        switch (PROP_TYPE(pPropArray[i].ulPropTag))
        {
            case PT_NULL :
            case PT_UNSPECIFIED :
            case PT_ERROR :
                continue;
        }
        hTagError = S_OK;
        switch (pPropArray[i].ulPropTag)
        {
            // This properties cannot be set by clients. They are computed after the a
            // modification of the recipient table through IMessage::ModifyRecipients()
            case PR_DISPLAY_BCC :
            case PR_DISPLAY_CC :
            case PR_DISPLAY_TO :
            case PR_NORMALIZED_SUBJECT :
                hTagError = MAPI_E_COMPUTED;
                break;
            // The PR_SEARCH_KEY may be set on messages by clients before the first call to
            // IMessage::SaveChanges is made. After that the property is computed by the service provider.
            case PR_SEARCH_KEY :
                 if (OBJECT_NEVER_SAVED & m_ulObjFlags)
                 {
                    hTagError = AddPropToList (&pPropArray[i], TRUE);
                 }
                 else
                 {
                    TraceMessage ("CMessage::SetProps: A client is trying to set PR_SEARCH_KEY after the message has been saved");
                    hTagError = MAPI_E_COMPUTED; // Return this error of IMAPISupport::DoCopyProps() will no behave properly
                 }
                 break;
            // The PR_MESSAGE_FLAGS may be set on messages by clients before the first call to
            // IMessage::SaveChanges is made. After that the property is computed by the service provider.
            case PR_MESSAGE_FLAGS :
                 if (OBJECT_NEVER_SAVED & m_ulObjFlags)
                 {
                    // Make sure clients are not setting some invalid flags:
                    //      MSGFLAG_HASATTACH
                    //      MSGFLAG_ASSOCIATED
                    SPropValue spvFlags = { 0 };
                    spvFlags.ulPropTag = PR_MESSAGE_FLAGS;
                    spvFlags.Value.l = pPropArray[i].Value.l;
                    spvFlags.Value.l &= ~(MSGFLAG_HASATTACH | MSGFLAG_ASSOCIATED);
                    hTagError = AddPropToList (&spvFlags, TRUE);
                    if (!hTagError)
                    {
                        m_ulObjFlags |= ((MSGFLAG_READ & spvFlags.Value.l) ? 0 : OBJECT_UNREAD_MESSAGE);
                    }
                 }
                 else
                 {
                    TraceMessage ("CMessage::SetProps: A client is trying to set PR_MESSAGE_FLAGS after the message has been saved");
                    hTagError = MAPI_E_COMPUTED; // Return this error of IMAPISupport::DoCopyProps() will no behave properly
                 }
                 break;
            default :
                hTagError = AddPropToList (&pPropArray[i]);
                if (!hTagError)
                {
                    // If the user sets a property that is currently opened as a stream,
                    // yank out the stream from the list and defer it actual deletion from
                    // the database until SaveChanges gets called.
                    if (m_pOpenPropList)
                    {
                        m_pOpenPropList->Delete (pPropArray[i].ulPropTag);
                    }
                }
                break;
        }
        if (hTagError && ppProblems)
        {
            hResult = AddTagToProbArray (cValues,
                                         ppProblems,
                                         i,
                                         pPropArray[i].ulPropTag,
                                         hTagError);
        }
    }
    LeaveCriticalSection (&m_csObj);
    // If an error occurred while adding errors to the problem array, the user
    // won't attempt to delete it so I should.
    if (hResult)
    {
        if (ppProblems)
        {
            PrivMAPIFree ((LPVOID *)ppProblems);
        }
        TraceResult ("CMessage::SetProps", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::DeleteProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function deletes a set of properties specified by the client. In
//      our implementation, we simply defer to the base class to let it delete
//      the properties in the property cache. On IMessage objects, the changes
//      are transacted and are only committed when IMessage::SaveChanges()
//      is invoked.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::DeleteProps (LPSPropTagArray       pPropTagArray,
                                    LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMessage::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    return CGenericProp::DeleteProps (pPropTagArray, ppProblems);
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::CopyTo()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy this object's content into another one. For the copy operation
//      we delegate to MAPI for the implemenation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::CopyTo (ULONG                 ciidExclude,
                               LPCIID                rgiidExclude,
                               LPSPropTagArray       pExcludeProps,
                               ULONG                 ulUIParam,
                               LPMAPIPROGRESS        pProgress,
                               LPCIID                pDestInterface,
                               LPVOID                pDestObj,
                               ULONG                 ulFlags,
                               LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMessage::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,   
                               rgiidExclude,  
                               pExcludeProps, 
                               ulUIParam,     
                               pProgress,     
                               pDestInterface,
                               pDestObj,      
                               ulFlags,       
                               ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyTo (&IID_IMessage,
                                                     this,
                                                     ciidExclude,
                                                     rgiidExclude,
                                                     pExcludeProps,
                                                     ulUIParam,
                                                     pProgress,
                                                     pDestInterface,
                                                     pDestObj,
                                                     ulFlags,
                                                     ppProblems);
    // Set the error in the cache and trace it out
    TraceResult ("CMessage::CopyTo", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::CopyProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy the properties from one object to the other. For the copy
//      operation, we defer to the MAPI support object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::CopyProps (LPSPropTagArray       pIncludeProps,
                                  ULONG                 ulUIParam,
                                  LPMAPIPROGRESS        pProgress,
                                  LPCIID                pDestInterface,
                                  LPVOID                pDestObj,
                                  ULONG                 ulFlags,
                                  LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMessage::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pDestInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyProps (&IID_IMessage,
                                                        this,
                                                        pIncludeProps,
                                                        ulUIParam,
                                                        pProgress,
                                                        pDestInterface,
                                                        pDestObj,
                                                        ulFlags,
                                                        ppProblems);

    // Set the error in the cache and trace it out
    TraceResult ("CMessage::CopyProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetNamesFromIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the list of names for the identifier list supplied by the
//      caller. Our implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                        LPGUID            pPropSetGuid,
                                        ULONG             ulFlags,
                                        ULONG *           pcPropNames,
                                        LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CMessage::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    LPSPropTagArray pTags;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptMsgObj, &pTags);
    if (!hResult)
    {
        hResult = CGenericProp::GetNamesFromIDs (ppPropTags,
                                                 pPropSetGuid,
                                                 ulFlags,
                                                 pcPropNames,
                                                 pppPropNames,
                                                 pTags);
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMessage::GetNamesFromIDs", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetIDsFromNames()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return a property identifier for each named listed by the client. Our
//      IMessage implementation, defers to the common IMAPIProp implementation
//      to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetIDsFromNames (ULONG             cPropNames,
                                        LPMAPINAMEID *    ppPropNames,
                                        ULONG             ulFlags,
                                        LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CMessage::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    
    LPSPropTagArray pTags;
    LPSTREAM pStream;
    LeaveCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptMsgObj, &pTags);
    if (!hResult)
    {
        hResult = m_pBEobj->GetNamedPropertiesStream (&pStream);
        if (!hResult)
        {
            hResult = CGenericProp::GetIDsFromNames (cPropNames,
                                                     ppPropNames,
                                                     ulFlags,
                                                     ppPropTags,
                                                     pTags,
                                                     pStream);
            pStream->Release();
        }
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMessage::GetIDsFromNames", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
// IMessage virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetAttachmentTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function returns a READ-only table of the atachments that
//      exists in this message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetAttachmentTable (ULONG         ulFlags,
                                           LPMAPITABLE * ppTable)
{
    InfoTrace ("CMessage::GetAttachmentTable method called");
    Validate_IMessage_GetAttachmentTable (this, ulFlags, ppTable);
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    // If we don't have an ITableData object yet, go to the back end
    // and retrieve the attachments on it.
    if (!m_pAttTable)
    {
        hResult = m_pBEobj->GetTable (TABLE_ATTACHMENTS,
                                      m_ObjectEID,
                                      m_lAccess,
                                      (LPVOID)&m_pStore->m_StoreEID,
                                      NULL,
                                      &m_pAttTable);
    }
    if (!hResult)
    {
        hResult = m_pAttTable->HrGetView (NULL, MessageTablesViewRelease, (ULONG)this, ppTable);
        TraceResult ("CMessage::GetAttachmentTable: Failed to get a view of the memory table", hResult);
        if (!hResult)
        {
            // For each view on the table data, that is going to be returned to a client, I
            // must AddRef the entire object so that the opened views on this data, upon
            // release, can access the object. The matching release call is in the callback
            // function especified in the HrGetView() call
            AddRef();
            m_dwAttTableViews++;
        }
    }
    LeaveCriticalSection (&m_csObj);
    // If any, save the last error in the object error cache
    TraceResult ("CMessage::GetAttachmentTable", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::OpenAttach()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Open an existing attachment in this message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::OpenAttach (ULONG      ulAttachmentNum,
                                   LPCIID     pInterface,
                                   ULONG      ulFlags,
                                   LPATTACH * ppAttach)
{
    InfoTrace ("CMessage::OpenAttach method called");
    Validate_IMessage_OpenAttach (this,
                                  ulAttachmentNum,
                                  pInterface,
                                  ulFlags,
                                  ppAttach);

    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & MAPI_MODIFY))
    {
        TraceMessage ("CMessage::OpenAttach: Requested WRITE access in a READ only object");
        TraceResult ("CMessage::OpenAttach", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    CPropList * pList = NULL;
    // Although, per the MAPI specs, attachment don't have entry ID, in this
    // implementation we have attachments for them for easier and uniform
    // access to the back end storage.
    PRIVATE_ENTRYID eidAttach = { 0 };
    eidAttach.uidGlobal = m_pBEobj->GetBackEndUID();
    eidAttach.bVersion = ENTRYID_VERSION;
    eidAttach.bObject = MAPI_ATTACH;
    eidAttach.dwObjID = ulAttachmentNum;
    HRESULT hResult = S_OK;

    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);

    // Before I even attempt to open the attachment, make sure the
    // message object HAS attachments stored.
    if (!(OBJECT_HAS_ATTACHMENT & m_ulObjFlags))
    {
        hResult = MAPI_E_NOT_FOUND;
        goto ErrorExit;
    }

    // Get the properties for this attachment from the back end storage
    hResult = m_pBEobj->OpenObject (eidAttach, &pList);
    if (!hResult)
    {
        long lAccess = (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                       (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                       MAPI_ACCESS_READ;
        CAttach * pAttachObj = new CAttach (m_pStore,
                                            &m_ObjectEID,
                                            &eidAttach,
                                            pList,
                                            !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                            lAccess,
                                            this,
                                            m_pNPList);
        if (pAttachObj)
        {
            *ppAttach = (LPATTACH)pAttachObj;
            pAttachObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            hResult = E_OUTOFMEMORY;
            delete pList;
        }
    }
ErrorExit:
    LeaveCriticalSection (&m_csObj);

    // If any, save the error in the object error cache
    TraceResult ("CMessage::OpenAttach", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::CreateAttach()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Creates an IAttach object in this message and
//      returns it to the caller
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::CreateAttach (LPCIID     pInterface,
                                     ULONG      ulFlags,
                                     ULONG *    pulAttachmentNum,
                                     LPATTACH * ppAttach)
{
    InfoTrace ("CMessage::CreateAttach method called");
    Validate_IMessage_CreateAttach (this,
                                    pInterface,
                                    ulFlags,
                                    pulAttachmentNum,
                                    ppAttach);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMessage::CreateAttach: Cannot create a new attachment in a READ only object");
        TraceResult ("CMessage::CreateAttach", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);

    // Although, per the MAPI specs, attachment don't have entry ID, in this
    // implementation we have attachments for them for easier and uniform
    // access to the back end storage.
    PRIVATE_ENTRYID eidNewAttachment = { 0 };
    CPropList * pList;
    CAttach * pAttachObj;
    // Tell the back end to create storage space for this
    // new object and fill the entries for the Attachment "entry ID".
    HRESULT hResult = m_pBEobj->CreateObject (m_ObjectEID, &eidNewAttachment, &pList, MAPI_ATTACH, NULL, NULL);
    if (!hResult)
    {
        // Inherit the parent's access rights.
        long lAccess = (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                       (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                       MAPI_ACCESS_READ;
        pAttachObj = new CAttach (m_pStore,
                                  &m_ObjectEID,
                                  &eidNewAttachment,
                                  pList,
                                  !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                  lAccess,
                                  this,
                                  m_pNPList);
        if (pAttachObj)
        {
            pAttachObj->m_ulObjFlags |= OBJECT_JUST_CREATED |
                                        OBJECT_NEVER_SAVED |
                                        ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            hResult = E_OUTOFMEMORY;
            delete pList;
        }
    }
    LeaveCriticalSection (&m_csObj);
    if (!hResult)
    {
        // The attachment number is the ID the back end
        // assigned in the "entry ID" of the attachment
        *pulAttachmentNum = eidNewAttachment.dwObjID;
        *ppAttach = (LPATTACH)pAttachObj;
    }
    // If any, save the error in the object error cache
    TraceResult ("CMessage::CreateAttach", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::DeleteAttach()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Deletes an attachment that exists in this message. Because IMessage
//      object are transacted objects, if we find the attachment in our
//      message we enter a transaction information node in an internal
//      linked list for the message object. When IMessage::SaveChanges() is
//      called, we walk this list and do the transaction action at that
//      time. In other words the actual deletion of the attachment is
//      deferred until the message object is committed.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::DeleteAttach (ULONG          ulAttachmentNum,
                                     ULONG          ulUIParam,
                                     LPMAPIPROGRESS pProgress,
                                     ULONG          ulFlags)
{
    InfoTrace ("CMessage::DeleteAttach method called");
    Validate_IMessage_DeleteAttach (this,
                                    ulAttachmentNum,
                                    ulUIParam,
                                    pProgress,
                                    ulFlags);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMessage::DeleteAttach: Cannot delete an attachment in a READ only object");
        TraceResult ("CMessage::DeleteAttach", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    HRESULT hResult = S_OK;
    TRANSACTED_OBJECTS * pNewNode;
    // Set up a property structure to search the attachment table
    SPropValue spvRow = { 0 };
    spvRow.ulPropTag = PR_INSTANCE_KEY;
    spvRow.Value.bin.cb = sizeof(ULONG);
    spvRow.Value.bin.lpb = (LPBYTE)&ulAttachmentNum;
    LPSRow pRow;

    // Lock the object from exclusive access
    EnterCriticalSection (&m_csObj);

    // Before I even attempt to delete the attachment, make sure the
    // message object HAS attachments stored.
    if (!(OBJECT_HAS_ATTACHMENT & m_ulObjFlags))
    {
        hResult = MAPI_E_NOT_FOUND;
        goto ErrorExit;
    }
    
    // If we don't have the attachment table around, get it from the back end storage.
    if (!m_pAttTable)
    {
        hResult = m_pBEobj->GetTable (TABLE_ATTACHMENTS,
                                      m_ObjectEID,
                                      m_lAccess,
                                      (LPVOID)&m_pStore->m_StoreEID,
                                      NULL,
                                      &m_pAttTable);
        if (hResult)
        {
            goto ErrorExit;
        }
    }

    // Query the table for the attachment row information
    hResult = m_pAttTable->HrQueryRow (&spvRow, &pRow, NULL);
    if (!hResult)
    {
        // Free the row we got from the table
        PrivMAPIFree ((LPVOID *)&pRow);
        // Allocate a node for the transacted actions taken on attachment objects
        hResult = PrivAllocate (sizeof(TRANSACTED_OBJECTS), (LPVOID *)&pNewNode, "Attach Action Node (Delete)");
        if (!hResult)
        {
            pNewNode->ObjEID.uidGlobal = m_pBEobj->GetBackEndUID();
            pNewNode->ObjEID.bVersion = ENTRYID_VERSION;
            pNewNode->ObjEID.bObject = MAPI_ATTACH;
            pNewNode->ObjEID.dwObjID = ulAttachmentNum;
            pNewNode->Action = TRANS_OBJ_ACTION_DELETE;
            InsertInList (&m_pActionList, pNewNode);
            hResult = m_pAttTable->HrDeleteRow (&spvRow);
            TraceResult ("CMessage::DeleteAttach: Failed to remove row from attachment table", hResult);
            hResult = S_OK;
        }
    }
ErrorExit:
    LeaveCriticalSection (&m_csObj);

    TraceResult ("CMessage::DeleteAttach", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GetRecipientTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function returns a READ-only view of the recipients of
//      this message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GetRecipientTable (ULONG         ulFlags,
                                          LPMAPITABLE * ppTable)
{
    InfoTrace ("CMessage::GetRecipientTable method called");
    Validate_IMessage_GetRecipientTable (this, ulFlags, ppTable);

    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    if (!m_pRecipTable)
    {
        BOOL fReturnEmptyTable = ((OBJECT_NEVER_SAVED & m_ulObjFlags) ? TRUE : FALSE);
        hResult = m_pBEobj->GetTable (TABLE_RECIPIENTS,
                                      m_ObjectEID,
                                      0,
                                      (LPVOID)&m_dwRecipRowID,
                                      (LPVOID)fReturnEmptyTable,
                                      &m_pRecipTable);
    }
    if (!hResult)
    {
        hResult = m_pRecipTable->HrGetView (NULL, MessageTablesViewRelease, (ULONG)this, ppTable);
        TraceResult ("CMessage::GetRecipientTable: Failed to get a view of the memory table", hResult);
        if (!hResult)
        {
            // For each view on the table data, that is going to be returned to a client, I
            // must AddRef the entire object so that the opened views on this data, upon
            // release, can access the object. The matching release call is in the callback
            // function especified in the HrGetView() call
            AddRef();
            m_dwRecipTableViews++;
        }
    }
    LeaveCriticalSection (&m_csObj);

    // If any, save the last error in the object error cache
    TraceResult ("CMessage::GetRecipientTable", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::ModifyRecipients()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This methods adds, removes. or modifies one or more receipient
//      entry from the message's recipient table.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::ModifyRecipients (ULONG     ulFlags,
                                         LPADRLIST pAdl)
{
    InfoTrace ("CMessage::ModifyRecipients method called");
    Validate_IMessage_ModifyRecipients (this, ulFlags, pAdl);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMessage::ModifyRecipients: Cannot modify recipient in a READ only object");
        TraceResult ("CMessage::ModifyRecipients", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    HRESULT hResult = S_OK;
    // If no flags are specified, then it means to replace the entire list of recipients passed in.
    // In our implementation, MODRECIP_ADD will call HrModifyRows() which will do the "replacement."
    if (!ulFlags)
    {
        ulFlags = MODRECIP_ADD;
        if (0 == pAdl->cEntries)
        {
            ulFlags = MODRECIP_REMOVE;
        }
    }

    // If I'm adding or modifying the recipients, prepare the given
	// recipients and initialize the column prop tag array
	if (ulFlags & (MODRECIP_MODIFY | MODRECIP_ADD))
	{
        LPADRBOOK pAddrBook;
		SizedSPropTagArray (1, pta)	=
		{
		    1,
		    {
		        PR_SEARCH_KEY
		    }
        };
        hResult = m_pStore->m_pSupObj->OpenAddressBook (0, 0, &pAddrBook);
		// OpenAddressBook may return warnings. Treat them like success.
		if (SUCCEEDED(hResult))
		{
			hResult = pAddrBook->PrepareRecips (0, (LPSPropTagArray)&pta, pAdl);
			pAddrBook->Release();
		}
		if (hResult)
        {
            TraceResult ("CMessage::ModifyRecipients: Address book failed to prepare the recipients", hResult);
            SetLastError (hResult);
            return hResult;
        }
	}

    ULONG j;
    if (ulFlags & (MODRECIP_MODIFY | MODRECIP_REMOVE))
    {
        // For a modification or removal of a row, I need PR_ROWID which is the index property
        // in our recipient table, without it, we would not be able to
        // locate the row for that recipient.
        for (j=0; j<pAdl->cEntries; j++)
        {
            if (!IsPropInRow ((LPSRow)&pAdl->aEntries[j], PR_ROWID))
            {
                TraceMessage ("CMessage::ModifyRecipients: Missing PR_ROWID for row delete/modify");
                SetLastError (E_INVALIDARG);
                return E_INVALIDARG;
            }
        }
    }
    else
    {
        // When adding new recipient to the table, I can't let the user pass us PR_ROWID
        // because that is something I calculate internally before adding the row to the table.
        for (j=0; j<pAdl->cEntries; j++)
        {
            LPSPropValue pProp;
            if (IsPropInRow ((LPSRow)&pAdl->aEntries[j], PR_ROWID, &pProp))
            {
                ulFlags = MODRECIP_MODIFY;
            }
        }
    }
    
    // Lock the object for exclusive access.
    EnterCriticalSection (&m_csObj);
    if (!m_pRecipTable)
    {
        LPMAPITABLE pTmpTable;
        hResult = GetRecipientTable (0, &pTmpTable);
        if (hResult)
        {
            goto ExitError;
        }
        else
        {
            pTmpTable->Release();
        }
    }
    if (ulFlags & MODRECIP_ADD)
    {
        SRow srTblRow = { 0 };
        SPropValue spvProps[64] = { 0 };
        srTblRow.lpProps = spvProps;
        ULONG i;
        for (j=0; j<pAdl->cEntries; j++)
        {
            if (0 == pAdl->aEntries[j].cValues)
            {
                continue;
            }
            if (pAdl->aEntries[j].cValues > 62) // Size of spvProps - 2 for our PR_ROWID and PR_INSTANCE_KEY properties
            {
                hResult = E_OUTOFMEMORY;
                ASSERTNOW();
            }
            else
            {
                srTblRow.cValues = pAdl->aEntries[j].cValues;
                for (i=0; i<srTblRow.cValues; i++)
                {
                    spvProps[i].ulPropTag = pAdl->aEntries[j].rgPropVals[i].ulPropTag;
                    spvProps[i].Value = pAdl->aEntries[j].rgPropVals[i].Value;
                }
                m_dwRecipRowID++;
                spvProps[i].ulPropTag = PR_ROWID;
                spvProps[i++].Value.l = m_dwRecipRowID;
                srTblRow.cValues++; // Bump the count for the new property added
                // In case the PR_INSTANCE_KEY property is not found for this recipient, add it.
                if (!IsPropInRow ((LPSRow)&pAdl->aEntries[j], PR_INSTANCE_KEY))
                {
                    // In the table of recipients, PR_INSTANCE_KEY and PR_ROWID have the same value
                    spvProps[i].ulPropTag = PR_INSTANCE_KEY;
                    spvProps[i].Value.bin.cb = sizeof(DWORD);
                    spvProps[i++].Value.bin.lpb = (LPBYTE)&m_dwRecipRowID;
                    srTblRow.cValues++; // Bump the count for the new property added
                }

                hResult = m_pRecipTable->HrModifyRow (&srTblRow);
            }
            if (hResult)
            {
                goto ExitError;
            }
        }
    }
    if (ulFlags & MODRECIP_MODIFY)
    {
        hResult = m_pRecipTable->HrModifyRows (0, (LPSRowSet)pAdl);
    }
    LPMAPITABLE pTableView;
    if (ulFlags & MODRECIP_REMOVE)
    {
        // 0 entries means that I need to wipe out the entire recipient list.
        if (0 == pAdl->cEntries)
        {
            hResult = m_pRecipTable->HrGetView (NULL, NULL, 0, &pTableView);
            if (!hResult)
            {
                LPSRowSet pRows;
                BOOL fContinue = TRUE;
                ULONG ulRow;
                SPropTagArray sptRecips = { 1, { PR_ROWID } };
                hResult = pTableView->SetColumns (&sptRecips, 0);
                while (fContinue && !hResult)
                {
                    hResult = pTableView->QueryRows (25, 0, &pRows);
                    if (!hResult)
                    {
                        if (pRows->cRows)
                        {
                            for (ulRow=0; ulRow<pRows->cRows; ulRow++)
                            {
                                hResult = m_pRecipTable->HrDeleteRows (0, pRows, NULL);
                            }
                        }
                        else
                        {
                            fContinue = FALSE;
                        }
                        FreeProws (pRows);
                    }
                }
                pTableView->Release();
            }
        }
        else
        {
            hResult = m_pRecipTable->HrDeleteRows (0, (LPSRowSet)pAdl, NULL);
        }
    }
    if (hResult)
    {
        goto ExitError;
    }

    HRESULT hOtherResult;
    hOtherResult = m_pRecipTable->HrGetView (NULL, NULL, 0, &pTableView);
    if (!hOtherResult)
    {
        // After the recipient has been modified, we call an internal method to generate
        // the string for PR_DISPLAY_xxx properties which are derived from
        // the information in the receipient table.
        hOtherResult = GenerateDisplayNames (pTableView);
        pTableView->Release();
    }
    // If any, trace out the error from generating PR_DISPLAY_TO, PR_DISPLAY_CC, and PR_DISPLAY_BCC.
    TraceResult ("CMessage::ModifyRecipients: An error occurred generating PR_DISPLAY_xxx", hOtherResult);

ExitError:
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMessage::ModifyRecipients", hResult);
    if (!hResult)
    {
        m_fRecipTableDirty = TRUE;
    }
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SubmitMessage()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CMessage::SubmitMessage (ULONG ulFlags)
{
    InfoTrace ("CMessage::SubmitMessage method called");
    Validate_IMessage_SubmitMessage (this, ulFlags);
    
    // I can't submit a message that is read-only because some properties will be altered.
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMessage::SubmitMessage: Cannot submit a READ only");
        TraceResult ("CMessage::SubmitMessage", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    // I don't support submitting embedded messages or for that matter associated messages.
    if (m_ulObjFlags & OBJECT_MESSAGE_IN_MESSAGE ||
        m_ulObjFlags & OBJECT_ASSOCIATED)
    {
        TraceMessage ("CMessage::SubmitMessage: Cannot submit embeded messages or associated messages");
        TraceResult ("CMessage::SubmitMessage", MAPI_E_NO_SUPPORT);
        SetLastError (MAPI_E_NO_SUPPORT);
        return MAPI_E_NO_SUPPORT;
    }
    
    SYSTEMTIME st;
    FILETIME ft;
    GetSystemTime (&st);
    SystemTimeToFileTime (&st, &ft);
    SPropValue spvProp = { 0 };
    BOOL fFlagChangedAndSaved = FALSE;

    // Lock the object for exclusive access.
    EnterCriticalSection (&m_csObj);

    HRESULT hResult = PrepareRecipsForSubmission();
    if (hResult)
    {
        goto ErrorExit;
    }
    m_pList->Find (PR_MESSAGE_FLAGS, &spvProp);
    if (spvProp.Value.l & MSGFLAG_SUBMIT)
	{
		ASSERTNOW();
        // TODO: Implement message re-submission.
        //          IrvingD 4/22/95
        goto ErrorExit;
	}
    spvProp.ulPropTag = PR_MESSAGE_FLAGS;
    spvProp.Value.l |= MSGFLAG_SUBMIT;
    m_pList->Insert (&spvProp);
    spvProp.ulPropTag = PR_CLIENT_SUBMIT_TIME;
    spvProp.Value.ft = ft;
    m_pList->Insert (&spvProp);
    hResult = SaveChanges (KEEP_OPEN_READWRITE);
    if (!hResult)
	{
        fFlagChangedAndSaved = TRUE;
		if (OBJECT_RUNNING_IN_SPOOLER & m_ulObjFlags)
        {
            ASSERTNOW();
            m_pStore->AddMsgToOutGoingQueue (ft, m_ObjectEID);
        }
        else
        {
            m_pStore->SubmitToSpoolerContextStore (ft, m_ObjectEID);
        }
    }
ErrorExit :
    if (hResult)
    {
        if (fFlagChangedAndSaved)
        {
            // If something failed, remove the submit bit from the message flags
            m_pList->Find (PR_MESSAGE_FLAGS, &spvProp);
            spvProp.Value.l &= ~MSGFLAG_SUBMIT;
            m_pList->Insert (&spvProp);
            SaveChanges (KEEP_OPEN_READWRITE);
        }
        TraceResult ("CMessage::SubmitMessage", hResult);
        SetLastError (hResult);
    }
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SetReadFlag()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method is called by client to togle the status of a message
//      object between READ and UNREAD and to send read receipt notifications
//      Clients may call this method with a combination of flags to suppress
//      or generate or suppress READ notification which are sent back to the
//      original sender of the message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::SetReadFlag (ULONG ulFlags)
{
    InfoTrace ("CMessage::SetReadFlag method called");
    Validate_IMessage_SetReadFlag (this, ulFlags);
    // Eliminate unsupported flags
    ulFlags &= ~MAPI_DEFERRED_ERRORS;
    // Call the internal method
    return InternalSetReadFlag (ulFlags, TRUE);
}

///////////////////////////////////////////////////////////////////////////////
// Other member functions specific to this class
//
///////////////////////////////////////////////////////////////////////////////
//    CMessage::InternalSetReadFlag()
//
//    Parameters
//      [IN]    ulFlags                     Flags for the call (See
//                                          IMessage::SetReadFlag())
//      [IN]    fSentFolderNotification     Wether or not we should sent
//                                          notification on the folder
//
//    Purpose
//      This method togles the status of a message object between READ and
//      UNREAD, sends read notification, or remote nofication requests bits
//      from the message flags on this message.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::InternalSetReadFlag (ULONG ulFlags,
                                            BOOL  fSentFolderNotification)
{
    HRESULT hResult = S_OK;
    BOOL fIncrementContentUnread = FALSE;
    CMAPIFolder * pFolder = NULL;

    // First lock the object for exclusive access
    EnterCriticalSection (&m_csObj);

    // Now get the PR_MESSAGE_FLAGS of this message object to examine the bits
    SPropValue spvFlags = { 0 };
    m_pList->Find (PR_MESSAGE_FLAGS, &spvFlags);

    long lOldFlags = spvFlags.Value.l;
    
    UPDATE_TABLE_ACTION Action = UPD_TABLE_REFRESH;

    if (OBJECT_NEVER_SAVED & m_ulObjFlags ||            // This does not work on unsaved items
        OBJECT_MESSAGE_IN_MESSAGE & m_ulObjFlags ||     // ... or embeeded messages
        OBJECT_DELETED_IN_CONTAINER & m_ulObjFlags)     // ... or deleted message
    {
        goto ExitNow;
    }

    if ((CLEAR_RN_PENDING | CLEAR_NRN_PENDING) & ulFlags)
    {
        if (CLEAR_RN_PENDING & ulFlags)
        {
            spvFlags.Value.l &= ~MSGFLAG_RN_PENDING;
        }
        if (CLEAR_NRN_PENDING & ulFlags)
        {
            spvFlags.Value.l &= ~MSGFLAG_NRN_PENDING;
        }
    }
    else
    {
        if (GENERATE_RECEIPT_ONLY & ulFlags)
        {
            // See if we have to generate a READ receipt message in this message
            SPropValue spvReadReceipt = { 0 };
            m_pList->Find (PR_READ_RECEIPT_REQUESTED, &spvReadReceipt);
            if (PR_READ_RECEIPT_REQUESTED != spvReadReceipt.ulPropTag ||
                FALSE == spvReadReceipt.Value.l ||
                MSGFLAG_UNSENT & spvFlags.Value.l) // watch that the user is not trying to confuse us
            {
                // Nothing to do.
                goto ExitNow;
            }
            // Looks like we have to create a Read Receipt. For this, we create a new
            // message in the ROOT folder. To optimize, we look in our list of open
            // objects and find out if we have the root folder open so we can ride on it.
            PRIVATE_ENTRYID eidRoot;
            eidRoot.uidGlobal = m_pStore->m_pBEobj->GetBackEndUID();
            eidRoot.bVersion = ENTRYID_VERSION;
            eidRoot.bObject = MAPI_FOLDER;
            eidRoot.dwObjID = ROOT_FOLDER_ID;
            pFolder = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (eidRoot, WALK_START, NULL);
            if (!pFolder)
            {
                // Since we don't have the root folder around, get its
                // properties from the back end storage.
                CPropList * pList = NULL;
                if (!m_pBEobj->OpenObject (eidRoot, &pList))
                {
                    pFolder = new CMAPIFolder (m_pStore,
                                               &eidRoot,
                                               &eidRoot,
                                               FOLDER_ROOT,
                                               pList,
                                               FALSE,
                                               MAPI_ACCESS_MODIFY |
                                               MAPI_ACCESS_DELETE |
                                               MAPI_ACCESS_READ |
                                               MAPI_ACCESS_CREATE_HIERARCHY |
                                               MAPI_ACCESS_CREATE_CONTENTS |
                                               MAPI_ACCESS_CREATE_ASSOCIATED,
                                               m_pStore->m_pNPList);
                    if (pFolder)
                    {
                        pFolder->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
                    }
                    else
                    {
                        delete pList;
                        hResult = E_OUTOFMEMORY;
                    }
                }
                else
                {
                    hResult = E_FAIL;
                }
            }
            if (pFolder)
            {
                // Now that we have the root folder object, create a new message in it.
                CMessage * pNewMessage;
                hResult = pFolder->CreateMessage (NULL, 0, (LPMESSAGE *)&pNewMessage);
                if (!hResult)
                {
                    // Call the MAPI support object and give it the new message. MAPI
                    // will do the copy of the properties.
                    hResult = m_pStore->m_pSupObj->ReadReceipt (0, this, (LPMESSAGE *)&pNewMessage);
                    if (!hResult)
                    {
                        hResult = pNewMessage->SubmitMessage (0);
                        if (!hResult)
                        {
                            // After the receipt has been generated mark the message
                            // so no more receipts are sent
                            spvReadReceipt.ulPropTag = PR_READ_RECEIPT_REQUESTED;
                            spvReadReceipt.Value.l = FALSE;
                            m_pList->Insert (&spvReadReceipt);

                            spvFlags.Value.l &= ~(MSGFLAG_RN_PENDING | MSGFLAG_NRN_PENDING);
                            m_pList->Insert (&spvFlags);
                        }
                    }
                    // Release the IMessage object
                    pNewMessage->Release();
                    if (hResult)
                    {
                        // If we failed to send the read receipt, then we delete the
                        // message from the folder where created it.
                        SBinary binEID;
                        binEID.cb = CB_PRIVATE_EID;
                        binEID.lpb = (LPBYTE)&pNewMessage->m_ObjectEID;
                        ENTRYLIST eidList;
                        eidList.cValues = 1;
                        eidList.lpbin = &binEID;
                        pFolder->DeleteMessages (&eidList, 0, NULL, 0);
                    }
                }
                pFolder->Release();
                pFolder = NULL;
            }
        }
        else
        {
            // If the caller wants to suppress notification receipts,
            // then remove the bit from the message flags.
            if (SUPPRESS_RECEIPT & ulFlags)
            {
                spvFlags.Value.l &= ~(MSGFLAG_RN_PENDING | MSGFLAG_NRN_PENDING);
            }
            // If this flags is set, the client wants to set the message to UNREAD
            if (CLEAR_READ_FLAG & ulFlags)
            {
                spvFlags.Value.l &= ~MSGFLAG_READ;
                Action = UPD_TABLE_REFRESH_ADD_UNREAD;
            }
            else
            {
                spvFlags.Value.l |= MSGFLAG_READ;
                Action = UPD_TABLE_REFRESH_DEL_UNREAD;
            }
        }
    }
    // In our implementation we only change the value of the property if there
    // was an actual change of state in the flags of the message
    if (lOldFlags != spvFlags.Value.l)
    {
        hResult = m_pList->Insert (&spvFlags);
        if (!hResult)
        {
            m_ulObjFlags |= ((MSGFLAG_READ & spvFlags.Value.l) ? 0 : OBJECT_UNREAD_MESSAGE);
            // Save the changes on this IMessage object if it is not a new object
            if (!(OBJECT_NEVER_SAVED & m_ulObjFlags))
            {
                CGenericProp::SaveChanges (KEEP_OPEN_READWRITE, NULL);
            }
        }
        // Update the unread content count of NON-associated messages
        // in the parent folder if the message has been saved
        if (!hResult &&
            !(OBJECT_ASSOCIATED & m_ulObjFlags) &&
            !(OBJECT_NEVER_SAVED & m_ulObjFlags))
        {
            pFolder = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ParentEID, WALK_START, NULL);
            if (pFolder)
            {
                pFolder->UpdateContentTables (Action,
                                              FALSE,  // The message is not associated
                                              OBJECT_UNREAD_MESSAGE & m_ulObjFlags, // Weather or not a message has been read
                                              &m_ObjectEID,
                                              m_pList);
                // ... We also need to update the hierarchy of the parent of
                // the parent folder of this message so that the new value for
                // PR_CONTENT_UNREAD, and PR_ASSOC_CONTENT_COUNT show up
                if (UPD_TABLE_REFRESH_ADD_UNREAD == Action || UPD_TABLE_REFRESH_DEL_UNREAD == Action)
                {    
                    pFolder->UpdateParentTable();
                    if (fSentFolderNotification)
                    {
                        // ... Finally, send a notification that some object properties have been altered (on the parent)
                        NOTIFICATION notif = { 0 };
                        notif.ulEventType         = fnevObjectModified;
                        notif.info.obj.ulObjType  = MAPI_FOLDER;
                        notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
                        notif.info.obj.lpEntryID  = (LPENTRYID)&pFolder->m_ObjectEID;
                        notif.info.obj.cbParentID = CB_PRIVATE_EID;
                        notif.info.obj.lpParentID = (LPENTRYID)&pFolder->m_ParentEID;
                        m_pStore->NotifyClients (&pFolder->m_ObjectEID, 1, &notif);
                    }
                }
                pFolder->Release();
            }
            if (!pFolder &&
                (UPD_TABLE_REFRESH_ADD_UNREAD == Action ||
                 UPD_TABLE_REFRESH_DEL_UNREAD == Action))
            {
                // If we could not get a hold of the parent folder, then we must open the
                // object from the back end storage and set the properties on the message
                SPropValue spvContentUnread = { 0 }, spvParentEID = { 0 };
                CPropList * pList = NULL;
                if (!m_pBEobj->OpenObject (m_ParentEID, &pList))
                {
                    pList->Find (PR_CONTENT_UNREAD, &spvContentUnread);
                    pList->Find (PR_PARENT_ENTRYID, &spvParentEID);
                    ASSERT (spvContentUnread.ulPropTag == PR_CONTENT_UNREAD);
                    ASSERT (spvParentEID.ulPropTag == PR_PARENT_ENTRYID);
                    PRIVATE_ENTRYID * pParentEID = (PRIVATE_ENTRYID *)spvParentEID.Value.bin.lpb;
                    // Change the PR_CONTENT_UNREAD count depending on the new state of the message (READ or UNREAD)
                    if (UPD_TABLE_REFRESH_ADD_UNREAD == Action)
                    {
                        spvContentUnread.Value.l++;
                    }
                    else
                    {
                        if (spvContentUnread.Value.l)
                        {
                            spvContentUnread.Value.l--;
                        }
                    }
                    pList->Insert (&spvContentUnread);
                    if (spvParentEID.ulPropTag == PR_PARENT_ENTRYID &&
                        spvContentUnread.ulPropTag == PR_CONTENT_UNREAD)
                    {
                        pFolder = new CMAPIFolder (m_pStore,
                                                   pParentEID,
                                                   &m_ParentEID,
                                                   FOLDER_GENERIC,
                                                   pList,
                                                   FALSE,
                                                   MAPI_ACCESS_MODIFY |
                                                   MAPI_ACCESS_DELETE |
                                                   MAPI_ACCESS_READ |
                                                   MAPI_ACCESS_CREATE_HIERARCHY |
                                                   MAPI_ACCESS_CREATE_CONTENTS |
                                                   MAPI_ACCESS_CREATE_ASSOCIATED,
                                                   m_pStore->m_pNPList);
                        if (pFolder)
                        {
                            pList = NULL;
                            pFolder->CommitChanges();

                            if (fSentFolderNotification)
                            {
                                // Send a notification that some object properties have been altered (on the parent)
                                NOTIFICATION notif = { 0 };
                                notif.ulEventType         = fnevObjectModified;
                                notif.info.obj.ulObjType  = MAPI_FOLDER;
                                notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
                                notif.info.obj.lpEntryID  = (LPENTRYID)&pFolder->m_ObjectEID;
                                notif.info.obj.cbParentID = CB_PRIVATE_EID;
                                notif.info.obj.lpParentID = (LPENTRYID)&pFolder->m_ParentEID;
                                m_pStore->NotifyClients (&pFolder->m_ObjectEID, 1, &notif);
                            }

                            pFolder->Release();
                            pFolder = NULL;
                        }
                        else
                        {
                            TraceMessage ("CMessage::SetReadFlag: Failed to create parent folder to commit changes");
                        }
                    }
                    if (pList)
                    {
                        delete pList;
                    }
                }
            }
        }
    }
ExitNow :
    LeaveCriticalSection (&m_csObj);
    SetLastError (hResult);
    TraceResult ("CMessage::SetReadFlag", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::LoadObjectTables()
//
//    Parameters
//      None
//
//    Purpose
//      This funciton is called during the object initialization to load up
//      the tables associated with an IMessage object. The tables are loaded
//      during initialization so that other instances of this message are not
//      affected when one commits it changes to disk.
//
//    Return Value
//      None
//
void WINAPI CMessage::LoadObjectTables()
{
    ASSERT (!m_pAttTable);
    HRESULT hResult = m_pBEobj->GetTable (TABLE_ATTACHMENTS,
                                          m_ObjectEID,
                                          m_lAccess,
                                          (LPVOID)&m_pStore->m_StoreEID,
                                          NULL,
                                          &m_pAttTable);
    TraceResult ("CMessage::LoadObjectTables: Failed to get the attachment table", hResult);
    BOOL fReturnEmptyTable = ((OBJECT_NEVER_SAVED & m_ulObjFlags) ? TRUE : FALSE);
    ASSERT (!m_pRecipTable);
    hResult = m_pBEobj->GetTable (TABLE_RECIPIENTS,
                                  m_ObjectEID,
                                  0,
                                  (LPVOID)&m_dwRecipRowID,
                                  (LPVOID)FALSE,
                                  &m_pRecipTable);
    TraceResult ("CMessage::LoadObjectTables: Failed to get the recipient table", hResult);
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::GenerateDisplayNames()
//
//    Parameters
//      [IN]    pTable      An IMAPITable pointer for the recipient table
//
//    Purpose
//      This function generates the strings for the PR_DISPLAY_xxx properties
//      on the message based on changes to the recipient table through
//      IMessage::ModifyRecipients()
//      When recipients are added, deleted, or modified, this function
//      recomputes the values for the forementioned properties so the new
//      value is available in a GetProps() call immediately following the
//      recipient table modification call.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::GenerateDisplayNames (LPMAPITABLE pTable)
{
    LPTSTR pszToNames = NULL, pszCcNames = NULL, pszBccNames = NULL;
    LPSRowSet pRows;
    SizedSPropTagArray(2, sptProps) =
    {
        2,
        {
            PR_RECIPIENT_TYPE,
            PR_DISPLAY_NAME
        }
    };
    // Get All the recipient from the recipient table
    HRESULT hResult = HrQueryAllRows (pTable,
                                      (LPSPropTagArray)&sptProps,
                                      NULL,
                                      NULL,
                                      0,
                                      &pRows);
    if (hResult)
    {   
        TraceResult ("CMessage::GenerateDisplayNames: Failed to get rows", hResult);
        return hResult;
    }
    LPSPropValue pProps;
    for (ULONG i=0; i<pRows->cRows; i++)
    {
        pProps = pRows->aRow[i].lpProps;
        ASSERT (pProps && pRows->aRow[i].cValues == 2)
        ASSERT (pProps[0].ulPropTag == PR_RECIPIENT_TYPE);
        ASSERT (pProps[1].ulPropTag == PR_DISPLAY_NAME);
        switch (pProps[0].Value.l)
        {
            case MAPI_TO :
                AddNameToString (pProps[1].Value.LPSZ, &pszToNames);
                break;
            case MAPI_CC :
                AddNameToString (pProps[1].Value.LPSZ, &pszCcNames);
                break;
            case MAPI_BCC :
                AddNameToString (pProps[1].Value.LPSZ, &pszBccNames);
                break;
            default :
                ASSERTNOW();
                break;
        }
    }
    FreeProws (pRows);

    // Set the new properties
    SPropValue spvProp = { 0 };
    spvProp.ulPropTag = PR_DISPLAY_TO;
    spvProp.Value.LPSZ = pszToNames ? pszToNames : "";
    m_pList->Insert (&spvProp);
    PrivFree ((LPVOID *)&pszToNames);

    spvProp.ulPropTag = PR_DISPLAY_CC;
    spvProp.Value.LPSZ = pszCcNames ? pszCcNames : "";
    m_pList->Insert (&spvProp);
    PrivFree ((LPVOID *)&pszCcNames);

    spvProp.ulPropTag = PR_DISPLAY_BCC;
    spvProp.Value.LPSZ = pszBccNames ? pszBccNames : "";
    m_pList->Insert (&spvProp);
    PrivFree ((LPVOID *)&pszBccNames);

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::AddNameToString()
//
//    Parameters
//      [IN]    pszNewName
//      [OUT]   ppszNameList
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::AddNameToString (LPTSTR pszNewName, LPTSTR * ppszNameList)
{
    ULONG cbSize = Cbtszsize (pszNewName);
    // If the string is empty, only space for a NULL terminator was
    // calculate. We don't want to add an empty string to the list.
    if (cbSize == sizeof(char))
    {
        return S_OK;
    }
    if (*ppszNameList)
    {
        cbSize += Cbtszsize (*ppszNameList) + sizeof("; ");
    }
    LPTSTR pszNewList;
    HRESULT hResult = PrivAllocate (cbSize, (LPVOID *)&pszNewList, "TO/CC/BCC Computed String");
    if (hResult)
    {
        TraceResult ("CMessage::AddNameToString: Failed to allocate memory for new string", hResult);
        return hResult;
    }
    if (*ppszNameList)
    {
        lstrcpy (pszNewList, *ppszNameList);
        lstrcat (pszNewList, "; ");
        lstrcat (pszNewList, pszNewName);
        PrivFree ((LPVOID *)ppszNameList);        
    }
    else
    {
        lstrcpy (pszNewList, pszNewName);
    }
    *ppszNameList = pszNewList;
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::SetNormalizedSubject()
//
//    Parameters
//      [IN]    pszSubject
//      [OUT]   ppNormSubject
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMessage::SetNormalizedSubject (LPTSTR   pszSubject,
                                             LPTSTR * ppNormSubject)
{
    // TODO: This function is not *really* normalizing the subject.
    HRESULT hResult = S_OK;
    if (NULL == *ppNormSubject)
    {
        LPTSTR pStr;
        hResult = PrivAllocate (Cbtszsize(pszSubject), (LPVOID *)&pStr, "Normalized String Text");
        if (hResult)
        {
            TraceResult ("CMessage::SetNormalizedSubject: Failed to allocate memory for new string", hResult);
            return hResult;
        }
        *ppNormSubject = pStr;
    }
    lstrcpy (*ppNormSubject, pszSubject);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::UpdateAttachmentTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CMessage::UpdateAttachmentTable (CPropList *         pList,
                                              PRIVATE_ENTRYID     AttachEID,
                                              UPDATE_TABLE_ACTION utaChange)
{
    HRESULT hResult = S_OK;
    LPSPropValue pProps = NULL;
    LPMAPITABLE pView;
    DWORD dwLastAllocated = 0;
    DWORD dwAttCount;
    BOOL fSkipAdd = FALSE;
    SPropValue spvFlags = { 0 };
    m_pList->Find (PR_MESSAGE_FLAGS, &spvFlags);
    if (UPD_TABLE_ADD_ROW == utaChange)
    {
        spvFlags.Value.l |= MSGFLAG_HASATTACH;
        m_ulObjFlags |= OBJECT_HAS_ATTACHMENT;
        m_pList->Insert (&spvFlags);
    }
    if (!m_pAttTable)
    {
        if (UPD_TABLE_ADD_ROW == utaChange)
        {
            goto ErrorExit; // No need to add it since there is no table in use
        }
        fSkipAdd = TRUE; // Set the flags to release the table table after the function is done
    }
    hResult = GetAttachmentTable (0, &pView);
    if (!hResult)
    {
        hResult = pView->GetRowCount (0, &dwAttCount);
        pView->Release();
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    if (UPD_TABLE_DELETE_ROW == utaChange)
    {
        ASSERT (MSGFLAG_HASATTACH & spvFlags.Value.l);
        if (dwAttCount)
        {
            dwAttCount--;
        }
        if (0 == dwAttCount)
        {
            spvFlags.Value.l &= ~MSGFLAG_HASATTACH;
            m_ulObjFlags &= ~OBJECT_HAS_ATTACHMENT;
            m_pList->Insert (&spvFlags);
        }
    }

    // Now update the attachment table on this object
    hResult = m_pBEobj->GetTableProps (TABLE_ATTACHMENTS,
                                       m_lAccess,
                                       NULL,
                                       &AttachEID,
                                       pList,
                                       &pProps,
                                       dwLastAllocated,
                                       m_pAttTable,
                                       (LPVOID)&m_pStore->m_StoreEID,
                                       NULL);
    PrivFree ((LPVOID *)&pProps);
ErrorExit:
    if (fSkipAdd && m_pAttTable)
    {
        m_pAttTable->Release();
        m_pAttTable = NULL;
    }
    TraceResult ("CMessage::UpdateAttachmentTable", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMessage::PrepareRecipsForSubmission()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CMessage::PrepareRecipsForSubmission()
{
    LPMAPITABLE pTableView;
    HRESULT hResult = S_OK;
    ULONG ulRecipCount, ulPrepareSubmitFlags = 0, ulExpandRecipsFlags = 0;
    if (!m_pRecipTable)
    {
        hResult = GetRecipientTable (0, &pTableView);
        if (!hResult)
        {
            pTableView->Release();
        }
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    hResult = m_pRecipTable->HrGetView (NULL, NULL, 0, &pTableView);
    if (!hResult)
    {
        hResult = pTableView->GetRowCount (0, &ulRecipCount);
        if (!hResult)
        {
            if (0 == ulRecipCount)
            {
                hResult = MAPI_E_NO_RECIPIENTS;
            }
            else
            {
                hResult = HrAddColumns (pTableView,
                                        (LPSPropTagArray)&sptSubmitRecipTable,
                                        GetMAPIAllocBuffer(),
                                        GetMAPIFreeBuffer());
            }
            BOOL fContinue = TRUE;
            LPSRowSet pRows;
            ULONG ulRow;
            while (fContinue && !hResult)
            {
                hResult = pTableView->QueryRows (15, 0, &pRows);
                if (!hResult)
                {
                    if (pRows->cRows)
                    {
                        for (ulRow=0; ulRow<pRows->cRows; ulRow++)
                        {
                            ASSERT (PR_ROWID == pRows->aRow[ulRow].lpProps[0].ulPropTag);
                            // We don't want responsibility for any recipient. If this isn't
                            // the spooler calling, then force ALL responsibilities to FALSE
                            // regardless of what they were
                            if (PR_RESPONSIBILITY != pRows->aRow[ulRow].lpProps[1].ulPropTag ||
                                !(m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER))
                            {
                                pRows->aRow[ulRow].lpProps[1].ulPropTag = PR_RESPONSIBILITY;
                                pRows->aRow[ulRow].lpProps[1].Value.b = FALSE;
                                m_pRecipTable->HrModifyRows (0, pRows);
                            }
                        }
                    }
                    else
                    {
                        fContinue = FALSE;
                    }
                    FreeProws (pRows);
                }
            }
        }
        pTableView->Release();
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    SPropValue spvSubmitFlags;
    hResult = m_pStore->m_pSupObj->PrepareSubmit (this, &ulPrepareSubmitFlags);
    if (!hResult)
    {
        hResult = m_pStore->m_pSupObj->ExpandRecips (this, &ulExpandRecipsFlags);
        if (!hResult)
        {
            spvSubmitFlags.ulPropTag = PR_SUBMIT_FLAGS;
            if (NEEDS_PREPROCESSING & ulExpandRecipsFlags)
            {
		        spvSubmitFlags.Value.l = SUBMITFLAG_PREPROCESS;
            }
	        else
            {
		        spvSubmitFlags.Value.l = 0;
            }
            m_pList->Insert (&spvSubmitFlags);
        }
    }
    if (!hResult)
    {
        m_fRecipTableDirty = TRUE;
    }
ErrorExit :
    TraceResult ("CMessage::PrepareRecipsForSubmission", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    MessageTablesViewRelease()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDAPI_(void) MessageTablesViewRelease (ULONG       ulCallBackData,
                                        LPTABLEDATA pTableData,
                                        LPMAPITABLE pTableView)
{
    DWORD * pdwViewsLeft = NULL;
    CMessage * pMessage = (CMessage *)ulCallBackData;
    __try
    {
        if (!pMessage || 0 == pMessage->m_cRef)
        {
            TraceMessage ("MessageTablesViewRelease: Invalid Message Object");
            return;
        }
        EnterCriticalSection (&pMessage->m_csObj);
        if (pMessage->m_pRecipTable == pTableData)
        {
            pdwViewsLeft = &(pMessage->m_dwRecipTableViews);
        }
        else
        {
            if (pMessage->m_pAttTable == pTableData)
            {
                pdwViewsLeft = &(pMessage->m_dwAttTableViews);
            }
            else
            {
                TraceMessage ("MessageTablesViewRelease: Called for an unknown table");
            }
        }
        if (pdwViewsLeft)
        {
            ASSERT (0 != (*pdwViewsLeft));
            --(*pdwViewsLeft);
        }
        LeaveCriticalSection (&pMessage->m_csObj);
        pMessage->Release();
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("MessageTablesViewRelease: An exception occured. Exception code: %#08x", GetExceptionCode());
        TraceMessage ("     MSLMS: WARNING - Possible deadlock if sync object was not released due to exception");
    }
}

// End of file for MSMSG.CPP
