///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSWIZARD.CPP
//
//  Description
//      This file implements some of the entry point for the provider Profile
//      Wizard installation. The functions in this file are:
//
//      MAPI SPI entry-points functions:
//          WizardEntry()
//          WizardDlgProc()
//
//      Helper functions:
//          TogglePage()
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.0
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSLMS.H"
#include <WINDOWSX.H>
#include <MAPIWZ.H>

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

HINSTANCE ghUIControl = NULL;
typedef BOOL (WINAPI CTL3DFUNCTION) (HINSTANCE hInstance);
typedef CTL3DFUNCTION *LPCTL3DFUNCTION;
typedef BOOL (WINAPI CTL3DSUBCLASS) (HWND hDlg, DWORD dwFlags);
typedef CTL3DSUBCLASS *LPCTL3DSUBCLASS;

LPCTL3DFUNCTION pfn3DRegister = NULL, pfn3DUnregister = NULL, pfn3DAutoSubclass = NULL;
LPCTL3DSUBCLASS pfnDlgSubclassEx = NULL;

// Prototypes of the function in this module
extern "C"
{
    WIZARDENTRY WizardEntry;
    SERVICEWIZARDDLGPROC WizardDlgProc;
};

void WINAPI TogglePage (HWND, BOOL);

// Global data used in the WizardDlgProc function
LPMAPIPROP pLocalPropObj = NULL;
char szStorePath[_MAX_PATH];
SPropValue spProp;
HINSTANCE hInstance = NULL;

///////////////////////////////////////////////////////////////////////////////
//    WizardEntry()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Returns the name of the resource where the controls for the wizard
//      pages are defined and outlined. The Wizard gives us a pointer to an
//      IMAPIProp object. We will release this object after we are finished
//      in the dialog procedure of our configuration pages.
//
//    Return Value
//      S_OK    Always
//
ULONG STDAPICALLTYPE WizardEntry (HINSTANCE  hProviderDLLInstance,
                                  LPTSTR *   ppcsResourceName,
                                  DLGPROC *  ppDlgProc,
                                  LPMAPIPROP pMapiProp,
                                  LPVOID     pMapiSupportObject)
{
    InfoTrace ("WizardEntry function called");
    // Quick sanity check on the input parameters
    ASSERT (hProviderDLLInstance);
    ASSERT (ppcsResourceName);
    ASSERT (pMapiProp);
    ASSERT (ppDlgProc);
    // The resource name needs to be an actual text string not
    // something generated with the MAKEINTRESOURCE() macro
    const static char szTemplate[] = "WIZARD_PAGES";
    *ppcsResourceName = (LPTSTR)szTemplate;
    *ppDlgProc = (DLGPROC)WizardDlgProc;
    pLocalPropObj = pMapiProp;
    pLocalPropObj->AddRef();
    hInstance = hProviderDLLInstance;
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    WizardDlgProc()
//
//    Parameters
//      { Refer to MAPI and Win32 SDK Documentation on this method }
//
//    Purpose
//      This is the function that the profile wizard will call while
//      interacting with our configuration pages.
//
//    Return Value
//      TRUE if we handle a message, FALSE if we don't or there was an error
//      in the data
//
BOOL STDAPICALLTYPE WizardDlgProc (HWND   hDlg,
                                   UINT   message,
                                   WPARAM wParam,
                                   LPARAM lParam)
{
    static BOOL fInitalizeControl = TRUE;
    static int iPage = 1;
    switch (message)
    {
        case WM_INITDIALOG:
            // If the control have not been initialized do so
    		if (fInitalizeControl)
    		{
                // Set the flag to avoid re-initialization
    			fInitalizeControl = FALSE;
                // Limit the maximum number of character users can put in the edit
    			Edit_LimitText (GetDlgItem (hDlg, IDC_STOREPATH), _MAX_PATH);
                PrivInitialize3DCtl (hInstance);
    		}
    		return TRUE;

        case WIZ_QUERYNUMPAGES :
            // Return the number of pages the our provider needs to display
            return (BOOL)1;

        case WM_CLOSE :
            PrivUninitialize3DCtl (hInstance);
            pLocalPropObj->Release();
            pLocalPropObj = NULL;
		    return TRUE;

        case WM_COMMAND :
		    switch (LOWORD(wParam))
		    {
                case IDC_BROWSE :
                    if (BrowseFileDialog (hInstance, hDlg, szStorePath))
                    {
                        SetWindowText (GetDlgItem (hDlg, IDC_STOREPATH), szStorePath);
                    }
                    return TRUE;

                case WIZ_PREV:
                    // Disable the page we were in and enable the next one
                    // before it is display by the wizard
                	if (2 == iPage)
                    {
                        TogglePage (hDlg, TRUE);
                        iPage--;
                    }
                	return TRUE;

                case WIZ_NEXT :
                    if (1 == iPage)
                    {
                        SetFocus (GetDlgItem (hDlg, IDC_STOREPATH));
                        TogglePage (hDlg, TRUE);
                        iPage++;
                        return TRUE;
                    }
                    if (!GetWindowText (GetDlgItem (hDlg, IDC_STOREPATH), szStorePath, _MAX_PATH))
                    {
                        break;
                    }
                    spProp.ulPropTag = PR_SMP_STORE_PATH;
                    spProp.Value.lpszA = szStorePath;

                    // Since this is the last page in our configuration,
                    // we must save the obtained data into the
                    // IMAPIProp object that the wizard gave us. In
                    // turn this data will be saved in the profile
                    // section of this provider
                    ASSERT (pLocalPropObj);
                    if (pLocalPropObj)
                    {
        			    HRESULT hResult = pLocalPropObj->SetProps (1, &spProp, NULL);
                        if (FAILED(hResult))
                        {
                            // Something went wrong
                            TraceResult ("WizardDlgPrc: Failed to set the prop", hResult);
                            break;
                        }
                    }
                    return TRUE;
		    }
		    break;
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    TogglePage()
//
//    Parameters
//      [IN]    hDlg    Handle to the dialogs where the controls are located
//      [IN]    fState  State of the controls: TRUE shows them,
//                      FALSE hides them
//
//    Purpose
//      This function toggles the controls in a wizard page from active
//      to inactive and viceversa.
//
//    Return Value
//      None
//
void WINAPI TogglePage (HWND hDlg, BOOL fState)
{
    // Toggle the controls ON or OFF
    EnableWindow (GetDlgItem (hDlg, IDC_STOREPATH), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_STOREPATH), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_BROWSE), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_BROWSE), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_STATIC1), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_STATIC1), fState ? SW_SHOW : SW_HIDE);
    EnableWindow (GetDlgItem (hDlg, IDC_INFO), fState);
    ShowWindow   (GetDlgItem (hDlg, IDC_INFO), fState ? SW_SHOW : SW_HIDE);
}

///////////////////////////////////////////////////////////////////////////////
//    PrivateMessageBox()
//
//    Parameters
//      [IN]    ids             ID of a string in the resource string table
//      [IN]    hOwnerWnd       Window which will own the Message Box
//      [IN]    uFlags          Flags (options) for the message box
//      [IN]    hInst           Handle of the instance where the resources are
//      
//    Purpose
//      This function display a message box with a string loaded from the
//      string table of this app. The ID of the string is passed in the ids
//      parameter. The dialog box is modal with respect to the window
//      identified in hOwnerWnd. The options to display the dialog box are
//      passed in uFlags
//      
//    Return Value
//      ID of the button pressed in the message box dialog box
//      
int WINAPI PrivateMessageBox (UINT      ids,
                              HWND      hOwnerWnd,
                              UINT      uFlags,
                              HINSTANCE hInst)
{
    // Default response
	int nResponse = -1;
    // Get the string from the string table. The size of the buffer is the
    // maximum number of character allowed in the character buffer, without
    // the accounting for the NULL terminator
    if (NULL == hInst)
    {
        hInst = hInstance;
    }
    char szBuffer[256];
    if (LoadString (hInst, ids, szBuffer, 255))
    {
        // Show the message box and get the ID of the button pressed
        nResponse = MessageBox (hOwnerWnd,
                                szBuffer,
                                PROVIDER_DISPLAY_NAME_STRING,
                                uFlags);
    }
    return nResponse;
}

///////////////////////////////////////////////////////////////////////////////
//    BrowseFileDialog()
//
//    Parameters
//      [IN]        hInstance           Instance that owns the common dialog
//      [IN]        hOwnerWnd           Window that owns the common dialog
//      [IN,OUT]    szFileNameBuffer    Buffer where we return the
//                                      selected or new file 
//
//    Purpose
//      This function displays a common dialog for files so that the use may
//      select a message store file to use. This is displayed during the
//      wizard configuration or during the provider configuration
//      in ServiceEntry.
//
//    Return Value
//      TRUE if the user selected a valid file. FALSE if not or if cancelled
//
BOOL WINAPI BrowseFileDialog (HINSTANCE hInstance,
                              HWND      hOwnerWnd,
                              LPTSTR    szFileNameBuffer)
{
    ZeroMemory (szFileNameBuffer, _MAX_PATH);
    // Initialize the OPENFILENAME structure
   	OPENFILENAME ofn = { 0 };
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner   = hOwnerWnd;
    ofn.lpstrFilter = "Message Files (*.MSF)\0*.MSF\0All Files (*.*)\0*.*\0";
    ofn.lpstrFile   = szFileNameBuffer;
    ofn.nMaxFile    = _MAX_PATH;
    ofn.lpstrTitle  = "Find Message Store File";
    ofn.Flags       = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
    ofn.hInstance   = hInstance;
    ofn.lpstrDefExt = "MSF";

    // Display the dialog box now
    BOOL bReturn = GetSaveFileName(&ofn);
    CharUpper (szFileNameBuffer);
    return bReturn;
}

///////////////////////////////////////////////////////////////////////////////
//    PrivInitialize3DCtl()
//
//    Parameters
//      hInstance   Handle to the instance the control 3d library will
//                  subclass
//
//    Purpose
//      This function dynamically binds to the CTL3D library for UI apperance
//      on Windows NT. On Windows 95 the library may or may not be on the
//      system. We do this to have a single binary file that works on
//      Windows NT and Windows 95 and ensure that the WINDS service DLL loads
//      correctly in both platforms.
//      The CTL3D provider a less disparate user interface between the two OS's
//
//    Return Value
//      None
//
void WINAPI PrivInitialize3DCtl (HINSTANCE hInstance)
{
    if (NULL == ghUIControl)
    {
        ghUIControl = LoadLibrary (TEXT("CTL3D32.DLL"));
        if (ghUIControl)
        {
            pfn3DRegister = (LPCTL3DFUNCTION)GetProcAddress (ghUIControl, TEXT("Ctl3dRegister"));
            pfn3DUnregister = (LPCTL3DFUNCTION)GetProcAddress (ghUIControl, TEXT("Ctl3dUnregister"));
            pfn3DAutoSubclass = (LPCTL3DFUNCTION)GetProcAddress (ghUIControl, TEXT("Ctl3dAutoSubclass"));
            pfnDlgSubclassEx = (LPCTL3DSUBCLASS)GetProcAddress (ghUIControl, TEXT("Ctl3dSubclassDlgEx"));
            
            if (pfn3DRegister &&
                pfn3DUnregister &&
                pfn3DAutoSubclass &&
                pfnDlgSubclassEx)
            {
                pfn3DRegister (hInstance);
                pfn3DAutoSubclass (hInstance);
            }
            else
            {
                PrivUninitialize3DCtl (hInstance);
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//    PrivUninitialize3DCtl()
//
//    Parameters
//      hInstance   Handle to the instance the control 3d library will
//                  de-attach from.
//
//    Purpose
//      Unregister the 3D UI control stuff and unload the library of the
//      3D controls.
//
//    Return Value
//      None
//
void WINAPI PrivUninitialize3DCtl (HINSTANCE hInstance)
{
    if (ghUIControl)
    {
        if (pfn3DUnregister)
        {
            pfn3DUnregister (hInstance);
        }
        FreeLibrary (ghUIControl);
        ghUIControl = NULL;
        pfn3DRegister = NULL;
        pfn3DUnregister = NULL;
        pfn3DAutoSubclass = NULL;
        pfnDlgSubclassEx = NULL;
    }
}

// End of file for MSWIZARD.CPP
