///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      NAMEPROP.CPP
//
//  Description
//      This file implements a generic named property manager that can be used
//      independent of the a specific IMAPIProp implementation
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "NAMEPROP.H"

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::CNamePropMgr()
//
//    Parameters
//      [IN]    pfnAllocateBuffer       Pointer to MAPI memory function
//      [IN]    pfnAllocateMore         Pointer to MAPI memory function
//      [IN]    pfnFreeBuffer           Pointer to MAPI memory function
//
//    Purpose
//      Initialize the object to a known state. Set the data members
//      to valid values.
//
//    Return Value
//      None
//
CNamePropMgr::CNamePropMgr (LPALLOCATEBUFFER pfnAllocateBuffer,
                            LPALLOCATEMORE   pfnAllocateMore,
                            LPFREEBUFFER     pfnFreeBuffer)
{
    // Save the pointer to the MAPI memory functions
    m_pfnAllocateBuffer = pfnAllocateBuffer;
    m_pfnAllocateMore = pfnAllocateMore;
    m_pfnFreeBuffer = pfnFreeBuffer;
    // Initialize the member variables
    m_bMappingChanged = FALSE;
    InitializeCriticalSection (&m_csObj);
    m_pNameTable = NULL;
    m_pMapData = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::~CNamePropMgr()
//
//    Parameters
//      None
//
//    Purpose
//      Release any internally allocated data and contained objects
//
//    Return Value
//      None
//
CNamePropMgr::~CNamePropMgr()
{
    // Release the objects if they have been allocated
    if (m_pNameTable)
    {
        m_pNameTable->Release();
    }
    if (m_pMapData)
    {
        m_pMapData->Release();
    }
    DeleteCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::GetTableView()
//
//    Parameters
//      None
//
//    Purpose
//      Makes sure we have been initialize by loading the named property
//      MAPI from the underlying data storage. When the map is loaded, we
//      allocate an ITableData. In this function, we get an IMAPITable
//      interface to view the name property map. The view of the data is
//      save in a data member of the object, so this function should only
//      execute completely once.
//
//    Return Value
//      An HRESULT
//
inline STDMETHODIMP CNamePropMgr::GetTableView()
{
    HRESULT hResult = S_OK;
    // If we already have a table view, skip it and get out of here.
    if (!m_pNameTable)
    {
        // Make sure we have an ITableData interface around
        if (m_pMapData)
        {
            // Get a view of the data
            hResult = m_pMapData->HrGetView (NULL, NULL, 0, &m_pNameTable);
            if (hResult)
            {
                TraceResult ("CNamePropMgr::GetTableView: Failed to get a view", hResult);
            }
            else
            {
                // Just for kicks, make sure we have the columns of the
				// view in a known state (order)
                m_pNameTable->SetColumns ((LPSPropTagArray)&sptNPProps, 0);
            }
        }
        else
        {
            TraceMessage ("CNamePropMgr::GetTableView: The mapping has not been loaded yet");
            hResult = MAPI_E_NOT_INITIALIZED;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::InitializeForInMemoryUse()
//
//    Parameters
//      None
//
//    Purpose
//      For providers that need to use the name property manager in memory and
//      will not load the map from an IStream, they must call this method
//      prior to using the CNamePropMgr object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::InitializeForInMemoryUse()
{
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    if (!m_pMapData)
    {
        hResult = CreateTable (&IID_IMAPITableData,
                               m_pfnAllocateBuffer,
                               m_pfnAllocateMore,
                               m_pfnFreeBuffer,
                               NULL,
                               TBLTYPE_DYNAMIC,
                               PR_NAMED_PROP_ID,
                               (LPSPropTagArray)&sptNPProps,
                               &m_pMapData);
        if (!hResult)
        {
            m_ulNextID = 0x8000;
        }
    }
    LeaveCriticalSection (&m_csObj);
    if (hResult)
    {
        TraceResult ("CNamePropMgr::InitializeForInMemoryUse", hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::LoadMappingFromStream()
//
//    Parameters
//      [IN]    pStream     Pointer to an IStream interface where the mapping
//                          has been saved.
//
//    Purpose
//      Load the mapping of the named properties from the permanent storage
//      onto memory.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::LoadMappingFromStream (LPSTREAM pStream)
{
    ASSERTMSG (pStream, "Where is the stream I asked for?");
    EnterCriticalSection (&m_csObj);
    ULONG cbRead, ulMapCount;
    STORED_NAMED_PROP snpEntry;
    LPWSTR wstrName = NULL;
    DWORD dwLastAlloc = 0;
    SRow srRow = { 0 };
    SPropValue spvProps[NAMED_PROPERTIES_TABLE_PROPS - 1] = { 0 };
    srRow.cValues = sizeof(spvProps)/sizeof(SPropValue);
    srRow.lpProps = spvProps;

    spvProps[0].ulPropTag = PR_NAMED_PROP_GUID;
    spvProps[0].Value.lpguid = (LPGUID)&(snpEntry.uidPropSet);
    spvProps[1].ulPropTag = PR_NAMED_PROP_ID;
    spvProps[2].ulPropTag = PR_NAMED_PROP_KIND;
    
    HRESULT hResult = pStream->Read (&m_ulNextID, sizeof(ULONG), &cbRead);
    if (!hResult)
    {
        hResult = CreateTable (&IID_IMAPITableData,
                               m_pfnAllocateBuffer,
                               m_pfnAllocateMore,
                               m_pfnFreeBuffer,
                               NULL,
                               TBLTYPE_DYNAMIC,
                               PR_NAMED_PROP_ID,
                               (LPSPropTagArray)&sptNPProps,
                               &m_pMapData);
        if (!hResult)
        {
            if (sizeof(ULONG) == cbRead)
            {
                hResult = pStream->Read (&ulMapCount, sizeof(ULONG), NULL);
                if (!hResult)
                {
                    while (ulMapCount && !hResult)
                    {
                        hResult = pStream->Read (&snpEntry, sizeof(STORED_NAMED_PROP), NULL);
                        if (!hResult)
                        {
                            spvProps[1].Value.l = snpEntry.ulPropID;
                            spvProps[2].Value.l = snpEntry.ulKind;
                            if (MNID_STRING == snpEntry.ulKind)
                            {
                                spvProps[3].ulPropTag = PR_NAMED_PROP_NAME;
                                if (dwLastAlloc < (DWORD)snpEntry.lIDNameOrSize)
                                {
                                    if (wstrName)
                                    {
                                        wstrName = (LPWSTR)HeapReAlloc (GetProcessHeap(), 0, wstrName, snpEntry.lIDNameOrSize);
                                    }
                                    else
                                    {
                                        wstrName = (LPWSTR)HeapAlloc (GetProcessHeap(), 0, snpEntry.lIDNameOrSize);
                                    }
                                    if (wstrName)
                                    {
                                        dwLastAlloc = snpEntry.lIDNameOrSize;
                                    }
                                    else
                                    {
                                        hResult = E_OUTOFMEMORY;
                                    }
                                }
                                if (!hResult)
                                {
                                    hResult = pStream->Read (wstrName, snpEntry.lIDNameOrSize, NULL);
                                    spvProps[3].Value.lpszW = wstrName;
                                }
                            }
                            else
                            {
                                spvProps[3].ulPropTag = PR_NAMED_PROP_IDNAME;
                                spvProps[3].Value.l = snpEntry.lIDNameOrSize;
                            }
                            if (!hResult)
                            {
                                hResult = m_pMapData->HrModifyRow (&srRow);
                            }
                        }
                        if (ulMapCount)
                        {
                            ulMapCount--;
                        }
                    }
                }
            }
            else
            {
                // The mapping has never been initialized. Do so now. The first ID
                // available for a mapping will be the first value in the named
                // properties range.
                m_ulNextID = 0x8000;
                m_bMappingChanged = TRUE;
            }
        }
    }
    if (wstrName)
    {
        HeapFree (GetProcessHeap(), 0, wstrName);
    }
    if (hResult)
    {
        if (m_pMapData)
        {
            m_pMapData->Release();
            m_pMapData = NULL;
        }
        TraceResult ("CNamePropMgr::LoadMappingFromStream", hResult);
    }
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::SaveMappingToStream()
//
//    Parameters
//      [IN]    pStream     Pointer to an IStream interface where the mapping
//                          is saved to.
//
//    Purpose
//      This function writes the contents of the name property mapping to an
//      IStream interface for permanent storage.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::SaveMappingToStream (LPSTREAM pStream)
{
    ASSERTMSG (pStream, "Where is the stream I asked for?");
    HRESULT hResult = S_OK;
    ULONG ulRows;
    BOOL fContinue = TRUE;
    LPSRowSet pRows;

    STORED_NAMED_PROP snpEntry;
    
    EnterCriticalSection (&m_csObj);
    hResult = GetTableView();
    if (hResult)
    {
        goto ErrorExit;
    }

    hResult = m_pNameTable->SeekRow (BOOKMARK_BEGINNING, 0, NULL);
    if (!hResult)
    {
        hResult = pStream->Write (&m_ulNextID, sizeof(ULONG), NULL);
        {
            hResult = m_pNameTable->GetRowCount (0, &ulRows);
            if (!hResult && ulRows)
            {
                hResult = pStream->Write (&ulRows, sizeof(ULONG), NULL);
                while (fContinue && !hResult)
                {
                    hResult = m_pNameTable->QueryRows (25, 0, &pRows);
                    if (!hResult)
                    {
                        if (pRows->cRows)
                        {
                            LPSPropValue pProps;
                            for (ULONG i=0; i<pRows->cRows && !hResult; i++)
                            {
                                pProps = pRows->aRow[i].lpProps;
                                ASSERT (PR_NAMED_PROP_GUID == pProps[NP_GUID].ulPropTag);
                                ASSERT (PR_NAMED_PROP_ID == pProps[NP_ID].ulPropTag);
                                ASSERT (PR_NAMED_PROP_KIND == pProps[NP_KIND].ulPropTag);

                                CopyMemory (&(snpEntry.uidPropSet), pProps[NP_GUID].Value.lpguid, sizeof(MAPIUID));
                                snpEntry.ulPropID = pProps[NP_ID].Value.l;
                                snpEntry.ulKind = pProps[NP_KIND].Value.l;
                                if (MNID_STRING == snpEntry.ulKind)
                                {
                                    snpEntry.lIDNameOrSize = (lstrlenW (pProps[NP_NAME].Value.lpszW) + 1) * sizeof(WCHAR);
                                }
                                else
                                {
                                    snpEntry.lIDNameOrSize = pProps[NP_IDNAME].Value.l;
                                }
                                hResult = pStream->Write (&snpEntry, sizeof(STORED_NAMED_PROP), NULL);
                                if (!hResult && (MNID_STRING == snpEntry.ulKind))
                                {
                                    hResult = pStream->Write (pProps[NP_NAME].Value.lpszW,
                                                              snpEntry.lIDNameOrSize,
                                                              NULL);
                                }
                            }
                        }
                        else
                        {
                            fContinue = FALSE;
                        }
                        FreeProws (pRows);
                    }
                }
                pStream->Commit (STGC_DEFAULT);
            }
        }
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::GetIDForName()
//
//    Parameters
//      [IN]    pPropSetGUID        GUID of the property set the name property
//                                  belongs to
//      [IN]    wstrName            UNICODE string with the name of the prop
//      [IN]    lIDName             Numeric "name" of the property
//      [IN]    ulKind              The kind of name for the property (numeric
//                                  or literal string). Based on this value we
//                                  use the wstrName or lIDName parameter
//      [OUT]   pulNewID            Where we return the ID of the PROPERTY if
//                                  found or if it was created
//      [IN]    fCreateIfNotFound   Flag to indicate that we should allocate
//                                  a new ID if we don't find one for the
//                                  specified property
//
//    Purpose
//      This function assigns a new ID to a named property. If the name
//      property exists in the given property set (indicated with the GUID
//      parameter), we return the existing ID. If the name does not exist
//      in the given property set we allocate a new one, if there is space.
//      Since we use a global allocation space of 32767 entries for named
//      properties, if we run out of IDs, we return
///     MAPI_E_NOT_ENOUGH_RESOURCES.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::GetIDForName (LPGUID  pPropSetGUID,
                                         LPWSTR  wstrName,
										 long	 lIDName,
										 ULONG	 ulKind,
                                         ULONG * pulNewID,
                                         BOOL    fCreateIfNotFound)
{
    SRestriction SResSearch = { 0 };
    SRestriction SRestrictions[2] = { 0 };
    SPropValue spvName = { 0 }, spvPSGuid = { 0 };
    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);
    // Get the table view on the data, if we don't already have one
    HRESULT hSearchResult, hResult = GetTableView();
    if (hResult)
    {
        goto ErrorExit;
    }
    
    // Set up the property and restriction for the search in the table
    if (MNID_STRING == ulKind)
    {
	    spvName.ulPropTag = PR_NAMED_PROP_NAME;
        spvName.Value.lpszW = wstrName;
    }
    else
    {
        spvName.ulPropTag = PR_NAMED_PROP_IDNAME;
        spvName.Value.l = lIDName;
    }

    spvPSGuid.ulPropTag = PR_NAMED_PROP_GUID;
    spvPSGuid.Value.lpguid = pPropSetGUID;
    
    SRestrictions[0].rt = RES_PROPERTY;
    SRestrictions[0].res.resProperty.relop = RELOP_EQ;
    SRestrictions[0].res.resProperty.ulPropTag = PR_NAMED_PROP_GUID;
    SRestrictions[0].res.resProperty.lpProp = &spvPSGuid;

    SRestrictions[1].rt = RES_PROPERTY;
    SRestrictions[1].res.resProperty.relop = RELOP_EQ;
    SRestrictions[1].res.resProperty.ulPropTag = spvName.ulPropTag;
    SRestrictions[1].res.resProperty.lpProp = &spvName;
    
    SResSearch.rt = RES_AND;
    SResSearch.res.resAnd.cRes = sizeof(SRestrictions)/sizeof(SRestriction);
    SResSearch.res.resAnd.lpRes = SRestrictions;

    // Look up the name in the property, based on an AND restriction
    // of the GUID of the property set and the name of the property.
    hSearchResult = m_pNameTable->FindRow (&SResSearch, BOOKMARK_BEGINNING, 0);
    if (S_OK == hSearchResult)
    {
        // If we found the property, get the row with the property values
        LPSRowSet pRows;
        hResult = m_pNameTable->QueryRows (1, 0, &pRows);
        if (!hResult)
        {
            ASSERT (PR_NAMED_PROP_GUID == pRows->aRow->lpProps[NP_GUID].ulPropTag);
            ASSERT (PR_NAMED_PROP_ID == pRows->aRow->lpProps[NP_ID].ulPropTag);
            ASSERT (PR_NAMED_PROP_KIND == pRows->aRow->lpProps[NP_KIND].ulPropTag);
            
            // Return the ID to the caller
            *pulNewID = pRows->aRow->lpProps[NP_ID].Value.l;
            FreeProws (pRows);
        }
    }
    else
    {
        // If the property was not set in the table, try to allocte a new ID for it
        if (MAPI_E_NOT_FOUND == hSearchResult && fCreateIfNotFound)
        {
            // If the next ID wraps the range, we can't allocate any
            // new more properties. 
            if (0xFFFF == m_ulNextID)
            {
                hResult = MAPI_E_NOT_ENOUGH_RESOURCES;
            }
            else
            {
                // Looks like we were successful at allocating the ID for the property,
                // now we need to add that information to the global map of property
                // names. We maintain that information in an ITableData object.
                SRow srRow = { 0 };
                SPropValue spvProps[4] = { 0 };
                srRow.cValues = sizeof(spvProps)/sizeof(SPropValue);
                srRow.lpProps = spvProps;

                spvProps[0].ulPropTag = PR_NAMED_PROP_GUID;
                spvProps[0].Value.lpguid = pPropSetGUID;
                spvProps[1].ulPropTag = PR_NAMED_PROP_ID;
                spvProps[1].Value.l = m_ulNextID;
                spvProps[2].ulPropTag = PR_NAMED_PROP_KIND;
                spvProps[2].Value.l = ulKind;

                if (MNID_STRING == ulKind)
                {
	                spvProps[3].ulPropTag = PR_NAMED_PROP_NAME;
                    spvProps[3].Value.lpszW = wstrName;
                }
                else
                {
                    spvProps[3].ulPropTag = PR_NAMED_PROP_IDNAME;
                    spvProps[3].Value.l = lIDName;
                }

                hResult = m_pMapData->HrModifyRow (&srRow);
                if (!hResult)
                {
                    // Return the new ID for the named property...
                    *pulNewID = m_ulNextID;
                    // ...increment the counter for the next one...
                    m_ulNextID++;
                    // ... and mark the object "dirty"
                    m_bMappingChanged = TRUE;
                }
            }
        }
        else
        {
            // Something has gone wrong
            hResult = hSearchResult;
        }
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);
    if (hResult && hResult != MAPI_E_NOT_FOUND)
    {
        TraceResult ("CNamePropMgr::GetIDForName", hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::GetNameOfID()
//
//    Parameters
//      [IN]    ulPropTag           Property tag for which we will look up
//                                  the name
//      [IN]    pParentMemBlock     MAPI-allocated memory block to which the
//                                  returned allocated MAPINAMEID structure
//                                  is linked to
//      [OUT]   ppNamedProp         Where we return the name of the
//                                  property tag
//
//    Purpose
//      This function looks up a property tag ID in our internal mapping. If
//      we find a match, we then return the name string (in UNICODE) for
//      that ID. Along with the name we also return the GUID for the
//      property set to which it belongs. The caller is responsible for
//      releasing the memory we allocate in this function
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::GetNameOfID (ULONG           ulPropTag,
                                        LPVOID          pParentMemBlock,
                                        LPMAPINAMEID  * ppNamedProp)
{
    HRESULT hResult = S_OK;

    SRestriction SResSearch = { 0 };
    SPropValue spvPropID = { 0 };

    EnterCriticalSection (&m_csObj);
    hResult = GetTableView();
    if (hResult)
    {
        goto ErrorExit;
    }

    spvPropID.ulPropTag = PR_NAMED_PROP_ID;
    spvPropID.Value.l = PROP_ID(ulPropTag);
    
    SResSearch.rt = RES_PROPERTY;
    SResSearch.res.resProperty.relop = RELOP_EQ;
    SResSearch.res.resProperty.ulPropTag = spvPropID.ulPropTag;
    SResSearch.res.resProperty.lpProp = &spvPropID;

    // Look up the name in the property, based on an AND restriction
    // of the GUID of the property set and the name of the property.
    hResult = m_pNameTable->FindRow (&SResSearch, BOOKMARK_BEGINNING, 0);
    if (hResult)
    {
        goto ErrorExit;
    }
    LPSRowSet pRows;
    hResult = m_pNameTable->QueryRows (1, 0, &pRows);
    if (hResult)
    {
        goto ErrorExit;
    }
    ASSERT (1 == pRows->cRows);

    hResult = CopyNamedProp (pParentMemBlock, ppNamedProp, pRows->aRow->lpProps);
    FreeProws (pRows);
    
ErrorExit:
    LeaveCriticalSection (&m_csObj);
    if (hResult && hResult != MAPI_E_NOT_FOUND)
    {
        TraceResult ("CNamePropMgr::GetNameOfID", hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::GetAllNamesAndTags()
//
//    Parameters
//      [OUT]   ppPropTags      Where we return an newly allocated array of the
//                              property tags for ALL the named properties we
//                              store. The type of these tags will be set to
//                              PT_UNSPECIFIED.
//      [OUT]   pppPropNames    Where we return the array of named property
//                              structures
//      [IN]    ulFlags         Flags for kinds of names we return to the
//                              caller
//      [OUT]   pcPropNames     Where we return how many properties we found
//
//    Purpose
//      Allocate an array of property tags and return the tags of all the
//      named properties in our mapping. Also allocate an array of MAPINAMEID
//      structure and return the information for all the named properties in
//      our mapping. We will allocate the memory for the arrays and its
//      elements. The client will be responsible for releasing that.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::GetAllNamesAndTags (LPSPropTagArray * ppPropTags,
                                               LPMAPINAMEID **   pppPropNames,
                                               ULONG             ulFlags,
                                               ULONG *           pcPropNames)
{
    *pppPropNames = NULL;
    *ppPropTags = NULL;
    *pcPropNames = 0;

    HRESULT hResult = S_OK;

    LPMAPINAMEID *	ppOneName;

    LPSPropTagArray pTags = NULL;
    EnterCriticalSection (&m_csObj);
    hResult = GetTableView();
    if (hResult)
    {
        goto ErrorExit;
    }
    
    // Get a count of how many rows are our table of properties
    ULONG ulCount;
    hResult = m_pNameTable->GetRowCount (0, &ulCount);
    if (!hResult)
    {
        // Allocate an array where we return the tags (with just the property IDs) back to client
        hResult = m_pfnAllocateBuffer (CbNewSPropTagArray (ulCount), (LPVOID *)&pTags);
        if (!hResult && ulCount)
        {
            // Allocate an array where we return an array of named properties
            hResult = m_pfnAllocateBuffer (ulCount * sizeof(LPMAPINAMEID), (LPVOID *)pppPropNames);
        }
    }
    if (pTags)
    {
        ZeroMemory (pTags, CbNewSPropTagArray (ulCount));
    }
    if (hResult || !ulCount)
	{
        goto ErrorExit;
	}
    
    // Move the table pointer to the origin of the table
    m_pNameTable->SeekRow (BOOKMARK_BEGINNING, 0, NULL);

    LPSRowSet pRows;
    BOOL fContinue;
    fContinue = TRUE;
    LPVOID pVoid;
    pVoid = (LPVOID)(*pppPropNames);
    ppOneName = *pppPropNames;
    while (!hResult && fContinue)
    {
        // Get chunks of the table and process them
        hResult = m_pNameTable->QueryRows (25, 0, &pRows);
        if (!hResult)
        {
            if (pRows->cRows)
            {
                ULONG ulKind;
                for (ULONG j=0; j<pRows->cRows && !hResult; j++)
                {
                    // Get kind of name for this property...
                    ulKind = pRows->aRow[j].lpProps[NP_KIND].Value.l;
                    // ... if the user does not want numeric names, skip
                    // this entry ...
                    if ((MAPI_NO_IDS & ulFlags) && (ulKind == MNID_ID))
                    {
                        continue;
                    }
                    // ... if the user does not want string names, skip this entry
                    if ((MAPI_NO_STRINGS & ulFlags) && (ulKind == MNID_STRING))
                    {
                        continue;
                    }
                    // Now we need to duplicate the named property information.
                    // All the memory to the entries is linked to the array of
                    // named properties.
                    hResult = CopyNamedProp (pVoid,
                                             ppOneName,
                                             pRows->aRow[j].lpProps);
                    if (!hResult)
                    {
                        // Set the property tag we are returning to the client
                        pTags->aulPropTag[pTags->cValues] =
                                PROP_TAG (PT_UNSPECIFIED,
                                          pRows->aRow[j].lpProps[NP_ID].Value.l);
                        pTags->cValues++;
                        ppOneName++;
                    }
                }
            }
            else
            {
                fContinue = FALSE;
            }
            FreeProws (pRows);
        }
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);
    if (hResult)
    {
        // If we failed clean up, the caller won't
        if (pTags)
        {
            m_pfnFreeBuffer (pTags);
        }
        if (*pppPropNames)
        {
            m_pfnFreeBuffer (*pppPropNames);
            *pppPropNames = NULL;
        }
        TraceResult ("CNamePropMgr::GetAllNamesAndTags", hResult);
    }
    else
    {
        // We are successful, return the stuff we found in our mapping
        *ppPropTags = pTags;
        *pcPropNames = pTags->cValues;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::CopyNamedProp()
//
//    Parameters
//      [IN]    pParentMem      Pointer to a MAPI-allocated memory block to
//                              which the data we return is linked to.
//      [OUT]   ppNamedProp     Where we return the copied named property
//      [IN]    pProps          Pointer to an array of property values which
//                              hold the values of the name property entry 
//                              which we will copy.
//
//    Purpose
//      This function copies the data from an name property entry in our
//      mapping into a MAPINAMEID structure ready to return to the client
//      consumers.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::CopyNamedProp (LPVOID          pParentMem,
                                          LPMAPINAMEID  * ppNamedProp,
                                          LPSPropValue    pProps)
{
    ASSERT (PR_NAMED_PROP_GUID == pProps[NP_GUID].ulPropTag);
    ASSERT (PR_NAMED_PROP_ID   == pProps[NP_ID].ulPropTag);
    ASSERT (PR_NAMED_PROP_KIND == pProps[NP_KIND].ulPropTag);

    // Allocate space for the named ID information
	HRESULT hResult = m_pfnAllocateMore (sizeof(MAPINAMEID),
								         pParentMem,
								         (LPVOID *)ppNamedProp);
	if (hResult)
	{
		return hResult;
	}
    // Copy the property set GUID
	hResult = m_pfnAllocateMore (sizeof(GUID),
								 pParentMem,
								 (LPVOID *)&((*ppNamedProp)->lpguid));
	if (hResult)
	{
		return hResult;
	}
    CopyMemory ((*ppNamedProp)->lpguid, pProps[NP_GUID].Value.lpguid, sizeof(GUID));
    (*ppNamedProp)->ulKind = pProps[NP_KIND].Value.l;
	
    // If the name of the property is a string, we must allocate memory for it.
	if (MNID_STRING == pProps[NP_KIND].Value.l)
	{
		ULONG cbString = (lstrlenW (pProps[NP_NAME].Value.lpszW)+1)*sizeof(WCHAR);
        hResult = m_pfnAllocateMore (cbString,
									 pParentMem,
									 (LPVOID *)&((*ppNamedProp)->Kind.lpwstrName));
		if (hResult)
		{
			return hResult;
		}
		CopyMemory ((*ppNamedProp)->Kind.lpwstrName, pProps[NP_NAME].Value.lpszW, cbString);
	}
    else
    {
        // otherwise just copy the numeric name of the property
        (*ppNamedProp)->Kind.lID = pProps[NP_IDNAME].Value.l;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CNamePropMgr::GetMAPINamedPropsAndTags()
//
//    Parameters
//      [IN]    pObjPropTags    Array of property tags of the object this name
//                              property map is linked to
//      [OUT]   ppPropTags      Where we return an newly allocated array of the
//                              property tags for ALL the named properties
//                              we store. The type of these tags will be set
//                              to PT_UNSPECIFIED.
//      [OUT]   pppPropNames    Where we return the array of named property
//                              structures
//      [OUT]   pcPropNames     Where we return how many properties we found
//
//    Purpose
//      Allocate an array of property tags and return the tags of all the
//      named properties for the PS_MAPI set. These are the properties whose
//      IDs are NOT in the named property range (i.e. less than 0x8000)
//      Along with those tags, we return an array of MAPINAMEID structure.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CNamePropMgr::GetMAPINamedPropsAndTags (LPSPropTagArray   pObjPropTags,
                                                     LPSPropTagArray * ppPropTags,
                                                     LPMAPINAMEID **   pppPropNames,
                                                     ULONG *           pcPropNames)
{
    *pppPropNames = NULL;
    *ppPropTags = NULL;
    *pcPropNames = 0;

    HRESULT hResult = S_OK;

    LPMAPINAMEID *	ppOneName;

    LPSPropTagArray pTags = NULL;
    LPGUID pMAPIGuid = NULL;
    ULONG ulCount = 0, i;
    EnterCriticalSection (&m_csObj);
    

    for (i=0; i<pObjPropTags->cValues; i++)
    {
        if (0x8000 < PROP_ID(pObjPropTags->aulPropTag[i]))
        {
            ulCount++;
        }
    }

    // Allocate an array where we return the tags (with just the property IDs) back to client
    hResult = m_pfnAllocateBuffer (CbNewSPropTagArray (ulCount), (LPVOID *)&pTags);
    if (!hResult && ulCount)
    {
        // Allocate where we return an array of named properties
        hResult = m_pfnAllocateBuffer (ulCount * sizeof(LPMAPINAMEID), (LPVOID *)pppPropNames);
    }
    if (pTags)
    {
        ZeroMemory (pTags, CbNewSPropTagArray (ulCount));
    }
    if (hResult || !ulCount)
	{
        goto ErrorExit;
	}
    
    LPVOID pVoid;
    pVoid = (LPVOID)(*pppPropNames);
    ppOneName = *pppPropNames;
    for (i=0; i<ulCount; i++)
    {
        if (0x8000 >= PROP_ID(pObjPropTags->aulPropTag[i]))
        {
            continue;
        }

        hResult = m_pfnAllocateMore (sizeof(MAPINAMEID),
								     pVoid,
								     (LPVOID *)ppOneName);
	    if (hResult)
	    {
		    goto ErrorExit;
	    }
        if (NULL == pMAPIGuid)
        {
            // Copy the property set GUID for PS_MAPI, we only do this once,
            // since all the named properties are of the same set.
	        hResult = m_pfnAllocateMore (sizeof(GUID),
								         pVoid,
								         (LPVOID *)&((*ppOneName)->lpguid));
	        if (hResult)
	        {
		        goto ErrorExit;
	        }
        }
        CopyMemory ((*ppOneName)->lpguid, &PS_MAPI, sizeof(GUID));
        (*ppOneName)->ulKind = MNID_ID;
	    (*ppOneName)->Kind.lID = PROP_ID(pObjPropTags->aulPropTag[i]);

        pTags->aulPropTag[pTags->cValues] = pObjPropTags->aulPropTag[i];
        pTags->cValues++;
        ppOneName++;
    }

ErrorExit:
    LeaveCriticalSection (&m_csObj);
    if (hResult)
    {
        // If we failed clean up, the caller won't
        if (pTags)
        {
            m_pfnFreeBuffer (pTags);
        }
        if (*pppPropNames)
        {
            m_pfnFreeBuffer (*pppPropNames);
            *pppPropNames = NULL;
        }
        TraceResult ("CNamePropMgr::GetMAPINamedPropsAndTags", hResult);
    }
    else
    {
        // We are successful, return the stuff we found in our mapping
        *ppPropTags = pTags;
        *pcPropNames = pTags->cValues;
    }
    return hResult;
}
// End of file for NAMEPROP.CPP
