///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      PROPLIST.CPP
//
//  Description
//      Implements the generic CPropList class and its auxiliary classes and
//      function. The CPropList class is used as the in-memory property
//      cache for MAPI object implementing the IMAPIProp interface.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "PROPLIST.H"

// Global pointers to the MAPI memory allocation functions
LPALLOCATEBUFFER gpfnAllocateBuffer     = NULL;  // MAPIAllocateBuffer function
LPALLOCATEMORE   gpfnAllocateMore       = NULL;  // MAPIAllocateMore function    
LPFREEBUFFER     gpfnFreeBuffer         = NULL;  // MAPIFreeBuffer function      
HANDLE           ghMemHeap              = NULL;
BOOL             fMemoryHeapInitialized = FALSE;

BOOL WINAPI GetTempFullPathFileName (LPSTR szFileName);

LPALLOCATEBUFFER GetMAPIAllocBuffer
                    () { return gpfnAllocateBuffer; }
LPALLOCATEMORE GetMAPIAllocMore
                    () { return gpfnAllocateMore; }
LPFREEBUFFER GetMAPIFreeBuffer
                    () { return gpfnFreeBuffer; }

//#define DO_LEAK_TRACES
#ifdef DO_LEAK_TRACES
#define LeakTrace(a)                TraceMessage(a)
#define LeakTraceRaw(a)             TraceRaw(a)
#define LeakTraceString(a,b,c,d,e)  TraceString4(a,b,c,d,e)
#else
#define LeakTrace(a)
#define LeakTraceRaw(a)
#define LeakTraceString(a,b,c,d,e)
#endif // DO_LEAK_TRACES

//#define USE_MAPIALLOC_FUNCTIONS        1

#ifdef _DEBUG

typedef struct _ALLOC_NODE
{
    struct _ALLOC_NODE *    pNext;
    LPVOID                  pAllocBlock;
    ULONG                   ulSize;
    TCHAR                   szName[32];
} ALLOC_NODE, *PALLOC_NODE;

class CAllocationsList
{
public:
    void WINAPI Insert
                    (LPVOID                     pBlock,
                     ULONG                      ulSize,
                     LPTSTR                     pszName = NULL);
    void WINAPI Delete
                    (LPVOID                     pBlock);
private :
    void WINAPI DumpLeaks
                    ();
///////////////////////////////////////////////////////////////////////////////
// Constructors and destructors
public :
    CAllocationsList () { m_pHead = NULL; };
    ~CAllocationsList () { DumpLeaks(); };

///////////////////////////////////////////////////////////////////////////////
// Data members
private :
    PALLOC_NODE             m_pHead;
};

CAllocationsList * pAllocList = NULL;

void WINAPI CAllocationsList::Insert (LPVOID  pBlock,
                                      ULONG   ulSize,
                                      LPTSTR  pszName)
{
    PALLOC_NODE pNewNode = (PALLOC_NODE)HeapAlloc (GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(ALLOC_NODE));
    if (pNewNode)
    {
        pNewNode->pAllocBlock = pBlock;
        pNewNode->ulSize = ulSize;
        pNewNode->pNext = NULL;
        lstrcpy (pNewNode->szName, pszName);
        InsertInList (&m_pHead, pNewNode);
    }
    else
    {
        TraceMessage ("CAllocationsList::Insert: Failed to allocate new node");
    }
}
void WINAPI CAllocationsList::Delete (LPVOID pBlock)
{
    PALLOC_NODE pNext, pPrev = NULL, pNode = m_pHead;
    while (pNode)
    {
        if (pNode->pAllocBlock == pBlock)
        {
            if (m_pHead == pNode)
            {
                m_pHead = m_pHead->pNext;
            }
            // Save the next pointer so that we may delete the current node
            pNext = pNode->pNext;
            HeapFree (GetProcessHeap(), 0, pNode);
            // If we have a previous node, we are doing a delete in the middle of a single link list
            if (pPrev)
            {
                pPrev->pNext = pNext;
            }
            break;
        }
        else
        {
            pPrev = pNode;
            pNode = pNode->pNext;
        }
    }
}

void WINAPI CAllocationsList::DumpLeaks()
{
    LeakTraceRaw ("Internal Allocation List: Checking for memory leaks...");
    if (m_pHead)
    {
        LeakTrace ("  ...Leak(s) Found!!");
        LeakTrace ("Dumping Leaks:");
        UINT i=1;
        PALLOC_NODE pNext, pNode = m_pHead;
        while (pNode)
        {
            LeakTraceString ("    Leak %d, Block Address: %#08x, Size: %d, Name: %s", i, pNode->pAllocBlock, pNode->ulSize, pNode->szName);
            pNext = pNode->pNext;
            HeapFree (GetProcessHeap(), 0, pNode);
            pNode = pNext;
            i++;
        }
        m_pHead = NULL;
    }
    else
    {
        LeakTrace ("  ...None Found");
    }
}

#endif // _DEBUG

///////////////////////////////////////////////////////////////////////////////
//    InitMemoryHeap()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI InitMemoryHeap (LPMAPISUP pSupObj)
{
    if (fMemoryHeapInitialized)
    {
        return MAPI_W_PARTIAL_COMPLETION;
    }
    LPALLOCATEBUFFER pfnAllocateBuffer;
    LPALLOCATEMORE pfnAllocateMore;
    LPFREEBUFFER pfnFreeBuffer;
    // Get the MAPI memory allocation routines we'll be needing
    HRESULT hResult = pSupObj->GetMemAllocRoutines (&pfnAllocateBuffer, &pfnAllocateMore, &pfnFreeBuffer);
    TraceResult ("InitMemoryHeap: Failed to get the MAPI memory allocation functions", hResult);
    if (!hResult)
    {
        hResult = InitMemoryHeap (pfnAllocateBuffer, pfnAllocateMore, pfnFreeBuffer);
    }
    return hResult;
}
///////////////////////////////////////////////////////////////////////////////
//    InitMemoryHeap()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI InitMemoryHeap (LPALLOCATEBUFFER pfnAllocateBuffer,
                               LPALLOCATEMORE   pfnAllocateMore,
                               LPFREEBUFFER     pfnFreeBuffer)
{
    if (fMemoryHeapInitialized)
    {
        return MAPI_W_PARTIAL_COMPLETION;
    }
    gpfnAllocateBuffer = pfnAllocateBuffer;
    gpfnAllocateMore   = pfnAllocateMore;
    gpfnFreeBuffer     = pfnFreeBuffer;
    #ifdef _DEBUG
    pAllocList = new CAllocationsList;
    if (!pAllocList)
    {
        TraceMessage ("InitMemoryHeap: Failed to allocate list object for internal allocations");
        return E_OUTOFMEMORY;
    }
    #endif // _DEBUG
    fMemoryHeapInitialized = TRUE;
    ghMemHeap = HeapCreate (0, 8192, 0);
    fMemoryHeapInitialized = (ghMemHeap ? TRUE : FALSE);
    return ghMemHeap ? S_OK : E_FAIL;
}

///////////////////////////////////////////////////////////////////////////////
//    CloseMemoryHeap()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI CloseMemoryHeap()
{
    HRESULT hResult = S_OK;
    #ifdef _DEBUG
    delete pAllocList; // This will call the destructor to dump any leaks
    pAllocList = NULL;
    #endif // _DEBUG
    if (ghMemHeap)
    {
        if (!HeapDestroy (ghMemHeap))
        {
            hResult = HRESULT_FROM_WIN32(GetLastError());
            TraceResult ("CloseMemoryHeap: Failed to destroy the process heap", hResult);
        }
        ghMemHeap = NULL;
    }
    gpfnAllocateBuffer      = NULL;  // MAPIAllocateBuffer function
    gpfnAllocateMore        = NULL;  // MAPIAllocateMore function    
    gpfnFreeBuffer          = NULL;  // MAPIFreeBuffer function    
    fMemoryHeapInitialized  = FALSE;
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    _PrivAllocate()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI _PrivAllocate (DWORD cbSize, LPVOID * ppVoid, LPTSTR pszBlockName)
{    
    HRESULT hResult = S_OK;
    ASSERTMSG (cbSize <= 0x12000, "Allocation size is bigger that 72Kb. Are you sure?"); 
    #ifdef USE_MAPIALLOC_FUNCTIONS
    hResult = gpfnAllocateBuffer (cbSize, ppVoid);
    #else  // USE_MAPIALLOC_FUNCTIONS
    *ppVoid = HeapAlloc (ghMemHeap, 0, cbSize);
    if (!(*ppVoid))
    {
        hResult = E_OUTOFMEMORY;
    }
    #endif // USE_MAPIALLOC_FUNCTIONS    
    if (hResult)
    {
        *ppVoid = NULL;
        TraceResult ("PrivAllocate: Failed to allocate buffer", hResult);
        TraceString2 ("     Allocation Size: %d, Name: %s", cbSize, pszBlockName);
    }
    else
    {
        ZeroMemory (*ppVoid, cbSize);
        #ifdef _DEBUG
        pAllocList->Insert (*ppVoid, cbSize, pszBlockName);
        #else
        #endif // _DEBUG
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    PrivFree()
//
//    Parameters
//      [IN,OUT] ppVoid     Pointer to a pointer to a block of memory that we
//                          should free. Upon return we NULL the pointer value
//
//    Purpose
//      Free a block of memory previously allocated with a PrivAllocate() call
//
//    Return Value
//      None
//
void WINAPI PrivFree (LPVOID * ppVoid)
{
    if (ppVoid && *ppVoid)
    {
        #ifdef USE_MAPIALLOC_FUNCTIONS
        gpfnFreeBuffer (*ppVoid);
        pAllocList->Delete (*ppVoid);
        #else  // USE_MAPIALLOC_FUNCTIONS
        HeapFree (ghMemHeap, 0, *ppVoid);
        #endif // USE_MAPIALLOC_FUNCTIONS
        *ppVoid = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
//    PrivMAPIAllocate()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI PrivMAPIAllocate (DWORD cbSize, LPVOID * ppVoid, LPVOID pParentMemBlock)
{
    HRESULT hResult;
    if (pParentMemBlock)
    {
        hResult = gpfnAllocateMore (cbSize, pParentMemBlock, ppVoid);
    }
    else
    {
        hResult = gpfnAllocateBuffer (cbSize, ppVoid);
    }
    if (hResult)
    {
        TraceResult("PrivAllocate: Failed to allocate buffer", hResult);
        TraceRaw ("              Buffer size: "); TraceString1 ("%d", cbSize);
	    return hResult;
    }
    ZeroMemory (*ppVoid, cbSize);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    PrivMAPIFree()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
void WINAPI PrivMAPIFree (LPVOID * ppVoid)
{
    gpfnFreeBuffer (*ppVoid);
    *ppVoid = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    _Insert()
//
//    Parameters
//      
//    Purpose
//
//    Return Value
//
void WINAPI _Insert (PGENERIC_NODE * ppHead, PGENERIC_NODE pNewNode)
{
    if (*ppHead)
    {
        PGENERIC_NODE pNode = *ppHead;
        while (pNode)
        {
            if (NULL == pNode->pNext)
            {
                pNode->pNext = pNewNode;
                return;
            }
            pNode = pNode->pNext;
        }
    }
    else
    {
        *ppHead = pNewNode;
    }
}

///////////////////////////////////////////////////////////////////////////////
//    _DeleteList()
//
//    Parameters
//      
//    Purpose
//
//    Return Value
//
void WINAPI _DeleteList (PGENERIC_NODE * ppHead)
{
    PGENERIC_NODE pNode = *ppHead;
    while (*ppHead)
    {
        pNode = (*ppHead)->pNext;
        PrivFree ((LPVOID *)ppHead);
        *ppHead = pNode;
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Insert()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::Insert (LPSPropValue pProp)
{
    HRESULT hResult = S_OK;
    if (InternalNodePropReplaced (pProp))
    {
        return hResult;
    }
    PFlatPropNode pNode;
    hResult = Flatten (pProp, &pNode);
    if (hResult)
    {
        if (MAPI_W_ERRORS_RETURNED == hResult)
        {
            hResult = S_OK;
        }
        return hResult;
    }
    if (!m_pHead)
    {   
        m_pHead = m_pTail = pNode;
        m_ulNodes++;
        return hResult;
    }

    // Check for a duplicated property tag, if found, replace it with the new node
    PFlatPropNode pTmpNode;
    if (PROP_ID(m_pHead->Prop.ulPropTag) == PROP_ID(pProp->ulPropTag))       // Hit on the head of the list
    {
        pTmpNode = m_pHead;
        pNode->pNext = m_pHead->pNext;
        m_pHead = pNode;
        PrivFree ((LPVOID *)&pTmpNode);        // Free the memory of the node
        return hResult;             // We are done
    }
    PFlatPropNode pPrevious = m_pHead;
    pTmpNode = m_pHead->pNext;
    while (pTmpNode)
    {
        if (PROP_ID(pTmpNode->Prop.ulPropTag) == PROP_ID(pProp->ulPropTag))  // Hit on the middle of the list
        {
            pPrevious->pNext = pNode;
            pNode->pNext = pTmpNode->pNext;
            PrivFree ((LPVOID *)&pTmpNode);    // Free the memory of the node
            return hResult;         // We are done
        }
        pPrevious = pTmpNode;
        pTmpNode = pTmpNode->pNext;
    }

    // The property is not in the list, put it at end.
    pPrevious->pNext = pNode;
    m_ulNodes++;
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Find()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::Find (ULONG         ulPropTag,
                              LPSPropValue  pProp,
                              BOOL          fMatchExact,
                              LPVOID        pVoid)
{
    PFlatPropNode pNode = m_pHead;
    HRESULT hResult = MAPI_E_NOT_FOUND;
    while (pNode)
    {
        if (PROP_ID(pNode->Prop.ulPropTag) == PROP_ID(ulPropTag))
        {
            if (fMatchExact)
            {
                // For an exact match, the property types must be the same
                if (PROP_TYPE(pNode->Prop.ulPropTag) != PROP_TYPE(ulPropTag))
                {
                    pNode = pNode->pNext;
                    continue;
                }
            }
            if (pVoid)
            {
                hResult = Copy (pNode, pProp, pVoid);
            }
            else
            {
                hResult = S_OK;
                pProp->ulPropTag = pNode->Prop.ulPropTag;
                pProp->Value = pNode->Prop.Value;
            }
            break; // Out of the WHILE() loop
        }
        pNode = pNode->pNext;
    }
    if (hResult)
    {
        pProp->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(ulPropTag));
        pProp->Value.err = hResult;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::GetAllProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::GetAllProps (LPSPropValue * ppProps, ULONG *  pulCount)
{
    PFlatSPropValue pProp;
    HRESULT hResult = S_OK;
    if (!m_ulNodes)
    {
        *ppProps = NULL;
        *pulCount = 0;
        return hResult;
    }
    ULONG i = 0;
    *pulCount = m_ulNodes;
    hResult = PrivAllocate (m_ulNodes * sizeof(SPropValue), (LPVOID *)ppProps, "CPropList's GetAllProps");
    if (!hResult)
    {
        pProp = Walk (WALK_START);
        while (pProp)
        {
            (*ppProps)[i].ulPropTag = pProp->ulPropTag;
            (*ppProps)[i++].Value = pProp->Value;
            pProp = Walk (WALK_CONTINUE);
        }
        ASSERT (i == m_ulNodes);
    }
    TraceResult ("CPropList::GetAllProps", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Delete()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::Delete (ULONG ulPropTag, LPSPropValue pProp)
{
    PFlatPropNode pTmpNode2 = NULL, pTmpNode = m_pHead;
    HRESULT hResult = S_OK;
    if (!pTmpNode)
    {
        return hResult;
    }

    if (PROP_ID(pTmpNode->Prop.ulPropTag) == PROP_ID(ulPropTag))
    {
        m_pHead = m_pHead->pNext;
    }
    else
    {
        hResult = MAPI_E_NOT_FOUND;
        pTmpNode = m_pHead->pNext;
        while (pTmpNode)
        {
            if (PROP_ID(pTmpNode->Prop.ulPropTag) == PROP_ID(ulPropTag))
            {
                hResult = S_OK;
                pTmpNode2->pNext = pTmpNode->pNext;
                break; // Out of the WHILE() loop
            }
            pTmpNode2 = pTmpNode;
            pTmpNode = pTmpNode->pNext;
        }
    }
    if (S_OK == hResult)
    {
        // If the caller requested to keep a hold of the node's property
        // value, save it in the returning parameter
        if (pProp)
        {
            pProp->ulPropTag = pTmpNode->Prop.ulPropTag;
            pProp->Value = pTmpNode->Prop.Value;
        }
        if (m_ulNodes)
        {
            m_ulNodes--;
        }
        PrivFree ((LPVOID *)&pTmpNode);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::GetTags()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::GetTags (LPSPropTagArray * ppTags)
{
    HRESULT hResult = S_OK;
    DWORD cbSize = CbNewSPropTagArray (m_ulNodes);
    ULONG i = 0;
    PFlatPropNode pNode = m_pHead;
    __try
    {
        hResult = PrivAllocate (cbSize, (LPVOID *)ppTags, "CPropList's GetTags");
        if (!hResult)
        {
            while (pNode)
            {
                (*ppTags)->cValues++;
                (*ppTags)->aulPropTag[i++] = pNode->Prop.ulPropTag;
                pNode = pNode->pNext;
            }
            ASSERT ((*ppTags)->cValues == m_ulNodes);
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("GetTags: An exception occured. Exception code: %#08x", GetExceptionCode());
        hResult = E_FAIL;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Walk()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
PFlatSPropValue WINAPI CPropList::Walk (WALKING_ACTION waAction)
{
    if (m_pHead)
    {
        if (WALK_START == waAction)
        {
            m_pWalkNext = m_pHead->pNext;
            return &m_pHead->Prop;
        }
        if (WALK_CONTINUE == waAction && m_pWalkNext)
        {
            PFlatPropNode pTmp = m_pWalkNext;
            m_pWalkNext = m_pWalkNext->pNext;
            return &pTmp->Prop;
        }
    }
    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Flatten()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::Flatten (LPSPropValue pProp, PFlatPropNode * ppNode)
{
    LPVOID pVoid;
    ULONG cbSize;
    PFlatPropNode pNode = NULL;
    LPBYTE pByte;
    HRESULT hResult = NULL;
    __try
    {
        hResult = GetSize (pProp->ulPropTag, pProp->Value, cbSize, &pVoid);
        if (!hResult)
        {
            hResult = PrivAllocate (cbSize + sizeof (FlatPropNode), (LPVOID *)&pNode, "Flatten Property Data");
            if (!hResult)
            {
                pNode->Prop.ulPropTag = pProp->ulPropTag;
                pNode->Prop.Value = pProp->Value;

                if (pVoid)
                {
                    pByte = (LPBYTE)pNode + sizeof(FlatPropNode);
                    CopyMemory (pByte, pVoid, cbSize);
                    FixUpPointers (pNode, pByte);
                }
                pNode->Prop.cbSize = cbSize + sizeof (FlatSPropValue);
            }
        }
        *ppNode = pNode;
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("Flatten: An exception occured. Exception code: %#08x", GetExceptionCode());
        PrivFree ((LPVOID *)&pNode);
        hResult = E_FAIL;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::FixUpPointers()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      None
//
void WINAPI CPropList::FixUpPointers (PFlatPropNode pNode, LPBYTE pByte)
{
    if (MV_FLAG & PROP_TYPE(pNode->Prop.ulPropTag))
    {
        // Since the Multi-value members of the "Value" structure
        // are in a UNION, setting one of the pointers, would be
        // sufficient for the others
        pNode->Prop.Value.MVl.lpl = (long *)pByte;
    }
    else
    {
        if (PT_BINARY == PROP_TYPE(pNode->Prop.ulPropTag))
        {
            pNode->Prop.Value.bin.lpb = pByte;
        }
        else
        {
            if (PT_STRING8 == PROP_TYPE(pNode->Prop.ulPropTag) ||
                PT_UNICODE == PROP_TYPE(pNode->Prop.ulPropTag) ||
                PT_CLSID   == PROP_TYPE(pNode->Prop.ulPropTag))
            {
                // The other members would fall by default
                // since they are in a UNION
                pNode->Prop.Value.lpszA = (LPSTR)pByte;
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::GetSize()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::GetSize (ULONG ulPropTag, union _PV Value, ULONG & cbSize, LPVOID * ppVoid)
{
    HRESULT hResult = S_OK;
    cbSize = 0;
    *ppVoid = NULL;
    if (MV_FLAG & PROP_TYPE(ulPropTag))
    {
        cbSize = Value.MVi.cValues;
        *ppVoid = (LPVOID)Value.MVi.lpi;
    }
    switch (PROP_TYPE(ulPropTag))
    {
        // Single value properites
        case PT_FLOAT :
        case PT_DOUBLE :
        case PT_CURRENCY :
        case PT_APPTIME :
        case PT_SYSTIME :
        case PT_BOOLEAN :
        case PT_SHORT :
        case PT_LONG :
        case PT_LONGLONG :
            // There is no need to allocate extra space for this types
            break;

        case PT_STRING8 :
            cbSize = lstrlenA (Value.lpszA) + 1;
            *ppVoid = (LPVOID)Value.lpszA;
            break;
        case PT_UNICODE :
            cbSize = lstrlenW (Value.lpszW) + sizeof(WCHAR);
            *ppVoid = (LPVOID)Value.lpszW;
            break;
        case PT_CLSID :
            cbSize = sizeof(GUID);
            *ppVoid = (LPVOID)Value.lpguid;
            break;
        case PT_BINARY :
            cbSize = Value.bin.cb;
            *ppVoid = (LPVOID)Value.bin.lpb;
            break;
        case PT_OBJECT :
            cbSize = 0;
            *ppVoid = NULL;
            break;
        // Multi-value properites
        case PT_MV_SHORT :
            cbSize *= sizeof(SHORT);
            break;
        case PT_MV_LONG :
            cbSize *= sizeof(long);
            break;
        case PT_MV_FLOAT :
            cbSize *= sizeof(float);
            break;
        case PT_MV_APPTIME :
        case PT_MV_DOUBLE :
            cbSize *= sizeof(double);
            break;
        case PT_MV_CURRENCY :
        	cbSize *= sizeof(CURRENCY);
			break;
		case PT_MV_SYSTIME:
			cbSize *= sizeof(FILETIME);
			break;
        case PT_MV_I8 :
            cbSize *= sizeof(LARGE_INTEGER);
            break;
        case PT_MV_CLSID :
            cbSize *= sizeof(GUID);
            break;
        case PT_MV_STRING8 :
            {
                cbSize *= sizeof(LPSTR);
                for (ULONG i=0; i<Value.MVszA.cValues; i++)
    			{
    				cbSize += lstrlenA (Value.MVszA.lppszA[i]) + 1;
    			}
            }
            break;
        case PT_MV_UNICODE :
            {
                cbSize *= sizeof(LPWSTR);
                for (ULONG i=0; i<Value.MVszW.cValues; i++)
    			{
    				cbSize += lstrlenW (Value.MVszW.lppszW[i]) + 1;
    			}
            }
            break;
        case PT_MV_BINARY :
            {
                cbSize *= sizeof(SBinary);
                for (ULONG i=0; i<Value.MVbin.cValues; i++)
                {
                    cbSize += Value.MVbin.lpbin[i].cb;
                }
            }
            break;

        // Custom property type for OPENED properties
        case PT_STREAM_STORAGE :
            cbSize = 0;
            *ppVoid = NULL;
            break;

        default :
            // We don't allow setting properties with unknown or
            // invalid (i.e PT_ERROR, PT_NULL) type of tag
            hResult = MAPI_W_ERRORS_RETURNED;
            break;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::Copy()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::Copy (PFlatPropNode pNode,
                              LPSPropValue  pProp,
                              LPVOID        pMemBlock)
{
    SPropValue spvOne;
    spvOne.ulPropTag = pNode->Prop.ulPropTag;
    spvOne.Value = pNode->Prop.Value;
    HRESULT hResult = S_OK;
    // If the type is our own PT_STREAM_STORAGE, then return. The caller
    // needs to get the data from the OPEN property list or storage
    if (PT_STREAM_STORAGE == PROP_TYPE(pNode->Prop.ulPropTag))
    {
        *pProp = spvOne;
    }
    else
    {
        hResult = PropCopyMore (pProp, &spvOne, GetMAPIAllocMore(), pMemBlock);
        TraceResult ("CPropList::Copy: PropCopyMore failed", hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::SaveToStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::SaveToStream (LPSTREAM pStream)
{
    LARGE_INTEGER liZero = { 0 };
    pStream->Seek (liZero, STREAM_SEEK_SET, NULL);

    ULARGE_INTEGER uliSize = { 0 };
    PFlatSPropValue pProp;
    DWORD dwTotal = 0, dwWritten;
    HRESULT hResult = pStream->Write (&m_ulNodes, sizeof(ULONG), &dwWritten);
    if (!hResult && m_ulNodes)
    {
        dwTotal = dwWritten;
        pProp = Walk (WALK_START);
        while (pProp && !hResult)
        {
            hResult = pStream->Write (pProp, pProp->cbSize, &dwWritten);
            if (!hResult)
            {
                dwTotal += dwWritten;
                pProp = Walk (WALK_CONTINUE);
            }
        }
    }
    if (!hResult)
    {
        uliSize.LowPart = (long)dwTotal;
        pStream->SetSize (uliSize); // We don't report errors here
    }
    if (!hResult)
    {
        m_fDirty = FALSE;
    }
    TraceResult ("SaveToStream: I/O operation failed", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::LoadFromStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::LoadFromStream (LPSTREAM pStream)
{
    // Save current values in temporary holders so that we may restore in case of failure
    ULONG ulNodeSize, i, tmpNodes = m_ulNodes;
    PFlatPropNode pNode, tmpTail = m_pTail, tmpHead = m_pHead;
    
    // Reset internal members
    Reset();

    ULONG ulStoredNodes;
    HRESULT hResult = pStream->Read (&ulStoredNodes, sizeof(ULONG), NULL);
    if (!hResult && ulStoredNodes)
    {
        for (i=0; i<ulStoredNodes && !hResult; i++)
        {
            hResult = pStream->Read (&ulNodeSize, sizeof(ULONG), NULL);
            if (!hResult)
            {
                hResult = PrivAllocate (ulNodeSize + sizeof(FlatPropNode), (LPVOID *)&pNode, "Stream Loaded Property Data");
                if (!hResult)
                {
                    hResult = pStream->Read (&pNode->Prop.ulPropTag, ulNodeSize - sizeof(ULONG), NULL);
                    if (!hResult)
                    {
                        LPBYTE pByte = (LPBYTE)pNode + sizeof(FlatPropNode);
                        FixUpPointers (pNode, pByte);
                        pNode->Prop.cbSize = ulNodeSize;
                        if (!m_pHead)
                        {   
                            m_pHead = m_pTail = pNode;
                        }
                        else
                        {
                            m_pTail->pNext = pNode;
                            m_pTail = pNode;
                        }
                        m_ulNodes++;
                    }
                    if (hResult)
                    {
                        PrivFree ((LPVOID *)&pNode);
                    }
                }
            }
        }
        if (hResult && i)
        {
            TraceResult ("CPropList::LoadFromStream: Failed to load all property nodes", hResult);
            hResult = S_OK;
        }
    }
    if (hResult)
    {
        // Free what ever we started to load from the stream ...
        DeleteList (&m_pHead);
        // ... and restore the original state of this list
        m_ulNodes = tmpNodes;
        m_pTail = tmpTail;
        m_pHead = tmpHead;
        TraceResult ("LoadFromStream: I/O operation failed", hResult);
    }
    else
    {
        // Success. Free the old list and reset the 'dirty' flag
        DeleteList (&tmpHead);
        m_fDirty = FALSE;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::CopyToStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropList::CopyToStream (ULONG ulPropTag, LPSTREAM pStream)
{
    LPVOID pVoid;
    DWORD cbSize;
    SPropValue spvOne;
    HRESULT hResult = Find (ulPropTag, &spvOne);
    if (!hResult)
    {
        if (PT_TSTRING == PROP_TYPE(spvOne.ulPropTag) || PT_BINARY == PROP_TYPE(spvOne.ulPropTag))
        {
            GetSize (ulPropTag, spvOne.Value, cbSize, &pVoid);
            ASSERT (pVoid);
            if (pVoid)
            {
                hResult = pStream->Write (pVoid, cbSize, NULL);
            }
            else
            {
                hResult = E_FAIL;
            }
        }
        else
        {
            hResult = MAPI_E_NO_SUPPORT;
        }
    }
    if (!hResult)
    {
        LARGE_INTEGER liZero = { 0 };
        pStream->Seek (liZero, STREAM_SEEK_SET, NULL);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::CopyToList()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CPropList::CopyToList (CPropList * pDestList)
{
    if (!m_ulNodes)
    {
        return S_OK;
    }
    pDestList->DeleteAllProps();

    PFlatSPropValue pProp = Walk (WALK_START);
    while (pProp)
    {
        pDestList->Insert (pProp);
        pProp = Walk (WALK_CONTINUE);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::InternalNodePropReplaced()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI CPropList::InternalNodePropReplaced (LPSPropValue pProp)
{
    if (MV_FLAG & PROP_TYPE(pProp->ulPropTag) ||
        PROP_TYPE(pProp->ulPropTag) == PT_TSTRING ||
        PROP_TYPE(pProp->ulPropTag) == PT_CLSID ||
        PROP_TYPE(pProp->ulPropTag) == PT_BINARY)
    {
        return FALSE;
    }
    PFlatPropNode pNode = m_pHead;
    while (pNode)
    {
        if (PROP_ID(pNode->Prop.ulPropTag) == PROP_ID(pProp->ulPropTag) &&
            PROP_TYPE(pNode->Prop.ulPropTag) == PROP_TYPE(pProp->ulPropTag))
        {
            // This will copy the value of any of this single-value property types:
            // PT_FLOAT, PT_DOUBLE, PT_CURRENCY , PT_APPTIME , PT_SYSTIME , 
            // PT_BOOLEAN, PT_SHORT, PT_LONG and PT_LONGLONG.
            pNode->Prop.Value.li = pProp->Value.li;
            return TRUE;
        }
        pNode = pNode->pNext;
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropList::CopySmallStreams()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CPropList::CopySmallStreams (COpenPropList * pLargeProps)
{
    if (!pLargeProps)
    {
        return S_OK;
    }
    SPropValue spvProp = { 0 }, spvOldProp = { 0 };
    HRESULT hResult;
    DWORD dwStreamSize;
#define MAX_SMALL_STREAM    2048
    BYTE bBuffer[MAX_SMALL_STREAM+1];
    POPEN_PROPS pNode = pLargeProps->Walk (WALK_START);
    while (pNode)
    {
        if (PROP_STREAM & pNode->ulFlags && !(PROP_DELETE_THIS & pNode->ulFlags))
        {
            if (!pNode->iface.pStream->IsStreamInUse())
            {
                dwStreamSize = 0;
                pNode->iface.pStream->GetSize (&dwStreamSize);
                if (dwStreamSize > 0 && dwStreamSize < MAX_SMALL_STREAM)
                {
                    if (PROP_TYPE(pNode->ulPropTag) == PT_STRING8)
                    {
                        dwStreamSize++; // Add space for NULL terminator
                    }
                    else
                    {
                        if (PROP_TYPE(pNode->ulPropTag) == PT_UNICODE)
                        {
                            dwStreamSize += sizeof(WCHAR); // Add space for NULL terminator
                        }
                    }
                    hResult = pNode->iface.pStream->SaveToBuffer (bBuffer);
                    if (!hResult)
                    {
                        spvProp.ulPropTag = pNode->ulPropTag;
                        if (PROP_TYPE(spvProp.ulPropTag) == PT_TSTRING)
                        {
                            // TODO: This is not suitable for UNICODE
                            //          IrvingD 7/19/95
                            dwStreamSize--;
                            bBuffer[dwStreamSize] = 0;
                            spvProp.Value.lpszA = (LPSTR)bBuffer;
                        }
                        else
                        {
                            spvProp.Value.bin.cb = dwStreamSize;
                            spvProp.Value.bin.lpb = bBuffer;
                        }
                        Delete (spvProp.ulPropTag, &spvOldProp);
                        hResult = Insert (&spvProp);
                        if (hResult)
                        {
                            Insert (&spvOldProp);
                        }
                        else
                        {
                            pLargeProps->Delete (spvProp.ulPropTag);
                        }
                    }
                }
            }
        }
        pNode = pLargeProps->Walk (WALK_CONTINUE);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::Insert()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP COpenPropList::Insert (LPUNKNOWN    pUnk,
                                    ULONG        ulFlags,
                                    DWORD        dwStreamID,
                                    ULONG        ulPropTag)
{
    POPEN_PROPS pNewNode;
    HRESULT hResult = PrivAllocate (sizeof(OPEN_PROPS), (LPVOID *)&pNewNode, "CPropList's Node structure");
    if (hResult)
    {
        return hResult;
    }
    pNewNode->ulPropTag = ulPropTag;
    pNewNode->ulFlags = ulFlags;
    pNewNode->dwID = dwStreamID;
    // Since iface is a UNION, setting the pointer in any of the members
    // will assign it to all the members, therefore, the right interface pointer.
    pNewNode->iface.pStream = (CPropStream *)pUnk;
    InsertInList (&m_pHead, pNewNode);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::FreeList()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      None
//
void WINAPI COpenPropList::FreeList (BOOL fForce)
{
    POPEN_PROPS pNext, pPrev = NULL, pNode = m_pHead;
    BOOL fInUse = TRUE;
    LPUNKNOWN pUnk = NULL;
    while (pNode)
    {
        // First, chech what interface is this node holding, get a pointer to
        // its IUnknown root and see if the interface is in use by a client
        if (PROP_STREAM & pNode->ulFlags)
        {
            fInUse = pNode->iface.pStream->IsStreamInUse();
            pUnk = (LPUNKNOWN)pNode->iface.pStream;
        }
        if (PROP_STORAGE & pNode->ulFlags)
        {
            fInUse = pNode->iface.pStorage->IsStorageInUse();
            pUnk = (LPUNKNOWN)pNode->iface.pStorage;
        }
        // If the interface pointer is not in use or the caller is FORCING
        // the purging of the list, release the interface pointer and free
        // the memory for the node.
        if (!fInUse || fForce || (PROP_DELETE_THIS & pNode->ulFlags))
        {
            #ifdef _DEBUG
            if (fInUse) TraceMessage ("COpenPropList::FreeList: WARNING interface pointer still in use");
            #endif // _DEBUG
            // If the node is the head of the list, make it point to the next node
            if (m_pHead == pNode)
            {
                m_pHead = m_pHead->pNext;
            }
            // Save the next pointer so that we may delete the current node
            pNext = pNode->pNext;
            // Release the interface and free the node's structure memory
            pUnk->Release();
            PrivFree ((LPVOID *)&pNode);
            pNode = pNext;
            // If we have a previous node, we are doing a delete in the middle of a single link list
            if (pPrev)
            {
                pPrev->pNext = pNode;
            }
        }
        else
        {
            // Don't delete this node, save is as the "previous" node in the linear list
            pPrev = pNode;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::Find()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
POPEN_PROPS WINAPI COpenPropList::Find (ULONG ulPropTag)
{
    POPEN_PROPS pNode = m_pHead;
    while (pNode)
    {
        if (PROP_ID(pNode->ulPropTag) == PROP_ID(ulPropTag) && !(PROP_DELETE_THIS & pNode->ulFlags))
        {
            return pNode;
        }
        pNode = pNode->pNext;
    }
    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::Find()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP COpenPropList::Find (ULONG        ulPropTag,
                                  LPSPropValue pProp,
                                  LPVOID       pMemBlock,
                                  ULONG        ulMaxBytes)
{
    // TODO: This is to avoid a compiler warning. Check to make sure we release
    //       need to use this parameter.
    //          Irving 2/26/96 
    ulMaxBytes = ulMaxBytes;

    DWORD dwAllocSize, dwStoreData;
    LPVOID pVoid;
    HRESULT hResult = MAPI_E_NOT_FOUND;
    POPEN_PROPS pNode = m_pHead;
    while (pNode)
    {
        if (PROP_ID(pNode->ulPropTag) == PROP_ID(ulPropTag) && !(PROP_DELETE_THIS & pNode->ulFlags))
        {
            if (PROP_STREAM & pNode->ulFlags)
            {
                hResult = pNode->iface.pStream->GetSize (&dwStoreData);
                if (!hResult)
                {
                    dwAllocSize = dwStoreData;
                    if (PROP_TYPE(pNode->ulPropTag) == PT_TSTRING)
                    {
                        dwAllocSize++; // Add space for NULL terminator
                    }
                    hResult = PrivMAPIAllocate (dwAllocSize, &pVoid, pMemBlock); // Alloc more
                    if (!hResult)
                    {
                        hResult = pNode->iface.pStream->SaveToBuffer (pVoid);
                        if (!hResult)
                        {
                            if (PROP_TYPE(pNode->ulPropTag) == PT_TSTRING)
                            {
                                pProp->Value.lpszA = (LPSTR)pVoid;
                                pProp->Value.lpszA[dwStoreData] = 0;
                            }
                            else
                            {
                                pProp->Value.bin.cb = dwStoreData;
                                pProp->Value.bin.lpb = (LPBYTE)pVoid;
                            }
                        }
                    }
                }
            }
            else
            {
                hResult = E_OUTOFMEMORY;
            }
            TraceResult ("COpenPropList::Find", hResult);
            break; // Out of the WHILE() loop
        }
        pNode = pNode->pNext;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::Delete()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP COpenPropList::Delete (ULONG ulPropTag)
{
    LPUNKNOWN pUnk;
    HRESULT hResult = MAPI_E_NOT_FOUND;
    POPEN_PROPS pNext, pPrev = NULL, pNode = m_pHead;
    while (pNode)
    {
        if (PROP_ID(pNode->ulPropTag) == PROP_ID(ulPropTag))
        {
            // If the properties has never been saved, delete the node and release the interface pointer open on it
            if (PROP_NEVER_SAVED & pNode->ulFlags)
            {
                pUnk = (LPUNKNOWN)pNode->iface.pStream;
                // If the node is the head of the list, make the head point to the next node
                if (m_pHead == pNode)
                {
                    m_pHead = m_pHead->pNext;
                }
                // Save the next pointer so that we may delete the current node
                pNext = pNode->pNext;
                // Release the interface and free the node's structure memory
                pUnk->Release();
                PrivFree ((LPVOID *)&pNode);
                // If we have a previous node, we are doing a delete in the middle of a single link list
                if (pPrev)
                {
                    pPrev->pNext = pNext;
                }
            }
            else
            {
                pNode->ulFlags |= PROP_DELETE_THIS;
            }
            hResult = S_OK;
            break; // Out of the WHILE() loop
        }
        else
        {
            // Don't delete this node, save is as the "previous" node in the linear list
            pPrev = pNode;
            pNode = pNode->pNext;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    COpenPropList::Walk()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
POPEN_PROPS WINAPI COpenPropList::Walk (WALKING_ACTION waAction)
{
    if (!m_pHead)
    {
        return NULL;
    }
    if (WALK_START == waAction)
    {
        m_pWalkNext = m_pHead->pNext;
        return m_pHead;
    }
    if (WALK_CONTINUE == waAction && m_pWalkNext)
    {
        POPEN_PROPS pTmp = m_pWalkNext;
        m_pWalkNext = m_pWalkNext->pNext;
        return pTmp;
    }
    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    AllocateNotifKey()
//
//    Parameters
//      [IN]    cbSize      Number of bytes in pKeyData
//      [IN]    pKeyData    Array of bytes with the information to the
//                          notification key.
//      [OUT]   ppKey       Pointer to a pointer where a newly allocated
//                          NOTIFKEY structure is returned. If the funciton
//                          is successful, the caller is responsible for
//                          freeing the memory for this structure with the
//                          MAPI alloc functions.
//
//    Purpose
//      This function allocates a notification key structure and copies
//      the data passed in by the caller.
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI AllocateNotifKey (DWORD              cbSize,
                                 LPBYTE             pKeyData,
                                 LPNOTIFKEY *       ppKey)
{
    HRESULT hResult = PrivAllocate (CbNewNOTIFKEY(cbSize), (LPVOID *)ppKey, "Notification Key");
    TraceResult ("AllocateNotifKey: Failed to allocate new key", hResult);
    if (!hResult)
    {
        (*ppKey)->cb = cbSize;
        CopyMemory ((*ppKey)->ab, pKeyData, cbSize);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    IncludePropInCopy()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI IncludePropInCopy (ULONG ulPropTag, LPSPropTagArray pTags)
{
    if (PR_MESSAGE_FLAGS          == ulPropTag ||
        PR_CREATION_TIME          == ulPropTag ||
        PR_LAST_MODIFICATION_TIME == ulPropTag ||
        PR_MESSAGE_SIZE           == ulPropTag ||
        PR_ATTACH_SIZE            == ulPropTag ||
        PR_SEARCH_KEY             == ulPropTag ||
        PR_PARENT_ENTRYID         == ulPropTag)
    {
        return FALSE;
    }
    if (pTags)
    {
        for (ULONG i=0; i<pTags->cValues; i++)
        {
            if (PROP_ID(ulPropTag) == PROP_ID(pTags->aulPropTag[i]))
            {
                return FALSE;
            }
        }
    }
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
//    IsPropInRow()
//
//    Parameters
//      
//    Purpose
//
//    Return Value
//
BOOL WINAPI IsPropInRow (LPSRow            pRow,
                         ULONG             ulPropTag,
                         LPSPropValue *    ppProp)
{
    for (ULONG i=0; i<pRow->cValues; i++)
    {
        if (pRow->lpProps[i].ulPropTag == ulPropTag)
        {
            if (ppProp)
            {
                *ppProp = &pRow->lpProps[i];
            }
            return TRUE;
        }
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    CreatePrivateIStorage()
//
//    Parameters
//      
//    Purpose
//
//    Return Value
//
HRESULT WINAPI CreatePrivateIStorage (CStorage ** ppStorage,
                                      BOOL        fWriteAccess,
                                      LPSTREAM    pStream)
{
    BOOL fReleaseStream = FALSE;
    *ppStorage = NULL;
    CStorage * pStg;
    LPSTORAGE pImpStg;
    HRESULT hResult = S_OK;
    char szFile[512];
    DWORD dwFlags = STGSTRM_MODIFY;
    if (pStream)
    {
        dwFlags |= STGSTRM_CURRENT;
    }
    else
    {
        dwFlags |= STGSTRM_CREATE;
        if (GetTempFullPathFileName (szFile))
        {
            hResult = OpenStreamOnFile (GetMAPIAllocBuffer(),
                                        GetMAPIFreeBuffer(),
        		                        STGM_CREATE | STGM_READWRITE | STGM_DELETEONRELEASE,
        		                        szFile,
        		                        NULL,
        		                        &pStream);
            fReleaseStream = TRUE;
        }
        else
        {
            hResult = E_FAIL;
        }
    }
    if (!hResult)
    {
        hResult = HrIStorageFromStream (pStream, NULL, dwFlags, &pImpStg);
        if (!hResult)
        {
            pStg = new CStorage (pImpStg, pStream, fWriteAccess);
            // The constructor did an AddRef() on these object so we can release
            // them. If the construction failed, this serves as cleanup code.
            pImpStg->Release();
            if (fReleaseStream)
            {
                pStream->Release();
            }
            // Were we successful at allocating the new object?
            if (pStg)
            {
                *ppStorage = pStg;
            }
            else
            {
                hResult = E_OUTOFMEMORY;
            }
        }
    }
    TraceResult ("CreatePrivateIStorage", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CStorage::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IStorage || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::CreateStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::CreateStream (LPSTR      pszName,
                                     DWORD      grfMode,
                                     DWORD      reserved1,
                                     DWORD      reserved2,
                                     LPSTREAM * ppstm)
{
    HRESULT hResult = S_OK;
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        hResult = HRESULT_FROM_WIN32(GetLastError());
        TraceResult ("CreateStream: Failed to convert string", hResult);
        return hResult;
    }
    hResult = m_pImpStorage->CreateStream (OleStrBuffer, grfMode, reserved1, reserved2, ppstm);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::OpenStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::OpenStream (LPSTR        pszName,
                                   void *       reserved1,
                                   DWORD        grfMode,
                                   DWORD        reserved2,
                                   LPSTREAM *   ppstm)
{
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("OpenStream: Failed to convert string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    HRESULT hResult = m_pImpStorage->OpenStream (OleStrBuffer, reserved1, grfMode, reserved2, ppstm);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::CreateStorage()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::CreateStorage (LPSTR       pszName,
                                      DWORD       grfMode,
                                      DWORD       dwStgFmt,
                                      DWORD       reserved2,
                                      CStorage ** ppstg)
{
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("CreateStorage: Failed to convert string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    LPSTORAGE pStg;
    HRESULT hResult = m_pImpStorage->CreateStorage (OleStrBuffer, grfMode, dwStgFmt, reserved2, &pStg);
    if (SUCCEEDED(hResult))
    {
        CStorage * pObj = new CStorage (pStg, m_pImpStream, m_fWriteOk);
        pStg->Release();
        if (pObj)
        {
            *ppstg = pObj;
            SetDirty (TRUE);
        }
        else
        {
            TraceMessage ("CreateStorage: Failed to allocate new storage object");
            hResult = E_OUTOFMEMORY;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::OpenStorage()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::OpenStorage (LPSTR       pszName,
                                    LPSTORAGE   pstgPriority,
                                    DWORD       grfMode,
                                    SNB         snbExclude,
                                    DWORD       reserved,
                                    CStorage ** ppstg)
{
    ASSERT (m_pImpStorage);
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("OpenStorage: Failed to convert string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    LPSTORAGE pStg;
    HRESULT hResult = m_pImpStorage->OpenStorage (OleStrBuffer, pstgPriority, grfMode, snbExclude, reserved, &pStg);
    if (SUCCEEDED(hResult))
    {
        CStorage * pObj = new CStorage (pStg, m_pImpStream, m_fWriteOk);
        pStg->Release();
        if (pObj)
        {
            *ppstg = pObj;
            SetDirty (TRUE);
        }
        else
        {
            TraceMessage ("OpenStorage: Failed to allocate new storage object");
            hResult = E_OUTOFMEMORY;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::MoveElementTo()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::MoveElementTo (LPSTR      pszName,
                                      LPSTORAGE  pstgDest,
                                      LPSTR      pszNewName,
                                      DWORD      grfFlags)
{
    OLECHAR OleStrBuffer1[256], OleStrBuffer2[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer1,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("MoveElementTo: Failed to convert first string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    iChars = MultiByteToWideChar (CP_ACP,
                                  MB_PRECOMPOSED,
                                  pszNewName,
                                  -1,
                                  OleStrBuffer2,
                                  256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("MoveElementTo: Failed to convert second string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    HRESULT hResult = m_pImpStorage->MoveElementTo (OleStrBuffer1, pstgDest, OleStrBuffer2, grfFlags);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::DestroyElement()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::DestroyElement (LPSTR pszName)
{
    ASSERT (m_pImpStorage);
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("DestroyElement: Failed to convert string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    HRESULT hResult = m_pImpStorage->DestroyElement (OleStrBuffer);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::RenameElement()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::RenameElement (LPSTR  pszOldName,
                                      LPSTR  pszNewName)
{
    ASSERT (m_pImpStorage);
    OLECHAR OleStrBuffer1[256], OleStrBuffer2[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszOldName,
                                      -1,
                                      OleStrBuffer1,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("RenameElement: Failed to convert first string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    iChars = MultiByteToWideChar (CP_ACP,
                                  MB_PRECOMPOSED,
                                  pszNewName,
                                  -1,
                                  OleStrBuffer2,
                                  256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("RenameElement: Failed to convert second string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    HRESULT hResult = m_pImpStorage->RenameElement (OleStrBuffer1, OleStrBuffer2);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CStorage::SetElementTimes()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CStorage::SetElementTimes (LPSTR            pszName,
                                        const FILETIME * pctime,
                                        const FILETIME * patime,
                                        const FILETIME * pmtime)
{
    ASSERT (m_pImpStorage);
    OLECHAR OleStrBuffer[256];
    int iChars = MultiByteToWideChar (CP_ACP,
                                      MB_PRECOMPOSED,
                                      pszName,
                                      -1,
                                      OleStrBuffer,
                                      256);
    if (0 == iChars)
    {
        DWORD dwLastError = GetLastError();
        TraceSysError ("SetElementTimes: Failed to convert string", dwLastError);
        return HRESULT_FROM_WIN32 (dwLastError);
    }
    HRESULT hResult = m_pImpStorage->SetElementTimes (OleStrBuffer, pctime, patime, pmtime);
    if (!hResult)
    {
        SetDirty (TRUE);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropStream::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPropStream::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IStream || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropStream::GetSize()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      None.
//
STDMETHODIMP CPropStream::GetSize (DWORD * pdwSize)
{
    STATSTG stat;
    HRESULT hResult = m_pImpStream->Stat (&stat, STATFLAG_NONAME);
    TraceResult ("CPropStream::GetSize", hResult);
    if (!hResult)
    {
        if (stat.cbSize.HighPart)
        {
            hResult = E_OUTOFMEMORY;
        }
        else
        {
            *pdwSize = stat.cbSize.LowPart;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropStream::SaveToStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      None.
//
STDMETHODIMP CPropStream::SaveToStream (LPSTREAM pStream)
{
    STATSTG stat = { 0 };
    LARGE_INTEGER liOldPos, liZeroOffSet = { 0 };
    HRESULT hResult = m_pImpStream->Seek (liZeroOffSet, STREAM_SEEK_CUR, (ULARGE_INTEGER *)&liOldPos);
    if (!hResult)
    {
        // Put the stream at the beginning
        hResult = m_pImpStream->Seek (liZeroOffSet, STREAM_SEEK_SET, NULL);
        if (!hResult)
        {
            // How big is it
            hResult = m_pImpStream->Stat (&stat, STATFLAG_NONAME);
            if (!hResult)
            {
                // Dump the data into the supplied stream, which should be at the start
                hResult = m_pImpStream->CopyTo (pStream, stat.cbSize, NULL, NULL);
                if (!hResult)
                {
                    hResult = pStream->SetSize (stat.cbSize);
                }
            }
        }
        // Restore the original position in the stream to leave it intact.
        m_pImpStream->Seek (liOldPos, STREAM_SEEK_SET, NULL);
    }
    TraceResult ("CPropStream::SaveToStream", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPropStream::SaveToBuffer()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CPropStream::SaveToBuffer (LPVOID pVoid)
{
    LARGE_INTEGER liOldPos, liZeroOffSet = { 0 };
    DWORD cbSize;
    HRESULT hResult = GetSize (&cbSize);
    if (!hResult)
    {
        hResult = m_pImpStream->Seek (liZeroOffSet, STREAM_SEEK_CUR, (ULARGE_INTEGER *)&liOldPos);
        if (!hResult)
        {
            // Put the stream at the beginning
            hResult = m_pImpStream->Seek (liZeroOffSet, STREAM_SEEK_SET, NULL);
            if (!hResult)
            {
                // Dump the data into the supplied buffer
                hResult = m_pImpStream->Read (pVoid, cbSize, NULL);
            }
            // Restore the original position in the stream to leave it intact.
            m_pImpStream->Seek (liOldPos, STREAM_SEEK_SET, NULL);
        }
    }
    TraceResult ("CPropStream::SaveToBuffer", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    GetTempFullPathFileName()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
BOOL WINAPI GetTempFullPathFileName (LPSTR szFileName)
{
    char szTmpDir[512];
    DWORD dwChars = GetTempPath (512, szTmpDir);
    if (dwChars)
    {
        dwChars = GetTempFileName (szTmpDir,
                                   "TMP",
                                   0,
                                   szFileName);
    }
    if (!dwChars)
    {
        TraceResult ("GetTempFullPathFileName: Failed to get temp path or file name", GetLastError());
    }
    return (dwChars ? TRUE : FALSE);
}

// End of file for PROPLIST.CPP
