///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSATTACH.CPP
//
//  Description
//      This file implements the IAttach interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSWDS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

#define NUM_ATTACH_PROPS        13
// Default internal properties available in our implementation of IAttach
// Other properties could be available if a client or MAPI added them using
// IMessage::SetProps()
SizedSPropTagArray(NUM_ATTACH_PROPS, sptAttachObj) =
{
    NUM_ATTACH_PROPS,
    {
        PR_ACCESS,                      // READ only
        PR_ACCESS_LEVEL,                // READ only
        PR_ATTACH_NUM,                  // READ only
        PR_ATTACH_SIZE,                 // READ only
        PR_STORE_ENTRYID,               // READ only
        PR_STORE_RECORD_KEY,            // READ only
        PR_RECORD_KEY,                  // READ only
        PR_MAPPING_SIGNATURE,           // READ only
        PR_OBJECT_TYPE,                 // READ only
        PR_STORE_SUPPORT_MASK,          // READ only
        PR_CREATION_TIME,               // READ only
        PR_CREATION_VERSION,            // READ only
        PR_LAST_MODIFICATION_TIME       // READ only
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CAttach::CAttach()
//
//    Parameters
//      [IN]    pStore          Pointer to the container IMsgStore object
//      [IN]    pParentEID      Entry ID of the parent folder
//      [IN]    pAttachEID       Private "Entry ID" of this attachment
//      [IN]    pList           In-memory property cache for this object
//      [IN]    fDeferErrors    Weather or not we should defer errors
//                              in this instance
//      [IN]    lAccess         Access level of this object
//      [IN]    pMsgObj         Pointer to the parent IMessage object
//                              where the attachment lives
//      [IN]    pNPList         Pointer to the named properties manager
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CAttach::CAttach (CMsgStore *       pStore,
                  PRIVATE_ENTRYID * pParentEID,
                  PRIVATE_ENTRYID * pAttachEID,
                  CPropList *       pList,
                  BOOL              fDeferErrors,
                  long              lAccess,
                  CMessage *        pMsgObj,
                  CNamePropMgr *    pNPList) : CGenericProp (pStore->m_pBEobj,
                                                             pParentEID,
                                                             pAttachEID,
                                                             pList,
                                                             fDeferErrors,
                                                             lAccess)
{
    InfoTrace ("CAttach: Constructor called");
    m_pNPList       = pNPList;
    m_pParentMsgObj = pMsgObj;
    m_pParentMsgObj->AddRef();
    m_pActionList = NULL;
    m_pStore = pStore;
    m_pStore->AddRef();
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::~CAttach()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CAttach.
//
//    Return Value
//      None
//
CAttach::~CAttach ()
{
    InfoTrace ("CAttach: Destructor called");
    m_pStore->Release();
    m_pStore = NULL;
    m_pParentMsgObj->Release();
    m_pParentMsgObj = NULL;
    if (m_pActionList)
    {
        TRANSACTED_OBJECTS * pNode = m_pActionList;
        while (pNode)
        {
            if (TRANS_OBJ_ACTION_SAVE == pNode->Action)
            {
                ASSERT (MAPI_MESSAGE == pNode->ObjEID.bObject);
                pNode->pUnkObj->Release();
                pNode->pUnkObj = NULL;
            }
            ASSERT (NULL == pNode->pUnkObj);
            pNode = pNode->pNext;
        }
        DeleteList (&m_pActionList);
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      S_OK            If successful. With the interface pointer in *ppvObj
//      E_NOINTERFACE   If interface requested is not supported by this object
//
STDMETHODIMP CAttach::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IAttachment ||
        riid == IID_IMAPIProp   ||
        riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CAttach::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method finds a string description for the last error cached in.
//      If the error cached does not match the error passed by the client, we
//      fail with MAPI_E_NOT_FOUND. The MAPIERROR structure is allocated
//      dynamically, so the caller is responsible for releasing it with
//      the MAPI allocation functions.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::GetLastError (HRESULT       hError,
                                    ULONG         ulFlags,
                                    LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CAttach::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this,
                                     hError,
                                     ulFlags,
                                     ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_pStore->m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::SaveChanges()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function will save the properties on this object from the cache
//      into the message database file.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CAttach::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);

    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CAttach::SaveChanges: Cannot SaveChanges in a READ only object");
        TraceResult ("CAttach::SaveChanges", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    EnterCriticalSection (&m_csObj);
    EnterCriticalSection (&m_pParentMsgObj->m_csObj);
    TRANSACTED_OBJECTS * pNewNode;
    HRESULT hResult = PrivAllocate (sizeof(TRANSACTED_OBJECTS), (LPVOID *)&pNewNode, "Attach Action Node (Save)");
    if (!hResult)
    {
        pNewNode->Action = TRANS_OBJ_ACTION_SAVE;
        pNewNode->ObjEID = m_ObjectEID;
        pNewNode->pUnkObj = (LPUNKNOWN)this;
        pNewNode->pUnkObj->AddRef();
        InsertInList (&(m_pParentMsgObj->m_pActionList), pNewNode);
        if (ulFlags & KEEP_OPEN_READONLY)
        {
            // The user need only read access from now on.
            m_lAccess = MAPI_ACCESS_READ;
        }
        else
        {
            // If this flag was not passed in, close the object and change the access flags.
            if (!(ulFlags & KEEP_OPEN_READWRITE))
            {
                m_lAccess = 0;
            }
        }
    }
    LeaveCriticalSection (&m_pParentMsgObj->m_csObj);
    LeaveCriticalSection (&m_csObj);

    TraceResult ("CAttach::SaveChanges", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::SaveChangesNow()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description
//        on IMAPIProp::SaveChanges() }
//
//    Purpose
//      Commits transacted actions on the object. Since attachments are
//      transacted changes to its properties are not save to permanenent
//      storage until the container parent IMessage object is called in
//      it own SaveChanges.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::SaveChangesNow (ULONG ulFlags)
{
    EnterCriticalSection (&m_csObj);
    long lOldAccess = m_lAccess;
    // Change the access flags just to make sure we have READ/WRITE access
    m_lAccess = MAPI_ACCESS_MODIFY | MAPI_ACCESS_DELETE | MAPI_ACCESS_READ;
    // Tell the base class to save itself
    HRESULT hResult = CGenericProp::SaveChanges (ulFlags, NULL);
    if (!hResult)
    {
        // Walk the list of transacted actions and execute them now
        TRANSACTED_OBJECTS * pNode = m_pActionList;
        while (pNode)
        {
            if (TRANS_OBJ_ACTION_DELETE == pNode->Action)
            {   
                pNode->Action = TRANS_OBJ_ACTION_UNDEFINED;
            }
            if (TRANS_OBJ_ACTION_SAVE == pNode->Action)
            {
                ASSERT (MAPI_MESSAGE == pNode->ObjEID.bObject);
                ((CMessage *)pNode->pUnkObj)->SaveChangesNow (FORCE_SAVE);
                pNode->pUnkObj->Release();
                pNode->pUnkObj = NULL;
                pNode->Action = TRANS_OBJ_ACTION_UNDEFINED;
            }
            if (TRANS_OBJ_ACTION_DELETE_OBJ_PROP == pNode->Action)
            {   
                PRIVATE_ENTRYID eid = { 0 };
                ASSERT (PT_STREAM_STORAGE == PROP_TYPE (pNode->Prop.ulPropTag));
                if (pNode->Prop.Value.li.HighPart)
                {
                    eid.dwObjID = pNode->Prop.Value.li.HighPart;
                    m_pBEobj->DeleteOneObject (eid, m_ObjectEID);
                }
                pNode->Action = TRANS_OBJ_ACTION_UNDEFINED;
            }
            pNode = pNode->pNext;
        }

        // After the attachment object has been updated, update the
        // attachment table in the parent message object.
        if (OBJECT_NEVER_SAVED & m_ulObjFlags)
        {
            m_ulObjFlags &= ~OBJECT_NEVER_SAVED;
            m_pParentMsgObj->UpdateAttachmentTable (m_pList,
                                                    m_ObjectEID,
                                                    UPD_TABLE_ADD_ROW);
        }
        else
        {
            m_pParentMsgObj->UpdateAttachmentTable (m_pList,
                                                    m_ObjectEID,
                                                    UPD_TABLE_REFRESH);
        }
    }
    // Restore the previous access level
    m_lAccess = lOldAccess;
    // Unlock the object
    LeaveCriticalSection (&m_csObj);

    TraceResult ("CAttach::SaveChanges", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::GetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the value of the properties the client especified in the
//      property tag array. If the tag array is NULL, the user wants all
//      the properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::GetProps (LPSPropTagArray pPropTagArray,
                                ULONG           ulFlags,
                                ULONG *         pcValues,
                                LPSPropValue *  ppPropArray)
{
    InfoTrace ("CAttach::GetProps method called");
    Validate_IMAPIProp_GetProps (this,
                                 pPropTagArray,
                                 ulFlags,
                                 pcValues,
                                 ppPropArray);
    #ifdef DO_INFO_TRACES
    if (!pPropTagArray)
    {
        TraceMessage ("GetProps: All properties requested");
    }
    else
    {
        TraceMessage ("Properties requested...");
        for (ULONG j=0; j<pPropTagArray->cValues; j++)
        { TraceProp (pPropTagArray->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES

    EnterCriticalSection (&m_csObj);
    BOOL fGotErrors = FALSE;
    HRESULT hResult = CGenericProp::GetProps (pPropTagArray,
                                              ulFlags,
                                              pcValues,
                                              ppPropArray,
                                              (LPSPropTagArray)&sptAttachObj);
    if (hResult)
    {
        goto ErrorExit;
    }
    
    LPVOID pSrc;
    ULONG i, cbSize;
    for (i=0; i<*pcValues; i++)
    {
        pSrc = NULL;
        hResult = S_OK;
        switch ((*ppPropArray)[i].ulPropTag)
        {
            // Get the PT_OBJECT/PR_NULL properties
            case PR_NULL :
                (*ppPropArray)[i].Value.err = S_OK;
                break;
            // Get the PT_LONG properties
            case PR_ACCESS :
                (*ppPropArray)[i].Value.l = m_lAccess;
                break;
            case PR_ACCESS_LEVEL :
                (*ppPropArray)[i].Value.l = (m_lAccess & MAPI_ACCESS_MODIFY ? MAPI_MODIFY : 0);
                break;
            case PR_OBJECT_TYPE :
                (*ppPropArray)[i].Value.l = MAPI_ATTACH;
                break;
            case PR_STORE_SUPPORT_MASK:
                (*ppPropArray)[i].Value.l = SUPPORT_MASK;
                break;
            case PR_ATTACH_NUM :
                (*ppPropArray)[i].Value.l = m_ObjectEID.dwObjID;
                break;
            // Get the PT_I8 properties
            case PR_CREATION_VERSION :
                (*ppPropArray)[i].Value.li.LowPart =  MSG_STORE_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = MSG_STORE_VERSION_MINOR;
                break;
            // Get the PT_BINARY properties
            case PR_STORE_RECORD_KEY :
            case PR_MAPPING_SIGNATURE :
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&m_pStore->m_StoreUID;
                break;
            case PR_STORE_ENTRYID :
                cbSize = m_pStore->m_StoreEID.cb;
                pSrc = (LPVOID)m_pStore->m_StoreEID.lpb;
                break;
            case PR_RECORD_KEY :
                cbSize = sizeof(DWORD);
                pSrc = (LPVOID)&m_ObjectEID.dwObjID;
                break;
            // Oh no, the property was not found.
            default :
                hResult = CGenericProp::GetProps (&(*ppPropArray)[i], *ppPropArray);
                #ifdef DO_INFO_TRACES
                if (MAPI_E_NOT_FOUND == hResult)
                {
                    TraceRaw ("CAttach::GetProps: Not Found - ");
                    TraceProp ((*ppPropArray)[i].ulPropTag);
                }
                #endif // #DO_INFO_TRACES
                break;
        }
        if (S_OK != CGenericProp::GetProps (hResult,
                                            cbSize,
                                            pSrc,
                                            &(*ppPropArray)[i],
                                            *ppPropArray))
        {
            fGotErrors = TRUE;
        }
    }
ErrorExit :
    LeaveCriticalSection (&m_csObj);
    if (fGotErrors)
	{
		hResult = MAPI_W_ERRORS_RETURNED;
    }
    TraceResult ("CAttach::GetProps", hResult);
    SetLastError (hResult);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::GetPropList()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return an array of property tags with all the properties
//      in this object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::GetPropList (ULONG ulFlags, LPSPropTagArray * ppAllTags)
{
    InfoTrace ("CMessage::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppAllTags);
    return CGenericProp::GetPropList (ulFlags, (LPSPropTagArray)&sptAttachObj, ppAllTags);
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::OpenProperty()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method is used to open an interface on a given property. Namely,
//      we support opening and IMessage, IStream, or IStorage interface on
//      the PR_ATTACH_DATA_BIN or PR_ATTACH_DATA_OBJ properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::OpenProperty (ULONG       ulPropTag,
                                    LPCIID      piid,
                                    ULONG       ulInterfaceOptions,
                                    ULONG       ulFlags,
                                    LPUNKNOWN * ppUnk)
{
    InfoTrace ("CAttach::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this,
                                     ulPropTag,
                                     piid,
                                     ulInterfaceOptions,
                                     ulFlags,
                                     ppUnk);
    HRESULT hResult = CGenericProp::OpenProperty (ulPropTag, piid, ulInterfaceOptions, ulFlags, ppUnk);
    if (PR_ATTACH_DATA_OBJ == ulPropTag && MAPI_E_NO_SUPPORT == hResult && *piid == IID_IMessage)
    {
        // Embeeded messages are not required to have a real entry ID,
        // but for our implementation it is easier to have one than not to.
        PRIVATE_ENTRYID MessageEID = { 0 };
        long lAccess;
        ULONG ulObjFlags = OBJECT_MESSAGE_IN_MESSAGE;
        CPropList * pList;
        CMessage * pMsgObj;
        SPropValue spvProp;
        // If we are creating a new message embedded in this
        // attachment, we need to do some special things...
        if (MAPI_CREATE & ulFlags)
        {
            // ...Set up some flags for the object
            ulObjFlags |= (OBJECT_JUST_CREATED | OBJECT_NEVER_SAVED);
            // Allocate a new UID for the Search Key property
            MAPIUID uidSearchKey;
            hResult = m_pStore->m_pSupObj->NewUID (&uidSearchKey);
            if (!hResult)
            {
                // Now we are ready to crteate
                hResult = m_pBEobj->CreateObject (m_ObjectEID,
                                                  &MessageEID,
                                                  &pList,
                                                  MAPI_MESSAGE,
                                                  (LPVOID)&uidSearchKey,
                                                  (LPVOID)FALSE);
            }
        }
        else
        {
            // Find the property in the set to see if we have this property stored in the database
            hResult = m_pList->Find (ulPropTag, &spvProp, FALSE);
            if (!hResult)
            {
                if (PT_STREAM_STORAGE != PROP_TYPE(spvProp.ulPropTag) ||
                    PROP_TYPE(ulPropTag) != LOWORD(spvProp.Value.li.LowPart) ||
                    PROP_MESSAGE != HIWORD(spvProp.Value.li.LowPart))
                {
                    hResult = MAPI_E_NOT_FOUND;
                }
                else
                {
                    MessageEID = m_ParentEID; // This will copy the fields of the entry ID
                    MessageEID.bObject = MAPI_MESSAGE;
                    MessageEID.dwObjID = spvProp.Value.li.HighPart;
                    hResult = m_pBEobj->OpenObject (MessageEID, &pList);
                }
            }
        }
        if (!hResult)
        {
            // Inherit the parent's access rights
            lAccess = (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                      (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                      MAPI_ACCESS_READ;
            pMsgObj = new CMessage (m_pStore,
                                    &m_ObjectEID, // For embeded messages, PR_PARENT_ENTRYID does not need to be supported.
                                    &MessageEID,
                                    pList,
                                    !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                    lAccess,
                                    this,
                                    m_pNPList);
            if (pMsgObj)
            {
                *ppUnk = (LPUNKNOWN)pMsgObj;
                pMsgObj->m_ulObjFlags |= ulObjFlags;
                if (MAPI_CREATE & ulFlags)
                {
                    m_pList->InsertOpenProp (ulPropTag, MessageEID.dwObjID, PROP_MESSAGE);
                }
            }
            else
            {
                hResult = E_OUTOFMEMORY;
                delete pList;
            }
        }
    }
    TraceResult ("CAttach::OpenProperty", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::SetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Modify the value of the properties in our object. On transacted
//      objects (such as this), the changes to the property cache are not
//      permanent until the SaveChanges() method is called.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::SetProps (ULONG                 cValues,
                                LPSPropValue          pPropArray,
                                LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CAttach::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CAttach::SetProps: Cannot set properties in a READ only object");
        TraceResult ("CAttach::SetProps", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    
    #ifdef DO_INFO_TRACES
    TraceMessage ("Properties to set...");
    for (ULONG j=0; j<cValues; j++) { TraceProp (pPropArray[j].ulPropTag); }
    #endif DO_INFO_TRACES

    HRESULT hTagError, hResult = S_OK;
    EnterCriticalSection (&m_csObj);

    for (ULONG i=0; i<cValues; i++)
    {
        // Ignore this property tags
        switch (PROP_TYPE(pPropArray[i].ulPropTag))
        {
            case PT_NULL :
            case PT_UNSPECIFIED :
            case PT_ERROR :
                continue;
        }
        // Add the property to the property cache
        hTagError = AddPropToList (&pPropArray[i]);
        if (hTagError && ppProblems)
        {
            // If we had an error adding the property return a problem array to the caller.
            hResult = AddTagToProbArray (cValues,
                                         ppProblems,
                                         i,
                                         pPropArray[i].ulPropTag,
                                         hTagError);
            if (hResult)
            {
                break; // Out of the FOR() loop
            }
        }
    }
    LeaveCriticalSection (&m_csObj);
    // If an error occurred while adding errors to the problem array, the user
    // won't attempt to delete it so we should.
    if (hResult)
    {
        if (ppProblems)
        {
            PrivMAPIFree ((LPVOID *)ppProblems);
        }
        TraceResult ("CAttach::SetProps", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::DeleteProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function deletes a set of properties specified by the client. In
//      our implementation, we simply defer to the base class to let it delete
//      the properties in the property cache. On IAttach objects, the changes
//      are transacted and are only committed when IAttach::SaveChanges()
//      is invoked.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::DeleteProps (LPSPropTagArray       pPropTagArray,
                                   LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CAttach::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    // Check the access level of the object
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CAttach::DeleteProps: Cannot delete properties in a READ only object");
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    HRESULT hTagError, hResult = S_OK;
    LPSPropTagArray pTags = NULL;
    EnterCriticalSection (&m_csObj);
    __try
    {
        // If we are supposed to return any problems, set the pointer to NULL;
        if (ppProblems)
        {
            *ppProblems = NULL;
        }

        // Copy the array of properties because we are going to modify some of the property tags
        DWORD cbSize = CbNewSPropTagArray (pPropTagArray->cValues);
        hResult = PrivAllocate (cbSize, (LPVOID *)&pTags, "CAttach::DeleteProps");
        if (!hResult)
        {
            CopyMemory (pTags, pPropTagArray, cbSize);
            // For the "Interface" properties, we must do a special kind of delete...
            for (ULONG i=0; i<pTags->cValues; i++)
            {
                if (PR_ATTACH_DATA_BIN == pTags->aulPropTag[i] ||
                    PR_ATTACH_DATA_OBJ == pTags->aulPropTag[i])
                {
                    // ... Attachments are a transacted object note the delete transaction
                    // and defer deleting the property until the caller invokes SaveChanges
                    // in this object and in the container IMessage object
                    TRANSACTED_OBJECTS * pNewNode;
                    SPropValue spvProp;
                    hTagError = m_pList->Find (pTags->aulPropTag[i], &spvProp, FALSE);
                    if (!hTagError)
                    {
                        // Allocate a node for the transacted actions taken on this object
                        hTagError = PrivAllocate (sizeof(TRANSACTED_OBJECTS), (LPVOID *)&pNewNode, "Action Node (Att:DeleteProps)");
                        if (!hTagError)
                        {
                            pNewNode->Action = TRANS_OBJ_ACTION_DELETE_OBJ_PROP;
                            pNewNode->ObjEID = m_ObjectEID;
                            pNewNode->Prop = spvProp;
                            InsertInList (&m_pActionList, pNewNode);
                            m_pList->Delete (pTags->aulPropTag[i]);
                        }
                    }
                    // Report any problems
                    if (hTagError && ppProblems)
                    {
                        AddTagToProbArray (pTags->cValues, ppProblems, i, pTags->aulPropTag[i], hTagError);
                    }
                    // Null the property tag so the base class won't attempt 
                    // to delete this property tag from the list
                    pTags->aulPropTag[i] = PR_NULL;
                }
            }
            // Call the base class to contine deleting the other property tags
            CGenericProp::DeleteProps (pTags, ppProblems);
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CAttach::DeleteProps: An exception occured. Exception code: %#08x", GetExceptionCode());
        hResult = E_FAIL;
    }
    if (pTags)
    {
        PrivFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CAttach::DeleteProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::CopyTo()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy this object's content into another one. For the copy operation
//      we delegate to MAPI for the implemenation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::CopyTo (ULONG                 ciidExclude,
                              LPCIID                rgiidExclude,
                              LPSPropTagArray       pExcludeProps,
                              ULONG                 ulUIParam,
                              LPMAPIPROGRESS        pProgress,
                              LPCIID                pDestInterface,
                              LPVOID                pDestObj,
                              ULONG                 ulFlags,
                              LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CAttach::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,   
                               rgiidExclude,  
                               pExcludeProps, 
                               ulUIParam,     
                               pProgress,     
                               pDestInterface,
                               pDestObj,      
                               ulFlags,       
                               ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyTo (&IID_IAttachment,
                                                     this,
                                                     ciidExclude,
                                                     rgiidExclude,
                                                     pExcludeProps,
                                                     ulUIParam,
                                                     pProgress,
                                                     pDestInterface,
                                                     pDestObj,
                                                     ulFlags,
                                                     ppProblems);
    // Set the error in the cache and trace it out
    TraceResult ("CAttach::CopyTo", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::CopyProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy the properties from one object to the other. For the copy
//      operation, we defer to the MAPI support object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::CopyProps (LPSPropTagArray       pIncludeProps,
                                 ULONG                 ulUIParam,
                                 LPMAPIPROGRESS        pProgress,
                                 LPCIID                pDestInterface,
                                 LPVOID                pDestObj,
                                 ULONG                 ulFlags,
                                 LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CAttach::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pDestInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyProps (&IID_IAttachment,
                                                        this,
                                                        pIncludeProps,
                                                        ulUIParam,
                                                        pProgress,
                                                        pDestInterface,
                                                        pDestObj,
                                                        ulFlags,
                                                        ppProblems);

    // Set the error in the cache and trace it out
    TraceResult ("CAttach::CopyProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::GetNamesFromIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the list of names for the identifier list supplied by the
//      caller. Our implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                       LPGUID            pPropSetGuid,
                                       ULONG             ulFlags,
                                       ULONG *           pcPropNames,
                                       LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CAttach::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    LPSPropTagArray pTags;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptAttachObj, &pTags);
    if (!hResult)
    {
        hResult = CGenericProp::GetNamesFromIDs (ppPropTags,
                                                 pPropSetGuid,
                                                 ulFlags,
                                                 pcPropNames,
                                                 pppPropNames,
                                                 pTags);
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CAttach::GetNamesFromIDs", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CAttach::GetIDsFromNames()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return a property identifier for each named listed by the client. Our
//      IAttach implementation, defers to the common IMAPIProp implementation
//      to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CAttach::GetIDsFromNames (ULONG             cPropNames,
                                       LPMAPINAMEID *    ppPropNames,
                                       ULONG             ulFlags,
                                       LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CAttach::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    LPSPropTagArray pTags;
    LPSTREAM pStream;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptAttachObj, &pTags);
    if (!hResult)
    {
        hResult = m_pBEobj->GetNamedPropertiesStream (&pStream);
        if (!hResult)
        {
            hResult = CGenericProp::GetIDsFromNames (cPropNames,
                                                     ppPropNames,
                                                     ulFlags,
                                                     ppPropTags,
                                                     pTags,
                                                     pStream);
            pStream->Release();
        }
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CAttach::GetIDsFromNames", hResult);
    SetLastError (hResult);
    return hResult;
}

// End of file for MSATTACH.CPP
