///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSFOLDER.CPP
//
//  Description
//      This file implements the IMAPIFolder interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSWDS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

#define NOTIFICATION_BATCH_SIZE  16

#define NUM_FOLDER_PROPS        25
// Properties available in our implementation of IMAPIFolder
SizedSPropTagArray(NUM_FOLDER_PROPS, sptFolderObj) =
{
    NUM_FOLDER_PROPS,
    {
        PR_ACCESS,                      // READ only
        PR_ACCESS_LEVEL,                // READ only
        PR_ENTRYID,                     // READ only
        PR_PARENT_ENTRYID,              // READ only
        PR_STORE_ENTRYID,               // READ only
        PR_STORE_RECORD_KEY,            // READ only
        PR_RECORD_KEY,                  // READ only
        PR_MAPPING_SIGNATURE,           // READ only
        PR_CONTENT_COUNT,               // READ only
        PR_CONTENT_UNREAD,              // READ only
        PR_ASSOC_CONTENT_COUNT,         // READ only
        PR_CONTAINER_HIERARCHY,         // READ only
        PR_CONTAINER_CONTENTS,          // READ only
        PR_FOLDER_ASSOCIATED_CONTENTS,  // READ only
        PR_FOLDER_TYPE,                 // READ only
        PR_OBJECT_TYPE,                 // READ only
        PR_STORE_SUPPORT_MASK,          // READ only
        PR_SUBFOLDERS,                  // READ only
        PR_CREATION_TIME,               // READ only
        PR_LAST_MODIFICATION_TIME,      // READ only
        PR_CREATION_VERSION,            // READ only
        PR_DISPLAY_TYPE,                // READ only
        PR_DISPLAY_NAME,                // READ/WRITE
        PR_COMMENT,                     // READ/WRITE
        PR_STATUS                       // READ/WRITE
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CMAPIFolder()
//
//    Parameters
//      [IN]    pStore          Pointer to the container IMsgStore object
//      [IN]    pParentEID      Entry ID of the parent folder
//      [IN]    pFolderEID      Entry ID of this folder
//      [IN]    lFolderType     Type of folder this is (generic, search, etc.)
//      [IN]    pList           In-memory property cache for this object
//      [IN]    fDeferErrors    Weather or not we should defer errors
//                              in this instance
//      [IN]    lAccess         Access level of this object
//      [IN]    pNPList         Pointer to the named properties manager
//      
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CMAPIFolder::CMAPIFolder (CMsgStore *       pStore,
                          PRIVATE_ENTRYID * pParentEID,
                          PRIVATE_ENTRYID * pFolderEID,
                          long              lFolderType,
                          CPropList *       pList,
                          BOOL              fDeferErrors,
                          long              lAccess,
                          CNamePropMgr *    pNPList) : CGenericProp (pStore->m_pBEobj,
                                                                     pParentEID,
                                                                     pFolderEID,
                                                                     pList,
                                                                     fDeferErrors,
                                                                     lAccess)
{
    InfoTrace ("CMAPIFolder: Constructor called");
    m_pNPList               = pNPList;
    m_pHierTable            = NULL;
    m_pContTable            = NULL;
    m_pAsscContTable        = NULL;
    m_dwHierTableViews      = 0;
    m_dwContTableViews      = 0;
    m_dwAsscContTableViews  = 0;
    m_lFolderType           = lFolderType;
    m_pStore                = pStore;
    m_pStore->AddRef();
    m_pStore->m_OpenObjList.Insert (this, &m_ObjectEID);
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::~CMAPIFolder()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CMAPIFolder.
//
//    Return Value
//      None
//
CMAPIFolder::~CMAPIFolder()
{
    InfoTrace ("CMAPIFolder: Destructor called");
    // Release the table objects if we have them around
    if (m_pHierTable)
    {
        m_pHierTable->Release();
        m_pHierTable = NULL;
    }
    if (m_pContTable)
    {
        m_pContTable->Release();
        m_pContTable = NULL;
    }
    if (m_pAsscContTable)
    {
        m_pAsscContTable->Release();
        m_pAsscContTable = NULL;
    }
    #ifdef _DEBUG
    if (m_dwHierTableViews)
    {
        TraceString1 ("CMAPIFolder: Object going away with %d views left on the hierarchy table", m_dwHierTableViews);
    }
    if (m_dwContTableViews)
    {
        TraceString1 ("CMAPIFolder: Object going away with %d views left on the contents table", m_dwContTableViews);
    }
    if (m_dwAsscContTableViews)
    {
        TraceString1 ("CMAPIFolder: Object going away with %d views left on the contents table", m_dwAsscContTableViews);
    }
    #endif // _DEBUG
    m_pStore->m_OpenObjList.Delete (this, &m_ObjectEID);
    // If the object has been deleted in the parent folder, I must delete the file storage for this message....
    if (OBJECT_DELETED_IN_CONTAINER & m_ulObjFlags)
    {
        // ... however, the message storage should only go away if this is the LAST object open. I search the
        // list again trying to find another folder with the same entry ID. If found, I don't delete it now,
        // Let that task for the last object instance opened by a client.
        CMAPIFolder * pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
        if (pFldObj)
        {
            pFldObj->Release();
        }
        else
        {
            m_pBEobj->DeleteOneObject (m_ObjectEID, m_ParentEID);
        }
    }
    m_pStore->Release();
    m_pStore = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;    
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IMAPIFolder    ||
        riid == IID_IMAPIContainer ||
        riid == IID_IMAPIProp      ||
        riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Returns a MAPIERROR structure with an error description string about
//      the last error that occurred in the object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetLastError (HRESULT       hError,
                                        ULONG         ulFlags,
                                        LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CMAPIFolder::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this,
                                     hError,
                                     ulFlags,
                                     ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_pStore->m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SaveChanges()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Place holder. Folder objects are not transacted so a call to
//      SaveChanges is a no-op method that does not do anything.
//      For consistency with other implementation we check the access level
//      and return an error to the client is the object is READ-only.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CMAPIFolder::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);
    HRESULT hResult = S_OK;
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        hResult = E_ACCESSDENIED;
    }
    TraceResult ("CMAPIFolder::SaveChanges", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the value of the properties the client especified in the
//      property tag array. If the tag array is NULL, the user wants all
//      the properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetProps (LPSPropTagArray pPropTagArray,
                                    ULONG           ulFlags,
                                    ULONG *         pcValues,
                                    LPSPropValue *  ppPropArray)
{
    InfoTrace ("CMAPIFolder::GetProps method called");
    Validate_IMAPIProp_GetProps (this,
                                 pPropTagArray,
                                 ulFlags,
                                 pcValues,
                                 ppPropArray);

    #ifdef DO_INFO_TRACES
    if (!pPropTagArray)
    {
        TraceMessage ("CMAPIFolder::GetProps: All properties requested");
    }
    else
    {
        TraceMessage ("CMAPIFolder::GetProps Properties requested...");
        for (ULONG j=0; j<pPropTagArray->cValues; j++)
        { TraceProp (pPropTagArray->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES

    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);
    BOOL fGotErrors = FALSE;
    // Call the base calls for the preliminary work on the property tag array.
    // This function will allocate the returned property value array and
    // set their tags.
    HRESULT hResult = CGenericProp::GetProps (pPropTagArray,
                                              ulFlags,
                                              pcValues,
                                              ppPropArray,
                                              (LPSPropTagArray)&sptFolderObj);
    if (hResult)
    {
        goto ErrorExit;
    }

    LPVOID pSrc;
    ULONG cbSize, i;
    // Get the values for all the properties requested
    for (i=0; i<*pcValues; i++)
    {
        pSrc = NULL;
        hResult = S_OK;
        // At this point, the property value array has the full property tag
        // of the properties set and we are ready to copy the data into
        // the array slots.
        switch ((*ppPropArray)[i].ulPropTag)
        {
            // Get the PT_OBJECT/PR_NULL properties
            case PR_NULL :
            case PR_FOLDER_ASSOCIATED_CONTENTS :
            case PR_CONTAINER_CONTENTS :
            case PR_CONTAINER_HIERARCHY :
                (*ppPropArray)[i].Value.err = S_OK;
                break;
            // Get the PT_LONG properties
            case PR_ACCESS :
                (*ppPropArray)[i].Value.l = m_lAccess;
                break;
            case PR_ACCESS_LEVEL :
                (*ppPropArray)[i].Value.l = (m_lAccess & MAPI_ACCESS_MODIFY ? MAPI_MODIFY : 0);
                break;
            case PR_FOLDER_TYPE :
                (*ppPropArray)[i].Value.l = m_lFolderType;
                break;
            case PR_OBJECT_TYPE :
                (*ppPropArray)[i].Value.l = MAPI_FOLDER;
                break;
            case PR_DISPLAY_TYPE :
                (*ppPropArray)[i].Value.l = DT_FOLDER;
                break;
            case PR_STORE_SUPPORT_MASK:
                (*ppPropArray)[i].Value.l = SUPPORT_MASK;
                break;
            // Get the PT_I8 properties
            case PR_CREATION_VERSION :
                (*ppPropArray)[i].Value.li.LowPart =  MSG_STORE_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = MSG_STORE_VERSION_MINOR;
                break;
            // Get the PT_BINARY properties
            case PR_STORE_RECORD_KEY :
            case PR_MAPPING_SIGNATURE :
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&m_pStore->m_StoreUID;
                break;
            case PR_PARENT_ENTRYID :
                cbSize = CB_PRIVATE_EID;
                pSrc = (LPVOID)&m_ParentEID;
                break;
            case PR_STORE_ENTRYID :
                cbSize = m_pStore->m_StoreEID.cb;
                pSrc = (LPVOID)m_pStore->m_StoreEID.lpb;
                break;
            case PR_RECORD_KEY :
            case PR_ENTRYID :
                cbSize = CB_PRIVATE_EID;
                pSrc = (LPVOID)&m_ObjectEID;
                break;
            default :
                // Oh no, the property was not found. Try the property cache.
                hResult = CGenericProp::GetProps (&(*ppPropArray)[i], *ppPropArray);
                #ifdef DO_INFO_TRACES
                if (MAPI_E_NOT_FOUND == hResult)
                {
                    TraceRaw ("CMAPIFolder::GetProps: Not Found - ");
                    TraceProp ((*ppPropArray)[i].ulPropTag);
                }
                #endif // #DO_INFO_TRACES
                break;
        }
        // Now we either found the property in our cache or not. The function
        // below takes care of checking hResult, and if an error occurred,
        // change the property tag in the slot to be of the type PT_ERROR.
        if (S_OK != CGenericProp::GetProps (hResult,
                                            cbSize,
                                            pSrc,
                                            &(*ppPropArray)[i],
                                            *ppPropArray))
        {
            fGotErrors = TRUE;
        }
    }
ErrorExit :
    LeaveCriticalSection (&m_csObj);
    if (fGotErrors)
	{
		hResult = MAPI_W_ERRORS_RETURNED;
    }
    if (hResult && hResult != MAPI_W_ERRORS_RETURNED)
    {
        TraceResult ("CMAPIFolder::GetProps", hResult);
    }
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetPropList()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return an array of property tags with all the properties
//      in this object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetPropList (ULONG             ulFlags,
                                       LPSPropTagArray * ppAllTags)
{
    InfoTrace ("CMAPIFolder::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppAllTags);
    return CGenericProp::GetPropList (ulFlags, (LPSPropTagArray)&sptFolderObj, ppAllTags);
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::OpenProperty()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. Our implementation does not support opening interfaces
//      on properties in a folder objects.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CMAPIFolder::OpenProperty (ULONG       ulPropTag,
                                        LPCIID      piid,
                                        ULONG       ulInterfaceOptions,
                                        ULONG       ulFlags,
                                        LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMAPIFolder::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this,
                                     ulPropTag,
                                     piid,
                                     ulInterfaceOptions,
                                     ulFlags,
                                     ppUnk);
    TraceResult ("CMAPIFolder::OpenProperty", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Modify the value of the properties in our object. In this
//      implementation of IMAPIFolder, a user cannot add arbitrary
//      properties to the object, only modify some of them if the proper
//      access has been requested when opening it.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::SetProps (ULONG                 cValues,
                                    LPSPropValue          pPropArray,
                                    LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMAPIFolder::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::SetProps: Cannot set properties in a READ only object");
        TraceResult ("CMAPIFolder::SetProps", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    #ifdef DO_INFO_TRACES
    TraceMessage ("CMAPIFolder::SetProps: Properties to set...");
    for (ULONG j=0; j<cValues; j++) { TraceProp (pPropArray[j].ulPropTag); }
    #endif DO_INFO_TRACES

    BOOL fObjectModified = FALSE;
    HRESULT hTagError, hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    for (ULONG i=0; i<cValues; i++)
    {
        switch (PROP_TYPE(pPropArray[i].ulPropTag))
        {
            case PT_NULL :
            case PT_UNSPECIFIED :
            case PT_ERROR :
            case PT_OBJECT :
                TraceMessage ("CMAPIFolder::SetProps: Ignoring property of PT_NULL, PT_UNSPEFICIED, PT_ERROR or PT_OBJECT type");
                continue;
        }
        hTagError = S_OK;
        switch (pPropArray[i].ulPropTag)
        {
            // If the user is trying to set the name of the folder, we must check
            // with the database to make sure the there is no folder in the same
            // hierarchy level that has that name.
            case PR_DISPLAY_NAME :
                // If the name already exists, the return value is MAPI_E_COLLISION, otherwise
                // it is S_OK if the new name is not a duplicate of another folder's name
                hTagError = m_pBEobj->SetObjNameInContainer (pPropArray[i].Value.lpszA,
                                                             m_ParentEID,
                                                             m_ObjectEID);
                break;
            // The only properties that we allow to be set on our IMAPIFolder objects are:
            //      PR_DISPLAY_NAME
            //      PR_COMMENT
            //      PR_STATUS
            // Anything else we deny access to
            case PR_COMMENT :
            case PR_STATUS :
                break;
            default :
                hTagError = GetPropAccess  (pPropArray[i].ulPropTag);
                if (!hTagError)
                {
                    hTagError = E_ACCESSDENIED;
                }
                break;
        }
        // Add the property to the array of properties
        if (!hTagError)
        {
            hTagError = AddPropToList (&pPropArray[i]);
            if (!hTagError)
            {
                fObjectModified = TRUE;
            }
        }
        // If there was a problem setting the properties in the object,
        // return this information if the caller requested it.
        if (hTagError && ppProblems)
        {
            hResult = AddTagToProbArray (cValues,
                                         ppProblems,
                                         i,
                                         pPropArray[i].ulPropTag,
                                         hTagError);
            if (hResult)
            {
                break; // Out of the FOR() loop
            }
        }
    }
    if (!hResult && fObjectModified)
    {
        // We need to commit the changes to the folder's properties before
        // sending the notification or update the parent's hierarchy table.
        CGenericProp::SaveChanges (KEEP_OPEN_READWRITE, NULL);

        // Replicate the changes made to the property cache, to other
        // instance of this same folder that may be open
        UpdatePropCacheInReplicas();

        NOTIFICATION notif = { 0 };
        notif.ulEventType           = fnevObjectModified;
        notif.info.obj.cbEntryID    = CB_PRIVATE_EID;
        notif.info.obj.lpEntryID    = (LPENTRYID)&m_ObjectEID;
        notif.info.obj.cbParentID   = CB_PRIVATE_EID;
        notif.info.obj.lpParentID   = (LPENTRYID)&m_ParentEID;
        notif.info.obj.ulObjType    = MAPI_FOLDER;
        m_pStore->NotifyClients (&m_ObjectEID, 1, &notif);

        // Update the any in-memory hierarchy table for the parents of this folder.
        UpdateParentTable();
    }
    LeaveCriticalSection (&m_csObj);
    // If an error occurred while adding errors to the problem array, the user
    // won't attempt to delete it so we should.
    if (hResult)
    {
        if (ppProblems)
        {
            PrivMAPIFree ((LPVOID *)ppProblems);
        }
        TraceResult ("CMAPIFolder::SetProps", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::DeleteProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Our implementation of IMAPIFolder does not allow deleting properties,
//      so any property that a client tries to delete, will result with
//      E_ACCESSDENIED (MAPI_E_NO_ACCESS) in the problem array returned.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::DeleteProps (LPSPropTagArray       pPropTagArray,
                                       LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMAPIFolder::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::SetProps: Cannot set properties in a READ only object");
        TraceResult ("CMAPIFolder::DeleteProps", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    HRESULT hResult = S_OK;
    // Note that we don't lock the object because we don't modify any internal data member.
    // If the caller gave us a problem array, fill it up with E_ACCESSDENIED.
    if (ppProblems)
    {
        *ppProblems = NULL;
        for (ULONG i=0; i<pPropTagArray->cValues; i++)
        {
            hResult = AddTagToProbArray (pPropTagArray->cValues,
                                         ppProblems,
                                         i,
                                         pPropTagArray->aulPropTag[i],
                                         E_ACCESSDENIED);
            if (hResult)
            {
                break; // Out of the FOR() loop
            }
        }
    }
    // If an error occurred while adding errors to the problem array, the user
    // won't attempt to delete it so we should.
    if (hResult)
    {
        if (ppProblems)
        {
            PrivMAPIFree ((LPVOID *)ppProblems);
        }
        TraceResult ("CMAPIFolder::DeleteProps", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CopyTo()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy this object's content into another one. For the copy operation
//      we delegate to MAPI for the implemenation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CopyTo (ULONG                 ciidExclude,
                                  LPCIID                rgiidExclude,
                                  LPSPropTagArray       pExcludeProps,
                                  ULONG                 ulUIParam,
                                  LPMAPIPROGRESS        pProgress,
                                  LPCIID                pInterface,
                                  LPVOID                pDestObj,
                                  ULONG                 ulFlags,
                                  LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMAPIFolder::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,   
                               rgiidExclude,  
                               pExcludeProps, 
                               ulUIParam,     
                               pProgress,     
                               pDestInterface,
                               pDestObj,      
                               ulFlags,       
                               ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyTo (&IID_IMAPIFolder,
                                                     this,
                                                     ciidExclude,
                                                     rgiidExclude,
                                                     pExcludeProps,
                                                     ulUIParam,
                                                     pProgress,
                                                     pInterface,
                                                     pDestObj,
                                                     ulFlags,
                                                     ppProblems);
    // Set the error in the cache and trace it out
    TraceResult ("CMAPIFolder::CopyTo", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CopyProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Copy the properties from one object to the other. For the copy
//      operation, we defer to the MAPI support object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CopyProps (LPSPropTagArray       pIncludeProps,
                                     ULONG                 ulUIParam,
                                     LPMAPIPROGRESS        pProgress,
                                     LPCIID                pInterface,
                                     LPVOID                pDestObj,
                                     ULONG                 ulFlags,
                                     LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMAPIFolder::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pDestInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    // Call MAPI to do the copy
    HRESULT hResult = m_pStore->m_pSupObj->DoCopyProps (&IID_IMAPIFolder,
                                                        this,
                                                        pIncludeProps,
                                                        ulUIParam,
                                                        pProgress,
                                                        pInterface,
                                                        pDestObj,
                                                        ulFlags,
                                                        ppProblems);

    // Set the error in the cache and trace it out
    TraceResult ("CMAPIFolder::CopyProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetNamesFromIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the list of names for the identifier list supplied by the
//      caller. Our implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                           LPGUID            pPropSetGuid,
                                           ULONG             ulFlags,
                                           ULONG *           pcPropNames,
                                           LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CMAPIFolder::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    LPSPropTagArray pTags;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptFolderObj, &pTags);
    if (!hResult)
    {
        hResult = CGenericProp::GetNamesFromIDs (ppPropTags,
                                                 pPropSetGuid,
                                                 ulFlags,
                                                 pcPropNames,
                                                 pppPropNames,
                                                 pTags);
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMAPIFolder::GetNamesFromIDs", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetIDsFromNames()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return a property identifier for each named listed by the client.
//      Our IMAPIFolder implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetIDsFromNames (ULONG             cPropNames,
                                           LPMAPINAMEID *    ppPropNames,
                                           ULONG             ulFlags,
                                           LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CMAPIFolder::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    LPSPropTagArray pTags;
    LPSTREAM pStream;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptFolderObj, &pTags);
    if (!hResult)
    {
        hResult = m_pBEobj->GetNamedPropertiesStream (&pStream);
        if (!hResult)
        {
            hResult = CGenericProp::GetIDsFromNames (cPropNames,
                                                     ppPropNames,
                                                     ulFlags,
                                                     ppPropTags,
                                                     pTags,
                                                     pStream);
            pStream->Release();
        }
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMAPIFolder::GetIDsFromNames", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIContainer virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetContentsTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Returns a READ-only table for the contained messages in this folder.
//      Depending on the flags in ulFlags, we return the table of associated
//      messages, or the table of regular messages.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetContentsTable (ULONG         ulFlags,
                                            LPMAPITABLE * ppTable)
{
    InfoTrace ("CMAPIFolder::GetContentsTable method called");
    Validate_IMAPIContainer_GetContentsTable (this, ulFlags, ppTable);
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    LPTABLEDATA pInMemoryTable;
    DWORD * pdwViews;
    if (ulFlags & MAPI_ASSOCIATED)
    {
        InfoTrace ("CMAPIFolder::GetContentsTable: An assocated contents table was requested");
        // If we don't have a copy of the table data in memory already, load it
        // up from the the file.
        if (!m_pAsscContTable)
        {
            hResult = m_pBEobj->GetTable (TABLE_CONTENTS,
                                          m_ObjectEID,
                                          m_lAccess,
                                          (LPVOID)&m_pStore->m_StoreEID,
                                          (LPVOID)TRUE,
                                          &m_pAsscContTable);
        }
        pInMemoryTable = m_pAsscContTable;
        pdwViews = &m_dwAsscContTableViews;
    }
    else
    {
        InfoTrace ("CMAPIFolder::GetContentsTable: A regular contents table was requested");
        // If we don't have a copy of the table data in memory already, load it
        // up from the the file.
        if (!m_pContTable)
        {
            hResult = m_pBEobj->GetTable (TABLE_CONTENTS,
                                          m_ObjectEID,
                                          m_lAccess,
                                          (LPVOID)&m_pStore->m_StoreEID,
                                          (LPVOID)FALSE,
                                          &m_pContTable);
        }
        pInMemoryTable = m_pContTable;
        pdwViews = &m_dwContTableViews;
    }
    if (!hResult)
    {
        ASSERT (pInMemoryTable);
        hResult = pInMemoryTable->HrGetView (NULL, FolderTablesViewRelease, (ULONG)this, ppTable);
        TraceResult ("CMAPIFolder::GetContentsTable: Failed to get a view of the memory table", hResult);
        if (!hResult)
        {
            // For each view on the table data, that is going to be returned to a client, we
            // must AddRef the entire object so that the opened views on this data, upon
            // release, can access the object. The matching release call is in the callback
            // function especified in the HrGetView() call
            AddRef();
            (*pdwViews)++;
        }
    }
    LeaveCriticalSection (&m_csObj);
    // Save the last error in the object error cache
    TraceResult ("CMAPIFolder::GetContentsTable", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetHierarchyTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Returns a READ-only table for the contained folders in this folder.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetHierarchyTable (ULONG         ulFlags,
                                             LPMAPITABLE * ppTable)
{
    InfoTrace ("CMAPIFolder::GetHierarchyTable method called");
    Validate_IMAPIContainer_GetHierarchyTable (this, ulFlags, ppTable);
    HRESULT hResult = S_OK;
    EnterCriticalSection (&m_csObj);
    if (!m_pHierTable)
    {
        // If we don't have a copy of the table data in memory already, load it
        // up from the the file.
        hResult = m_pBEobj->GetTable (TABLE_HIERARCHY,
                                      m_ObjectEID,
                                      m_lAccess,
                                      (LPVOID)&m_pStore->m_StoreEID,
                                      NULL,
                                      &m_pHierTable);
    }
    if (!hResult)
    {
        hResult = m_pHierTable->HrGetView (NULL, FolderTablesViewRelease, (ULONG)this, ppTable);
        TraceResult ("CMAPIFolder::GetHierarchyTable: Failed to get a view of the memory table", hResult);
        if (!hResult)
        {
            // For each view on the table data, that is going to be returned to a client, we
            // must AddRef the entire object so that the opened views on this data, upon
            // release, can access the object. The matching release call is in the callback
            // function especified in the HrGetView() call
            AddRef();
            m_dwHierTableViews++;
        }
    }
    LeaveCriticalSection (&m_csObj);
    // Save the last error in the object error cache
    TraceResult ("CMAPIFolder::GetHierarchyTable", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::OpenEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method opens an object that exists in this folder. The object
//      could be a direct subfolder of this object or a direct child message
//      object contained in this folder. We don't support opening objects
//      that are not direct descendants of the parent container.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::OpenEntry (ULONG       cbEntryID,
                                     LPENTRYID   pEntryID,
                                     LPCIID      pInterface,
                                     ULONG       ulFlags,
                                     ULONG *     pulObjType,
                                     LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMAPIFolder::OpenEntry method called");
    Validate_IMAPIContainer_OpenEntry (this,
                                       cbEntryID,
                                       pEntryID,
                                       pInterface,
                                       ulFlags,
                                       pulObjType,
                                       ppUnk);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & MAPI_MODIFY))
    {
        TraceMessage ("CMAPIFolder::OpenEntry: Requested WRITE access in a READ only object");
        TraceResult ("CMAPIFolder::OpenEntry", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    // If the user requested MAPI_BEST_ACCESS, a provider should give the client the
    // maximum access level it can. In our object, if we are write-able, give the
    // user WRITE access. Is the easiest to handle for us.
    if (ulFlags & MAPI_BEST_ACCESS)
    {
        if (m_lAccess & MAPI_ACCESS_MODIFY)
        {
            ulFlags |= MAPI_MODIFY;
        }
    }

    CMessage * pMsgObj;
    CMAPIFolder * pFldObj;
    CPropList * pList = NULL;
    long lAccess;
    HRESULT hResult = S_OK;
    PRIVATE_ENTRYID * pEID = (PRIVATE_ENTRYID *)pEntryID;
    if (cbEntryID)
    {
        if (!m_pBEobj->IsValidEntryID (pEID))
        {
            return MAPI_E_INVALID_ENTRYID;
        }
    }

    EnterCriticalSection (&m_csObj);
    if (pEID->bObject == MAPI_FOLDER || pEID->bObject == MAPI_MESSAGE)
    {
        hResult = m_pBEobj->OpenObject (*pEID, &pList);
    }
    else
    {   
        hResult = MAPI_E_NOT_FOUND;
    }
    if ((MAPI_FOLDER == pEID->bObject) && !hResult)
    {
        lAccess = (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                  (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                  MAPI_ACCESS_READ |
                  MAPI_ACCESS_CREATE_HIERARCHY |
                  MAPI_ACCESS_CREATE_CONTENTS |
                  MAPI_ACCESS_CREATE_ASSOCIATED;
        pFldObj = new CMAPIFolder (m_pStore,
                                   &m_ObjectEID,
                                   pEID,
                                   FOLDER_GENERIC,
                                   pList,
                                   !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                   lAccess,
                                   m_pNPList);
        if (pFldObj)
        {
            *pulObjType = MAPI_FOLDER;
            *ppUnk = (LPUNKNOWN)pFldObj;
            pFldObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    if ((MAPI_MESSAGE == pEID->bObject) && !hResult)
    {
        lAccess = (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                  (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                  MAPI_ACCESS_READ;
        pMsgObj = new CMessage (m_pStore,
                                &m_ObjectEID,
                                pEID,
                                pList,
                                !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                lAccess,
                                NULL,
                                m_pNPList);
        if (pMsgObj)
        {
            *pulObjType = MAPI_MESSAGE;
            *ppUnk = (LPUNKNOWN)pMsgObj;
            pMsgObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
            pMsgObj->LoadObjectTables();
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    LeaveCriticalSection (&m_csObj);
    if (hResult)
    {
        if (pList)
        {
            delete pList;
        }
        TraceResult ("CMAPIFolder::OpenEntry", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SetSearchCriteria()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't implement search folders in this release.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMAPIFolder::SetSearchCriteria (LPSRestriction pRestriction,
                                             LPENTRYLIST    pContainerList,
                                             ULONG          ulSearchFlags)
{
    InfoTrace ("CMAPIFolder::SetSearchCriteria method called");
    Validate_IMAPIContainer_SetSearchCriteria (this,
                                               pRestriction,
                                               pContainerList,
                                               ulSearchFlags);
    TraceResult ("CMAPIFolder::SetSearchCriteria", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetSearchCriteria()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't implement search folders in this release.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMAPIFolder::GetSearchCriteria (ULONG            ulFlags,
                                             LPSRestriction * ppRestriction,
                                             LPENTRYLIST *    ppContainerList,
                                             ULONG *          pulSearchState)
{
    InfoTrace ("CMAPIFolder::GetSearchCriteria method called");
    Validate_IMAPIContainer_GetSearchCriteria (this,
                                               ulFlags,
                                               ppRestriction,
                                               ppContainerList,
                                               pulSearchFlags);
    TraceResult ("CMAPIFolder::GetSearchCriteria", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIFolder virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CreateMessage()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Here we create a message object as a child object of the parent
//      container. Depending on the flags in ulFlags, we create the message
//      ASSOCIATED to the folder or not. The actual creation of the underlying
//      data storage is delegated to the underlying database implementation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CreateMessage (LPCIID      pInterface,
                                         ULONG       ulFlags,
                                         LPMESSAGE * ppMessage)
{
    InfoTrace ("CMAPIFolder::CreateMessage method called");
    Validate_IMAPIFolder_CreateMessage (this,
                                        pInterface,
                                        ulFlags,
                                        ppMessage);    
    PRIVATE_ENTRYID eidNewMessage = { 0 };
    long lAccess;
    CPropList * pList;
    CMessage * pMsgObj;
    MAPIUID uidSearchKey;
    HRESULT hResult = m_pStore->m_pSupObj->NewUID (&uidSearchKey);
    if (!hResult)
    {
        hResult = m_pBEobj->CreateObject (m_ObjectEID,
                                          &eidNewMessage,
                                          &pList,
                                          MAPI_MESSAGE,
                                          (LPVOID)&uidSearchKey,
                                          (LPVOID)!!(ulFlags & MAPI_ASSOCIATED));
        if (!hResult)
        {
            if (OBJECT_RUNNING_IN_SPOOLER & m_ulObjFlags)
            {
                SPropValue spvMsgFlags = { 0 };
                if (S_OK == pList->Find (PR_MESSAGE_FLAGS, &spvMsgFlags))
                {
                    spvMsgFlags.Value.l &= ~(MSGFLAG_READ | MSGFLAG_UNSENT);
                    pList->Insert (&spvMsgFlags);
                }
            }
            // Inherit the parent's access rights
            lAccess = MAPI_ACCESS_READ |
                      MAPI_ACCESS_MODIFY |
                      MAPI_ACCESS_DELETE;
            pMsgObj = new CMessage (m_pStore,
                                    &m_ObjectEID,
                                    &eidNewMessage,
                                    pList,
                                    !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                    lAccess,
                                    NULL,
                                    m_pNPList);
            if (pMsgObj)
            {
                *ppMessage = (LPMESSAGE)pMsgObj;
                pMsgObj->m_ulObjFlags |= OBJECT_JUST_CREATED | OBJECT_NEVER_SAVED |
                                         ((ulFlags & MAPI_ASSOCIATED) ? OBJECT_ASSOCIATED : 0) |
                                         ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
                pMsgObj->LoadObjectTables();
            }
            else
            {
                hResult = E_OUTOFMEMORY;
                delete pList;
            }
        }
    }
    TraceResult ("CMAPIFolder::CreateMessage", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CopyMessages()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method copies a lists of messages in a folder to the destination
//      folder specified by the caller. In our implementation we delegate to
//      the IMAPISupport object for the copy of the messages.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CopyMessages (LPENTRYLIST    pMsgList,
                                        LPCIID         pDestInterface,
                                        LPVOID         pDestFolder,
                                        ULONG          ulUIParam,
                                        LPMAPIPROGRESS pProgress,
                                        ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::CopyMessages method called");
    Validate_IMAPIFolder_CopyMessages (this,
                                       pMsgList,
                                       pDestInterface,
                                       pDestFolder,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    // Check the access level of this folder
    HRESULT hResult;
    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & MESSAGE_MOVE))
    {
        TraceMessage ("CMAPIFolder::CopyMessages: Cannot move messages from a READ only folder");
        hResult = E_ACCESSDENIED;
    }
    else
    {
        // If it is not an object that we now off, then let MAPI do the copying...
        hResult = m_pStore->m_pSupObj->CopyMessages (&IID_IMAPIFolder,
                                                     this,
                                                     pMsgList,
                                                     pDestInterface,
                                                     pDestFolder,
                                                     ulUIParam,
                                                     pProgress,
                                                     ulFlags);
    }
    TraceResult ("CMAPIFolder::CopyMessages", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::DeleteMessages()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Given a list of message entry IDs, delete them in this folder. If the
//      user specified so, the message store will put up or update the
//      indicator telling how much is has progressed.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::DeleteMessages (LPENTRYLIST    pMsgList,
                                          ULONG          ulUIParam,
                                          LPMAPIPROGRESS pProgress,
                                          ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::DeleteMessages method called");
    Validate_IMAPIFolder_DeleteMessages (this,
                                         pMsgList,
                                         ulUIParam,
                                         pProgress,
                                         ulFlags);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::DeleteMessages: Cannot delete messages in a READ only folder");
        TraceResult ("CMAPIFolder::DeleteMessages", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    // Initialize the members of the custom structure which is passed to 
    // the backend. The back end uses the information in this structure to
    // delete the folders in the storage and update the container object
    // that may be open.
    OPENOBJECTS_DATA Data = { 0 };
    Data.pSrcEIDList = pMsgList;
    Data.pOpenObjList = &(m_pStore->m_OpenObjList);
    Data.dwFlags = OPENOBJDATA_USE_MSGEIDLIST;
    
    // Initialize the array of notifications and eid which are used to send
    // fnevObjectDeleted notifications to subscribers of this store.
    HRESULT hResult = S_OK;
    NOTIFICATION notifs[NOTIFICATION_BATCH_SIZE] = { 0 };
    PRIVATE_ENTRYID EIDList[NOTIFICATION_BATCH_SIZE] = { 0 };
    for (ULONG j=0; j<NOTIFICATION_BATCH_SIZE; j++)
    {
        EIDList[j].uidGlobal = m_pStore->m_StoreUID;
        EIDList[j].bVersion = ENTRYID_VERSION;
    }

    DWORD dwBatch = 0;
    BOOL fReleaseProgress = FALSE, fObjectChanged = FALSE;
    // If the client wants UI, but did not bother to pass one in, call MAPI and
    // create a default progress indicator.
    if (ulFlags & MESSAGE_DIALOG)
    {
        if (NULL == pProgress)
        {
            hResult = m_pStore->m_pSupObj->DoProgressDialog (ulUIParam, MAPI_TOP_LEVEL, &pProgress);
            if (hResult)
            {
                TraceResult ("CMAPIFolder::DeleteMessages: Failed to get default progress interface", hResult);
                pProgress = NULL;
            }
            else
            {
                fReleaseProgress = TRUE;
            }
        }
    }
    // Lock the folder object for exclusive access
    EnterCriticalSection (&m_csObj);
    // Call the back end to do the actual delete of the messages
    hResult = m_pBEobj->DeleteObjects (m_ObjectEID, CONTAINER_OBJECTS, pProgress, &Data);
    if (SUCCEEDED(hResult)) // We might get partial success
    {
        // Walk through the list of object ID returned by the back end
        for (ULONG i=0; i<Data.dwCount; i++)
        {
            // If the database could not delete this object, the object ID would be set to 0,
            // so on anything that is non-Zero, we assume got deleted and a notification is in order.
            if (Data.pSubEIDList[i].dwObjID)
            {
                // If at least one message was deleted, then the count properties of this
                // folder have changed
                fObjectChanged = TRUE;

                EIDList[dwBatch].bObject = Data.pSubEIDList[i].bObject;
                EIDList[dwBatch].dwObjID = Data.pSubEIDList[i].dwObjID;
                
                notifs[dwBatch].ulEventType         = fnevObjectDeleted;
                notifs[dwBatch].info.obj.cbEntryID  = CB_PRIVATE_EID;
                notifs[dwBatch].info.obj.lpEntryID  = (LPENTRYID)(&EIDList[dwBatch]);
                notifs[dwBatch].info.obj.cbParentID = CB_PRIVATE_EID;
                notifs[dwBatch].info.obj.lpParentID = (LPENTRYID)&m_ObjectEID;
                notifs[dwBatch].info.obj.ulObjType  = MAPI_MESSAGE;
                dwBatch++;
                // This implementation of a store provider batches NOTIFICATION_BATCH_SIZE notifications
                // of deleted objects before sending them out.
                if (NOTIFICATION_BATCH_SIZE == dwBatch)
                {
                    // Send the notification to subscribers of the STORE object only
                    m_pStore->NotifyClients (NULL, dwBatch, notifs);
                    dwBatch = 0;
                }
            }
        }
        // If any nofications where not sent because
        // there were fewer than the batch limit, send them now.
        if (dwBatch)
        {
            // Send the notification to subscribers of the STORE object only
            m_pStore->NotifyClients (NULL, dwBatch, notifs);
        }

        if (fObjectChanged)
        {
            // Since the properties for this folder have changed, send a notification about it.
            notifs[0].ulEventType         = fnevObjectModified;
            notifs[0].info.obj.cbEntryID  = CB_PRIVATE_EID;
            notifs[0].info.obj.lpEntryID  = (LPENTRYID)&m_ObjectEID;
            notifs[0].info.obj.cbParentID = CB_PRIVATE_EID;
            notifs[0].info.obj.lpParentID = (LPENTRYID)&m_ParentEID;
            notifs[0].info.obj.ulObjType  = MAPI_FOLDER;
            m_pStore->NotifyClients (&m_ObjectEID, 1, notifs);
        }
    }
    // Free up the progress indicator if we are using MAPI's object
    if (fReleaseProgress)
    {
        pProgress->Release();
    }
    LeaveCriticalSection (&m_csObj);
    // Release any data allocated by the back end
    m_pBEobj->FreeDeleteObjectsData (&Data);
    TraceResult ("CMAPIFolder::DeleteMessages", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CreateFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Create an IMAPIFolder object with the name and comment specified by
//      the caller. If a folder with the given name already exists in the
//      hiearachy of this parent folder, we may or may not fail the call,
//      based on the callers flags.
//      We call our backend storage to allocate space for the properties and
//      to return us a valid entry ID.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CreateFolder (ULONG          ulFolderType,
                                        LPTSTR         pszFolderName,
                                        LPTSTR         pszFolderComment,
                                        LPCIID         pInterface,
                                        ULONG          ulFlags,
                                        LPMAPIFOLDER * ppFolder)
{
    InfoTrace ("CMAPIFolder::CreateFolder method called");
    Validate_IMAPIFolder_CreateFolder (this,
                                       ulFolderType,
                                       pszFolderName,
                                       pszFolderComment,
                                       pInterface,
                                       ulFlags,
                                       ppFolder);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::CreateFolder: Cannot delete messages in a READ only folder");
        TraceResult ("CMAPIFolder::CreateFolder", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    // Allocate an empty entry ID structure for the new folder
    PRIVATE_ENTRYID eidNewFolder = { 0 };
    BOOL fFolderCreated = TRUE;
    CMAPIFolder * pFldObj;
    CPropList * pList;

    // Lock for exclusive access to the data on this object
    EnterCriticalSection (&m_csObj);
    // Let the back end storage create the space for the new object
    HRESULT hResult = m_pBEobj->CreateObject (m_ObjectEID,
                                              &eidNewFolder,
                                              &pList,
                                              MAPI_FOLDER,
                                              pszFolderName,
                                              pszFolderComment);
    if ((MAPI_E_COLLISION == hResult) && (ulFlags & OPEN_IF_EXISTS))
    {
        // If there is a folder with that name already, and the user said so, open it.
        hResult = m_pBEobj->OpenObject (eidNewFolder, &pList);
        fFolderCreated = FALSE;
    }
    long lAccess;
    if (!hResult)
    {
        // Inherit the parent's access rights.
        lAccess = (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                  (m_lAccess & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                   MAPI_ACCESS_READ |
                   MAPI_ACCESS_CREATE_HIERARCHY |
                   MAPI_ACCESS_CREATE_CONTENTS |
                   MAPI_ACCESS_CREATE_ASSOCIATED;
        pFldObj = new CMAPIFolder (m_pStore,
                                   &m_ObjectEID,
                                   &eidNewFolder,
                                   ulFolderType,
                                   pList,
                                   !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                   lAccess,
                                   m_pNPList);
        if (pFldObj)
        {
            // Set the returning object and set some internal flags
            *ppFolder = (LPMAPIFOLDER)pFldObj;
            pFldObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            // Darn..., we must get more RAM!!!
            hResult = E_OUTOFMEMORY;
            delete pList;
        }
    }
    // If we are successful at creating a new folder on the folder,
    // get ready to send the client notifications...
    if (!hResult && fFolderCreated)
    {
        NOTIFICATION notif = { 0 };
        BOOL fObjectChanged = FALSE;
        
        // When a new folder is created, if it is the same on the folder, we must
        // update the PR_SUBFOLDERS property and if modified send the appropiate
        // notification to the client subscribers. If this property is modified,
        // we must also update any in-memory hiereachy table of the parent of
        // this folder object.
        SPropValue spvSubFolders = { 0 }, spvLastModTime = { 0 };
        m_pList->Find (PR_SUBFOLDERS, &spvSubFolders);
        if (!spvSubFolders.Value.b)
        {
            fObjectChanged = TRUE;
            spvSubFolders.ulPropTag = PR_SUBFOLDERS;
            spvSubFolders.Value.b = TRUE;
            // The properties must be updated in our list BEFORE we send out notifications
            // so that clients can get the proper value after receiving the
            // notication from our store.
            m_pList->Insert (&spvSubFolders);
        
            // Send a notification the the parent folder has been modified if this
            // is the first folder created in it.
            notif.ulEventType         = fnevObjectModified;
            notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
            notif.info.obj.lpEntryID  = (LPENTRYID)&m_ObjectEID;
            notif.info.obj.cbParentID = CB_PRIVATE_EID;
            notif.info.obj.lpParentID = (LPENTRYID)&m_ParentEID;
            notif.info.obj.ulObjType  = MAPI_FOLDER;
            m_pStore->NotifyClients (&m_ObjectEID, 1, &notif);
            // Update the any in-memory hierarchy table for the parents of this folder.
            UpdateParentTable();
         }

        // Send a notification that a new object has been created. This is a the store level
        notif.ulEventType         = fnevObjectCreated;
        notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
        notif.info.obj.lpEntryID  = (LPENTRYID)&eidNewFolder;
        notif.info.obj.cbParentID = CB_PRIVATE_EID;
        notif.info.obj.lpParentID = (LPENTRYID)&m_ObjectEID;
        notif.info.obj.ulObjType  = MAPI_FOLDER;
        m_pStore->NotifyClients (NULL, 1, &notif);

        // We must update the hierarchy table of any instance of this folder that
        // may be open on the store.
        LPSPropValue pProps = NULL;
        DWORD dwLastAllocated = 0;
        CMAPIFolder * pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
        while (pFldObj)
        {
            if (fObjectChanged && (pFldObj != this))
            {
                pFldObj->m_pList->Insert (&spvSubFolders);
                pFldObj->m_pList->Insert (&spvLastModTime);
            }
            if (pFldObj->m_pHierTable)
            {
                hResult = m_pBEobj->GetTableProps (TABLE_HIERARCHY,
                                                   m_lAccess,
                                                   &m_ObjectEID,
                                                   &eidNewFolder,
                                                   pList,
                                                   &pProps,
                                                   dwLastAllocated,
                                                   pFldObj->m_pHierTable,
                                                   (LPVOID)&m_pStore->m_StoreEID,
                                                   (LPVOID)ulFolderType);
                TraceResult ("CMAPIFolder::CreateFolder: Failed to add row to table", hResult);
            }
            pFldObj->Release();
            pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_CONTINUE, pFldObj);
        }
        PrivFree ((LPVOID *)&pProps);
        // We don't care if we couldn't add the folder to the in-memory hierarchy table
        hResult = S_OK;        
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMAPIFolder::CreateFolder", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::CopyFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method copies the contents of a folder to the destination
//      folder specified by the caller. In our implementation we delegate to
//      the IMAPISupport object for the copy operation.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::CopyFolder (ULONG          cbEntryID,
                                      LPENTRYID      pEntryID,
                                      LPCIID         pInterface,
                                      LPVOID         pDestFolder,
                                      LPTSTR         pszNewFolderName,
                                      ULONG          ulUIParam,
                                      LPMAPIPROGRESS pProgress,
                                      ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::CopyFolder method called");
    Validate_IMAPIFolder_CopyFolder (this,
                                     cbEntryID,
                                     pEntryID,
                                     pInterface,
                                     pDestFolder,
                                     pszNewFolderName,
                                     ulUIParam,
                                     pProgress,
                                     ulFlags);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & FOLDER_MOVE))
    {
        TraceMessage ("CMAPIFolder::CopyFolder: Cannot move folder from a READ only folder");
        TraceResult ("CMAPIFolder::CopyFolder", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    BOOL fUseMAPICopy = FALSE;
    LPUNKNOWN pUnk;
    // See if the destination folder is one of our own, if it is, the QueryInterface()
    // method DID NOT increment the reference count on the object.
    HRESULT hResult = ((LPUNKNOWN)pDestFolder)->QueryInterface (IID_SAMPLE_IMAPIFolder, (LPVOID *)&pUnk);
    if (hResult)
    {
        // MAPI_DECLINE_OK means that if a store implementation does not natively knows
        // the destination object, a client can handle the copying of the messages because
        // it has a more efficient method than a default implementation by IMAPISupport.
        if (ulFlags & MAPI_DECLINE_OK)
        {
            // Let the client handle it
            hResult = MAPI_E_DECLINE_COPY;
        }
        else
        {
            fUseMAPICopy = TRUE;
        }        
    }
    else
    {
        CMAPIFolder * pDstFld = (CMAPIFolder *)pDestFolder;
        SBinaryArray * pEIDList = NULL;
        // If the folders are on different database files, then let MAPI
        // do the copy. I don't want to bother.
        if (!m_pBEobj->IsSameBackEnd (pDstFld->m_pBEobj->GetBackEndUID()))
        {
            fUseMAPICopy = TRUE;
            goto DelegateToMAPI;
        }
        EnterCriticalSection (&m_csObj);
        
        // In a future release we will implement our own CopyTo() methods
        // which should be more efficient that letting MAPI do the copy.
        // In the current release of this sample, we simply delegate to MAPI.
        //          IrvingD 3/11/96
        fUseMAPICopy = TRUE;

        LeaveCriticalSection (&m_csObj);
    }
DelegateToMAPI:
    if (fUseMAPICopy)
    {
        // If it is not an object that we now off, then let MAPI do the copying...
        hResult = m_pStore->m_pSupObj->CopyFolder (&IID_IMAPIFolder,
                                                   (LPVOID)this,
                                                   cbEntryID,
                                                   pEntryID,
                                                   pInterface,
                                                   pDestFolder,
                                                   pszNewFolderName,
                                                   ulUIParam,
                                                   pProgress,
                                                   ulFlags);
    }
//ErrorExit:
    TraceResult ("CMAPIFolder::CopyFolder", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::DeleteFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function delete a subfolder in the current folder. The contents
//      of the deleted folder (sub folders and messages) are also deleted if
//      the client wanted to.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::DeleteFolder (ULONG          cbEntryID,
                                        LPENTRYID      pEntryID,
                                        ULONG          ulUIParam,
                                        LPMAPIPROGRESS pProgress,
                                        ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::DeleteFolder method called");
    Validate_IMAPIFolder_DeleteFolder (this,
                                       cbEntryID,
                                       pEntryID,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    // Check the integrity of the entry ID passed in
    if (!pEntryID ||
        !cbEntryID ||
        CB_PRIVATE_EID != cbEntryID ||
        !m_pBEobj->IsValidEntryID ((PRIVATE_ENTRYID *)pEntryID))
    {
        return MAPI_E_INVALID_ENTRYID;
    }
    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);
    
    // Decode the entry ID into a private one
    PRIVATE_ENTRYID * pEID = (PRIVATE_ENTRYID *)pEntryID;
    // Use the common delete folder function to do the action deletion
    HRESULT hResult = DeleteFolderContents (m_ObjectEID,
                                            *pEID,
                                            ulUIParam,
                                            pProgress,
                                            ulFlags);
    if (!hResult)
    {
        // If there were no warnings, delete the object from the storage.
        // If even one object could not be deleted, the parent of that
        // object should not be deleted.
        hResult = m_pBEobj->DeleteOneObject (*pEID, m_ObjectEID);
        if (!hResult)
        {
            // Now we must delete the deleted folder from the hierarchy
            // table in this folder and in other instances of the "this" folder
            SPropValue spvInstKey = { 0 };
            spvInstKey.ulPropTag = PR_INSTANCE_KEY;
            spvInstKey.Value.bin.cb = sizeof(DWORD);
            spvInstKey.Value.bin.lpb = (LPBYTE)&(pEID->dwObjID);
            CMAPIFolder * pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
            while (pFldObj)
            {
                if (pFldObj->m_pHierTable)
                {
                    pFldObj->m_pHierTable->HrDeleteRow (&spvInstKey);
                }
                pFldObj->Release();
                // Get the next folder, continue after the last hit found in the list
                pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_CONTINUE, pFldObj);
            }
        }
    }
    // Unlock the object
    LeaveCriticalSection (&m_csObj);

    // Trace and set the last error code
    TraceResult ("CMAPIFolder::DeleteFolder", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SetReadFlags()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Modify the READ/UNREAD flags of messages in this folder. If the
//      messages have read receipt notifications in them, we supress or
//      send the notification message based on the flags passed.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::SetReadFlags (LPENTRYLIST    pMsgList,
                                        ULONG          ulUIParam,
                                        LPMAPIPROGRESS pProgress,
                                        ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::SetReadFlags method called");
    Validate_IMAPIFolder_SetReadFlags (this,
                                       pMsgList,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::SetReadFlags: Cannot set read flags in a READ only folder");
        TraceResult ("CMAPIFolder::SetReadFlags", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    
    // Initialize the array of notifications and eid which are used to send
    // fnevObjectModified notifications to subscribers of this store.
    HRESULT hResult = S_OK;
    NOTIFICATION notif = { 0 };
    notif.ulEventType         = fnevObjectModified;
    notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
    notif.info.obj.cbParentID = CB_PRIVATE_EID;
    notif.info.obj.lpParentID = (LPENTRYID)&m_ObjectEID;
    notif.info.obj.ulObjType  = MAPI_MESSAGE;
    PRIVATE_ENTRYID * pEID;
    
    // Eliminate unsupported flags
    ulFlags &= ~MAPI_DEFERRED_ERRORS;

    CMessage * pMessage;
    ULONG i = 0, ulObjType, cbSize;
    LPSRowSet pRows = NULL;
    BOOL fReleaseProgress = FALSE, fAtLeastOne = FALSE, fContinue = TRUE;
    // If the client wants UI, but did not bother to pass one in, call MAPI and
    // create a default progress indicator.
    if (ulFlags & MESSAGE_DIALOG)
    {
        if (NULL == pProgress)
        {
            hResult = m_pStore->m_pSupObj->DoProgressDialog (ulUIParam, MAPI_TOP_LEVEL, &pProgress);
            if (hResult)
            {
                TraceResult ("CMAPIFolder::SetReadFlags: Failed to get default progress interface", hResult);
                pProgress = NULL;
            }
            else
            {
                fReleaseProgress = TRUE;
            }
        }
    }
    LPMAPITABLE pTable = NULL;
    if (!pMsgList)
    {
        hResult = GetContentsTable (0, &pTable);
        if (!hResult)
        {
            SPropTagArray sptEntryID = { 1, { PR_ENTRYID } };
            hResult = pTable->SetColumns (&sptEntryID, 0);
        }
        if (hResult)
        {
            goto ErrorExit;
        }
    }

    // Lock the folder object for exclusive access
    EnterCriticalSection (&m_csObj);

    __try
    {
        hResult = S_OK;
        while (fContinue)
        {
            if (pMsgList)
            {
                if (i == pMsgList->cValues)
                {
                    break; // Out of the WHILE() loop
                }
                cbSize = pMsgList->lpbin[i].cb;
                pEID = (PRIVATE_ENTRYID *)pMsgList->lpbin[i].lpb;
            }
            else
            {
                hResult = pTable->QueryRows (1, 0, &pRows);
                if (!hResult)
                {
                    if (pRows->cRows)
                    {
                        ASSERT (PR_ENTRYID == pRows->aRow->lpProps->ulPropTag);
                        cbSize = pRows->aRow->lpProps->Value.bin.cb;
                        pEID = (PRIVATE_ENTRYID *)pRows->aRow->lpProps->Value.bin.lpb;
                    }
                    else
                    {
                        FreeProws (pRows);
                        pRows = NULL;
                        break; // Out of the WHILE() loop
                    }
                }
                if (hResult)
                {
                    break; // Out of the WHILE() loop
                }
            }
            i++;
            if (!m_pStore->OpenEntry (cbSize,
                                      (LPENTRYID)pEID,
                                      NULL,
                                      MAPI_MODIFY,
                                      &ulObjType,
                                      (LPUNKNOWN *)&pMessage))
            {
                if (MAPI_MESSAGE == ulObjType)
                {
                    if (!pMessage->InternalSetReadFlag (ulFlags, FALSE))
                    {
                        fAtLeastOne = TRUE;
                        notif.info.obj.lpEntryID = (LPENTRYID)pEID;
                        // Send the notification to subscribers
                        m_pStore->NotifyClients (pEID, 1, &notif);
                    }
                    else
                    {
                        hResult = MAPI_W_PARTIAL_COMPLETION;
                    }
                }
                else
                {
                    hResult = MAPI_W_PARTIAL_COMPLETION;
                }
                pMessage->Release();
            }
            else
            {
                hResult = MAPI_W_PARTIAL_COMPLETION;
            }
            if (pRows)
            {
                FreeProws (pRows);
                pRows = NULL;
            }
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        hResult = E_FAIL;
    }
    if (fAtLeastOne)
    {
        // Since the properties for this folder have changed, send a notification about it.
        notif.info.obj.lpEntryID  = (LPENTRYID)&m_ObjectEID;
        notif.info.obj.lpParentID = (LPENTRYID)&m_ParentEID;
        notif.info.obj.ulObjType  = MAPI_FOLDER;
        m_pStore->NotifyClients (&m_ObjectEID, 1, &notif);
    }
    LeaveCriticalSection (&m_csObj);
ErrorExit:
    // Free up the progress indicator if we are using MAPI's object
    if (fReleaseProgress)
    {
        pProgress->Release();
    }
    if (pTable)
    {
        pTable->Release();
    }
    TraceResult ("CMAPIFolder::SetReadFlags", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::GetMessageStatus()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Retrieves the status associated with a message in a folder. The
//      status of a message is stored in the message object itself in the
//      property PR_MSG_STATUS.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::GetMessageStatus (ULONG     cbEntryID,
                                            LPENTRYID pEntryID,
                                            ULONG     ulFlags,
                                            ULONG *   pulMessageStatus)
{
    InfoTrace ("CMAPIFolder::GetMessageStatus method called");
    Validate_IMAPIFolder_GetMessageStatus (this,
                                           cbEntryID,
                                           pEntryID,
                                           ulFlags,
                                           pulMessageStatus);

    PRIVATE_ENTRYID * pEID = (PRIVATE_ENTRYID *)pEntryID;
    HRESULT hResult = S_OK;
    
    // Lock the folder object for exclusive access
    EnterCriticalSection (&m_csObj);

    if (m_pStore->m_pBEobj->IsValidEntryID (pEID) && MAPI_MESSAGE == pEID->bObject)
    {
        CPropList * pList;
        hResult = m_pBEobj->OpenObject (*pEID, &pList);
        if (!hResult)
        {
            SPropValue spvOne = { 0 };
            if (S_OK == pList->Find (PR_MSG_STATUS, &spvOne))
            {
                ASSERT (PR_MSG_STATUS == spvOne.ulPropTag);
                *pulMessageStatus = spvOne.Value.ul;
            }
            else
            {
                hResult = MAPI_E_NOT_FOUND;
            }
            delete pList;
        }
    }
    else
    {
        hResult = MAPI_E_INVALID_ENTRYID;
    }
    LeaveCriticalSection (&m_csObj);

	TraceResult ("CMAPIFolder::GetMessageStatus", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SetMessageStatus()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Sets the status associated with a message in a folder. The
//      status of a message is stores in the message object itself in the
//      property PR_MSG_STATUS.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::SetMessageStatus (ULONG     cbEntryID,
                                            LPENTRYID pEntryID,
                                            ULONG     ulNewStatus,
                                            ULONG     ulNewStatusMask,
                                            ULONG *   pulOldStatus)
{
    InfoTrace ("CMAPIFolder::SetMessageStatus method called");
    Validate_IMAPIFolder_SetMessageStatus (this,
                                           cbEntryID,
                                           pEntryID,
                                           ulNewStatus,
                                           ulNewStatusMask,
                                           pulOldStatus);
    // Lock the folder object for exclusive access
    EnterCriticalSection (&m_csObj);

    ULONG ulObjType;
    CMessage * pMessage;
    HRESULT hResult = m_pStore->OpenEntry (cbEntryID,
                                           pEntryID,
                                           NULL,
                                           MAPI_MODIFY,
                                           &ulObjType,
                                           (LPUNKNOWN *)&pMessage);
    if (!hResult)
    {
        if (MAPI_MESSAGE == ulObjType)
        {
            SPropValue spvOne = { 0 };
            if (S_OK == pMessage->m_pList->Find (PR_MSG_STATUS, &spvOne))
            {
                ULONG ulOldStatus = spvOne.Value.ul;

                // If the client wants to know the old status, return it
	            if (pulOldStatus)
                {
		            *pulOldStatus = ulOldStatus;
                }

	            // If the client wants to change the status, then get if better have supplied a mask
	            if (ulNewStatusMask)
	            {
		            ulNewStatus &= ulNewStatusMask;
		            spvOne.Value.ul &= ~ulNewStatusMask;
		            spvOne.Value.ul |= ulNewStatus;
                    // If there was a change on the status of the message save it
                    if (ulOldStatus != spvOne.Value.ul)
                    {
		                spvOne.ulPropTag = PR_MSG_STATUS;
                        hResult = pMessage->m_pList->Insert (&spvOne);
                        if (!hResult)
                        {
                            // Save the message object. The call to save changes will take care
                            // of broadcasting notifications on the object and updating
                            // the contents table in this folder.
                            pMessage->SaveChanges (0);
                        }
                    }
	            }
            }
        }
        pMessage->Release();
    }
    LeaveCriticalSection (&m_csObj);

	TraceResult ("CMAPIFolder::SetMessageStatus", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::SaveContentsSort()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't saving the sort order of the
//      contents tables in this folder.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMAPIFolder::SaveContentsSort (LPSSortOrderSet pSortCriteria,
                                            ULONG           ulFlags)
{
    InfoTrace ("CMAPIFolder::SaveContentsSort method called");
    Validate_IMAPIFolder_SaveContentsSort (this, pSortCriteria, ulFlags);
    TraceResult ("CMAPIFolder::SaveContentsSort", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::EmptyFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function empties a folder by deleting all of its contained
//      messages and subfolders
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::EmptyFolder (ULONG          ulUIParam,
                                       LPMAPIPROGRESS pProgress,
                                       ULONG          ulFlags)
{
    InfoTrace ("CMAPIFolder::EmptyFolder method called");
    Validate_IMAPIFolder_EmptyFolder (this, ulUIParam, pProgress, ulFlags);
    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);
    // Use the common function to delete contained folders
    HRESULT hResult = DeleteFolderContents (m_ParentEID,
                                            m_ObjectEID,
                                            ulUIParam,
                                            pProgress,
                                            ulFlags | DEL_FOLDERS | DEL_MESSAGES);
    if (!hResult)
    {
        // If there were now warnings, delete ALL the rows in all
        // running instances of this folder
        CMAPIFolder * pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
        while (pFldObj)
        {
            if (pFldObj->m_pHierTable)
            {
                pFldObj->m_pHierTable->HrDeleteRows (TAD_ALL_ROWS, NULL, NULL);
            }
            pFldObj->Release();
            // Get the next folder, continue after the last hit found in the list
            pFldObj = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_CONTINUE, pFldObj);
        }
    }
    // Unlock the object
    LeaveCriticalSection (&m_csObj);

    // Trace and set the last error code
    TraceResult ("CMAPIFolder::EmptyFolder", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
// Other member functions specific to this class
//

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::UpdateContentTables()
//
//    Parameters
//      [IN]    Action              Type of update action on the table: ADD,
//                                  DELETE, MODIFY, etc.
//      [IN]    fAssociatedCount    TRUE if the change to the table affects
//                                  the ASSOCIATED content count
//      [IN]    fUnReadCount        TRUE if the change to the table affects
//                                  the UNREAD content count
//      [IN]    pMessageEID         Entry ID of the message for the given
//                                  row that will be modified.
//      [IN]    pMsgProps           Properties of the message modified on
//                                  the table
//
//    Purpose
//      This function modifies the contents table of the folder where a
//      message object has been modified. A contents table is modified when
//      ever a message is saved, when a message is created or deleted, and
//      when some of its properties are altered.
//      This function is RECURSIVE.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMAPIFolder::UpdateContentTables (UPDATE_TABLE_ACTION Action,
                                               BOOL                fAssociatedCount,
                                               BOOL                fUnReadCount,
                                               PRIVATE_ENTRYID *   pMessageEID,
                                               CPropList *         pMsgProps,
                                               BOOL                bRecursing)
{
    // Lock the object for exclusive access
    EnterCriticalSection (&m_csObj);
    SPropValue spvOne = { 0 };
    // If a new row has been added or one has been deleted the content count
    // properties of this folder have been changed. So get the property
    // values read for modification
    if (UPD_TABLE_ADD_ROW == Action || UPD_TABLE_DELETE_ROW == Action)
    {
        m_pList->Find ((fAssociatedCount ? PR_ASSOC_CONTENT_COUNT : PR_CONTENT_COUNT) , &spvOne);
        ASSERT (spvOne.ulPropTag == PR_CONTENT_COUNT || spvOne.ulPropTag == PR_ASSOC_CONTENT_COUNT);
        if (UPD_TABLE_ADD_ROW == Action)
        {
            spvOne.Value.l++;
        }
        else
        {
            ASSERT (Action == UPD_TABLE_DELETE_ROW);
            if (spvOne.Value.l)
            {
                spvOne.Value.l--;
            }
        }
        m_pList->Insert (&spvOne);
        if (fUnReadCount)
        {
            m_pList->Find (PR_CONTENT_UNREAD, &spvOne);
            ASSERT (spvOne.ulPropTag == PR_CONTENT_UNREAD);
            if (UPD_TABLE_ADD_ROW == Action)
            {
                spvOne.Value.l++;
            }
            else
            {
                ASSERT (Action == UPD_TABLE_DELETE_ROW);
                if (spvOne.Value.l)
                {
                    spvOne.Value.l--;
                }
            }
            m_pList->Insert (&spvOne);
        }
    }
    if (UPD_TABLE_REFRESH_ADD_UNREAD == Action || UPD_TABLE_REFRESH_DEL_UNREAD == Action)
    {
        m_pList->Find (PR_CONTENT_UNREAD, &spvOne);
        ASSERT (spvOne.ulPropTag == PR_CONTENT_UNREAD);
        if (UPD_TABLE_REFRESH_ADD_UNREAD == Action)
        {
            spvOne.Value.l++;
        }
        else
        {
            ASSERT (Action == UPD_TABLE_REFRESH_DEL_UNREAD);
            if (spvOne.Value.l)
            {
                spvOne.Value.l--;
            }
        }
        m_pList->Insert (&spvOne);
        if (!bRecursing)
        {
            // We need to commit the changes to the folder's properties
            CommitChanges();
        }
    }
    // Now figure out which table needs to be updated and if the
    // folder has it open update it
    LPTABLEDATA pTable = (fAssociatedCount ? m_pAsscContTable : m_pContTable);
    if (pTable)
    {
        ASSERT (pMessageEID);
        ASSERT (m_pBEobj->IsValidEntryID (pMessageEID));
        if (UPD_TABLE_DELETE_ROW == Action)
        {
            spvOne.ulPropTag = PR_INSTANCE_KEY;
            spvOne.Value.bin.cb = sizeof(DWORD);
            spvOne.Value.bin.lpb = (LPBYTE)&(pMessageEID->dwObjID);
            pTable->HrDeleteRow (&spvOne);
        }
        else
        {
            DWORD dwLastAllocated = 0;
            LPSPropValue pProps = NULL;
            CPropList * pList = NULL;
            HRESULT hResult = S_OK;
            // If the caller did not give up the property list we must open
            // the object from the database and get the obj's properties.
            if (NULL == pMsgProps)
            {
                hResult = m_pBEobj->OpenObject (*pMessageEID, &pList);
            }
            if (!hResult)
            {
                // Now we are ready to modify the contents table in this folder.
                // The back end object will modify the table object.
                hResult = m_pBEobj->GetTableProps (TABLE_CONTENTS,
                                                   m_lAccess,
                                                   &m_ObjectEID,
                                                   pMessageEID,
                                                   (pMsgProps ? pMsgProps : pList),
                                                   &pProps,
                                                   dwLastAllocated,
                                                   pTable,
                                                   (LPVOID)&m_pStore->m_StoreEID,
                                                   NULL);
                delete pList;
            }
            PrivFree ((LPVOID *)&pProps);
        }
    }
    // If this function is not in a recursive iteration, update other instances of this folder.
    if (!bRecursing)
    {
        // Now we need to update other instances of this folder
        CMAPIFolder * pFolder = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
        while (pFolder)
        {
            // Update any folder that is not 'this'
            if (pFolder != this)
            {
                pFolder->UpdateContentTables (Action,
                                              fAssociatedCount,
                                              fUnReadCount,
                                              pMessageEID,
                                              pMsgProps,
                                              TRUE);
            }
            pFolder->Release();
            pFolder = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_CONTINUE, pFolder);
        }
    }

    // Unlock the object
    LeaveCriticalSection (&m_csObj);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::UpdateParentTable()
//
//    Parameters
//      None
//
//    Purpose
//      When the folder properties are modified in any way (folders added,
//      properties modified, added or deleted, etc.) We need to update the
//      hierarchy table in our parent folder where
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMAPIFolder::UpdateParentTable()
{
    DWORD dwLastAllocated = 0;
    LPSPropValue pProps = NULL;
    // Start the walk in the list looking for this entry ID
    CMAPIFolder * pParent = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ParentEID, WALK_START, NULL);
    while (pParent)
    {
        // If the hierarchy table is opened, then update its contents
        if (pParent->m_pHierTable)
        {
            m_pBEobj->GetTableProps (TABLE_HIERARCHY,
                                     m_lAccess,
                                     &(pParent->m_ObjectEID),
                                     &m_ObjectEID,
                                     m_pList,
                                     &pProps,
                                     dwLastAllocated,
                                     pParent->m_pHierTable,
                                     (LPVOID)&m_pStore->m_StoreEID,
                                     (LPVOID)m_lFolderType);
        }
        pParent->Release();
        pParent = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ParentEID, WALK_CONTINUE, pParent);
    }
    PrivFree ((LPVOID *)&pProps);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::DeleteFolderContents()
//
//    Parameters
//      [IN]    ParentFolderEID     Entry ID of the parent of the folder whose
//                                  contents we are deleting
//      [IN]    FolderEID           Entry ID of the folder whose contents
//                                  we are deleting
//      [IN]    ulUIParam           Handle to the window that will own any
//                                  UI displayed
//      [IN]    pProgress           Pointer to an IMAPIProgress interface
//      [IN]    ulFlags             Flags to control the behaviour of
//                                  the operation
//
//    Purpose
//      Delete the messages and subfolders with the folder specified with
//      by the caller. Here we verify the caller passed the appropiate flags
//      to delete the contained items and if not return an error.
//      For the actual delete of the items, we call the back end
//      implementation with the list of object IDs that must be deleted.
//      Once the objects have been deleted, I send the appropiate
//      notifications to tell interested clients that objects have been
//      deleted in this message store that the container folder has
//      been modified.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::DeleteFolderContents (PRIVATE_ENTRYID ParentFolderEID,
                                                PRIVATE_ENTRYID FolderEID,
                                                ULONG           ulUIParam,
                                                LPMAPIPROGRESS  pProgress,
                                                ULONG           ulFlags)
{
    // Check the access level of the folder
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMAPIFolder::DeleteFolderContents: Cannot delete contents from a READ only folder");
        return E_ACCESSDENIED;
    }
    BOOL fDeleteInThisObject = TRUE;
    CPropList * pList = m_pList;
    HRESULT hResult = S_OK;
    // If the entry IDs are not the same as 'this' object's entry ID
    // then this function was called by DeleteFolder()
    if (0 != memcmp (&FolderEID, &m_ObjectEID, CB_PRIVATE_EID))
    {
        fDeleteInThisObject = FALSE;
        hResult = m_pBEobj->OpenObject (FolderEID, &pList);
    }
    if (!hResult)
    {
        // Get some properties of our internal list. These properties are
        // used to accept the behaviour indicated in the flags passed in by the user.
        SPropValue spvSubFolders, spvContentCount, spvAsscContentCount;
        hResult = pList->Find (PR_SUBFOLDERS, &spvSubFolders);
        if (!hResult)
        {
            hResult = pList->Find (PR_CONTENT_COUNT, &spvContentCount);
            if (!hResult)
            {
                hResult = pList->Find (PR_ASSOC_CONTENT_COUNT, &spvAsscContentCount);
            }
        }
        if (hResult)
        {
            hResult = MAPI_E_CORRUPT_DATA;
        }
        else
        {
            // Return the appropiate error if the flags are not set
            if (spvSubFolders.Value.b && !(ulFlags & DEL_FOLDERS))
            {
                hResult = MAPI_E_HAS_FOLDERS;
            }
            else
            {
                if ((spvContentCount.Value.l || spvAsscContentCount.Value.l) && !(ulFlags & DEL_MESSAGES))
                {
                    hResult = MAPI_E_HAS_MESSAGES;
                }
            }
        }
    }
    if (pList && !fDeleteInThisObject)
    {
        delete pList;
        pList = NULL;
    }
    // "houston... we have a problem!..." bail out
    if (hResult)
    {
        return hResult;
    }
    
    // If the client application wants to display a progress UI but did not bother to
    // give us one, we call MAPI to use the built-in progress indicator
    if (ulFlags & FOLDER_DIALOG)
    {
        if (NULL == pProgress)
        {
            hResult = m_pStore->m_pSupObj->DoProgressDialog (ulUIParam, MAPI_TOP_LEVEL, &pProgress);
            if (hResult)
            {
                TraceResult ("CMAPIFolder::DeleteFolderContents: Failed to get default progress interface", hResult);
                pProgress = NULL;
            }
        }
        else
        {
            pProgress->AddRef();
        }
    }
    // Initialize the array of entry ID used for sending
    // batch notification of deleted objects.
    NOTIFICATION notifs[NOTIFICATION_BATCH_SIZE] = { 0 };
    PRIVATE_ENTRYID EIDList[NOTIFICATION_BATCH_SIZE] = { 0 };
    DWORD dwBatch = 0;
    for (ULONG j=0; j<NOTIFICATION_BATCH_SIZE; j++)
    {
        EIDList[j].uidGlobal = m_pStore->m_StoreUID;
        EIDList[j].bVersion = ENTRYID_VERSION;
    }

    // Set up the data structure with the information we pass to the back end
    OPENOBJECTS_DATA Data = { 0 };
    Data.pOpenObjList = &(m_pStore->m_OpenObjList);
    // If we are getting called from EmptyFolder(), and the caller said that we
    // should not delete the associated messages, then remind ourselves to skip
    // the associated messages.
    if (fDeleteInThisObject && !(ulFlags & DEL_ASSOCIATED))
    {
        Data.dwFlags |= OPENOBJDATA_SKIP_ASSOCIATED;
    }

    BOOL fObjectChanged = FALSE;
    // Call the back end for the actual delete
    hResult = m_pBEobj->DeleteObjects (FolderEID,
                                       EMPTY_CONTAINER,
                                       pProgress,
                                       &Data);
    if (SUCCEEDED(hResult)) // We might get warnings
    {
        PRIVATE_ENTRYID SubFolderEID = { 0 };
        SubFolderEID.uidGlobal = m_pStore->m_StoreUID;
        SubFolderEID.bVersion = ENTRYID_VERSION;
        SubFolderEID.bObject = MAPI_FOLDER;
        // For each message and for the contents of each deleted subfolders and for
        // the subfolders as well, send a notification to the client
        for (ULONG i=0; i<Data.dwSubFldCount; i++)
        {
            SubFolderEID.dwObjID = Data.pSubFldInfo[i].dwFldObjID;
            for (ULONG j=0; j<Data.pSubFldInfo[i].dwCount; j++)
            {
                if (Data.pSubFldInfo[i].pSubEIDList[j].dwObjID)
                {
                    EIDList[dwBatch].bObject = Data.pSubFldInfo[i].pSubEIDList[j].bObject;
                    EIDList[dwBatch].dwObjID = Data.pSubFldInfo[i].pSubEIDList[j].dwObjID;
                
                    notifs[dwBatch].ulEventType         = fnevObjectDeleted;
                    notifs[dwBatch].info.obj.cbEntryID  = CB_PRIVATE_EID;
                    notifs[dwBatch].info.obj.lpEntryID  = (LPENTRYID)(&EIDList[dwBatch]);
                    notifs[dwBatch].info.obj.ulObjType  = EIDList[dwBatch].bObject;

                    notifs[dwBatch].info.obj.cbParentID = CB_PRIVATE_EID;
                    notifs[dwBatch].info.obj.lpParentID = (LPENTRYID)&SubFolderEID;
                    dwBatch++;
                    // This implementation of a store provider batches NOTIFICATION_BATCH_SIZE notifications
                    // of deleted objects before sending them out.
                    if (NOTIFICATION_BATCH_SIZE == dwBatch)
                    {
                        m_pStore->NotifyClients (NULL, dwBatch, notifs);
                        dwBatch = 0;
                    }
                }
            }
            // If any nofications where not sent because send them now.
            if (dwBatch)
            {
                m_pStore->NotifyClients (NULL, dwBatch, notifs);
                dwBatch = 0;
            }
        }

        // Walk through the list of entry ID passed in by the client
        for (i=0; i<Data.dwCount; i++)
        {
            // If the database could not delete this object, the object ID size would be set to 0,
            // so on anything that is non-Zero, we assume got deleted and a notification is in order.
            if (Data.pSubEIDList[i].dwObjID)
            {
                fObjectChanged = TRUE;
                EIDList[dwBatch].bObject = Data.pSubEIDList[i].bObject;
                EIDList[dwBatch].dwObjID = Data.pSubEIDList[i].dwObjID;
                
                notifs[dwBatch].ulEventType         = fnevObjectDeleted;
                notifs[dwBatch].info.obj.cbEntryID  = CB_PRIVATE_EID;
                notifs[dwBatch].info.obj.lpEntryID  = (LPENTRYID)(&EIDList[dwBatch]);
                notifs[dwBatch].info.obj.ulObjType  = EIDList[dwBatch].bObject;

                notifs[dwBatch].info.obj.cbParentID = CB_PRIVATE_EID;
                notifs[dwBatch].info.obj.lpParentID = (LPENTRYID)&FolderEID;
                dwBatch++;
                // This implementation of a store provider batches NOTIFICATION_BATCH_SIZE notifications
                // of deleted objects before sending them out.
                if (NOTIFICATION_BATCH_SIZE == dwBatch)
                {
                    m_pStore->NotifyClients (NULL, dwBatch, notifs);
                    dwBatch = 0;
                }
            }
        }
        // If any nofications where not sent because
        // there where fewer than the batch limit, send them now.
        if (dwBatch)
        {
            m_pStore->NotifyClients (NULL, dwBatch, notifs);
        }

        // From DeleteFolder() or EmptyFolder() we must modify the properties of the folder
        // container and make sure we update PR_LAST_MODIFICATION_TIME and PR_SUBFOLDERS.
        SPropValue spvLastMod = { 0 }, spvSavedHasFolders =  { 0 }, spvThisHasFolders = { 0 };
        if (!m_pBEobj->OpenObject (m_ObjectEID, &pList))
        {
            pList->Find (PR_SUBFOLDERS, &spvSavedHasFolders);
            m_pList->Find (PR_SUBFOLDERS, &spvThisHasFolders);
            if (spvSavedHasFolders.Value.b != spvThisHasFolders.Value.b)
            {
                ASSERT (PR_SUBFOLDERS == spvSavedHasFolders.ulPropTag);
                m_pList->Insert (&spvSavedHasFolders);
            }

            pList->Find (PR_LAST_MODIFICATION_TIME, &spvLastMod);
            ASSERT (PR_LAST_MODIFICATION_TIME == spvLastMod.ulPropTag);
            m_pList->Insert (&spvLastMod);

            delete pList;
            pList = NULL;
        }

        // If fDeleteInThisObject is set to TRUE, it means that we are getting called from
        // EmptyFolder() in which case we need to send notifications out that the container
        // changed its content count properties.
        // If on the other hand, fDeleteInThisObject is set to FALSE, we are getting called
        // from DeleteFolder() to delete the contents of the folder we are deleting, in
        // which case it not useful to send notifications that the folder we are deleting
        // has changed because by the time the client receives the notification, the object
        // in question will be gone from the storage.
        if (fObjectChanged && fDeleteInThisObject)
        {
            // We must update other instances of this folders so they reflect the same
            // properties as this one.
            UpdatePropCacheInReplicas();

            // Since the properties for this folder have changed, send a notification about it.
            notifs[0].ulEventType         = fnevObjectModified;
            notifs[0].info.obj.cbEntryID  = CB_PRIVATE_EID;
            notifs[0].info.obj.lpEntryID  = (LPENTRYID)&FolderEID;
            notifs[0].info.obj.cbParentID = CB_PRIVATE_EID;
            notifs[0].info.obj.lpParentID = (LPENTRYID)&ParentFolderEID;
            notifs[0].info.obj.ulObjType  = MAPI_FOLDER;
            m_pStore->NotifyClients (&FolderEID, 1, notifs);
            UpdateParentTable();
        }
    }
    if (pProgress)
    {
        pProgress->Release();
    }
    // Free the data that the back end returned to us.
    m_pBEobj->FreeDeleteObjectsData (&Data);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::UpdatePropertyCache()
//
//    Parameters
//      [IN]    pSourceList     Pointer to the property cache in
//                              the source folder
//
//    Purpose
//      Given the pointer to the property cache of the folder where the
//      properties have been altered, we use this to copy it into our
//      own property cache.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMAPIFolder::UpdatePropertyCache (CPropList * pSourceList)
{
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = pSourceList->CopyToList (m_pList);
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMAPIFolder::UpdatePropCacheInReplicas()
//
//    Parameters
//      None
//
//    Purpose
//      When a user modifies the list or values of the properties in a folder
//      object, all opened instances of this folder must have the same
//      in-memory property cache. This function walks the list of opened
//      folder with this entry ID and updates their prop cache
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMAPIFolder::UpdatePropCacheInReplicas()
{
    // Update other instances of this folder that are running.
    // Remember that IMAPIFolder objects are not transacted Changes
    // to one object are reflected inmediately in ALL instances of that objects.
    // Start the walk in the list looking for this entry ID
    CMAPIFolder * pOtherFld = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_START, NULL);
    while (pOtherFld)
    {
        // If the folder is NOT THIS, update it
        if (pOtherFld != this)
        {
            pOtherFld->UpdatePropertyCache (m_pList);
        }
        pOtherFld->Release();
        pOtherFld = (CMAPIFolder *)m_pStore->m_OpenObjList.Walk (m_ObjectEID, WALK_CONTINUE, pOtherFld);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    FolderTablesViewRelease()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function is called when a table returned by the IMAPIFolder
//      objects is released, so that we may release the source IMAPIFolder
//      object associated with it.
//
//    Return Value
//      None
//
STDAPI_(void) FolderTablesViewRelease (ULONG       ulCallBackData,
                                       LPTABLEDATA pTableData,
                                       LPMAPITABLE pTableView)
{
    DWORD * pdwViewsLeft = NULL;
    LPTABLEDATA *ppTable = NULL;
    CMAPIFolder * pFolder = (CMAPIFolder *)ulCallBackData;
    __try
    {
        if (!pFolder || 0 == pFolder->m_cRef)
        {
            TraceMessage ("FolderTablesViewRelease: Invalid Folder Object");
            return;
        }
        EnterCriticalSection (&pFolder->m_csObj);
        if (pFolder->m_pHierTable == pTableData)
        {
            pdwViewsLeft = &(pFolder->m_dwHierTableViews);
            ppTable= &(pFolder->m_pHierTable);
            InfoTrace ("FolderTablesViewRelease: A hierarchy table view has been released");
        }
        else
        {
            if (pFolder->m_pContTable == pTableData)
            {
                pdwViewsLeft = &(pFolder->m_dwContTableViews);
                ppTable= &(pFolder->m_pContTable);
                InfoTrace ("FolderTablesViewRelease: A contents table view has been released");
            }
            else
            {
                if (pFolder->m_pAsscContTable == pTableData)
                {
                    pdwViewsLeft = &(pFolder->m_dwAsscContTableViews);
                    ppTable= &(pFolder->m_pAsscContTable);
                    InfoTrace ("FolderTablesViewRelease: An associated contents table view has been released");
                }
                else
                {
                    TraceMessage ("FolderTablesViewRelease: Called for an unknown table");
                    goto Exit;
                }
            }
        }
        // Decrement the count of views and release the ITableData object
        // if there are no more active views on that data
        ASSERT (0 != (*pdwViewsLeft));
        --(*pdwViewsLeft);
Exit:
        LeaveCriticalSection (&pFolder->m_csObj);
        pFolder->Release();
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("FolderTablesViewRelease: An exception occured. Exception code: %#08x", GetExceptionCode());
        TraceMessage ("     MSWDS: WARNING - Possible deadlock if sync object was not released due to exception");
    }
}

// End of file for MSFOLDER.CPP
