///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSGSTORE.CPP
//
//  Description
//      This file implements the IMsgStore interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSWDS.H"
#include "MSSRVBE.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

#define NUM_MSGSTORE_PROPS      21
// Properties available in our implementation of IMsgStore
SizedSPropTagArray(NUM_MSGSTORE_PROPS, sptMsgStoreObj) =
{
    NUM_MSGSTORE_PROPS,
    {
        PR_RECEIVE_FOLDER_SETTINGS,
        PR_ACCESS,
        PR_ACCESS_LEVEL,
        PR_OBJECT_TYPE,
        PR_STORE_SUPPORT_MASK,
        PR_VALID_FOLDER_MASK,
        PR_FOLDER_TYPE,
        PR_CONTENT_UNREAD,
        PR_STORE_STATE,
        PR_SUBFOLDERS,
        PR_CURRENT_VERSION,
        PR_COMMENT,
        PR_DISPLAY_NAME,
        PR_MAPPING_SIGNATURE,
        PR_STORE_RECORD_KEY,
        PR_RECORD_KEY,
        PR_IPM_SUBTREE_ENTRYID,
        PR_PARENT_ENTRYID,
        PR_ENTRYID,
        PR_STORE_ENTRYID,
        PR_MDB_PROVIDER
    }
};

extern "C"
{
    extern MAPIUID guidProvider;
};

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::CMsgStore()
//
//    Parameters
//      pMSProviderObj  Pointer to the associated IMSProvider object
//      pSupObj         MAPI support object for this session
//      pszServer       Server name we are connecting to
//      StoreEID        MAPI-wrapped entry ID of this store instance
//      pBEobj          Pointer to the CBackEnd object used to access the
//                      message database file
//      fDeferErrors    Weather or not we should defer errors in this instance
//      lAccess         Access level of this object
//      uidStore
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CMsgStore::CMsgStore (CMSProvider * pMSProviderObj,
                      LPMAPISUP     pSupObj,
                      LPSTR         szServer,
                      SBinary       StoreEID,
                      CBackEnd *    pBEobj,
                      BOOL          fDeferErrors,
                      long          lAccess,
                      MAPIUID       uidStore) : CGenericProp (pBEobj,
                                                              NULL,
                                                              NULL,
                                                              NULL,
                                                              fDeferErrors,
                                                              lAccess)
{
    InfoTrace ("CMsgStore: Constructor called");
    m_pNext             = NULL;
    m_StoreEID.cb       = StoreEID.cb;
    m_StoreEID.lpb      = StoreEID.lpb;
    m_hInstance         = pMSProviderObj->m_hInstance;
    m_pSupObj           = pSupObj; // Don't AddRef it here
    m_pProvObj          = pMSProviderObj;
    m_pProvObj->AddRef();
    m_ulNotifAllObjects = 0;
    m_pNotifList        = NULL;
    m_StoreUID          = uidStore;

    lstrcpy (m_szServer, szServer);

    SPropValue spvOne = { 0 };
    spvOne.ulPropTag = PR_DISPLAY_NAME;
    spvOne.Value.LPSZ = WINDS_STORE_NAME;
    AddPropToList (&spvOne);

    HRESULT hResult = S_OK;
    m_pNPList = new CNamePropMgr (GetMAPIAllocBuffer(),
                                  GetMAPIAllocMore(),
                                  GetMAPIFreeBuffer());
    if (m_pNPList)
    {
        LPSTREAM pStream = NULL;
        hResult = pBEobj->GetNamedPropertiesStream (&pStream);
        if (!hResult)
        {
            hResult = m_pNPList->LoadMappingFromStream (pStream);
            pStream->Release();
        }
    }
    else
    {
        hResult = E_OUTOFMEMORY;
    }
    if (hResult)
    {
        if (m_pNPList)
        {
            delete m_pNPList;
            m_pNPList = NULL;
        }
        throw CException(hResult);
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::~CMsgStore()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CMsgStore.
//
//    Return Value
//      None
//
CMsgStore::~CMsgStore()
{
    InfoTrace ("CMsgStore: Destructor called");
    // Before the store object goes aways, we must call this method to let
    // the spooler finish transport session that may be using our store
    // provider.
    m_pSupObj->StoreLogoffTransports (&m_ulCachedStoreLogOffFlags);
    
    ASSERT (m_pProvObj);
    m_pProvObj->RemoveOpenStoreObject (this);
    m_pProvObj->Release();
    m_pProvObj = NULL;
    if (m_pNPList)
    {
        delete m_pNPList;
    }
    ASSERTMSG (m_OpenObjList.IsEmpty(), "The open object list is not empty!");
    ASSERTMSG (NULL == m_pNotifList, "The notification object list is not empty!");

    if (((CMAPISupport *)m_pSupObj)->m_cRef)
    {
        ASSERTMSG (1 == ((CMAPISupport *)m_pSupObj)->m_cRef, "A support object has been leaked!");
        m_pSupObj->Release();
        m_pSupObj = NULL;
    }
    
    PrivMAPIFree ((LPVOID *)&m_StoreEID.lpb);
    m_StoreEID.lpb      = NULL;
    m_StoreEID.cb       = 0;
    m_pNext             = NULL;
    m_hInstance         = NULL;
    m_lAccess           = 0;
    m_ulNotifAllObjects = 0;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      S_OK            If successful. With the interface pointer in *ppvObj
//      E_NOINTERFACE   If interface requested is not supported by this object
//
STDMETHODIMP CMsgStore::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IMsgStore || riid == IID_IMAPIProp || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetLastError()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Returns a MAPIERROR structure with an error description string about
//      the last error that occurred in the object.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetLastError (HRESULT       hError,
                                      ULONG         ulFlags,
                                      LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CMsgStore::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this,
                                     hError,
                                     ulFlags,
                                     ppMAPIError);
    return CGeneric::GetLastError (hError,
                                   ulFlags,
                                   ppMAPIError,
                                   m_hInstance,
                                   PROVIDER_DISPLAY_NAME_STRING);
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::SaveChanges()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. The store object itself is not transacted. Changes
//      occurs immediately without the need for committing changes.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMsgStore::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CMsgStore::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the value of the properties the client especified in the
//      property tag array. If the tag array is NULL, the user wants all
//      the properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetProps (LPSPropTagArray pPropTagArray,
                                  ULONG           ulFlags,
                                  ULONG *         pcValues,
                                  LPSPropValue *  ppPropArray)
{
    InfoTrace ("CMsgStore::GetProps method called");
    Validate_IMAPIProp_GetProps (this,
                                 pPropTagArray,
                                 ulFlags,
                                 pcValues,
                                 ppPropArray);
    #ifdef DO_INFO_TRACES
    if (!pPropTagArray)
    {
        InfoTrace ("CMsgStore::GetProps: All properties requested");
    }
    else
    {
        InfoTrace ("CMsgStore::Properties requested...");
        for (ULONG j=0; j<pPropTagArray->cValues; j++)
        { TraceProp (pPropTagArray->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES

    EnterCriticalSection (&m_csObj);
    BOOL fGotErrors = FALSE;
    PRIVATE_ENTRYID eid = { 0 };
    HRESULT hResult = CGenericProp::GetProps (pPropTagArray,
                                              ulFlags,
                                              pcValues,
                                              ppPropArray,
                                              (LPSPropTagArray)&sptMsgStoreObj);
    if (hResult)
    {
        goto ErrorExit;
    }

    LPVOID pSrc;
    ULONG cbSize, i;
    // Get the values for all the properties requested
    for (i=0; i<*pcValues; i++)
    {
        pSrc = NULL;
        hResult = S_OK;
        switch ((*ppPropArray)[i].ulPropTag)
        {
            // Get the PT_OBJECT/PR_NULL properties
            case PR_NULL:
            case PR_RECEIVE_FOLDER_SETTINGS:
                (*ppPropArray)[i].Value.err = S_OK;
                break;
            // Get the PT_LONG properties
            case PR_ACCESS:
                (*ppPropArray)[i].Value.l = m_lAccess;
                break;
            case PR_ACCESS_LEVEL:
                (*ppPropArray)[i].Value.l = (m_lAccess & MAPI_ACCESS_MODIFY ? MAPI_MODIFY : 0);
                break;
            case PR_OBJECT_TYPE:
                (*ppPropArray)[i].Value.l = MAPI_STORE;
                break;
            case PR_STORE_SUPPORT_MASK:
                (*ppPropArray)[i].Value.l = SUPPORT_MASK;
                break;
            case PR_VALID_FOLDER_MASK :
                (*ppPropArray)[i].Value.l = FOLDER_IPM_SUBTREE_VALID;
                break;
            case PR_FOLDER_TYPE:
                (*ppPropArray)[i].Value.l = FOLDER_ROOT;
                break;
            case PR_CONTENT_UNREAD:
                (*ppPropArray)[i].Value.l = 0;
                break;
            case PR_STORE_STATE:
                // If a provider has active searches (because it support Search Folders),
                // should set the STORE_HAS_SEARCHES bit in this property. If a store could
                // not determine wheather or not it has searches active, it should set the bit ON.
                // Since we don't support seaches yet, the bit is off.
                (*ppPropArray)[i].Value.l = 0;
                break;
            // Get the PT_BOOLEAN properties
            case PR_SUBFOLDERS:
                (*ppPropArray)[i].Value.b = TRUE;
                break;
            // Get the PT_I8 properties
            case PR_CURRENT_VERSION:
                (*ppPropArray)[i].Value.li.LowPart = MSG_STORE_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = MSG_STORE_VERSION_MINOR;
                break;
            // Get the PT_BINARY properties
            // In our store implementation the PR_MAPPING_SIGNATURE,
            // PR_STORE_RECORD_KEY and PR_RECORD_KEY and the same on the IMsgStore object
            // The value of the PR_MAPPING_SIGNATURE is the same in all of our objects
            // because the mapping of named properties is done at the provider level.
            case PR_MAPPING_SIGNATURE:
            case PR_STORE_RECORD_KEY:
            case PR_RECORD_KEY:
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&m_StoreUID;
                break;
            case PR_PARENT_ENTRYID :
            case PR_IPM_SUBTREE_ENTRYID :
                {
                    eid.uidGlobal = m_pBEobj->GetBackEndUID();
                    eid.bVersion = ENTRYID_VERSION;
                    eid.bObject = MAPI_FOLDER;
                    switch ((*ppPropArray)[i].ulPropTag)
                    {
                        case PR_PARENT_ENTRYID :
                            eid.dwObjID = 0;
                            break;
                        case PR_IPM_SUBTREE_ENTRYID :
                            eid.dwObjID = 1;
                            break;
                    }
                    cbSize = CB_PRIVATE_EID;
                    pSrc = (LPVOID)&eid;
                }
                break;
            case PR_ENTRYID:
            case PR_STORE_ENTRYID:
                cbSize = m_StoreEID.cb;
                pSrc = (LPVOID)m_StoreEID.lpb;
                break;
            case PR_MDB_PROVIDER:
                cbSize = sizeof(MAPIUID);
                pSrc = (LPVOID)&guidProvider;
                break;
            // Oh no, the property was not found. Try the list of properties
            default :
                hResult = CGenericProp::GetProps (&(*ppPropArray)[i], *ppPropArray);
                #ifdef DO_INFO_TRACES
                if (MAPI_E_NOT_FOUND == hResult)
                {
                    TraceRaw ("CMsgStore::GetProps: Not Found - ");
                    TraceProp ((*ppPropArray)[i].ulPropTag);
                }
                #endif // #DO_INFO_TRACES
                break;
        }
        if (S_OK != CGenericProp::GetProps (hResult,
                                            cbSize,
                                            pSrc,
                                            &(*ppPropArray)[i],
                                            *ppPropArray))
        {
            fGotErrors = TRUE;
        }
    }
ErrorExit :
    LeaveCriticalSection (&m_csObj);
    if (fGotErrors)
	{
		hResult = MAPI_W_ERRORS_RETURNED;
    }
    if (MAPI_W_ERRORS_RETURNED != hResult)
    {
        TraceResult ("CMsgStore::GetProps", hResult);
    }
    SetLastError (hResult);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetPropList()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Duplicate the tag array with ALL the available properties
//      on this object. Properties added with SetProps() will not be returned.
//      The must be explicitely requested in a GetProps() call.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetPropList (ULONG             ulFlags,
                                     LPSPropTagArray * ppAllTags)
{
    InfoTrace ("CMsgStore::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppAllTags);
    return CGenericProp::GetPropList (ulFlags, (LPSPropTagArray)&sptMsgStoreObj, ppAllTags);
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::OpenProperty()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't support opening properties directly on the
//      store object.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CMsgStore::OpenProperty (ULONG       ulPropTag,
                                      LPCIID      piid,
                                      ULONG       ulInterfaceOptions,
                                      ULONG       ulFlags,
                                      LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMsgStore::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this,
                                     ulPropTag,
                                     piid,
                                     ulInterfaceOptions,
                                     ulFlags,
                                     ppUnk);
    TraceResult ("CMsgStore::OpenProperty", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::SetProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Modify the value of the properties in our object. On IMsgStore
//      objects, changes to the properties are committed immediately.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::SetProps (ULONG                 cValues,
                                  LPSPropValue          pPropArray,
                                  LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMsgStore::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY))
    {
        TraceMessage ("CMsgStore::SetProps: Cannot set properties in a READ only object");
        TraceResult ("CMsgStore::SetProps", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }

    #ifdef DO_INFO_TRACES
    TraceMessage ("CMsgStore::Properties to set...");
    for (ULONG j=0; j<cValues; j++) { TraceProp (pPropArray[j].ulPropTag); }
    #endif DO_INFO_TRACES

    BOOL fObjectModified = FALSE;
    HRESULT hTagError, hResult = S_OK;
    LPPROFSECT pProfileObj = NULL;
    EnterCriticalSection (&m_csObj);
    
    for (ULONG i=0; i<cValues; i++)
    {
        switch (PROP_TYPE(pPropArray[i].ulPropTag))
        {
            case PT_NULL :
            case PT_UNSPECIFIED :
            case PT_ERROR :
                continue;
        }
        switch (pPropArray[i].ulPropTag)
        {
            // Our implementation does not allow clients to set IPM subtree entry ID because we
            // have a hard coded entry ID to it. The same goes for the rest of the special
            // folders (IPM subtree, views, etc.)
            case PR_COMMON_VIEWS_ENTRYID :
            case PR_FINDER_ENTRYID :
            case PR_IPM_SUBTREE_ENTRYID :
            case PR_IPM_OUTBOX_ENTRYID :
            case PR_IPM_SENTMAIL_ENTRYID :
            case PR_IPM_WASTEBASKET_ENTRYID :
                hTagError = MAPI_E_COMPUTED;
                break;
            case PR_COMMENT :
            case PR_DISPLAY_NAME :
                hTagError = E_ACCESSDENIED;
                break;
            default :
                hTagError = MAPI_E_NO_SUPPORT;
                break;
        }
        if (hTagError && ppProblems)
        {
            hResult = AddTagToProbArray (cValues,
                                         ppProblems,
                                         i,
                                         pPropArray[i].ulPropTag,
                                         hTagError);
            if (hResult)
            {
                break; // Out of the FOR() loop
            }
        }
    }
    if (pProfileObj)
    {
        pProfileObj->Release();
    }
    LeaveCriticalSection (&m_csObj);
    // If not error occurred, send a notification that the stored changed. We send the
    // notification after the releasing the critical section because we no longer need
    // exclusive access to the object's data members.
    if (!hResult && fObjectModified)
    {
        NOTIFICATION notif = { 0 };
        notif.ulEventType        = fnevObjectModified;
        notif.info.obj.cbEntryID = m_StoreEID.cb;
        notif.info.obj.lpEntryID = (LPENTRYID)m_StoreEID.lpb;
        notif.info.obj.ulObjType = MAPI_STORE;
        NotifyClients (NULL, 1, &notif);
    }
    TraceResult ("CMsgStore::SetProps", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::DeleteProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We do not allow clients to delete properties in the
//      IMsgStore object.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMsgStore::DeleteProps (LPSPropTagArray       pPropTagArray,
                                     LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMsgStore::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    TraceResult ("CMsgStore::DeleteProps", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::CopyTo()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      MAPI 1.0 does not require IMsgStore to support copying itself to
//      other store, so we always return MAPI_E_NO_SUPPORT.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CMsgStore::CopyTo (ULONG                 ciidExclude,
                                LPCIID                rgiidExclude,
                                LPSPropTagArray       pExcludeProps,
                                ULONG                 ulUIParam,
                                LPMAPIPROGRESS        pProgress,
                                LPCIID                pInterface,
                                LPVOID                pDestObj,
                                ULONG                 ulFlags,
                                LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMsgStore::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,   
                               rgiidExclude,  
                               pExcludeProps, 
                               ulUIParam,     
                               pProgress,     
                               pDestInterface,
                               pDestObj,      
                               ulFlags,       
                               ppProblems);
    TraceResult ("CMsgStore::CopyTo", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::CopyProps()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. We don't support copying our properties onto
//      another object.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always
//
STDMETHODIMP CMsgStore::CopyProps (LPSPropTagArray       pIncludeProps,
                                   ULONG                 ulUIParam,
                                   LPMAPIPROGRESS        pProgress,
                                   LPCIID                pInterface,
                                   LPVOID                pDestObj,
                                   ULONG                 ulFlags,
                                   LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CMsgStore::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pDestInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    TraceResult ("CMsgStore::CopyProps", MAPI_E_NO_SUPPORT);
    SetLastError (MAPI_E_NO_SUPPORT);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetNamesFromIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return the list of names for the identifier list supplied by the
//      caller. Our implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                         LPGUID            pPropSetGuid,
                                         ULONG             ulFlags,
                                         ULONG *           pcPropNames,
                                         LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CMsgStore::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    LPSPropTagArray pTags;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptMsgStoreObj, &pTags);
    if (!hResult)
    {
        hResult = CGenericProp::GetNamesFromIDs (ppPropTags,
                                                 pPropSetGuid,
                                                 ulFlags,
                                                 pcPropNames,
                                                 pppPropNames,
                                                 pTags);
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMsgStore::GetNamesFromIDs", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetIDsFromNames()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Return a property identifier for each named listed by the client. Our
//      IMsgStore implementation, defers to the common IMAPIProp
//      implementation to do the actual work.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetIDsFromNames (ULONG             cPropNames,
                                         LPMAPINAMEID *    ppPropNames,
                                         ULONG             ulFlags,
                                         LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CMsgStore::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    LPSPropTagArray pTags;
    LPSTREAM pStream;
    EnterCriticalSection (&m_csObj);
    HRESULT hResult = CGenericProp::GetPropList (0, (LPSPropTagArray)&sptMsgStoreObj, &pTags);
    if (!hResult)
    {
        hResult = m_pBEobj->GetNamedPropertiesStream (&pStream);
        if (!hResult)
        {
            hResult = CGenericProp::GetIDsFromNames (cPropNames,
                                                     ppPropNames,
                                                     ulFlags,
                                                     ppPropTags,
                                                     pTags,
                                                     pStream);
            pStream->Release();
        }
        PrivMAPIFree ((LPVOID *)&pTags);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CMsgStore::GetIDsFromNames", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
// IMsgStore virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::Advise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Add a notification node to the list of notification subscribers for
//      the object pointer in the entry ID passed in. In this implementation
//      we use the MAPI notification engine so we simply call MAPI to add a
//      new notification subscriber to the object. The token returned from
//      MAPI is save in our internal list of subscribers along with the
//      notification mask. This is used by us when our object want to
//      send notification about them.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::Advise (ULONG            cbEntryID,
                                LPENTRYID        pEntryID,
                                ULONG            ulEventMask,
                                LPMAPIADVISESINK pAdviseSink,
                                ULONG *          pulConnection)
{
    InfoTrace ("CMsgStore::Advise method called");
    Validate_IMsgStore_Advise (this,
                               cbEntryID,
                               pEntryID,
                               ulEventMask,
                               pAdviseSink,
                               pulConnection);
    PRIVATE_ENTRYID * pEID = (PRIVATE_ENTRYID *)pEntryID;
    HRESULT hResult = S_OK;
    *pulConnection = 0;
    DWORD cbSize;
    LPBYTE pKeyData;
    // If an entry ID was supplied and the size of it matches our internal entry ID structure
    // size, we assume the are trying to set a notification with a particular object in the store.
    // We must check the size to not be confuse by clients that passed the entry ID of the store
    // to set an advise with the store object. We handle that in the ELSE case below.
    if ((cbEntryID == CB_PRIVATE_EID) && pEID)
    {
        if (!m_pBEobj->IsValidEntryID (pEID))
        {
            hResult = MAPI_E_INVALID_ENTRYID;
            goto ErrorExit;
        }
        cbSize = CB_PRIVATE_EID;
        pKeyData = (LPBYTE)pEID;
    }
    else
    {
        // If the entry ID size is non-zero, it better be the entry ID of the store object.
        // Otherwise, we assume the client is giving us a bogus (to us) entry ID which
        // doesn't now mean anything in this store.
        if (cbEntryID)
        {
            if (cbEntryID != m_StoreEID.cb || memcmp (pEntryID, m_StoreEID.lpb, cbEntryID))
            {
                hResult = MAPI_E_INVALID_ENTRYID;
                goto ErrorExit;
            }
        }
        cbSize = sizeof(MAPIUID);
        pKeyData = (LPBYTE)&m_StoreUID;
        pEID = NULL;
    }
    LPNOTIFKEY pKey;
    hResult = AllocateNotifKey (cbSize, pKeyData, &pKey);
    if (hResult)
    {
        goto ErrorExit;
    }

    EnterCriticalSection (&m_csObj);
    // In our implementation of notifications in this store, we utilize the MAPI notification
    // engine. When we set an advise link for an object we must supply MAPI with a
    // UNIQUE key (an SBinary structure) that identifies that object. Since the entry IDs in
    // our store are unique for ALL the objects, the entry ID of the object will be
    // enough to satisfy MAPI. For the store object, we pass the provider UID as the key.
    hResult = m_pSupObj->Subscribe (pKey, ulEventMask, 0, pAdviseSink, pulConnection);
    PrivFree ((LPVOID *)&pKey);
    TraceResult ("CMsgStore::Advise: Failed to set internal advise link", hResult);
    if (!hResult)
    {
        if (pEID)
        {
            hResult = UpdateSubscriptionList (pEID, *pulConnection, TRUE);
            if (hResult)
            {
                m_pSupObj->Unsubscribe (*pulConnection);
            }
        }
        else
        {
            m_ulNotifAllObjects++;
        }
    }
    LeaveCriticalSection (&m_csObj);
ErrorExit:
    if (hResult)
    {
        *pulConnection = 0;
        TraceResult ("CMsgStore::Advise", hResult);
        SetLastError (hResult);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::Unadvise()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Terminates the advise link for a particular object. The connection
//      number passed in is given to the MAPI notification engine so that it
//      removes the connection that matches it. If the MAPI successfully
//      removes the connection, we remove it from our subscription list.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::Unadvise (ULONG ulConnection)
{
    InfoTrace ("CMsgStore::Unadvise method called");
    Validate_IMsgStore_Unadvise (this, ulConnection);
    HRESULT hResult = m_pSupObj->Unsubscribe (ulConnection);
    if (!hResult)
    {
        EnterCriticalSection (&m_csObj);
        
        UpdateSubscriptionList (NULL, ulConnection, FALSE);
        
        LeaveCriticalSection (&m_csObj);
    }
    TraceResult ("CMsgStore::Unadvise", hResult);
    SetLastError (hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::CompareEntryIDs()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function compares the two entry ID structures passed in and if
//      they are pointing to the same object, meaning all the members of the
//      entry ID structure are identical, it return TRUE. Otherwise false.
//
//    Return Value
//      An HRESULT. The comparison result is returned in the
//      pulResult argument.
//
STDMETHODIMP CMsgStore::CompareEntryIDs (ULONG     cbEntryID1,
                                         LPENTRYID pEntryID1,
                                         ULONG     cbEntryID2,
                                         LPENTRYID pEntryID2,
                                         ULONG     ulFlags,
                                         ULONG *   pulResult)
{
    InfoTrace ("CMsgStore::CompareEntryIDs method called");
    // There is not need to protect this method with a critical section
    Validate_IMsgStore_CompareEntryIDs (this,
                                        cbEntryID1,
                                        pEntryID1,
                                        cbEntryID2,
                                        pEntryID2,
                                        ulFlags,
                                        pulResult);
    
    PRIVATE_ENTRYID * pEID1 = (PRIVATE_ENTRYID *)pEntryID1;
    PRIVATE_ENTRYID * pEID2 = (PRIVATE_ENTRYID *)pEntryID2;
    HRESULT hResult = S_OK;
    __try
    {
        // If the size passed in, does not match the size of our own structures,
        // then don't bother comparing the members.
        if (cbEntryID1 != CB_PRIVATE_EID || cbEntryID2 != CB_PRIVATE_EID)
        {
            hResult = MAPI_E_UNKNOWN_ENTRYID;
            *pulResult = FALSE;
        }
        else
        {
            // All the fields in the private entry ID structure must be exaclty the same or there is no hit.
            if (pEID1 &&
                pEID2 &&
                cbEntryID1 == cbEntryID2 &&
                0 == memcmp (&pEID1->uidGlobal, &pEID2->uidGlobal, sizeof(MAPIUID)) &&
                pEID1->bVersion == pEID2->bVersion &&
                pEID1->bObject == pEID2->bObject &&
                pEID1->dwObjID == pEID2->dwObjID)
            {
                *pulResult = TRUE;
            }
            else
            {
                *pulResult = FALSE;
            }
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CMsgStore::CompareEntryIDs: An exception occured. Exception code: %#08x", GetExceptionCode());
        hResult = MAPI_E_INVALID_ENTRYID;
    }
    TraceResult ("CMsgStore::CompareEntryIDs", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::OpenEntry()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This method opens an object that exists in this message store. The object
//      could be a folder or a message in any subfolder.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::OpenEntry (ULONG       cbEntryID,
                                   LPENTRYID   pEntryID,
                                   LPCIID      pInterface,
                                   ULONG       ulFlags,
                                   ULONG *     pulObjType,
                                   LPUNKNOWN * ppUnk)
{
    InfoTrace ("CMsgStore::OpenEntry method called");
    Validate_IMsgStore_OpenEntry (this,
                                  cbEntryID,
                                  pEntryID,
                                  pInterface,
                                  ulFlags,
                                  pulObjType,
                                  ppUnk);
    if (!(m_lAccess & MAPI_ACCESS_MODIFY) && (ulFlags & MAPI_MODIFY))
    {
        TraceMessage ("CMsgStore::OpenEntry: Requested WRITE access in a READ only object");
        TraceResult ("CMsgStore::OpenEntry", E_ACCESSDENIED);
        SetLastError (E_ACCESSDENIED);
        return E_ACCESSDENIED;
    }
    // If the user requested MAPI_BEST_ACCESS, a provider should give the client the
    // maximum access level it can. In our object, if we are write-able, give the
    // user WRITE access. Is the easiest to handle for us.
    if (ulFlags & MAPI_BEST_ACCESS)
    {
        if (m_lAccess & MAPI_ACCESS_MODIFY)
        {
            ulFlags |= MAPI_MODIFY;
        }
    }
    PRIVATE_ENTRYID eidRoot = { 0 };
    eidRoot.uidGlobal = m_StoreUID;
    eidRoot.bVersion = ENTRYID_VERSION;
    eidRoot.bObject = MAPI_FOLDER;
    eidRoot.dwObjID = ROOT_FOLDER_ID;
    PRIVATE_ENTRYID * pEID = cbEntryID ? (PRIVATE_ENTRYID *)pEntryID : NULL;
    CMAPIFolder * pFldObj;
    CMessage * pMsgObj;
    CPropList * pList = NULL;
    HRESULT hResult = S_OK;
    if (cbEntryID)
    {
        if (!m_pBEobj->IsValidEntryID (pEID))
        {
            return MAPI_E_INVALID_ENTRYID;
        }
    }

    EnterCriticalSection (&m_csObj);

    long lAccess, lType = FOLDER_GENERIC;
    if (!pEID || pEID->bObject == MAPI_FOLDER) // This is the entry ID of the ROOT FOLDER { 0, NULL }
    {
        if (!pEID || ROOT_FOLDER_ID == pEID->dwObjID)
        {
            lType = FOLDER_ROOT;
            // Note that for the ROOT folder only, its parent entry ID is reflexive to its own entry ID.
            pEID = &eidRoot;
        }
    }
    if (pEID->bObject == MAPI_FOLDER || pEID->bObject == MAPI_MESSAGE)
    {
        hResult = m_pBEobj->OpenObject (*pEID, &pList);
    }
    else
    {   
        hResult = MAPI_E_NOT_FOUND;
    }
    PRIVATE_ENTRYID * pParentEID = NULL;
    if (!hResult)
    {
        SPropValue spvOne = { 0 };
        VERIFY (!pList->Find (PR_PARENT_ENTRYID, &spvOne));
        ASSERT (CB_PRIVATE_EID == spvOne.Value.bin.cb);
        pParentEID = (PRIVATE_ENTRYID *)spvOne.Value.bin.lpb;
    }
    if (!hResult && (pEID->bObject == MAPI_FOLDER))
    {
        // Grant access rights to the object according to what the user requested,
        // provided the store object can grant these.
        lAccess = (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                  (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                  MAPI_ACCESS_READ |
                  MAPI_ACCESS_CREATE_HIERARCHY |
                  MAPI_ACCESS_CREATE_CONTENTS |
                  MAPI_ACCESS_CREATE_ASSOCIATED;
        pFldObj = new CMAPIFolder (this,
                                   pParentEID,
                                   pEID,
                                   lType,
                                   pList,
                                   !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                   lAccess,
                                   m_pNPList);
        if (pFldObj)
        {
            pFldObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    if (!hResult && (pEID->bObject == MAPI_MESSAGE))
    {
        // Grant access rights to the object according to what the user requested,
        // provided the store object can grant these.
        lAccess = (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_MODIFY : 0) |
                  (ulFlags & MAPI_MODIFY ? MAPI_ACCESS_DELETE : 0) |
                  MAPI_ACCESS_READ;
        pMsgObj = new CMessage (this,
                                pParentEID,
                                pEID,
                                pList,
                                !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                lAccess,
                                NULL,
                                m_pNPList);
        if (pMsgObj)
        {
            pMsgObj->m_ulObjFlags |= ((m_ulObjFlags & OBJECT_RUNNING_IN_SPOOLER) ? OBJECT_RUNNING_IN_SPOOLER : 0);
        }
        else
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    LeaveCriticalSection (&m_csObj);
    
    if (hResult)
    {
        if (pList)
        {
            delete pList;
        }
        TraceResult ("CMsgStore::OpenEntry", hResult);
        SetLastError (hResult);
    }
    else
    {
        if (!pEID || MAPI_FOLDER == pEID->bObject)
        {
            *pulObjType = MAPI_FOLDER;
            *ppUnk = (LPUNKNOWN)pFldObj;
        }
        else
        {
            if (MAPI_MESSAGE == pEID->bObject)
            {
                *pulObjType = MAPI_MESSAGE;
                *ppUnk = (LPUNKNOWN)pMsgObj;
            }
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::SetReceiveFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::SetReceiveFolder (LPTSTR    szMessageClass,
                                          ULONG     ulFlags,
                                          ULONG     cbEntryID,
                                          LPENTRYID pEntryID)
{
    InfoTrace ("CMsgStore::SetReceiveFolder method called");
    Validate_IMsgStore_SetReceiveFolder (this,
                                         szMessageClass,
                                         ulFlags,
                                         cbEntryID,
                                         pEntryID);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetReceiveFolder()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetReceiveFolder (LPTSTR      szMessageClass,
                                          ULONG       ulFlags,
                                          ULONG *     pcbEntryID,
                                          LPENTRYID * ppEntryID,
                                          LPTSTR *    ppszExplicitClass)
{
    InfoTrace ("CMsgStore::GetReceiveFolder method called");
    Validate_IMsgStore_GetReceiveFolder (this,
                                         szMessageClass,
                                         ulFlags,
                                         pcbEntryID,
                                         ppEntryID,
                                         ppszExplicitClass);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetReceiveFolderTable()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetReceiveFolderTable (ULONG         ulFlags,
                                               LPMAPITABLE * ppTable)
{
    InfoTrace ("CMsgStore::GetReceiveFolderTable method called");
    Validate_IMsgStore_GetReceiveFolderTable (this, ulFlags, ppTable);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::StoreLogoff()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      This function is used to cleanup and close up any resources owned by
//      the store since the client plans to no longer use the message store.
//      In our implementation we simply save the flags in our object an return
//      immediately. We have no special cleanup to do.
//
//    Return Value
//      S_OK always
//
STDMETHODIMP CMsgStore::StoreLogoff (ULONG * pulFlags)
{
    InfoTrace ("CMsgStore::StoreLogoff method called");
    Validate_IMsgStore_StoreLogoff (this, pulFlags);
    EnterCriticalSection (&m_csObj);
    m_ulCachedStoreLogOffFlags = *pulFlags;    
    LeaveCriticalSection (&m_csObj);
    *pulFlags = LOGOFF_COMPLETE;
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::AbortSubmit()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      MAPI_E_UNABLE_TO_ABORT always
//
STDMETHODIMP CMsgStore::AbortSubmit (ULONG     cbEntryID,
                                     LPENTRYID pEntryID,
                                     ULONG     ulFlags)
{
    InfoTrace ("CMsgStore::AbortSubmit method called");
    CheckParameters_IMsgStore_AbortSubmit (this, cbEntryID, pEntryID, ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::GetOutgoingQueue()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::GetOutgoingQueue (ULONG ulFlags, LPMAPITABLE * ppTable)
{
    InfoTrace ("CMsgStore::GetOutgoingQueue method called");
    CheckParameters_IMsgStore_GetOutgoingQueue (this, ulFlags, ppTable);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::SetLockState()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::SetLockState (LPMESSAGE pMessageObj,
                                      ULONG     ulLockState)
{
    InfoTrace ("CMsgStore::SetLockState method called");
    CheckParameters_IMsgStore_SetLockState (this, pMessageObj, ulLockState);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::FinishedMsg()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::FinishedMsg (ULONG     ulFlags,
                                     ULONG     cbEntryID,
                                     LPENTRYID pEntryID)
{
    InfoTrace ("CMsgStore::FinishedMsg method called");
    CheckParameters_IMsgStore_FinishedMsg (this, ulFlags, cbEntryID, pEntryID);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::NotifyNewMail()
//
//    Parameters
//      { Refer to the MAPI documentation for parameter description }
//
//    Purpose
//      Stub method. This is not supported in a Public Folder store.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::NotifyNewMail (LPNOTIFICATION pNotification)
{
    InfoTrace ("CMsgStore::NotifyNewMail method called");
    CheckParameters_IMsgStore_NotifyNewMail (this, pNotification);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// Other member functions specific to this class
//

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::UpdateSubscriptionList()
//
//    Parameters
//      [IN]    pEID            Pointer to the entry ID of the object we just
//                              set a notification advise with.
//      [IN]    ulConnectionNum Connection number returned by MAPI.
//      [IN]    fAddToList      Flag to indicate if the object is being added
//                              to the list or removed from the list.
//
//    Purpose
//      Given an entry ID structure we add it to the list of objects clients
//      have notifications established with.
//
//    Return Value
//      S_OK            If successful updating the list.
//      E_OUTOFMEMORY   If we failed to allocate a new node.
//
STDMETHODIMP CMsgStore::UpdateSubscriptionList (PRIVATE_ENTRYID * pEID,
                                                ULONG             ulConnectionNum,
                                                BOOL              fAddToList)
{
    PNOTIF_OBJS pNode = m_pNotifList;
    if (!fAddToList)
    {
        PNOTIF_OBJS pPrev;
        // See if we have the object in our list already
        while (pNode)
        {
            if (ulConnectionNum == pNode->ulConnection)
            {
                if (pNode == m_pNotifList)
                {
                    // If the node we are deleting is the head of the list, reset the list head pointer.
                    m_pNotifList = m_pNotifList->pNext;
                }
                else
                {
                    // Take this node out from the list, and make the previous node
                    // point to the one in front of the current node.
                    pPrev->pNext = pNode->pNext;
                }
                PrivFree ((LPVOID *)&pNode);
                return S_OK;
            }
            pPrev = pNode;
            pNode = pNode->pNext;
        }
        return S_OK;
    }

    // Allocate a new node and add it to the list
    PNOTIF_OBJS pNewNode;
    HRESULT hResult = PrivAllocate (sizeof(NOTIF_OBJS), (LPVOID *)&pNewNode, "NOTIF Object List Node");
    if (hResult)
    {
        return hResult;
    }
    CopyMemory (&pNewNode->eidObject, pEID, sizeof (PRIVATE_ENTRYID));
    pNewNode->ulConnection = ulConnectionNum;
    pNewNode->pNext = NULL;
    InsertInList (&m_pNotifList, pNewNode);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::IsObjectInNotifList()
//
//    Parameters
//      [IN]    pEID    Pointer to the entry ID of the object we are going to
//                      search for in our list of objects.
//
//    Purpose
//      Given an entry ID structure search for it in the list of objects
//      clients have established notifications with.
//
//    Return Value
//      TRUE if the object is found in the list, FALSE if not.
//
BOOL WINAPI CMsgStore::IsObjectInNotifList (PRIVATE_ENTRYID * pEID)
{
    PNOTIF_OBJS pNode = m_pNotifList;
    while (pNode)
    {
        if (0 == memcmp (pEID, &pNode->eidObject, CB_PRIVATE_EID))
        {
            return TRUE;
        }
        pNode = pNode->pNext;
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::NotifyClients()
//
//    Parameters
//      [IN]    pEID            Pointer to the entry ID of the object sending
//                              the notification. If NULL, the notification
//                              is sent in behalf of the store.
//      [IN]    ulNotifsCount   Number of notifications in pNotifs.
//      [IN]    pNotifs         List of notification structures with all the
//                              information from the sending object.
//
//    Purpose
//      This function makes a request to the MAPI notification engine to send
//      notifications to all the clients that have an advise sink
//      established with the object indicated in pEID. If the object pointed
//      by pEID is not in the list of notifications, nothing is requested
//      to MAPI.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::NotifyClients (PRIVATE_ENTRYID * pEID,
                                       ULONG             ulNotifsCount,
                                       LPNOTIFICATION    pNotifs)
{
    LPNOTIFKEY pKey;
    HRESULT hResult = S_OK;
    ULONG ulFlags = fMapiUnicode;
    // First, check if there are notifications set at the STORE level.
    if (m_ulNotifAllObjects)
    {
        hResult = AllocateNotifKey (sizeof(MAPIUID), (LPBYTE)&m_StoreUID, &pKey);
        if (!hResult)
        {
            hResult = m_pSupObj->Notify (pKey, ulNotifsCount, pNotifs, &ulFlags);
            PrivFree ((LPVOID *)&pKey);
        }
    }
    // If an entry ID was passed in, see if that specific entry ID is
    // the list objects with notification links.
    if (pEID && !hResult)
    {
        if (IsObjectInNotifList (pEID))
        {
            hResult = AllocateNotifKey (CB_PRIVATE_EID, (LPBYTE)pEID, &pKey);
            if (!hResult)
            {
                hResult = m_pSupObj->Notify (pKey, ulNotifsCount, pNotifs, &ulFlags);
                PrivFree ((LPVOID *)&pKey);
            }
        }
    }
    TraceResult ("CMsgStore::NotifyClients: Failed to send notifications", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMsgStore::UpdateFolder()
//
//    Parameters
//      [IN]    fUpdateOutgoing     TRUE if the update is on the OUTBOX
//                                  folder. FALSE for incoming message and
//                                  the update is on the receive folder of
//                                  the incoming message's class.
//      [IN]    pParentEID          The entry ID of the folder where the
//                                  message is (RECEPTION) or was (SUBMISSION)
//      [IN]    pMsgEID             Entry ID of the message that will be
//                                  added or removed from the appropiate
//                                  contents table
//
//    Purpose
//      Update the content table and properties of the OUTBOX folder or
//      receive folder of a new incoming mail.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMsgStore::UpdateFolder (BOOL              fUpdateOutgoing,
                                      PRIVATE_ENTRYID * pParentEID, 
                                      PRIVATE_ENTRYID * pMsgEID)
{
    ASSERT (m_pBEobj->IsValidEntryID (pParentEID));
    ASSERT (m_pBEobj->IsValidEntryID (pMsgEID));
    NOTIFICATION notif = { 0 };
    EnterCriticalSection (&m_csObj);
    // Get the folder whose contents table we will update. If the folder
    // is not open then there is nothing to update.
    CMAPIFolder * pFolder = (CMAPIFolder *)m_OpenObjList.Walk (*pParentEID, WALK_START, NULL);
    if (pFolder)
    {
        // If the update is on submission, the OUTBOX folder is updated. The row
        // with the submitted message is deleted from the contents table.
        if (fUpdateOutgoing)
        {
            pFolder->UpdateContentTables (UPD_TABLE_DELETE_ROW,
                                          FALSE,                 // The message is NOT associated
                                          FALSE,                 // The message has been read
                                          pMsgEID,
                                          NULL);
            pFolder->UpdateParentTable();
            // The parent folder object changed, send a notification
            notif.ulEventType        = fnevObjectModified;
            notif.info.obj.cbEntryID = CB_PRIVATE_EID;
            notif.info.obj.lpEntryID = (LPENTRYID)pParentEID;
            notif.info.obj.ulObjType = MAPI_FOLDER;
            NotifyClients (pParentEID, 1, &notif);
        }
        else
        {
            pFolder->UpdateContentTables (UPD_TABLE_ADD_ROW,
                                          FALSE,                 // The message is NOT associated
                                          TRUE,                  // The message has not ben read
                                          pMsgEID,
                                          NULL);
            // The a new object was created (the new incoming message), send a notification
            notif.ulEventType         = fnevObjectCreated;
            notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
            notif.info.obj.lpEntryID  = (LPENTRYID)pMsgEID;
            notif.info.obj.cbParentID = CB_PRIVATE_EID;
            notif.info.obj.lpParentID = (LPENTRYID)&pFolder->m_ObjectEID;
            notif.info.obj.ulObjType  = MAPI_MESSAGE;
            NotifyClients (&pFolder->m_ObjectEID, 1, &notif);

            pFolder->UpdateParentTable();

            // The parent folder object changed, send a notification
            notif.ulEventType         = fnevObjectModified;
            notif.info.obj.cbEntryID  = CB_PRIVATE_EID;
            notif.info.obj.lpEntryID  = (LPENTRYID)&pFolder->m_ObjectEID;
            notif.info.obj.cbParentID = CB_PRIVATE_EID;
            notif.info.obj.lpParentID = (LPENTRYID)&pFolder->m_ParentEID;
            notif.info.obj.ulObjType  = MAPI_FOLDER;
            NotifyClients (&pFolder->m_ObjectEID, 1, &notif);            
        }
        pFolder->Release();
    }
    LeaveCriticalSection (&m_csObj);
    return S_OK;
}

// End of file for MSGSTORE.CPP