///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSPROV.CPP
//
//  Description
//      This file implements the IMSProvider interface with the methods 
//      specified in the MAPI SPI 1.0 specifications.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#define INITGUID
#define USES_IID_IMAPIProp
#define USES_IID_IMAPIPropData
#define USES_IID_IMAPIControl
#define USES_IID_IMSProvider
#define USES_IID_IMSLogon
#define USES_IID_IMsgStore
#define USES_IID_IMAPIContainer
#define USES_IID_IMAPIFolder
#define USES_IID_IMessage
#define USES_IID_IAttachment
#define USES_IID_IMAPITableData
#define USES_IID_IMAPITable
#define USES_IID_IStreamDocfile
#define USES_PS_PUBLIC_STRINGS
#define USES_PS_MAPI
#include "MSWDS.H"
#include "MSSRVBE.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::CMSProvider()
//
//    Parameters
//      hInst       Handle to instance of this MS DLL
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CMSProvider::CMSProvider (HINSTANCE hInst)
{
    InfoTrace ("CMSProvider: Constructor called");
    ASSERT (hInst);
    m_cRef = 1;
    m_hInstance = hInst;
    m_pMsgStoreList = NULL;
    InitializeCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::~CMSProvider()
//
//    Parameters
//      None
//
//    Purpose
//      Release any memory used by the members of this object.
//
//    Return Value
//      None
//
CMSProvider::~CMSProvider()
{
    InfoTrace ("CMSProvider: Destructor called");
    m_hInstance = NULL;
    ASSERT (NULL == m_pMsgStoreList);
    
    // Deallocate the resources for the critical section
    DeleteCriticalSection (&m_csObj);

    // Dump any memory leaks
    CloseMemoryHeap();

    // Close down the log file
    UnInitTraces();
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::QueryInterface()
//
//    Parameters
//      { Refer to the OLE documentation for parameter description }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it 
//      returns NULL.
//
//    Return Value
//      S_OK            If successful. With the interface pointer in *ppvObj
//      E_NOINTERFACE   If interface requested is not supported by this object
//
STDMETHODIMP CMSProvider::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If the interface requested is supported by this object, return a pointer
    // to ourselves, with the reference count incremented by one.
    if (riid == IID_IMSProvider || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IMSProvider virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::Shutdown()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      To terminate all session objects attached to this provider object
//
//    Return Value
//      S_OK always.
//
STDMETHODIMP CMSProvider::Shutdown (ULONG * pulFlags)
{
    InfoTrace ("\r\nCMSProvider::Shutdown method called");
    CheckParameters_IMSProvider_Shutdown (this, pulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::Logon()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      This method logs on the store by opening the message store database
//      and allocating the the IMsgStore and IMSLogon objects that are
//      returned to MAPI. During this function the provider gets the
//      configuration properties stored in its corresponding profile section.
//      This infomation include the location of the database file and other
//      options important at logon time.
//      If the client is trying to logon onto a message store database file
//      which has already been open then we don't bother creating new object
//      we simply return an AddRef()'ed pointer to the IMsgStore logon object
//      and a NULL IMSLogon object. This are the rules of MAPI.
//      If new objects are created they are added to the chain of open
//      message store objects in this IMSProvider.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSProvider::Logon (LPMAPISUP     pSupObj,
                                 ULONG         ulUIParam,
                                 LPTSTR        pszProfileName,
                                 ULONG         cbEntryID,
                                 LPENTRYID     pEntryID,
                                 ULONG         ulFlags,
                                 LPCIID        pInterface,
                                 ULONG *       pcbSpoolSecurity,
                                 LPBYTE *      ppbSpoolSecurity,
                                 LPMAPIERROR * ppMAPIError,
                                 LPMSLOGON *   ppMSLogon,
                                 LPMDB *       ppMDB)
{
    InfoTrace ("CMSProvider::Logon method called");
    CheckParameters_IMSProvider_Logon (this,
                                       pSupObj,
                                       ulUIParam,
                                       pszProfileName,
                                       cbEntryID,
                                       pEntryID,
                                       ulFlags,
                                       pInterface,
                                       pcbSpoolSecurity,
                                       ppbSpoolSecurity,
                                       ppMAPIError,
                                       ppMSLogon,
                                       ppMDB);

    // This trivial wrapper of IMAPISupport is here for debugging purposes. This is not
    // required when implementing a message store provider
    CMAPISupport * pPrivSupObj = new CMAPISupport (pSupObj);
    if (NULL == pPrivSupObj)
    {
        TraceMessage ("CMSProvider::Logon: Failed to allocate Private Support Object");
        return E_OUTOFMEMORY;
    }
    
    // This is used to return a valid MAPIERROR structure in case of a failure
    CGeneric obj;
    
    // NULL this parameters
    *ppMAPIError = NULL;
    *ppbSpoolSecurity = NULL;
    *pcbSpoolSecurity = 0;

    LPSPropValue pProps = NULL;
    CBackEnd * pBEobj        = NULL;
    CMsgStore * pMsgStoreObj = NULL;
    CMSLogon * pMSLogonObj   = NULL;
    SBinary StoreEID         = { 0 };
    MAPIUID uidResource      = { 0 };
    long lAccess = (MDB_WRITE ? MAPI_ACCESS_MODIFY : 0) |
                   (MDB_WRITE ? MAPI_ACCESS_DELETE : 0) |
                    MAPI_ACCESS_READ |
                    MAPI_ACCESS_CREATE_HIERARCHY |
                    MAPI_ACCESS_CREATE_CONTENTS |
                    MAPI_ACCESS_CREATE_ASSOCIATED;

    // Since we are going to allocate a new object and manipulate the list of
    // sessions, lock the provider until we are finished
    EnterCriticalSection (&m_csObj);

    // The configuration options from the profile section
    HRESULT hResult = GetProfileProps (pPrivSupObj, &pProps, (HWND)ulUIParam);
    if (hResult)
    {
        goto LogonFinished;
    }
    // If we can piggyback on an existing open message store on this session
    // then do so. If we piggyback, then don't return a logon object or
    // addref the support object
    if (IsStoreAlreadyOpen (pProps[SERVER_NAME].Value.lpszA, &pMsgStoreObj))
    {
        goto LogonFinished;
    }

    ASSERT (PR_SMP_REMOTE_SERVER == pProps[SERVER_NAME].ulPropTag);
    uidResource = guidProvider;

    // Open the coonection with the public folder server and get
    // back an CBackEnd object for uniform access to data sources.
    hResult = OpenRemoteDataBase (pProps[SERVER_NAME].Value.lpszA,
                                  &uidResource,
                                  &pBEobj);
    if (!hResult)
    {
        // Create a MAPI-wrapped store entry ID with this' instance information
        hResult = CreateStoreEntryID (&uidResource, &StoreEID);
    }
    if (hResult)
    {
        goto LogonFinished;
    }
    try
    {
        // Allocate an instance of an IMsgStore object
        pMsgStoreObj = new CMsgStore (this,
                                      pPrivSupObj,
                                      pProps[SERVER_NAME].Value.lpszA,
                                      StoreEID,
                                      pBEobj,
                                      !!(ulFlags & MAPI_DEFERRED_ERRORS),
                                      lAccess,
                                      uidResource);
        if (!pMsgStoreObj)
        {
            hResult = E_OUTOFMEMORY;
        }
    }
    catch (CException & Exception)
    {
        hResult = Exception.GetError();
    }
    if (hResult)
    {
        goto LogonFinished;
    }
    // This is the UID of the entry IDs in our message store objects (IMAPIFolder, IMessage, etc.)
    hResult = pPrivSupObj->SetProviderUID (&uidResource, 0);
    if (!hResult)
    {
        // Allocate a new instance of an IMSLogon object
        pMSLogonObj = new CMSLogon (pMsgStoreObj);
        if (NULL == pMSLogonObj)
        {
            hResult = E_OUTOFMEMORY;
        }
        else
        {
            // Modify our profile section so that we appear in the session store table
            pPrivSupObj->ModifyProfile (0);
            // Add this object to the list of opened message stores
            AddStoreObjectToList (pMsgStoreObj);
        }
    }

LogonFinished :
    // Unlock the critical section
    LeaveCriticalSection (&m_csObj);
    if (pBEobj)
    {
        pBEobj->Release();
    }
    PrivMAPIFree ((LPVOID *)&pProps);

    // If we failed cleanup before leaving
    if (hResult)
    {
        TraceResult ("CMSProvider::Logon", hResult);
        if (pMsgStoreObj)
        {
            pMsgStoreObj->Release();
        }
        if (pMSLogonObj)
        {
            pMSLogonObj->Release();
        }
        PrivMAPIFree ((LPVOID *)&StoreEID.lpb);
        obj.SetLastError (hResult);
        obj.GetLastError (hResult, fMapiUnicode, ppMAPIError, m_hInstance, PROVIDER_DISPLAY_NAME_STRING);

        delete pPrivSupObj;
    }
    else
    {
        // We were successful, so copy the objects into the returning parameters
        ASSERT (pMsgStoreObj);
        // If the logon object that we are supposed to return still NULL, then it
        // means that we are piggybacking on an open message store, so the private
        // support object that we use must be deleted.
        if (NULL == pMSLogonObj)
        {
            delete pPrivSupObj;
        }
        else
        {
            // This call has opened a brand new message store.
            // AddRef() the support object now that we are successful
            pPrivSupObj->AddRef();
            pMSLogonObj->InitializeStatusRow();
        }
        *ppMSLogon = (LPMSLOGON)pMSLogonObj;
	    *ppMDB = (LPMDB)pMsgStoreObj;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::SpoolerLogon()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      Do the same as CMSProvider::Logon() but in the MAPI spooler context
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSProvider::SpoolerLogon (LPMAPISUP     pSupObj,
                                        ULONG         ulUIParam,
                                        LPTSTR        pszProfileName,
                                        ULONG         cbEntryID,
                                        LPENTRYID     pEntryID,
                                        ULONG         ulFlags,
                                        LPCIID        pInterface,
                                        ULONG         cbSpoolSecurity,
                                        LPBYTE        pbSpoolSecurity,
                                        LPMAPIERROR * ppMAPIError,
                                        LPMSLOGON *   ppMSLogon,
                                        LPMDB *       ppMDB)
{
    InfoTrace ("CMSProvider::SpoolerLogon method called");
    CheckParameters_IMSProvider_SpoolerLogon (this,
                                              pSupObj,
                                              ulUIParam,
                                              pszProfileName,
                                              cbEntryID,
                                              pEntryID,
                                              ulFlags,
                                              pInterface,
                                              cbSpoolSecurity,
                                              pbSpoolSecurity,
                                              ppMAPIError,
                                              ppMSLogon,
                                              ppMDB);

    // A public folder store does not have to support the
    // IMSProvider::SpoolerLogon method because we don't
    // support submitting messages. The messages on a PF
    // store are "posted" which means a client set
    // properties on a message and calls
    // IMessage::SaveChanges to save it in the folder
    // where the message is "posted"
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::CompareStoreIDs()
//
//    Parameters
//      { Refer to the MAPI documentation parameter description }
//
//    Purpose
//      Give two entry IDs we compare them to find out if they are the
//      same. (i.e. they refer to the same object)
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSProvider::CompareStoreIDs (ULONG     cbEntryID1,
                                           LPENTRYID pEntryID1,
                                           ULONG     cbEntryID2,
                                           LPENTRYID pEntryID2,
                                           ULONG     ulFlags,
                                           ULONG *   pulResult)
{
    InfoTrace ("CMSProvider::CompareStoreIDs method called");
    CheckParameters_IMSProvider_CompareStoreIDs  (this,
                                                  cbEntryID1,
                                                  pEntryID1,
                                                  cbEntryID2,
                                                  pEntryID2,
                                                  ulFlags,
                                                  pulResult);
    PRIVATE_ENTRYID * pEID1 = (PRIVATE_ENTRYID *)pEntryID1;
    PRIVATE_ENTRYID * pEID2 = (PRIVATE_ENTRYID *)pEntryID2;
    HRESULT hResult = S_OK;
    __try
    {
        // If the size passed in, does not match the size of our own structures,
        // then don't bother comparing the members.
        if (cbEntryID1 != CB_PRIVATE_EID || cbEntryID2 != CB_PRIVATE_EID)
        {
            hResult = MAPI_E_UNKNOWN_ENTRYID;
            *pulResult = FALSE;
        }
        else
        {
            // All the fields in the private entry ID structure must be exaclty the same or there is no hit.
            if (pEID1 &&
                pEID2 &&
                cbEntryID1 == cbEntryID2 &&
                0 == memcmp (&pEID1->uidGlobal, &pEID2->uidGlobal, sizeof(MAPIUID)) &&
                pEID1->bVersion == pEID2->bVersion &&
                pEID1->bObject == pEID2->bObject &&
                pEID1->dwObjID == pEID2->dwObjID)
            {
                *pulResult = TRUE;
            }
            else
            {
                *pulResult = FALSE;
            }
        }
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CMSProvider::CompareStoreIDs: An exception occured. Exception code: %#08x", GetExceptionCode());
        hResult = MAPI_E_INVALID_ENTRYID;
    }
    TraceResult ("CMSProvider::CompareStoreIDs", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
// Other member functions specific to this class
//

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::GetProfileProps()
//
//    Parameters
//      [IN]    pMAPISupObj     Pointer to the MAPI support object
//      [OUT]   ppProps         Where we return the logon properties for
//                              the provider
//      [IN]    hOwnerWnd       Handle to the windows that will own
//                              any UI required
//
//    Purpose
//      Retrieve the logon/configuration properties for this provider.
//      These properties are stored in the profile section allocated for the
//      message store in the current profile.
//      If the properties are not where they should be because the service
//      has not been configured, we display the configuration dialog for
//      the provider.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CMSProvider::GetProfileProps (LPMAPISUP      pMAPISupObj,
                                           LPSPropValue * ppProps,
	                                       HWND           hOwnerWnd)
{
    *ppProps = NULL;
    // Get information for the provider's profile section
    LPPROFSECT pProfileObj;
    HRESULT hResult = OpenServiceProfileSection (pMAPISupObj, &pProfileObj, GetMAPIFreeBuffer());
    if (hResult)
    {
        TraceResult ("GetProfileProps: OpenServiceProfileSection failed", hResult);
        return hResult;
    }

    ULONG ulPropCount;
    hResult = pProfileObj->GetProps ((LPSPropTagArray)&sptLogonProps,
                                     fMapiUnicode,
                                     &ulPropCount,
                                     ppProps);
    if (SUCCEEDED(hResult))
    {
        // In case we get MAPI_W_ERRORS_RETURNED display the logon configuration
        // dialog to get the missing properties
        if (MAPI_W_ERRORS_RETURNED == hResult)
        {
            hResult = S_OK;
            if (PR_SMP_REMOTE_SERVER != (*ppProps)[SERVER_NAME].ulPropTag)
            {
                hResult = MAPI_E_UNCONFIGURED;
            }
            if (!hResult)
            {
                // The CfgDialog is passed to our private function to display
                // the configuration page for the message store
                CFGDLG CfgDialog;
                // Fill in the logon UI structure
                CfgDialog.hInst   = m_hInstance;
                CfgDialog.hWnd    = hOwnerWnd;
                CfgDialog.ppProps = ppProps;
                CfgDialog.pSupObj = pMAPISupObj;
                CfgDialog.ulFlags = 0;

                hResult = DoLogonDlg (&CfgDialog);
                if (hResult)
                {
                    if (hResult == MAPI_E_USER_CANCEL)
                    {
                        hResult = MAPI_E_UNCONFIGURED;
                    }
                }
                else
                {
                    hResult = pProfileObj->SetProps (ulPropCount, *ppProps, NULL);
                    if (!hResult)
                    {
                        hResult = pProfileObj->SaveChanges (0);
                    }
                }
            }
        }
    }
    pProfileObj->Release();
    TraceResult ("CMSProvider::GetProfileProps", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::AddStoreObjectToList()
//
//    Parameters
//      [IN]    pNewMsgStoreObj     Pointer to the IMsgStore object we
//                                  will insert in the list
//
//    Purpose
//      Insert the newly instatiated IMsgStore object to the list
//      of open objects
//
//    Return Value
//      None
//
void WINAPI CMSProvider::AddStoreObjectToList (CMsgStore * pNewMsgStoreObj)
{
    ASSERT (pNewMsgStoreObj);
    CMsgStore * pStore;
    if (m_pMsgStoreList)
    {
        pStore = m_pMsgStoreList;
        while (pStore)
        {
            if (NULL == pStore->m_pNext)
            {
                pStore->m_pNext = pNewMsgStoreObj;
                break; // out of the WHILE() loop
            }
            pStore = pStore->m_pNext;
        }
    }
    else
    {
        m_pMsgStoreList = pNewMsgStoreObj;
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::RemoveOpenStoreObject()
//
//    Parameters
//      [IN]    pStore  Pointer to the IMsgStore object that we will remove
//
//    Purpose
//      Remove the specified IMsgStore object from the list of running
//      session maintained by the IMSProvider
//
//    Return Value
//      None
//
void WINAPI CMSProvider::RemoveOpenStoreObject (CMsgStore * pStore)
{
    EnterCriticalSection (&m_csObj);
    CMsgStore *pPrevStore, *pCurrStore = m_pMsgStoreList;
    while (pCurrStore)
    {
        if (pCurrStore == pStore)
        {
            if (pCurrStore == m_pMsgStoreList)
            {
                // If the node we are deleting is the head of the list, reset the list head pointer.
                m_pMsgStoreList = m_pMsgStoreList->m_pNext;
            }
            else
            {
                // Take this node out from the list, and make the previous node
                // point to the one in front of the current node.
                pPrevStore->m_pNext = pCurrStore->m_pNext;
            }
            break; // Out of the WHILE loop
        }
        pPrevStore = pCurrStore;
        pCurrStore = pCurrStore->m_pNext;
    }
    LeaveCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CMSProvider::IsStoreAlreadyOpen()
//
//    Parameters
//      [IN]    szServer        Path to the WINDS Public Folder server
//      [OUT]   ppStoreObj      Where we return the available IMsgStore object
//
//    Purpose
//      Determines if a message store database file is already opened by a
//      session owned by the IMSProvider object. If the file is opened and
//      the IMsgStore object exists, we return it, AddRef() so that MAPI
//      piggybacks on this session as oppossed to instatiating a new object.
//
//    Return Value
//      TRUE if the database is already open and we have a store object
//      to return. FALSE otherwise.
//
BOOL WINAPI CMSProvider::IsStoreAlreadyOpen (LPSTR        szServer,
                                             CMsgStore ** ppStoreObj)
{
    *ppStoreObj = NULL;
    CMsgStore * pStore = m_pMsgStoreList;
    while (pStore)
    {
        if (0 == lstrcmpi (szServer, pStore->m_szServer))
        {
            *ppStoreObj = pStore;
            (*ppStoreObj)->AddRef();
            return TRUE;
        }
		pStore = pStore->m_pNext;
    }
    return FALSE;
}

// End of File for MSPROV.CPP