///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSSRVBE.CPP
//
//  Description
//      Implements the CPFServerBackEnd class which derives from CBackEnd.
//      The methods in this class are used by our address book implementation
//      to request the data from the remote server host to fill contents and
//      hierarchy tables, and request the property of object in the messaging
//      host.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "MSWDS.H"
#include "MSSRVBE.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceMessage(a)
#define InfoTraceRaw(a)     TraceRaw(a)
#define InfoTraceProp(a)    TraceProp(a)
#else
#define InfoTrace(a)        
#define InfoTraceRaw(a)     
#define InfoTraceProp(a)    
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    OpenRemoteDataBase()
//
//    Paremeters
//      pszServer           Name of the server the back end connects to
//      pUID                UID of the entry ID in returned by the back end
//      ppBackEnd           Where we return the new CBackEnd object
//
//    Purpose
//      Create a instance of a CBackEnd object returned to the public folder
//      store. This object is used to access the server and retrieve the
//      data. All the code for the Remote procedure call and protocol of
//      the server is contained in this class implementation.
//      CBackEnd methods are also used in my generic IMAPIProp implementation
//      (CGenericProp) to open object and request stuff from
//      the data storage back end (file, server, etc)
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI OpenRemoteDataBase (LPTSTR       pszServer,
                                   MAPIUID *    pUID,
                                   CBackEnd **  ppBackEnd)
{
    // Allocate the new database object. The constructor initializes internal data in the object.
    CPFServerBackEnd * pSBEobj = new CPFServerBackEnd (pUID, pszServer);
    if (pSBEobj)
    {
        *ppBackEnd = pSBEobj;
        return S_OK;
    }
    TraceMessage ("OpenRemoteDataBase: Failed to allocate a new back end object");
    return E_OUTOFMEMORY;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::CPFServerBackEnd()
//
//    Parameters
//      pUIDinFile          UID that identifies the entry ID returned from
//                          this backend storage (the server)
//      pszRemoteServer     Name of the remote server we are connecting to
//      
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CPFServerBackEnd::CPFServerBackEnd (MAPIUID * pUIDinFile,
                                    LPTSTR    szRemoteServer) : CBackEnd (*pUIDinFile)
{
    InfoTrace ("CPFServerBackEnd: Constructor called");
    m_cRef = 1;
    ASSERT (szRemoteServer);
    lstrcpy (m_szServer, szRemoteServer);
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::~CPFServerBackEnd()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CABServerBackEnd.
//
//    Return Value
//      None
//
CPFServerBackEnd::~CPFServerBackEnd()
{
    InfoTrace ("CPFServerBackEnd: Destructor called");
    // Terminate connections with the remote server
    BindToServer (NULL);
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::GetNamedPropertiesStream()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::GetNamedPropertiesStream (LPSTREAM * ppStream)
{
    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceString1 ("CPFServerBackEnd::GetNamedPropertiesStream: Failed to bind to remote server %s", m_szServer);
        return hResult;
    }
    hResult = CreateStreamOnHGlobal (NULL, TRUE, ppStream);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::GetNamedPropertiesStream: Failed to create stream in memory", hResult);
        return hResult;
    }

    DWORD dwBytesRead;
    long lPipeNumber;
    HANDLE hPipe;
    TCHAR szPipeName[64];
    char szBuffer[IO_BUFFERSIZE];
    RpcTryExcept
    {
        hResult = RemotePFGetNamePropStream (0, &lPipeNumber);
        if (S_OK == hResult)
        {
            wsprintf (szPipeName, PIPE_NAME_FORMAT, m_szServer, lPipeNumber);
            hPipe = CreateFile (szPipeName,
                                GENERIC_READ,
                                0,
                                NULL,
                                OPEN_EXISTING,
                                0,
                                NULL);
            if (INVALID_HANDLE_VALUE == hPipe)
            {
                hResult = HRESULT_FROM_WIN32(GetLastError());
                TraceResult ("CPFServerBackEnd::GetNamedPropertiesStream: Failed to open pipe to server", hResult);
            }
            else
            {
                do
                {
                    // Read the pipe handle.
                    if (!ReadFile (hPipe, szBuffer, IO_BUFFERSIZE, &dwBytesRead, NULL))
                    {
                        hResult = HRESULT_FROM_WIN32(GetLastError());
                        if (HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)   != hResult &&   // For Windows NT
                            HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED) != hResult)     // For Windows 95
                        {
                            // There was an error and we can't continue
                            TraceResult ("CPFServerBackEnd::GetNamedPropertiesStream: Failed to read from the pipe", hResult);
                        }
                        else
                        {
                            // If the pipe was broken, it means the server finished writing
                            // to the it, so we are finished reading from it.
                            hResult = S_OK;
                        }
                    }
                    if (dwBytesRead)
                    {
                        (*ppStream)->Write (szBuffer, dwBytesRead, NULL);
                    }
                } while (dwBytesRead && !hResult);
                CloseHandle (hPipe);
                LARGE_INTEGER liZero = { 0 };
                (*ppStream)->Seek (liZero, STREAM_SEEK_SET, NULL);
            }
        }
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::GetNamedPropertiesStream: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    if (hResult)
    {
        (*ppStream)->Release();
        *ppStream = NULL;
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::CreateObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::CreateObject (PRIVATE_ENTRYID   ParentEID,
                                             PRIVATE_ENTRYID * pNewObjEID,
                                             CPropList **      ppList,
                                             ULONG             ulObjType,
                                             LPVOID            pData1,
                                             LPVOID            pData2)
{
    if (ulObjType != MAPI_MESSAGE && ulObjType != MAPI_FOLDER && ulObjType != MAPI_ATTACH)
    {
        TraceMessage ("CPFServerBackEnd::CreateObject: Invalid object to create");
        return MAPI_E_NO_SUPPORT;
    }
    if (!IsValidEntryID (&ParentEID))
    {
        TraceMessage ("CPFServerBackEnd::CreateObject: Invalid parent entry ID");
        return MAPI_E_INVALID_ENTRYID;
    }
    // Initialize the entry ID of the new object.
    ZeroMemory (pNewObjEID, CB_PRIVATE_EID);
    pNewObjEID->uidGlobal = m_uidBackEnd;
    pNewObjEID->bVersion  = ENTRYID_VERSION;
    pNewObjEID->bObject   = (BYTE)ulObjType;

    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::CreateObject", hResult);
        return hResult;
    }
    RpcTryExcept
    {
        if (MAPI_MESSAGE == ulObjType)
        {
            // Messages in our PF store will not a valid entry ID until
            // we return from the first SaveChanges();
            hResult = InitMessageProps ((BOOL)pData2, ParentEID, (LPBYTE)pData1, ppList);
        }
        if (MAPI_ATTACH == ulObjType)
        {
            //hResult = InitAttachProps (ppList);
            hResult = MAPI_E_NO_SUPPORT;
        }
        if (MAPI_FOLDER == ulObjType)
        {
            PUBLIC_FOLDER_INFO_A pf = { 0 };
            pf.dwParentID = ParentEID.dwObjID;
            lstrcpy ((LPSTR)pf.szFldName, (LPSTR)pData1);
            if (pData2)
            {
                lstrcpy ((LPSTR)pf.szFldComment, (LPSTR)pData2);
            }
            GetSystemTimeAsFileTime ((FILETIME *)&pf.ftCreateTime);
            pf.ftLastModTime = pf.ftCreateTime;
            hResult = InitFolderProps ((LPSTR)pf.szFldName,
                                       (LPSTR)pf.szFldComment,
                                       ParentEID,
                                       ppList);
            if (!hResult)
            {
                hResult = RemotePFCreateFolder (&pf);
                pNewObjEID->dwObjID = pf.dwObjID;
                if (HRESULT_FROM_WIN32(ERROR_ALREADY_EXISTS) == hResult)
                {
                    hResult = MAPI_E_COLLISION;
                }
            }
        }
    }
    RpcExcept(1)
    {
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::CreateObject: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    TraceResult ("CPFServerBackEnd::CreateObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::OpenObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::OpenObject (PRIVATE_ENTRYID ObjectEID,
                                           CPropList **    ppList)
{
    PUBLIC_FOLDER_INFO_A pf = { 0 };
    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::OpenObject", hResult);
        return hResult;
    }
    RpcTryExcept
    {
        if (MAPI_FOLDER == ObjectEID.bObject)
        {
            hResult = RemotePFGetFolderProps (ObjectEID.dwObjID, &pf);
        }
        else
        {
            ASSERT (FALSE); // TODO:
        }
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::OpenObject: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    if (!hResult)
    {
        DWORD i = 0;
        SPropValue spvProps[16];
        PRIVATE_ENTRYID ParentEID = { 0 };
        ParentEID.uidGlobal = m_uidBackEnd;
	    ParentEID.bVersion = ENTRYID_VERSION;
        if (MAPI_FOLDER == ObjectEID.bObject)
        {
            ParentEID.bObject = MAPI_FOLDER;
            ParentEID.dwObjID = pf.dwParentID;
            spvProps[i].ulPropTag = PR_PARENT_ENTRYID;
            spvProps[i].Value.bin.cb = CB_PRIVATE_EID;
            spvProps[i++].Value.bin.lpb = (LPBYTE)&ParentEID;
            spvProps[i].ulPropTag = PR_CONTENT_COUNT;
            spvProps[i++].Value.l = pf.dwCount;
            spvProps[i].ulPropTag = PR_CONTENT_UNREAD;
            spvProps[i++].Value.l = pf.dwUnread;
            spvProps[i].ulPropTag = PR_ASSOC_CONTENT_COUNT;
            spvProps[i++].Value.l = pf.dwAssoc;
            spvProps[i].ulPropTag = PR_SUBFOLDERS;
            spvProps[i++].Value.b = (BOOL)pf.fHasSubFolders;
            spvProps[i].ulPropTag = PR_DISPLAY_NAME;
            spvProps[i++].Value.lpszA = (LPSTR)pf.szFldName;
            spvProps[i].ulPropTag = PR_COMMENT;
            spvProps[i++].Value.lpszA = (LPSTR)pf.szFldComment;
            spvProps[i].ulPropTag = PR_CREATION_TIME;
            CopyMemory (&spvProps[i].Value.ft, &pf.ftCreateTime, sizeof(FILETIME));
            i++;
            spvProps[i].ulPropTag = PR_LAST_MODIFICATION_TIME;
            CopyMemory (&spvProps[i].Value.ft, &pf.ftLastModTime, sizeof(FILETIME));
            i++;
            spvProps[i].ulPropTag = PR_STATUS;
            spvProps[i++].Value.l = 0;
        }
        else
        {
            ASSERT (FALSE); // TODO:
        }
        hResult = SetPropsInList (i, spvProps, ppList);
    }
    TraceResult ("CPFServerBackEnd::OpenObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::DeleteOneObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::DeleteOneObject (PRIVATE_ENTRYID ObjectEID,
                                                PRIVATE_ENTRYID ParentEID)
{
    PUBLIC_FOLDER_INFO_A pf = { 0 };
    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::DeleteOneObject", hResult);
        return hResult;
    }
    RpcTryExcept
    {
        if (MAPI_FOLDER == ObjectEID.bObject)
        {
            hResult = RemotePFDeleteFolder (ObjectEID.dwObjID);
        }
        else
        {
            ASSERT (FALSE); // TODO:
        }
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::DeleteOneObject: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    TraceResult ("CPFServerBackEnd::DeleteOneObject", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::DeleteObjects()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::DeleteObjects (PRIVATE_ENTRYID   ContainerEID,
                                              ULONG             ulDeleteFlags,
                                              LPMAPIPROGRESS    pProgress,
                                              POPENOBJECTS_DATA pData)
{
    HRESULT hResult = S_OK;
    // TODO:
    TraceResult ("CPFServerBackEnd::DeleteObjects", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::GetTable()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::GetTable (const TABLE_TYPE  tblType,
										 PRIVATE_ENTRYID   ContainerEID,
										 long              lAccess,
										 LPVOID            pData1,
										 LPVOID            pData2,
										 LPTABLEDATA *     ppTable)
{
    LPSPropTagArray pDefaultColumnSet;
    ULONG ulIndexProp = PR_INSTANCE_KEY;
    HRESULT hResult;
    if (!IsValidEntryID (&ContainerEID))
    {
        hResult = MAPI_E_INVALID_ENTRYID;
        goto ErrorExit;
    }
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            pDefaultColumnSet = (LPSPropTagArray)&sptFolderHierProps;
            break;
        case TABLE_CONTENTS :
            pDefaultColumnSet = (LPSPropTagArray)&sptFolderContProps;
            break;
        default :
            hResult = MAPI_E_NO_SUPPORT;
            goto ErrorExit;
    }
    
    // Before locking the database, create a new In-Memory table
    hResult = CreateTable (&IID_IMAPITableData,
                           GetMAPIAllocBuffer(),
                           GetMAPIAllocMore(),
                           GetMAPIFreeBuffer(),
                           NULL,
                           TBLTYPE_DYNAMIC,
                           ulIndexProp,
                           pDefaultColumnSet,
                           ppTable);
    if (hResult)
    {
        goto ErrorExit;
    }
	switch (tblType)
    {
        case TABLE_HIERARCHY :
            hResult = FillHierarchyTable (ContainerEID,
                                          *ppTable,
                                          (LPSBinary)pData1,
                                          lAccess);
            break;
        case TABLE_CONTENTS :
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            if (FALSE == (BOOL)pData2)
            {
                hResult = FillContentsTable (ContainerEID,
                                             *ppTable,
                                             (LPSBinary)pData1,
                                             lAccess);
            }
            break;
    }
    if (hResult)
    {
        (*ppTable)->Release();
        *ppTable = NULL;
    }
ErrorExit:
    TraceResult ("CPFServerBackEnd::GetTable", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::FillHierarchyTable()
//
//    Parameters
//      FolderEID	Entry ID of the folder we are building the
//                  hierarchy table for.
//      pTable      ITableData allocated by the caller of the method
//
//    Purpose
//      Here we fill the rows of the hierachy table of the given container.
//      The data for the rows is comming from the PF server.
//      The columns in the rows added will have at least the default columns
//      in the tag array specified when the ITableData was created.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::FillHierarchyTable (PRIVATE_ENTRYID FolderEID,
                                                   LPTABLEDATA     pTable,
                                                   LPSBinary       pStoreEID,
                                                   long            lFolderAccess)
{
    SPropValue spvProps[MIN_HIERARCHY_TABLE_PROPS] = { 0 };
    SRow row = { 0 };
    row.cValues = MIN_HIERARCHY_TABLE_PROPS;
    row.lpProps = spvProps;

    PRIVATE_ENTRYID eidObject = { 0 }, eidParent = { 0 };
    eidObject.uidGlobal = eidParent.uidGlobal = m_uidBackEnd;
    eidObject.bVersion = eidParent.bVersion = ENTRYID_VERSION;
    eidObject.bObject = eidParent.bObject = MAPI_FOLDER;
    
    PUBLIC_FOLDER_INFO_A Info;

    for (ULONG i=0; i<MIN_HIERARCHY_TABLE_PROPS; i++)
    {
        spvProps[i].ulPropTag = sptFolderHierProps.aulPropTag[i];
    }

    long lPipeNum;
    TCHAR szPipeName[64];
    HANDLE hSource = NULL;
    DWORD dwInstKey, dwBytesRead;

    spvProps[HT_INST_KEY].Value.bin.cb  = sizeof(DWORD);
    spvProps[HT_INST_KEY].Value.bin.lpb = (LPBYTE)&dwInstKey;
    spvProps[HT_EID].Value.bin.cb       = CB_PRIVATE_EID;
    spvProps[HT_EID].Value.bin.lpb      = (LPBYTE)&eidObject;
    spvProps[HT_RKEY].Value.bin.cb      = CB_PRIVATE_EID;
    spvProps[HT_RKEY].Value.bin.lpb     = (LPBYTE)&eidObject;
    spvProps[HT_PEID].Value.bin.cb      = CB_PRIVATE_EID;
    spvProps[HT_PEID].Value.bin.lpb     = (LPBYTE)&eidParent;
    spvProps[HT_SEID].Value.bin.cb      = pStoreEID->cb;
    spvProps[HT_SEID].Value.bin.lpb     = pStoreEID->lpb;
    spvProps[HT_SRKEY].Value.bin.cb     = sizeof(MAPIUID);
    spvProps[HT_SRKEY].Value.bin.lpb    = (LPBYTE)&m_uidBackEnd;
    spvProps[HT_MAPPING].Value.bin.cb   = sizeof(MAPIUID);
    spvProps[HT_MAPPING].Value.bin.lpb  = (LPBYTE)&m_uidBackEnd;
    spvProps[HT_FLDTYPE].Value.l        = FOLDER_GENERIC;
    spvProps[HT_OBJTYPE].Value.l        = MAPI_FOLDER;
    spvProps[HT_DISPTYPE].Value.l       = DT_FOLDER;
    spvProps[HT_NAME].Value.lpszA       = (LPSTR)Info.szFldName;
    spvProps[HT_ACCESS].Value.l         = lFolderAccess;
    spvProps[HT_COMMENT].Value.lpszA    = (LPSTR)Info.szFldComment;

    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::FillHierarchyTable", hResult);
        return hResult;
    }
    RpcTryExcept
    {
        // Make the remote call to request the server to open a pipe for the download
        hResult = RemotePFGetHierarchyTable (FolderEID.dwObjID, &lPipeNum);
        if (!hResult)
        {
            // Construct the download pipe name
            wsprintf (szPipeName, PIPE_NAME_FORMAT, m_szServer, lPipeNum);
            // Create our endpoint and connect    
            hSource = CreateFile (szPipeName, 
                                  GENERIC_READ,
                                  0,
                                  NULL,
                                  OPEN_EXISTING,
                                  0,
                                  NULL);
            if (INVALID_HANDLE_VALUE == hSource)
            {
                hResult = HRESULT_FROM_WIN32(GetLastError());
            }
        }
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::FillHierarchyTable: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    
    if (hResult)
    {
        return hResult;
    }
    
    do
    {
        // Read the pipe handle.
        if (!ReadFile (hSource, &Info, sizeof(PUBLIC_FOLDER_INFO_A), &dwBytesRead, NULL))
        {
            hResult = HRESULT_FROM_WIN32(GetLastError());
            if (HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)   != hResult &&   // For Windows NT
                HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED) != hResult)     // For Windows 95
            {
                // There was an error and we can't continue
                TraceResult ("CPFServerBackEnd::FillHierarchyTable: Failed to read from the source", hResult);
            }
            else
            {
                // If the pipe was broken, it means the server finished writing
                // to the it, so we are finished reading from it.
                hResult = S_OK;
            }
        }
        else
        {
            eidObject.dwObjID = dwInstKey = Info.dwObjID;
            eidParent.dwObjID = Info.dwParentID;
            spvProps[HT_COUNT].Value.l = Info.dwCount;
            spvProps[HT_UNREAD].Value.l = Info.dwUnread;
            spvProps[HT_ASSCNT].Value.l = Info.dwAssoc;
            spvProps[HT_SUBFLD].Value.b = (BOOL)Info.fHasSubFolders;
            CopyMemory (&spvProps[HT_CTIME].Value.ft, &Info.ftCreateTime, sizeof(FILETIME));
            CopyMemory (&spvProps[HT_LMTIME].Value.ft, &Info.ftLastModTime, sizeof(FILETIME));
            hResult = pTable->HrModifyRow (&row);
            hResult = S_OK;
        }
    } while (dwBytesRead && !hResult);
    CloseHandle (hSource);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::FillContentsTable()
//
//    Parameters
//      FolderEID	Entry ID of the folder we are building the
//                  hierarchy table for.
//      pTable      ITableData allocated by the caller of the method
//
//    Purpose
//      Here we fill the rows of the hierachy table of the given container.
//      The data for the rows is comming from the PF server.
//      The columns in the rows added will have at least the default columns
//      in the tag array specified when the ITableData was created.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::FillContentsTable (PRIVATE_ENTRYID FolderEID,
                                                  LPTABLEDATA     pTable,
                                                  LPSBinary       pStoreEID,
                                                  long            lFolderAccess)
{
    SRow row = { 0 };
    //row.cValues = ;
    //row.lpProps = ;

    PRIVATE_ENTRYID eidObject = { 0 }, eidParent = { 0 };
    eidObject.uidGlobal = eidParent.uidGlobal = m_uidBackEnd;
    eidObject.bVersion = eidParent.bVersion = ENTRYID_VERSION;
    eidObject.bObject = eidParent.bObject = MAPI_FOLDER;
    
    long lPipeNum;
    TCHAR szPipeName[64];
    HANDLE hSource = NULL;
    DWORD dwInstKey = 0, dwBytesRead;

    HRESULT hResult = BindToServer (m_szServer);
    if (hResult)
    {
        TraceResult ("CPFServerBackEnd::FillContentsTable", hResult);
        return hResult;
    }
    RpcTryExcept
    {
        // Make the remote call to request the server to open a pipe for the download
        hResult = RemotePFGetContentsTable (FolderEID.dwObjID, &lPipeNum);
        if (!hResult)
        {
            // Construct the download pipe name
            wsprintf (szPipeName, PIPE_NAME_FORMAT, m_szServer, lPipeNum);
            // Create our endpoint and connect    
            hSource = CreateFile (szPipeName, 
                                  GENERIC_READ,
                                  0,
                                  NULL,
                                  OPEN_EXISTING,
                                  0,
                                  NULL);
            if (INVALID_HANDLE_VALUE == hSource)
            {
                hResult = HRESULT_FROM_WIN32(GetLastError());
            }
        }
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::FillContentsTable: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept
    
    if (hResult)
    {
        return hResult;
    }
    
    DWORD dwLastAlloc = 0, dwSize = 0;
    LPBYTE pBuffer = NULL;
    do
    {
        // Read the pipe handle.
        if (!ReadFile (hSource, &dwSize, sizeof(DWORD), &dwBytesRead, NULL))
        {
            hResult = HRESULT_FROM_WIN32(GetLastError());
            if (HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)   != hResult &&   // For Windows NT
                HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED) != hResult)     // For Windows 95
            {
                // There was an error and we can't continue
                TraceResult ("CPFServerBackEnd::FillContentsTable: Failed to read from the source", hResult);
            }
            else
            {
                // If the pipe was broken, it means the server finished writing
                // to the it, so we are finished reading from it.
                hResult = S_OK;
            }
        }
        else
        {
            if (dwSize)
            {
                if (dwSize > dwLastAlloc)
                {
                    PrivFree ((LPVOID *)&pBuffer);
                    hResult = PrivAllocate (dwSize, (LPVOID *)&pBuffer, "FillContentsTable Buffer");
                    if (!hResult)
                    {
                        dwLastAlloc = dwSize;
                    }
                }
            }
            if (!hResult && dwSize)
            {
                // Read the pipe handle.
                if (!ReadFile (hSource, pBuffer, dwSize, &dwBytesRead, NULL))
                {
                    hResult = HRESULT_FROM_WIN32(GetLastError());
                    if (HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)   != hResult &&   // For Windows NT
                        HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED) != hResult)     // For Windows 95
                    {
                        // There was an error and we can't continue
                        TraceResult ("CPFServerBackEnd::FillContentsTable: Failed to read from the source", hResult);
                    }
                    else
                    {
                        // If the pipe was broken, it means the server finished writing
                        // to the it, so we are finished reading from it.
                        hResult = S_OK;
                    }
                }
                else
                {
                    // TODO: 
                    //hResult = pTable->HrModifyRow (&row);
                    //hResult = S_OK;
                }
            }
        }
    } while (dwBytesRead && !hResult);
    PrivFree ((LPVOID *)&pBuffer);
    CloseHandle (hSource);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::GetTableProps()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CPFServerBackEnd::GetTableProps (const TABLE_TYPE  tblType,
                                              long              lAccess,
                                              PRIVATE_ENTRYID * pParentEID,
                                              PRIVATE_ENTRYID * pObjectEID,
                                              CPropList *       pList,
                                              LPSPropValue *    ppProps,
                                              DWORD &           dwLastAllocated,
                                              LPTABLEDATA       pTable,
                                              LPVOID            pData1,
                                              LPVOID            pData2)
{
    HRESULT hResult;
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            hResult = GetHierarchyTableProps (pList,
                                              ppProps,
                                              *pObjectEID,
                                              *pParentEID,
                                              lAccess,
                                              (LPSBinary)pData1,
                                              (long)pData2,
                                              dwLastAllocated,
                                              pTable,
                                              m_uidBackEnd);
            break;
        case TABLE_CONTENTS :
            /*
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = GetContentsTableProps (pList,
                                             ppProps,
                                             *pObjectEID,
                                             lAccess,
                                             (LPSBinary)pData1,
                                             dwLastAllocated,
                                             pTable);
            break;
            */
        case TABLE_ATTACHMENTS :
            /*
            lAccess &= (MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY | MAPI_ACCESS_READ);
            hResult = GetAttachmentTableProps (pList,
                                               ppProps,
                                               pObjectEID->dwObjID,
                                               lAccess,
                                               (LPSBinary)pData1,
                                               dwLastAllocated,
                                               pTable);
            break;
            */
        default :
            hResult = MAPI_E_NO_SUPPORT;
    }
    TraceResult ("CPFServerBackEnd::GetTableProps", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CPFServerBackEnd::SaveObject()
//
//    Parameters
//
//    Purpose
//
//    Return Value
//
STDMETHODIMP CPFServerBackEnd::SaveObject (PRIVATE_ENTRYID ParentEID,
                                           PRIVATE_ENTRYID ObjectEID,
                                           CPropList *     pList,
                                           COpenPropList * pOpenList,
                                           LPTABLEDATA     pTable)
{
    short sSize = 4096;
    char szBuffer[4096] = { 0 };
    lstrcpy ((LPSTR)&szBuffer[16], "Hello World!!!!");
    HRESULT hResult;
    RpcTryExcept
    {
        hResult = RemovePFCreateMessage (ParentEID.dwObjID, sSize, (unsigned char *)szBuffer, &ObjectEID.dwObjID);
    }
    RpcExcept(1)
    {        
        // If we got here is because there was an error while call was made
        // or when it was about to be made.
        hResult = RpcExceptionCode();
        TraceRPCError ("CPFServerBackEnd::SaveObject: Remote call failed", hResult);
        if (RPC_S_SERVER_UNAVAILABLE == hResult)
        {
            hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
        }
        else
        {
            hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
        }
    }
    RpcEndExcept

    /*
    SPropValue spvSize = { 0 }, spvTime = { 0 };
    LPSTREAM pStream;
    STATSTG stat;
    HRESULT hResult = GetStream (ObjectEID.dwObjID, RECFLD_PROPERTIES, &pStream);
    if (!hResult)
    {
        if (pTable)
        {
            ASSERT (MAPI_MESSAGE == ObjectEID.bObject);
            // TODO: hResult = SaveRecipTable (ObjectEID.dwObjID, pTable);
        }
        if (!hResult && pOpenList)
        {
            // TODO: hResult = SaveLargeProps (pOpenList, pList);
        }
        if (!hResult)
        {
            hResult = pList->SaveToStream (pStream);
            // If the object we are saving is not a folder, we most calculate PR_xxx_SIZE
            if (!hResult && (MAPI_FOLDER != ObjectEID.bObject))
            {
                if (MAPI_MESSAGE == ObjectEID.bObject)
                {
                    spvSize.ulPropTag = PR_MESSAGE_SIZE;
                }
                else
                {
                    ASSERT (MAPI_ATTACH == ObjectEID.bObject);
                    spvSize.ulPropTag = PR_ATTACH_SIZE;
                }
                CalculateSize (ObjectEID, pList, spvSize.Value.l);
                pStream->Stat (&stat, STATFLAG_NONAME);
                spvSize.Value.l += stat.cbSize.LowPart;
                pList->Insert (&spvSize);
                pList->SaveToStream (pStream);
            }
            if (!hResult && (MAPI_ATTACH == ObjectEID.bObject))
            {
                // TODO: UpdateAttachmentsCount (ParentEID, ObjectEID, FALSE);
            }
        }
        pStream->Release();
    }
    */
    TraceResult ("CPFServerBackEnd::SaveObject", hResult);
    return hResult;
}

// End of file for MSSRVBE.CPP