///////////////////////////////////////////////////////////////////////////////
//
//  OTTO.CPP
//
//  This module contains the code for a sample extended MAPI mailbox agent client
//  This sample demonstrates how to process new mail notifications, create,move,and
//  submit messages, set/get properties and simple table operations.
//
//  This example is slightly contrived for the following reasons:
//  1. a mailbox agent will be server-based instead of running on an end-user's machine
//  2. it will run as an NT service
//  3. it will logon to a special mailbox instead of a user's mailbox
//  4. the account it runs under will have special privileges on the server, e.g.Admin
//     privileges on MS Exchange
//     
//  The sample models a mailbox agent that accepts admissions applications to a fictitious
//  university. An applicant sends mail with the name of the college s/he is applying to in
//  the subject field and sends the application to a special email address corresponding to
//  the university admissions office's mailbox. The mailbox agent monitors this mailbox and
//  routes inbound applications to one of three public folders (College of Arts, College of
//  Science, or College of Business) depending on which college the applicant is applying to.
//
//  Each college's admissions officer(s) picks up applications from his/her college's
//  public folder. If this mailbox agent is run against a Microsoft Exchange account with Admin
//  privileges on Exchange, (i.e. read-write access to the Exchange public store's IPM subtree)
//  the scenario is realistic. The only difference in that case is the logon sequence and
//  the fact that the agent isn't an NT service.
//
//  If this program is run in an ordinary user account, or against a profile without a public
//  folders store, or there is an error opening/creating public folders, then the program 
//  simulates the public folder store by creating the college folders in the default store's
//  IPM subtree.
//
//  Author :    Les Thaler
//  Copyright (c) 1996 Irving De La Cruz and Les Thaler
//  
//  NOTE: for Windows NT and Windows 95 ONLY. No 16 bit version. Requires Extended MAPI 1.0
//  libraries and headers. See the Microsoft Win32 SDK for details.
//

#include <windows.h>

#define USES_IID_IMAPITable
#define USES_IID_IMessage
#define USES_IID_IMAPIAdviseSink
#define INITGUID

#include <initguid.h>
#include <mapiguid.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapitags.h>
#include "otto.h"   
 
LPSTR gszFolderNames[NUM_PUB_FOLDERS] = {"Admissions Applications",
                                          "College of Arts",
                                          "College of Science",
                                          "College of Business"};


//////////////////////////////////////////////////////////////////////////////
//
//  Main Window procedure.
//  
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static  PCSession pCSess;
            HINSTANCE hInst   = (HINSTANCE)GetWindowLong(hWnd,GWL_HINSTANCE);
    	    WORD      wmId    = LOWORD(wParam), 
	                  wmEvent = HIWORD(wParam);
 
    if (!hInst)
        return 0;

	switch (message) 
	{ 
          case WM_CREATE:
              PostMessage(hWnd,WM_INIT,0,0);
              break;
          case WM_INIT:
            OnError(hWnd,InitSession(hWnd,hInst,&pCSess));
            break;                    

		case WM_COMMAND:
			switch (wmId) 
			{
                case IDM_ABOUT:
					DialogBox(hInst, "AboutBox", hWnd, (DLGPROC)About);
					break;
                case ID_ERROR_EXIT:
				case IDM_EXIT:
                    if (pCSess)
                        delete pCSess;
                    
					DestroyWindow(hWnd);
					break;
            	default:
					return (DefWindowProc(hWnd, message, wParam, lParam));
			}
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
        case WM_SYSCOMMAND:
            if (SC_CLOSE == wParam)
                if (pCSess)
                    delete pCSess;
            // fall through
		default:
			return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	return (0);
}

///////////////////////////////////////////////////////////////////////////////
//    CAdviseSink::AddRef()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          To increase the reference count of this object,
//          and of any contained (or aggregated) object within
//
//    Return Value
//      Usage count for this object
//
STDMETHODIMP CAdviseSink::QueryInterface(REFIID   riid,LPVOID * ppvObj)
{
    if (IsBadWritePtr(ppvObj,sizeof LPUNKNOWN))
        return MAPI_E_INVALID_PARAMETER;

    *ppvObj = NULL;

    // If this is one of the two IID return an interface pointer to it
    if (riid == IID_IUnknown || riid == IID_IMAPIAdviseSink)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        m_cRef++;
        return NOERROR;
    }

    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
//    CAdviseSink::AddRef()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          To increase the reference count of this object,
//          and of any contained (or aggregated) object within
//
//    Return Value
//      Usage count for this object
//
STDMETHODIMP_ (ULONG) CAdviseSink::AddRef (void)
{   
    return ++m_cRef;
}

///////////////////////////////////////////////////////////////////////////////
//    CAdviseSink::Release()
//
//    Parameters
//          Refer to OLE Documentation on this method
//
//    Purpose
//          Decreases the usage count of this object. If the 
//          count is down to zero, object deletes itself.
//
//    Return Value
//      Usage of this object
//
STDMETHODIMP_ (ULONG) CAdviseSink::Release (void)
{
    ULONG ulRefCount;
    ulRefCount = --m_cRef;

    // If nobody is using this object, delete it
    if (0 == ulRefCount)
    {
        // Destroy the object
        delete this;        // don't need to leave the critical section
        return ulRefCount;  // its deleted when the object is destroyed
    }
        
    return ulRefCount;
}

///////////////////////////////////////////////////////////////////////////////
//    CAdviseSink:: OnNotify()
//
//    Parameters
//          Refer to MAPI Documentation on this method
//
//    Purpose
//          Called when an event we've Advised on occurs. In this case,
//          we only Advise on fnevNewMail. We respond by moving the
//          inbound admissions application message to the correct college's
//          public folder and replying to the sender. If the inbound message
//          isn't correctly addressed, we send an error reply and leave the
//          incoming message in the Admissions folder.
//
//    Return Value
//      S_OK in all cases
//
STDMETHODIMP_ (ULONG) CAdviseSink::OnNotify(ULONG cNotifs,LPNOTIFICATION lpNotifications)
{
    EnterCriticalSection(&m_pCSess -> m_cs);
    for (int i = 0; i < (int) cNotifs; i++)
        if (fnevNewMail == lpNotifications[i].ulEventType)
        {
            // don't respond to any but IPM.Note messages, e.g. don't reply to our replies
            if (!lstrcmpi("IPM.NOTE",lpNotifications[i].info.newmail.lpszMessageClass))
            {
                LPMESSAGE   pMsg;
                ULONG       ulObjType;
                int         iCollege;
                SBinary     sb;

                sb.cb  = lpNotifications[i].info.newmail.cbEntryID;
                sb.lpb = (LPBYTE)lpNotifications[i].info.newmail.lpEntryID;
            
                if (FAILED(m_pCSess -> m_pMDB -> OpenEntry(sb.cb,
                                                           (LPENTRYID)sb.lpb,
                                                            &IID_IMessage,
                                                            MAPI_BEST_ACCESS,
                                                            &ulObjType,
                                                            (LPUNKNOWN *)&pMsg)))
                    return S_OK;

                if (MAPI_MESSAGE != ulObjType || !pMsg)
                    return S_OK;
                
                if (SUCCEEDED(m_pCSess -> SendReply(pMsg,&iCollege)))
                {
                    pMsg -> Release();
                
                    if (iCollege < NUM_COLLEGES)
                    {   
                        SBinaryArray sba;
                        sba.cValues = 1;
                        sba.lpbin   = &sb;
                        m_pCSess -> RouteToFolder(&sba, iCollege);
                    }
                }
            }
        }
    
    LeaveCriticalSection(&(m_pCSess -> m_cs));

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//   CSession::RouteToFolder
//
//    Parameters
//          iCollege    which college's folder this application goes to
//          psba        the entry ID of the message to move
//
//    Purpose
//          Moves the message to the folder given by iCollege
//
//    Return Value
//      none.
//
void CSession::RouteToFolder(LPENTRYLIST psba, int iCollege)
{

    int  iFldr = iCollege + NUM_DEF_FOLDERS;

    if (psba -> cValues)
        m_pFolders[INBOX] -> CopyMessages((LPENTRYLIST) psba,
                                            NULL,
                                            (LPVOID) m_pFolders[iFldr],
                                            NULL,
                                            NULL,
                                            MESSAGE_MOVE);
}

///////////////////////////////////////////////////////////////////////////////
//   CSession::SendReply
//
//    Parameters
//          pInMsg      inbound message we're replying to
//          pCollege    points to numeric ID of the college the application was
//                      bound for
//
//    Purpose
//          Sends a reply message to an inbound application message, if the
//          the inbound subject specifies a valid college. Sends an error
//          reply otherwise.
//
//    Return Value
//      none.
//
STDMETHODIMP CSession::SendReply(LPMESSAGE pInMsg,int * pCollege)
{
    HRESULT     hRes;
    LPMESSAGE   pReplyMsg   = NULL;

    if (FAILED(hRes = CreateReplyNote(pInMsg,pCollege,&pReplyMsg)))
        goto Quit;

    hRes = pReplyMsg -> SubmitMessage(0);

Quit:

    if (pReplyMsg)
        pReplyMsg -> Release();

    return hRes;

}

//////////////////////////////////////////////////////////////////////////////
//
//  Initializes MAPI,logs on, creates and initializes the CSession object.
//  

STDMETHODIMP InitSession(HWND hWnd,HINSTANCE hInst,PCSession * ppCSess)
{
    LPMAPISESSION pSession    = NULL;
    PCSession     pCSess      = NULL;
    PCAdviseSink  pAdviseSink = NULL;
    HRESULT       hRes;
        
    if (FAILED(hRes = MAPIInitialize(NULL)))
        return hRes;

    if (FAILED(hRes = MAPILogonEx((ULONG)hWnd,
                                  NULL,
                                  NULL,
                                  LOGON_FLAGS,
                                  &pSession)))
        goto Quit;
    
    pCSess = new CSession(pSession,hInst);

    if (!pCSess)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

    // initialize CSession object    
    if (FAILED(hRes = pCSess -> Init()))
        goto Quit;

    // open default msg store
    if (FAILED(hRes = pCSess -> OpenDefStore(hWnd)))
        goto Quit;

    pAdviseSink = new CAdviseSink(pCSess);
    pAdviseSink -> AddRef();
    
    if (!pAdviseSink)
    {
        hRes = MAPI_E_NOT_ENOUGH_MEMORY;
        goto Quit;
    }

    EnterCriticalSection(&(pCSess -> m_cs));

    if (SUCCEEDED(hRes = pCSess -> SetNotifications(pAdviseSink)))
        // get IPM folder entry IDs
        if (SUCCEEDED(hRes = pCSess -> OpenIPMSubtree(hWnd)))
        {
            // Creates public Admissions folders or opens them if they exist
            // We try this in the Public Folder Store's IPM_SUBTREE, If there's
            // no public folder store, we simulate it by using the default store.
            hRes = pCSess -> ValidatePublicFolders();
        }
    LeaveCriticalSection(&(pCSess -> m_cs));

Quit:
    pCSess -> ReleasePubStore();

    if (FAILED(hRes))
    {
        // destructor releases session and uninitializes
        if (pCSess)
            delete pCSess;
        else
        {
            if (pSession)
                pSession -> Release();        

            MAPIUninitialize();
        }
        return hRes;
    }

    *ppCSess = pCSess;
    return hRes;
}

//////////////////////////////////////////////////////////////////////////////
//
//  CSession::CreateSubfolder
//  
//    Parameters
//          pPubIPM     ptr to public folder store's IPM subtree
//
//    Purpose
//          Creates the admission's office folder,and each college's foler
//          in the folder pointed to by pPubIPM..
//
//    Return Value
//      S_OK on success, an error code otherwise.
//
STDMETHODIMP CSession :: CreateSubfolders(LPMAPIFOLDER pPubIPM)
{
    HRESULT  hRes;

    for (int i = PUB_ADM; i < NUM_FOLDERS; i++)
    {
        LPSPropValue pspv = NULL;

        if (m_pFolders[i])
        {
            m_pFolders[i] -> Release();
            m_pFolders[i] = NULL;
        }

        if (FAILED(hRes = pPubIPM -> CreateFolder(FOLDER_GENERIC,
                                                  gszFolderNames[i - PUB_ADM],
                                                  "applications routed to this college",
                                                  NULL,
                                                  OPEN_IF_EXISTS,
                                                  &m_pFolders[i])))
            goto Quit;
    
        // create the others as children of "Admissions Applications"
        pPubIPM = m_pFolders[PUB_ADM];
    
        if (FAILED(hRes = HrGetOneProp(m_pFolders[i],PR_ENTRYID,&pspv)))
            goto Quit;
    
        // copy the folder's entry ID to class member
        hRes = CopySBinary(& m_psbFolderEID[i], 
                           & pspv -> Value.bin,
                           (LPVOID) m_psbFolderEID);
    
        MAPIFreeBuffer(pspv);
        
        if (hRes)
            goto Quit;
    }
Quit:
    return hRes;
}


//////////////////////////////////////////////////////////////////////////////
//
//  CSession::ValidatePublicFolders
//  
//    Parameters
//          none.
//
//    Purpose
//          Creates the admission's office folder,and each college's foler
//          in the public folder store's IPM subtree. If the profile doesn't 
//          contain a public folder store, or the user doesn't have write access,
//          creates the subfolders in the default store.
//
//    Return Value
//      S_OK on success, an error code otherwise.
//
STDMETHODIMP CSession::ValidatePublicFolders()
{
    HRESULT         hRes = S_OK;
    ULONG           ulObjType;
    LPMAPITABLE     pHTbl = NULL;
    LPSRowSet       pRow = NULL;
    LPMAPIFOLDER    pPubRoot = NULL,
                    pPubIPM = NULL;

    SizedSPropTagArray(2,sptCols) = {2,PR_DISPLAY_NAME,PR_ENTRYID};

    // if IPM subtree is really in public folder store, we have extra work
    if (m_pMDB != m_pPubMDB)
    {
        if (FAILED(hRes = m_pPubMDB -> OpenEntry(0,
                                                NULL,
                                                NULL,
                                                MAPI_BEST_ACCESS,
                                                &ulObjType,
                                                (LPUNKNOWN *) &pPubRoot)))
            goto Quit;
    
        if (MAPI_FOLDER != ulObjType) 
        {
            hRes = MAPI_E_INVALID_TYPE;
            goto Quit;
        }

        if (FAILED(hRes = pPubRoot -> GetHierarchyTable(0,&pHTbl)))
            goto Quit;

        // get IPM Subtree's Entry ID
        if (FAILED(hRes = HrQueryAllRows(pHTbl,
                                         (LPSPropTagArray) &sptCols,
                                         NULL,
                                         NULL,
                                         0,
                                         &pRow)))
            goto Quit;

        hRes = MAPI_E_NOT_FOUND;

        if (pRow && pRow -> cRows)
        {
            SBinary sbEID;

            // search manually because some HTbls don't support Restrict or FindRow
            for (int i = 0; i < (int) pRow -> cRows; i++)
            {
                if (pRow -> aRow[i].lpProps[0].ulPropTag != PR_DISPLAY_NAME ||
                    pRow -> aRow[i].lpProps[1].ulPropTag != PR_ENTRYID)
                    continue;

                if (lstrcmpi("IPM_SUBTREE",pRow -> aRow[i].lpProps[0].Value.lpszA))
                    continue;
                else
                {
                    sbEID  = pRow -> aRow[i].lpProps[1].Value.bin;
                    hRes = S_OK;
                    break;
                }
            }

            if (hRes)
                goto Quit;

            // Open public folders IPM_SUBTREE       
            hRes = m_pPubMDB -> OpenEntry(sbEID.cb,
                                          (LPENTRYID)sbEID.lpb,
                                          NULL,
                                          MAPI_MODIFY,
                                          &ulObjType,
                                          (LPUNKNOWN *) &pPubIPM);

            if (S_OK != hRes || MAPI_FOLDER != ulObjType) 
            {   
                hRes = MAPI_E_NOT_FOUND;
                goto Quit;
            }

            hRes = CreateSubfolders(pPubIPM);
        }
    }
    else  
        // we're simulating public folders in default store
        return CreateSubfolders(m_pFolders[IPM_ROOT]);

Quit:
    FreeProws(pRow);
    pHTbl -> Release();

    if (pPubRoot)
        pPubRoot -> Release();

    if (pPubIPM)
        pPubIPM -> Release();

    if (hRes)       // error in public folders, simulate in default store
        return CreateSubfolders(m_pFolders[IPM_ROOT]);
    
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Create an array of entry IDs for the IPM subtree folders
//  
STDMETHODIMP CSession::Init()
{
    HRESULT hRes;
    
    if (FAILED(hRes = MAPIAllocateBuffer(NUM_FOLDERS * sizeof SBinary, 
                                            (LPVOID *) & m_psbFolderEID)))
        goto Quit;

    ZeroMemory((LPVOID) m_psbFolderEID, NUM_FOLDERS * sizeof SBinary);
    InitializeCriticalSection(&m_cs);
Quit:        
    return hRes;    
}

///////////////////////////////////////////////////////////////////////////////
//
//  Open the default store for this session
//  
STDMETHODIMP CSession::SetNotifications(LPMAPIADVISESINK pAdviseSink)
{
    return m_pMDB -> Advise(0,NULL, fnevNewMail, pAdviseSink,&m_ulAdviseCnx);
}
///////////////////////////////////////////////////////////////////////////////
//
//  Open the default store for this session
//  
STDMETHODIMP CSession::OpenDefStore(HWND hWnd)
{
    LPMAPITABLE         pStoresTbl = NULL;
    LPSRowSet           pRow       = NULL,
                        pRowPub    = NULL;
    SBinary             sbEID      = {0,NULL};
    static SRestriction sres;
    SPropValue          spv, 
                      * pspvMask = NULL;
    HRESULT             hRes; 
    int                 i;

    static SizedSPropTagArray(2,sptCols) = {2,PR_ENTRYID,PR_DEFAULT_STORE};
    
    if (FAILED(hRes = m_pSess -> GetMsgStoresTable(0, &pStoresTbl)))
    {
        ShowLastError(hRes,hWnd,m_pSess);
        goto Quit;
    }

    sres.rt = RES_PROPERTY;
    sres.res.resProperty.relop = RELOP_EQ;
    sres.res.resProperty.ulPropTag = PR_DEFAULT_STORE;
    sres.res.resProperty.lpProp = &spv;

    spv.ulPropTag = PR_DEFAULT_STORE;
    spv.Value.b   = TRUE;

    if (FAILED(hRes = HrQueryAllRows(pStoresTbl,
                                     (LPSPropTagArray) &sptCols,
                                     &sres,
                                     NULL,
                                     0,
                                     &pRow)))
        goto Quit;

    if (pRow                    &&
        pRow -> cRows           && 
        pRow -> aRow[0].cValues &&
        PR_ENTRYID == pRow -> aRow[0].lpProps[0].ulPropTag)
    {
        sbEID  = pRow -> aRow[0].lpProps[0].Value.bin;
    }
    else
    {
        hRes = MAPI_E_NOT_FOUND;
        goto Quit;
    }        

    if (FAILED(hRes = m_pSess -> OpenMsgStore((ULONG)hWnd,
                                              sbEID.cb,
                                              (LPENTRYID)sbEID.lpb,
                                              NULL,
                                              MDB_WRITE,
                                              &m_pMDB)))
    {
        ShowLastError(hRes,hWnd,m_pSess);
        goto Quit;
    }
    
    hRes = HrGetOneProp(m_pMDB, PR_STORE_SUPPORT_MASK,&pspvMask);
    
    if (hRes || !(pspvMask -> Value.l & STORE_NOTIFY_OK))
    {
        hRes = MAPI_E_NOT_FOUND;
        goto Quit;
    }

    MAPIFreeBuffer(pspvMask);
    sres.res.resProperty.relop = RELOP_NE;

    // See if we can get public folder store. If we can't, fake it by
    // using the default store
    if (FAILED(hRes = HrQueryAllRows(pStoresTbl,
                                     (LPSPropTagArray) &sptCols,
                                     &sres,
                                     NULL,
                                     0,
                                     &pRowPub)))
        goto Quit;


    for (i = 0; i < (int) pRowPub -> cRows; i++)
    {
        if (pRowPub                    &&
            pRowPub -> cRows           && 
            pRowPub -> aRow[i].cValues &&
            PR_ENTRYID == pRowPub -> aRow[i].lpProps[0].ulPropTag)
        {

            sbEID  = pRowPub -> aRow[i].lpProps[0].Value.bin;

            if (FAILED(m_pSess -> OpenMsgStore((ULONG)hWnd,
                                               sbEID.cb,
                                               (LPENTRYID)sbEID.lpb,
                                               NULL,
                                               MDB_WRITE,
                                               &m_pPubMDB)))
                continue;
            else
                if (SUCCEEDED(HrGetOneProp(m_pPubMDB, PR_STORE_SUPPORT_MASK,&pspvMask)))
                    if (PUB_FLAGS != (pspvMask -> Value.l & PUB_FLAGS))
                    {   
                        m_pPubMDB -> Release();
                        m_pPubMDB = NULL;
                        MAPIFreeBuffer(pspvMask);
                        pspvMask = NULL;
                        continue;
                    }    
                    else
                    {
                        MAPIFreeBuffer(pspvMask);
                        pspvMask = NULL;
                        break;
                    }
        }
    }

    if (!m_pPubMDB)
        m_pPubMDB = m_pMDB;

Quit:
    FreeProws(pRow);
    FreeProws(pRowPub);
    MAPIFreeBuffer(pspvMask);
    
    if (pStoresTbl)
        pStoresTbl -> Release();
                
    return hRes;
}

///////////////////////////////////////////////////////////////////////////////
//
//  
//  Open each folder and save a pointer to open interfaces,
//
STDMETHODIMP CSession::OpenIPMSubtree(HWND hWnd)
{
    LPSPropValue  pProps = NULL;
    HRESULT       hRes;
    ULONG         cProps,
                  ulObjType;
    int           i;
    static SizedSPropTagArray(NUM_DEF_FOLDERS,sptFldrs) = {NUM_DEF_FOLDERS,
                                                                   PR_NULL,     //save room for Inbox
                                                                   PR_IPM_OUTBOX_ENTRYID,
                                                                   PR_IPM_WASTEBASKET_ENTRYID,
                                                                   PR_IPM_SENTMAIL_ENTRYID,
                                                                   PR_IPM_SUBTREE_ENTRYID};    //save room for root

    if (FAILED(hRes = m_pMDB -> GetProps((LPSPropTagArray)&sptFldrs,
                                          0,
                                          &cProps,
                                          &pProps)))
        goto Quit;

    // get Inbox folder's entry ID
    if (FAILED(hRes = m_pMDB -> GetReceiveFolder("IPM.Note",
                                                 0,
                                                 &pProps[INBOX].Value.bin.cb,
                                                 (LPENTRYID *) &pProps[INBOX].Value.bin.lpb,
                                                 NULL)))
    {
        ShowLastError(hRes, hWnd, m_pMDB);
        goto Quit;
    }

    for (i = INBOX; i <= IPM_ROOT; i++)
    {
        if (pProps[i].ulPropTag != sptFldrs.aulPropTag[i])
        {
            hRes = MAPI_E_NOT_FOUND;
            goto Quit;
        }

        // copy the folder's entry ID to class member
        if (FAILED(hRes = CopySBinary(& m_psbFolderEID[i], 
                                      & pProps[i].Value.bin,
                                      (LPVOID) m_psbFolderEID)))
                goto Quit;
    }

    for (i = INBOX; i <= IPM_ROOT; i++)
    {
        if (FAILED(hRes = m_pSess -> OpenEntry(m_psbFolderEID[i].cb,
                                               (LPENTRYID) m_psbFolderEID[i].lpb,
                                               NULL,
                                               MAPI_MODIFY,
                                               &ulObjType,
                                               (LPUNKNOWN *) &m_pFolders[i])))
        {
            ShowLastError(hRes,hWnd,m_pSess);
            goto Quit;
        }

        if (MAPI_FOLDER != ulObjType)
        {
            hRes = MAPI_E_INVALID_OBJECT;
            goto Quit;
        }
    }                         
    

Quit:
    MAPIFreeBuffer(pProps[INBOX].Value.bin.lpb);    // separate allocation,free separately
    pProps[INBOX].Value.bin.cb = 0;
    MAPIFreeBuffer((LPVOID) pProps);
    return hRes;
}


///////////////////////////////////////////////////////////////////////////////
//
//  Creates a reply message, copies/sets appropriate properties,enables 
//  dialog box controls.
//
STDMETHODIMP CSession :: CreateReplyNote(LPMESSAGE   pInMsg,
                                         int     *   pCollege,
                                         LPMESSAGE * ppReplyMsg)
{
    enum {SNDR_NAME,SNDR_ADDRTYPE,SNDR_ADDR,SNDR_EID,SNDR_SKEY,SNDR_TYPE,NUM_PROPS};
    enum {SUBJ,CLASS,BODY};

    LPSTR        szToLine = NULL;
    LPMESSAGE    pNewMsg  = NULL;
    LPSPropValue pspvIn;
    HRESULT      hRes;
    ULONG        cProps;
    int          i;
    char         buf[MAX_SUBJ + sizeof SZ_REPLY_SUBJ];
    LPADRLIST    palReplyRecips;
    LPSPropProblemArray pProbs = NULL;
    SPropValue   spvOut[3];
    

    static const SizedSPropTagArray(NUM_PROPS,sptSndr) = 
                                   {NUM_PROPS, PR_SENT_REPRESENTING_NAME,         
                                        PR_SENT_REPRESENTING_ADDRTYPE,     
                                        PR_SENT_REPRESENTING_EMAIL_ADDRESS,
                                        PR_SENT_REPRESENTING_ENTRYID,      
                                        PR_SENT_REPRESENTING_SEARCH_KEY,   
                                        PR_NULL};     // save spot for PR_RECIPIENT_TYPE
    static const SizedSPropTagArray(3,sptMsg) = 
                                   {3,  PR_SUBJECT,         
                                        PR_MESSAGE_CLASS,
                                        PR_BODY};


    static const SizedSPropTagArray(NUM_PROPS,sptRecip) = 
                                   {NUM_PROPS, PR_DISPLAY_NAME,         
                                        PR_ADDRTYPE,     
                                        PR_EMAIL_ADDRESS,
                                        PR_ENTRYID,      
                                        PR_SEARCH_KEY,   
                                        PR_RECIPIENT_TYPE};

    *pCollege = NUM_COLLEGES;

    if (FAILED(hRes = MAPIAllocateBuffer(CbNewADRLIST(1),(LPVOID*) &palReplyRecips)))
        goto Quit;

    if (FAILED(hRes = pInMsg -> GetProps((LPSPropTagArray)&sptSndr,
                                         0,
                                         &cProps,
                                         &pspvIn)))
        goto Quit;


    // check that we have all PR_SENT_REPRESENTING_XX props
    for (i = SNDR_NAME; i < NUM_PROPS; i++)
        if (sptSndr.aulPropTag[i] != pspvIn[i].ulPropTag)
        {
            hRes = MAPI_E_NOT_FOUND;
            goto Quit;
        }
        else
            pspvIn[i].ulPropTag = sptRecip.aulPropTag[i];


    if (FAILED(hRes = m_pFolders[OUTBOX] -> CreateMessage(NULL,
                                                          0,
                                                          &pNewMsg)))
        goto Quit;


    if ((S_OK != hRes) || ((pProbs) && (pProbs -> cProblem)))
    {   
        hRes = MAPI_E_NOT_FOUND;
        goto Quit;
    }

    palReplyRecips->cEntries = 1;
    palReplyRecips->aEntries[0].cValues = NUM_PROPS; // only recip props, no subj
    palReplyRecips->aEntries[0].rgPropVals = pspvIn;

    pspvIn[SNDR_TYPE].ulPropTag = PR_RECIPIENT_TYPE;
    pspvIn[SNDR_TYPE].Value.ul = MAPI_TO;


    hRes = pNewMsg -> ModifyRecipients(MODRECIP_ADD,palReplyRecips);
    FreePadrlist(palReplyRecips);

    if (S_OK != hRes)
        goto Quit;

    if (FAILED(hRes = pInMsg -> GetProps((LPSPropTagArray)&sptMsg,
                                         0,
                                         &cProps,
                                         &pspvIn)))
        goto Quit;

    // if there's no subject, or the subject isn't a application, it's an error
    if (PR_SUBJECT == pspvIn[SUBJ].ulPropTag)
        for (i = ARTS; i < NUM_COLLEGES; i++)
            if (!lstrcmpi(gszFolderNames[i],pspvIn[SUBJ].Value.lpszA))
            {
                *pCollege = i;
                break;
            }
    
    // No subject or subject not an admission application
    if (NUM_COLLEGES == *pCollege)
    {
        wsprintf(buf,"RE:%s", pspvIn[SUBJ].Value.lpszA);
        spvOut[BODY].Value.lpszA = SZ_ERROR_REPLY;
    }
    else
    {
        wsprintf(buf,
                 "RE:%s, %s",
                 pspvIn[SUBJ].Value.lpszA,
                 (*pCollege == NUM_COLLEGES ? "" : SZ_REPLY_SUBJ));
        spvOut[BODY].Value.lpszA = SZ_BODY;

    }
    spvOut[BODY].ulPropTag = PR_BODY;

    spvOut[SUBJ].ulPropTag = PR_SUBJECT;
    spvOut[SUBJ].Value.lpszA = buf;

    spvOut[CLASS].ulPropTag = PR_MESSAGE_CLASS;
    spvOut[CLASS].Value.lpszA = SZOOF_MSGCLASS;

    hRes = pNewMsg -> SetProps(3,spvOut,&pProbs);

    if (FAILED(hRes) || pProbs)
        goto Quit;

    hRes = pNewMsg -> SaveChanges(KEEP_OPEN_READWRITE);

Quit:
    if (FAILED(hRes))
    {
        if (pNewMsg)
            pNewMsg -> Release();

        pNewMsg = NULL;
    }

    MAPIFreeBuffer(pProbs);
    MAPIFreeBuffer((LPVOID) pspvIn);
    *ppReplyMsg = pNewMsg;

    return hRes;
}
