///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      ABSRVBE.CPP
//
//  Description
//      Implements the CABServerBackEnd class which derives from CBackEnd.
//      The methods in this class are used by our address book implementation
//      to request the data from the remote server host to fill contents and
//      hierarchy tables, and request the property of object in the messaging
//      host.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "ABWDS.H"
#include "ABSRVBE.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)        TraceInfoMessage(a)
#define InfoTraceRaw(a)     TraceRaw(a)
#define InfoTraceProp(a)    TraceProp(a)
#else
#define InfoTrace(a)        0
#define InfoTraceRaw(a)     0
#define InfoTraceProp(a)    0
#endif // DO_INFO_TRACES

#define GAL_NAME_FORMAT             "WINDS Users on %s"
#define GW_CONTAINERS_NAME_FORMAT   "WINDS Foreign Systems (%s)"
#define EXCHANGE_GW_NAME            "Exchange Recipients"
#define SMTP_GW_NAME                "SMTP Recipients"
#define FAX_GW_NAME                 "FAX Recipients"

enum
{
    HT_INST,
    HT_DEPTH,
    HT_ENTRYID,
    HT_DISPNAME,
    HT_DISPTYPE,
    HT_OBJTYPE,
    HT_CONTFLAGS,
    HT_AB_PROVID,
    HIERARCHY_TABLE_PROPS
};
const static SizedSPropTagArray(HIERARCHY_TABLE_PROPS, sptContainerHierProps) =
{
    HIERARCHY_TABLE_PROPS,
    {
        PR_INSTANCE_KEY,
        PR_DEPTH,
        PR_ENTRYID,
        PR_DISPLAY_NAME,
        PR_DISPLAY_TYPE,
        PR_OBJECT_TYPE,
        PR_CONTAINER_FLAGS,
        PR_AB_PROVIDER_ID
    }
};

///////////////////////////////////////////////////////////////////////////////
//    OpenRemoteDataBase()
//
//    Paremeters
//      pszServer           Name of the server the back end connects to
//      pUID                UID of the entry ID in returned by the back end
//      fWorkOffLine        Weather or not to work offline with the local
//                          address book directory data file
//      pszABDLocalFile     Name of the local ABD file.
//      ppBackEnd           Where we return the new CBackEnd object
//
//    Purpose
//      Create a instance of a CBackEnd object returned to the address book.
//      This object is used to access the server and retrieve the data. All
//      the code for the Remote procedure call and protocol of the server is
//      contained in this class implementation.
//      CBackEnd methods are also used in my generic IMAPIProp
//      implementation (CGenericProp) to open object and request stuff from
//      the data storage back end (file, server, etc)
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI OpenRemoteDataBase (LPTSTR       pszServer,
                                   MAPIUID *    pUID,
                                   BOOL         fWorkOffLine,
                                   LPTSTR       pszABDLocalFile,
                                   CBackEnd **  ppBackEnd)
{
    // To verify that we have a local address book directory file, we attempt
    // to create the data directory of this service. If the Win32 API is
    // SUCCEESSFUL, it means that the director was create and it didn't exist
    // so we must create an empty file for the ABD data file
    TCHAR szTmpDir[_MAX_PATH];
    GetTempPath (_MAX_PATH, szTmpDir);
    lstrcat (szTmpDir, WINDS_DATA_DIRECTORY);
    if (TRUE == CreateDirectory (szTmpDir, NULL))
    {
        HANDLE hFile = CreateFile (pszABDLocalFile,
                                   GENERIC_WRITE,
                                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                                   NULL,
                                   CREATE_ALWAYS,
                                   FILE_ATTRIBUTE_NORMAL,
                                   NULL);
        CloseHandle (hFile);
    }
    // Allocate the new database object. The constructor initializes internal data in the object.
    CABServerBackEnd * pSBEobj = new CABServerBackEnd (pUID,
                                                       pszServer,
                                                       fWorkOffLine,
                                                       pszABDLocalFile);
    
    if (pSBEobj)
    {
        *ppBackEnd = pSBEobj;
        return S_OK;
    }
    TraceMessage ("OpenRemoteDataBase: Failed to allocate a new back end object");
    return E_OUTOFMEMORY;
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::CABServerBackEnd()
//
//    Parameters
//      pUIDinFile          UID that identifies the entry ID returned from
//                          this backend storage (the server)
//      pszRemoteServer     Name of the remote server we are connecting to
//      fWorkOffLine        Weather or not we are offline
//      pszABDLocalFile     Name of the off-line address book directory data
//                          file from where we get the AB information while
//                          offline
//      
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CABServerBackEnd::CABServerBackEnd (MAPIUID *   pUIDinFile,
                                    LPTSTR      pszRemoteServer,
                                    BOOL        fWorkOffLine,
                                    LPTSTR      pszABDLocalFile) : CBackEnd (*pUIDinFile)
{
    InfoTrace ("CABServerBackEnd: Constructor called");
    m_cRef = 1;
    ASSERT (pszRemoteServer);
    lstrcpy (m_szServer, pszRemoteServer);
    m_fUseLocalABD = fWorkOffLine;
    lstrcpy (m_szABDLocalFile, pszABDLocalFile);
    m_pCachedLocalGAL = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::~CABServerBackEnd()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CABServerBackEnd.
//
//    Return Value
//      None
//
CABServerBackEnd::~CABServerBackEnd()
{
    InfoTrace ("CABServerBackEnd: Destructor called");
    if (m_pCachedLocalGAL)
    {
        m_pCachedLocalGAL->Release();
    }
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::GetTable()
//
//    Parameters
//      tblType         Type of table requested
//      ContainerEID    Entry ID of the container where the table will
//                      be obtained
//      lAccess         Access to the object in the rows of the tables
//      pData1          Varies depending on the type of table requested
//      pData2          Varies depending on the type of table requested
//      ppTable         Where we return the new ITableData object
//
//    Purpose
//      Allocate a new ITableData object with the default columns and index
//      for the given table requested.
//      If requested the function returns just and empty table, otherwise, the
//      implemtation fills the rows from the data source with at least the
//      default columns properties.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABServerBackEnd::GetTable (const TABLE_TYPE  tblType,
                                         PRIVATE_ENTRYID   ContainerEID,
                                         long              lAccess,
                                         LPVOID            pData1,
                                         LPVOID            pData2,
                                         LPTABLEDATA *     ppTable)
{
    LPSPropTagArray pDefaultColumnSet;
    ULONG ulIndexProp = PR_INSTANCE_KEY;
    HRESULT hResult = S_OK;
    BOOL fReturnEmptyTable = (TRUE == (BOOL)pData2);  // TRUE on pData2 means the user wants an empty table returned
    if (!IsValidEntryID (&ContainerEID))
    {
        hResult = MAPI_E_INVALID_ENTRYID;
        goto ErrorExit;
    }
    switch (tblType)
    {
        case TABLE_HIERARCHY :
            pDefaultColumnSet = (LPSPropTagArray)&sptContainerHierProps;
            break;
        case TABLE_CONTENTS :
            pDefaultColumnSet = (LPSPropTagArray)&sptContainerContProps;
            break;
        default :
            hResult = MAPI_E_NO_SUPPORT;
            goto ErrorExit;
    }
    // If we are using the local ABD, use the cached table to return the GAL table.
    if (m_fUseLocalABD && m_pCachedLocalGAL)
    {
        if (TABLE_CONTENTS == tblType)
        {
            if (GAL_CONTAINER_ID == ContainerEID.dwObjID)
            {
                *ppTable = m_pCachedLocalGAL;
                (*ppTable)->AddRef();
                return S_OK;
            }
            else
            {
                // if the container is not the GAL (i.e. a distribution list), we
                // can't supply it while OFF line so just return an empty table
                fReturnEmptyTable = TRUE;
            }
        }
    }
    
    // Before locking the backend object, create a new In-Memory table
    hResult = CreateTable (&IID_IMAPITableData,
                           GetMAPIAllocBuffer(),
                           GetMAPIAllocMore(),
                           GetMAPIFreeBuffer(),
                           NULL,
                           TBLTYPE_DYNAMIC,
                           ulIndexProp,
                           pDefaultColumnSet,
                           ppTable);
    if (hResult || fReturnEmptyTable)
    {
        goto ErrorExit;
    }
    // Looks like we need to get the information of the data source
    EnterCriticalSection (&m_csObj);
    if (TABLE_HIERARCHY  == tblType)
    {
        FillHierarchyTable (ContainerEID, (DWORD)pData1, *ppTable);
    }
    else
    {
        ASSERT (TABLE_CONTENTS  == tblType);
        hResult = FillContentsTable (ContainerEID, *ppTable);
    }
    LeaveCriticalSection (&m_csObj);
    TraceResult ("CABServerBackEnd::GetTable", hResult);
ErrorExit:
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::OpenObject()
//
//    Parameters
//      ObjectEID   Entry ID of the object we are opening
//      ppList      Where we return the property cache, CPropList, with
//                  the stored properties of the object on the back end
//                  data source.
//
//    Purpose
//      This method goes to the data source and loads all the properties
//      available for the particular object being open. The entry ID
//      structure has the object type.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABServerBackEnd::OpenObject (PRIVATE_ENTRYID  ObjectEID,
                                           CPropList **     ppList)
{
    PRIVATE_ENTRYID ManagerEID = { 0 };
    MAILBOX_INFO MBInfo = { 0 };
    DIST_LIST_INFO DLInfo = { 0 };
    HRESULT hResult = S_OK;
    ULONG ulObjType, ulContentCount = 0, i = 0;
    SPropValue spvProps[21] = { 0 };
    TCHAR szEmailAddress[64], szContainerName[64], szSearchKey[64];
    long lDTType;
    BOOL fSpecialContainer = FALSE;
    // Allocate a new property cache object
    //*ppList = new CPropList;
    *ppList = AllocatePropList();
    if (!(*ppList))
    {
        TraceMessage ("CABServerBackEnd::OpenObject: Failed to allocate property cache");
        hResult = E_OUTOFMEMORY;
        goto FinishNow;
    }
    
    // For some special ID, we don't need to go the server, so just open them here, using local data
    switch (ObjectEID.dwObjID)
    {
        case ROOT_CONTAINER_ID :
        case GATEWAY_CONTAINERS_ID :
        case FAX_CONTAINER_ID :
        case SMTP_CONTAINER_ID :
        case EXCHANGE_CONTAINER_ID :
        case GAL_CONTAINER_ID :
            ASSERT (MAPI_ABCONT == ObjectEID.bObject);
            fSpecialContainer = TRUE;
            if (ROOT_CONTAINER_ID == ObjectEID.dwObjID)
            {
                lDTType = DT_NOT_SPECIFIC;
                lstrcpy (szContainerName, "ABWDS Root Container");
            }
            else
            {
                if (GAL_CONTAINER_ID == ObjectEID.dwObjID ||
                    GATEWAY_CONTAINERS_ID == ObjectEID.dwObjID)
                {
                    lDTType = DT_GLOBAL;
                    wsprintf (szContainerName,
                              (GAL_CONTAINER_ID == ObjectEID.dwObjID ? GAL_NAME_FORMAT : GW_CONTAINERS_NAME_FORMAT),
                              m_szServer);
                }
                else
                {
                    lDTType = DT_WAN;
                    if (FAX_CONTAINER_ID == ObjectEID.dwObjID)
                    {
                        lstrcpy (szContainerName, FAX_GW_NAME);
                    }
                    else
                    {
                        if (SMTP_CONTAINER_ID == ObjectEID.dwObjID)
                        {
                            lstrcpy (szContainerName, SMTP_GW_NAME);
                        }
                        else
                        {
                            lstrcpy (szContainerName, EXCHANGE_GW_NAME);
                        }
                    }
                }
            }
            spvProps[0].ulPropTag = PR_DISPLAY_TYPE;
            spvProps[i++].Value.l = lDTType;
            spvProps[1].ulPropTag = PR_DISPLAY_NAME;
            spvProps[i++].Value.lpszA = szContainerName;
            spvProps[2].ulPropTag = PR_CONTENT_COUNT;
            spvProps[i++].Value.ul = ulContentCount;
            break;
    }
    if (fSpecialContainer)
    {
        ASSERT (3 == i);
        goto FinishNow;
    }

    if (MAPI_MAILUSER != ObjectEID.bObject && MAPI_DISTLIST != ObjectEID.bObject)
    {
        TraceMessage ("CABServerBackEnd::OpenObject: Opening object that is not supported");
        delete *ppList;
        *ppList = NULL;
        return MAPI_E_NOT_FOUND;
    }
    // If we are using the local address book directory, all the information need for Mail
    // Users and Distribution Lists is in the contents table of the GAL so get the row properties
    if (m_fUseLocalABD)
    {
        ASSERT (0 == i);
        if (NULL == m_pCachedLocalGAL)
        {
            PRIVATE_ENTRYID GALObjectEID = { 0 };
            GALObjectEID.uidGlobal = m_uidBackEnd;
            GALObjectEID.bVersion = ENTRYID_VERSION;
            GALObjectEID.bObject = MAPI_ABCONT;
            GALObjectEID.dwObjID = GAL_CONTAINER_ID;
            hResult = GetTable (TABLE_CONTENTS,
                                GALObjectEID,
                                0,
                                NULL,
                                NULL,
                                &m_pCachedLocalGAL);
        }
        if (!hResult)
        {
            LPSRow pRow;
            LPSPropValue pOneProp;
            SPropValue spvInstKey = { 0 };
            spvInstKey.ulPropTag = PR_INSTANCE_KEY;
            spvInstKey.Value.bin.cb = sizeof(DWORD);
            spvInstKey.Value.bin.lpb = (LPBYTE)&(ObjectEID.dwObjID);

            hResult = m_pCachedLocalGAL->HrQueryRow (&spvInstKey, &pRow, NULL);
            if (!hResult)
            {
                for (ULONG k=0; k<pRow->cValues; k++)
                {
                    if (PR_INSTANCE_KEY != pRow->lpProps[k].ulPropTag &&
                        PR_OBJECT_TYPE != pRow->lpProps[k].ulPropTag &&
                        PR_ENTRYID != pRow->lpProps[k].ulPropTag &&
                        PR_ADDRTYPE != pRow->lpProps[k].ulPropTag)
                    {
                        (*ppList)->Insert (&(pRow->lpProps[k]));
                    }
                }
                // The transmitable display name is not in the table, but in our
                // implementation it is equal to the display name.
                pOneProp = PpropFindProp (pRow->lpProps, pRow->cValues, PR_DISPLAY_NAME);
                if (pOneProp)
                {
                    pOneProp->ulPropTag = PR_TRANSMITABLE_DISPLAY_NAME;
                    (*ppList)->Insert (pOneProp);
                }
                PrivMAPIFree ((LPVOID *)&pRow);
            }
        }
    }
    else // connect to the server a get the latest user properties
    {
        hResult = BindToServer (m_szServer);
        if (hResult)
        {
            TraceString1 ("CABServerBackEnd::OpenObject: Failed to bind to remote server %s", m_szServer);
            return S_OK;
        }
        RpcTryExcept
        {
            if (MAPI_MAILUSER == ObjectEID.bObject)
            {
                hResult = RemoteGetRecipientProps (ObjectEID.dwObjID,
                                                   &ulObjType,
                                                   (WINDS_RPC_STRING)MBInfo.szMailboxName,
                                                   (WINDS_RPC_STRING)MBInfo.szFullName,
                                                   (WINDS_RPC_STRING)MBInfo.szJobTitle,
                                                   (WINDS_RPC_STRING)MBInfo.szOffice,
                                                   (WINDS_RPC_STRING)MBInfo.szPhone,
                                                   (WINDS_RPC_STRING)MBInfo.szAltPhone,
                                                   (WINDS_RPC_STRING)MBInfo.szFax,
                                                   (WINDS_RPC_STRING)MBInfo.szComments,
                                                   (WINDS_RPC_STRING)MBInfo.szCompany,
                                                   (WINDS_RPC_STRING)MBInfo.szDepartment,
                                                   (WINDS_RPC_STRING)MBInfo.szManagerName,
                                                   (WINDS_RPC_STRING)MBInfo.szManagerAlias,
                                                   &MBInfo.dwManagerID);
                if (S_OK == hResult)
                {
                    spvProps[i].ulPropTag = PR_DISPLAY_TYPE;
                    if (SERVER_USER_MAILBOX == ulObjType)
                    {
                        spvProps[i++].Value.l = DT_MAILUSER;
                    }
                    else
                    {
                        ASSERT (GATEWAY_RECIPIENT == ulObjType);
                        spvProps[i++].Value.l = DT_REMOTE_MAILUSER;
                    }
                    spvProps[i].ulPropTag       = PR_ACCOUNT;
                    spvProps[i++].Value.lpszA   = MBInfo.szMailboxName;
                    spvProps[i].ulPropTag       = PR_DISPLAY_NAME;
                    spvProps[i++].Value.lpszA   = MBInfo.szFullName;
                    spvProps[i].ulPropTag       = PR_TRANSMITABLE_DISPLAY_NAME;
                    spvProps[i++].Value.lpszA   = MBInfo.szFullName;
                    spvProps[i].ulPropTag       = PR_TITLE;
                    spvProps[i++].Value.lpszA   = MBInfo.szJobTitle;
                    spvProps[i].ulPropTag       = PR_OFFICE_LOCATION;
                    spvProps[i++].Value.lpszA   = MBInfo.szOffice;
                    spvProps[i].ulPropTag       = PR_OFFICE_TELEPHONE_NUMBER;
                    spvProps[i++].Value.lpszA   = MBInfo.szPhone;
                    spvProps[i].ulPropTag       = PR_OTHER_TELEPHONE_NUMBER;
                    spvProps[i++].Value.lpszA   = MBInfo.szAltPhone;
                    spvProps[i].ulPropTag       = PR_PRIMARY_FAX_NUMBER;
                    spvProps[i++].Value.lpszA   = MBInfo.szFax;
                    spvProps[i].ulPropTag       = PR_COMMENT;
                    spvProps[i++].Value.lpszA   = MBInfo.szComments;
                    spvProps[i].ulPropTag       = PR_COMPANY_NAME;
                    spvProps[i++].Value.lpszA   = MBInfo.szCompany;
                    spvProps[i].ulPropTag       = PR_DEPARTMENT_NAME;
                    spvProps[i++].Value.lpszA   = MBInfo.szDepartment;
                    if (0 != MBInfo.dwManagerID)
                    {
                        spvProps[i].ulPropTag = PR_WINDS_AB_MANAGER_NAME;
                        spvProps[i++].Value.lpszA = MBInfo.szManagerName;

                        spvProps[i].ulPropTag       = PR_WINDS_AB_MANAGER_ENTRYID;
                        spvProps[i].Value.bin.cb    = CB_PRIVATE_EID;
                        spvProps[i++].Value.bin.lpb = (LPBYTE)&ManagerEID;
                    
                        ManagerEID.uidGlobal = m_uidBackEnd;
                        ManagerEID.bVersion = ENTRYID_VERSION;
                        ManagerEID.bObject = MAPI_MAILUSER;
                        ManagerEID.dwObjID = MBInfo.dwManagerID;
                    }
                    wsprintf (szEmailAddress, "%s\\%s", m_szServer, MBInfo.szMailboxName);
                }
            }
            if (MAPI_DISTLIST == ObjectEID.bObject)
            {
                hResult = RemoteGetDistListProps (ObjectEID.dwObjID,
                                                  (WINDS_RPC_STRING)DLInfo.szDLAlias,
                                                  (WINDS_RPC_STRING)DLInfo.szDLFullName,
                                                  &ulContentCount,
                                                  (WINDS_RPC_STRING)DLInfo.szOwnerAlias,
                                                  (WINDS_RPC_STRING)DLInfo.szOwnerName,
                                                  (WINDS_RPC_STRING)DLInfo.szComments);
                if (S_OK == hResult)
                {
                    spvProps[i].ulPropTag       = PR_DISPLAY_TYPE;
                    spvProps[i++].Value.l       = DT_DISTLIST;
                    spvProps[i].ulPropTag       = PR_ACCOUNT;
                    spvProps[i++].Value.lpszA   = DLInfo.szDLAlias;
                    spvProps[i].ulPropTag       = PR_DISPLAY_NAME;
                    spvProps[i++].Value.lpszA   = DLInfo.szDLFullName;
                    spvProps[i].ulPropTag       = PR_TRANSMITABLE_DISPLAY_NAME;
                    spvProps[i++].Value.lpszA   = DLInfo.szDLFullName;
                    spvProps[i].ulPropTag       = PR_CONTENT_COUNT;
                    spvProps[i++].Value.ul      = ulContentCount;
                    spvProps[i].ulPropTag       = PR_WINDS_AB_DL_OWNER_ACCT;
                    spvProps[i++].Value.lpszA   = DLInfo.szOwnerAlias;
                    spvProps[i].ulPropTag       = PR_WINDS_AB_DL_OWNER_NAME;
                    spvProps[i++].Value.lpszA   = DLInfo.szOwnerName;
                    spvProps[i].ulPropTag       = PR_COMMENT;
                    spvProps[i++].Value.lpszA   = DLInfo.szComments;

                    wsprintf (szEmailAddress, "%s\\%s", m_szServer, DLInfo.szDLAlias);
                }
            }
            if (S_OK == hResult && (MAPI_DISTLIST == ObjectEID.bObject || MAPI_MAILUSER == ObjectEID.bObject))
            {
                spvProps[i].ulPropTag       = PR_EMAIL_ADDRESS;
                spvProps[i++].Value.lpszA   = szEmailAddress;
                wsprintf (szSearchKey, "%s:%s", WINDS_ADDRESS_TYPE, szEmailAddress);
                CharUpper (szSearchKey);
                spvProps[i].ulPropTag       = PR_SEARCH_KEY;
                spvProps[i].Value.bin.cb    = Cbtszsize(szSearchKey);
                spvProps[i++].Value.bin.lpb = (LPBYTE)szSearchKey;
            }
        }
        RpcExcept(1)
        {        
            // If we got here is because there was an error while call was made
            // or when it was about to be made.
            hResult = RpcExceptionCode();
            TraceRPCError ("CABServerBackEnd::OpenObject: Remote call failed", hResult);
            if (RPC_S_SERVER_UNAVAILABLE == hResult)
            {
                hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
            }
            else
            {
                hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
            }
        }
        RpcEndExcept
        // Terminate connections with the remote server
        BindToServer (NULL);
    }
FinishNow:
    if (hResult)
    {
        delete *ppList;
        *ppList = NULL;
    }
    else
    {
        // If the connection was ON line to the server, we need to copy the
        // properties from the temporaty property array onto the property cache
        // we return to the caller.
        for (UINT j=0; j<i; j++)
        {
            (*ppList)->Insert (&spvProps[j]);
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::FillHierarchyTable()
//
//    Parameters
//      ContainerEID    Entry ID of the container we are building the
//                      hierarchy table for.
//      dwFlags         Flags for the table: 0 or CONVENIENT_DEPTH
//      pTable          ITableData allocated by the caller of the method
//
//    Purpose
//      Here we fill the rows of the hierarchy table of the given container,
//      and if CONVENIENT_DEPTH was requested, will build the ENTIRE tree of
//      subcontainer under the one indicated by the entry ID in ContainerEID.
//      For building hierarchy table this implementation of an address book
//      does not need to request any information from the server because the
//      hierarchy is fixed. However, if the provider is working offline, the
//      container for foreign recipients, (i.e server gateway directory) is not
//      available, although the recipients on those gateway containers
//      directories have been cached in the GAL contents table.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABServerBackEnd::FillHierarchyTable (PRIVATE_ENTRYID ContainerEID,
                                                   DWORD           dwFlags,
                                                   LPTABLEDATA     pTable)
{
    ASSERT (MAPI_ABCONT == ContainerEID.bObject);
    SPropValue spvProps[HIERARCHY_TABLE_PROPS] = { 0 };
    SRow row = { 0 };
    row.cValues = HIERARCHY_TABLE_PROPS;
    row.lpProps = spvProps;

    PRIVATE_ENTRYID eidSubContainer = { 0 };
    eidSubContainer.uidGlobal = m_uidBackEnd;
    eidSubContainer.bVersion = ENTRYID_VERSION;
    eidSubContainer.bObject = MAPI_ABCONT;
    
    DWORD dwInst = 0;
    TCHAR szContainerName[64];
    for (ULONG i=0; i<HIERARCHY_TABLE_PROPS; i++)
    {
        spvProps[i].ulPropTag = sptContainerHierProps.aulPropTag[i];
    }

    spvProps[HT_INST].Value.bin.cb = sizeof(DWORD);
    spvProps[HT_INST].Value.bin.lpb = (LPBYTE)&dwInst;
    spvProps[HT_DEPTH].Value.l = 0;
    spvProps[HT_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
    spvProps[HT_ENTRYID].Value.bin.lpb = (LPBYTE)&eidSubContainer;
    spvProps[HT_DISPNAME].Value.lpszA = szContainerName;
    spvProps[HT_OBJTYPE].Value.l = MAPI_ABCONT;
    spvProps[HT_AB_PROVID].Value.bin.cb = sizeof(MAPIUID);
    spvProps[HT_AB_PROVID].Value.bin.lpb = (LPBYTE)&guidABEntries;

    // If we are building the hierachy table for the root container list
    // the GAL and the Gateways directory container
    if (ROOT_CONTAINER_ID == ContainerEID.dwObjID)
    {
        // The GAL container
        dwInst++;
        wsprintf (szContainerName, GAL_NAME_FORMAT, m_szServer);
        eidSubContainer.dwObjID = GAL_CONTAINER_ID;
        spvProps[HT_CONTFLAGS].Value.l = AB_UNMODIFIABLE | AB_RECIPIENTS; // This container may have recipients
        spvProps[HT_DISPTYPE].Value.l = DT_GLOBAL;
        pTable->HrModifyRow (&row);

        // The Gateways directory
        dwInst++;
        wsprintf (szContainerName, GW_CONTAINERS_NAME_FORMAT, m_szServer);
        eidSubContainer.dwObjID = GATEWAY_CONTAINERS_ID;
        spvProps[HT_CONTFLAGS].Value.l = AB_UNMODIFIABLE | AB_SUBCONTAINERS; // This container HAS subcontainers
        spvProps[HT_DISPTYPE].Value.l = DT_NOT_SPECIFIC;    // This tell MAPI not to add this container
                                                            // on the search path when adding this address book to a profile.
        pTable->HrModifyRow (&row);

        spvProps[1].Value.l++; // In case the user set CONVENIENT_DEPTH, increment the level
    }

    // When the back end is working offline, the gateways directory is empty.
    // However the gateway recipients are already listed in the GAL saved
    // in the local address book directory.
    if (m_fUseLocalABD)
    {
        return S_OK;
    }

    if ((ROOT_CONTAINER_ID == ContainerEID.dwObjID && CONVENIENT_DEPTH & dwFlags) ||
        GATEWAY_CONTAINERS_ID == ContainerEID.dwObjID)
    {
        // Gateways have a different display type and they have recipients
        spvProps[HT_DISPTYPE].Value.l = DT_WAN;
        spvProps[HT_CONTFLAGS].Value.l = AB_UNMODIFIABLE | AB_RECIPIENTS;

        dwInst++;
        lstrcpy (szContainerName, EXCHANGE_GW_NAME);
        eidSubContainer.dwObjID = EXCHANGE_CONTAINER_ID;
        pTable->HrModifyRow (&row);

        dwInst++;
        lstrcpy (szContainerName, SMTP_GW_NAME);
        eidSubContainer.dwObjID = SMTP_CONTAINER_ID;
        pTable->HrModifyRow (&row);

        dwInst++;
        lstrcpy (szContainerName, FAX_GW_NAME);
        eidSubContainer.dwObjID = FAX_CONTAINER_ID;
        pTable->HrModifyRow (&row);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CABServerBackEnd::FillContentsTable()
//
//    Parameters
//      ContainerEID    Entry ID of the container we are building the
//                      contents table for.
//      pTable          ITableData allocated by the caller of the method
//
//    Purpose
//      Here we fill the rows of the hierachy table of the given container.
//      The data for the rows is comming from one of two source:
//          1. The server if we are online.
//          2. The local address book directory if the address book is offline.
//      The columns in the rows added will have at least the default columns
//      in the tag array specified when the ITableData was created.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CABServerBackEnd::FillContentsTable (PRIVATE_ENTRYID ContainerEID,
                                                  LPTABLEDATA     pTable)
{
    HRESULT hResult = S_OK;
    // These guys don't have contents tables
    ASSERT (ROOT_CONTAINER_ID != ContainerEID.dwObjID && GATEWAY_CONTAINERS_ID != ContainerEID.dwObjID);
    
    SPropValue spvProps[CONTENTS_TABLE_PROPS] = { 0 };
    SRow row = { 0 };
    row.cValues = CONTENTS_TABLE_PROPS;
    row.lpProps = spvProps;

    PRIVATE_ENTRYID eidObject = { 0 };
    eidObject.uidGlobal = m_uidBackEnd;
    eidObject.bVersion = ENTRYID_VERSION;
    
    DWORD dwInstKey;
    AB_ENTRY_INFO abEntry = { 0 };

    for (ULONG i=0; i<CONTENTS_TABLE_PROPS; i++)
    {
        spvProps[i].ulPropTag = sptContainerContProps.aulPropTag[i];
    }

    long lPipeNum;
    TCHAR szPipeName[64], szEmailAddress[64], szSearchKey[128];
    HANDLE hSource = NULL;
    DWORD dwBytesRead;
    BOOL fBadEntry;

    spvProps[CONT_INST_KEY].Value.bin.cb = sizeof(DWORD);
    spvProps[CONT_INST_KEY].Value.bin.lpb = (LPBYTE)&dwInstKey;
    spvProps[CONT_ADDRTYPE].Value.lpszA = WINDS_ADDRESS_TYPE;
    spvProps[CONT_ENTRYID].Value.bin.cb = CB_PRIVATE_EID;
    spvProps[CONT_ENTRYID].Value.bin.lpb = (LPBYTE)&eidObject;
    spvProps[CONT_RTF].Value.b = TRUE;
    spvProps[CONT_EMAILADDR].Value.lpszA = szEmailAddress;
    spvProps[CONT_SKEY].Value.bin.lpb = (LPBYTE)szSearchKey;

    // Offline. Use the data file
    if (m_fUseLocalABD)
    {
        ASSERTMSG (GAL_CONTAINER_ID == ContainerEID.dwObjID, "We only cached the GAL in the Local ABD");
        hSource = CreateFile (m_szABDLocalFile,
                              GENERIC_READ,
                              FILE_SHARE_READ | FILE_SHARE_WRITE,
                              NULL,
                              OPEN_ALWAYS,
                              FILE_ATTRIBUTE_NORMAL,
                              NULL);
        if (INVALID_HANDLE_VALUE == hSource)
        {
            hResult = HRESULT_FROM_WIN32(GetLastError());
        }
    }
    else
    {
        // Online. The data is coming down a pipe from the server
        hResult = BindToServer (m_szServer);
        if (hResult)
        {
            TraceString1 ("CABServerBackEnd::FillContentsTable: Failed to bind to remote server %s", m_szServer);
            return S_OK;
        }
        RpcTryExcept
        {
            // Make the remote call to request the server to open a pipe for the download
            hResult = RemoteGetContainerRecipients (0, ContainerEID.dwObjID, &lPipeNum);
            if (!hResult)
            {
                // Construct the download pipe name
                wsprintf (szPipeName, PIPE_NAME_FORMAT, m_szServer, lPipeNum);
                // Create our endpoint and connect    
                hSource = CreateFile (szPipeName, 
                                      GENERIC_READ,
                                      0,
                                      NULL,
                                      OPEN_EXISTING,
                                      0,
                                      NULL);
                if (INVALID_HANDLE_VALUE == hSource)
                {
                    hResult = HRESULT_FROM_WIN32(GetLastError());
                }
            }
        }
        RpcExcept(1)
        {        
            // If we got here is because there was an error while call was made
            // or when it was about to be made.
            hResult = RpcExceptionCode();
            TraceRPCError ("CABServerBackEnd::FillContentsTable: Remote call failed", hResult);
            if (RPC_S_SERVER_UNAVAILABLE == hResult)
            {
                hResult = HRESULT_FROM_WIN32(ERROR_HOST_UNREACHABLE);
            }
            else
            {
                hResult = MAKE_HRESULT(1, FACILITY_RPC, hResult);
            }
        }
        RpcEndExcept
        // Terminate connections with the remote server
        BindToServer (NULL);
    }
    if (!hResult)
    {
        do
        {
            // Read the pipe handle.
            if (!ReadFile (hSource, &abEntry, sizeof(AB_ENTRY_INFO), &dwBytesRead, NULL))
            {
                hResult = HRESULT_FROM_WIN32(GetLastError());
                if (HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE)   != hResult &&   // For Windows NT
                    HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED) != hResult)     // For Windows 95
                {
                    // There was an error and we can't continue
                    TraceResult ("CABServerBackEnd::FillContentsTable: Failed to read from the source", hResult);
                }
                else
                {
                    // If the pipe was broken, it means the server finished writing
                    // to the it, so we are finished reading from it.
                    hResult = 0;
                }
            }
            else
            {
                fBadEntry = FALSE;
                switch (abEntry.Type)
                {
                    case GATEWAY_RECIPIENT :
                    case SERVER_USER_MAILBOX :
                        if (HIDE_IN_DIR & abEntry.Info.MB.dwFlags)
                        {
                            // We should not list the entry in the contents table
                            fBadEntry = TRUE;
                            continue;
                        }
                        spvProps[CONT_OBJTYPE].Value.l = MAPI_MAILUSER;
                        spvProps[CONT_DISPTYPE].Value.l = (abEntry.Type == SERVER_USER_MAILBOX ? DT_MAILUSER : DT_REMOTE_MAILUSER);
                        spvProps[CONT_ACCOUNT].Value.lpszA = abEntry.Info.MB.szMailboxName;
                        wsprintf (szEmailAddress, "%s\\%s", m_szServer, abEntry.Info.MB.szMailboxName);
                        spvProps[CONT_DISPNAME].Value.lpszA = abEntry.Info.MB.szFullName;
                        spvProps[CONT_TITLE].Value.lpszA = abEntry.Info.MB.szJobTitle;
                        spvProps[CONT_OFFICE_LOC].Value.lpszA = abEntry.Info.MB.szOffice;
                        spvProps[CONT_OFFICE_PHONE].Value.lpszA = abEntry.Info.MB.szPhone;
                        spvProps[CONT_OTHER_PHONE].Value.lpszA = abEntry.Info.MB.szAltPhone;
                        spvProps[CONT_FAX].Value.lpszA = abEntry.Info.MB.szFax;
                        spvProps[CONT_COMMENT].Value.lpszA = abEntry.Info.MB.szComments;
                        spvProps[CONT_COMPANY].Value.lpszA = abEntry.Info.MB.szCompany;
                        spvProps[CONT_DEPT].Value.lpszA = abEntry.Info.MB.szDepartment;
                        spvProps[CONT_MANAGER].Value.lpszA = abEntry.Info.MB.szManagerName;
                        eidObject.dwObjID = abEntry.Info.MB.dwObjID;
                        break;
                    case SERVER_DISTRIBUTION_LIST :
                        if (HIDE_IN_DIR & abEntry.Info.DL.dwFlags)
                        {
                            // We should not list the entry in the contents table
                            fBadEntry = TRUE;
                            continue;
                        }
                        spvProps[CONT_DISPTYPE].Value.l = DT_DISTLIST;
                        spvProps[CONT_OBJTYPE].Value.l = MAPI_DISTLIST;
                        spvProps[CONT_ACCOUNT].Value.lpszA = abEntry.Info.DL.szDLAlias;
                        wsprintf (szEmailAddress, "%s\\%s", m_szServer, abEntry.Info.DL.szDLAlias);
                        spvProps[CONT_DISPNAME].Value.lpszA = abEntry.Info.DL.szDLFullName;
                        spvProps[CONT_TITLE].Value.lpszA = szBlank;
                        spvProps[CONT_OFFICE_LOC].Value.lpszA = szBlank;
                        spvProps[CONT_OFFICE_PHONE].Value.lpszA = szBlank;
                        spvProps[CONT_OTHER_PHONE].Value.lpszA = szBlank;
                        spvProps[CONT_FAX].Value.lpszA = szBlank;
                        spvProps[CONT_COMMENT].Value.lpszA = szBlank;
                        eidObject.dwObjID = abEntry.Info.DL.dwObjID;
                        break;
                    default :
                        fBadEntry = TRUE;
                        break;
                }
                if (!fBadEntry)
                {
                    wsprintf (szSearchKey, "%s:%s", WINDS_ADDRESS_TYPE, szEmailAddress);
                    CharUpper (szSearchKey);
                    spvProps[CONT_SKEY].Value.bin.cb = Cbtszsize(szSearchKey);
                    
                    dwInstKey = eidObject.dwObjID;
                    eidObject.bObject = (BYTE)spvProps[CONT_OBJTYPE].Value.l;
                    pTable->HrModifyRow (&row);
                }
            }
        } while (dwBytesRead && !hResult);
        // If we are working off line, save the GAL table in the object so that we avoid reloading it next time we are asked for it.
        if (m_fUseLocalABD && !m_pCachedLocalGAL && (GAL_CONTAINER_ID == ContainerEID.dwObjID))
        {
            m_pCachedLocalGAL = pTable;
            m_pCachedLocalGAL->AddRef();
        }
    }
    CloseHandle (hSource);
    return hResult;
}

// End of file for ABSRVBE.CPP
