///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      XPPROV.CPP
//
//  Description
//      This file implements the IXPProvider interface with the methods 
//      specified in the MAPI SPI 1.0 specifications. Also some helper
//      functions are implemented in this file.
//
//  Author
//      Irving De la Cruz
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "WDS.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceInfoMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

///////////////////////////////////////////////////////////////////////////////
//    CXPProvider::CXPProvider()
//
//    Parameters
//      hInst                   Handle to instance of this XP DLL
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CXPProvider::CXPProvider (HINSTANCE hInst)
{
    InfoTrace ("CXPProvider: Constructor called");
    m_hInstance    = hInst;
    m_cRef         = 1;
    // Initialize critical sections for this transport
    InitializeCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CXPProvider::~CXPProvider()
//
//    Parameters
//      None
//
//    Purpose
//      Close down and release resources and libraries
//
//    Return Value
//      None
//
CXPProvider::~CXPProvider()
{
    InfoTrace ("CXPProvider: Destructor called");
    m_hInstance = NULL;

    // This is the last method called on a transport, close down the traces
    UnInitTraces();
    DeleteCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CXPProvider::QueryInterface()
//
//    Parameters
//      { Refer to OLE Documentation on this method }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it 
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPProvider::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If this is one of the two IID return an interface pointer to it
    if (riid == IID_IXPProvider || riid == IID_IUnknown)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
        return S_OK;
    }
    // This object does not support the interface requested
    return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////////////////
// IXPProvider virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CXPProvider::Shutdown()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPProvider::Shutdown (ULONG * pulFlags)
{
    InfoTrace ("CXPProvider::Shutdown method called");
    CheckParameters (IXPProvider_Shutdown, pulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPProvider::TransportLogon()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Display the logon dialog to show the options saved in the profile for
//      this provider and allow changes to it. Save new configuration settings
//      back in the profile.
//      Create a new CXPLogon object and return it to the spooler. Also, 
//      initialize the properties array for each address type handled
//      by this transport. Check all the flags and return them to the spooler
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPProvider::TransportLogon (LPMAPISUP     pSupObj,
                                          ULONG         ulUIParam,
                                          LPTSTR        pszProfileName,
                                          ULONG *       pulFlags,
                                          LPMAPIERROR * ppMAPIError,
                                          LPXPLOGON *   ppXPLogon)
{
    InfoTrace ("CXPProvider::TransportLogon method called");
    CheckParameters (IXPProvider_TransportLogon, pSupObj);
    
    CXPLogon * pXPLogon = NULL;
    ULONG ulPropCount;
    LPSPropValue pProps = NULL;
    MAILBOX_INFO UserMBInfo = { 0 };
    CFGDLG CfgDialog = { 0 };
    LPPROFSECT pProfileObj;
    DWORD dwConnectionID;
    HRESULT hResult = OpenServiceProfileSection (pSupObj, &pProfileObj, GetMAPIFreeBuffer());
    if (hResult)
    {
        TraceResult ("CXPProvider::TransportLogon: Failed to open the service profile section", hResult);
        return hResult;
    }

    HANDLE hUIMutex = CreateMutex (NULL, FALSE, CONFIG_UI_MUTEX);
    if (NULL == hUIMutex)
    {
        HRESULT hResult = HRESULT_FROM_WIN32 (GetLastError());
        TraceResult ("CXPProvider::TransportLogon: Failed to create UI mutext", hResult);
    }
    
    hResult = pProfileObj->GetProps ((LPSPropTagArray)&sptLogonProps,
                                     fMapiUnicode,
                                     &ulPropCount,
                                     &pProps);
    if (FAILED(hResult))
    {
        TraceResult ("CXPProvider::TransportLogon: Failed to get the logon props", hResult);
        goto ErrorExit;
    }

    // Fill in the logon UI structure
    CfgDialog.hInst    = m_hInstance;
    CfgDialog.hWnd     = (HWND)ulUIParam;
    CfgDialog.ppProps  = &pProps;
    CfgDialog.pSupObj  = pSupObj;
    CfgDialog.hUIMutex = hUIMutex;

    // In case we get MAPI_W_ERRORS_RETURNED, ignore it and reset 
    // to S_OK. Now display the logon configuration dialog
    if (MAPI_W_ERRORS_RETURNED == hResult)
    {
        if (PR_SMP_MAILBOX_ID      != pProps[MAILBOX_ID].ulPropTag  ||
            PR_SMP_CONNECTION_TYPE != pProps[NET_CON].ulPropTag)
        {
            TraceMessage ("CXPProvider::TransportLogon: Where are the provider properties?");
            hResult = MAPI_E_UNCONFIGURED;
            if (!(LOGON_NO_DIALOG & *pulFlags))
            {
                PrivInitialize3DCtl (m_hInstance);
                PrivateMessageBox (IDS_MSG_SERVICE_NOT_PROPERLY_CFG, (HWND)ulUIParam);
                PrivUninitialize3DCtl (m_hInstance);
            }
            goto ErrorExit;
        }

        if (PR_SMP_REMOTE_SERVER != pProps[SERVER_NAME].ulPropTag ||
            PR_SMP_MAILBOX_NAME != pProps[MAILBOX_NAME].ulPropTag ||
            PR_SMP_USER_NAME != pProps[USER_NAME].ulPropTag ||
            PR_SMP_MAILBOX_PASSWORD != pProps[PASSWORD].ulPropTag)
        {
            if (LOGON_NO_DIALOG & *pulFlags)
            {
                TraceMessage ("CXPProvider::TransportLogon: UI not allowed but it is needed");
                hResult = MAPI_E_UNCONFIGURED;
                goto ErrorExit;
            }
            PrivInitialize3DCtl (m_hInstance);
ReStartLogonDlg :
            hResult = DoLogonDlg (&CfgDialog);
            if (hResult)
            {
                goto ErrorExit;
            }
            hResult = pProfileObj->SetProps (4, pProps, NULL);
            if (hResult)
            {
                TraceResult ("CXPProvider::TransportLogon: Failed to set the properties", hResult);
                goto ErrorExit;
            }
        }
    }
    ASSERTMSG (PR_SMP_CONNECTION_TYPE == pProps[NET_CON].ulPropTag, "Where is the connection information for the logon?");
    BOOL fRemoteXP;
    fRemoteXP = (OFFLINE_CONNECT == pProps[NET_CON].Value.l);
    // Connect to the server with the stored information. If necessary (and allowed) display UI.
    hResult = DoServerLogon (&UserMBInfo,
                             pProps,
                             pProfileObj,
                             !(LOGON_NO_DIALOG & *pulFlags),
                             (HWND)ulUIParam,
                             !fRemoteXP,
                             (fRemoteXP ? 0 : WINDS_NOTIF_ON_XP | WINDS_NOTIF_ON_USER),
                             (fRemoteXP ? NULL : &dwConnectionID),
                             fRemoteXP);
    if (S_FALSE == hResult)
    {
        goto ReStartLogonDlg;
    }
        
    if (OFFLINE_CONNECT == pProps[NET_CON].Value.l ||
        MAPI_W_NO_SERVICE == hResult ||      // This is the error DoServerLogon() returns in the CLIENT context
        MAPI_E_NETWORK_ERROR == hResult)    // This is the error DoServerLogon() returns in the SPOOLER context
    {
        hResult = S_OK;
        pProps[NET_CON].Value.l = OFFLINE_CONNECT;
        // To avoid create a mailslot notification sink in the CXPLogon constructor
        dwConnectionID = 0;
    }
    if (hResult)
    {
        goto ErrorExit;
    }

    try
    {
        // Allocate the IXPLogon-derived object. Initialize its data members with the necessary values
        pXPLogon = new CXPLogon (m_hInstance,
                                 pSupObj,
                                 pProps[HEADER_FILE].Value.lpszA,
                                 pProps[GET_HEADERS].Value.b,
                                 &UserMBInfo,
                                 pProps[MAILBOX_ID].Value.l,
                                 pProps[SERVER_NAME].Value.lpszA,
                                 hUIMutex,
                                 dwConnectionID,
                                 (CONNECTION_TYPE)pProps[NET_CON].Value.l);
        if (!pXPLogon)
        {
            // New failed, propably due to memory shortage
            TraceMessage ("CXPProvider::TransportLogon: Failed to allocate new CXPLogon object");
            hResult = E_OUTOFMEMORY;
        }
    }
    catch (CException & Exception)
    {
        hResult = Exception.GetError();
    }
    if (hResult)
    {
        goto ErrorExit;
    }
    // To avoid closing this in the cleanup code
    hUIMutex = NULL;

    // Lock other threads because we are going to work on data that is 
    // global to all threads (sessions) using this transport
    EnterCriticalSection (&m_csObj);
    // Check the mode in which MAPI is logging into our transport and set
    // the appropiate internal (this CXPLogon object) transport session flags
    pXPLogon->InitializeTransportStatusFlags (*pulFlags);
    // Initialize the transport ID prop array for this session
    hResult = pXPLogon->SetIdentityProps();
    if (!hResult)
    {
        // Build the transport status row for this session
        hResult = pXPLogon->InitializeStatusRow();
        if (!hResult)
        {
            // Set the session flags returned to MAPI by the transport
            pXPLogon->SetSessionFlags (pulFlags);
            // Copy our allocated object back to the returned MAPI object pointer
            *ppXPLogon = (LPXPLOGON)pXPLogon;
        }
    }
    // Release the critical section
    LeaveCriticalSection (&m_csObj);


ErrorExit:
    // Release pProfileObj. This object was AddRef() by MAPI before it 
    // came to us. The transport is in charge of cleaning it up (Release()ing it).
    if (pProfileObj)
    {
        pProfileObj->Release();
    }
    PrivUninitialize3DCtl (m_hInstance);

    if (hUIMutex)
    {
        CloseHandle (hUIMutex);
    }
    PrivMAPIFree ((LPVOID *)&pProps);

    if (hResult)
    {   
        // Something failed, so clean the session if it has been allocated.
        // The Release() method of CXPLogon does the memory deallocation 
        // of this object. The destructor of CXPLogon does the rest of 
        // the clean up
        if (pXPLogon)
        {
            pXPLogon->Release();
        }
    }
    return hResult;
}                                        

///////////////////////////////////////////////////////////////////////////////
//    XPMailslotListenThreadProc()
//
//    Parameters
//      pLogon      Pointer to the XP logon session the thread is
//                  listening for
//
//    Purpose
//      This is function runs in a separate low priority thread listening for
//      notification that a WINDS server send to client provider. When a
//      notification of interest is received, an appropiate action is taken.
//      The notification arrive on a mailslot created by the logon object.
//
//    Return Value
//      S_OK always.
//
DWORD WINAPI XPMailslotListenThreadProc (CXPLogon * pLogon)
{
    HANDLE hMailslot = pLogon->GetListenMailslot();
    DWORD dwRead;
    FILETIME ftLastNotifTime = { 0 };
    WINDS_NOTIF_EVENT LastEvent = GENERAL_NOTIF_MIN;
    WINDS_NOTIFICATION Notif;
    HRESULT hReadError;
    #pragma warning (disable : 4127)
    while (TRUE)
    {
    #pragma warning (default : 4127)
        if (!ReadFile (hMailslot, &Notif, sizeof(WINDS_NOTIFICATION), &dwRead, NULL)) 
        { 
            hReadError = HRESULT_FROM_WIN32 (GetLastError());
            if (HRESULT_FROM_WIN32(ERROR_INVALID_HANDLE) == hReadError ||
                HRESULT_FROM_WIN32(ERROR_HANDLE_EOF) == hReadError)
            {
                break; // out of the WHILE() loop and terminate the thread
            }
            TraceResult ("XPMailslotListenThreadProc: ReadFile", HRESULT_FROM_WIN32(GetLastError()));
            continue;
        }
        // Mailslot are received in each network protocol stack installed on the system.
        // Check the time stamp and event of the last notification received. If it
        // is the same, drop it and continue listen for new and different incoming
        // data on the mailslot.
        if (ftLastNotifTime.dwLowDateTime  == Notif.ftEventTime.dwLowDateTime &&
	        ftLastNotifTime.dwHighDateTime == Notif.ftEventTime.dwHighDateTime &&
            LastEvent                      == Notif.Event)
        {
            continue;
        }
        // Save the last event information
        ftLastNotifTime = Notif.ftEventTime;
        LastEvent = Notif.Event;

        #ifdef DO_INFO_TRACES
        switch (Notif.Event)
        {
            case XP_NEW_MAIL_ARRIVED :
                InfoTrace ("XP_NEW_MAIL_ARRIVED notification received from the WINDS server");
                break;
            case SERVER_IS_SHUTTING_DOWN :
                InfoTrace ("SERVER_IS_SHUTTING_DOWN notification received from the WINDS server");
                break;
            case RESET_LINKS_WITH_SERVER :
                InfoTrace ("RESET_LINKS_WITH_SERVER notification received from the WINDS server");
                break;
            case SERVER_HAS_RESTARTED :
                InfoTrace ("SERVER_HAS_RESTARTED notification received from the WINDS server");
                break;
            default :
                InfoTrace ("UNWANTED notification received from the WINDS server");
                break;
        }
        #endif // DO_INFO_TRACES

        __try
        {
            switch (Notif.Event)
            {
                case XP_NEW_MAIL_ARRIVED :
                    pLogon->m_pSupObj->SpoolerNotify (NOTIFY_NEWMAIL, NULL);
                    break;
                case SERVER_IS_SHUTTING_DOWN :
                    pLogon->SetConnectionType (CONNECTION_BROKEN);
                    pLogon->AddStatusBits (STATUS_OFFLINE);
                    pLogon->UpdateStatus();
                    pLogon->SetTransportState (SERVER_UNAVAILABLE);
                    break;
                case SERVER_HAS_RESTARTED :
                    pLogon->SetConnectionType (LAN_CONNECT);
                    pLogon->RemoveStatusBits (STATUS_OFFLINE);
                    pLogon->UpdateStatus();
                    pLogon->SetTransportState (READY);
                    break;
                case RESET_LINKS_WITH_SERVER :
                    {
                        pLogon->ResetServerNotifications();
                        DWORD dwConnectionID = pLogon->GetConnectionID();
                        char szMailslotName[64];
                        wsprintf (szMailslotName,
                                  CLIENT_MAILSLOT_SINK_NAME_FORMAT,
                                  XP_WINDS_NOTIFICATION_MAILSLOT,
                                  dwConnectionID);
                        hMailslot = CreateMailslot (szMailslotName, 0, MAILSLOT_WAIT_FOREVER, NULL);
                        if (INVALID_HANDLE_VALUE == hMailslot)
                        {
                            TraceResult ("XPMailslotListenThreadProc: Failed to create notification mailslot", HRESULT_FROM_WIN32(GetLastError()));
                        }
                        else
                        {
                            pLogon->SetListenMailslot (hMailslot);
                        }
                    }
                    break;
            }
        }
        __except (EXCEPTION_EXECUTE_HANDLER)
        {
            continue;
        }
    } 
    TraceMessage ("XPMailslotListenThreadProc: Thread is terminating");
    return S_OK;
}

// End of File for XPPROV.CPP
