///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      XPFOLDER.CPP
//
//  Description
//      This file implements the IMAPIFolder interface with the methods 
//      specified in the MAPI SPI 1.0.
//
//  Authors
//      Irving De la Cruz
//      Les Thaler
//
//  Revision: 1.8
//
// Written for Microsoft Corporation
// Copyright (c) 1995-1996 Microsoft Corporation. All rights reserved.
//
#include "XPWDSR.H"

// Remark this line to turn verbose tracing OFF
//#define DO_INFO_TRACES
#ifdef DO_INFO_TRACES
#define InfoTrace(a)   TraceInfoMessage(a)
#else
#define InfoTrace(a)   
#endif // DO_INFO_TRACES

// Available properties in the folder object 
#define NUM_REMOTE_FOLDER_PROPS     12
static const SizedSPropTagArray (NUM_REMOTE_FOLDER_PROPS, sptRemoteFolderProps) =
{
    NUM_REMOTE_FOLDER_PROPS,
    {
        PR_ACCESS,
        PR_ACCESS_LEVEL,
        PR_CONTENT_COUNT,
        PR_ASSOC_CONTENT_COUNT,
        PR_FOLDER_TYPE,
        PR_OBJECT_TYPE,
        PR_SUBFOLDERS,
        PR_CREATION_VERSION,
        PR_CURRENT_VERSION,
        PR_CREATION_TIME,
        PR_DISPLAY_NAME,
        PR_DISPLAY_TYPE
    }
};

// Indexes of the properties in each row in the contents table of the remote mailbag folder.
enum
{
    HDR_EID,                // Array Indexes
    HDR_SNDR,
    HDR_REPR,
    HDR_DISP_TO,
    HDR_SUBJ,
    HDR_CLASS,
    HDR_FLAGS,
    HDR_SIZE,
    HDR_PRIOR,
    HDR_IMPORTANCE,
    HDR_SENS,
    HDR_TIME,
    HDR_STAT,
    HDR_DOWNLOAD_TIME,
    HDR_HASATTACH,
    HDR_OBJ_TYPE,
    HDR_INST_KEY,
    HDR_NORM_SUBJ,
    NUM_HDR_PROPS           // Array size
};

static const SizedSPropTagArray(NUM_HDR_PROPS, sptRemoteFldContents) =
{
    NUM_HDR_PROPS,
    {
        PR_ENTRYID,                 // Index column in the contents table
        PR_SENDER_NAME,
        PR_SENT_REPRESENTING_NAME,
        PR_DISPLAY_TO,
        PR_SUBJECT,
        PR_MESSAGE_CLASS,
        PR_MESSAGE_FLAGS,
        PR_MESSAGE_SIZE,
        PR_PRIORITY,
        PR_IMPORTANCE,
        PR_SENSITIVITY,
        PR_MESSAGE_DELIVERY_TIME,
        PR_MSG_STATUS,
        PR_MESSAGE_DOWNLOAD_TIME,
        PR_HASATTACH,
        PR_OBJECT_TYPE,
        PR_INSTANCE_KEY,
        PR_NORMALIZED_SUBJECT,
    }
};

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CXPMAPIFolder()
//
//    Parameters
//      pStatusObj      Pointer to parent status object
//      pLogonObj       Pointer to logon session
//
//    Purpose
//      Constructor of the object. Parameters are passed to initialize the
//      data members with the appropiate values.
//
//    Return Value
//      None
//
CXPMAPIFolder::CXPMAPIFolder (CXPMAPIStatus * pStatusObj,
                              CXPLogon *      pLogonObj)
{
    InfoTrace ("CXPMAPIFolder: Constructor called");
    m_cRef = 1;
    m_pTableData = NULL;
    m_pLogon = pLogonObj;
    m_pLogon->AddRef();
    m_pStatusObj = pStatusObj;
    m_fNoHeadersFile = FALSE;
    InitializeCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::~CXPMAPIFolder()
//
//    Parameters
//      None
//
//    Purpose
//      Destructor of CXPMAPIFolder. Releases memory allocated for internal 
//      properties during the life of this transport logon object.
//
//    Return Value
//      None
//
CXPMAPIFolder::~CXPMAPIFolder ()
{
    InfoTrace ("CXPMAPIFolder: Destructor called");
    if (m_pTableData)
    {
        m_pTableData->Release();
        m_pTableData = NULL;
    }
    m_pStatusObj = NULL;
    m_pLogon->Release();
    m_pLogon = NULL;
    m_fNoHeadersFile = FALSE;
    DeleteCriticalSection (&m_csObj);
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::QueryInterface()
//
//    Parameters
//      { Refer to OLE Documentation on this method }
//
//    Purpose
//      Returns a pointer to a interface requested if the interface is 
//      supported and implemented by this object. If it is not supported, it
//      returns NULL
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::QueryInterface (REFIID riid, LPVOID * ppvObj)
{
    HRESULT hResult = S_OK;
    // OLE requires NULLing parameter
    *ppvObj = NULL;
    // If this is one of the two IID return an interface pointer to it
    if (riid == IID_IMAPIFolder || riid == IID_IMAPIContainer || riid == IID_IMAPIProp)
    {
        *ppvObj = (LPVOID)this;
        // Increase usage count of this object
        AddRef();
    }
    else
    {
        // OLE's COM rules for IUnknown establish that if an interface B
        // was obtained using QueryInterface() on interface A, querying B for
        // its IUnknown IID, should return a pointer to the A interface.
        // Since this IMAPIFolder interface was obtained from IMAPIStatus::QueryInterface,
        // when requested for IUnknown, return the contained pointer to the parent status
        // object
        if (riid == IID_IUnknown)
        {
            EnterCriticalSection (&m_csObj);
            *ppvObj = (LPVOID)m_pStatusObj;
            // Increase usage count of this object
            m_pStatusObj->AddRef();
            LeaveCriticalSection (&m_csObj);
        }
        else
        {
            hResult = E_NOINTERFACE;
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::Release()
//
//    Parameters
//      None
//
//    Purpose
//      Decrements the usage count of the Folder object.
//
//    Return Value
//      Usage count of this object.
//
STDMETHODIMP_(ULONG) CXPMAPIFolder::Release()
{
    EnterCriticalSection (&m_csObj);
    ULONG ulCount = --m_cRef;
    if (!ulCount)
    {
        delete this;
    }
    // If the count is 1, is only the internal reference kept in the parent
    // status object. This is the time to flush the contents table to
    // the headers file.
    if (1 == ulCount)
    {
        CopyTableToFile();
    }
    LeaveCriticalSection (&m_csObj);
    return ulCount;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIProp virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetLastError()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      This function finds a string for the hError passed. The hError is the
//      last error that occurred in one of our methods. Here we call a
//      function to allocate a MAPIERROR structure and passed the information
//      to the caller.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::GetLastError (HRESULT       hError,
                                          ULONG         ulFlags,
                                          LPMAPIERROR * ppMAPIError)
{
    InfoTrace ("CXPMAPIFolder::GetLastError method called");
    Validate_IMAPIProp_GetLastError (this, hError, ulFlags, ppMAPIError);

    // Allocate and fill a MAPIERROR structure with the error passed in.
    // Use the common helper function GetMAPIError
    return GetMAPIError (ppMAPIError, ulFlags, hError, m_pLogon->GetInstance());
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SaveChanges()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      We don't implement this. Besides, folder objects are not transacted.
//
//    Return Value
//      S_OK always.
//
STDMETHODIMP CXPMAPIFolder::SaveChanges (ULONG ulFlags)
{
    InfoTrace ("CXPMAPIFolder::SaveChanges method called");
    Validate_IMAPIProp_SaveChanges (this, ulFlags);
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Gets the properties of the remote message folder container. A client
//      requests the properties it needs by passing an array of property tags
//      in pTags. If the array pointer is NULL, we return ALL the properties
//      we have.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::GetProps (LPSPropTagArray pTags,
                                      ULONG           ulFlags,
                                      ULONG *         pcValues,
                                      LPSPropValue *  ppPropArray)
{
    InfoTrace ("CXPMAPIFolder::GetProps method called");
    Validate_IMAPIProp_GetProps (this, pTags, ulFlags, pcValues, ppPropArray);

    #ifdef DO_INFO_TRACES
    if (!pTags)
    {
        TraceInfoMessage ("CXPMAPIFolder::GetProps: All properties requested");
    }
    else
    {
        TraceMessage ("CXPMAPIFolder::GetProps Properties requested...");
        for (ULONG j=0; j<pTags->cValues; j++) { TraceProp (pTags->aulPropTag[j]); }
    }
    #endif // DO_INFO_TRACES

    BOOL fGetAllProps = pTags ? FALSE : TRUE;
    if (!pTags)
	{
		pTags = (LPSPropTagArray)&sptRemoteFolderProps;
	}
    *pcValues = pTags->cValues;
	HRESULT hResult = gpfnAllocateBuffer (sizeof(SPropValue)*(*pcValues), (LPVOID *)ppPropArray);
	if (hResult)
	{
        TraceResult ("CXPMAPIFolder::GetProps: Memory allocation failed", hResult);
		return hResult;
	}
    if (fGetAllProps)
    {
        for (ULONG i=0; i<sptRemoteFolderProps.cValues; i++)
        {
            (*ppPropArray)[i].ulPropTag = sptRemoteFolderProps.aulPropTag[i];
        }
    }
    else
    {
        // A client of the IMAPIProp::GetProps() method can pass PT_UNSPECIFIED in the
        // tag of a property. Is the provider's responsiblity to return
        // the proper type of the property.
        for (ULONG k=0; k<pTags->cValues; k++)
        {
            for (ULONG i=0; i<sptRemoteFolderProps.cValues; i++)
            {
                if (PROP_ID(pTags->aulPropTag[k]) == PROP_ID(sptRemoteFolderProps.aulPropTag[i]))
                {
                    (*ppPropArray)[k].ulPropTag = sptRemoteFolderProps.aulPropTag[i];
                }
                else
                {
                    (*ppPropArray)[k].ulPropTag = pTags->aulPropTag[k];
                }
            }
        }
    }

    EnterCriticalSection (&m_csObj);
    char szBuffer[64];
    BOOL fGotErrors = FALSE;    
	LPVOID pSrc = NULL, pDst;
    ULONG cbSize = 0;
    for (ULONG i=0; i<*pcValues; i++)
	{
        hResult = S_OK;
        switch ((*ppPropArray)[i].ulPropTag)
		{
            case PR_NULL :
                (*ppPropArray)[i].Value.err = S_OK;
                break;
		    case PR_CREATION_TIME :
                if (m_fNoHeadersFile)
                {
                    hResult = MAPI_E_NOT_FOUND;
                }
                else
                {
                    (*ppPropArray)[i].Value.ft = m_ftLastUpdate;
                }
                break;
            case PR_CONTENT_COUNT :
                hResult = MAPI_E_NOT_FOUND;
                if (m_pTableData)
                {
                    LPMAPITABLE pTable;
                    if (!m_pTableData->HrGetView (NULL, NULL, 0, &pTable))
                    {
                        if (!pTable->GetRowCount (0, &(*ppPropArray)[i].Value.ul))
                        {
                            hResult = S_OK;
                        }
                        pTable->Release();
                    }
                }
                break;
            case PR_ASSOC_CONTENT_COUNT :
                (*ppPropArray)[i].Value.l = 0;
                break;
            case PR_ACCESS :
                (*ppPropArray)[i].Value.l = MAPI_ACCESS_READ;
                break;
            case PR_ACCESS_LEVEL :
                (*ppPropArray)[i].Value.l = 0;
                break;
            case PR_FOLDER_TYPE :
                (*ppPropArray)[i].Value.l = FOLDER_GENERIC;
                break;
            case PR_OBJECT_TYPE :
                (*ppPropArray)[i].Value.l = MAPI_FOLDER;
                break;
            case PR_DISPLAY_TYPE :
                (*ppPropArray)[i].Value.l = DT_FOLDER;
                break;
            case PR_SUBFOLDERS :
                (*ppPropArray)[i].Value.b = FALSE;
                break;
            case PR_CREATION_VERSION :
                (*ppPropArray)[i].Value.li.LowPart = TRANSPORT_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = TRANSPORT_VERSION_MINOR;
                break;
            case PR_CURRENT_VERSION :
                (*ppPropArray)[i].Value.li.LowPart = TRANSPORT_VERSION_MAJOR;
                (*ppPropArray)[i].Value.li.HighPart = TRANSPORT_VERSION_MINOR;
                break;
            case PR_DISPLAY_NAME :
                szBuffer[0] = 0;
                LoadString (m_pLogon->GetInstance(), IDS_MSG_REMOTE_FOLDER_VIEW_NAME, szBuffer, 64);
                pSrc = (LPVOID)szBuffer;
                break;
		    default:
				hResult = MAPI_E_NOT_FOUND;
                #ifdef DO_INFO_TRACES
                TraceRaw ("CXPMAPIFolder::GetProps: Not Found - "); TraceProp ((*ppPropArray)[i].ulPropTag);
                #endif // #DO_INFO_TRACES
                break;
		}
        if (!hResult && 
            (PROP_TYPE((*ppPropArray)[i].ulPropTag) == PT_BINARY ||
             PROP_TYPE((*ppPropArray)[i].ulPropTag) == PT_TSTRING))
        {
            if (PROP_TYPE((*ppPropArray)[i].ulPropTag) == PT_TSTRING)
            {
                cbSize = Cbtszsize ((LPSTR)pSrc);
            }
            hResult = gpfnAllocateMore (cbSize, *ppPropArray, &pDst);
            if (!hResult)
            {
                CopyMemory (pDst, pSrc, cbSize);
                if (PROP_TYPE((*ppPropArray)[i].ulPropTag) == PT_BINARY)
                {
                    (*ppPropArray)[i].Value.bin.cb  = cbSize;
                    (*ppPropArray)[i].Value.bin.lpb = (LPBYTE)pDst;
                }
                else
                {
                    (*ppPropArray)[i].Value.lpszA = (LPSTR)pDst;
                }
            }
        }
        if (hResult)
        {
            (*ppPropArray)[i].ulPropTag = PROP_TAG(PT_ERROR, PROP_ID((*ppPropArray)[i].ulPropTag));
            (*ppPropArray)[i].Value.err = hResult;
            fGotErrors = TRUE;
        }
	}
	if (fGotErrors)
	{
		hResult = MAPI_W_ERRORS_RETURNED;
	}
    LeaveCriticalSection (&m_csObj);
	return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetPropList()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Returns an array with all the property tags that we support in
//      this folder
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::GetPropList (ULONG ulFlags, LPSPropTagArray * ppTags)
{
    InfoTrace ("CXPMAPIFolder::GetPropList method called");
    Validate_IMAPIProp_GetPropList (this, ulFlags, ppTags);

	ULONG cbTagsArraySize = CbNewSPropTagArray(NUM_REMOTE_FOLDER_PROPS);
	// Allocate the required amount of memory
	HRESULT hResult = gpfnAllocateBuffer (cbTagsArraySize, (LPVOID *)ppTags);
	TraceResult ("CXPMAPIFolder::GetPropList: Memory allocation failed", hResult);
	if (!hResult)
    {
	    // Copy the contents of our property tag array into the buffer
	    CopyMemory (*ppTags, &sptRemoteFolderProps, cbTagsArraySize);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::OpenProperty()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::OpenProperty (ULONG       ulPropTag,
                                          LPCIID      piid,
                                          ULONG       ulInterfaceOptions,
                                          ULONG       ulFlags,
                                          LPUNKNOWN * ppUnk)
{
    InfoTrace ("CXPMAPIFolder::OpenProperty method called");
    Validate_IMAPIProp_OpenProperty (this, ulPropTag,  piid, ulInterfaceOptions, ulFlags, ppUnk);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SetProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::SetProps (ULONG                 cValues,
                                      LPSPropValue          pPropArray,
                                      LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CXPMAPIFolder::SetProps method called");
    Validate_IMAPIProp_SetProps (this, cValues, pPropArray, ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::DeleteProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::DeleteProps (LPSPropTagArray       pPropTagArray,
                                         LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CXPMAPIFolder::DeleteProps method called");
    Validate_IMAPIProp_DeleteProps (this, pPropTagArray, ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CopyTo()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CopyTo (ULONG                 ciidExclude,
                                    LPCIID                rgiidExclude,
                                    LPSPropTagArray       pExcludeProps,
                                    ULONG                 ulUIParam,
                                    LPMAPIPROGRESS        pProgress,
                                    LPCIID                pInterface,
                                    LPVOID                pDestObj,
                                    ULONG                 ulFlags,
                                    LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CXPMAPIFolder::CopyTo method called");
    Validate_IMAPIProp_CopyTo (this,
                               ciidExclude,
                               rgiidExclude,
                               pExcludeProps,
                               ulUIParam,
                               pProgress,
                               pInterface,
                               pDestObj,
                               ulFlags,
                               ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CopyProps()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CopyProps (LPSPropTagArray       pIncludeProps,
                                       ULONG                 ulUIParam,
                                       LPMAPIPROGRESS        pProgress,
                                       LPCIID                pInterface,
                                       LPVOID                pDestObj,
                                       ULONG                 ulFlags,
                                       LPSPropProblemArray * ppProblems)
{
    InfoTrace ("CXPMAPIFolder::CopyProps method called");
    Validate_IMAPIProp_CopyProps (this,
                                  pIncludeProps,
                                  ulUIParam,
                                  pProgress,
                                  pInterface,
                                  pDestObj,
                                  ulFlags,
                                  ppProblems);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetNamesFromIDs()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::GetNamesFromIDs (LPSPropTagArray * ppPropTags,
                                             LPGUID            pPropSetGuid,
                                             ULONG             ulFlags,
                                             ULONG *           pcPropNames,
                                             LPMAPINAMEID **   pppPropNames)
{
    InfoTrace ("CXPMAPIFolder::GetNamesFromIDs method called");
    Validate_IMAPIProp_GetNamesFromIDs (this,
                                        ppPropTags,
                                        pPropSetGuid,
                                        ulFlags,
                                        pcPropNames,
                                        pppPropNames);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetIDsFromNames()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::GetIDsFromNames (ULONG             cPropNames,
                                             LPMAPINAMEID *    ppPropNames,
                                             ULONG             ulFlags,
                                             LPSPropTagArray * ppPropTags)
{
    InfoTrace ("CXPMAPIFolder::GetIDsFromNames method called");
    Validate_IMAPIProp_GetIDsFromNames (this,
                                        cPropNames,
                                        ppPropNames,
                                        ulFlags,
                                        ppPropTags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIContainer virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetContentsTable()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::GetContentsTable (ULONG ulFlags, LPMAPITABLE * ppTable)
{
    InfoTrace ("CXPMAPIFolder::GetContentsTable method called");
    Validate_IMAPIContainer_GetContentsTable (this, ulFlags, ppTable);

    if (MAPI_ASSOCIATED & ulFlags)
    {
        TraceMessage ("CXPMAPIFolder::GetContentsTable: We don't have associated messages");
        return MAPI_E_NO_SUPPORT;
    }
    HRESULT hResult;
    // If we don't have an ITableData interface pointer instantiated, create one.
    if (!m_pTableData)
    {
        EnterCriticalSection (&m_csObj);

#ifdef  WINDS_SINGLE_SERVICE
#define pfnMAPIAllocateBuffer   GetMAPIAllocBuffer()
#define pfnMAPIAllocateMore     GetMAPIAllocMore()
#define pfnMAPIFreeBuffer       GetMAPIFreeBuffer()
#else
#define pfnMAPIAllocateBuffer   gpfnAllocateBuffer
#define pfnMAPIAllocateMore     gpfnAllocateMore
#define pfnMAPIFreeBuffer       gpfnFreeBuffer
#endif // WINDS_SINGLE_SERVICE

        hResult = CreateTable (&IID_IMAPITableData,
                               pfnMAPIAllocateBuffer,
                               pfnMAPIAllocateMore,
                               pfnMAPIFreeBuffer,
                               NULL,
                               TBLTYPE_DYNAMIC,
                               PR_ENTRYID,
                               (LPSPropTagArray)&sptRemoteFldContents,
                               &m_pTableData);
        if (hResult)
        {
            TraceResult ("CXPMAPIFolder::GetContentsTable: Failed to create an ITableData object", hResult);
            return hResult;
        }

        // Populate the data for the table here
        hResult = FillContentsTable (m_pLogon->GetLocalHeadersCache());
        if (hResult)
        {
            TraceResult ("CXPMAPIFolder::GetContentsTable: Failed to populate table data", hResult);
            return hResult;
        }
        LeaveCriticalSection (&m_csObj);
    }
    hResult = m_pTableData->HrGetView (NULL, NULL, 0, ppTable);
    // If any, trace out the error code.
    TraceResult ("CXPMAPIFolder::GetContentsTable: Failed to create a view of the table data", hResult);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetHierarchyTable()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::GetHierarchyTable (ULONG         ulFlags,
                                               LPMAPITABLE * ppTable)
{
    InfoTrace ("CXPMAPIFolder::GetHierarchyTable method called");
    Validate_IMAPIContainer_GetHierarchyTable (this, ulFlags, ppTable);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::OpenEntry()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::OpenEntry (ULONG       cbEntryID,
                                       LPENTRYID   pEntryID,
                                       LPCIID      pInterface,
                                       ULONG       ulFlags,
                                       ULONG *     pulObjType,
                                       LPUNKNOWN * ppUnk)
{
    InfoTrace ("CXPMAPIFolder::OpenEntry method called");
    Validate_IMAPIContainer_OpenEntry (this,
                                       cbEntryID,
                                       pEntryID,
                                       pInterface,
                                       ulFlags,
                                       pulObjType,
                                       ppUnk);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SetSearchCriteria()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::SetSearchCriteria (LPSRestriction pRestriction,
                                               LPENTRYLIST    pContainerList,
                                               ULONG          ulSearchFlags)
{
    InfoTrace ("CXPMAPIFolder::SetSearchCriteria method called");
    Validate_IMAPIContainer_SetSearchCriteria (this, pRestriction, pContainerList, ulSearchFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetSearchCriteria()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::GetSearchCriteria (ULONG            ulFlags,
                                               LPSRestriction * ppRestriction,
                                               LPENTRYLIST *    ppContainerList,
                                               ULONG *          pulSearchState)
{
    InfoTrace ("CXPMAPIFolder::GetSearchCriteria method called");
    Validate_IMAPIContainer_GetSearchCriteria (this, ulFlags, ppRestriction, ppContainerList, pulSearchState);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// IMAPIFolder virtual member functions implementation
//

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CreateMessage()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CreateMessage (LPCIID      pInterface,
                                           ULONG       ulFlags,
                                           LPMESSAGE * ppMessage)
{
    InfoTrace ("CXPMAPIFolder::CreateMessage method called");
    Validate_IMAPIFolder_CreateMessage (this, pInterface, ulFlags, ppMessage);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CopyMessages()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CopyMessages (LPENTRYLIST    pMsgList,
                                          LPCIID         pInterface,
                                          LPVOID         pDestFolder,
                                          ULONG          ulUIParam,
                                          LPMAPIPROGRESS pProgress,
                                          ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::CopyMessages method called");
    Validate_IMAPIFolder_CopyMessages (this,
                                       pMsgList,
                                       pInterface,
                                       pDestFolder,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::DeleteMessages()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::DeleteMessages (LPENTRYLIST    pMsgList,
                                            ULONG          ulUIParam,
                                            LPMAPIPROGRESS pProgress,
                                            ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::DeleteMessages method called");
    Validate_IMAPIFolder_DeleteMessages (this,
                                         pMsgList,
                                         ulUIParam,
                                         pProgress,
                                         ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CreateFolder()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CreateFolder (ULONG          ulFolderType,
                                          LPTSTR         pszFolderName,
                                          LPTSTR         pszFolderComment,
                                          LPCIID         pInterface,
                                          ULONG          ulFlags,
                                          LPMAPIFOLDER * ppFolder)
{
    InfoTrace ("CXPMAPIFolder::CreateFolder method called");
    Validate_IMAPIFolder_CreateFolder (this,
                                       ulFolderType,
                                       pszFolderName,
                                       pszFolderComment,
                                       pInterface,
                                       ulFlags,
                                       ppFolder);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CopyFolder()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::CopyFolder (ULONG          cbEntryID,
                                        LPENTRYID      pEntryID,
                                        LPCIID         pInterface,
                                        LPVOID         pDestFolder,
                                        LPTSTR         pszNewFolderName,
                                        ULONG          ulUIParam,
                                        LPMAPIPROGRESS pProgress,
                                        ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::CopyFolder method called");
    Validate_IMAPIFolder_CopyFolder (this,
                                     cbEntryID,
                                     pEntryID,
                                     pInterface,
                                     pDestFolder,
                                     pszNewFolderName,
                                     ulUIParam,
                                     pProgress,
                                     ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::DeleteFolder()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::DeleteFolder (ULONG          cbEntryID,
                                          LPENTRYID      pEntryID,
                                          ULONG          ulUIParam,
                                          LPMAPIPROGRESS pProgress,
                                          ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::DeleteFolder method called");
    Validate_IMAPIFolder_DeleteFolder (this,
                                       cbEntryID,
                                       pEntryID,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SetReadFlags()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::SetReadFlags (LPENTRYLIST    pMsgList,
                                          ULONG          ulUIParam,
                                          LPMAPIPROGRESS pProgress,
                                          ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::SetReadFlags method called");
    Validate_IMAPIFolder_SetReadFlags (this,
                                       pMsgList,
                                       ulUIParam,
                                       pProgress,
                                       ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::GetMessageStatus()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::GetMessageStatus (ULONG     cbEntryID,
                                              LPENTRYID pEntryID,
                                              ULONG     ulFlags,
                                              ULONG *   pulMessageStatus)
{
    InfoTrace ("CXPMAPIFolder::GetMessageStatus method called");
    Validate_IMAPIFolder_GetMessageStatus (this,
                                           cbEntryID,
                                           pEntryID,
                                           ulFlags,
                                           pulMessageStatus);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SetMessageStatus()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::SetMessageStatus (ULONG     cbEntryID,
                                              LPENTRYID pEntryID,
                                              ULONG     ulNewStatus,
                                              ULONG     ulNewStatusMask,
                                              ULONG *   pulOldStatus)
{
    InfoTrace ("CXPMAPIFolder::SetMessageStatus method called");
    Validate_IMAPIFolder_SetMessageStatus (this,
                                           cbEntryID,
                                           pEntryID,
                                           ulNewStatus,
                                           ulNewStatusMask,
                                           pulOldStatus);
    EnterCriticalSection (&m_csObj);
    ASSERT (TRANSPORT_MESSAGE_EID_SIZE == cbEntryID);
    SPropValue spvSearchProp;
    spvSearchProp.ulPropTag     = PR_ENTRYID;
    spvSearchProp.Value.bin.cb  = cbEntryID;
    spvSearchProp.Value.bin.lpb = (LPBYTE)pEntryID;

    LPSRow psrTargetRow;
    HRESULT hResult = m_pTableData->HrQueryRow (&spvSearchProp, &psrTargetRow, NULL);
    if (hResult)
    {
        TraceResult ("CXPMAPIFolder::SetMessageStatus: Failed to query row from internal table", hResult);
    }
    else
    {
        BOOL bFound = FALSE;
        ULONG ulOldStatus;
        for (ULONG i=0; i<psrTargetRow->cValues; i++)
        {
            if (PR_MSG_STATUS == psrTargetRow->lpProps[i].ulPropTag)
            {
                bFound = TRUE;
                ulOldStatus = psrTargetRow->lpProps[i].Value.l;
                if (pulOldStatus)
                {
                    *pulOldStatus = ulOldStatus;
                }
                ulOldStatus &= ulNewStatusMask;
                ulNewStatus &= ulNewStatusMask;
                if (ulNewStatus != ulOldStatus)
                {
                    psrTargetRow->lpProps[i].Value.l &= ~ulNewStatusMask;
                    psrTargetRow->lpProps[i].Value.l |= ulNewStatus;
                    hResult= m_pTableData->HrModifyRow (psrTargetRow);
                    // If any, trace out the error code.
                    TraceResult ("CXPMAPIFolder::SetMessageStatus: Failed to modify row in the internal table", hResult);
                }
                break; // Out of the FOR() loop
            }
        }
        if (!hResult && !bFound)
        {
            TraceMessage ("CXPMAPIFolder::SetMessageStatus: Could not find property in table row");
            hResult = MAPI_E_NOT_FOUND;
        }
        gpfnFreeBuffer (psrTargetRow);
    }
    LeaveCriticalSection (&m_csObj);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::SaveContentsSort()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::SaveContentsSort (LPSSortOrderSet pSortCriteria,
                                              ULONG           ulFlags)
{
    InfoTrace ("CXPMAPIFolder::SaveContentsSort method called");
    Validate_IMAPIFolder_SaveContentsSort (this, pSortCriteria, ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::EmptyFolder()
//
//    Parameters
//      { Refer to MAPI Documentation on this method }
//
//    Purpose
//      Stub method.
//
//    Return Value
//      MAPI_E_NO_SUPPORT always.
//
STDMETHODIMP CXPMAPIFolder::EmptyFolder (ULONG          ulUIParam,
                                         LPMAPIPROGRESS pProgress,
                                         ULONG          ulFlags)
{
    InfoTrace ("CXPMAPIFolder::EmptyFolder method called");
    Validate_IMAPIFolder_EmptyFolder (this, ulUIParam, pProgress, ulFlags);
    return MAPI_E_NO_SUPPORT;
}

///////////////////////////////////////////////////////////////////////////////
// Other member functions specific to this class
//

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::FillContentsTable()
//
//    Parameters
//      pszHeaderFileName       String with a fully qualified path to the
//                              headers data file
//    Purpose
//      This function reads the file with the headers information to
///     populate the table of contents of the remote mail bag folder.
//      The data is store in a file whose path is given in the
//      pszHeaderFileName parameter.
//
//    Return Value
//      S_OK  If the function is successful or an hResult with the
//      error that occurr
//
STDMETHODIMP CXPMAPIFolder::FillContentsTable (LPTSTR pszHeaderFileName)
{
    ASSERT(pszHeaderFileName);
    if (NULL == m_pTableData)
    {
        // Nothing to do
        return S_OK;
    }

    HRESULT hResult = m_pTableData->HrDeleteRows (TAD_ALL_ROWS, NULL, NULL);
    TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to delete all rows in internal table", hResult);

    // Check that the file exists. If it doesn't exist, then there is no headers file.
    HANDLE hFile = CreateFile (pszHeaderFileName,
                               GENERIC_READ,
                               0,
                               NULL,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
    if (INVALID_HANDLE_VALUE == hFile)
    {
        // Set the internal flag that the user HAS NOT downloaded headers from the server
        m_fNoHeadersFile = TRUE;
        return S_OK;
    }
    else
    {
        // The time the headers file was last updated, is the last time the file was written to
        if (GetFileTime (hFile, NULL, NULL, &m_ftLastUpdate))
        {
            // Set the internal flag that the user HAS downloaded headers from the server
            m_fNoHeadersFile = FALSE;
        }
        TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to get info about the headers file", ::GetLastError()); // Call the Win32 API
    }
    if (!GetFileSize (hFile, NULL))
    {
        CloseHandle(hFile);
        return S_OK;
    }

    SRow srTblRow;
    // Allocate an array of properties for the colums in the contents
    // tables of the folder
    hResult = gpfnAllocateBuffer (NUM_HDR_PROPS*sizeof(SPropValue), (LPVOID *)&(srTblRow.lpProps));
    if (hResult)
    {
        TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to allocate memory for properties", hResult);
        CloseHandle (hFile);
        return hResult;
    }

    // Initialize the property tags in the row set
    srTblRow.cValues = NUM_HDR_PROPS;
    for (UINT i=0; i<NUM_HDR_PROPS; i++)    
    {
        srTblRow.lpProps[i].ulPropTag = sptRemoteFldContents.aulPropTag[i];
    }
    srTblRow.lpProps[HDR_OBJ_TYPE].Value.l = MAPI_MESSAGE;
    
    DWORD dwBytesRead, dwInstanceKey = 1;
    srTblRow.lpProps[HDR_INST_KEY].Value.bin.cb  = sizeof (DWORD); // This is a fixed size.
    srTblRow.lpProps[HDR_INST_KEY].Value.bin.lpb = (LPBYTE)&dwInstanceKey;
    ULONG ulHeaderSize, iNextField;
    LONG lFieldSize, lValue = 0;
    LPSTR pStr, pszMsgClass;
    char szDisplayTo[256], szSubject[256], szSender[256];
    BYTE abBuffer[IO_BUFFERSIZE];
    __try
    {
        do
        {
            ZeroMemory (szDisplayTo, 256);
            ZeroMemory (szSubject, 256);
            ZeroMemory (szSender, 256);
            if (ReadFile (hFile, &ulHeaderSize, sizeof(ULONG), &dwBytesRead, NULL))
            {
                if (ReadFile (hFile, abBuffer, ulHeaderSize, &dwBytesRead, NULL))
                {
                    if (dwBytesRead)
                    {
                        // Put the data into the property value members
                        srTblRow.lpProps[HDR_EID].Value.bin.cb  = TRANSPORT_MESSAGE_EID_SIZE; // This is a fixed size.
                        srTblRow.lpProps[HDR_EID].Value.bin.lpb = abBuffer;                   // It will only read the first 16 bytes
                        iNextField = TRANSPORT_MESSAGE_EID_SIZE;                              // Skip the entry ID

                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lFieldSize = atol (pStr);
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        CopyMemory (szSender, &abBuffer[iNextField], lFieldSize);
                        iNextField += lstrlen (szSender) + 1;   // +1 to skip the comma
                        srTblRow.lpProps[HDR_SNDR].Value.lpszA = szSender;
                        srTblRow.lpProps[HDR_REPR].Value.lpszA = szSender;

                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lFieldSize = atol (pStr);
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        CopyMemory (szDisplayTo, &abBuffer[iNextField], lFieldSize);
                        srTblRow.lpProps[HDR_DISP_TO].Value.lpszA = szDisplayTo;
                    
                        iNextField += lFieldSize + 1;       // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lFieldSize = atol (pStr);
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        CopyMemory (szSubject, &abBuffer[iNextField], lFieldSize);
                        srTblRow.lpProps[HDR_SUBJ].Value.lpszA = szSubject;
                        srTblRow.lpProps[HDR_NORM_SUBJ].Value.lpszA = szSubject;

                        iNextField += lFieldSize + 1;       // +1 to skip the comma
                        pszMsgClass = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        srTblRow.lpProps[HDR_CLASS].Value.lpszA = pszMsgClass;
                    
                        iNextField += lstrlen (pszMsgClass) + 1;       // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_FLAGS].Value.l = lValue;

                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_SIZE].Value.l = lValue;

                        // The PR_MESSAGE_DOWNLOAD_TIME property contains how many secods it will take the transport
                        // to download a message from a server. In our sample, we compute this simply assuming a 28800
                        // bits per second rate connection speed to the server (even thougn an RPC call over the land
                        // will be alot faster) but this is just for demonstration purposes. Transport providers that
                        // really compute this property will probably have code to detect the actual link speed and
                        // compute the download time against the size of the message.
                        srTblRow.lpProps[HDR_DOWNLOAD_TIME].Value.l = (lValue * 8) / 28800;
                        if (!srTblRow.lpProps[HDR_DOWNLOAD_TIME].Value.l)
                        {
                            srTblRow.lpProps[HDR_DOWNLOAD_TIME].Value.l = 1;
                        }
                    
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_PRIOR].Value.l = lValue;
                    
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_IMPORTANCE].Value.l = lValue;

                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_SENS].Value.l = lValue;

                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_TIME].Value.ft.dwLowDateTime = lValue;
                    
                        iNextField += lstrlen (pStr) + 1;   // +1 to skip the comma
                        pStr = strtok ((LPSTR)&abBuffer[iNextField], ",");
                        lValue = atol (pStr);
                        srTblRow.lpProps[HDR_TIME].Value.ft.dwHighDateTime = lValue;

                        srTblRow.lpProps[HDR_STAT].Value.l = 0;

                        srTblRow.lpProps[HDR_HASATTACH].Value.l = (BOOL)!!(srTblRow.lpProps[HDR_FLAGS].Value.l & MSGFLAG_HASATTACH);
                    
                        hResult = m_pTableData->HrModifyRow (&srTblRow);
                        if (hResult)
                        {
                            TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to add row to table data", hResult);
                        }
                        dwInstanceKey++;
                    }
                }
                else
                {
                    hResult = ::GetLastError(); // Call the Win32 API
                    TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to read header data", hResult);
                    hResult = HRESULT_FROM_WIN32 (hResult);
                }
            }
            else
            {
                hResult = ::GetLastError(); // Call the Win32 API
                TraceResult ("CXPMAPIFolder::FillContentsTable: Failed to read header size", hResult);
                hResult = HRESULT_FROM_WIN32 (hResult);
            }
        } while (dwBytesRead && !hResult);
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        TraceString1 ("CXPMAPIFolder::FillContentsTable: An exception occured. Exception code: %d", GetExceptionCode());
        hResult = MAPI_E_CORRUPT_DATA;
    }

    // Free the array and close the file
    gpfnFreeBuffer (srTblRow.lpProps);
    CloseHandle (hFile);
    // If the headers file is corrupt, delete it and have the user get updated information from the server
    if (MAPI_E_CORRUPT_DATA == hResult)
    {
        DeleteFile (pszHeaderFileName);
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    CXPMAPIFolder::CopyTableToFile()
//
//    Parameters
//      None.
//
//    Purpose
//      This function copies the data from the in-memory contents table into
//      the headers file. If the user selected any action on the messages in
//      the contents table, those changes will be reflected in the headers
//      file and restored the next time the remote viewer is brought up.
//
//    Return Value
//      An HRESULT
//
STDMETHODIMP CXPMAPIFolder::CopyTableToFile()
{
    if (!m_pTableData)
    {
        return S_OK;
    }
    HRESULT hResult;
    HANDLE hFile = CreateFile (m_pLogon->GetLocalHeadersCache(),
                               GENERIC_WRITE,
                               0,
                               NULL,
                               CREATE_ALWAYS,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
    if (INVALID_HANDLE_VALUE == hFile)
    {
        hResult = ::GetLastError();
        TraceResult ("CXPMAPIFolder::CopyTableToFile: Failed to open local headers file", hResult);
        return HRESULT_FROM_WIN32(hResult);
    }
    BOOL fDeleteFile = TRUE;
    LPMAPITABLE pTable = NULL;
    hResult = m_pTableData->HrGetView (NULL, NULL, 0, &pTable);
    if (hResult)
    {
        TraceResult ("CXPMAPIFolder::CopyTableToFile: Failed to get a view of the table data", hResult);
        goto ErrorExit;
    }
    
    hResult = pTable->SetColumns ((LPSPropTagArray)&sptRemoteFldContents, 0);
    if (hResult)
    {
        TraceResult ("CXPMAPIFolder::CopyTableToFile: Failed to set the columns", hResult);
        goto ErrorExit;
    }
    
    LPSPropValue pProps;
    BYTE abBuffer[IO_BUFFERSIZE];
    LPSRowSet pRows;
    LPVOID pVoid;
    DWORD i, cbSize, dwBytesWrittem;
    BOOL fSuccess;
    #pragma warning (disable : 4127)
    while (TRUE)
    {
    #pragma warning (default : 4127)
        #define QUERY_SIZE  10
        hResult = pTable->QueryRows (QUERY_SIZE, 0, &pRows);
        if (hResult)
        {
            TraceResult ("CXPMAPIFolder::CopyTableToFile: Failed to query the rows in the table", hResult);
            goto ErrorExit;
        }
        // Did we finish reading all the rows?
        if (!pRows->cRows)
        {
            FreeProws (pRows);
            break; // Out of the WHILE() loop
        }
        for (i=0; i<pRows->cRows; i++)
        {
            pProps = pRows->aRow[i].lpProps;
            if (pProps[HDR_EID].ulPropTag != PR_ENTRYID ||
                pProps[HDR_SNDR].ulPropTag != PR_SENDER_NAME ||
                pProps[HDR_DISP_TO].ulPropTag != PR_DISPLAY_TO ||
                pProps[HDR_SUBJ].ulPropTag != PR_SUBJECT ||
                pProps[HDR_CLASS].ulPropTag != PR_MESSAGE_CLASS ||
                pProps[HDR_FLAGS].ulPropTag != PR_MESSAGE_FLAGS ||
                pProps[HDR_SIZE].ulPropTag != PR_MESSAGE_SIZE ||
                pProps[HDR_PRIOR].ulPropTag != PR_PRIORITY ||
                pProps[HDR_IMPORTANCE].ulPropTag != PR_IMPORTANCE ||
                pProps[HDR_SENS].ulPropTag != PR_SENSITIVITY ||
                pProps[HDR_TIME].ulPropTag != PR_MESSAGE_DELIVERY_TIME ||
                pProps[HDR_STAT].ulPropTag != PR_MSG_STATUS)
            {
                TraceMessage ("CXPMAPIFolder::CopyTableToFile: Missing property in a row. Skipping it");
                continue; // The FOR() loop
            }
            
            ZeroMemory (abBuffer, IO_BUFFERSIZE);
        
            // Put the properties' data into the buffer.
            pVoid = (LPVOID)abBuffer;
            CopyMemory (pVoid, pProps[HDR_EID].Value.bin.lpb, TRANSPORT_MESSAGE_EID_SIZE);
            cbSize = (DWORD)pVoid + TRANSPORT_MESSAGE_EID_SIZE;
            pVoid = (LPVOID)cbSize;

            wsprintf ((LPSTR)pVoid,
                      "%d,%s,%d,%s,%d,%s,%s,%d,%d,%d,%d,%d,%d,%d,",
                      lstrlen (pProps[HDR_SNDR].Value.lpszA),
                      pProps[HDR_SNDR].Value.lpszA,
                      lstrlen (pProps[HDR_DISP_TO].Value.lpszA),
                      pProps[HDR_DISP_TO].Value.lpszA,
                      lstrlen (pProps[HDR_SUBJ].Value.lpszA),
                      pProps[HDR_SUBJ].Value.lpszA,
                      pProps[HDR_CLASS].Value.lpszA,
                      pProps[HDR_FLAGS].Value.l,
                      pProps[HDR_SIZE].Value.l,
                      pProps[HDR_PRIOR].Value.l,
                      pProps[HDR_IMPORTANCE].Value.l,
                      pProps[HDR_SENS].Value.l,
                      pProps[HDR_TIME].Value.ft.dwLowDateTime,
                      pProps[HDR_TIME].Value.ft.dwHighDateTime);

            cbSize = TRANSPORT_MESSAGE_EID_SIZE + lstrlen ((LPSTR)pVoid);
            fSuccess = WriteFile (hFile, &cbSize, sizeof(DWORD), &dwBytesWrittem, NULL);
            if (fSuccess)
            {
                fSuccess = WriteFile (hFile, abBuffer, cbSize, &dwBytesWrittem, NULL);
                if (fSuccess)
                {
                    fDeleteFile = FALSE;
                }
            }
            if (!fSuccess)
            {
                hResult = ::GetLastError();
                TraceResult ("CXPMAPIFolder::CopyTableToFile: Write header data to file", hResult);
                hResult = HRESULT_FROM_WIN32(hResult);
                break; // Ouf of the FOR() loop
            }
        }
        FreeProws (pRows);
        if (hResult)
        {
            goto ErrorExit;
        }
    }

ErrorExit:
    if (pTable)
    {
        pTable->Release();
    }
    CloseHandle (hFile);
    if (hResult && fDeleteFile)
    {
        DeleteFile (m_pLogon->GetLocalHeadersCache());
    }
    m_pTableData->Release();
    m_pTableData = NULL;
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    ParseTime()
//
//    Parameters                        
//      [IN]    pszTime     String in the form HH:MM.
//      [OUT]   pwHour      Hour of day, 0-23
//      [OUT]   pwMinute    Minute of hour, 0-59
//
//    Purpose
//      This function parses the string in pszTime and returns the numbers
//      for the hour and minute of day in a 24 hour clock format.
//      
//    Return Value
//      TRUE on success, FALSE if we failed or the string passed
//      in is invalid.
//      
BOOL WINAPI ParseTime (LPTSTR pszTime, WORD * pwHour, WORD * pwMinute)
{
    char szTemp[6];
    lstrcpy (szTemp, pszTime);
    LPSTR pszSubStr = strtok(szTemp, ":");
    if (!pszSubStr)
    {
        return FALSE;
    }
    *pwHour = (WORD)atoi (pszSubStr);
    if (*pwHour < 0 || *pwHour > 23)
    {
        return FALSE;
    }
    pszSubStr = strtok (NULL, " ");
    if (!pszSubStr)
    {
        return FALSE;
    }
    *pwMinute = (WORD)atoi(pszSubStr);
    if (*pwMinute < 0 || *pwMinute > 59)
    {
        return FALSE;
    }
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
//    GetTimeString()
//
//    Parameters
//      [OUT]   pszTime     Pointer to a buffer where the function copies the
//                          formatted time string. The buffer must have been
//                          allocated by the caller.
//      [IN]    FileTime    FILETIME structure with the time for the string.
//
//    Purpose
//      This function returns a string formatted as "HH:MM"
//      The formatted string is returned in a buffer supplied by the caller
//      of the function.
//      
//    Return Value
//      None.
//
void WINAPI GetTimeString (LPTSTR pszTime, FILETIME FileTime)
{
    // Convert the FILETIME structure to a more readable form
    SYSTEMTIME sysTime;
    FileTimeToSystemTime (&FileTime, &sysTime);
    wsprintf (pszTime, "%d:%02d", sysTime.wHour, sysTime.wMinute);
}

///////////////////////////////////////////////////////////////////////////////
//    GetMAPIError()
//
//    Parameters
//      ppMAPIError             Pointer to a location where the newly
//                              allocated MAPIERROR structure is returened. 
//      ulFlags                 Flags for the allocation of the structure
//      hError                  Handle to the error for which a string will
//                              be returned
//      hInstance               Handle to instance of DLL where the resource
//                              strings are located
//
//    Purpose
//      This provider implements a shared function for the two different
//      GetLastError methods in the IMAPIFolder and IMAPIStatus interfaces.
//      These interfaces do not maintain (cached) the last error result, so
//      we trust the client that the hError passed in is indeed the last error.
//      This simplifies implementation of GetLastError, but it will give up
//      giving error strings on the exact context of the actual last error
//      in the case a client passes the wrong hResult.
//      ISV's should weight the benefits and trade-off of this implementation.
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI GetMAPIError (LPMAPIERROR * ppMAPIError,
                             ULONG         ulFlags,
                             HRESULT       hError,
                             HINSTANCE     hInstance)
{
    // The following array maps a string identifier (IDS) to an hResult
    // The order of HRESULTs in the array has an external dependency:
    // the order of elements in the array is dictated by the IDS definitions
    // in RESOURCE.H. This implicit association must be maintained for the 
    // strings associated with string identifiers to make sense. Thus, if 
    // either this structure or the RESOURCE.H defines changes, the other 
    // must change to match it
    static const HRESULT MapOfhResults[] =
    {
    	S_OK,

    	E_FAIL,
    	E_OUTOFMEMORY,
    	E_INVALIDARG,
    	E_NOINTERFACE,
    	E_ACCESSDENIED,

        MAPI_E_BUSY,
    	MAPI_E_NO_SUPPORT,
    	MAPI_E_NOT_FOUND,
    	MAPI_E_UNKNOWN_FLAGS,
    	MAPI_E_VERSION,
    	MAPI_E_INVALID_ENTRYID,
        MAPI_E_USER_CANCEL,
        MAPI_E_DISK_ERROR,
        MAPI_E_CORRUPT_DATA,
        MAPI_E_BAD_CHARWIDTH,
        MAPI_E_UNCONFIGURED,
        

        CO_E_NOTINITIALIZED,    // This error could only be comming from the server

        MAPI_W_ERRORS_RETURNED
    };

    // Default the a return value in case we don't have the specified error
    // Allocate the memory for the MAPIERROR structure. The caller is
    // responsible for free this
    HRESULT hResult = gpfnAllocateBuffer (sizeof(MAPIERROR), (LPVOID *)ppMAPIError);
    if (hResult)
    {
        TraceResult ("GetMAPIError: Failed to allocate memory for the MAPIERROR structure", hResult);
        return hResult;
    }

    // Zero-out all the members in the structure
    ZeroMemory (*ppMAPIError, sizeof(MAPIERROR));
    (*ppMAPIError)->ulVersion = MAPI_ERROR_VERSION;
    // See if we have the string for the error. If so, make a copy 
    // and pass it back. The memory for the string will be linked to
    // the memory of the structure so that the caller can make a
    // single release on the structure
    
	// Search for the hResult in the HRESULT map
    UINT i, uMaxNum, uStringID;
    uStringID = 0;
    uMaxNum = sizeof (MapOfhResults) / sizeof (HRESULT);
    for (i=0; i<uMaxNum; i++ )
	{
		if (MapOfhResults[i] == hError)
		{
            break;
		}
	}

    if (i == uMaxNum)
    {
        (*ppMAPIError)->ulLowLevelError = LOWORD(hError);
        if (HRESULT_FACILITY(hError) == FACILITY_RPC)
        {
            uStringID = IDS_RPC_ERROR;
        }
        else
        {
            if (HRESULT_FACILITY(hError) == FACILITY_WIN32 ||
                HRESULT_FACILITY(hError) == FACILITY_NULL)
            {
                uStringID = IDS_SYSTEM_ERROR;
            }
            else
            {
                uStringID = IDS_E_UNKNOWN_ERROR;
            }
        }
    }
	else
    {
        uStringID = i + IDS_SUCCESS;
    }
    
    char szBuffer[512];
	int nCharsRead = LoadString (hInstance, uStringID, szBuffer, sizeof(szBuffer) -1);
    if (!nCharsRead)
    {
        TraceString1 ("GetMAPIError: Could not find error string for ID %d", uStringID);
        hResult = MAPI_E_NOT_FOUND;
    }
    else
    {
    	// Allocate memory for return variable and set it.
        // The memory for this string is linked to the memory in the pMemBlock parameter
    	hResult = gpfnAllocateMore (Cbtszsize (szBuffer),
                                    *ppMAPIError,
    	                            (LPVOID *)&(*ppMAPIError)->lpszError);
    	if (!hResult)
        {
            lstrcpy ((*ppMAPIError)->lpszError, szBuffer);
            hResult = gpfnAllocateMore (Cbtszsize (TRANSPORT_DISPLAY_NAME_STRING),
                                        *ppMAPIError,
    	                                (LPVOID *)&(*ppMAPIError)->lpszComponent);
            if (!hResult)
            {
                lstrcpy ((*ppMAPIError)->lpszComponent, TRANSPORT_DISPLAY_NAME_STRING);
            }
            else
            {
                TraceResult ("GetMAPIError: Failed to allocate the memory for the component string", hResult);
            }
        }
        else
    	{
            TraceResult ("GetMAPIError: Failed to allocate the memory for the error string", hResult);
    	}
    }

    if (hResult)
    {
        gpfnFreeBuffer (*ppMAPIError);
        *ppMAPIError = NULL;
    }
    return hResult;
}

// End of file for XPFOLDER.CPP
